local ev = require 'ev'
local ts = require 'timestamps'
local u = require 'unix'
local struct = require 'struct'
local rational = require 'rational'

local channel_descriptions = {
    BRG_AIN = 'Low Level Bridge',
    HLSS_AIN = 'High Level SS',
}

local supported_layers = {
    ELBRG = { lun = 'BRG_AIN',
              channel_count = 4,
              coefficient_set_count = 24,
              category = 'source/somat_edaq_lite_brg' },
    MSBRG = { lun = 'BRG_AIN',
              channel_count = 16,
              coefficient_set_count = 24,
              category = 'source/somat_edaq_brg' },
    ELHLS = { lun = 'HLSS_AIN',
              channel_count = 4,
              coefficient_set_count = 24,
              category = 'source/somat_edaq_lite_hls' },
    MSHLS = { lun = 'HLSS_AIN',
              channel_count = 16,
              coefficient_set_count = 24,
              category = 'source/somat_edaq_hls' },
}

local channel_proxy = require 'channel_proxy'
local expmod_proxy = require 'expmod_proxy'
local datapath = require 'edaq-datapath'

local all_off = "ADC_RUN=OFF FILTERING_RUN=OFF RUN=OFF SYNC_START=OFF"

-- Read in all the digital filter files
local filters = {}
for fn in io.popen('ls -a /usr/share/HSFilter/*.aff'):lines() do
    io.input(fn)
    local fn_parts = {}
    for token in string.gmatch(fn, "[^/]+") do
        table.insert(fn_parts, token)
    end
    local filter_name = fn_parts[#fn_parts]:lower()
    assert(filters[filter_name] == nil)
    filters[filter_name] = {}
    line = io.read('*line')
    local words = {}
    for token in string.gmatch(line, "[^%s]+") do
        table.insert(words, tonumber(token))
    end
    local taps = table.remove(words, 1)
    local delay = table.remove(words, 1)
    local gain = table.remove(words, 1)
    assert(taps == #words)
    filters[filter_name] = { taps = words, delay = delay, gain = gain }
end

local filter_tables = {}
for fn in io.popen('ls -a /usr/share/HSFilter/*.tbl'):lines() do
    io.input(fn)
    local table_name
    local adc_sample_rate = 100000
    if fn:find('hpwr') then adc_sample_rate = 98304 end
    for line in io.lines() do
        local words = {}
        for token in string.gmatch(line, "[^%s]+") do
            table.insert(words, token)
        end
        if #words > 0 and words[1]:sub(1, 1) ~= ';' then
            if not table_name then
                table_name = words[1]:lower()
                filter_tables[table_name] = filter_tables[table_name] or {}
            else
                local freq = tonumber(table.remove(words, 1))
                local def = table.remove(words, 1):lower() == 'y'
                local alias = table.remove(words, 1):lower() == 'y'
                local brk_freq = tonumber(table.remove(words, 1))
                assert(table_name)
                assert(filter_tables[table_name])
                filter_tables[table_name][freq] =
                    filter_tables[table_name][freq] or {}
                if filter_tables[table_name][freq][brk_freq] == nil then
                    local entry = { def = def,
                                    alias = alias,
                                    adc_sample_rate = adc_sample_rate }
                    repeat
                        filter = table.remove(words, 1):lower()
                        decimation = table.remove(words, 1)
                        assert(filters[filter])
                        table.insert(entry, { filter = filter,
                                              decimation = decimation })
                    until #words == 0
                    filter_tables[table_name][freq][brk_freq] = entry
                end
            end
        end
    end
end


local analog_channel = class('analog_channel', channel_proxy)
function analog_channel:initialize(proxy, lun, ch_num)
    self.proxy = proxy
    self.proxy_type = 'analog'
    self.lun = lun
    self.ch_num = ch_num
    self.name = string.format('ch%d', ch_num)
    self.lun_channel = string.format('LUN=%s CHANNEL=%d', lun, ch_num - 1)
    self.channel_set = 'SET ' .. self.lun_channel

    self.comet_sinks = { }
    self.num_comet_sinks = 0
    self.comet_value = nil
    self.comet_timestamp = 0
    self.last_comet_timestamp = 0

    self.num_channel_sinks = 0

    self.channel_setup = {}

    self.comet_configured = false
    self.proxy:cmd(self.channel_set, all_off)
    self.proxy:cmd(self.channel_set, "ATTENUATOR1=1/1")
    self.run_port = nil

    channel_proxy.initialize(self, self.name .. '.' .. self.proxy.name)
end

function analog_channel:available_channels(hostname)
    return {{ name = hostname:gsub('%.', '_'),
            connector = '@' .. hostname,
            description = channel_descriptions[self.lun],
            electrical_units = 'Volts',
            data_type = 'sequential_int16',
            sample_rate = 100,
            board_serial_number = self.proxy.sn,
            category = self.proxy.layer_info.category,
    }}
end

function analog_channel:allocate_run_port(socket)
    local old_port = self.run_port
    if not self.run_port  then
        self.run_port = self.ch_num
    else
        local instance = math.floor(self.run_port / 16)
        if instance < 1000 then instance = instance + 1 else instance = 0 end
        self.run_port = self.ch_num + (instance * 16)
    end
    if old_port then self.proxy.dp:remap(old_port, self.run_port) end
    self.proxy.dp:add(self.run_port, 'INT16', socket)
end

function analog_channel:switch_run_port()
    self.proxy:cmd(self.channel_set, string.format("RUN_PORT=%d",self.run_port))
end

function analog_channel:comet_timer(loop)
    if self.num_comet_sinks > 0 then
        local last_sample = self.proxy.dp:get_last_sample(self.run_port)
        if self.channel_setup.filtered_scalar and
           self.channel_setup.offset_voltage and
           last_sample then
            self.comet_value = (last_sample *
                                self.channel_setup.filtered_scalar) +
                                self.channel_setup.offset_voltage
        else
            self.comet_value = '-'
        end
        for _, sink in pairs(self.comet_sinks) do
            sink(self.comet_timestamp, self.comet_value)
        end
        self.comet_timestamp = self.comet_timestamp + 1
    end
end

function analog_channel:stop()
    if self.num_channel_sinks == 0 then
        self.proxy:cmd(self.channel_set, all_off)
        self.channel_setup = {}
        if self.num_comet_sinks > 0 then
            self:start({})
        end
    end
end

function analog_channel:start(spec)
    spec.sample_rate = spec.sample_rate or 1000
    local rate, rate_num, rate_denom =
            rational:new(spec.sample_rate):tonumbers()

    spec.digital_filter_type = spec.digital_filter_type or 'none'
    if rate_denom ~= 1 then
        spec.error_message = 'Sample rate not supported'
        self:switch_run_port()
        return true
    end
    if filter_tables[spec.digital_filter_type] == nil then
        spec.error_message = 'Filter type not supported'
        self:switch_run_port()
        return true
    end
    if filter_tables[spec.digital_filter_type][rate_num] == nil then
        spec.error_message = 'Sample rate not supported'
        self:switch_run_port()
        return true
    end
    local filter_set = filter_tables[spec.digital_filter_type][rate_num]
    local def_break_frequency 
    for break_freq,filter_taps in pairs(filter_set) do
        if filter_taps.def then
            def_break_frequency = break_freq
            break
        end
    end
    assert(def_break_frequency)
    spec.digital_filter_attenuation_frequency =
        spec.digital_filter_attenuation_frequency or def_break_frequency
    spec.digital_filter_attenuation_frequency =
        tonumber(spec.digital_filter_attenuation_frequency)
    if filter_set[spec.digital_filter_attenuation_frequency] == nil then
        spec.error_message = 'Break frequency not supported'
        self:switch_run_port()
        return true
    end
    local filter = filter_set[spec.digital_filter_attenuation_frequency]

    local start_40m = 0
    if spec.start_time then
        start_40m = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
        start_40m = math.floor(start_40m + 0.5)
    end

    spec._timescale = rate_denom..'/'..rate_num

    self.proxy:cmd(self.channel_set, all_off)
    self.channel_setup = {}

    local adc_sample_rate = filter.adc_sample_rate
    self.proxy:cmd(self.channel_set, 'ADC_SAMPLE_RATE=' .. adc_sample_rate)

    self.proxy:cmd('FILTER_CLEAR', self.lun_channel, 'FILTER_INDEX=ALL')
    for filter_index,entry in ipairs(filter) do
        local set_index = self.proxy:load_coefficient_set(entry.filter,
                                                          self.lun,
                                                          self.channel)
        if set_index == 0 then
            self.proxy:clear_coefficient_sets(self.channel)
            spec.error_message = 'Insufficient coefficient sets'
            self:switch_run_port()
            return true
        end
        self.proxy:cmd('FILTER_SET', self.lun_channel, string.format(
            'FILTER_INDEX=%d DECIMATION_FACTOR=%d COEFFICIENT_SET=%d',
            filter_index - 1, entry.decimation, set_index - 1))
    end

    if not spec.sensor_input then spec.sensor_input = {} end
    if spec.sensor_input.mode then
        if spec.sensor_input.mode == 'bridge' then
            local brg_type = string.format('BRIDGE_TYPE=%s',
                                           spec.sensor_input.type:upper())
            self.proxy:cmd(self.channel_set, brg_type)
        end
    end

    if self.lun == 'BRG_AIN' then
        local exc_level = spec.excitation_range or 'OFF'
        local exc_set = 'EXCITATION_LEVEL=' .. exc_level:upper()
        self.proxy:cmd(self.channel_set, exc_set)
    end

    if self.lun == 'HLSS_AIN' then
        if spec.sensor_input.mode and spec.sensor_input.mode == 'iepe' then
            self.proxy:cmd(self.channel_set, 'EXCITATION_VOLTAGE=24')
        else
            local power = spec.transducer_power or 0
            local power_set = string.format('EXCITATION_VOLTAGE=%d',
                                            tonumber(power))
            self.proxy:cmd(self.channel_set, power_set)
        end
    end

    local fs_max_requested = tonumber(spec._input_max) or
        spec.sensor_input.max or 10.0
    local fs_min_requested = tonumber(spec._input_min) or
        spec.sensor_input.min or -10.0

    local msg = self.proxy:cmd_yield('CENTER', self.lun_channel,
                string.format('REQUESTED_FS_UPPER_LIMIT=%g', fs_max_requested),
                string.format('REQUESTED_FS_LOWER_LIMIT=%g', fs_min_requested))
    if self.proxy:msg_validate(msg, 'CANNOT_CENTER') then
        self.proxy:cmd('CENTER', self.lun_channel, 
                       'REQUESTED_FS_UPPER_LIMIT=10',
                       'REQUESTED_FS_LOWER_LIMIT=-10')
    end

    self.proxy:cmd(self.channel_set, 'ADC_RUN=ON')
    self.proxy:cmd('REZERO', self.lun_channel)
    self.proxy:cmd(self.channel_set, 'ADC_RUN=OFF')
    msg = self.proxy:cmd_yield('CHANNEL_GET_SETUP', self.lun_channel)
    self.proxy:msg_validate(msg)
    for _,resp in ipairs(msg) do
        for k,v in pairs(resp) do
            if k ~= 'CHANNEL' and k ~= 'LUN' and k ~= 'ERRNO' then
                spec[k:lower()] = v
                self.channel_setup[k:lower()] = v
            end
        end
    end

    if start_40m == 0 then
        -- No start time specified, unsynchronized start
        self:switch_run_port()
        self.proxy:cmd(self.channel_set, "FILTERING_RUN=ON ADC_RUN=ON RUN=ON")
        return
    end

    local start = self.proxy.mmio:compute_layer_start_time(self.proxy.devname,
                                                           start_40m)

    self:switch_run_port()
    self.proxy:cmd(self.channel_set, "SYNC_START=ON START_TIME=" .. start)
    self.proxy:cmd(self.channel_set, "FILTERING_RUN=ON RUN=ON")
    msg = self.proxy:cmd_yield(self.channel_set, "ADC_RUN=ON")
    if self.proxy:msg_validate(msg, 'ARMING_FAILURE') then
        spec.error_message = 'Failed to start at requested time'
        self.proxy:cmd(self.channel_set, all_off)
        self.channel_setup = {}
        return
    end
end

function analog_channel:start_comet_stream(w, channels, sink)
    for _, ch in ipairs(channels) do
        local name = ch:match('@([^.]+)')
        self.comet_sinks = self.comet_sinks or { }
        self.comet_sinks[w.socket] = function(timestamp, value)
            sink(timestamp, ch, value)
        end
        self.num_comet_sinks = self.num_comet_sinks + 1
        if self.num_comet_sinks == 1 then
            self.proxy:launch_coroutine(
                function ()
                    self:allocate_run_port()
                    self:start({})
                end)
        end
    end
    w.cleanup = function()
        for _, ch in ipairs(channels) do
            self.comet_sinks[w.socket] = nil
            self.num_comet_sinks = self.num_comet_sinks - 1
            if self.num_comet_sinks == 0 then
                self.proxy:launch_coroutine(
                    function ()
                        self:stop()
                        self.proxy.dp:drop(self.run_port)
                    end)
            end
        end
    end
end

function analog_channel:start_channel_stream(w, spec, md_sink, sink)
    if self.num_channel_sinks > 0 then
        spec.error_message = 'Multiple tests not allowed'
        md_sink(spec)
        w.cleanup = function() end
        return
    end
    spec.socket = w.socket
    self.num_channel_sinks = self.num_channel_sinks + 1
    self.proxy:launch_coroutine(
        function ()
            self:allocate_run_port(spec.socket)
            self:start(spec)
            md_sink(spec)
        end)
    w.cleanup = function()
        self.num_channel_sinks = self.num_channel_sinks - 1
        self.proxy:launch_coroutine(
            function ()
                self:stop()
                self.proxy.dp:drop(self.run_port, w.socket)
            end)
    end
end

              -- -- -- -- -- -- -- -- -- -- -- --

local analog_proxy = class('analog_proxy', expmod_proxy)
function analog_proxy:initialize(name, devname)
    self.pending_tasks = {}
    self.debug_expmod_com = false   -- true to debug expmod_proxy activity
    expmod_proxy.initialize(self, name, devname)
    self.debug_expmod_com = false   -- true to debug analog_proxy activity

    self.dp = datapath.datapipe(self.data_dev, 1/20, self.sn)  -- 20 Hz rate
    self.dp:start(ev.Loop.default)

    self.pending_tasks = {}

    self.analog_channels = {}
    self.coefficient_sets = {}
    self.num_coefficient_sets_used = 0
    self.proxy_type = 'analog'

    self:cmd("SET SHUTDOWN=OFF DMA_RUN=YES")

    self.layer_info = supported_layers[self.sn:sub(1,5)]
    assert(self.layer_info, 'Invalid Serial Number')

    -- Generate a channel_proxy for each connector
    for ch_num = 1,self.layer_info.channel_count do
        self.analog_channels[ch_num] =
            analog_channel:new(self, self.layer_info.lun, ch_num)
    end
    for i= 1,self.layer_info.coefficient_set_count do
        self.coefficient_sets[i] = { set_index = i - 1, channels = {} }
    end

    local function comet_timer(loop, w, revents)
        for _,ch in ipairs(self.analog_channels) do
            ch:comet_timer(loop)
        end
    end
    self.comet_timer = ev.Timer.new(comet_timer, 0.1, 0.1)
    self.comet_timer:start(ev.Loop.default)

    channel_proxy.initialize(self, name)
end

function analog_proxy:load_coefficient_set(cs_name, lun, channel)
    for i = 1,self.layer_info.coefficient_set_count do
        if self.coefficient_sets[i].set_name then
            if self.coefficient_sets[i].set_name == cs_name then
                table.insert(self.coefficient_sets[i].channels, channel)
                return i
            end
        end
    end
    -- Not found, need to add a new one
    if self.num_coefficient_sets_used ==
       self.layer_info.coefficient_set_count then
        return 0   -- Invalid index, indicates no room available for a new set
    end
    self.num_coefficient_sets_used = self.num_coefficient_sets_used + 1
    local set_index = 1
    while self.coefficient_sets[set_index].set_name do
        set_index = set_index + 1
    end
    assert(set_index <= self.layer_info.coefficient_set_count)
    self.coefficient_sets[set_index].set_name = cs_name
    table.insert(self.coefficient_sets[set_index].channels, channel)

    -- Now load the set onto the board
    local cmdprefix_format = "FILTER_SET_TAP LUN=%s COEFFICIENT_SET=%d"
    local cmdprefix = string.format(cmdprefix_format, lun, set_index - 1)
    for i=1,#filters[cs_name].taps do
        local c = math.floor(filters[cs_name].taps[i] * 32767 + 0.5)
        local settap = string.format('TAP_INDEX=%d TAP_COEFFICIENT=%d', i-1, c)
        self:cmd(cmdprefix, settap)
    end
    local cmdprefix_format = "FILTER_SET_PARAM LUN=%s COEFFICIENT_SET=%d"
    local cmdprefix = string.format(cmdprefix_format, lun, set_index - 1)
    local prm_format = "FILTER_LENGTH=%d FILTER_GAIN=%f FILTER_DELAY=%d"
    local prm = string.format(prm_format, #filters[cs_name].taps,
                              1.0 - filters[cs_name].gain,
                              filters[cs_name].delay)
    self:cmd(cmdprefix, prm)
    return set_index
end

function analog_proxy:clear_coefficient_sets(channel)
    for set_index = 1,self.layer_info.coefficient_set_count do
        if self.coefficient_sets[set_index].set_name then
            for i,ch in ipairs(self.coefficient_sets[set_index].channels) do
                if ch == channel then
                    table.remove(self.coefficient_sets[set_index].channels, i)
                    break
                end
            end
            if #self.coefficient_sets[set_index].channels == 0 then
                self.coefficient_sets[set_index].set_name = nil
                self.num_coefficient_sets_used =
                    self.num_coefficient_sets_used - 1
            end
        end
    end
    assert(self.num_coefficient_sets_used >= 0)
end

function analog_proxy:available_channels(hostname)
    return {}
end

function analog_proxy:launch_coroutine(func)
    if func then
        table.insert(self.pending_tasks, coroutine.create(func))
    end
    while not self.active_task or coroutine.status(self.active_task) == 'dead'
    do
        self.active_task = table.remove(self.pending_tasks, 1)
        if not self.active_task then break end
        coroutine.resume(self.active_task)
    end
end

function analog_proxy:start_comet_stream(w, channels, sink)
end

function analog_proxy:start_channel_stream(w, spec, md_sink, sink)
end

return analog_proxy
