--! \file

require "strict"

require('lj_include'){
    'stdio',
    'string',
    'unistd',
    'sys/types',
    'signal',
    'linux/time',
    'sys/time',
    'sys/resource',
}

local __ = require "__"
local log = require "log"
local class = require "class"
local bit = require "bit"
local bits = require "bits"
local cjson = require "cjson.safe"
local crypto = require "crypto"
local dbus = require "dbus"
local ev = require "ev"
local ffi = require "ffi"
local http_parser = require "http_parser"
local lpeg = require "lpeg"
local membuf = require "membuf"
local scgi_server = require "scgi_server"
local u = require "unix"
local setup_spec = require "setup_spec"
local sie_parser = require "sie_parser"
local sqlite3 = require "lsqlite3"
local sqlite = require "sqlite_ffi"
local timestamps = require "timestamps"
local us = require "underscore"
local uuid = require "uuid"
local network_db = require "network_db"
local lfs = require "lfs"
local setup_migrator = require 'setupdb_migrate'
local system_config = require "system_config"
local fs_info = require 'fs_info'

local storage_path = system_config.storage.path or '/storage'

local sql_exec

local C = ffi.C

local bit_band = bit.band
local bit_lshift = bit.lshift
local bit_rshift = bit.rshift
local ffi_string = ffi.string

local ev_loop = ev.Loop.default

local ledlib = require('led_lib'):new(ev_loop)
assert(ledlib)

local setup_db_file = '/var/lib/armada/setup.db'

local local_authorization_key = "locally_authorized"
local start_is_locked = true

-- Fail safe to unlock test start API if test restarter fails to do so for
-- some reason.  NOTE: This is a one shot timer
ev.Timer.new(function(l, t, e)
        start_is_locked = nil
    end, 5 * 60):start(ev_loop)


-- Wrapper for executing OS shell commands
local function shellf(fmt, ...)
    local cmd = string.format(fmt, ...)
    local ok, term, status = os.execute(cmd)
    if term == nil then
        -- 5.1 semantics
        ok = bit.band(ok, 0xFF) == 0
    end
    return ok
end

-- Function used to generate a hostname based on target type and MAC address of
-- the specified interface.
local function make_default_hostname(iface)
    local f = io.popen('ip link show dev ' .. (iface or 'eth0') ..
        ' | grep ether')
    if f then
        for line in f:lines() do
            local mac = line:match('link/ether (.-) ')
            if mac then
                f:close()
                local v1, v2, v3 = mac:match('^.-:.-:.-:(.-):(.-):(.-)$')
                local unit = ''
                if system_config.target == 'armada' then
                    unit = 'cx23r-'
                else
                    unit = 'ex23r-'
                end
                local hname = string.format('%s%s%s%s', unit, v1, v2, v3)
                return hname
            end
        end
        f:close()
    end
end
local hostname = make_default_hostname()

local function dbg(fmt, ...)
    log:log(log.DEBUG, string.format(fmt, ...))
end

local uri_prefixes = {'/comet/', '/api/'}
local uri_unprefix_matches = {}
for _, uri_prefix in ipairs(uri_prefixes) do
    table.insert(uri_unprefix_matches, '^' .. uri_prefix .. '(.*)$')
end

local handlers = {}
local always_authorized_handlers = {
    sign_in = true,
    sign_out = true,
}
local requests = {}
local always_authorized_requests = {}

local function epoch_timestamp()
    local sec, nsec = u.clock_gettime()
    return sec + 1e-9 * nsec
end

local function iso_timestamp()
    return timestamps.epoch_to_iso(epoch_timestamp())
end

local function child_watcher(server, loop, child, revents)
    local status = child:getstatus()
    local pid = child:getrpid()
    if pid and pid ~= 0 then
        local kid = server.kids[pid]
        if kid.fn then
            kid.fn(server, loop, child, revents)
            server.kids[pid] = nil
        end
    end
end

--! A Comet server.
--? class api_server scgi_server
local m = class(scgi_server)
--? alias m api_server

function m:init(settings)
    m.super().init(self, settings)

    for k, v in pairs(handlers) do
        for _, uri_prefix in ipairs(uri_prefixes) do
            requests[uri_prefix .. k] = v
        end
    end
    for k, v in pairs(always_authorized_handlers) do
        for _, uri_prefix in ipairs(uri_prefixes) do
            always_authorized_requests[uri_prefix .. k] = v
        end
    end

    self.kids = {} -- indexed by child pid
    ev.Child.new(function(l, c, r) child_watcher(self, l, c, r) end,
        0, false):start(ev_loop)

    ev.Signal.new(function() -- Ignore SIGPIPE -- KLUDGE move to server.lua
    end, u.SIGPIPE):start(ev_loop)

    local function die()
        self.ev_loop:unloop()
    end

    for _, sig in ipairs{u.SIGINT, u.SIGTERM} do
        ev.Signal.new(die, sig):start(ev_loop)
    end

    self.comet_clients = {}
    self.accumulators = {}
    self.fake_comet_setup = {}
    self.running_tests_kludge = {}

    local bus = dbus.bus_get('system')
    if bus then
        self.bus = bus
        bus:hook_libev_default_loop()
        self.master = assert(bus:auto_proxy_wait('com.somat.armada.Master',
                                                '/com/somat/armada/Master'))
        self.network = assert(bus:auto_proxy_wait('com.somat.armada.Network',
                                                 '/com/somat/armada/Network'))
        self.network.ChangedRRs:bind(
            function(removed, added)
                self:changed_rr(removed, added)
            end
        )
    end

    self.setup_db = sqlite:new(setup_db_file)

    self:clean_sie()
    self:check_sie_files()

    self.throttled_live_data = {}
    self.comet_proxies_by_name = {}
    self.comet_proxies_by_socket = {}
end

-- Get the list of SIE files.  This is basically a mapping between the SIE file
-- name as stored on the storage pack and the "friendly" name assigned by the
-- user.
function m:get_sie_file_list()
    local db = self.setup_db

    local result = {}
    local r, ok = sql_exec(db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                local c = cjson.decode(v.content)
                if c then
                    local stat =
                        lfs.attributes(storage_path .. '/' .. v.id .. '.sie')
                    if stat then
                        result[#result + 1] = {
                            name = c.friendly_name,
                            size = stat.size,
                            access_time =
                                os.date('!%FT%H:%M:%S', stat.access),
                            change_time =
                                os.date('!%FT%H:%M:%S', stat.change),
                            modification_time =
                                os.date('!%FT%H:%M:%S', stat.modification),
                        }
                    end
                end
            end
        end
    end
    return result
end

-- This function returns the path to the actual SIE file associated with the
-- "friendly" name given.
function m:get_sie_path_on_disk(name)
    local db = self.setup_db

    local result = {}
    local r, ok = sql_exec(db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                local c = cjson.decode(v.content)
                if c then
                    if c.friendly_name and c.friendly_name == name then
                        return storage_path .. '/' .. v.id .. '.sie'
                    end
                end
            end
        end
    end
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk.
function m:delete_sie_friendly_name(name)
    local path_on_disk = self:get_sie_path_on_disk(name)
    if not path_on_disk then return end

    local txt_file = path_on_disk:gsub('%.sie', '%.txt')

    local id = path_on_disk:match(storage_path .. '/(.+)%.sie')

    local db = self.setup_db
    local result = {}
    local request = string.format(
        [[DELETE FROM "%%s" WHERE db_name == 'sie_name' AND id == '%s';]], id)
    local r, ok = sql_exec(db, 'tsd_users', request)
    C.unlink(txt_file)
    return path_on_disk
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk AND removes the SIE from the disk.
function m:delete_sie_file(name)
    local path_on_disk = self:delete_sie_friendly_name(name)
    if path_on_disk then
        C.unlink(path_on_disk)
        return true
    end
end

-- This function associates a "friendly" name with an SIE file on the disk.
-- If the friendly name already exists, a new one will be generated.  The name
-- used will be returned.
function m:assign_friendly_sie_name(requested_name, path_on_disk)
    local name = requested_name
    for idx = 1, 999 do
        local n = self:get_sie_path_on_disk(name)
        if not n then
            local info = {
                friendly_name = name,
            }
            local id = path_on_disk:match(storage_path .. '/(.+)%.sie')
            local req = string.format(
                [[INSERT INTO "%%s" (db_name, id, content, added)
                    VALUES ('sie_name', '%s', '%s', '%s')]],
                    id, cjson.encode(info), os.date('!%FT%H:%M:%S', os.time()))
            sql_exec(self.setup_db, 'tsd_users', req)
            local f = io.open(path_on_disk:gsub('%.sie$', '%.txt'), 'w')
            if f then
                f:write(name)
                f:close()
            end
            return name
        end
        name = string.format("%s-%03d", requested_name, idx)
    end
end

-- This function changes the "friendly" name of an SIE file and returns the new
-- friendly name.
function m:rename_sie(requested_name, current_name)
    local pod = self:delete_sie_friendly_name(current_name)
    if pod then
        local new_pod = pod:match('^(/.+/)') .. uuid.generate() .. '.sie'
        if new_pod then
            local assigned_name =
                self:assign_friendly_sie_name(requested_name, new_pod)
            if assigned_name then
                os.rename(pod, new_pod)
                return assigned_name
            end
        end
        -- Any failure restores the old mapping
        self:assign_friendly_sie_name(current_name, pod)
    end
end

-- This function is used to clean out friendly name to actual file associations
-- from the database for friendly names whos actual file no longer exists.
function m:clean_sie()
    local db = self.setup_db

    local result = {}
    local r, ok = sql_exec(db, 'tsd_users',
        [[SELECT content, id FROM "%s"
        WHERE db_name == 'sie_name' AND deleted == '';]])
    if ok then
        for _, v in ipairs(r) do
            if v.content and v.id then
                if not lfs.attributes(
                    storage_path .. '/' .. v.id .. '.sie') then
                    local req = string.format(
                        [[DELETE FROM "%%s" WHERE db_name == 'sie_name'
                        AND id == '%s';]], v.id)
                    sql_exec(db, 'tsd_users', req)
                end
            end
        end
    end
    return result
end

-- This function is used to get the "friendly" name associated with a
-- particular SIE file.
function m:get_sie_mapping(hash_name)
    local db = self.setup_db

    hash_name = string.gsub(hash_name, '%.sie', '')
    local result = {}
    local req = string.format(
        [[SELECT content FROM "%%s"
        WHERE db_name == 'sie_name' AND deleted == '' AND id == '%s';]],
        hash_name)
    local r, ok = sql_exec(db, 'tsd_users', req)
    if ok then
        for _, v in ipairs(r) do
            if v.content then
                local c = cjson.decode(v.content)
                if c then
                    if c.friendly_name then return c.friendly_name end
                end
            end
        end
    end
end

-- Function used to check the SIE hash name to friendly name mappings
function m:check_sie_files()
    -- Process all of the SIE files in the storage path
    for fh in lfs.dir(storage_path) do
        -- Is this an SIE file ?
        if fh:match('%.sie$') then
            -- Yes, get the existing hash to friendly mapping from the database
            local fn = self:get_sie_mapping(fh)

            -- Generate the path to the backup mapping file
            local bu = storage_path .. '/' ..  fh:gsub('%.sie', '%.txt')

            -- Is there an entry in the database for this SIE file?
            if not fn then
                -- Nope, is there a backup name mapping file for it?
                local fi = io.open(bu, 'r')
                if fi then
                    -- Yes, get the mapping
                    fn = fi:read('*a')
                    fi:close()
                else
                    -- No backup mapping of friendly name to hash name, attempt
                    -- to use the test name in the SIE file
                    -- KLUDGE:  This should probably be done by actuall reading
                    -- and parsing the XML for the information.
                    fn = fh:match('(.+)%.sie$')
                    local f = io.popen('grep -m 1 \'"test_name": "\' ' ..
                        storage_path .. '/' .. fh)
                    if f then
                        for l in f:lines() do
                            fn = l:match('"test_name": "(.+)",')
                            if fn then
                                break
                            end
                        end
                        f:close()
                    end
                end

                -- Add the mapping to the database
                fn = self:assign_friendly_sie_name(
                    fn, storage_path .. '/' .. fh)
            end
            -- Make sure backup mapping is set properly.
            local f = io.open(bu, 'w')
            if f then
                f:write(fn)
                f:close()
            end
        end
    end

    -- Clean out all friendly name mapping files for which the associated SIE
    -- no longer exists
    for fh in lfs.dir(storage_path) do
        -- Is this an friendly name mapping file?
        -- FIXME:  The pattern is not selective enough to prevent removal of
        -- all TXT files in the directory.
        if fh:match('%.txt$') then
            local sie_path = storage_path .. '/' ..  fh:gsub('%.txt', '%.sie')
            local stat = lfs.attributes(sie_path)
            if not stat then
                -- No associated SIE file, remove the TXT file
                C.unlink(storage_path .. '/' .. fh)
            end
        end
    end
end

local function read_file(fname)
    local fh = io.open(fname)
    if not fh then
        return nil
    end
    local content = fh:read('*a')
    fh:close()
    return content
end

local send_response -- forward decl

local function coro_post_mortem(server, loop, child, revents)
    local pid = child:getrpid()
    local status = child:getstatus()
    local kid = server.kids[pid]
    local r = {
        ok = status.exit_status == 0,
        pid = pid,
        exited = status.exited,
        exit_status = status.exit_status,
        stopped = status.stopped,
        signaled = status.signaled,
        term_signal = status.term_signal,
        stdout = read_file(kid.stdout),
        stderr = read_file(kid.stderr),
    }
    local ok, value = coroutine.resume(kid.w.coro, r)
    if coroutine.status(kid.w.coro) == 'dead' then
        -- KLUDGE KLUDGE KLUDGE correct???
        if not ok then
            pcall(function() server.setup_db:exec("ROLLBACK;") end)
            send_response(kid.w, 500, "Server Error", "text/plain", "%s\n",
                value)
        end
    end
end

function m:fork(loop, w, cfg, cmd, ...)
    local kid = {}
    kid.w = w
    kid.stdout, kid.stdout_fd = self:temp_file(w, '/tmp/kid_')
    kid.stderr, kid.stderr_fd = self:temp_file(w, '/tmp/kid_')
    kid.pid = C.fork()
    if kid.pid == 0 then
        loop:fork()
        -- We are the child
        local dev_null_fd = u.open('/dev/null', u.O_RDONLY) -- KLUDGE
        C.dup2(dev_null_fd, 0)
        C.dup2(kid.stdout_fd, 1)
        C.dup2(kid.stderr_fd, 2)
        C.close(dev_null_fd)
        C.close(kid.stdout_fd)
        C.close(kid.stderr_fd)

        if type(cfg) == 'number' then
            cfg = { cpu_limit = cfg > 0 and cfg or nil }
        end
        assert(type(cfg) == 'table')
        if cfg.cpu_limit then
            local lim = cfg.cpu_limit
            if type(lim) == 'number' then
                lim = {lim, lim + 1}
            end
            assert(type(lim) == 'table')
            local rlim = ffi.new('struct rlimit', lim)
            C.setrlimit(C.RLIMIT_CPU, rlim)
        end
        if type(cmd) == 'function' then
            local ok = cmd()
            os.exit(ok or 0)
        end
        local in_args = type(cmd) == 'table' and cmd or {cmd, ...}
        local args = {}
        for i, v in ipairs(in_args) do
            args[i] = tostring(v)
        end
        local cargs = ffi.new('char *[?]', #args + 1)
        local t = {} -- Keep stuff from being GC'd
        for i, arg in ipairs(args) do
            t[i] = ffi.new('char[?]', #arg + 1, arg)
            cargs[i - 1] = t[i]
        end
        C.execvp(cargs[0], cargs)
        log:error("execvp() failed") -- KLUDGE more descriptive
        kid.ok = false
        kid.exec_failed = true
        return kid
    elseif kid == -1 then
        log:error("fork() failed") -- KLUDGE more descriptive
        kid.ok = false
        kid.fork_failed = true
        return kid
    end
    kid.fn = coro_post_mortem
    self.kids[kid.pid] = kid
    return coroutine.yield(kid.w.coro)
end

function m:changed_rr(removed, added)
    self.changed = {}

    for i,rr in ipairs(removed) do
        self:rr_change("remove", rr)
    end
    for i,rr in ipairs(added) do
        self:rr_change("add", rr)
    end

    self:emit_accumulated_changes()
end

-- Function used to make sure the yellow LED gets turned off when there are
-- no tests running.
function m:tests_monitor()
    if not self.tests_monitor_timer then
        self.net_db = self.net_db or network_db:new()
        self.tests_monitor_timer = ev.Timer.new(function(l, t, e)
                if not self.net_db:tests_running() then
                    ledlib:SetLED("off", "yellow")
                    self.running_tests = 0
                    t:stop(l)
                end
            end, 1, 1)
    end
    self.tests_monitor_timer:start(ev_loop)
end

local function reload_nginx_config()
    -- Tell NGINX to re-read it's configuration files.
    local ch =
        io.popen('ps -o pid,args | grep nginx | grep master | grep -v grep')
    if ch then
        for line in ch:lines() do
            local pid = line:match('(%d+) nginx: master')
            if pid then
                C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGHUP)
                break
            end
        end
        ch:close()
    end
end

-- Function used to immediately kill all locally running tests
local function killall_tests()
    local plist = io.popen('ps -o pid,group | grep acq')
    if plist then
        local pid_list = {}
        for p in plist:lines() do
            local pid = p:match('(%d+) ')
            if pid then
                pid_list[#pid_list + 1] = pid
            end
        end
        plist:close()
        if #pid_list > 0 then
            for _, pid in ipairs(pid_list) do
                C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGINT)
            end
        end
    end
end

function m:rr_change(action, rr)
--print("rr_change(" .. action .. ", " .. cjson.encode(rr))
    if rr[1] == "_armada-connectors" then
        local cmd = "#" .. action .. "_connector"
        self:dispatch_to_clients{{cmd, rr[3]}}
    elseif rr[1] == "_armada-hardware" then
        local cmd = "#" .. action .. "_hardware"
        self:dispatch_to_clients{{cmd, rr[3]}}
    elseif rr[2] == "TXT" and rr[1]:find("^_system%-notification") then
        local m = cjson.decode(rr[3])
        if action == 'add' then
            if m then
                if m.type and m.type == 'error' then
                    ledlib:SetRedLED("on")
                end
                if m.type and m.type == '_notice' and
                    m.msg:find(__'Data acquisition has begun') then
                    ledlib:SetLED("eight_hz", "yellow")
                end
                if m.led and m.led_mode then
                    ledlib:SetLED(m.led_mode, m.led)
                end
                local cmd = "#" .. action .. "_notification"
                self:dispatch_to_clients{{cmd, rr[3]}}
            end
        elseif m and m.reboot then
            if system_config.max_test_restarts > 0 then
                -- If reboot was specified when the notification is removed, do
                -- the reboot after a short delay
                ev.Timer.new(function(l, t, e)
                        log:log(log.ERR, "System reboot due to error")
                        os.execute('/usr/bin/reboot error')
                    end, 2, 0):start(ev_loop)
            else
                -- If system error resets are overridden, just kill the test(s)
                log:log(log.ERR, "Stopping test due to critical error\n" ..
                    "Test will be not be restarted on next reboot")
                killall_tests()
            end
        elseif action == 'remove' and m and m.announce_remove then
            local cmd = "#" .. action .. "_notification"
            self:dispatch_to_clients{{cmd, rr[3]}}
        end
    elseif rr[2] == "TXT" and rr[3]:match("^available_channels:") then
        local cmd = "#" .. action .. "_channel"
        local js = rr[3]:gsub("^.-%:", "")
        local chs = cjson.decode(js)
--print("rr_change(AVAILABLE_CHANNELS): " .. cjson.encode(chs))
        local proxy_chs = {} -- KLUDGE
        for _, ch in ipairs(chs) do
            self:dispatch_to_clients{{cmd, {ch}}} -- KLUDGE
            proxy_chs[#proxy_chs + 1] = ch.connector
            if action == "remove" then
                self.throttled_live_data[ch.connector] = nil
            end
        end
        if action == "add" and #self.comet_clients > 0 then
            if not self.comet_proxies_by_name[rr[1]] then
                local host, port, ip = self.network:ResolveService(rr[1])
                if host then
                    self:connect_proxy(rr[1], ip, port, proxy_chs)
                end
            end
        end
    elseif rr[2] == "TXT" and rr[1]:match("._armada%-tests$") then
        self:accumulate_changes("tests", action, rr)
    elseif rr[2] == "SRV" and rr[1]:match("._armada%-tests$") then
        self:accumulate_changes("tests", action, rr)
        if action == 'add' then
            self.running_tests = self.running_tests or 0
            self.running_tests = self.running_tests + 1
--            ledlib:SetLED("eight_hz", "yellow")
            self:tests_monitor()
        elseif action == 'remove' and self.running_tests then
            local hash = rr[1]:match('^(.+)%._armada%-tests')
            if hash then
                os.execute('rm /tmp/nginx/conf.d/' .. hash .. '* 2>/dev/null')
                reload_nginx_config()
            end
            self.running_tests = self.running_tests - 1
            if self.running_tests <= 0 then
                if self.tests_monitor_timer then
                    self.tests_monitor_timer:stop(ev_loop)
                    self.tests_monitor_timer:clear_pending(ev_loop)
                end
                self.running_tests = 0
                ledlib:SetLED("off", "yellow")
            end
        end
    elseif rr[2] == "TXT" and rr[1]:match("%._armada%-files$") then
        self:accumulate_changes("file", action, rr)
    elseif rr[2] == "TXT" and rr[1]:match("._cx23r%-storage$") then
        self:accumulate_changes("storage", action, rr)
    elseif rr[2] == "TXT" and rr[1]:match("interactive%-trigger") then
        if action == 'add' then
            self:accumulate_changes("inter_trig", action, rr)
        end
    end
end

function m:dispatch_to_clients(message)
    for i = self.start_comet_client or 1, #self.comet_clients do
        self:write_comet(self.comet_clients[i], message)
    end
end

function m:accumulate_changes(tag, action, rr)
    self.accumulators[tag] = self.accumulators[tag] or {}
    local acc = self.accumulators[tag]
    local id = rr[1]
    local k, v

    if rr[2] == "TXT" then
        k, v = rr[3]:match("^(.-):%s*(.*)$")
        if not k then
            return
        end
    elseif rr[2] == "SRV" then
        local port, host = rr[3]:match("^%s*%d+%s+%d+%s+(%d+)%s+(%S+)")
        k, v = "srv_port", port -- KLUDGE for now, should have host
    end
    self.changed[tag] = self.changed[tag] or {}
    self.changed[tag][id] = true
    if action == "add" then
        acc[id] = acc[id] or {}
        acc[id][k] = v
    else
        if acc[id] then
            acc[id][k] = nil
        end
    end
end

function m:emit_accumulated_changes()
    for tag, changes in pairs(self.changed) do
        local acc = self.accumulators[tag]
        local update, remove = "#update_" .. tag, "#remove_" .. tag
        for f, _ in pairs(changes) do
            local rec = acc[f]
            if rec and next(rec) then
                rec.id = f -- KLUDGE
                self:dispatch_to_clients{{update, rec}}
                rec.id = nil -- KLUDGE
            else
                self:dispatch_to_clients{{remove, f}}
                acc[f] = nil
            end
        end
    end
end

local comet_parse_table
ffi.cdef[[
    enum {
        COMET_PARSE_ERROR = 0xFF,
        COMET_PARSE_DONE = 0xFE,
        COMET_STATE_MASK = 0x1F,
        COMET_OP_MASK = 0xE0,
        COMET_OP_SHIFT = 5,
        COMET_MARK_PREV = 1,
        COMET_MARK_THIS = 2,
        COMET_MARK_NEXT = 3,
        COMET_MARK_BACKSLASH = 4,
        COMET_MAX_MARKS = 1024,
    };
]]
do
    local IDLE = 0
    local OBJECT_START = 1
    local KEY = 2
    local KEY_BACKSLASH = 3
    local KEY_DONE = 4
    local COLON = 5
    local VALUE_STRING = 6
    local VALUE_STRING_BACKSLASH = 7
    local VALUE_NUMBER = 8
    local VALUE_DONE = 9
    local COMMA = 10
    local MAX_STATE = 11

    local table_size = 256 * MAX_STATE
    comet_parse_table = ffi.new('uint8_t[?]', table_size, C.COMET_PARSE_ERROR)

    local function trans(from_state, to_state, flags, ...)
        local a = {...}
        local f = bit.bor(to_state, flags)
        if #a == 2 and type(a[1]) == 'number' and type(a[2]) == 'number' then
            for c = a[1], a[2] do
                comet_parse_table[from_state * 256 + c] = f
            end
        elseif #a == 1 and type(a[1]) == 'table' then
            for _, c in ipairs(a[1]) do
                comet_parse_table[from_state * 256 + c] = f
            end
        else
            error('Bad trans() call')
        end
    end

    local whitespace = {0x0A, 0x0D, 0x20}
    local digits = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39}
    local plus_minus = {0x2B, 0x2D}
    local dot = {0x2E}
    local anycase_e = {0x45, 0x65}
    local double_quote = {0x22}
    local colon = {0x3A}
    local comma = {0x2C}
    local backslash = {0x5C}
    local left_curly = {0x7B}
    local right_curly = {0x7D}
    local end_packet = {0x04}

    local mark_prev = bit.lshift(C.COMET_MARK_PREV, C.COMET_OP_SHIFT)
    local mark_this = bit.lshift(C.COMET_MARK_THIS, C.COMET_OP_SHIFT)
    local mark_next = bit.lshift(C.COMET_MARK_NEXT, C.COMET_OP_SHIFT)
    local mark_backslash = bit.lshift(C.COMET_MARK_BACKSLASH, C.COMET_OP_SHIFT)

    trans(IDLE, IDLE, 0, whitespace)
    trans(IDLE, OBJECT_START, 0, left_curly)
    trans(IDLE, IDLE, C.COMET_PARSE_DONE, end_packet)

    trans(OBJECT_START, OBJECT_START, 0, whitespace)
    trans(OBJECT_START, KEY, mark_next, double_quote)
    trans(OBJECT_START, IDLE, 0, right_curly)

    trans(KEY, KEY, 0, 0x00, 0xFF)
    trans(KEY, KEY_BACKSLASH, mark_backslash, backslash)
    trans(KEY, KEY_DONE, mark_prev, double_quote)

    trans(KEY_BACKSLASH, KEY, 0, 0x00, 0xFF)

    trans(KEY_DONE, KEY_DONE, 0, whitespace)
    trans(KEY_DONE, COLON, 0, colon)

    trans(COLON, COLON, 0, whitespace)
    trans(COLON, VALUE_STRING, mark_next, double_quote)
    trans(COLON, VALUE_NUMBER, mark_this, digits)
    trans(COLON, VALUE_NUMBER, mark_this, plus_minus)

    trans(VALUE_STRING, VALUE_STRING, 0, 0x00, 0xFF)
    trans(VALUE_STRING, VALUE_STRING_BACKSLASH, mark_backslash, backslash)
    trans(VALUE_STRING, VALUE_DONE, mark_prev, double_quote)

    trans(VALUE_STRING_BACKSLASH, VALUE_STRING, 0, 0x00, 0xFF)

    -- KLUDGE too-lenient number parsing!!!
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, digits)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, dot)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, anycase_e)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, plus_minus) -- E-013, e+12 ...
    trans(VALUE_NUMBER, VALUE_DONE, mark_prev, whitespace)
    trans(VALUE_NUMBER, COMMA, mark_prev, comma)
    trans(VALUE_NUMBER, IDLE, mark_prev, right_curly)

    trans(VALUE_DONE, VALUE_DONE, 0, whitespace)
    trans(VALUE_DONE, COMMA, 0, comma)
    trans(VALUE_DONE, IDLE, 0, right_curly)

    trans(COMMA, COMMA, 0, whitespace)
    trans(COMMA, KEY, mark_next, double_quote)
end

local marks = ffi.new('int[?]', C.COMET_MAX_MARKS)

local function comet_parse(buf, len, out)
    local mk_i = 0
    local state = 0
    local consumed = 0
    local backslash = false
    for i = 0, len - 1 do
        state = comet_parse_table[buf[i] +
            bit_lshift(bit_band(C.COMET_STATE_MASK, state), 8)]
        local flags = bit_band(C.COMET_OP_MASK, state)
        if flags ~= 0 then
            if state < C.COMET_PARSE_DONE then -- KLUDGE
                local op = bit_rshift(flags, C.COMET_OP_SHIFT)
                if op == C.COMET_MARK_BACKSLASH then
                    backslash = true
                else
                    marks[mk_i] = i + op - 2
                    if backslash then
                        marks[mk_i] = -marks[mk_i]
                        backslash = false
                    end
                    mk_i = mk_i + 1
                end
            elseif state == C.COMET_PARSE_DONE then
                assert(bit_band(mk_i, 3) == 0)
                for j = 0, mk_i - 1, 4 do
                    local s1, e1 = marks[j], marks[j + 1]
                    local s2, e2 = marks[j + 2], marks[j + 3]
                    if e1 > 0 and e2 > 0 then
                        local key = ffi_string(buf + s1, e1 - s1 + 1)
                        out[key] = ffi_string(buf + s2, e2 - s2 + 1)
                    else
                        e1 = math.abs(e1)
                        e2 = math.abs(e2)
                        -- backslash special case
                        -- We know both are strings, so include leading
                        -- and trailing double quotes for cjson.decode()
                        local key = ffi_string(buf + s1 - 1, e1 - s1 + 3)
                        local val = ffi_string(buf + s2 - 1, e2 - s2 + 3)
                        key = cjson.decode(key)
                        out[key] = cjson.decode(val)
                    end
                end
                mk_i = 0
                state = 0
                consumed = i + 1
            elseif state == C.COMET_PARSE_ERROR then
                return consumed, false
            end
        end
    end
    return consumed, true
end

function m:comet_proxy_reader(loop, w, pxy)
    local buf = pxy.buf
    buf:assure_available()
    local got = C.read(pxy.socket, buf:tail(), buf:available())
    if got <= 0 then
        -- KLUDGE error here?
        self.close(loop, w)
        self.comet_proxies_by_socket[pxy.socket] = nil
        self.comet_proxies_by_name[pxy.name] = nil
        return
    end
    buf.size = buf.size + got

    if not pxy.got_headers then
        local http, end_pos = http_parser.response:match(tostring(buf))
        if not http then
            return
        end
        buf:remove(0, end_pos - 1)
        pxy.got_headers = true
    end

    local len = #buf
    local consumed, ok = comet_parse(buf.buf, len, self.throttled_live_data)
    if not ok then
        -- KLUDGE handle errors somehow here...
    end
    if consumed == len then
        buf:clear()
    else
        buf:remove(0, consumed)
    end
end

function m:connect_proxy(name, proxy_ip, proxy_port, channels)
    local pxy = {
        name = name,
        buf = membuf:new(8192),
        got_headers = false,
    }
    local proxy_w = self:connect(proxy_ip, proxy_port,
        function(loop, w) self:comet_proxy_reader(loop, w, pxy) end)
    if not proxy_w then
        -- KLUDGE emit some kind of better error message here
        log:error(__"Couldn't connect to proxy server")
        return nil
    end
    pxy.watcher = proxy_w
    pxy.socket = proxy_w.socket
    self.comet_proxies_by_socket[pxy.socket] = pxy
    self.comet_proxies_by_name[name] = pxy
    local json = cjson.encode(channels) .. "\n"
    local req = "GET /comet_stream HTTP/1.0\r\n"
        .. "Content-type: application/json\r\n"
        .. "Content-length: " .. #json .. "\r\n"
        .. "\r\n"
        .. json
    C.write(pxy.socket, req, #req)
end

function m:live_updater(loop, timer)
    if self.fake_live_data then
        self.fake_live_data(self.throttled_live_data,
            epoch_timestamp() - self.base_time)
    end
    local json = cjson.encode({{'#live_data', self.throttled_live_data}})
    self.throttled_live_data = {}
    self:dispatch_to_clients(json)
end

local function send_response_headers(w, code, http_msg, content_type, fmt, ...)
    local extras = #w.out_headers > 0 and table.concat(w.out_headers) or ''
    local msg =
        "HTTP/1.1 " .. code .. " " .. http_msg .. "\n"
        .. "Cache-Control: no-cache, no-store, must-revalidate\n"
        .. "Pragma: no-cache\n"
        .. "Expires: -1\n"
        .. "Content-Type: " .. content_type .. "\n"
        .. extras
        .. "\n"
        .. (#{...} == 0 and fmt or string.format(fmt, ...))
    while true do
        local amt, estr, eno = u.write(w.socket, msg)
        if not amt then
            if eno ~= u.EAGAIN then return end
        elseif amt == msg:len() then return
        else msg = string.sub(msg, amt + 1)
        end
    end
end

function send_response(w, code, http_msg, content_type, fmt, ...)
    send_response_headers(w, code, http_msg, content_type, fmt, ...)
    w.server.close(w.server.ev_loop, w)
end

local function redirect(w, code, http_msg, location)
    u.write(w.socket,
        "HTTP/1.1 " .. code .. " " .. http_msg .. "\n"
        .. "Location: " .. location .. "\n"
        .. "\n"
    )
    w.server.close(w.server.ev_loop, w)
end

function m:start_comet_stream(loop, w, begin_func)
    local op = w.headers.REQUEST_METHOD
    if op == 'POST' and w.content then
        self.fake_comet_setup[w.session_key] = assert(cjson.decode(w.content))
        send_response(w, 200, "OK", "text/plain", "")
        return
    else
        w.posted = self.fake_comet_setup[w.session_key] or {}
    end
    if w.posted.fake_live_data then
        self.fake_live_data =
            (loadstring('return ' .. w.posted.fake_live_data))()
    end

    if #self.comet_clients == 0 then
        local rate = tonumber(w.query.update_rate or w.posted.update_rate)
        self.live_update_rate = rate or 5
        self.live_w = function(t_loop, t_timer)
            self:live_updater(t_loop, t_timer)
        end
        local interval = 1.0 / self.live_update_rate
        self.live_data_timer = ev.Timer.new(self.live_w, interval, interval)
        self.live_data_timer:start(ev_loop)
    end

    self[begin_func](self, w)
    self.comet_clients[#self.comet_clients + 1] = w
    self.base_time = self.base_time or epoch_timestamp()

    -- Propagate current conditions to our Comet client
    -- KLUDGE horrifically inefficient
    if self.network then
        self.start_comet_client = #self.comet_clients
        self:changed_rr({}, self.network:QueryRR('*', '*', '*', '*'))
        self.start_comet_client = nil
    end

    if w.posted.fake_preamble then
        self:write_comet(w, w.posted.fake_preamble)
    end

    w.cleanup = function() self:cleanup_comet_stream(loop, w) end
end

function m:cleanup_comet_stream(loop, w)
    for i = 1, #self.comet_clients do
        if self.comet_clients[i] == w then
            table.remove(self.comet_clients, i)
            break
        end
    end
    if #self.comet_clients == 0 then
        for name, pxy in pairs(self.comet_proxies_by_name) do
            self.close(loop, pxy.watcher)
        end
        self.comet_proxies_by_name = {}
        self.comet_proxies_by_socket = {}
        self.live_data_timer:stop(loop)
        self.live_data_timer = nil
    end
end

handlers['stream.comet'] = function(loop, w)
    w.server:start_comet_stream(loop, w, 'begin_comet')
end

handlers['stream.eventsource'] = function(loop, w)
    w.server:start_comet_stream(loop, w, 'begin_eventsource')
end

local function out_sorted(out, t, fmt, stringify)
    if type(t) ~= 'table' then
        return
    end
    stringify = stringify or tostring
    local heads = {}
    for k, _ in pairs(t) do
        heads[#heads + 1] = k
    end
    table.sort(heads)
    for _, head in ipairs(heads) do
        out[#out + 1] = string.format(fmt, tostring(head), stringify(t[head]))
    end
end

function handlers.echo(loop, w, frag)
    local out = {}
    out[#out + 1] = string.format("frag=%s\n", frag)
    out_sorted(out, w.headers, "header %s => %s\n")
    out[#out + 1] = "\n"
    out[#out + 1] = string.format("accept_language: %s\n",
        cjson.encode(w.accept_language))
    out_sorted(out, w.content_type, "content_type %s => %s\n")
    out_sorted(out, w.query, "query %s => %s\n")
    out_sorted(out, w.cookies, "cookie %s => %s\n")
    out[#out + 1] = 'CONTENT_VALUES:\n'
    out_sorted(out, w.content_values, "content_values %s => %s\n")
    out[#out + 1] = 'MULTIPART:\n'
    out_sorted(out, w.multipart, "multipart %s => %s\n", cjson.encode)
    out[#out + 1] = 'END MULTIPART:\n'
    out[#out + 1] = "\n\n==== DEBUG ====\n\n"
    if w.debug then
        for _, v in ipairs(w.debug) do
            out[#out + 1] = v
        end
    end
    out[#out + 1] = "\n\n==== CONTENT ====\n\n"
    local content = string.gsub(w.content, "([\r\n])", function(c)
        return string.format("%%%02X%s", string.byte(c), c == "\n" and c or "")
    end)
    out[#out + 1] = content
    send_response(w, 200, "OK", "text/plain", table.concat(out))
end

function handlers.delete_file(loop, w, f)
    local files = {}
    if w.headers.REQUEST_METHOD == 'POST' and w.content and
        not f:find('^/.*%.sie$') then
        files = cjson.decode(w.content) or {}
    else
        files[#files + 1] = f:match('^/(.*)$')
    end

    local removed_files = {}
    local nonremoved_files = {}
    for _, frag in ipairs(files) do
        if frag:find('%.sie$') then
            if w.server:delete_sie_file(frag:gsub('%.sie', '')) then
                removed_files[#removed_files + 1] = frag
            else
                nonremoved_files[#nonremoved_files + 1] = frag
            end
        end
    end
    local response = {
        result = (#nonremoved_files == 0 and #removed_files > 0) and
            'success' or 'fail',
        removed_files = removed_files,
        nonremoved_files = nonremoved_files,
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(response))
end

function handlers.start_test(loop, w, frag)
    -- run-test look-mom-a-file.json '{"stop_time":"+30"}'

    if not frag or frag == "" then
        send_response(w, 500, "Error", "text/plain", "No test name\n")
        return
    end

    local test_file = '/tmp/' .. frag

    local setup = w.content

    log:log(log.DEBUG, "start_test test_file: " .. test_file)
    log:log(log.DEBUG, "start_test setup: " .. setup)

    local fh = assert(io.open(test_file, "w"))
    fh:write(setup)
    fh:close()

    local params
    if w.query then
        params = cjson.encode(w.query)
    end

    local cmd = '/usr/bin/run-test'
    w.server.master:Spawn(cmd, {cmd, test_file, params})

    local ok, err = 1, nil
    if ok then
        send_response(w, 200, "OK", "text/plain", "Started test\n")
    else
        send_response(w, 500, "Error", "text/plain", "Couldn't start test\n")
    end
end

function handlers.stop_test(loop, w, frag)
    -- wget 'http://[fda9:aac8:be27:0:800:3e03:181:0]:41824/stop'
    local cmd = '/usr/bin/wget'
    local ip = w.server.network:GetModuleIP()
    local port = frag:sub(2) -- KLUDGE???
    if not port:match('^%d+$') then
        local test_id = port .. '._armada-tests'
        local rec = w.server.network:QueryRR(test_id, 'SRV', '*', '*')
        if rec and #rec == 1 then
            port = tonumber(rec[1][3]:match('^%d+%s+%d+%s+(%d+)'))
        end
    end
    local uri = string.format('http://[%s]:%d/stop', ip, port)
    w.server.master:Spawn(cmd, {cmd, '-q', '-O', '/dev/null', uri})
    local ok, err = 1, nil
    if ok then
        send_response(w, 200, "OK", "text/plain", "Stopped test\n")
    else
        send_response(w, 500, "Error", "text/plain", "Couldn't stop test\n")
    end
end

local function html_escape(str)
    return str:gsub('[&"<]', {
                        ['&'] = '&amp;',
                        ['"'] = '&quot;',
                        ['<'] = '&lt;'
                    })
end

function handlers.get_log(loop, w, frag)
    local entries = { '<th>Time</th><th>Component</th>' ..
                      '<th>Level</th><th>Message</th>'}
    local db = sqlite3.open('/var/lib/armada/log.db')
    for r in db:rows([[
        SELECT time, component, level, message FROM log
               ORDER BY rowid DESC LIMIT 50
    ]]) do
        table.insert(entries, string.format(
            '<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>',
            html_escape(r[1]), html_escape(r[2]),
            html_escape(r[3]), html_escape(r[4])))
    end
    db:close()
    send_response(w, 200, "OK", "text/html", "<table>%s</table>",
                  table.concat(entries))
end

local function db_error(loop, w, status, message, result)
    result = result or {ok=false}
    send_response(w, status, message, "application/json", cjson.encode(result))
end

local function sql_raw(db, sql)
    local ok, result = pcall(function() return db:exec(sql) end)
    if ok then
        return {ok=true}, ok
    else
        return {ok=false, error_code=result}, ok
    end
end

sql_exec = function(db, table_name, sql, args)
    local query = string.format(sql, table_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query, args) end)
    if ok then
        return result, ok
        -- KLUDGE broadcast results to listeners based on table names here?
    else
        return {ok=false, error=result}, ok
    end
end

local function sql_do(db, tbl_name, sql)
    local query = sql:format(tbl_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query) end)
    if ok then
        return {ok=true}, ok
    else
        return {ok=false, error_code=result}, ok
    end
end

function handlers.raw_db(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local ok, result, ignore = false, nil, nil

    local db_name, doc = frag:match("^/([%w%_%-]+)/([%w%_%-]+)$")
    if not db_name then
        db_name = frag:match("^/([%w%_%-]+)/?$")
        doc = ""
    end
    local tbl_name = "rdb_" .. db_name

    if not db_name then
        return db_error(loop, w, 404, "No such database", result)
    end

    local db = w.server.setup_db
    ignore, ok = sql_raw(db, "BEGIN TRANSACTION;")

    local sec, nsec = u.clock_gettime()
    local stamp = string.format("(%022.0f)", sec * 1e9 + nsec)

    if doc ~= "" then -- Document level
        if op == "GET" then
            if doc:match("^_") then
                if doc == "_all_docs" then
                    local r
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content FROM "%s";
                    ]])
                    if ok then
                        result = {}
                        for _, v in ipairs(r) do
                            result[#result + 1] = v.content
                        end
                        result = "[" .. table.concat(result, ",\n") .. "]"
                    else
                        result = {ok=false, error='No such database'}
                    end
                end
            else
                local got, ok = sql_exec(db, tbl_name, [[
                    SELECT * FROM "%s" WHERE id = ?;
                ]], {doc})
                if ok and #got == 1 and got[1].content then
                    result = got[1].content
                else
                    result = {ok=false, error="No such document"}
                end
            end
        elseif op == "PUT" then
            result, ok = sql_exec(db, tbl_name, [[
                UPDATE "%s" SET content = ?, updated = ? WHERE id = ?;
            ]], {w.content, stamp, doc})
        elseif op == "POST" then
            result, ok = sql_exec(db, tbl_name, [[
                INSERT INTO "%s" (id, content, updated) VALUES (?, ?, ?);
            ]], {doc, w.content, stamp})
        elseif op == "DELETE" then
            result, ok = sql_exec(db, tbl_name, [[
                DELETE FROM "%s" WHERE id = ?;
            ]], {doc})
        end
    else -- Database level
        if op == "GET" and db_name == "_all_dbs" then
            result, ok = sql_exec(db, "sqlite_master", [[
                SELECT substr(name, 5) AS name
                FROM "%s"
                WHERE type = 'table' AND name LIKE 'rdb*_%%' escape '*';
            ]])
            if ok then
                for i, v in pairs(result) do
                    result[i] = v.name
                end
            end
        elseif op == "PUT" then
            result, ok = sql_do(db, tbl_name, [[
                CREATE TABLE "%s" (
                    id TEXT NOT NULL,
                    content TEXT NOT NULL,
                    updated INTEGER NOT NULL DEFAULT 0
                );
            ]])
        elseif op == "DELETE" then
            result, ok = sql_do(db, tbl_name, [[
                DROP TABLE "%s";
            ]])
        end
    end

    ignore, ok = sql_raw(db, ok and "COMMIT;" or "ROLLBACK;")

    if type(result) == 'table' then
        result = cjson.encode(result)
    end

    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "application/json", result)
end

local function json_concat(r)
    local result = {}
    for _, v in ipairs(r) do
        result[#result + 1] = v.content
    end
    result = "[" .. table.concat(result, ",\n") .. "]"
    return result
end

local function deep_equal(a, b)
    if a == b then
        return true
    end
    if type(a) ~= type(b) then
        return false
    end
    if type(a) == 'table' then
        for k, v in pairs(a) do
            if not deep_equal(v, b[k]) then
                return false
            end
        end
        for k, v in pairs(b) do
            if not a[k] then
                return false
            end
        end
        return true
    end
    return false
end

local function table_delta(old, new)
    local del, add = {}, {}
    for k, v in pairs(old) do
        if not deep_equal(v, new[k]) then
            del[k] = v
            add[k] = (new[k] == nil or new[k] == cjson.null)
                and cjson.null or v
        end
    end
    for k, v in pairs(new) do
        if not old[k] then
            del[k] = cjson.null
            add[k] = v
        end
    end
    return del, add
end

function handlers.tracked_db(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local ok, result, ignore = false, nil, nil

    local _1, schema, db_name, doc
        = frag:match("^/(([%w%_%-%/]+)/)([%w%_%-]+)/([%w%_%-%.%@%~%%]*)$")

    if not schema then
        doc = ""
        _1, schema, db_name
            = frag:match("^/(([%w%_%-%/]+)/)([%w%_%-]*)$")
    end

    if not schema then
        return db_error(loop, w, 404, "Malformed path", result)
    end

    schema = schema:gsub("%/+$", "")
    doc = w.server.url_unescape(doc)

    local tbl_name = string.format("tsd_%s", schema)
    local history_tbl_name = string.format("tsh_%s", schema)

    if not db_name then
        return db_error(loop, w, 404, "No such database", result)
    end

    local db = w.server.setup_db
    ignore, ok = sql_raw(db, "BEGIN TRANSACTION;")

    local stamp = iso_timestamp()
    local description = w.query.description
    local track_history = false
    local comet_updates = nil

    if doc ~= "" then -- Document level
        if op == "GET" then
            if doc:match("^_") then
                if doc == "_all_docs" then
                    local r
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content
                        FROM "%s"
                        WHERE db_name = ? AND id != '' AND deleted = '';
                    ]], {db_name})
                    if ok and #r > 0 then
                        if w.api_internal.export then
                            local setup = {}
                            for i = 1, #r do
                                setup[i] = cjson.decode(r[i].content)
                            end
                            local ss = setup_spec:new(setup)
                            local ok, err = pcall(function() result = 
                                ss:export("user_order") end)
                            if not ok then
                                --ok is true because we want a 200 return code,
                                --since error is probably not network related
                                result = {ok=true, error_message=err}
                            end
                        else
                            local o = {'['}
                            for i = 1, #r do
                                if i > 1 then
                                    o[#o + 1] = ',\n'
                                end
                                o[#o + 1] = r[i].content
                            end
                            o[#o + 1] = ']\n'
                            result = table.concat(o)
                        end
                    else
                        result = {ok=false, error='No such database'}
                    end
                elseif doc == "_historical" then
                    local r
                    local at = w.query.at or iso_timestamp()
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content
                        FROM "%%s"
                        WHERE db_name = ? AND id != ''
                            AND added <= ?
                            AND (deleted = '' OR deleted > ?);
                    ]], {db_name, at, at})
                    if ok then
                        result = json_concat(r)
                    else
                        result = {ok=false, error='No such database'}
                    end
                elseif doc == "_most_recent" then
                    local r
                    local extras = ''
                    if w.query.limit then
                        extras = 'LIMIT ' .. w.query.limit
                    end
                    local match_id = w.query.match_id or ''
                    r, ok = sql_exec(db, tbl_name, string.format([[
                        SELECT content, added
                        FROM "%%s"
                        WHERE db_name = ? AND id != '' AND deleted = ''
                        ORDER BY added DESC
                        %s;
                    ]], extras), {w.session.user_id}) -- KLUDGE ???
                    if ok then
                        result = {}
                        for _, v in ipairs(r) do
                            local e = cjson.decode(v.content)
                            e._timestamp = v.added
                            result[#result + 1] = e
                        end
                    else
                        result = {ok=false, error='No such database'}
                    end
                end
            else
                local got, ok = sql_exec(db, tbl_name, [[
                    SELECT *
                    FROM "%s"
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], {db_name, doc})
                if ok and #got == 1 and got[1].content then
                    result = got[1].content
                else
                    result = {ok=false, error="No such document"}
                end
            end
        elseif op == "POST" and doc == "_delta" then
            local payload = cjson.decode(w.content)
            local d_old, d_new = payload.old, payload.new
            if w.api_internal.import then
                description = description or 'Import setup'
            end
            description = description or payload.description
            local err = nil
            if type(d_old) ~= 'table' or type(d_new) ~= 'table' then
                err = "Missing or non-{} old or new field"
            end
            if not err then
                -- Check that old and new have exactly the same set of keys
                for k, v in pairs(d_old) do
                    if not d_new[k] then
                        err = k .. " in old not in new" -- KLUDGE I18N
                        break
                    end
                end
                if not err then
                    for k, v in pairs(d_new) do
                        if not d_old[k] then
                            err = k .. " in new not in old" -- KLUDGE I18N
                            break
                        end
                    end
                end
            end
            if not err then
                -- First we check to make sure that the old state matches
                -- the current state of the database. Abort if not.
                -- KLUDGE add coalesce capability:
                --  if coalesce_flag_or_condition
                --      UPDATE %s SET content = ?
                --      WHERE db_name = ? AND id = ? AND deleted = ''
                local get_cur = db:prepare(string.format([[
                    SELECT content
                    FROM "%s"
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], tbl_name))
                local do_delete = db:prepare(string.format([[
                    UPDATE "%s"
                    SET deleted = ?
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], tbl_name))
                local do_insert = db:prepare(string.format([[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], tbl_name))
                local cu = {}
                local changed = false
                for id, old in pairs(d_old) do
                    local new = d_new[id]
                    local deletable = nil
                    local insertable = nil
                    local got = get_cur:exec{db_name, id}
                    if type(got) ~= 'table' or #got > 1 then
                        err = string.format("Conflict at %s", id)
                        break
                    end
                    if old == cjson.null then
                        if #got == 0 then
                            insertable = new
                            cu[id] = new
                        else
                            err = string.format(
                                "Conflict at %s, insert on existing", id)
                            break
                        end
                    elseif new == cjson.null then
                        if #got > 0 then
                            deletable = id
                            cu[id] = cjson.null
                        else
                            err = string.format(
                                "Conflict at %s, delete on nonexisting", id)
                            break
                        end
                    else
                        local cur
                        if #got == 1 then
                            cur = cjson.decode(got[1].content)
                        elseif #got == 0 then
                            cur = {} -- KLUDGE WRONG BROKEN (???)
                        else
                            err = string.format(
                                "Internal error at %s, multiple existing", id)
                            break
                        end
                        -- We only match keys in delta, not in current
                        -- content, as it's a delta
                        for k, v in pairs(old) do
                            if not us.is_equal(cur[k], v) then
                                -- The following makes nil and "" equivalent
                                -- v will not be nil by definition
                                if not (cur[k] == nil
                                        and (v == "" or v == cjson.null)) then
                                    err = string.format(
                                        "Conflict at %s/%s [%s <=> %s]",
                                        id, k, tostring(cur[k]), tostring(v))
                                    break
                                end
                            end
                            if new[k] == cjson.null then
                                cur[k] = nil
                            else
                                cur[k] = new[k]
                            end
                        end
                        deletable = id
                        insertable = cur
                    end
                    if deletable then
                        do_delete:exec{stamp, db_name, deletable}
                        changed = true
                    end
                    if insertable then
                        insertable = cjson.encode(insertable)
                        do_insert:exec{db_name, id, insertable, stamp}
                        changed = true
                    end
                end
                if not err and changed then
                    comet_updates = {{'#tracked_db_delta', {
                        schema = schema, db = db_name,
                        source_session_key = w.session_key,
                        description = description,
                        old = d_old, new = d_new,
                    }}}
                end
                get_cur:finalize()
                do_delete:finalize()
                do_insert:finalize()
                result = {ok = not err, error = err}
            end
        elseif op == "POST" or op == "PUT" or op == "DELETE" then
            local d_old, d_new
            local got, ok = sql_exec(db, tbl_name, [[
                SELECT content
                FROM "%s"
                WHERE db_name = ? AND id = ? AND deleted = '';
            ]], {db_name, doc})
            if ok then
                if type(got) ~= 'table' or #got > 1 then
                    err = string.format("Internal error at %s", id)
                    ok = false
                else
                    d_old, d_new = {}, {}
                    d_new[doc] = cjson.null -- DELETE default case
                    d_old[doc] = (#got == 0)
                        and cjson.null
                        or cjson.decode(got[1].content)
                end
            end
            if ok then
                -- Here, PUT equivalent to POST
                result, ok = sql_exec(db, tbl_name, [[
                    UPDATE "%s"
                    SET deleted = ?
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], {stamp, db_name, doc})
            end
            if ok and (op == "PUT" or op == "POST") then
                result, ok = sql_exec(db, tbl_name, [[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], {db_name, doc, w.content, stamp})
                d_new[doc] = cjson.decode(w.content)
            end
            if ok then
                comet_updates = {{'#tracked_db_delta', {
                    schema = schema, db = db_name,
                    source_session_key = w.session_key,
                    description = description,
                    old = d_old, new = d_new,
                }}}
            end
            track_history = ok
        end
    else -- Database level
        if op == "GET" then
            if db_name == "_all_dbs" then
                local r
                local match_id = w.query.match_id or ''
                r, ok = sql_exec(db, tbl_name, [[
                    SELECT db_name, content
                    FROM "%s"
                    WHERE id = ? AND deleted = '';
                ]], {match_id})
                if ok then
                    result = {}
                    for _, v in ipairs(r) do
                        result[v.db_name] = cjson.decode(v.content)
                    end
                end
            elseif db_name == "_most_recent" then
                local r
                local extras = ''
                if w.query.limit then
                    extras = 'LIMIT ' .. w.query.limit
                end
                local match_id = w.query.match_id or ''
                r, ok = sql_exec(db, tbl_name, string.format([[
                    SELECT content, added
                    FROM "%%s"
                    WHERE deleted = ''
                    ORDER BY added DESC
                    %s;
                ]], extras), {})
                if ok then
                    result = {}
                    for _, v in ipairs(r) do
                        local e = cjson.decode(v.content)
                        r.timestamp = v.added
                        result[#result + 1] = e
                    end
                else
                    result = {ok=false, error='No such database'}
                end
            end
        elseif op == "PUT" then
            description = description
                or string.format("Create database '%s'", db_name)
            result, ok = sql_do(db, tbl_name, [[
                CREATE TABLE "%s" (
                    db_name TEXT NOT NULL,
                    id TEXT NOT NULL,
                    content TEXT NOT NULL,
                    added TEXT NOT NULL,
                    deleted TEXT NOT NULL DEFAULT ''
                );
            ]])
            if ok then
                result, ok = sql_do(db, '', string.format([[
                    CREATE INDEX "%s_idx"
                    ON "%s" (db_name, id, deleted);
                ]], tbl_name, tbl_name))
            end
            if ok then
                result, ok = sql_do(db, history_tbl_name, [[
                    CREATE TABLE "%s" (
                        db_name TEXT NOT NULL,
                        stamp TEXT NOT NULL,
                        author TEXT NOT NULL DEFAULT '',
                        description TEXT NOT NULL DEFAULT ''
                    );
                ]])
            end
            if ok then
                result, ok = sql_do(db, '', string.format([[
                    CREATE INDEX "%s_idx"
                    ON "%s" (db_name, stamp);
                ]], history_tbl_name, history_tbl_name))
            end
            if ok then
                result, ok = sql_exec(db, tbl_name, [[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], {db_name, "", w.content, stamp})
            end
            if ok then
                comet_updates = {
                    {'#tracked_db_create', {schema=schema, db=db_name}}
                }
            end
            track_history = ok
        elseif op == "DELETE" then
            local db_names
            if w.content and cjson.decode(w.content) then
                db_names = cjson.decode(w.content)
            else
                db_names = {db_name}
            end
            ok = true
            db_names = db_names or {}
            local _deleted, _nondeleted = {}, {}
            for _, db_name in ipairs(db_names) do
                -- get records that will be deleted
                -- will be sent over comet stream
                local deleted, lok = sql_exec(db, tbl_name, [[
                    SELECT * FROM "%s"
                    WHERE db_name = ? AND deleted = '';
                ]], {db_name})

                if not lok or #deleted == 0 then
                    _nondeleted[#_nondeleted + 1] = db_name
                else
                    -- This allows undo. Maybe add a 'force' query flag to
                    -- actually delete rows?
                    local res, lok = sql_exec(db, tbl_name, [[
                        UPDATE "%s" SET deleted = ?
                        WHERE db_name = ? AND deleted = '';
                    ]], {stamp, db_name})
                    if lok then
                        _deleted[#_deleted + 1] = db_name
                        comet_updates = {
                            {
                                '#tracked_db_delete',
                                {
                                    schema = schema,
                                    db = db_name,
                                    deleted = deleted
                                }
                            }
                        }
                        w.server:dispatch_to_clients(comet_updates)
                    end
                end
            end
            result = {
                requested = db_names,
                deleted = _deleted,
                nondeleted = _nondeleted,
            }
            comet_updates = nil
        end
    end

    if track_history then
        local author = w.query.author -- KLUDGE user_id, or ??? Backdoor???
        if not author and w.session then
            author = w.session.user_id
        end
        result, ok = sql_exec(db, history_tbl_name, [[
            INSERT INTO "%s" (db_name, stamp, author, description)
            VALUES (?, ?, ?, ?);
        ]], {db_name, stamp, author or "", description or ""})
        if ok then
            result = {ok=true}
        end
    end
    ignore, ok = sql_raw(db, ok and "COMMIT;" or "ROLLBACK;")

    if ok and comet_updates then
        w.server:dispatch_to_clients(comet_updates) -- KLUDGE not selective
    end

    if w.api_internal.import then
        return {ok = ok, result = result}
    end

    if type(result) == 'table' then
        result = cjson.encode(result)
    end

    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "application/json", result)
end

local function _sign_in(loop, w, frag)
    local user_name, password = w.query.user_name, w.query.password
    local status = nil -- NOT, Repeat, NOT 401 Unauthorized

    if type(user_name) ~= "string" or type("password") ~= "string" then
        error('Missing user_name or password')
    end
    if user_name:len() == 0 then
        error('Empty user_name')
    end

    local db = w.server.setup_db
    local got = sql_exec(db, "tsd_users", [[
        SELECT * FROM "%s" WHERE db_name = 'users' AND id = ? AND deleted = '';
    ]], {w.query.user_name or ""})

    if #got < 1 then
        error('No user entry for user_name')
    elseif #got > 1 then
        error('Too many user entries for user_name')
    elseif not got[1].content then
        error('No user record for user_name')
    end

    local user = cjson.decode(got[1].content)
    if type(user) ~= "table" or type(user.password_sha) ~= "string" then
        error('Bad user entry for user_name')
    end

    local salt = user.salt or '' -- KLUDGE
    local hash = bits.tohex(crypto.SHA1(password .. salt))
    if hash ~= user.password_sha then
        error('Incorrect password')
    end

    if not w.authorization_key then
        w.authorization_key = uuid.generate()
    end

    user.password_sha, user.salt = nil, nil -- Clear out private stuff
    local session_key = uuid.generate()
    w.server.session_db:update_session{
        session_key = session_key,
        authorization_key = w.authorization_key,
        timestamp = epoch_timestamp(),
        user_id = user.id,
        state = {user=user, hostname=hostname}, -- KLUDGE
    }

    local result = {
        ok = true,
        session_key = session_key,
        authorization_key = w.authorization_key,
        url = "/console.html", -- KLUDGE make url reflect user history
    }

    local cookie = "\nSet-Cookie: authorization_key=" .. w.authorization_key
        .. "; Secure"

    send_response(w, 200, "OK", "application/json" .. cookie,
        cjson.encode(result))
end

function handlers.hash(loop, w, frag)
    local s = string.sub(frag, 2)
    local ret = bits.tohex(crypto.SHA1(s))
    send_response(w, 200, "OK", "application/json", cjson.encode({s, ret}))
end

function handlers.sign_in(loop, w, frag)
    -- This is a special case to make sure that internal information
    -- doesn't leak outside during the authorization process
    local ok, result = pcall(function() return _sign_in(loop, w, frag) end)
    if not ok then
        log(log.WARNING, __'/api/sign_in error: "{0}"\nTraceback: {1}\n',
            result, debug.traceback())
        result = {ok=false, error="Forbidden"}
        send_response(w, 403, "Forbidden", "application/json",
            cjson.encode(result))
    end
end

function handlers.sign_out(loop, w, frag)
    local result = false
    if w.query.session_key then
        w.server.session_db:end_session(w.query.session_key)
        result = true;
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({result=result}))
end

function handlers.update_settings(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local up = assert(cjson.decode(w.content))
    for socket, client in pairs(w.server.comet_clients) do
        if client.session_key == w.session_key then
            for _, update in ipairs(up.updates) do
                client.server:write_comet(client, {{'#console_log', update}})
                local connector = update.connector:gsub("^.-@", "")
                local host, port, ip
                    = w.server.network:ResolveService(connector)
                if host then
                    local cb_w = w.server:connect(ip, port,
                        function(loop, w)
                            local got = u.read(w.socket, 8192)
                            if not got or got == "" then
                                w.server.close(loop, w)
                                return
                            end
                        end
                    )
                    local json = cjson.encode(update) .. "\n"
                    local req = "POST /update_settings/ HTTP/1.0\r\n"
                        .. "Content-type: application/json\r\n"
                        .. "Content-length: " .. #json .. "\r\n"
                        .. "\r\n"
                        .. json
                    u.write(cb_w.socket, req)
                end
            end
        end
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({ok=1}))
end

function handlers.session_data(loop, w, frag)
    local session = w.server.session_db:get_session(w.session_key) -- KLUDGE ? nil case?
    send_response(w, 200, "OK", "application/json",
        cjson.encode(session.state))
end

function handlers.dump_sessions(loop, w, frag)
    send_response(w, 200, "OK", "application/json",
        cjson.encode(w.server.session_db:get_all_sessions()))
end

local api_url_parser
do
    local C, Cc, Cg, Ct, P = lpeg.C, lpeg.Cc, lpeg.Cg, lpeg.Ct, lpeg.P
    local V = lpeg.V
    local uuid_digit = lpeg.R('09', 'af', 'AF')

    local function method(s)
        return Cg(P(s), 'method') * Cg(Cc(true), string.lower(s)) * P':/'
    end
    local function prefix(s)
        return Cg(C(s), 'prefix') * P'/'
    end
    local function action(s, action)
        if action then
            return P(s) * Cg(Cc(action), 'action')
        else
            return Cg(P(s), 'action')
        end
    end
    local function d(n)
        return P(uuid_digit^n - uuid_digit^(n+1))
    end

    api_url_parser = P{'api',
        base = P'api/',
        delete = method('DELETE'),
        get = method('GET'),
        post = method('POST'),
        put = method('PUT'),
        fake_uuid = Cg(P(P(1) - P'/')^1, 'id'),
        uuid = Cg(P(
            d(8) * '-' * d(4) * '-' * d(4) * '-' * d(4) * '-' * d(12)
          ) / string.lower, 'id'),
        api = V'session_api' + V'sie_api' + V'system_api' + V'test_api',
        session_prefix = V'base' * prefix('session'),
        session_api = Ct(
            V'get' * V'session_prefix' * action('state')
          + V'get' * V'base' * Cg(C'session', 'prefix')
                * Cg(Cc'session', 'action')
        ) * -P(1),
        sie_prefix = V'base' * prefix('sie'),
        sie_id = V'sie_prefix' * V'fake_uuid',
        sie_api = Ct(
            V'get' * V'sie_prefix' * V'uuid' * ( 
                action('.setup', 'export')
              + action('.setup', 'get_tags')
              + action('.setup', 'get_file_tags')
              + action('.setup', 'get_test_tags')
              + action('.setup', 'get_channel_tags')
              + action('.setup', 'get_system_version')
              + action('.setup', 'get_xml_metadata')
              + action('.setup', 'get_block_index')
              + action('.setup', 'get_armada_input_metadata')
            )
          + V'get' * V'sie_id' * P'/'
              * Cg(C((P(1) - '.setup')^1 * '.setup'), 'filename')
              * (
                action('', 'export')
              + action('', 'get_tags')
              + action('', 'get_file_tags')
              + action('', 'get_test_tags')
              + action('', 'get_channel_tags')
              + action('', 'get_system_version')
              + action('', 'get_xml_metadata')
              + action('', 'get_block_index')
              + action('', 'get_armada_input_metadata')
            )
          + V'get' * V'sie_id' * '/' * (
                action('export')
              + action('get_tags')
              + action('get_file_tags')
              + action('get_test_tags')
              + action('get_channel_tags')
              + action('get_system_version')
              + action('get_xml_metadata')
              + action('get_block_index')
              + action('get_armada_input_metadata')
            )
              * ('/' * Cg(P(1)^0, 'filename'))^-1
        ) * -P(1),
        system_prefix = V'base' * prefix('system'),
        system_api = Ct(
            V'get' * V'system_prefix' * (
                action('version')
              + action('_dev_config')
              + action('get_datetime')
              + action('log')
              + action('available_channels')
              + action('export_log')
              + action('unlock_start')
              + action('cancel_restart')
              + action('get_cal_cert')
              + action('setupdb_status')
              + action('delete_databases')
              + action('customer_service_package')
              + action('download_local_file')
            )
          + V'post' * V'system_prefix' * (
                action('update_firmware')
              + action('set_datetime')
              + action('acknowledge_notify')
              + action('reboot')
              + action('control')
              + action('set_cal_cert')
              + action('set_inter_trig')
              + action('read_local_file')
              + action('upload_local_file')
            )
          + V'delete' * V'system_prefix' * (
                action('log')
            )
        ) * -P(1),
        test_prefix = V'base' * prefix('test'),
        test_id = V'test_prefix' * V'fake_uuid',
        test_api = Ct(
            V'get' * V'test_prefix' * V'uuid' * action('.sie', 'data')
          + V'get' * V'test_id' * P'/'
              * Cg(C((P(1) - '.sie')^1 * '.sie'), 'filename')
              * action('', 'data')
          + V'get' * V'test_prefix' * action('data_file_list')
          + V'get' * V'test_prefix' * action('rename')
          + V'get' * V'test_prefix' * V'uuid' * action('.setup', 'export')
          + V'get' * V'test_id' * P'/'
              * Cg(C((P(1) - '.setup')^1 * '.setup'), 'filename')
              * action('', 'export')
          + V'get' * V'test_id' * '/' * action('export')
              * ('/' * Cg(P(1)^0, 'filename'))^-1
          + V'get' * V'test_id' * P'/' * action('data')
              * ('/' * Cg(P(1)^0, 'filename'))^-1
          + V'put' * V'test_id' * action('', 'import')
          + V'put' * V'test_prefix' * action('import')
          + V'post' * V'test_prefix' * action('create')
          + V'post' * V'test_prefix' * action('start')
          + V'post' * V'test_prefix' * action('restart')
          + V'post' * V'test_id' * P'/' * action(
              P 'import'
              + 'modify'
              + 'replace'
              + 'start'
              + 'restart'
              + 'stop'
              + 'stream'
              + 'truncate_data'
              -- + 'add' + 'remove' -- KLUDGE think modify/replace sufficient
            )
          + V'post' * V'test_prefix' * action('import')
          + V'delete' * V'test_id'
        ) * -P(1),
    }
end

local api = {
    session = {},
    sie = {},
    system = {},
    test = {},
}

local function multipart_file(w, r)
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                return mp
            end
        end
    end
end

function api.session.session(loop, w, r)
    if w.session then
        send_response(w, 200, "OK", "application/json", cjson.encode(w.session))
    else
        local err = {__"No session with key {0}", w.session_key}
        send_response(w, 500, "ERROR", "application/json", cjson.encode(err))
    end
end

function api.session.state(loop, w, r)
    if w.session and w.session.state then
        send_response(w, 200, "OK", "application/json",
            cjson.encode(w.session.state))
    else
        local err = {__"No session with key {0}", w.session_key}
        send_response(w, 500, "ERROR", "application/json", cjson.encode(err))
    end
end

function api.sie.export(loop, w, r)
    local f = w.server:fork(loop, w, {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            local setup = sp:get_setup()
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..
                r.filename .. '"\n'
            )
            send_response(w, 200, "OK", "application/json", setup)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_tags(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json", sp:get_tags())
end

function api.sie.get_file_tags(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json", sp:get_file_tags())
end

function api.sie.get_test_tags(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json", sp:get_test_tags())
end

function api.sie.get_channel_tags(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json", sp:get_channel_tags())
end

function api.sie.get_xml_metadata(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/xml", sp:get_xml_metadata())
end

function api.sie.get_block_index(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "text/plain", sp:get_block_index())
end

function api.sie.get_system_version(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json", sp:get_system_version())
end

function api.sie.get_armada_input_metadata(loop, w, r)
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    local sp = sie_parser:new(sie_name)
    send_response(w, 200, "OK", "application/json",
        sp:get_armada_input_metadata())
end

function api.system.log(loop, w, r)
    local f = w.server:fork(loop, w, 10, function()
        -- KLUDGE maybe try using HTTP Range or Content-Range with unit "dates"
        local q = w.query
        local count = q.count or 10
        local offset = q.offset or 0
        local lower = q.start_time or '0000-01-01'
        local upper = q.end_time or '9999-12-31'
        local order = q.order or 'newest'
        local method = r.method:lower()

        assert(order == 'newest' or order == 'oldest',
            "'order' must be 'newest' or 'oldest'")

        -- KLUDGE use database.open() for versioning
        local db = sqlite3.open('/var/lib/armada/log.db')

        if order == 'oldest' and not q.start_time then
            local rows
            for r in db:rows[[
                SELECT count(*) FROM log;
            ]] do
                rows = tonumber(r[1])
            end
            offset = rows - count - offset
        end

        send_response_headers(w, 200, "OK", "application/json", '[')

        if method == 'delete' then
            local ret = db:exec(string.format(
                "DELETE FROM log WHERE time >= '%s' and time <= '%s';",
                lower, upper))
            if ret == sqlite3.OK then
                u.write(w.socket, cjson.encode({purged=true}))
            else
                u.write(w.socket, cjson.encode({purged=false,
                    error = db:errmsg()}))
            end
        else
            local req = string.format([[
                SELECT time, component, level, message
                FROM log WHERE time >= ? and time <= ?
                AND ((component NOT LIKE '%%kern'
                AND component NOT LIKE '%%auth%%'
                AND (level LIKE 'warn%%' OR level LIKE 'err%%'
                    OR level LIKE 'crit%%'))
                OR message LIKE 'test-engine%%'
                OR message LIKE 'run-test%%'
                OR message LIKE 'restart-%%'
                OR component == 'powerd')
                ORDER BY time %s LIMIT ? OFFSET ?;
            ]], (order == 'newest' and 'DESC' or 'ASC'))
            if system_config.get_all_log_entries then
                req = string.format([[
                    SELECT time, component, level, message
                    FROM log
                    WHERE time >= ? and time <= ?
                    ORDER BY time %s LIMIT ? OFFSET ?;
                ]], (order == 'newest' and 'DESC' or 'ASC'))
            end
            local sth = db:prepare(req)
            local i = 0
            sth:bind_values(lower, upper, count, offset)
            for r in sth:rows() do
                local t = {
                    date = r[1],
                    facility = r[2],
                    priority = r[3],
                    message = r[4],
                }
                u.write(w.socket, (i == 0 and '' or ',\n') .. cjson.encode(t))
                i = i + 1
            end
        end

        db:close()
        u.write(w.socket, ']\n')
    end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
        ledlib:SetLED("off", "red")
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

function api.system.export_log(loop, w, r)
    local f = w.server:fork(loop, w, 10, function()
        local q = w.query
        local count = q.count or -1  -- unlimited unless specified
        local lower = q.start_time or '0000-01-01'
        local upper = q.end_time or '9999-12-31'
        local order = q.order or 'newest'
        local filename = q.filename or 'exported_log'
        filename = tostring(filename)

        assert(order == 'newest' or order == 'oldest',
            "'order' must be 'newest' or 'oldest'")

        -- KLUDGE use database.open() for versioning
        local db = sqlite3.open('/var/lib/armada/log.db')

        local header_lines = {}
        table.insert(header_lines, "HTTP/1.1 200 OK\n")
        table.insert(header_lines, "Cache-Control: no-cache\n")
        table.insert(header_lines, "Pragma: no-cache\n")
        table.insert(header_lines, "Content-Type: text/plain\n")
        table.insert(header_lines, 'Content-Disposition: attachment; ')
        table.insert(header_lines, 'filename="' .. filename .. '"\n\n')
        u.write(w.socket, table.concat(header_lines))
        
        local log_header = {}
        table.insert(log_header, "Time                              ")
        table.insert(log_header, "Component           ")
        table.insert(log_header, "Level       ")
        table.insert(log_header, "Message\n")
        u.write(w.socket, table.concat(log_header))

        local sth = db:prepare(string.format([[
            SELECT time, component, level, message
            FROM log
            WHERE time >= ? and time <= ?
            ORDER BY time %s LIMIT ?;
        ]], (order == 'newest' and 'DESC' or 'ASC')))
        local i = 0
        sth:bind_values(lower, upper, count)
        for r in sth:rows() do
            local line = {}
            local date_str = string.format("%s     ", r[1])
            table.insert(line, date_str)
            local comp_str = string.sub(r[2], 1, 20)
            table.insert(line, comp_str)
            if #comp_str < 20 then
                local comp_padding = string.rep(" ", 20 - #comp_str)
                table.insert(line, comp_padding)
            end
            local level_str = r[3]
            table.insert(line, level_str)
            if #level_str < 12 then
                local level_padding = string.rep(" ", 12 - #level_str)
                table.insert(line, level_padding)
            end
            table.insert(line, r[4])  -- log message string

            u.write(w.socket, (i == 0 and '' or ',\n') .. table.concat(line))
            i = i + 1
        end

        db:close()
        u.write(w.socket, '\n')
    end)

    if f.ok then
        w.server.close(w.server.ev_loop, w)
        ledlib:SetLED("off", "red")
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

local function get_fpga_version()
    local fh = io.open('/var/log/fpga-hardware', 'r')
    if not fh then return 'Unknown' end
    local last_line
    for line in fh:lines() do
        if line:find('FPGA') then last_line = line end
    end
    fh:close()
    if not last_line then return 'Unknown' end
    local version = last_line:match('%[FPGA%-(.+)%]')
    fh = io.popen('spi-flash-tool verify /usr/share/fpga-config/digcomm.bin')
    if fh then
        for line in fh:lines() do
            if line:find('Verified') then
                fh:close()
                return version or 'Unknown', true
            end
            break
        end
        fh:close()
    end
    return version or 'Unknown'
end

local sdm = require 'setupdb_migrate'
local expected_setup_version = sdm.get_info().expected_setup_file_version

function api.system.version(loop, w, r)
    -- append the setup version to the response
    local version = {}
    local f = io.open('/etc/version.json')
    if f then
        version = cjson.decode(f:read('*a')) or {}
        f:close()
    end
    f = io.open('/tmp/power-micro-version.json', 'r')
    version.power_micro_version = 'Unknown'
    if f then
        local t = cjson.decode(f:read('*a'))
        if t then
            version.power_micro_version = {
                running =
                    string.format("%d.%d", t.running.major, t.running.minor),
                expected =
                    string.format("%d.%d", t.expected.major, t.expected.minor),
                is_expected =
                    t.running.major == t.expected.major and
                    t.running.minor == t.expected.minor,
            }
        end
        f:close()
    end
    local running, expected = get_fpga_version()
    version.fpga_version = {
        running = running,
        is_expected = expected,
    }
    version.setup_version = expected_setup_version or 'Unknown'
    send_response(w, 200, "OK", "application/json", cjson.encode(version))
end

function api.system.setupdb_status(loop, w, r)
    local sdm = require 'setupdb_migrate'
    send_response(w, 200, "OK", "application/json",
        cjson.encode(sdm.get_info()))
end

-- API used to remove databases
function api.system.delete_databases(loop, w, r)
    send_response(w, 200, "OK", "application/json", '')
    ev.Timer.new(function(l, t, e)
            os.execute('rm -f /var/lib/armada/*.db*;sync;sync;/usr/bin/reboot')
        end, .25, 0):start(loop)
end

-- API function used to download GZIP'd tarball of databases
function api.system.customer_service_package(loop, w, r)
    local f = w.server:fork(loop, w, 10, function()

            local disp = w.query.filename
            if not disp then disp = 'customer_service_package.tgz' end
            if disp and not disp:find('%.tgz$') then
                disp = disp .. '.tgz'
            end

            local cmds = {
                {
                    heading = 'DMESG CONTENT',
                    cmd = '/bin/dmesg',
                },
                {
                    heading = 'PROCESS STATUS',
                    cmd = '/bin/ps -eaf',
                },
                {
                    heading = 'DISK SPACE',
                    cmd = '/bin/df -kh'
                },
            }

            -- File to dump stuff into
            local my_uuid = uuid.generate()
            local cs_file = 'cs-' .. my_uuid .. '.txt'
            local tar_file = 'csp-' .. my_uuid .. '.tgz'
            local fh = io.open('/var/lib/armada/' .. cs_file, 'w')

            -- Run all the commands and append the content to the file
            for _, ci in ipairs(cmds) do
                fh:write(string.rep('-', 80) .. '\n')
                fh:write(string.format('%s (%s)\n', ci.heading, ci.cmd))
                fh:write(string.rep('-', 80) .. '\n')
                local fh1 = io.popen(ci.cmd)
                if fh1 then
                    local v = fh1:read('*a')
                    fh:write(v)
                    fh1:close()
                end
            end
            fh:close()

            -- Create and send the compressed TAR archive
            local fh =
                io.popen('cd /var/lib/armada;tar -czf - *.db ' ..  cs_file)
            if fh then
                local v = fh:read('*a')
                fh:close()

                -- Send the file
                table.insert(w.out_headers,
                    'Content-Length: ' .. tostring(v:len()) .. '\n' ..
                    'Content-Disposition: attachment; filename="' ..disp.. '"\n'
                )
                send_response_headers(w, 200, "OK", "application/tgz", '')

                while true do
                    local amt, estr, err = u.write(w.socket, v)
                    if not amt then
                        if eno ~= u.EAGAIN then break end
                    elseif amt == v:len() then break
                    else v = string.sub(v, amt + 1)
                    end
                end
                C.unlink('/var/lib/armada/' .. cs_file)
            else
                send_response(w, 404, "NOT FOUND", '')
            end
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        f.error = 'Failed'
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
end

function api.system._dev_config(loop, w, r)
    local c = system_config
    -- NOTE:  Does not deal with type 'function' in subtables!!
    for k, v in pairs(c) do
        if type(v) == 'function' then
            c[k] = nil
        end
    end
    send_response(w, 200, "OK", "application/json", cjson.encode(c))
end

function api.system.get_datetime(loop, w, r)
    local dt_str = os.date("!*t")
    send_response(w, 200, "OK", "application/json", cjson.encode(dt_str))
end

function api.system.set_datetime(loop, w, r)
    local f = w.server:fork(loop, w, 10, function()

        local dt_input = assert(cjson.decode(w.content))
        -- allow for number values in date & time fields
        if type(dt_input.year) == 'number' then
            dt_input.year = string.format("%d", dt_input.year)
        end
        if type(dt_input.month) == 'number' then
            dt_input.month = string.format("%d", dt_input.month)
        end
        if type(dt_input.day) == 'number' then
            dt_input.day = string.format("%d", dt_input.day)
        end
        if type(dt_input.hour) == 'number' then
            dt_input.hour = string.format("%d", dt_input.hour)
        end
        if type(dt_input.min) == 'number' then
            dt_input.min = string.format("%d", dt_input.min)
        end
        if type(dt_input.sec) == 'number' then
            dt_input.sec = string.format("%d", dt_input.sec)
        end
        -- validate date & time input
        local field_errmsg = {}
        if not dt_input.year or #dt_input.year ~= 4 then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid year value: %s", tostring(dt_input.year))
        end
        if not dt_input.month or
            (#dt_input.month ~= 1 and #dt_input.month ~= 2) then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid month value: %s",
                tostring(dt_input.month))
        end
        if not dt_input.day or (#dt_input.day ~= 1 and #dt_input.day ~= 2) then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid day value: %s", tostring(dt_input.day))
        end
        if not dt_input.hour or
            (#dt_input.hour ~= 1 and #dt_input.hour ~= 2) then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid hour value: %s", tostring(dt_input.hour))
        end
        if not dt_input.min or (#dt_input.min ~= 1 and #dt_input.min ~= 2) then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid minutes value: %s",
                tostring(dt_input.min))
        end
        if not dt_input.sec or (#dt_input.sec ~= 1 and #dt_input.sec ~= 2) then
            field_errmsg[#field_errmsg + 1] =
                string.format("invalid seconds value: %s",
                tostring(dt_input.sec))
        end
        if #field_errmsg > 0 then
            local field_errmsg_str = table.concat(field_errmsg, ", ")
            local errmsg =
                cjson.encode{result = "error", msg = field_errmsg_str}
            send_response(w, 200, "OK", "application/json", errmsg)
            return
        end
        -- set date & time with input
        local datetime = {}
        datetime[#datetime + 1] = dt_input.year
        datetime[#datetime + 1] = "."
        if #dt_input.month < 2 then
            datetime[#datetime + 1] = "0"
            datetime[#datetime + 1] = dt_input.month
        else
            datetime[#datetime + 1] = dt_input.month
        end
        datetime[#datetime + 1] = "."
        if #dt_input.day < 2 then
            datetime[#datetime + 1] = "0"
            datetime[#datetime + 1] = dt_input.day
        else
            datetime[#datetime + 1] = dt_input.day
        end
        datetime[#datetime + 1] = "-"
        if #dt_input.hour < 2 then
            datetime[#datetime + 1] = "0"
            datetime[#datetime + 1] = dt_input.hour
        else
            datetime[#datetime + 1] = dt_input.hour
        end
        datetime[#datetime + 1] = ":"
        if #dt_input.min < 2 then
            datetime[#datetime + 1] = "0"
            datetime[#datetime + 1] = dt_input.min
        else
            datetime[#datetime + 1] = dt_input.min
        end
        datetime[#datetime + 1] = ":"
        if #dt_input.sec < 2 then
            datetime[#datetime + 1] = "0"
            datetime[#datetime + 1] = dt_input.sec
        else
            datetime[#datetime + 1] = dt_input.sec
        end
    
        local dt_str = table.concat(datetime)
        local ok = shellf("date -u -s %s", dt_str)
        if ok then
            send_response(w, 200, "OK", "application/json",
                        cjson.encode(os.date("!*t")))
        else
            send_response(w, 200, "OK", "application/json", cjson.encode{
                result = 'error',
                msg = string.format('shell command failed using string %s',
                    dt_str)
            })
        end
    end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

-- Note, can not fork this function because the HTTP response is actually
-- proxied from the channel proxy back to the requesting entity.  For whatever
-- reason if that is attempted via a forked version of this function, the
-- information never gets back to the requester and a 502 error occurs.
--function api.system.control(loop, w, r)
function api.system.control(loop, w, r)
    local cw = w
    local op = w.headers.REQUEST_METHOD
    local control = assert(cjson.decode(w.content))

    local connector = control.connector:gsub("^.-@", "")
    local host, port, ip =
        w.server.network:ResolveService(connector)
    local emsg = string.format("Can not resolve service for '%s'",
        control.connector or 'UNSPECIFIED CONNECTOR')
    if host then
        local cb_w = w.server:connect_nb(ip, port,
            function(loop, w)
                local got = u.read(w.socket, 8192)
                if not got or got == "" then
                    w.server.close(loop, w)
                    cw.server.close(loop, cw)
                else
                    u.write(cw.socket, got)
                end
            end,
            function(loop, cbs)
                if not cbs then
                    send_response(w, 200, 'OK', 'application/json',
                        cjson.encode{
                            result = 'error',
                            msg = 'Connect to remote server failed'
                        })
                    return
                end
                local json = cjson.encode(control) .. "\n"
                local req = "POST /control/ HTTP/1.0\r\n" ..
                    "Content-type: application/json\r\n" ..
                    "Content-length: " .. #json .. "\r\n" ..
                    "\r\n" ..  json
                u.write(cbs.socket, req)
                return
            end)
        if cb_w then return end
        emsg = string.format("NB connect to %s failed", control.connector)
    end
    local f = {
        result = 'error',
        msg = emsg,
    }
    send_response(w, 404, "NOT FOUND", "application/json", cjson.encode(f))
end

function api.system.unlock_start(loop, w, r)
    start_is_locked = nil
    local f = {
        result = 'success',
        msg = 'Test start is unlocked'
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

function api.system.cancel_restart(loop, w, r)
    os.execute('killall -SIGUSR1 restart-tests')
    local f = {
        result = 'success',
        msg = 'Restart of interrupted test cancelled',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
    start_is_locked = nil
end

function api.system.acknowledge_notify(loop, w, r)
    -- Make sure escape crap is removed!
    w.content = string.gsub(w.content, '\\"', '"')
    w.content = string.gsub(w.content, '\\', '')

    -- Decode the JSON
    local cj = cjson.decode(w.content)
    if not cj then
        local f = {
            result = 'error',
            msg = 'Unable to decode JSON content',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    if not cj.id then
        local f = {
            result = 'error',
            msg = 'JSON content does not contain id field',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    local owner = '_system-notification-' .. cj.id

    local action = w.query.action
    if action then
        if action == 'cancel_restart' then
            os.execute('killall -SIGUSR1 restart-tests')
        end
    end

    -- Remove all notifications found with the given owner
    w.server.network:RemoveLocalRROwner(owner)

    -- Send success
    local f = {
        result = 'success',
        msg = 'OK',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

function api.system.update_firmware(loop, w, r)
    -- KLUDGE allow PUT here as well
    local mp = assert(multipart_file(w, r))
    assert(shellf("/bin/gunzip -t '%s'", mp.tmp_filename))
    assert(shellf("/bin/mkdir -p /var/cache/update"))
    if mp.filename:match('\\') then
        mp.filename = mp.filename:match('^.+\\(.+%..+)') or mp.filename
    end
    assert(shellf("/bin/cp -a '%s' /var/cache/update/%s",
                  mp.tmp_filename, mp.filename or '__ng-update.tar.gz'))
    assert(shellf("/bin/mv '%s' /__ng-update.tar.gz", mp.tmp_filename))
    ev.Timer.new(function() shellf("/usr/bin/reboot") end, 1, 0):start(loop)
    send_response(w, 200, "OK", "text/plain", '{"ok":true}\n')
end

function api.system.reboot(loop, w, r)
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    ev.Timer.new(function() shellf("/usr/bin/reboot") end, 1, 0):start(loop)
end

local function available_channels(loop, w, r)
    local net_db = network_db:new()
    net_db:begin_transaction()
    local channels = net_db:available_channels()
    net_db:commit()
    net_db:close()
    local ss = setup_spec:new(channels)
    local result = ss:export()
    send_response(w, 200, "OK", "application/json", result)
end

function api.system.available_channels(loop, w, r)
    local f = w.server:fork(loop, w, 10,
        function() available_channels(loop, w, r) end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

local function get_unique_name(db, name)
    if not name then
        return nil
    end
    local unique_name
    local last = 0
    local found = 0
    -- Create SQL statement
    local find_setups = db:prepare(string.format([[
        SELECT content
        FROM "%s"
        WHERE id = 'summary' AND deleted = '';
    ]], 'tsd_setups'))
    -- Run SQL statement
    local got = find_setups:exec{}
    if type(got) == 'table' and #got > 0 then
        local t = {}
        -- Does exact 'name' exist?
        for i = 1, #got do
            if got[i].content then
                if got[i].content:match('"test_name":"(.-)",.+') == name then
                    found = found + 1
                end
            end
        end
        -- Does exact 'name_nnn' exist?
        for i = 1, #got do
            if got[i].content then
                local _name = got[i].content:match('"test_name":"(.-)",.+')
                for j = 1, 999 do
                    if _name == string.format("%s_%03d", name, j) then
                        last = j
                    end
                end
            end
        end
        if found > 0 then
            unique_name = string.format("%s_%03d", name, last + 1)
        else
            unique_name = name
        end
    end
    -- Cleanup SQL statement
    find_setups:finalize()
    return unique_name
end

local function get_extension(filename)
    if filename then
        return filename:match('.+%.(.+)$') or filename
    end
end

local function extract_setup_name(filename)
    if filename then
        if filename:match('\\') then
            return filename:match('^.+\\(.+)%..+') or filename
        end
        return filename:match('(.+)%..+')
    end
end

function api.test.import(loop, w, r)
    local ftype = w.query.type
    local origin = w.query.origin
    local extended =
        w.query.extended_names and ' enable_extended_signal_names' or ''

    if (ftype ~= "sxr" and ftype ~= "dbc" and ftype ~= "sdb") then
        log:error('Invalid query field "type", must be "sxr", "dbc", or '..
            '"sdb", got: '..tostring(ftype)..'"')
        send_response(w, 500, "ERROR", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = 'Invalid query field "type", must be "sxr", "dbc", or '..
                    '"sdb", got: '..tostring(ftype)..'"',
            })
        return
    elseif (origin ~= "database" and origin ~= "setup") then
        log:error('Invalid query field "origin", must be "database" or '..
            '"setup", got: "'..tostring(origin)..'"')
        send_response(w, 500, "ERROR", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = 'Invalid query field "origin", must be "database" or '..
                    '"setup", got: "'..tostring(origin)..'"',
            })
        return
    end
    if origin == "setup" and ftype ~= "sxr" then
        log:warn("Invalid setup. File must be of type *.sxr.")
        send_response(w, 200, "OK", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = "Invalid setup. "..
                "File must be of type *.sxr."
            })
        return
    end
    local filename
    r.query = r.query or {}
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                local fh = io.open(mp.tmp_filename)
                w.content = fh:read('*a')
                fh:close()
                filename = mp.filename
                if ftype == 'dbc' or ftype == 'sdb' then
                    assert(shellf("mv %s '/tmp/%s'", mp.tmp_filename, filename))
                    if ftype == 'dbc' then
			assert(shellf("dbc2sxr '/tmp/%s'%s",
                            filename, extended))
		    else
			assert(shellf("mdb-translation '/tmp/%s'", filename))
		    end
                    fh = io.open(string.format("/tmp/%s.sxr",
                        extract_setup_name(filename)))
                    if fh then
                        w.content = fh:read('*a')
                        fh:close()
                    end
                end
                break
            end
        end
    end
    filename = filename or r.query.filename or 'imported_file'
    local migration_result
    do
        r.id = r.id or uuid.generate()
        local ss = setup_spec:new()
        local ok, setup = pcall(ss.import, ss, w.content)
        if not ok or not setup then
            log:warn('Unable to read SXR file content.\n' ..
                        'File content may be corrupt')
            send_response(w, 500, "ERROR", "application/json",
                cjson.encode{
                    setup_id = r.id,
                    msg = 'Unable to read SXR file content.\n' ..
                        'File content may be corrupt',
                })
            return
        end
        local ok, err = pcall(ss.validate_structure, ss)
        if not ok then
            log:warn('SXR file content validation failure.\n' ..
                        'File content may be corrupt')
            send_response(w, 500, "ERROR", "application/json",
                cjson.encode{
                    setup_id = r.id,
                    msg = 'SXR file content validation failure.\n' ..
                        'File content may be corrupt',
                })
            return
        end
        if ftype == 'sxr' then
            local ok, err = pcall(ss.user_order, ss)
            if not ok then
                --"Changing to text/plain fixes IE
                send_response(w, 200, "OK", "text/plain",
                    cjson.encode{
                        setup_id = r.id,
                        msg = err,
                    })
                return
            end
        end
        local summary
        for i = 1, #setup do
            local item = setup[i]
            -- KLUDGE this next bit is disconcerting
            if item.id == 'summary' or item.category == 'summary' then
                item.id = 'summary'
                item.category = 'summary'
                item.setup_id = r.id
                summary = item
            else
                item.id = uuid.generate()
            end
        end
        if not summary then
            summary = {
                setup_id = r.id,
                id = 'summary',
                category = 'summary',
            }
            for k, v in pairs{
                test_author = 'Default Author', -- KLUDGE
                test_title = 'Default Title', -- KLUDGE
                test_synopsis = 'Default Synopsis', -- KLUDGE
            } do
                summary[k] = r.query[k] or v
            end
            table.insert(setup, 1, summary)
        end

        -- Migrate the setup from it's source version to the version needed on
        -- this machine.
        setup, migration_result =
            setup_migrator.migrate_setup(summary.setup_id, setup)

        local test_name = extract_setup_name(filename) or summary.test_name
        summary.test_name = get_unique_name(w.server.setup_db, test_name)

        local delta = {old={}, new={}}
        for i = 1, #setup do
            delta.old[setup[i].id] = cjson.null
            delta.new[setup[i].id] = setup[i]
        end
        w.content = cjson.encode(delta)
    end
    local frag = string.format('/setups/%s/_delta', r.id)
    if not r.query.description then
        r.query.description = string.format('Import "%s"', filename)
    end
    w.api_internal.import = true
    w.headers.REQUEST_METHOD = 'POST' -- KLUDGE to make tracked_db happy
    local got = handlers.tracked_db(loop, w, frag)
    local ok, result = got.ok, got.result
    if ok then
        result = cjson.encode{
            setup_id = r.id,
            migration_result = migration_result or 'Migration failed\n',
        }
    end
    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "text/plain", result)
end

function api.test.export(loop, w, r)
    local sie_name = r.id .. '.sxr'
    local disp = r.filename and r.filename or sie_name
    w.api_internal.export = true
    local frag = string.format('/setups/%s/_all_docs', r.id)
    table.insert(w.out_headers,
        'Content-Disposition: attachment; filename="' .. disp .. '"\n'
    )
    return handlers.tracked_db(loop, w, frag)
end


function api.test.start_nosie(loop, w, r)
    -- KLUDGE error on w.server.running_tests_kludge[r.id]
    local setup_name, fd = w.server:temp_file(w, '/tmp/nginx/test_', '', true)
    local compiled_name = setup_name .. '.lua'
    local sie_name = '/dev/null'
    local spec = setup_spec:new(cjson.decode(w.content))
    local setup = spec:export()
    assert(C.write(fd, setup, #setup) == #setup)
    assert(C.close(fd) == 0)

    if system_config.null_sie_name then
        sie_name = string.format('%s/%s.sie',
            storage_path, system_config.null_sie_name)
    end

    local got
    local fork_settings = {}
    fork_settings.cpu_limit = math.ceil(2 + #spec.setup * 0.5) -- KLUDGE guess
    local compiler_settings = w.query and w.query or {}
    got = w.server:fork(loop, w, fork_settings, '/usr/bin/setup-compiler',
        '-o', compiled_name, '-s', cjson.encode(compiler_settings), setup_name)
    if not got.ok then
        got.error_message = "Couldn't compile test setup" -- KLUDGE i18n!!!
        send_response(w, 500, "Error", "application/json", cjson.encode(got))
        error() -- KLUDGE for testing
    end

    local function add(t, a)
        if a then
            if type(a) == 'table' then
                for _, x in ipairs(a) do
                    add(t, x)
                end
            else
                t[#t + 1] = a
            end
        end
    end

    local cmd = {}
    add(cmd, w.query.strace and {
        '/usr/bin/strace',
        '-s1024',
        '-x',
        string.format('-o%s.strace', setup_name),
    })
    add(cmd, {
        '/usr/bin/luajit',
        w.query.jdump and string.format('-jdump=+rsx,%s.dump', setup_name)
            or false,
        '--',
    })
    add(cmd, {
        compiled_name,
        sie_name,
        w.query.duration or 'null',
        r.id,
    })

    local test_pid = w.server.master:Spawn(cmd[1], cmd)

    -- KLUDGE use a coro timer "sleep" construct here instead of forking
    -- KLUDGE constant time won't work for very short-lived tests
    got = w.server:fork(loop, w, 0, function() C.sleep(4) os.exit(0) end)
    assert(got.exit_status == 0)

    got = C.kill(test_pid, 0)
    if got ~= 0 then
        got = {
            error_message = "Test failed to start"
        }
        send_response(w, 500, "Error", "application/json", cjson.encode(got))
        error() -- KLUDGE for testing
    end
    w.server.running_tests_kludge[r.id] = test_pid
    send_response(w, 200, "OK", "application/json", cjson.encode{
        ok = true,
        test_id = r.id,
        test_sie_file = '',
    })
end

function api.test._start_new_te(loop, w, r)
    -- KLUDGE error on w.server.running_tests_kludge[r.id]
    local setup_name, fd = w.server:temp_file(w, '/tmp/nginx/test_', '', true)
    local spec = setup_spec:new(cjson.decode(w.content))
    local summary
    for _, stanza in ipairs(cjson.decode(w.content)) do
        if stanza.category and stanza.category == 'summary' then
            summary = stanza
            break
        end
    end
    
    local setup = nil
    local ok, err = pcall(function() setup = spec:export('compiler_order') end)
    if not ok then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({ok=ok, error_message=err}))
        return
    end
    assert(C.write(fd, setup, #setup) == #setup)
    assert(C.close(fd) == 0)

    if not w.query.friendly_name then
        w.query.friendly_name = r.id
    end

    -- Make a UUID...E.G.  Ignore UUID from user
    r.id = uuid.generate()

    local friendly_name = 'dummy_name'
    local sie_name = '/dev/null'

    if not summary.purpose then
        friendly_name = w.query.friendly_name or r.id
        sie_name = string.format('%s/%s.sie', storage_path, r.id)
        if friendly_name and w.query.append then
            sie_name = w.server:get_sie_path_on_disk(friendly_name) or sie_name
        else
            friendly_name =
                w.server:assign_friendly_sie_name(friendly_name, sie_name)
        end
    end

    -- If this is NOT a test restart, reset ALL of the interactive triggers
    if not w.query.restart then
        -- initialize all triggers to false
        local owner
        for trig_i = 1, 8 do
            owner = '_interactive-trigger-' .. tostring(trig_i)
            w.server.network:RemoveLocalRROwner(owner)
            w.server.network:ModifyRRs({}, {{owner, 'TXT', owner ..
                ":" .. "false"}})
        end
    end

    local function add(t, a)
        if a then
            if type(a) == 'table' then
                for _, x in ipairs(a) do
                    add(t, x)
                end
            else
                t[#t + 1] = a
            end
        end
    end

    local cmd = {}
    add(cmd, w.query.strace and {
        '/usr/bin/strace',
        '-s1024',
        '-x',
        string.format('-o%s.strace', setup_name),
    })
    add(cmd, {
        '/usr/bin/run-test',
        setup_name,
        sie_name,
        w.query.duration or summary.duration or 'null',
        r.id,
        friendly_name,
        w.query.restart or 'null',
    })

    -- OK spawn the test
    local test_pid = w.server.master:Spawn(cmd[1], cmd)

    -- Defer the reponse for 1 second.
    ev.Timer.new(function(l, t, e)
            -- Is the test running?
            local got = C.kill(test_pid, 0)
            if got ~= 0 then
                -- No send error
                send_response(w, 200, "OK", "application/json", cjson.encode{
                    ok = false,
                    error_message = __'Test failed to start',
                    test_id = r.id,
                    test_sie_file = friendly_name,
                })
                return
            end

            -- Test is running, send success
            w.server.running_tests_kludge[r.id] = test_pid
            send_response(w, 200, "OK", "application/json", cjson.encode{
                ok = true,
                test_id = r.id,
                test_sie_file = friendly_name,
            })
        end, 1, 0):start(loop)
end

function api.test._start(loop, w, r)

    -- If the system is out of SIE storage space, don't allow test to start
    if fs_info.get_storage_available() < (2048*1024) then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            ok = false,
            msg =
                "Can't start test because the system is out of storage space.",
        })
        return
    end

    if not system_config.dont_use_new_test_engine then
        return api.test._start_new_te(loop, w, r)
    end
    if r.id and r.id == 'null' then
        api.test.start_nosie(loop, w, r)
        test_start_in_progress = nil
        return
    end

    -- KLUDGE error on w.server.running_tests_kludge[r.id]
    local setup_name, fd = w.server:temp_file(w, '/tmp/nginx/test_', '', true)
    local spec = setup_spec:new(cjson.decode(w.content))
    local setup = spec:export()
    assert(C.write(fd, setup, #setup) == #setup)
    assert(C.close(fd) == 0)

    local compiled_name = setup_name .. '.lua'

    if not w.query.friendly_name then
        w.query.friendly_name = r.id
    end

    -- Make a UUID...E.G.  Ignore UUID from user
    r.id = uuid.generate()

    local friendly_name = w.query.friendly_name or r.id

    local sie_name = string.format('%s/%s.sie', storage_path, r.id)
    if friendly_name and w.query.append then
        sie_name = w.server:get_sie_path_on_disk(friendly_name) or sie_name
    else
        friendly_name =
            w.server:assign_friendly_sie_name(friendly_name, sie_name)
    end

    -- OK now fire up the setup compiler
    local got
    local fork_settings = {}
    fork_settings.cpu_limit = math.ceil(2 + #spec.setup * 0.5) -- KLUDGE guess
    local compiler_settings = w.query and w.query or {}
    got = w.server:fork(loop, w, fork_settings, '/usr/bin/setup-compiler',
        '-o', compiled_name, '-s', cjson.encode(compiler_settings), setup_name)

    -- We don't get here until the set compiler exits regardless of whether it
    -- exits successfully or not.
    if not got.ok then
        got.error_message = "Couldn't compile test setup" -- KLUDGE i18n!!!
        send_response(w, 500, "Error", "application/json", cjson.encode(got))
        test_start_in_progress = nil
        error() -- KLUDGE for testing
    end

    -- initialize all triggers to false
    local owner
    for trig_i = 1, 8 do
        owner = '_interactive-trigger-' .. tostring(trig_i)
        w.server.network:RemoveLocalRROwner(owner)
        w.server.network:ModifyRRs({}, {{owner, 'TXT', owner ..
            ":" .. "false"}})
    end

    local function add(t, a)
        if a then
            if type(a) == 'table' then
                for _, x in ipairs(a) do
                    add(t, x)
                end
            else
                t[#t + 1] = a
            end
        end
    end

    local cmd = {}
    add(cmd, w.query.strace and {
        '/usr/bin/strace',
        '-s1024',
        '-x',
        string.format('-o%s.strace', setup_name),
    })
    add(cmd, {
        '/usr/bin/luajit',
        w.query.jdump and string.format('-jdump=+rsx,%s.dump', setup_name)
            or false,
        '--',
    })
    add(cmd, {
        compiled_name,
        sie_name,
        w.query.duration or 'null',
        r.id,
        friendly_name,
        w.query.restart or 'null',
    })

    -- OK spawn the test
    local test_pid = w.server.master:Spawn(cmd[1], cmd)

    -- KLUDGE use a coro timer "sleep" construct here instead of forking
    -- KLUDGE constant time won't work for very short-lived tests
    got = w.server:fork(loop, w, 0, function() C.sleep(4) os.exit(0) end)
    assert(got.exit_status == 0)

    got = C.kill(test_pid, 0)
    if got ~= 0 then
        got = {
            error_message = "Test failed to start"
        }
        send_response(w, 500, "Error", "application/json", cjson.encode(got))
        error() -- KLUDGE for testing
    end
    w.server.running_tests_kludge[r.id] = test_pid
    send_response(w, 200, "OK", "application/json", cjson.encode{
        ok = true,
        test_id = r.id,
        test_sie_file = friendly_name,
    })
end

function api.test.start(loop, w, r)
    -- If start is locked due to interrupted test restart in progress, deny the
    -- start request
    if start_is_locked then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                msg =
                    'Restart of test stopped by system shutdown is ' ..
                    'in progress.  Click "cancel" if you want to abort ' ..
                    'interrupted test restart and enable user initiated ' ..
                    'test start.',
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Interrupted Test Restart In Progress',
                buttons = {
                    {
                        text='Cancel',
                        uri='/api/system/cancel_restart',
                    }
                },
            }))
        return
    end

    -- If a test is already running in this system, deny the start request.
    -- FIXME:  This is system wide...so test start will be denied if a a test
    -- is running on ANY of the devices that are interconnected.
    if w.server.running_tests and w.server.running_tests > 0 then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                msg =
                    "A test is already running in this system.  This " ..
                    "is sometimes a result of clicking 'start' in rapid " ..
                    "succession.  The currently running test must be " ..
                    "stopped before this operation can be performed.",
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Test Engine Busy',
            }))
        return
    end

    -- If a test start is in progress, deny the start request
    if w.server.test_start_in_progress then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                msg =
                    "A test is in the process of starting on this system.  " ..
                    "This is sometimes a result of clicking 'start' in " ..
                    "rapid succession.  The currently startng test must be " ..
                    "stopped before this operation can be performed.",
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Test Engine Busy',
            }))
        return
    end

    w.server.test_start_in_progress = true
    pcall(api.test._start, loop, w, r)
    -- Error or not gets us back to here
    w.server.test_start_in_progress = nil
end

function api.test.restart(loop, w, r)
    return api.test._start(loop, w, r)
end

function api.test.stop(loop, w, r)
    -- KLUDGE until we have stuff hooked in for real
    local test_pid = w.server.running_tests_kludge[r.id]
    w.server.running_tests_kludge[r.id] = nil
    local err = nil
    if not test_pid then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'No such test',
        })
        return
    end
    local got = system_config.force_brutal_stop and 0 or
        C.kill(test_pid, u.SIGINT)
    if got ~= 0 then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'The test was not running when test stop was attempted',
        })
    else
        -- Defer the response until the test actually stops
        local timeout = 10 * 20
        ev.Timer.new(function(l, t, e)
                if C.kill(test_pid, 0) ~= 0 then
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            result = 'success',
                            msg = 'Test stopped',
                        })
                    t:stop(l)
                    return
                end
                timeout = timeout - 1
                if timeout == 0 then
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            result = 'error',
                            msg = 'Test would not stop, killed!',
                        })
                    t:stop(l)
                    -- Be brutal about getting rid of the test!
                    log:log(log.ERR, "Timeout waiting for test to exit, killed")
                    C.kill(test_pid, u.SIGKILL)

                    -- Clean up the database
                    w.server.network:RemoveLocalRROwner("_restart-armada-tests")
                end
            end, .1, .1):start(loop)
        end
end

function api.test.data(loop, w, r)
    local disp = (r.filename or r.id)
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = w.server:get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    table.insert(w.out_headers,
        'Content-Disposition: attachment; filename="' .. disp .. '"\n' ..
        'X-Accel-Redirect: ' .. sie_name .. '\n'
    )
    send_response(w, 200, "OK", "application/sie", '')
end

function api.test.rename(loop, w, r)
    local old_name = w.query.old_name
    local new_name = w.query.new_name

    if not old_name or not new_name then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'fail',
                msg = 'Both old and new name parameters must be given',
            }))
        return
    end

    local name = w.server:rename_sie(new_name, old_name)
    if name then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'success',
                msg = 'Name changed',
                name = name,
            }))
    else
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'fail',
                msg = 'Unable to change name',
            }))
    end
end

function api.test.data_file_list(loop, w, r)
    send_response(w, 200, "OK", "application/json",
        cjson.encode(w.server:get_sie_file_list()))
end

function api.system.get_cal_cert(loop, w, r)
    local f = w.server:fork(loop, w, 10, function()
        local filename
        local fp = io.popen('ls /var/etc/armada/CX23R*|sort -r', 'r')
        if fp then
            for line in fp:lines() do
                filename = line
                break
            end
            fp:close()
            if not filename or filename == '' then
                send_response(w, 200, "OK", "text/plain", '{"ok":false}\n')
                return
            end
            local cert = io.open(filename, 'r')
            if cert then
                local header_lines = {}
                table.insert(header_lines, "HTTP/1.1 200 OK\n")
                table.insert(header_lines, "Cache-Control: no-cache\n")
                table.insert(header_lines, "Pragma: no-cache\n")
                table.insert(header_lines, "Content-Type: application/pdf\n")
                table.insert(header_lines, 'Content-Disposition: attachment; ')
                table.insert(header_lines, 'filename="' ..
                                           filename:match('^.+/(.+%..+)') ..
                                           '"\n\n')
                u.write(w.socket, table.concat(header_lines))
                u.write(w.socket, cert:read('*a'))
                u.write(w.socket, '\n\n')
                cert:close()
            end
        else
            send_response(w, 200, "OK", "text/plain", '{"ok":false}\n')
            return
        end
    end)
    if f.ok then
        send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

function api.system.set_cal_cert(loop, w, r)
    local mp = multipart_file(w, r)
    if mp and mp.filename and mp.tmp_filename then
        if mp.filename:match('\\') then
            mp.filename = mp.filename:match('^.+\\(.+%..+)') or mp.filename
        end
        shellf("/bin/mv '%s' /var/etc/armada/%s", mp.tmp_filename, mp.filename)
        send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    else
        send_response(w, 500, "ERROR", "application/json",
                      '{"error_message":"Incomplete multipart file!"}\n')
    end
end

function api.system.read_local_file(loop, w, r)
    local uploadFileName = "read_local_file"
    if w.query.uploadFileName then
        uploadFileName = w.query.uploadFileName
    end
    local filename
    r.query = r.query or {}
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                local fh = io.open(mp.tmp_filename)
                w.content = fh:read('*a')
                fh:close()
                filename = uploadFileName or mp.filename
                assert(shellf("mv %s '/tmp/%s'", mp.tmp_filename, filename))
                fh = io.open(string.format("/tmp/%s", filename))
                if fh then
                    w.content = fh:read('*a')
                    fh:close()
                end
            end
        end
    end
    send_response(w, 200, "OK", "text/plain", w.content)
end

function api.system.upload_local_file(loop, w, r)
    local result = false
    local uploadFileName = "upload_local_file"
    if w.query.uploadFileName then
        uploadFileName = '/tmp/' .. w.query.uploadFileName
    end
    local f = io.open(uploadFileName, 'w')
    if f then
        f:write(w.content)
        f:close()
        result = true
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({result=result}))
end

function api.system.download_local_file(loop, w, r)
    local downloadFileName = "download_local_file"
    if w.query.downloadFileName then
        downloadFileName = w.query.downloadFileName
    end
    local filename = '/tmp/' .. downloadFileName
    local f = io.open(filename, 'r')
    if f then
        local header_lines = {}
        table.insert(header_lines, "HTTP/1.1 200 OK\n")
        table.insert(header_lines, "Cache-Control: no-cache\n")
        table.insert(header_lines, "Pragma: no-cache\n")
        table.insert(header_lines, "Content-Type: text/plain\n")
        table.insert(header_lines, 'Content-Disposition: attachment; ')
        table.insert(header_lines, 'filename="' .. downloadFileName .. '"\n\n')
        u.write(w.socket, table.concat(header_lines))
        u.write(w.socket, f:read('*a'))
        u.write(w.socket, '\n\n')
        f:close()
    end
    w.server.close(w.server.ev_loop, w)
end

function api.system.set_inter_trig(loop, w, r)
    -- Make sure escape crap is removed! (quoting MikeL)
    w.content = string.gsub(w.content, '\\"', '"')
    w.content = string.gsub(w.content, '\\', '')

    -- Decode the JSON
    local cj = cjson.decode(w.content)
    if not cj then
        local f = {
            result = 'error',
            msg = 'Unable to decode JSON content',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    if not cj.trigger_index then
        local f = {
            result = 'error',
            msg = 'JSON content does not have trigger id',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end

    if not cj.value or (cj.value ~= 'true' and cj.value ~= 'false') then
        local f = {
            result = 'error',
            msg = 'JSON content does not have legal trigger value',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end

    -- set interactive trigger in database to specified value
    local owner = '_interactive-trigger-' .. cj.trigger_index
    w.server.network:RemoveLocalRROwner(owner)
    w.server.network:ModifyRRs({}, {{owner, 'TXT', owner ..
        ":" .. cj.value}})
    -- Send success
    local f = {
        result = 'success',
        msg = 'OK',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

function handlers.test(loop, w, frag)
    local method = w.headers.REQUEST_METHOD
    local tomatch = w.headers.REQUEST_METHOD .. ':' .. w.headers.DOCUMENT_URI
    local r = api_url_parser:match(tomatch)

    --[[
    dbg('handlers.test(): %s\n', tomatch)
    dbg('query_string = %s\n', tostring(w.headers.QUERY_STRING))
    dbg('content_type = %s\n', tostring(w.headers.CONTENT_TYPE))
    dbg("r=%s\n", (require 'data_dumper')(r))
    dbg("content=%s\n", tostring(w.content))
    dbg('==============================\n\n')
    --]]

    if not r then
        return send_response(w, 404, "Not Found", "text/plain",
            "No such URL: %s\n", w.headers.DOCUMENT_URI)
    end
    r.action = r.action or r.method:lower()
    assert(r.prefix, 'prefix required')
    api[r.prefix][r.action](loop, w, r)
end

handlers.system = handlers.test
handlers.sie = handlers.test
handlers.session = handlers.test

function m.handler(loop, w)
    local uri = w.headers.DOCUMENT_URI
    local unp
    for _, uri_unprefix_match in ipairs(uri_unprefix_matches) do
        unp = uri:match(uri_unprefix_match)
        if unp then
            break
        end
    end
    local req = nil
    local frag = ""
    local req_name = "" -- KLUDGE

    if unp then
        req = handlers[unp]
        if req then
            req_name = unp
        else
            -- Find the longest matching request
            for k, v in pairs(handlers) do
                if #k > #req_name and #unp >= #k and k == unp:sub(1, #k) then
                    req_name = k
                    req = v
                end
            end
            if req then
                frag = unp:sub(#req_name + 1)
                local f1 = string.sub(frag, 1, 1)
                if f1 ~= '' and f1 ~= '/' and f1 ~= '?' then
                    req = nil
                end
            end
        end
    end

    if not w.authorized then
        local addr = w.headers.REMOTE_ADDR
        if (addr == "::ffff:127.0.0.1" or addr == "127.0.0.1")
                and w.authorization_key == local_authorization_key then
            w.authorized = true -- Special case for initial database loads
        else
            w.authorization_key = nil
            w.session_key = nil
            w.authorized = always_authorized_handlers[req_name]
        end
    end

    w.out_headers = {}
    if req and w.authorized then
        w.coro = coroutine.create(function() req(loop, w, frag) end)
        local ok, err = coroutine.resume(w.coro)
        if not ok then
            pcall(function() w.server.setup_db:exec("ROLLBACK;") end)
            send_response(w, 500, "Server Error", "text/plain", err)
        end
    else
        send_response(w, 404, "Not Found", "text/plain",
            "No such URL: %s\n", w.headers.DOCUMENT_URI)
    end
end

return m
