--
-- Armada Memory Mapped I/O access functions
--
-- Memory mapped through the 1M block at 0xE0000000 corresponds to the
-- PCI mapped space inside the PowerPC microprocessor.
--

local u = require 'unix'
local b = require 'bit'
local bits = require 'bits'
local s = require 'struct'
local us = require 'underscore'
local ev = require 'ev'
local sample_clock = require 'sample_clock'
local loop = ev.Loop.default

local armada_mmio = {}

local config = require 'system_config'
local tcfg = config.timing

local fd = assert(u.open('/dev/mem', u.O_RDWR))
local map = assert(u.mmap(0x100000, fd, 0xe0000000))
assert(u.close(fd))

local calibration = { si514 = { offset = 0 } }

local function sleep(t)
    ev.Timer.new(function() loop:unloop() end, t, 0):start(loop)
    loop:loop()
end

local function offset_ppb(base, ppb)
    return base * (1 + ppb / 1000000000)
end

function armada_mmio:oscillator_offset(offset)
    local cjson = require 'cjson.safe'
    local fname = '/rootfs/root/hbm/calibration.json'
    os.execute('mkdir -p /rootfs')
    if not offset then
        os.execute('mount -t jffs2 /dev/mtdblock2 /rootfs -o,ro')
        local fh = io.open(fname, 'r')
        if fh then
            calibration = cjson.decode(fh:read("*a"))
            fh:close()
            if not calibration then
                self.log:log(self.log.ERR,
                    "Unable to decode oscillator calibration file - " ..
                    "unit may be out of calibration")
            end
        else
            self.log:log(self.log.ERR,
                "Oscillator calibration not found - " ..
                "unit may be out of calibration")
        end
    else
        os.execute('mount -t jffs2 /dev/mtdblock2 /rootfs')
        os.execute('mkdir -p /rootfs/root/hbm')
        calibration.si514.offset = offset
        local fh = io.open(fname .. '.new', 'w+')
        if fh then
            fh:write(cjson.encode(calibration))
            fh:close()
            os.execute('mv -f ' .. fname .. ' ' .. fname .. '.old')
            os.execute('mv -f ' .. fname .. '.new ' .. fname)
        else
            self.log:warning('armada_mmio:oscillator_offset(): ' ..
                        'could not write ' .. fname)
        end
    end
    os.execute('umount /rootfs')
    return (calibration and calibration.si514 and calibration.si514.offset) or 0
end

-- Apply the oscillitor characterization.  NOTE:  This assumes the value in the
-- characterization file is a difference between the measured number of ticks
-- per second and the desired number of ticks per second (40,000,000).
function armada_mmio:apply_oscillator_characterization()
    self.characterization_ppb = 0
    local ppb = math.floor((self:oscillator_offset() or 0) / 40 * 1000)
    self:adjust_oscillator(ppb)
    self.characterization_ppb = ppb
end

function armada_mmio:initialize(log)
    os.execute('/sbin/rmmod gianfar_clocksource')

    self.log = log
    self.fd = assert(u.open(tcfg.i2c_dev, u.O_RDWR))

    self.characterization_ppm = 0

    if tcfg.oscillator == 'Si598' then
        -- reset si598
        local msgs = {
            { addr = 0x55, write = s.pack('BB', 135, 0x01) },
        }
        assert(u.i2c_rdwr(self.fd, msgs))
        sleep(.100)

        local out_freq = sample_clock.master_rate

        -- read out si598 defaults
        local msgs = {
            { addr = 0x55, write = s.pack('B', 7) },
            { addr = 0x55, read = 6 },
            { addr = 0x55, write = s.pack('B', 135) },
            { addr = 0x55, read = 1 },
            { addr = 0x55, write = s.pack('B', 137) },
            { addr = 0x55, read = 1 },
        }
        local r7, r135, r137 = assert(u.i2c_rdwr(self.fd, msgs))
        local b1, b2, rest = s.unpack('>BBc4', r7)

        local hs_div_i = b.rshift(b1, 5)
        local n1 = b.bor(b.lshift(b.band(b1, 0x1f), 2), b.rshift(b2, 6))
        local rfreq = s.unpack('>I5', s.pack('B', b.band(b2, 0x3f)) .. rest)

        local to_hs_div =
            { [0] = 4, [1] = 5, [2] = 6, [3] = 7, [5] = 9, [7] = 11 }
        local from_hs_div =
            { [4] = 0, [5] = 1, [6] = 2, [7] = 3, [9] = 5, [11] = 7 }

        local f_dco = out_freq * to_hs_div[hs_div_i] * (n1 + 1)
        local f_xtal = f_dco / (rfreq / 268435456)

        function armada_mmio:adjust_oscillator(ppb)
            if ppb > 3000000 then ppb = 3000000 end
            if ppb < -3000000 then ppb = -3000000 end
            local new_rfreq = math.floor(offset_ppb(rfreq, ppb) + 0.5)
            local b2, rest = s.unpack('>Bc4', s.pack('>I5', new_rfreq))
            b2 = b.bor(b.lshift(n1, 6), b.band(b2, 0x3f))
            local w2 = s.pack('Bc4', b2, rest)
            local msgs = {
                { addr = 0x55, write = s.pack('BB', 135, 0x20) },
                { addr = 0x55, write = s.pack('Bc5', 8, w2) },
                { addr = 0x55, write = s.pack('BB', 135, 0x00) },

                { addr = 0x55, write = s.pack('B', 8) },
                { addr = 0x55, read = 5 },
            }
            local rb2, err = u.i2c_rdwr(self.fd, msgs)
            if not rb2 then
                u.close(self.fd)
                self.log:error('armada_mmio:adjust_oscillator(): ' ..
                               'i2c_rdwr - ' .. err)
            elseif rb2 ~= w2 then
                u.close(self.fd)
                self.log:error('armada_mmio:adjust_oscillator(): ' ..
                               'wrote ' .. bits.tohex(w2) .. ' ' ..
                               'read back ' .. bits.tohex(rb2))
            end
            return ppb
        end
    elseif tcfg.oscillator == 'Si514' or not tcfg.oscillator then

        local function reset_oscillator()
            local msgs = {
                { addr = 0x55, write = s.pack('BB', 128, 0x80) },
            }
            u.i2c_rdwr(self.fd, msgs)
            sleep(.100)
            local msgs = {
                { addr = 0x55, write = s.pack('B', 5) },
                { addr = 0x55, read = 5 },
            }
            local regs = assert(u.i2c_rdwr(self.fd, msgs))
            return s.unpack('<I5', regs)
        end

        local m = reset_oscillator()

        function armada_mmio:oscillator_write(ppb)
            local new_m = math.floor(offset_ppb(m, ppb) + 0.5)
            local pack_m = s.pack('<I5', new_m)
            local msgs = {
                { addr = 0x55, write = s.pack('Bc5', 5, pack_m) },
                { addr = 0x55, write = s.pack('B', 5) },
                { addr = 0x55, read = 5 },
            }
            local rb, err = u.i2c_rdwr(self.fd, msgs)
            if not rb then
                return err
            else
                return bits.tohex(rb)
            end
        end

        function armada_mmio:oscillator_read()
            local msgs = {
                { addr = 0x55, write = s.pack('B', 5) },
                { addr = 0x55, read = 5 },
            }
            local rb, err = u.i2c_rdwr(self.fd, msgs)
            if not rb then
                return err
            else
                return bits.tohex(rb)
            end
        end

        function armada_mmio:adjust_oscillator(ppb)
            ppb = ppb + self.characterization_ppb
            -- per Si514 data sheet, Rev. 1.0, page 13, it is not
            -- possible to implement a frequency change >= +/- 1000 ppm
            -- using multiple small frequency changes without changing
            -- the center frequency and recalibrating the VCO.
            if ppb >= 1000000 then
                self.log:debug(
                    string.format('Limiting Si514 ppb %d to 999999', ppb))
                ppb = 999999
            elseif ppb <= -1000000 then
                self.log:debug(
                    string.format('Limiting Si514 ppb %d to -999999', ppb))
                ppb = -999999
            end
            local new_m = math.floor(offset_ppb(m, ppb) + 0.5)
            local pack_m = s.pack('<I5', new_m)
            local msgs = {
                { addr = 0x55, write = s.pack('Bc5', 5, pack_m) },
                { addr = 0x55, write = s.pack('B', 5) },
                { addr = 0x55, read = 5 },
            }
            local count = 0
            local rb, err = u.i2c_rdwr(self.fd, msgs)
            while not rb do
                self.log:warn('armada_mmio:adjust_oscillator(): [' ..
                              tostring(count) .. '] ' ..
                              'i2c_rdwr - ' .. err)
                if count < 2 then
                    u.close(self.fd)
                    sleep(.100)
                    self.fd = assert(u.open(tcfg.i2c_dev, u.O_RDWR))
                    rb, err = u.i2c_rdwr(self.fd, msgs)
                    count = count + 1
                else
                    u.close(self.fd)
                    self.log:error('armada_mmio:adjust_oscillator(): ' ..
                                  'i2c_rdwr - ' .. err .. ', exiting')
                    break
                end
            end
            if rb ~= pack_m then
                u.close(self.fd)
                self.log:error('armada_mmio:adjust_oscillator(): ' ..
                               'wrote ' .. bits.tohex(pack_m) .. ' ' ..
                               'read back ' .. bits.tohex(rb))
            end
            return ppb
        end
    else
        error('Unknown oscillator type')
    end

    -- Enable GTM2_TIN4 and GTM2_TOUT1 signals
    map:w32(0x00114, b.bor(b.band(map:r32(0x00114), 0xff73ffff), 0x00140000))

    os.execute('/sbin/modprobe gianfar_clocksource'
               ..' sysclk_freq='..tcfg.sysclk_freq
               ..' target_freq='..tcfg.target_freq)

    -- If the pulser is not being used (this is for testing only)
    if not tcfg.ptpv2 or not tcfg.ptpv2.pulse_hz then
        -- Enable TOUT1 pulse mode
        map:w16(0x00610, b.bor(map:r16(0x00610), 0x0020))
        map:w16(0x00612, b.bor(map:r16(0x00612), 0x0020))
        map:w16(0x00620, b.bor(map:r16(0x00620), 0x0020))
        map:w16(0x00622, b.bor(map:r16(0x00622), 0x0020))
    end
end

function armada_mmio:sample_timers()
    local f =
        io.open("/sys/devices/e0000000.immr/e0024e00.clocksource/sample64", "r")
    assert(f, 'gianfar_clocksource driver failure')
    local line = f:read("*l")
    f:close()
    local sys, ptp, sec, nsec =
        line:match("^(%d+) (%d+) [+-]?%d+ (%d+)%.(%d+)")
    return tonumber(sys), tonumber(ptp), tonumber(sec), tonumber(nsec)
end

function armada_mmio:send_sync_pulse()
    -- Synchronize channel board
    collectgarbage('stop')

    local tmr_alarm = self:sample_timers() + 1000000

    -- Set alarm
    local alarm_hi, alarm_lo = s.unpack("I4I4", s.pack("I8", tmr_alarm))
    map:w32(0x00614, alarm_hi)
    map:w32(0x00624, alarm_lo)

    -- Wait until alarm is past
    repeat
        local tmr_now = self:sample_timers()
    until tmr_now > tmr_alarm + 10000

    collectgarbage('restart')

    return tmr_alarm
end

local function addend(base, rate)
    return math.floor(4294967296 / (base / rate) + 0.5)
end

function armada_mmio:adjust_ptp_counter(ppb)
    local add = addend(tcfg.sysclk_freq, offset_ppb(tcfg.target_freq, ppb))
    map:w32(0x24e20, add)
    return add
end

function armada_mmio:get_local_timestamp()
    local ts_hi = map:r32(0x0061c)
    local ts_lo = map:r32(0x0062c)
    local ts_hi_reread = map:r32(0x0061c)
    if ts_hi ~= ts_hi_reread then
        -- Rollover occurred, reread the lo 32 bits
        ts_lo = map:r32(0x0062c)
        ts_hi = ts_hi_reread
    end
    return (ts_hi * 2^32) + ts_lo
end

return armada_mmio

