local bit = require 'bit'
local ffi = require 'ffi'

require 'strict'

local M = {}

local function next_bound(step, pos)
    return math.floor(pos / step) * step + step - 1
end

local function bit_partition(fmt, step, spec)
    -- bit_partition(8, {0, 15}) --> {0, 7, 8, 15}
    -- bit_partition(8, {0, 4}) --> {0, 4}
    -- bit_partition(8, {7, 8}) --> {7, 7, 8, 8}
    -- bit_partition(8, {22, 36}) --> {22, 23, 24, 31, 32, 36}
    local max_bound = math.max(unpack(spec))
    local bound_count = math.floor(max_bound / step) + 1
    local bounds = {}
    local bit_count = 0
    for si = 1, #spec, 2 do
        local lb, ub = spec[si], spec[si + 1]
        if lb <= ub then
            bit_count = bit_count + ub - lb + 1
            table.insert(bounds, lb)
            local cb = lb
            while true do
                local nb = next_bound(step, cb)
                if nb >= ub then
                    break
                end
                table.insert(bounds, nb)
                cb = nb + 1
                table.insert(bounds, cb)
            end
            table.insert(bounds, ub)
        else -- Handle bit-reversed stuff in the most naive way possible
            -- KLUDGE eventually want to do something intelligent with
            -- table lookups and whatnot
            bit_count = bit_count + lb - ub + 1
            for i = lb, ub, -1 do
                table.insert(bounds, i)
                table.insert(bounds, i)
            end
        end
    end
    if fmt == 's' and bounds[1] ~= bounds[2] then
        table.insert(bounds, 2, bounds[1])
        table.insert(bounds, 3, bounds[1] + 1)
    end
    return bounds, bit_count
end

local function bit_mangle_specs(fmt, step, spec)
    local bounds, bit_count = bit_partition(fmt, step, spec)
    assert(bit_count <= 64) -- KLUDGE make proper error KLUDGE until bitops64
    local ms = {}
    local ctype = ffi.typeof(string.format('uint%d_t[1]', step))
    local pos = 0
    for bi = 1, #bounds, 2 do
        local lb, ub = bounds[bi], bounds[bi + 1]
        local mlb, mub = lb % step, ub % step
        local mask_val = math.pow(2, step - mlb) - math.pow(2, step - mub - 1)
        local urshift = step - mub - 1

        local new_pos = pos + ub - lb + 1
        ms[#ms + 1] = {
            lb = lb,
            ub = ub,
            pos = pos,
            bpos = bit_count - pos,
            idx = math.floor(lb / step),
            mask = mask_val,
            xmask = bit.tohex(mask_val, step / 4),
            urshift = urshift,
            lshift = bit_count - new_pos - urshift,
        }
        pos = new_pos
    end
    return ms, bit_count
end

function M:compile_bit(fmt, input, spec, step)
    -- This should all work for fmt 's' (signed) or fmt 'u' (unsigned)
    -- Floating point is a different matter
    -- Consider something like ieee754(exponent_bits, spec) ???
    -- Where ieee754(8, ...) is float32, and ieee754(11, ...) is float64
    -- The ... is (sign_bit, exponent_bits, mantissa_bits)
    -- Could also consider something like ieee754(sign, exponent, mantissa)
    local meta = {
        min_input_bytes = 0,
    }
    step = step or 8
    if #spec == 1 then
        spec[2] = spec[1]
    end
    assert(#spec > 0 and #spec % 2 == 0)
    for i = 1, #spec do
        -- KLUDGE eventually want to allow run-time expressions, in which
        -- case a run-time bit() function must be used
        local v = spec[i]
        assert((tostring(v)):match('^%d+$')) -- Spec must be numeric constants
        spec[i] = tonumber(v)
    end
    local step_mask = math.pow(2, step) - 1
    local ms, bit_count = bit_mangle_specs(fmt, step, spec)
    assert(bit_count <= 64) -- Limit for LuaJIT Bitops
    local out_bits = bit_count > 32 and 64 or 32
    local iv = {}
    if out_bits == 64 then
        iv[#iv + 1] = "0LL" -- First arg for bit.bor(), assure 64-bitness
    end
    for i, m in ipairs(ms) do
        meta.min_input_bytes = math.max(meta.min_input_bytes, m.idx + 1)
        local sv = string.format("%s.value[%d]", input, m.idx)
        local mv = sv
        if m.mask ~= step_mask then
            mv = string.format("bit_band(%s, 0x%s)",
                sv, bit.tohex(m.mask, step / 4))
        end
        local v = mv
        if i == 1 and fmt == 's' then -- Deal with signed stuff
            if out_bits == 32 then
                local neg_v = math.pow(2, out_bits) - math.pow(2, bit_count - 1)
                v = string.format("bit_tobit(%s == 0 and 0 or 0x%s)",
                    v, bit.tohex(neg_v))
            else
                local neg_v = bit.lshift(bit.bnot(0LL), bit_count - 1)
                v = string.format("(%s == 0 and 0 or 0x%sLL)",
                    v, bit.tohex(neg_v))
            end
        elseif m.lshift ~= 0 then
            v = string.format("bit_%sshift(%s, %d)",
                m.lshift >= 0 and 'l' or 'r', mv, math.abs(m.lshift))
        end
        iv[#iv + 1] = v
    end
    local output_types = {
        s = 'message_int' .. out_bits, -- KLUDGE maybe not hardcode 'message'
        u = 'message_uint' .. out_bits, -- KLUDGE maybe not hardcode 'message'
    }
    meta.data_type = output_types[fmt]
    local code = string.format(#iv > 1 and "bit_bor(%s)" or "%s",
        table.concat(iv, ", "))

    -- Corner case, LuaJIT BitOps are signed, so fix up 32 bit values
    if fmt == 'u' and bit_count == out_bits then
        code = string.format('bit%d_unsigned_fixup(%s)', out_bits, code)
    end
    code = string.format('(%s:byte_length() >= %d and %s or nil)',
        input, meta.min_input_bytes, code)
    return code, meta
end

function M:compile_byte(fmt, input, byte_spec, step)
    if #byte_spec == 1 then
        byte_spec[2] = byte_spec[1]
    end
    assert(#byte_spec > 0 and #byte_spec % 2 == 0)
    local spec = {}
    for i = 1, #byte_spec, 2 do
        local b1, b2 = byte_spec[i], byte_spec[i + 1]
        if b1 <= b2 then
            table.insert(spec, b1 * 8)
            table.insert(spec, b2 * 8 + 7)
        else
            for b = b1, b2, -1 do
                table.insert(spec, b * 8)
                table.insert(spec, b * 8 + 7)
            end
        end
    end
    return self:compile_bit(fmt, input, spec, step)
end

return M
