local __ = require '__'
local log = require 'log'
local class = require 'class'
local video = {
    ['mjpeg'] = require 'video_mjpeg'
}
local sio = require 'sio'
local pthread = require 'pthread'
local net = require 'net_ffi'
local ffi = require 'ffi'
local C = ffi.C

-- GNU C Library 'libc'

ffi.cdef[[
struct timeval {
    unsigned long       tv_sec;
    long                tv_usec;
};
typedef unsigned int size_t;
struct _IO_FILE;
typedef struct _IO_FILE FILE;
FILE *fopen(const char *path, const char *mode);
int fclose(FILE *fp);
int fflush(FILE *fp);
size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream);
void rewind(FILE *stream);
int sscanf(const char *str, const char *format, ...);
int sprintf(char *str, const char *format, ...);
typedef uint32_t useconds_t;
int usleep(useconds_t usec);
]]

-- Network

ffi.cdef[[
int setsockopt(int socket, int level, int option_name,
               const void *option_value, socklen_t option_len);
]]

-- Camera Data Structures

ffi.cdef[[
typedef enum video_state {
    VIDEO_SHUTDOWN,
    VIDEO_CONNECT,
    VIDEO_REQUEST,
    VIDEO_STREAMING,
    VIDEO_RETRY,
    VIDEO_ERROR
} video_state_t;

typedef struct camera_data {
    video_state_t state;
    bool debug;
    int socket;
    char name[81];
    char vendor[33];
    char model[33];
    char serial[18];
    char firmware[33];
    char host[65];
    char port[7];
    char page[129];
    char query[257];
    char spec_frame_rate[3];
    char spec_image_size[10];
    char spec_image_compression[4];
    uint8_t frame_rate;
    uint32_t frame_count;
    uint32_t frame_len;
    pthread_mutex_t lock[1];
    void *(*api_write)(uint32_t, uint8_t);
    void *(*channel_write)(void *, size_t);
} camera_data_t;
]]

local m = class()

function m:init(settings)
    if settings.data ~= nil then
        self.camera = settings.data
    else
        self.camera = ffi.new('camera_data_t [1]')
        self.camera[0].debug = false
        self.camera[0].state = C.VIDEO_SHUTDOWN
        self.camera[0].socket = -1
        self.camera[0].frame_count = 0
        self.camera[0].frame_rate = 0
        self.camera[0].frame_len = 0
        ffi.copy(self.camera[0].name, settings.vendor .. '-' ..
                                  settings.model .. '-' ..
                                  settings.serial:sub(7))
        ffi.copy(self.camera[0].vendor, settings.vendor)
        ffi.copy(self.camera[0].model, settings.model)
        ffi.copy(self.camera[0].serial, settings.serial)
        ffi.copy(self.camera[0].firmware, settings.firmware)
        ffi.copy(self.camera[0].spec_frame_rate, '10')
        ffi.copy(self.camera[0].spec_image_size, '320x240')
        ffi.copy(self.camera[0].spec_image_compression, '30')
        ffi.copy(self.camera[0].host, settings.host)
        ffi.copy(self.camera[0].port, settings.port)
        pthread:mutex_init(self.camera[0].lock, nil)
        self.camera[0].channel_write = nil
    end
end

function m:setup(data, model, serial, host, firmware, port)
    if ffi.istype('camera_data_t [1]', data) or
                ffi.istype('camera_data_t *', data) then
        return m:new(self, { data = data })
    else
        return m:new(self, {
            vendor = data,
            model = model,
            serial = serial,
            host = host,
            firmware = firmware,
            port = port or '80'
        })
    end
end

function m:start(codec)
    if self.camera[0].state ~= C.VIDEO_SHUTDOWN then
        self:stop()
    end
    self:state(C.VIDEO_CONNECT)
    if not pthread:create(video[codec], self.camera) then
        return false
    end
    local wait = 0
    while self.camera[0].state ~= C.VIDEO_STREAMING do
        if wait == 100 then
            self:stop()
            return false
        end
        wait = wait + 1
        C.usleep(10000)
    end
    return true
end

function m:stop()
    self:state(C.VIDEO_SHUTDOWN)
    local wait = 0
    while self.camera[0].state ~= C.VIDEO_SHUTDOWN do
        if wait == 100 then
            break
        end
        wait = wait + 1
        C.usleep(10000)
    end
    self.camera[0].frame_count = 0
    self.camera[0].frame_rate = 0
    self.camera[0].frame_len = 0
    self:disconnect()
    pthread:join()
end

function m:api_write(x, len)
    if len then
        if self.camera[0].api_write ~= nil then
            self.camera[0].api_write(x, len)
        end
    else
        pthread:lock(self.camera[0].lock)
        if x == false then
            self.camera[0].api_write = nil
        else
            self.camera[0].api_write = x
        end
        pthread:unlock(self.camera[0].lock)
    end
end

function m:channel_write(x, len)
    if len then
        if self.camera[0].channel_write ~= nil then
            self.camera[0].channel_write(x, len)
        end
    else
        pthread:lock(self.camera[0].lock)
        if x == false then
            self.camera[0].channel_write = nil
        else
            self.camera[0].channel_write = x
        end
        pthread:unlock(self.camera[0].lock)
    end
end

function m:data()
    return self.camera
end

function m:debug(x)
    if x then
        self.camera[0].debug = x
    else
        return self.camera[0].debug
    end
end

function m:state(x)
    pthread:lock(self.camera[0].lock)
    if x then
        self.camera[0].state = x
    else
        local o = self.camera[0].state
        pthread:unlock(self.camera[0].lock)
        return o
    end
    pthread:unlock(self.camera[0].lock)
end

function m:socket(x)
    if x then
        self.camera[0].socket = x
    else
        return self.camera[0].socket
    end
end

function m:frame_length(x)
    if x then
        self.camera[0].frame_len = x
    else
        return self.camera[0].frame_len
    end
end

function m:frame_count(x)
    if x then
        self.camera[0].frame_count = x
    else
        return self.camera[0].frame_count
    end
end

function m:frame_rate(x)
    if x then
        self.camera[0].frame_rate = x
    else
        return self.camera[0].frame_rate
    end
end

function m:host(x)
    if x then
        ffi.copy(self.camera[0].host, x)
    else
        return ffi.string(self.camera[0].host)
    end
end

function m:port(x)
    if x then
        ffi.copy(self.camera[0].port, x)
    else
        return ffi.string(self.camera[0].port)
    end
end

function m:name(x)
    if x then
        ffi.copy(self.camera[0].name, x)
    else
        return ffi.string(self.camera[0].name)
    end
end

function m:vendor(x)
    if x then
        ffi.copy(self.camera[0].vendor, x)
    else
        return ffi.string(self.camera[0].vendor)
    end
end

function m:model(x)
    if x then
        ffi.copy(self.camera[0].model, x)
    else
        return ffi.string(self.camera[0].model)
    end
end

function m:serial(x)
    if x then
        ffi.copy(self.camera[0].serial, x)
    else
        return ffi.string(self.camera[0].serial)
    end
end

function m:firmware(x)
    if x then
        ffi.copy(self.camera[0].firmware, x)
    else
        return ffi.string(self.camera[0].firmware)
    end
end

function m:page()
    return ffi.string(self.camera[0].page)
end

function m:query()
    return ffi.string(self.camera[0].query)
end

function m:spec_frame_rate(x)
    if x then
        ffi.copy(self.camera[0].spec_frame_rate, tostring(x))
    else
        local rate = ffi.new('uint8_t [1]')
        C.sscanf(self.camera[0].spec_frame_rate, "%c", rate)
        return ffi.string(self.camera[0].spec_frame_rate), rate[0]
    end
end

function m:spec_image_size(x)
    if x then
        ffi.copy(self.camera[0].spec_image_size, x)
    else
        local hres = ffi.new('uint32_t [1]')
        local vres = ffi.new('uint32_t [1]')
        C.sscanf(self.camera[0].spec_image_size, "%ux%u", hres, vres)
        return ffi.string(self.camera[0].spec_image_size), hres[0], vres[0]
    end
end

function m:spec_image_compression(x)
    if x then
        ffi.copy(self.camera[0].spec_image_compression, tostring(x))
    else
        local comp = ffi.new('uint8_t [1]')
        C.sscanf(self.camera[0].spec_image_compression, "%d", comp)
        if comp[0] == 0 then
            comp[0] = 1
        end
        return ffi.string(self.camera[0].spec_image_compression), comp[0]
    end
end

function m:lock()
    pthread:lock(self.camera[0].lock)
end

function m:unlock()
    pthread:unlock(self.camera[0].lock)
end

function m:update()
    return true
end

function m:write(sd, buf, len)
    local n = -1
    local sent = 0
    local data = ffi.cast('char *', buf)
    while sent < len do
        n = C.send(sd, data + sent, len - sent, 0)
        if n == -1 then
            log:error(__"camera:write(): {0} on {1}",
                      ffi.string(C.strerror(ffi.errno())), self:name())
            break
        end
        sent = sent + n
    end
    return n
end

function m:request()
    local n = ffi.new('int', -1)
    local rc = ffi.new('int', -1)
    local sent = ffi.new('int', 0)
    local query = "GET /" .. self:page() .. self:query() .. " HTTP/1.0\r\n" ..
                  "Host: " .. self:host() .. "\r\n" ..
                  "User-Agent: HBM CX23R\r\n\r\n"
    local sbuf = ffi.new('char [?]', #query + 1)
    ffi.copy(sbuf, query, #query)
    while sent < #query do
        local wfds = ffi.new('fd_set [1]')
        local tv = ffi.new('struct timeval [1]')
        sio.FD_ZERO(wfds)
        sio.FD_SET(self.camera[0].socket, wfds)
        tv[0].tv_sec = ffi.new('long', 1)
        tv[0].tv_usec = 0
        rc = C.select(self.camera[0].socket + 1, nil, wfds, nil, tv)
        if rc > 0 then
            n = C.send(self.camera[0].socket, sbuf + sent, #query - sent, 0)
            if n == -1 then
                io.stderr:write(self:name() .. ":request(): send " ..
                                ffi.string(C.strerror(ffi.errno())) .. "\n")
                break
            end
            sent = sent + n
        elseif rc == -1 then
            io.stderr:write(self:name() .. ":request(): select " ..
                            ffi.string(C.strerror(ffi.errno())) .. "\n")
            n = -1
            break
        else
            io.stderr:write(self:name() .. ":request(): timeout\n")
            n = -1
            break
        end
    end
    return n
end

function m:connect()
    local hints = ffi.new('struct addrinfo [1]')
    local res = ffi.new('struct addrinfo *[1]')
    local ressave = ffi.new('struct addrinfo *[1]')
    hints[0].ai_family = C.AF_UNSPEC
    hints[0].ai_socktype = C.SOCK_STREAM
    local rc = C.getaddrinfo(self.camera[0].host, self.camera[0].port,
                             hints, res)
    if rc ~= 0 then
        io.stderr:write(self:name() .. ":connect(): " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
        return sd
    end
    ressave = res
    while res ~= nil do
        self.camera[0].socket = C.socket(res[0].ai_family, res[0].ai_socktype,
                                         res[0].ai_protocol)
        if self.camera[0].socket >= 0 then
            rc = C.connect(self.camera[0].socket, res[0].ai_addr,
                           res[0].ai_addrlen)
            if rc ~= 0 then
                io.stderr:write(self:name() .. ":connect(): " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                C.close(self.camera[0].socket)
                self.camera[0].socket = -1
            else
                -- socket is connected
                break
            end
        end
        res = res[0].ai_next
    end
    C.freeaddrinfo(ressave[0])
    return self.camera[0].socket
end

function m:disconnect()
    net.shutdown(self.camera[0].socket)
    local o = { camera = ffi.new('camera_data_t [1]') }
    ffi.copy(o.camera, self.camera, ffi.sizeof(o.camera))
    self.camera = nil
    collectgarbage()
    self.camera = o.camera
end

function m:upnp()
    local host = ffi.new('char [16]', '239.255.255.250')
	local rc = ffi.new('int')
    local addr = ffi.new('struct sockaddr_in [1]')
    -- setup socket
	local sd = ffi.new('int')
    sd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if sd < 0 then
        io.stderr:write("camera:upnp(): " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
        return
    end
    ffi.fill(addr, ffi.sizeof('struct sockaddr_in'), 0)
    addr[0].sin_family = C.AF_INET
    addr[0].sin_port = C.htons(ffi.new('uint16_t', 1900))
    addr[0].sin_addr.s_addr = C.inet_addr(host)
    -- build query
    local msearch = 'M-SEARCH * HTTP/1.1\r\n' ..
                    'HOST: ' .. ffi.string(host) .. ':1900\r\n' ..
                    'ST: "ssdp:all"\r\n' ..
                    'MAN: "ssdp:discover"\r\n' ..
                    'MX: 3\r\n\r\n'
	local query = ffi.new('unsigned char [?]', #msearch)
    ffi.copy(query, msearch, #msearch)
    -- send query to destination address
    rc = C.sendto(sd, query, ffi.sizeof(query), 0,
                  ffi.cast('struct sockaddr *', addr), ffi.sizeof(addr))
    if rc < 0 then
        io.stderr:write("camera:upnp(): sendto " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
    end
    C.close(sd)
    C.usleep(1000000)
end

function m:probe()
    self:upnp()
end

function m:info(ip, query, rbuf)
    local port
    local host
    if self.camera then
        port = ffi.cast('char *', self.camera[0].port)
        host = ffi.cast('char *', self.camera[0].host)
    else
        port = ffi.new('char [3]', '80')
        host = ffi.new('char [?]', #ip + 1)
        ffi.copy(host, ip, #ip)
    end
    local rc = ffi.new('int')
    local sd = ffi.new('int')
    -- connect
    local hints = ffi.new('struct addrinfo [1]')
    local res = ffi.new('struct addrinfo *[1]')
    local ressave = ffi.new('struct addrinfo *[1]')
    hints[0].ai_family = C.AF_UNSPEC
    hints[0].ai_socktype = C.SOCK_STREAM
    rc = C.getaddrinfo(host, port, hints, res)
    if rc ~= 0 then
        io.stderr:write("camera:info(): getaddrinfo " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
        return false
    end
    ressave = res
    while res ~= nil do
        sd = C.socket(res[0].ai_family, res[0].ai_socktype,
                                         res[0].ai_protocol)
        if sd >= 0 then
            rc = C.connect(sd, res[0].ai_addr,
                           res[0].ai_addrlen)
            if rc ~= 0 then
                io.stdout:write("camera:info(): connect " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                net.shutdown(sd)
                sd = -1
            else
                -- socket is connected
                break
            end
        end
        res = res[0].ai_next
    end
    C.freeaddrinfo(ressave[0])
    if sd < 0 then
        return false
    end
    -- send
    local n = ffi.new('int')
    local sent = ffi.new('int')
    local rfds = ffi.new('fd_set [1]')
    local wfds = ffi.new('fd_set [1]')
    local tv = ffi.new('struct timeval [1]')
    local sbuf = ffi.new('char [?]', #query + 1)
    ffi.copy(sbuf, query, #query)
    while sent < #query do
        sio.FD_ZERO(wfds)
        sio.FD_SET(sd, wfds)
        tv[0].tv_sec = ffi.new('long', 1)
        tv[0].tv_usec = 0
        rc = C.select(sd + 1, nil, wfds, nil, tv)
        if rc > 0 then
            n = C.send(sd, sbuf + sent, #query - sent, 0)
            if n == -1 then
                io.stderr:write("camera:info(): send " ..
                                ffi.string(C.strerror(ffi.errno())) .. "\n")
                net.shutdown(sd)
                return false
            end
            sent = sent + n
        elseif rc == -1 then
            io.stderr:write("camera:info(): select " ..
                            ffi.string(C.strerror(ffi.errno())) .. "\n")
            net.shutdown(sd)
            return false
        else
            io.stderr:write("camera:info(): send timeout\n")
            net.shutdown(sd)
            return false
        end
    end
    -- recv
    n = 0
    local http = ffi.new('char *')
    local http_response = {
        version = ffi.new('char [4]'),
        code = ffi.new('char [4]'),
        reason = ffi.new('char [33]')
    }
    repeat
        sio.FD_ZERO(rfds)
        sio.FD_SET(sd, rfds)
        tv[0].tv_sec = ffi.new('long', 3)
        tv[0].tv_usec = ffi.new('long', 0)
        rc = C.select(sd + 1, rfds, nil, nil, tv)
        if rc > 0 then
            n = C.recv(sd, rbuf + n, ffi.sizeof(rbuf) - n, 0)
            if n < 0 then
                io.stderr:write("camera:info(): recv " ..
                         ffi.string(C.strerror(ffi.errno())) .. "\n")
                break
            end
            if http == nil then
                http = C.strstr(rbuf, "HTTP/")
                if http ~= nil then
                    local ptr = C.strstr(http, "\r\n")
                    if ptr ~= nil then
                        ptr[0] = 0
                        C.sscanf(http, "HTTP/%s %s %32[^\n]s",
                                        http_response.version,
                                        http_response.code,
                                        http_response.reason)
                        ptr[0] = 13
                        if ffi.string(http_response.code) ~= '200' then
                            io.stderr:write("video: HTTP/" ..
                                ffi.string(http_response.version) .. " " ..
                                ffi.string(http_response.code) .. " " ..
                                ffi.string(http_response.reason) .. "\n")
                            net.shutdown(sd)
                            return false
                        end
                    end
                end
            end
        elseif rc == -1 then
            io.stderr:write("camera:info(): select " ..
                 ffi.string(C.strerror(ffi.errno())) .. "\n")
            break
        else
            io.stderr:write("camera:info(): recv timeout\n")
            break
        end
    until n <= 0
    net.shutdown(sd)
    return true
end

function m:snapshot(proxy, file)
    log:notice(__"camera:snapshot unimplemented for {0}", proxy.proxy_hostname)
end

return m
