local __ = require '__'
local log = require 'log'

if not arg[1] and not arg[2] and not arg[3] and not arg[4] and not arg[5] then
    error("camera_proxy:  USAGE: luajit camera_proxy.lua " ..
        "VENDOR MODEL SERIAL IP FIRMWARE")
end

local nf = require 'net_functions'
nf.wait_carrier('any', 'ignore', {'eth1', 'eth2'})

local api = {}

local channel_proxy = require 'channel_proxy'

local camera_proxy = class('camera_proxy', channel_proxy)

local ev = require 'ev'
local ffi = require 'ffi'
local ts = require 'timestamps'
local C = ffi.C

ffi.cdef([[
typedef int32_t pid_t;
pid_t getpid(void);
struct msg_dhdr {
    unsigned int   len;
    long long      ts;
    unsigned char  data[0];
} __attribute__((packed));
struct msg_iovec {
    void   *iov_base;
    size_t  iov_len;
};
]])

local video
local fp
local last_ts
local iovec = ffi.new('struct msg_iovec[2]')
local dhdr = ffi.new('struct msg_dhdr[1]')
local dhdr_size = ffi.sizeof('struct msg_dhdr')
iovec[0].iov_len = dhdr_size
iovec[0].iov_base = dhdr

local function image_sink(buf, len)
    -- 'video' channel only (message_raw)
    if video and video.socket then
        local timestamp = ts.ts_to_stamp(ts.get_ts())
        if not timestamp then return end
        dhdr[0].len = len + dhdr_size - 4
        dhdr[0].ts = timestamp - video.start_stamp
        iovec[1].iov_len = len
        iovec[1].iov_base = ffi.cast('void *', buf)
        for i = 0, 1 do
            api:write(video.socket, iovec[i].iov_base, iovec[i].iov_len)
        end
        if (timestamp - last_ts) >= 4000000 then
            C.rewind(fp)
            C.fwrite(iovec[1].iov_base, 1, iovec[1].iov_len, fp)
            C.fflush(fp)
            last_ts = timestamp
        else
            last_ts = dhdr[0].ts
        end
    end
end

function camera_proxy:initialize(camera)
    api = camera
    self.camera = camera
    self.proxy_type = 'video'
    self.pid = string.format("%.0f", tonumber(C.getpid()))
    self.streaming = false
    self.channels = {}
    self.channel_sinks = {}
    table.insert(self.channels, {
        name = 'video',
        description = 'Video Input',
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/msg_camera'
    })
    self.channels['video'] = self.channels[1]
    table.insert(self.channels, {
        name = 'control',
        description = 'Camera Control Channel',
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/msg_camera',
    })
    self.channels['control'] = self.channels[2]
    channel_proxy.initialize(self, 'camera.' .. camera:name())
    fp = camera:snapshot(self, camera:name() .. '.jpg')
end

function camera_proxy:start_stream(w, spec)
    if not self.streaming then
        self.streaming = true
        self.camera:spec_frame_rate(spec.video_frame_rate)
        self.camera:spec_image_size(spec.video_image_size)
        self.camera:spec_image_compression(spec.video_image_compression)
        if not self.camera:start('mjpeg') then
            os.exit(1)
        end
    end
end

function camera_proxy:available_channels(hostname)
    local o = {}
    for _, ch in ipairs(self.channels) do
        if ch.name == 'control' then
            ch.connector = 'control.status@' .. self.proxy_hostname
        else
            ch.connector = ch.name .. '@' .. self.proxy_hostname
        end
        table.insert(o, {
            name = ch.name .. '_' .. hostname:gsub('%W', '_'),
            connector = ch.connector,
            description = ch.description,
            electrical_units = ch.electrical_units,
            data_type = ch.data_type,
            category = ch.category,
            ip_address = self.camera:host(),
            firmware_version = self.camera:firmware(),
            serial_number = self.camera:serial(),
            snapshot = self.camera:name() .. '.jpg',
            comet_message = self.camera:name() .. '.jpg'
        })
    end
    return o
end

function camera_proxy:start_comet_stream(w, chans, sink)
-- required channel_proxy callback function
end

function camera_proxy:start_channel_stream(w, spec, md_sink, sink)
    local name = spec.name:match('(%a+)_%a+')
    if self.channels[name] then
        if name == 'video' then
            video = self.channels[name]
            video.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
            video.socket = w.socket
            last_ts = video.start_stamp
            w.proxy = self
            w.cleanup = function ()
                self.streaming = false
                self.camera:stop()
                video.socket = nil
                video = nil
                self.camera:channel_write(false)
            end
            self.camera:channel_write(ffi.cast('void *(*)(void *, size_t)',
                                               image_sink))
--[[ FUTURE CHANNELS
        else
            table.insert(self.channel_sinks, sink)
            self.channel_sinks[spec.connector] = sink
            w.cleanup = function()
                self.channel_sinks[spec.connector] = nil
                for i = #self.channel_sinks, 1, -1 do
                    if self.channel_sinks[spec.connector] == sink then
                        table.remove(self.channel_sinks, i)
                    end
                end
            end
]]
        end
        w.proxy = self
        md_sink(spec)
        self:start_stream(w, spec)
    end
end

function camera_proxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

function camera_proxy:snapshot(params)
    local fn = self.camera:name() .. '.jpg'
    if not self.streaming then
        fp = self.camera:snapshot(self, fn)
    end
    return {
        msg = string.format("%s on %s", fn, self.proxy_hostname),
        result = 'success',
    }
end

-- Build table of known camera types by vendor
local camera = {
    ['axis'] = require 'axis_camera',
    ['vivotek'] = require 'vivotek_camera'
}

-- Store the given vendor
local vendor = arg[1]

-- Setup Camera Parameters
camera[vendor]:setup(vendor, arg[2], arg[3], arg[4], arg[5])

-- Create new instance of the Camera Proxy
camera_proxy:new(camera[vendor])

-- Main Event Loop
ev.Loop.default:loop()
