local __ = require '__'
local log = require 'log'

local cjson = require 'cjson.safe'
local ev = require 'ev'
local ffi = require 'ffi'
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_new = ffi.new
local ts = require 'timestamps'
local u = require 'unix'

local bit = require 'bit'
local bor = bit.bor
local band = bit.band
local rshift = bit.rshift
local arshift = bit.arshift
local lshift = bit.lshift
local bswap = bit.bswap
local tobit = bit.tobit
local htonl = ffi.abi('le') and bswap or tobit

local sample_clock = require 'sample_clock'

local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'
local message_can = require 'message_can'
local message_float32 = require 'message_float32'
local message_float64 = require 'message_float64'

local can = require 'can'

local C = ffi.C

require 'strict'

local update_rate = 10 -- Hz -- KLUDGE is this heartbeat rate?

local dumper = require 'data_dumper'

local can_buf_count = 512 -- KLUDGE!!!

ffi.cdef[[
    ssize_t write(int fd, const void *buf, size_t count);

    typedef struct __attribute__((packed)) can_sink {
        uint64_t base_timestamp;
        int socket;
        int max_count;
        int count;
        struct {
            uint32_t len;
            message_can msg[512]; // KLUDGE can_buf_count!!!
        } buf;
    } can_sink;

    typedef struct __attribute__((packed)) data_sink_f32 {
        uint64_t base_timestamp;
        int socket;
        int max_count;
        int count;
        struct {
            uint32_t len;
            message_float32     msg[512]; // KLUDGE can_buf_count!!!
        } buf;
    } data_sink_f32;

    typedef struct __attribute__((packed)) data_sink_f64 {
        uint64_t base_timestamp;
        int socket;
        int max_count;
        int count;
        struct {
            uint32_t len;
            message_float64     msg[512]; // KLUDGE can_buf_count!!!
        } buf;
    } data_sink_f64;
]]

local can_sink = ffi.metatype('can_sink', { __index = {
    flush = function(self, pkt)
        if self.count <= 0 then
            return -- KLUDGE HEARTBEATS HERE???
        end
        self.buf.len = self.count * ffi.sizeof'message_can'
        local towrite = ffi.sizeof'uint32_t' + self.buf.len
        local wrote = C.write(self.socket, self.buf, towrite)
        if (wrote ~= towrite) then
            -- KLUDGE do some error thing here! (maybe)
        end
        self.count = 0
    end,
    enqueue = function(self, msg)
        if self.count >= self.max_count then
            self:flush()
        end
        local dmsg = self.buf.msg + self.count
        dmsg.timestamp[0] = msg.timestamp[0] - self.base_timestamp
        ffi.copy(dmsg.value, msg.value, 13)
        self.count = self.count + 1
    end,
}})

local data_sink_f32 = ffi.metatype('data_sink_f32', { __index = {
    flush = function(self)
        if self.count <= 0 then
            return -- KLUDGE HEARTBEATS HERE???
        end
        self.buf.len = self.count * ffi.sizeof'message_float32'
        local towrite = ffi.sizeof'uint32_t' + self.buf.len
        local wrote = C.write(self.socket, self.buf, towrite)
        if (wrote ~= towrite) then
            -- KLUDGE do some error thing here! (maybe)
        end
        self.count = 0
    end,
    enqueue = function(self, msg)
        if self.count >= self.max_count then
            self:flush()
        end
        local dmsg = self.buf.msg + self.count
        dmsg.timestamp[0] = msg.timestamp[0] - self.base_timestamp
        ffi.copy(dmsg.value, msg.value, ffi.sizeof('float'))
        self.count = self.count + 1
    end,
}})

local data_sink_f64 = ffi.metatype('data_sink_f64', { __index = {
    flush = function(self)
        if self.count <= 0 then
            return -- KLUDGE HEARTBEATS HERE???
        end
        self.buf.len = self.count * ffi.sizeof'message_float64'
        local towrite = ffi.sizeof'uint32_t' + self.buf.len
        local wrote = C.write(self.socket, self.buf, towrite)
        if (wrote ~= towrite) then
            -- KLUDGE do some error thing here! (maybe)
        end
        self.count = 0
    end,
    enqueue = function(self, msg)
        if self.count >= self.max_count then
            self:flush()
        end
        local dmsg = self.buf.msg + self.count
        dmsg.timestamp[0] = msg.timestamp[0] - self.base_timestamp
        ffi.copy(dmsg.value, msg.value, ffi.sizeof('double'))
        self.count = self.count + 1
    end,
}})

----------------------------------------------------------------------

local can_proxy = class('can_proxy', channel_proxy)

function can_proxy:initialize(name, devname)
    self.proxy_type = 'can'
    self.device = string.format("CAN%d",
                                tonumber(devname:match('/dev/cca(%d+)')) + 1)
    cca_handler.new(devname, self, can, 'CAN_' .. tonumber(devname:match('/dev/cca(%d+)')) .. '.bin', sample_clock.master_rate)

    self:cmd_block('txdr_passthru_power_on')    -- Need to fix someday

    self.streaming = false

    self.device_spec = {}

    self.heartbeat_msg = message_can.new()
    self.heartbeat_msg:set_heartbeat(true)

    self.event_counts = 0

    self.can_signal_channel = {}
    self.channel_sinks = {}

    channel_proxy.initialize(self, name)

    self.connector = '@' .. self.proxy_hostname
end

-- Function used to decode raw data from the FPGA
function can_proxy:buffer(msg)
    if not self.streaming then
        -- Idle
        return
    end
    if msg.cmd == C.CAN_RX_START then
        local last_non_hb = msg.element_count - (msg.heartbeat and 2 or 1)
        for i = 0, last_non_hb do
            -- Process raw can channel if included in the test
            if self.channel_sinks[self.connector] then
                self.channel_sinks[self.connector]:enqueue(
                    ffi_cast(message_can.ptr_type, msg.bufdata + i))
            end
            -- All other channels processed with CAN ID Mapping
            local buf = ffi_cast('can_rxtx_buffer_t *', msg.bufdata + i)

            local sig_list = self.can_signal_channel[buf.can_id]
            if sig_list then
                for _, info in ipairs(sig_list) do
                    info.handler(buf, info)
                end
            end
        end
    elseif msg.cmd == C.CAN_EVENT_START then
        self.event_counts = self.event_counts + 1
        -- Drive raw message heartbeats with event timer
        if self.channel_sinks[self.connector] then
            if self.event_counts == 10 then
                self.heartbeat_msg.timestamp[0] = msg.bufdata[0].ts
                self.channel_sinks[self.connector]:enqueue(self.heartbeat_msg)
            end
            self.channel_sinks[self.connector]:flush()
        end

        -- Stale data and timed rate processing via heartbeat
        for can_id, info_list in pairs(self.can_signal_channel) do
            for _, info in ipairs(info_list) do

                -- Send heartbeats at a maximum rate of 1Hz
                if self.event_counts == 10 then
                -- If we have not reached the start time yet, dont bother
                    info.msg.timestamp[0] = msg.bufdata[0].ts
                    info.msg:set_heartbeat(true)
                    info.sink:enqueue(info.msg)
                end
                info.sink:flush()
            end
        end
        if self.event_counts == 10 then
            self.event_counts = 0
        end

        --
        -- Add other Event Processing Here (loop through all buffered messages)
        --
        --local buf
        --for i = 0, msg.element_count do
        --    buf = msg.bufdata[i]
        --end
    end
end

can_proxy.spec_translate = {
    bit_rate = 'baud_rate',
}

-- Function used to start CAN data from the FPGA
function can_proxy:start_stream(w, spec)
    w.proxy = self
    if not self.streaming then
        for k, v in pairs(spec) do
            self.device_spec[self.spec_translate[k] or k] = v
        end
        if not spec.internal_termination then
            self:cmd('disable_terminator')
        else
            self:cmd('enable_terminator')
        end
        self:cmd('bus_setup', self.device_spec)
        self:cmd('event_setup', {bus_off=true, bus_off_deassert=true})
        local start_time =
            ts.ts_to_stamp(ts.get_ts()) + sample_clock.master_rate / 10
        self:cmd('rx_start', start_time, update_rate, 16, 128)
        self:cmd('event_start', start_time, update_rate, 16, 128)
        self.streaming = true
        self.event_counts = 0
    end
end

-- Function used to stop the FPGA from sending CAN data
function can_proxy:stop_stream()
    if self.streaming then
        self.streaming = false
        self:cmd('event_stop')
        self:cmd('rx_stop')
        self.device_spec = {}
    end
end

-- Function used to assign a comet sink to channels
function can_proxy:start_comet_stream(w, channels)
end

-- Decoder for signed little endian signal data
function can_proxy.handle_signed_le_signal(buf, info)
    local v = ffi_cast('int64_t *', buf.can_message)[0]
    local shift = info.lshift_bits
    local rshift_amt = info.rshift_bits

    v = bswap(v)
    v = lshift(v, shift)
    v = arshift(v, rshift_amt)
    
    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(v) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for unsigned little endian signal data
function can_proxy.handle_unsigned_le_signal(buf, info)
    local v = ffi_cast('uint64_t *', buf.can_message)[0]
    local shift = info.lshift_bits
    local rshift_amt = info.rshift_bits

    v = bswap(v)
    v = lshift(v, shift)
    v = rshift(v, rshift_amt)

    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(v) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for signed big endian signal data
function can_proxy.handle_signed_be_signal(buf, info)
    local v = ffi_cast('int64_t *', buf.can_message)[0]
    local shift = info.lshift_bits

    v = lshift(v, shift)
    v = arshift(v, info.rshift_bits)
    
    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(v) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for unsigned big endian signal data
function can_proxy.handle_unsigned_be_signal(buf, info)
    local v = ffi_cast('int64_t *', buf.can_message)[0]
    local shift = info.lshift_bits

    v = lshift(v, shift)
    v = rshift(v, info.rshift_bits)
    
    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(v) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

local fv = ffi_new('float[2]')

-- Decoder for float little endian signal data
-- FIXE:  Not sure if this one works properly!
function can_proxy.handle_float_le_signal(buf, info)
    ffi_copy(fv, buf.can_message, 8)
    local v = ffi_cast('int64_t *', fv)

    local rshift_amt = info.rshift_bits
    local shift = info.lshift_bits

    v[0] = bswap(v[0])
    v[0] = lshift(v[0], shift)
    v[0] = rshift(v[0], rshift_amt)

    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(fv[1]) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for float big endian signal data
function can_proxy.handle_float_be_signal(buf, info)
    local fv = ffi_cast('float *', buf.can_message)

    local idx = 0
    if info.lshift_bits > 0 then
        idx = 1
    end

    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(fv[idx]) * info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for double little endian signal data
function can_proxy.handle_double_le_signal(buf, info)
    ffi_copy(fv, buf.can_message, 8)
    local v = ffi_cast('int64_t *', fv)

    v[0] = bswap(v[0])

    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(ffi_cast('double *', fv)[0]) *
        info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Decoder for double big endian signal data
function can_proxy.handle_double_be_signal(buf, info)
    info.msg.timestamp[0] = buf.ts
    info.msg.value[0] = tonumber(ffi_cast('double *', buf.can_message)[0]) *
        info.scale + info.offset
    info.sink:enqueue(info.msg)
end

-- Function called to start acquisiton on a channel
function can_proxy:start_channel_stream(w, spec, md_sink, sink)
    -- Initial channel setup
    self.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
    spec.ts = self.start_stamp

    -- If this is NOT the raw CAN channel default some stuff and add the 
    -- setup information to the CAN signal channel list
    if spec.connector ~= self.connector then
        spec.db_msg_id = tonumber(spec.db_msg_id or 0)

        -- Assign the signal data handler
        local handlers = {
            ULSB = self.handle_unsigned_le_signal,
            UMSB = self.handle_unsigned_be_signal,
            SLSB = self.handle_signed_le_signal,
            SMSB = self.handle_signed_be_signal,
            FLSB = self.handle_float_le_signal,
            FMSB = self.handle_float_be_signal,
        }
        local handler = handlers[spec.db_data_format]

        -- Switch in the proper floating point handler if needed
        if (spec.db_data_format == 'FLSB' or spec.db_data_format == 'FMSB') and
            spec.db_bit_length > 32 then
            if spec.db_data_format == 'FLSB' then
                handler = self.handle_double_le_signal
            else
                handler = self.handle_double_be_signal
            end
        end

        -- Calculate the aggregate scale and offset
        local slope = tonumber(spec.db_scale or 1.0)
        local intercept = tonumber(spec.db_offset or 0.0) 
        local csteps = {}
        for _, cal in pairs(spec.calibration) do
            if cal.mode then
                if cal.mode == 'defined_slope_intercept' then
                    -- Don't need to do anything here
                elseif cal.mode == 'defined_zero_span' then
                    cal.slope = tonumber(cal.physical_span) /
                        tonumber(cal.electrical_span)
                    cal.intercept =
                        tonumber(cal.electrical_zero) * cal.slope * -1
                elseif cal.mode == 'defined_two_point' or
                    cal.mode == 'experimental_two_point' then
                    cal.slope =
                        (tonumber(cal.physical[2]) -
                            tonumber(cal.physical[1])) /
                        (tonumber(cal.electrical[2]) -
                            tonumber(cal.electrical[1]))
                    cal.intercept = tonumber(cal.physical[2] -
                        cal.slope * tonumber(cal.electrical[2]))
                else
                    log:log(log.ERR, string.format(
                        "Unrecognized calibration mode: %s", cal.mode))
                end
            end
            if cal.purpose and cal.applied then
                if cal.purpose == 'sensor_scaling' then
                    table.insert(csteps, 1, cal)
                elseif cal.purpose == 'units_conversion' then
                    table.insert(csteps, 2, cal)
                elseif cal.purpose == 'rezero' then
                    table.insert(csteps, 3, cal)
                end
            end
        end
        for _, cal in ipairs(csteps) do
            slope = slope * cal.slope
            intercept = cal.slope * intercept + cal.intercept
        end

        self.can_signal_channel[spec.db_msg_id] =
            self.can_signal_channel[spec.db_msg_id] or {}
        self.can_signal_channel[spec.db_msg_id][
            #self.can_signal_channel[spec.db_msg_id] + 1] = {
            connector = spec.connector,
            sink = spec.data_type == 'message_float32' and
                ffi_new('data_sink_f32', self.start_stamp, w.socket, 100, 0) or
                ffi_new('data_sink_f64', self.start_stamp, w.socket, 100, 0),
            msg = spec.data_type == 'message_float32' and
                message_float32:new() or message_float64:new(),
            handler = handler,
            -- FIXME:  Add support for bit reversal
            lshift_bits = spec.db_data_format:find('.MSB') and
                (((math.floor(spec.db_bit_start/8)*8)+7)-(spec.db_bit_start%8))
-- THIS MAY BE WHAT THE ABOVE SHOULD BE:  63 - spec.db_start_bit
                    or
--                spec.db_bit_start,
                64 - (spec.db_bit_start + spec.db_bit_length),
            rshift_bits = 64 - spec.db_bit_length,
            start_stamp = self.start_stamp,
            scale = slope,
            offset = intercept,
        }

        w.cleanup = function()
            self:stop_stream()
            -- Remove from CAN Signal table
            local list = self.can_signal_channel[spec.db_msg_id]
            if list then
                for i, info in ipairs(list) do
                    if info.connector == spec.connector then
                        table.remove(list, i)
                        return
                    end
                end
            end
        end

    else
        -- Special 'sink' for raw CAN
        sink = ffi_new('can_sink', self.start_stamp, w.socket, can_buf_count, 0)

        -- Insert into Channel Sink table
        table.insert(self.channel_sinks, sink)
        self.channel_sinks[spec.connector] = sink

        w.cleanup = function()
            self:stop_stream()
            -- Remove from Channel Sink table
            for i = #self.channel_sinks, 1, -1 do
                if self.channel_sinks[spec.connector] == sink then
                    table.remove(self.channel_sinks, i)
                end
            end
            self.channel_sinks[spec.connector] = nil
        end
    end
    md_sink(spec)
    self:start_stream(w, spec)
end

function can_proxy:available_channels(hostname)
    local o = {}
    -- These two channels always exist at the proxy level
    table.insert(o, {
        name = 'raw_' .. hostname:gsub("%W", "_"),
        connector = '@' .. hostname,
        description = "Raw CAN messages",
        data_type = 'message_can',
        category = 'source/hbm_cx23r/msg_can',
    })
    table.insert(o, {
        name = 'control_' .. hostname:gsub("%W", "_"),
        connector = 'control.status@' .. hostname,
        description = 'CAN Control Channel',
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/msg_can',
    })
    return o
end

return can_proxy
