local can = require 'can'
local ffi = require 'ffi'
local listener = require 'listener'
local dbus = require 'dbus'
local sample_clock = require 'sample_clock'
local bus = dbus.bus_get('system')
local timing = assert(bus:auto_proxy('com.somat.armada.Timing',
                                     '/com/somat/armada/Timing'))
local clkfreq = sample_clock.master_rate
local cmd_timeout = 0.01

local function interactive_power_test(devname)
    local can_device = can:new(devname, "can.bin", clkfreq)
    io.write("Ready to turn on passthru power") io.read()
    can_device:txdr_passthru_power_on()
    assert(listener({can_device}, cmd_timeout))
    io.write("Ready to turn off passthru power") io.read()
    can_device:txdr_passthru_power_off()
    assert(listener({can_device}, cmd_timeout))
    io.write("Off now") io.read()
    can_device:halt()
    print("Done")
    os.exit(0)
end

local function get_ts_test(devname)
    local can_device = can:new(devname, "can.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get current timestamp") io.read()
        can_device:get_current_ts()
        local msg = assert(listener({can_device}, cmd_timeout))
        print(msg.ts)
    end
    can_device:halt()
    print("Done")
    os.exit(0)
end

local function get_sync_test(devname)
    local can_device = can:new(devname, "can.bin", clkfreq)
    for i=1,5 do
        io.write("Press key to get latched timestamp") io.read()
        can_device:get_sync_ts()
        local msg = assert(listener({can_device}, cmd_timeout))
        print(msg.ts)
    end
    can_device:halt()
    print("Done")
    os.exit(0)
end

local function synchronization_test(devname)
    local can_device = can:new(devname, "can.bin", clkfreq)
    can_device:setup_sync_ts()
    assert(listener({can_device}, cmd_timeout))
    local sync_time = timing:SendSyncPulse()
    can_device:get_sync_ts()
    local msg = assert(listener({can_device}, cmd_timeout))
    local offset = sync_time - msg.ts + 1
    can_device:ts_offset(offset)
    assert(listener({can_device}, cmd_timeout))
    can_device:setup_sync_ts()
    assert(listener({can_device}, cmd_timeout))
    sync_time = timing:SendSyncPulse()
    can_device:get_sync_ts()
    msg = assert(listener({can_device}, cmd_timeout))
    offset = sync_time - msg.ts
    if offset == 0 then
        print("Passed")
    else
        print("Failed, offset is", offset)
    end
    can_device:halt()
    print("Done")
    os.exit(0)
end

local function display_heartbeat_messages(channel_obj, name, ts)
    io.write("HB ") io.flush()
    channel_obj.heartbeat_count = channel_obj.heartbeat_count + 1
end

local function heartbeat_test(devname)
    local can_device = can:new(devname, "can.bin", clkfreq)
    can_device.heartbeat_func = display_heartbeat_messages
    can_device.heartbeat_count = 0
    can_device:get_current_ts()
    local msg = assert(listener({can_device}, cmd_timeout))
    local start_time = msg.ts + clkfreq
    can_device:event_start(start_time - 3, 4, 16, 128)
    assert(listener({can_device}, cmd_timeout))
    listener({can_device}, 10.0)
    print("\nCleaning up")
    can_device:event_stop()
    assert(listener({can_device}, cmd_timeout))
    can_device:halt()
    print("Heartbeats detected = "..can_device.heartbeat_count)
    print("Done")
    os.exit(0)
end

local function display_can_message(channel_obj, name, pktstruct)
    local ff = pktstruct.frame_format
    local ecc = pktstruct.error_capture_code
    local rxe = pktstruct.rx_error_count
    local txe = pktstruct.tx_error_count
    local fmt = "CAN %s => FF=%02X ECC=%02X RXE=%02X TXE=%02X "
    local ext_frame = false
    if ff > 127 then
        ff = ff - 128
        ext_frame = true
    end
    local rtr = false
    if ff > 63 then
        ff = ff - 64
        rtr = true
    end
    local direction = "RX"
    if ff > 15 then
        ff = ff - 16
        direction = "TX"
    end
    io.write(string.format(fmt, direction, ff, ecc, rxe, txe))
    print("TS "..tonumber(pktstruct.ts))
    if ext_frame then
        io.write(string.format("          ID=%08X MSG=", pktstruct.can_id))
    else
        io.write(string.format("          ID=%04X MSG=", pktstruct.can_id))
    end
    for di=1,ff do
        io.write(string.format("%02X ", pktstruct.can_message[di - 1]))
    end
    print("")
end

local function setup_device(can_device, terminate, br, sjw, tseg1, tseg2)
    can_device.element_action_func = display_can_message
    can_device:bus_setup({ baud_rate = br,
                           sjw = sjw,
                           tseg1 = tseg1,
                           tseg2 = tseg2 })
    assert(listener({can_device}, cmd_timeout))
    if terminate == 'true' then
        can_device:enable_terminator()
        assert(listener({can_device}, cmd_timeout))
    end
    can_device:get_current_ts()
    local msg = assert(listener({can_device}, cmd_timeout))
    if not msg then
        io.stderr:write("listener(" .. can_device.devname ..
                                  ") returned nil msg\n")
        os.exit(1)
    end
    local start_time = msg.ts + clkfreq
    can_device:rx_start(start_time - 3, 2, 16, 128)
    assert(listener({can_device}, cmd_timeout))
    can_device:tx_start(0, 16, 128)
    assert(listener({can_device}, cmd_timeout))
end

local function monitor_test(devname, terminate, br, sjw, tseg1, tseg2)
    local can_device = can:new(devname, "can.bin", clkfreq)
    setup_device(can_device,
                 terminate,
                 tonumber(br),
                 tonumber(sjw),
                 tonumber(tseg1),
                 tonumber(tseg2))
    -- Run for 10.0 seconds
    assert(listener({ can_device }, 10.0))
    can_device:halt()
    print("Done")
    os.exit(0)
end

local function usage ()
    print("Usage: can_test <device name> [POWER_TEST|GET_TS|GET_SYNC|" ..
          "SYNC_TEST|HEARTBEAT_TEST]")
    print("       can_test <device name> [true|false] <baud> <sjw> " ..
          "<tseg1> <tseg2>")
end

if #arg < 1 then
    usage()
    os.exit(1)
end

if #arg >= 1 then
    if arg[2] == "HEARTBEAT_TEST" then heartbeat_test(arg[1]) end
    if arg[2] == "POWER_TEST" then interactive_power_test(arg[1]) end
    if arg[2] == "GET_TS" then get_ts_test(arg[1]) end
    if arg[2] == "GET_SYNC" then get_sync_test(arg[1]) end
    if arg[2] == "SYNC_TEST" then synchronization_test(arg[1]) end
    if arg[2] == "HEARTBEAT_TEST" then heartbeat_test(arg[1]) end
    if arg[2] == 'true' or arg[2] == 'false' then
        monitor_test(arg[1], arg[2], arg[3], arg[4], arg[5], arg[6])
        os.exit(0)
    end
    usage()
    os.exit(1)
end
