--! \file

local armada = require 'armada'
local dbus = require 'dbus'
local ev = require 'ev'
local http_server = require 'http_server_noffi'
local u = require 'unix'
local cjson = require 'cjson'
local ts = require 'timestamps'
local data_dumper = require 'data_dumper'
local net = require 'net_noffi'
local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
struct dhdr {
    unsigned int        len;
    long long           ts;
    unsigned char       data[0];
} __attribute__((packed));

struct my_iovec {
    void        *iov_base;
    size_t      iov_len;
};

typedef int ssize_t;
ssize_t send(int sockfd, const void *buf, size_t len, int flags);
ssize_t writev(int fd, const struct my_iovec *, int num);

union htonll_u {
    double d;
    uint64_t li;
    uint32_t i[2];
};
]]

local iovec = ffi.new('struct my_iovec[2]')
local dhdr = ffi.new('struct dhdr[1]')

ffi.fill(iovec, ffi.sizeof(iovec))

iovec[0].iov_len = ffi.sizeof('struct dhdr')
iovec[0].iov_base = dhdr

local dhdr_size = ffi.sizeof('struct dhdr')

--! The base channel proxy implementation.
--? class channel_proxy
local m = class('channel_proxy')
--? alias m channel_proxy

function m:initialize(name, no_module_hostname)
    self.proxy_name = name
    assert(self.proxy_type)

    ev.Signal.new(function() -- Ignore SIGPIPE -- KLUDGE move to server.lua
--        print('SIGPIPE')
    end, u.SIGPIPE):start(ev.Loop.default)
    ev.Signal.new(function() -- Allow single Ctrl-C to kill
        print('SIGINT')
        ev.Loop.default:unloop()
    end, u.SIGINT):start(ev.Loop.default)

    self.module_hostname = armada.network:GetModuleHostname()
    self.module_ip = armada.network:GetModuleIP()
    self.channel_listen_ip = armada.network:GetModuleLinkLocalIP()
    self.channel_setup_ip = self.channel_listen_ip:gsub("%%.*$", "")

    if not no_module_hostname then
        self.proxy_hostname = string.format("%s.%s",
            self.proxy_name, self.module_hostname)
    else
        self.proxy_hostname = self.proxy_name
    end

--[[
    self:logf("notice", "%s channel_proxy starting up: %s",
        self.proxy_type, self.proxy_hostname)
]]

    self.proxy_address = self.module_ip

    local http = http_server:new{
        ev_loop = ev.Loop.default,
        handler = self.http_handler,
        add_to_watchers = {
            channel_proxy = self,
        }
    }
    local http_w = http:listen{address = self.proxy_address}
    self.proxy_port = http_w.port
--[[
    self:logf("notice", "Listening for HTTP on %s %d",
        self.proxy_address, http_w.port)
]]

    local ch_list = self:available_channels(self.proxy_hostname)

    local host = self.proxy_hostname
    self.old_available_channels = cjson.encode(ch_list)
    armada.network:ModifyRRs({}, {
        {'_armada-connectors', 'PTR', host},
        {host, 'SRV', '0 0 ' .. http_w.port .. ' ' .. self.module_hostname},
        {host, 'TXT', 'type: ' .. self.proxy_type},
        {host, 'TXT', 'available_channels: ' .. self.old_available_channels},
    })
end

function m:logf(level, fmt, ...)
    armada.logger:Log('channel_proxy('..self.proxy_name..')', level,
        string.format(fmt, ...))
end

function m:write_metadata(w, metadata)
    u.write(w.socket, net.packmsg("\128\000\000\000\000\000\000\000" ..
        cjson.encode(metadata)))
    if metadata.error_message then
        w.server.http_write_shutdown(loop, w)
    end
end

function m.http_handler(loop, w, http)
--print("channel_proxy http_handler, http=" .. data_dumper(http, "", true))
    local self = assert(w.channel_proxy)
    if http.uri:match("^/update_settings/") then
--print("channel_proxy UPDATE_SETTINGS: " .. data_dumper(cjson.decode(http.content), "", true))
        local result = self:update_settings(loop, w, http, cjson.decode(http.content))
        w.write(
            "HTTP/1.0 200 OK\r\n"
            .. "Content-Type: application/json\r\n"
            .. "\r\n"
            .. cjson.encode(result)
            .. "\n"
        )
        result.acknowledged = "ACK"
        u.write(w.socket, cjson.encode(result))
        u.write(w.socket, "\n")
        w.server.http_write_shutdown(loop, w)
    elseif http.uri:match("^/comet_stream$") then
        w.write(
            "HTTP/1.0 200 OK\r\n"
            .. "Content-Type: application/octet-stream\r\n"
            .. "\r\n"
        )
        local sink = function(timestamp, name, value)
            local o = {[name]=value}
            w.write(cjson.encode(o) .. "\n\004")
        end
        local channels = {}
        if http.headers.content_type == "application/json" then
            channels = cjson.decode(http.content)
        end
        self:start_comet_stream(w, channels, sink) -- KLUDGE for now!!

    elseif http.uri:match("^/channel_stream$") then
        -- "somat:connector": "latitude@gps-1.ppc1",
        -- "core:start_time": "2010-11-09T23:14:39.000000000",

        local spec = cjson.decode(http.content)
        local md_sink = function(metadata)
            u.write(w.socket, net.packmsg("\128\000\000\000\000\000\000\000" ..
                                cjson.encode(metadata)))
            if metadata.error_message then
                self:logf("error", "%s - %s", self.proxy_hostname,
                    metadata.error_message)
                w.server.http_write_shutdown(loop, w)
            end
        end
        local sock = w.socket
        local sink = function(timestamp, payload, plen)
            local len = plen or payload:len()
            dhdr[0].len = len + dhdr_size - 4;  -- Adjusted for length field
            dhdr[0].ts = timestamp
            dhdr[0].len = self.htonl(dhdr[0].len)
            dhdr[0].ts = self.htonll(dhdr[0].ts)
            iovec[1].iov_len = len
            iovec[1].iov_base = ffi.cast('void *', payload)
            C.writev(sock, iovec, 2);
        end

        self:start_channel_stream(w, spec, md_sink, sink) -- KLUDGE for now!!
    end
end

-- endian conversion
local htonll_data = ffi.new('union htonll_u')
if ffi.abi("be") then -- nothing to do
    function m.htonl(b) return b end
    function m.htons(b) return b end
    function m.htonll(b) return b end
    function m.htonll_double(b) return b end
else
    function m.htonl(b) return bit.bswap(b) end
    function m.htons(b) return bit.rshift(bit.bswap(b), 16) end
    function m.htonll_double(b)    -- assumes b is ffi C type double
        htonll_data.d = b
        local i1 = m.htonl(htonll_data.i[1])
        local i0 = m.htonl(htonll_data.i[0])
        htonll_data.i[1] = i0
        htonll_data.i[0] = i1
        return htonll_data.d
    end
    function m.htonll(b)    -- assumes b is ffi C type long long
        htonll_data.li = b
        local i1 = m.htonl(htonll_data.i[1])
        local i0 = m.htonl(htonll_data.i[0])
        htonll_data.i[1] = i0
        htonll_data.i[0] = i1
        return htonll_data.li
    end
end
m.ntohl = m.htonl -- reverse is the same
m.ntohs = m.htons -- reverse is the same
m.ntohll = m.htonll -- reverse is the same
m.ntohll_double = m.htonll_double -- reverse is the same

return m
