local pcre = require 'pcre_ffi'
local sample_types = require 'sample_types'
local setup_spec = require 'setup_spec'

local filter_re = pcre:new('^filter(/.*)?$')
local sink_re = pcre:new('^sink(/.*)?$')
local sink_sie_re = pcre:new('^sink/sie(/.*)?$')
local source_re = pcre:new('^source(/.*)?$')

local m = {}

local acc = {}

function acc.is_fixed_size(self)
    return sample_types.is_fixed_size(self.data_type)
end

function acc.is_message(self)
    return sample_types.is_message(self.data_type)
end

function acc.is_sequential(self)
    return sample_types.is_sequential(self.data_type)
end

function acc.data_class(self)
    return sample_types.data_class(self.data_type)
end

function acc.is_numeric(self)
    return sample_types.is_numeric(self.data_type)
end

function acc.is_filter(self)
    return filter_re:match(self.meta.category)
end

function acc.is_sink(self)
    return sink_re:match(self.meta.category)
end

function acc.is_sink_sie(self)
    return sink_sie_re:match(self.meta.category)
end

function acc.is_source(self)
    return source_re:match(self.meta.category)
end

function acc.has_heartbeats(self)
    return sample_types.has_heartbeats(self.data_type)
end

-- These channels are explitily excluded from the stream (for now)
local exclude_channel_from_stream = {
    ['elapsed_test_run_time.status'] = true,
    ['sie_file_size.status'] = true,
}

-- Add to these as more are supported
local streamable_types = {
    ['sequential_float32'] = true,
    ['sequential_float64'] = true,
    ['message_float32'] = true,
    ['message_float64'] = true,
}

function acc.is_streamable(self)
    if not sink_re:match(self.meta.category) and
        not self.internal_name and
        not exclude_channel_from_stream[self.meta.name] and
        streamable_types[self.data_type] then
        return true
    end
end

function acc.data_type(node)
    if node.meta.data_type then
        return node.meta.data_type
    end
    -- This bit is inherently (and mutually) recursive
    local edges = node.input_edges
    local dt = {}
    for i = 1, #edges do
        dt[i] = edges[i].data_type
        if i ~= 1 and dt[i] ~= dt[1] then
            error(string.format(
                'Mismatched input data types, node[%s], category=%s'
                .. ', edges[1].data_type=%s, edges[%s].data_type=%s',
                node.idx, node.meta.category, edges[1].data_type,
                i, edges[i].data_type))
        end
    end
    return dt[1]
end

function acc.input_edges(node)
    local edges = node.compiler.edges
    local r = {}
    for i = 1, #edges do
        if edges[i].to_idx == node.idx then
            r[#r + 1] = edges[i]
        end
    end
    return r
end

function acc.output_edges(node)
    local edges = node.compiler.edges
    local r = {}
    for i = 1, #edges do
        if edges[i].from_idx == node.idx then
            r[#r + 1] = edges[i]
        end
    end
    return r
end

function acc.input_nodes(node)
    local edges = node.compiler.edges
    local nodes = {}
    for i = 1, #edges do
        if edges[i].to_idx == node.idx then
            nodes[#nodes + 1] = node.compiler.nodes[edges[i].from_idx]
        end
    end
    return nodes
end

function acc.output_nodes(node)
    local edges = node.compiler.edges
    local nodes = {}
    for i = 1, #edges do
        if edges[i].from_idx == node.idx then
            nodes[#nodes + 1] = node.compiler.nodes[edges[i].to_idx]
        end
    end
    return nodes
end

local function terminal_nodes(node, traverser)
    local candidates = {node}
    local results = {}
    for i = 1, 10000 do -- Cycle safety
        local done = true
        local new_candidates = {}
        for _, n in ipairs(candidates) do
            local cl = n[traverser]
            if #cl == 0 then
                results[#results + 1] = n
            else
                done = false
                for _, nn in ipairs(cl) do
                    new_candidates[#new_candidates + 1] = nn
                end
            end
        end
        if done then
            return results
        end
        candidates = new_candidates
    end
end

function acc.source_nodes(node)
    return terminal_nodes(node, 'input_nodes')
end

function acc.sink_nodes(node)
    return terminal_nodes(node, 'output_nodes')
end

function acc.input_count(node)
    local o = {}
    for i, edge in ipairs(node.input_edges) do
        o[#o + 1] = string.format('%s.count', edge.buf)
    end
    if #o == 1 then
        return o[1]
    end
    return string.format('math_min(%s)', table.concat(o, ', '))
end

function acc.output_available_count(node)
    local o = {}
    for i, edge in ipairs(node.output_edges) do
        o[#o + 1] = string.format('%s:available()', edge.buf)
    end
    if #o == 1 then
        return o[1]
    end
    return string.format('math_min(%s)', table.concat(o, ', '))
end

function acc.output_available_count_expand(node)
    local o = {}
    for i, edge in ipairs(node.output_edges) do
        o[#o + 1] = string.format('%s:available(true)', edge.buf)
    end
    if #o == 1 then
        return o[1]
    end
    return string.format('math_min(%s)', table.concat(o, ', '))
end

function acc.output_edge_list(node)
    local o = {}
    for i, edge in ipairs(node.output_edges) do
        o[#o + 1] = string.format('%s', edge.buf)
    end
    if #o == 1 then
        return o[1]
    end
    return string.format('ipairs({%s})', table.concat(o, ', '))
end

function acc.processable_count(node)
    local o = {}
    for i, edge in ipairs(node.input_edges) do
        local buf = edge.buf
        if buf then
            o[#o + 1] = string.format('%s.count', buf)
        end
    end
    for i, edge in ipairs(node.output_edges) do
        local buf = edge.buf
        if buf then
            o[#o + 1] = string.format('%s:available()', buf)
        end
    end
    if #o == 1 then
        return o[1]
    end
    return string.format('math_min(%s)', table.concat(o, ', '))
end

local fn = {}

function fn.dump(self)
    return setup_spec:item_export(self)
end

function fn.only_sink_node(node)
    if node.is_source and #node.output_nodes == 1 then
        local sink_node = node.output_nodes[1]
        if sink_node.meta.category == 'sink/sie/time_history' then
            return sink_node
        end
    end
    return nil
end

function fn.upstream_meta(node, tag)
    local inputs = {node}
    local values = {}
    if node.meta[tag] then
        return {node.meta[tag]}
    end
    for i = 1, 10000 do -- Cycle safety
        local done = true
        local new_inputs = {}
        for _, n in ipairs(inputs) do
            local tin = n.input_nodes
            for _, nn in ipairs(tin) do
                if nn.meta[tag] then
                    values[nn.meta[tag]] = true
                else
                    new_inputs[#new_inputs + 1] = nn
                    done = false
                end
            end
        end
        if done then
            local r = {}
            for k, v in pairs(values) do
                r[#r + 1] = k
            end
            return r
        end
        inputs = new_inputs
    end
    return {}
end

function fn.set_internal_name(node)
    node.meta.name = node.compiler:uuid()
    node.internal_name = true
end

local node_mt = {
    __index = function(table, key)
        if acc[key] then
            return acc[key](table)
        end
        return fn[key]
    end,
}

function m:new(compiler, meta, extra)
    fn.compiler = compiler
    local node = {
        idx = #compiler.nodes + 1,
        meta=meta,
        -- For walker
        inputs = {},
        input_names = {},
    }
    for k, v in pairs(extra) do
        node[k] = v
    end
    setmetatable(node, node_mt)
    if not node.meta.name then
        node:set_internal_name()
    end
    node.compiler.nodes[node.idx] = node
    return node
end

return m
