require 'strict'

local m = {}

function m:oc_buffered(buf_name, ...)
    local args = {...}
    self.buffers[buf_name] = self.buffers[buf_name] or {}
    local buf = self.buffers[buf_name]
    if type(args[1]) ~= 'table' then
        table.insert(args, 1, {})
    end
    local opt = args[1]
    opt.output = false
    opt.indent = args[1].indent or ''
    if opt.inline then
        opt.indent = ''
    end
    local value = self:template(unpack(args))
    if opt.inline then
        value = value:gsub('^%s*(.-)%s*$', '%1')
    else
        value = value:gsub('([^\n])$', "%1\n")
    end
    if opt.no_lhs then
        buf[#buf + 1] = value
        return value
    end
    if opt.key then
        local name = string.format("%s.%s", buf_name, opt.key)
        buf[name] = string.format("%s = %s", name, value)
        return name, value
    end
    if opt.op_local then
        local name = opt.op_local
        buf[name] = string.format("local %s = %s", name, value)
        return name, value
    end
    local idx
    if opt.unique then
        for i = 1, #buf do
            if value == buf[i] then
                idx = i
                break
            end
        end
    end
    local name = string.format("%s[%d]", buf_name, idx or #buf + 1)
    if not idx then
        table.insert(buf, string.format("%s = %s", name, value))
    end
    return name, value, idx or #buf
end

function m:oc_declare(...)
    return self:oc_buffered('decl', ...)
end

function m:oc_cached(namespace, key, generator)
    self.cache[namespace]= self.cache[namespace] or {}
    if not self.cache[namespace][key] then
        self.cache[namespace][key] = generator(self, key)
    end
    return self.cache[namespace][key]
end

function m:oc_can_constant(hex)
    return self:oc_cached('can_constant', hex:lower(), function(self, v)
        return self:oc_declare("message_can:init('(=v=)')", {v = v})
    end)
end

function m:oc_sie_buf()
    local idx = self.sie_buf_idx or 1
    self.sie_buf_idx = idx + 1
    return string.format('sie_bufs[%s]', idx), idx
end

function m:oc_finalize(...)
    return self:oc_buffered('finalize', {no_lhs = true}, ...)
end

function m:oc_expression(node)
    if not node.parse_tree then
        return
    end
    node.input_names = {}
    for _, edge in ipairs(node.input_edges) do
        local accessor
        if edge.is_logical then
            accessor = '.value[0] ~= 0'
        elseif edge.from_node.is_numeric then
            accessor = '.value[0]'
        else
            accessor = ''
        end
        node.input_names[edge.from_name] = string.format('(t_%d%s)',
            edge.idx, accessor)
    end
    -- KLUDGE FIX CONSTANT NAME r_1
    self:template([==[
    |        local r_1 = (=lua_expr=)(=xform_expr=)
    ]==], {
        lua_expr = self:walk(node, node.parse_tree),
    }, node)
end

function m:oc_input_node(node, tv)
    self:template([==[
    |    md = [====[(=cjson_md=)]====]
    |    self:create_input_node(md, (=reader=), (=pbuf=))
    ]==], tv)
end

function m:oc_proxy_buffer(node, tv)
    tv.proxy_buf_size = tv.proxy_buf_size or
        '(=buf_size=) * (=proxy_sample_size=)'
    tv.proxy_sample_size = tv.proxy_sample_size or "(=sizeof'(=data_type=)'=)"
    tv.proxy_timestamped_samples = tv.proxy_timestamped_samples or 0
    tv.float_sample_rate = node.float_sample_rate or 0

    local _
    tv.pbuf, _, node.pbuf_idx = self:oc_buffered('proxy_buffers', [==[
    |proxy_buffer:new((=proxy_buf_size=), (=proxy_sample_size=),
    |+   "(=meta.connector=)", (=proxy_timestamped_samples=),
    |+   (=float_sample_rate=))
    ]==], tv, node)
end

function m:oc_proxy_reader_loop(node, data_handler, tv, ...)
    -- KLUDGE set indent, linefeed fixes
    local _
    _, tv.data_handler = self:oc_buffered('_oc_buffered_tmp', data_handler,
        tv, node)

    tv.sbuf = ''
    -- KLUDGE consolidate oc_conditional_sie_buf_boilerplate()
    if node:only_sink_node() then
        tv.sbuf = self:tmpl([==[
        |    local sbuf = (=sie_buf=) -- (=sie_meta._sie_name=)
        ]==], tv, node:only_sink_node())

        self:oc_finalize([==[
        |    (=sie_buf=):flush() -- (NOS) (=sie_meta._sie_name=)
        ]==], tv, node:only_sink_node())
        node:only_sink_node().only_sink_cleanup = true
    end

    self:push_indent(8)
    local mh = self:template({output = false}, [==[
    |    -- KLUDGE not currently relevant
    |    -- pb.xform_scale = 1
    |    -- pb.xform_offset = 0
    ]==])
    local sink_nodes = node.sink_nodes
    local flush_xml = false
    for _, onode in ipairs(sink_nodes) do
        -- KLUDGE this is now in the wrong place
        if onode.sie_meta and onode.xform_expr ~= '' then
            -- KLUDGE test id and ch id are very wrong for multiple runs
            -- KLUDGE PRECISION!!!!!!
            mh = mh .. self:template({output = false}, [==[
            |    test_engine.sw:xml(
            |        [[<test id="0"><xform ch="(=_sie_ch=)" dim="1"
            |+           scale="(=finalized_xform.m=)"
            |+          offset="(=finalized_xform.b=)"/></test>]])
            ]==], onode.sie_meta, onode)
            flush_xml = true
        end
    end
    if flush_xml then
        mh = mh .. self:template({output = false}, [==[
        |    -- KLUDGE Don't flush here once it can be done all at once
        |    test_engine.sw:xml()
        ]==])
    end
    self:pop_indent()
    tv.meta_handler = mh

    tv.pre_loop = tv.pre_loop or ''
    tv.post_loop = tv.post_loop or ''

    tv.reader = self:oc_buffered('proxy_readers', [==[
    |function(loop, w)
    |    local pb = (=pbuf=)
    |    (=sbuf=)
    |*
    |(=pre_loop=)
    |*
    |    ::there::
    |        local inbuf, count, ts = pb:read_from_proxy(test_engine,
    |+           '(=data_type=)_buf *')
    |        if count then
    |(=data_handler=)
    |*
    |            check_stop_bracket(test_engine, pb.bracket, (=pbuf_idx=))
    |        elseif inbuf ~= nil then -- Metadata
    |            local md = inbuf -- KLUDGE fix naming mismatch
    |(=meta_handler=)
    |*
    |        else -- finished
    |            if pb.failed ~= 0 then
    |                w:stop(loop)
    |                test_engine:stop_test() -- KLUDGE
    |            end
    |        end
    |    if inbuf ~= nil then goto there end
    |(=post_loop=)
    |*
    |end
    |
    ]==], tv, node)

    self:oc_input_node(node, tv)
end

function m:oc_conditional_sie_buf_boilerplate(node)
    if node.is_sink_sie then
        assert(#node.input_edges == 1) -- KLUDGE better error handling?
        self:template([==[
        |    local sbuf = (=sie_buf=) -- (=_sie_name=)
        ]==], node.sie_meta, node)

        self:oc_finalize([==[
        |    (=sie_buf=):flush() -- (=_sie_name=)
        ]==], node.sie_meta, node)
    end
end

function m:oc_source_sequential_to_sie(node, tv)
    node.output_nodes[1].only_sink_cleanup = true

    self:oc_proxy_buffer(node, tv)

    self:oc_proxy_reader_loop(node, [==[
    |            -- oc_source_sequential_to_sie()
    |            sbuf.timestamp[1] = sbuf.timestamp[0]
    |            local i, bc = pb.bracket:bracket_sequential(sbuf.timestamp[0],
    |+               0, count)
    |            ::here::
    |            if i < bc then
    |                if sbuf:pos() == 0 then
    |                    sbuf:enq(sbuf.timestamp, (=sizeof'timestamp'=))
    |                end
    |                local writable = sbuf:elements_available()
    |                local towrite = math_min(writable, bc - i)
    |                sbuf:enq(inbuf + i, (=sizeof'(=data_type=)'=) * towrite)
    |                i = i + towrite
    |                if towrite == writable then
    |                    sbuf:flush()
    |                    sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + i
    |                end
    |                goto here
    |            end
    |            sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + bc
    ]==], tv, node)
end

function m:oc_source_sequential_to_bufs(node, tv)
    tv.proxy_timestamped_samples = 0
    self:oc_proxy_buffer(node, tv)

    local oe = {}

    if node.is_numeric then -- KLUDGE do static xform check here
        oe[#oe + 1] = self:template({output = false}, [==[
        |                    local v = inbuf[j].value[0](=xform_expr=)
        ]==], node)
    else
        oe[#oe + 1] = self:template({output = false}, [==[
        |                    local v = inbuf[j] -- is_numeric FALSE
        ]==])
    end

    for i, edge in ipairs(node.output_edges) do
        if edge.data_type ~= tv.data_type then
            error('NYI: No data type conversion yet')
        end
        oe[#oe + 1] = self:template({output = false}, [==[
        |                    (=edge.buf=):enq()[0].value[0] = v
        ]==], {edge = edge})
    end

    tv.out_enqueue = table.concat(oe)

    self:oc_proxy_reader_loop(node, [==[
    |            local i, bc =
    |+               pb.bracket:bracket_sequential(pb.bracket.in_count,
    |+                   0, count)
    |            ::here::
    |            if i < bc then
    |                local towrite =
    |+                   math_min(bc - i, (=output_available_count_expand=))
    |                if towrite <= 0 then
    |                    log(log.ERR, __'stop_test "{0}" edge buffer full',
    |                        test_engine.test_name)
    |                    loop:unloop()
    |                    return
    |                end
    |                for j = i, i + towrite - 1 do
    |(=out_enqueue=)
    |*
    |                end
    |                i = i + towrite
    |                test_engine:run()
    |                goto here
    |            end
    ]==], tv)
end

function m:oc_source_fixed_message_to_sie(node, tv)
    node.output_nodes[1].only_sink_cleanup = true

    tv.proxy_timestamped_samples = 1
    self:oc_proxy_buffer(node, tv)

    self:oc_proxy_reader_loop(node, [==[
    |            -- oc_source_fixed_message_to_sie()
    |            for i = 0, count - 1 do
    |                if not inbuf[i]:is_heartbeat() and
    |                        pb.bracket:is_in_bracket(inbuf[i].timestamp[0])
    |                        then
    |                    sbuf:flush_unless_available((=sizeof'(=data_type=)'=))
    |                    sbuf:enq(inbuf + i, (=sizeof'(=data_type=)'=))
    |                end
    |            end
    ]==], tv, node)
end

function m:oc_source_fixed_message_no_hb_to_sie(node, tv)
    node.output_nodes[1].only_sink_cleanup = true

    tv.proxy_timestamped_samples = 1
    self:oc_proxy_buffer(node, tv)

    self:oc_proxy_reader_loop(node, [==[
    |        local i = 0
    |        ::here::
    |        if i < count then
    |            local writeable = sbuf:elements_available()
    |            local towrite = math_min(writeable, count - i)
    |            sbuf:enq(inbuf + i, (=sizeof'(=data_type=)'=) * towrite)
    |            i = i + towrite
    |            if towrite == writeable then
    |                sbuf:flush()
    |            end
    |            goto here
    |        end
    ]==], tv, node)
end

function m:oc_source_fixed_message_to_bufs(node, tv)
    tv.proxy_timestamped_samples = 1
    self:oc_proxy_buffer(node, tv)

    local oe = {}

    oe[#oe + 1] = self:template({output = false}, [==[
    |                    local ts = inbuf[j].timestamp[0]
    |                    pb.bracket:is_in_bracket(ts)
    ]==])
    if node.is_numeric then -- KLUDGE do static xform check here
        oe[#oe + 1] = self:template({output = false}, [==[
        |                    local v = inbuf[j].value[0](=xform_expr=)
        ]==], node)
    end

    for i, edge in ipairs(node.output_edges) do
        if edge.data_type ~= tv.data_type then
            error('NYI: No data type conversion yet')
        end
        if node.is_numeric then
            oe[#oe + 1] = self:template({output = false}, [==[
            |                    do
            |                        local ov = (=edge.buf=):enq()
            |                        ov.timestamp[0] = ts
            |                        ov.value[0] = v
            |                    end
            ]==], {edge = edge})
        else
            oe[#oe + 1] = self:template({output = false}, [==[
            |                        (=edge.buf=):enq()[0] = inbuf[j]
            ]==], {edge = edge})
        end
    end

    tv.out_enqueue = table.concat(oe)

    self:oc_proxy_reader_loop(node, [==[
    |            local i = 0
    |            ::here::
    |            if i < count then
    |                local towrite =
    |+                   math_min(count - i, (=output_available_count=))
    |                for j = i, i + towrite - 1 do
    |(=out_enqueue=)
    |*
    |                end
    |                i = i + towrite
    |                test_engine:run()
    |                goto here
    |            end
    ]==], tv)
end

function m:oc_source_var_message_to_sie(node, tv)
    node.output_nodes[1].only_sink_cleanup = true

    tv.proxy_sample_size = 1
    self:oc_proxy_buffer(node, tv)

    self:oc_proxy_reader_loop(node, [==[
    |            if count > 0 and pb.bracket:is_in_bracket(ts) then
    |                local size = count + (=sizeof'proxy_header'=)
    |                sbuf:flush_unless_available(size)
    |                -- KLUDGE skipping back in inbuf like this is way horrible
    |                sbuf:enq(inbuf - (=sizeof'proxy_header'=), size)
    |            end
    ]==], tv)
end

function m:oc_source_var_message_to_bufs(node, tv)
    tv.proxy_sample_size = 1
    self:oc_proxy_buffer(node, tv)

    local oe = {}
    for i, edge in ipairs(node.output_edges) do
        edge.value_malloced = true

        oe[#oe + 1] = self:template({output = false}, [==[
        |            do
        |                local p = (=edge.buf=):enq()
        |                -- KLUDGE the following should be done better,
        |                -- potential alignment traps as-written
        |                p.length = count + (=sizeof'timestamp'=)
        |                p.timestamp[0] = ts
        |                p.value = C.malloc(count)
        |                ffi.copy(p.value, inbuf, count)
        |            end
        ]==], {edge = edge})
    end
    tv.out_enqueue = table.concat(oe)

    tv.post_loop = self:template({output = false}, [==[
    |    test_engine:run()
    ]==])

    self:oc_proxy_reader_loop(node, [==[
    |            if (=output_available_count=) == 0 then
    |                test_engine:run()
    |                assert((=output_available_count=) > 0)
    |            end
    |            pb.bracket:is_in_bracket(ts)
    |(=out_enqueue=)
    ]==], tv, node)
end

function m:oc_init_sink_sie(node, tv)
    local meta = node.sie_meta
    if not meta._sie_buffer_size then
        -- KLUDGE specialize on type
        meta._sie_buffer_size = -- KLUDGE!!!!
            node.is_fixed_size and self.default_sie_buffer_size or 0x8000
    end
    meta.sie_data_type = node.data_type
    if meta.sie_data_type == 'message_raw' then -- KLUDGE
        meta.sie_data_type = 'uint8_t'
    end
    meta.sie_payload_overhead = node.is_sequential
        and "(=sizeof'timestamp'=)" or 0
    tv.cjson_setup_meta = node.sie_meta['somat:armada_input_metadata']
    self:template([==[
    |    do
    |        local md = cjson.decode([====[(=cjson_setup_meta=)]====])
    |        md._sie_test = 0 -- KLUDGE WRONG
    |        local ch_md = self.sw:channel(md)
    |        self.sw:new_buffer((=sie_buf_idx=), ch_md._sie_group,
    |+           (=_sie_buffer_size=), (=sizeof'(=sie_data_type=)'=),
    |+           (=sie_payload_overhead=))
    |+           -- (=_sie_ch=) => (=_sie_name=)
    |    end
    ]==], tv, node.sie_meta, node)
end

function m:oc_process_module(node)
    local tv = {}
    if node.mod.cm_run then
        -- The gsub() is to fix twice-applied initial indentation
        tv.mod_run = node.mod:cm_run(node.mod_name):gsub('^    ', '')
        self:template([==[
        |(=mod_run=)
        |*
        ]==], tv, node)
    end
    if node.mod.cm_finalize then
        tv.mod_finalize = node.mod:cm_finalize(node.mod_name)
        self:oc_finalize([==[
        |(=mod_finalize=)
        ]==], tv, node)
    end
end

function m:oc_process_fixed_message(node)
    -- No general way to combine multiple message input buffers (???)
    assert(#node.input_edges == 1) -- KLUDGE better error handling?
    self:oc_conditional_sie_buf_boilerplate(node)
    local edge = node.input_edges[1]
    local tv = {
        edge = edge,
        t_name = 't_' .. edge.idx,
    }
    self:template([==[
    |    local count = (=processable_count=)
    |    for i = 0, count - 1 do
    |        local (=t_name=) = (=edge.buf=):deq() -- (=edge.from_name=)
    ]==], tv, node)
    if node.is_sink_sie and node.has_heartbeats then
        -- KLUDGE eventually do timestamp deltas and whatnot
        self:template([==[
        |        if not (=t_name=):is_heartbeat() then
        |            sbuf:flush_unless_available((=sizeof'(=data_type=)'=))
        |            sbuf:enq((=t_name=), (=sizeof'(=data_type=)'=))
        |        end
        |    end
        ]==], tv, node, node.sie_meta)
    elseif node.is_sink_sie then
        -- KLUDGE eventually do timestamp deltas and whatnot
        self:template([==[
        |        sbuf:flush_unless_available((=sizeof'(=data_type=)'=))
        |        sbuf:enq((=t_name=), (=sizeof'(=data_type=)'=))
        |    end
        ]==], tv, node, node.sie_meta)
    else
        self:oc_expression(node)
        for _, edge in ipairs(node.output_edges) do
            self:template([==[
            |        if r_1 ~= nil or (=t_name=):is_heartbeat() then
            |+               -- (=edge.to_name=)
            |            local om = (=edge.buf=):enq()
            |            om.timestamp[0] = (=t_name=).timestamp[0]
            |            om.value[0] = r_1 or NaN
            |        end
            ]==], {
                edge = edge,
            }, tv)
        end
        for _, edge in ipairs(node.input_edges) do
            if edge.value_malloced then
                self:template([==[
                |        C.free((=t_name=).value)
                ]==], tv) -- KLUDGE???
            end
        end
        self:template([==[
        |    end
        |    more = more or count > 0
        ]==])
    end
end

function m:oc_process_variable_message(node)
    -- No general way to combine multiple message input buffers
    assert(#node.input_edges == 1) -- KLUDGE better error handling?
    self:oc_conditional_sie_buf_boilerplate(node)
    local edge = node.input_edges[1]
    local tv = {
        edge = edge,
        t_name = 't_' .. edge.idx,
    }
    self:template([==[
    |    local count = (=processable_count=)
    |    for i = 0, count - 1 do
    |        local (=t_name=) = (=edge.buf=):deq() -- (=edge.from_name=)
    ]==], tv, node)
    if node.is_sink_sie then
        -- KLUDGE eventually do timestamp deltas and whatnot
        self:template([==[
        |        local size = (=t_name=).length + (=sizeof'uint32_t'=)
        |        local payload_size = (=t_name=).length - (=sizeof'uint64_t'=)
        |        local writeable = sbuf:elements_available()
        |        if size > writeable then
        |            sbuf:flush()
        |            writeable = sbuf:elements_available()
        |        end
        |        assert(size <= writeable,
        |+           '(=meta.name=) Input buffer too small')
        |        sbuf:enq((=t_name=), (=sizeof'uint32_t'=) +
        |+           (=sizeof'uint64_t'=))
        |        sbuf:enq((=t_name=).value, payload_size)
        |        C.free((=t_name=).value)
        |    end
        ]==], tv, node)
    else
        self:oc_expression(node)
        for i, edge in ipairs(node.output_edges) do
            self:template([==[
            |        local om_(=i=) = (=edge.buf=):enq() -- (=edge.to_name=)
            |        om_(=i=).timestamp[0] = (=t_name=).timestamp[0]
            |        om_(=i=).value[0] = r_1
            ]==], {
                i = i,
                edge = edge,
            }, tv)
        end
        self:template([==[
        |        C.free((=t_name=).value)
        |    end
        |    more = more or count > 0
        ]==], tv)
    end
end

function m:oc_process_fixed_sequential(node)
    self:oc_conditional_sie_buf_boilerplate(node)
    if node.is_sink_sie then
        self:template([==[
        |    sbuf.timestamp[1] = sbuf.timestamp[0]
        |    if sbuf:pos() == 0 then
        |        sbuf:enq(sbuf.timestamp, (=sizeof'timestamp'=))
        |    end
        ]==], node.sie_meta, node)
    end
    self:template([==[
    |    local count = (=processable_count=)
    |    for i = 0, count - 1 do
    ]==], node)
    local last_edge
    for _, edge in ipairs(node.input_edges) do
        last_edge = edge
        local tv = {
            edge = edge,
            t_name = 't_' .. edge.idx,
        }
        -- KLUDGE?
        self:template([==[
        |        local (=t_name=) = (=edge.buf=):deq() -- (=edge.from_name=)
        ]==], tv)
        if node.is_sink_sie then
            -- KLUDGE specialize here on type!!!
            self:template([==[
            |        if sbuf:flush_unless_available((=sizeof'(=data_type=)'=))
            |+               then
            |            sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + i
            |            sbuf:enq(sbuf.timestamp, (=sizeof'timestamp'=))
            |        end
            |        sbuf:enq((=t_name=).value, (=sizeof'(=data_type=)'=))
            ]==], tv, node.meta, node)
        end
    end
    if node.is_sink_sie then
        self:template([==[
        |    end
        |    sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + count
        ]==])
    else
        self:oc_expression(node)
        self:template([==[
        |        if r_1 ~= nil then
        ]==])
        for _, edge in ipairs(node.output_edges) do
            self:template([==[
            |            (=edge.buf=):enq().value[0] = r_1 -- (=edge.to_name=)
            ]==], {
                edge = edge,
            })
        end
        self:template([==[
        |        end
        |    end
        |    more = more or count > 0
        ]==])
    end
end

return m
