local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'
log:open('db-initialize', log.USER)

ffi.cdef[[
void sleep(int);
]]

local function dbg_msg(fmt, ...)
    log:log(log.DEBUG, fmt, ...)
end

-- This is only ever set when the powerd has detected that the user has done
-- the magic "hold the power button down for the magic length of time to reset
-- the administrators credentials" bit.
local reset_admin = arg[1]

local db_client = require 'armadb_client'

local db
local retries = 10
while retries > 0 do
    db = db_client:new({host='127.0.0.1'})
    if db then
        break
    end
    C.sleep(1)
    retries = retries - 1
end
if retries == 0 then
    log:log(log.ERR, "Unable to open database")
    os.exit(1)
end

local query = {authorization_key='locally_authorized', author='admin'}
db:set({query=query})

if not reset_admin then
    local ok = false

    for i = 0, 60 do
        local got = db:tracked_db('PUT', 'users/')
        if got == nil then
            os.execute('sleep 1') -- KLUDGE use syscall here
        else
            ok = got.ok
            if not ok then
                dbg_msg("User database already initialized, no changes made")
                os.exit(1)
            end
            break
        end
    end

    if not ok then
        error("Timeout trying to initialize the user database")
    end

    query.description = "Add user 'admin'"
else
    query.description = "Reset user 'admin'"
end

local admin_role = {
    id="admin",
    description="Administrator",
    test_configuration="edit",
    test_control="all",
    sie_data_files="all",
    calibration="run",
    hardware_control=true,
    logs_control="purge",
    system_configuration={
        profiles=true,
        databases=true,
        networks=true,
        timezone=true,
        firmware=true,
        preferences=true,
        support=true,
    },
}

local system_preferences = {
    version=2,
    system={
        sample_rate_domain="decimal",
        sample_rate_correction_factor="1.0",
        aux_port_mode="edaq_sync",
        teds_enabled=false,
        full_both="down_scale",
        half_both="up_scale",
        quarter_regulated="up_scale",
        quarter_unregulated="down_scale",
        max_test_restarts=0
    },
    user={
        authentication_required=false,
        minimum_password_length=8,
        default_locale="en"
    },
    test={
        data_storage_format="SIE",
        delete_old_data_files_on_test_initialization=false,
        sys_max_sie_tests=1,
        allow_run_if_warnings=true,
        test_duration_prompt=false
    },
    calibration={
        use_calibration_filters=true
    }
}

local config = require 'system_config'
local host_name = {
    id="hostname",
    hostname=config.hostname
}

local default_network = require 'default_network'

retries = 10
while retries > 0 do
    local ret = db:tracked_db('POST', 'users/users/admin', { -- KLUDGE
        id="admin",
        name="Administrator",
        password_sha="ddf32cd4eb1caa4260d0ab626b2b3f1f7dec4a8c", -- "password"
        salt="4d2800fe62271497dfe4975a0e80d6eb",
        role=admin_role,
    })
    if ret and ret.ok then break end
    C.sleep(1)
    retries = retries - 1
end
if retries == 0 then
    log:log(log.ERR, "Unable to set adminstrators credentials to their defaults")
    os.exit(1)
end

retries = 10
while retries > 0 do
    local ret = db:tracked_db('POST', 'users/roles/admin', admin_role)
    if ret and ret.ok then break end
    C.sleep(1)
    retries = retries - 1
end
if retries == 0 then
    log:log(log.ERR, "Unable to set adminstrators profile to their defaults")
    os.exit(1)
end

if not reset_admin then
    retries = 10
    while retries > 0 do
        local ret = db:tracked_db('POST', 'users/system/preferences', system_preferences)
        if ret and ret.ok then break end
        C.sleep(1)
        retries = retries - 1
    end
    if retries == 0 then
        log:log(log.ERR, "Unable to set system preferences to their defaults")
        os.exit(1)
    end

    retries = 10
    while retries > 0 do
        local ret = db:tracked_db('POST', 'users/networks/hostname', host_name)
        if ret and ret.ok then break end
        C.sleep(1)
        retries = retries - 1
    end
    if retries == 0 then
        log:log(log.ERR, "Unable to set inside network defaults")
        os.exit(1)
    end

    retries = 10
    while retries > 0 do
        local ret = db:tracked_db('POST', 'users/networks/default', default_network)
        if ret and ret.ok then break end
        C.sleep(1)
        retries = retries - 1
    end
    if retries == 0 then
        log:log(log.ERR, "Unable to set default network")
        os.exit(1)
    end

    db:tracked_db('PUT', 'setups/')

    db:tracked_db('PUT', 'setup_user_states/') -- KLUDGE should be raw?

    dbg_msg("User database initialized\n")
else
    dbg_msg("Admin credentials reset\n")
end

os.exit(0)
