local __ = require '__'
local database_schemas = require 'database_schemas'
local log = require 'log'
local sqlite3 = require 'lsqlite3'
local cjson = require 'cjson'
local dbg = require 'dbg'

log.log_fh = io.stderr

local function logf(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

local function db_exec(db, fmt, ...)
    local sql = string.format(fmt, ...)
    logf('db_exec, sql=%s\n', sql)
    assert(db:exec(sql) == 0)
end

local function make_backup_fname(orig_fname, version)
    local file_base = (orig_fname:gsub('%.db$', ''))
    return string.format('%s_v%s.db', file_base, version)
end

local function backup_db(orig_fname, backup_fname)
    if type(backup_fname) == 'number' then
        backup_fname = make_backup_fname(orig_fname, backup_fname)
    end
    os.remove(backup_fname) -- ignore error
    os.execute(string.format("sqlite3 '%s' '.backup %s'",
        orig_fname, backup_fname)) -- KLUDGE check status
    local fh = io.open(backup_fname)
    if fh then
        fh:close()
        return true, backup_fname
    end
    return false, backup_fname
end

local function migrate_db(file, schema_spec)
    local db = assert(sqlite3.open(file))
    db_exec(db, 'PRAGMA journal_mode = WAL;')
    db_exec(db, 'PRAGMA synchronous = FULL;')
    local current_version
    for r in db:rows('PRAGMA user_version') do
        current_version = assert(tonumber(r[1]))
    end
    db:close()

    local max_version = -1
    local migratable = {}
    for _, spec in ipairs(schema_spec) do
        migratable[spec.from] = true
        max_version = math.max(max_version, spec.to)
    end

    if current_version ~= max_version then
        backup_db(file, current_version)
    end
    if current_version > max_version then
        -- DOWNGRADE!
        -- look for a "<name>_v<ver>.db" file, where <ver> <= current_version
        for ver = current_version, 1, -1 do
            -- Make sure it's a useful version for us
            if ver == max_version or migratable[ver] then
                local backup_file = make_backup_fname(file, ver)
                local fh = io.open(backup_file, 'r')
                if fh then
                    -- We've found a backup
                    fh:close()
                    -- Copy it to the main database (already backed up above)
                    backup_db(backup_file, file)
                    -- And use the normal migration path
                    current_version = ver
                    -- KLUDGE this is all nasty
                    local msg = log:notice(
                        __"{0}: Using archived {1} due to system downgrade",
                        file, backup_file)
                    logf('%s\n', __:format(msg))
                    dbg('%s\n', msg)
                    break
                end
            end
        end
        if current_version > max_version then
            -- No archived version found, blow away the already-backed-up
            -- database
            os.remove(file)
            current_version = 0
        end
    end

    local orig_version = current_version

    db = assert(sqlite3.open(file))
    assert(db:exec('BEGIN;'))
    local err
    repeat
        local updates_finished = true
        for _, spec in ipairs(schema_spec) do
            if spec.from == current_version then
                logf('%s, from=%s, to=%s\n', file, spec.from, spec.to)
                local good, ret = pcall(spec.updater, db)
                if not good then
                    err = ret
                    break
                end
                db = ret
                current_version = spec.to
                db_exec(db, 'PRAGMA user_version = %s', current_version)
                updates_finished = false
            end
        end
    until updates_finished
    db_exec(db, err and 'ROLLBACK;' or 'COMMIT;')
    db:close()
    local msg
    if err then
        msg = log:error(__"{0}: Error migrating from version {1} to {2} - {3}",
            file, current_version, max_version, err)
    elseif current_version ~= max_version then
        msg = log:error(__"{0}: No migration path from version {1} to {2}",
            file, current_version, max_version)
    elseif orig_version ~= current_version then
        msg = log:notice(__"{0}: Version {1} to {2}",
            file, orig_version, current_version)
    end
    if msg then
        -- KLUDGE this is all nasty
        logf('%s\n', __:format(msg))
        dbg('%s\n', msg)
    end
end

-- Function used to get information from the sql database
local function sql_get(db, sql)
    local vm = db:prepare(sql)
    if not vm then return end
    local result = {}
    while vm:step() == sqlite3.ROW do
       result[#result + 1] = vm:get_uvalues()
    end
    vm:finalize()
    if #result > 0 then
        return result, true
    end
    return
end

-- network configuration migration
-- will make sure that the network configurations are all in order
--   if a default network configuration does not exist, it will create one
--   there must always be an active network configuration. if not, the default will be made active
--   if multiple active configurations exist, keep first (latest) and make all others inactive
--   if a single active network configuration already exists, it will not be disturbed
local function migrate_networks()
    dbg('migrate networks...\n')
    local db = assert(sqlite3.open('/var/lib/armada/setup.db'))
    -- get the default network
    local r, ok = sql_get(db,[[
	SELECT content FROM tsd_users
	WHERE db_name = 'networks' AND id = 'default' AND deleted = '';
    ]])
    local default_network = nil
    if not ok then
	default_network = require 'default_network'
        default_network.active=false
        local req = string.format(
            [[INSERT INTO tsd_users (db_name, id, content, added)
              VALUES ('networks', 'default', '%s', '%s')]],
              cjson.encode(default_network), os.date('!%FT%H:%M:%S', os.time()))
        db_exec(db, req)
        log:notice("Default network configuration created")
    else
	-- default network does exist, store it
	default_network = r[1]
    end
    -- get the active networks
    r, ok = sql_get(db,[[
	SELECT content FROM tsd_users
	WHERE db_name = 'networks' AND id != 'hostname' AND deleted = ''
	AND content LIKE ('%"active":true%')
	ORDER BY added DESC
    ;]])
    if ok then
	if #r == 1 then
	    dbg('all is good, no networks need to be migrated\n')
	end
	if #r > 1 then
	    dbg('many active records - keep first and make others not active\n')
            for idx, v in ipairs(r) do
                if v and idx > 1 then
                    local c = cjson.decode(v)
                    if c and c.active then
                        c.active = false;
                        local query = string.format(
                                                [[ UPDATE tsd_users  SET content = '%s'
                                                   WHERE db_name = 'networks' AND id = '%s' AND deleted = '';]],  cjson.encode(c), c.id)
                        db_exec(db, query)
                        local msg = string.format("Network '%s' is no longer active", c.description)
                        log:notice(msg)
                    end
                end
           end
        end
    else
	default_network.active = true
        local query = string.format(
		[[ UPDATE tsd_users  SET content = '%s'
                  WHERE db_name = 'networks' AND id = 'default' AND deleted = '';]],  cjson.encode(default_network))
        db_exec(db, query)
        log:notice("Default network configuration is now the active one")
    end
    db:close()
end

for file, schema_spec in pairs(database_schemas) do
    dbg('db-migrate %s...\n', file)
    migrate_db(file, schema_spec)
end

migrate_networks()
