--
-- eDAQ Memory Mapped I/O access functions
--
-- Memory mapped through the 64K block at 0xEDAC0000 corresponds to the
-- PCI mapped space inside the MPB FPGA.
--

local u = require 'unix'
local bit = require 'bit'

local edaq_mmio = {}

local fd = assert(u.open('/dev/mem', u.O_RDWR))
local e10 = assert(u.mmap(0x10000, fd, 0xedac0000))
assert(u.close(fd))

local f = io.open("/sys/devices/platform/edaq-clocksource/sample", "r")

edaq_mmio.e10 = e10
edaq_mmio.clocksource_fd = f

local layer_address = {}
for layer=0,7 do
    layer_address['/dev/edaq_expmod' .. layer] =  0xFF00 + (layer * 0x20)
end

function edaq_mmio:initialize()
end

function edaq_mmio:adjust_oscillator(ppb)
    if ppb > 200000 then ppb = 200000 end
    if ppb < -200000 then ppb = -200000 end
    if ppb == 0 then -- Run at unadjusted oscillator rate
        self.e10:w32(0x8074, 0)
    else
        local rate_adjust = math.floor(1000000000 / (29 * ppb) + 0.5)
        local write_value = 0
        if rate_adjust > 0 then -- Go ppb faster
            write_value = bit.bor(0x40000000, rate_adjust)
        end
        if rate_adjust < 0 then -- Go ppb slower
            write_value = bit.bor(0x20000000, rate_adjust * -1)
        end
        self.e10:w32(0x8074, write_value)
    end
end

function edaq_mmio:set_extra_static_data_words_to_receive(static_data_words)
    local bits_to_receive = static_data_words or 0  -- default to zero extra
    bits_to_receive = bits_to_receive * 32  -- convert from words to bits
    bits_to_receive = bits_to_receive + 64  -- include the timestamp
    assert(bits_to_receive >= 64 and bits_to_receive <= 256)
    self.e10:w16(0x8042, bits_to_receive)
end

function edaq_mmio:remote_packet_arrived()
    local status = self.e10:r32(0x8070) -- Check status
    if status == 3 or status == 1 then  -- Mask 0x1=Packet, 0x2=Good
        self.e10:r32(0x8068)            -- If packet arrived, clear status
        return status == 3              -- Only report good packets
    else
        return false
    end
end

function edaq_mmio:get_remote_packet_contents()
    local ts0 = self.e10:r32(0x8068)    -- Low 32 bits of latched arrival time
    local ts1 = self.e10:r32(0x806C)    -- High 32 bits of latched arrival time
    local local_ts = (ts1 * 4294967296) + ts0
    local uf0 = self.e10:r32(0x8060)    -- Low 32 bits of packet contents
    local uf1 = self.e10:r32(0x8064)    -- High 32 bits of packet contents
    local remote_ts = (uf1 * 4294967296) + uf0
    local static_data = {}
    static_data[1] = self.e10:r32(0x8080)
    static_data[2] = self.e10:r32(0x8084)
    static_data[3] = self.e10:r32(0x8088)
    static_data[4] = self.e10:r32(0x808C)
    static_data[5] = self.e10:r32(0x8090)
    static_data[6] = self.e10:r32(0x8094)
    return local_ts, remote_ts, static_data
end

function edaq_mmio:get_local_timestamp()
    self.e10:w32(0x8070, 0)     -- Write causes latch of current counter
    local ts0 = self.e10:r32(0x8078)    -- Low 32 bits of latched count
    local ts1 = self.e10:r32(0x807C)    -- High 32 bits of latched count
    local local_ts = (ts1 * (2^32)) + ts0
    return local_ts
end

function edaq_mmio:sample_timers()
    self.clocksource_fd:seek("set")
    local line = self.clocksource_fd:read("*a")
    local sys, sec, nsec = line:match("^(%d+) (%d+)%.(%d+)")
    return tonumber(sys), nil, tonumber(sec), tonumber(nsec)
end

function edaq_mmio:reset_layer(devname)
    local layer_addr = assert(layer_address[devname])
    local hcsr = self.e10:r16(layer_addr + 0x00)
    hcsr = bit.band(hcsr, 0x7FFF)
    self.e10:w16(layer_addr + 0x00, bit.bor(hcsr, 0x8000))
    self.e10:w16(layer_addr + 0x00, hcsr)
end

--
-- Layer timestamp conversions are needed in two situations
-- 1. Converting received sample data timestamps from 32 to 64 bits
-- 2. Converting 64-bit start times to 32-bits for synchronous channels
--
function edaq_mmio:convert_layer_timestamp(devname, layer_timestamp)
    local layer_ts, local_ts = self:get_layer_timestamp(devname)
    local ts = (local_ts - layer_ts) + layer_timestamp
    if (local_ts - ts) > 2^31 then
        ts = ts + 2^32      -- Compensate for undetected rollover
    end
    return ts
end

function edaq_mmio:compute_layer_start_time(devname, target_local_timestamp)
    -- Apply offset to local timestamp and drop the upper 32 bits.
    -- This function only works on target timestamps that are within ~53
    -- seconds of occuring.
    local layer_ts, local_ts = self:get_layer_timestamp(devname)
    return (target_local_timestamp - (local_ts - layer_ts)) % (2^32)
end

function edaq_mmio:get_layer_timestamp(devname)
    local layer_addr = assert(layer_address[devname])
    self.e10:w16(layer_addr + 0x08, 0)  -- Write MAGIC1 reg to latch counters
    local ts0 = self.e10:r16(layer_addr + 0x0C)  -- Low 16 bits of timestamp
    local ts1 = self.e10:r16(layer_addr + 0x0E)  -- High 16 bits of timestamp
    local layer_ts = (ts1 * (2^16)) + ts0
    local ts0 = self.e10:r32(0x8078)    -- Low 32 bits of MPB timestamp
    local ts1 = self.e10:r32(0x807C)    -- High 32 bits of MPB timestamp
    local local_ts = (ts1 * (2^32)) + ts0
    return layer_ts, local_ts
end

return edaq_mmio

