local bit = require 'bit'
local dbg = require 'dbg'
local ffi = require 'ffi'
local C = ffi.C

local bit_band = bit.band
local ffi_copy = ffi.copy

local ctor_cache = {}
local mt = {}

function mt:asserts()
    if false then
        assert(self.size - 1 == self.mask)
        assert(self.front >= 0 and self.front <= self.mask)
        assert(self.count >= 0 and self.count <= self.mask)
    end
end

function mt:dump()
    dbg('front=%d, count=%d, size=%d, mask=%d\n',
        self.front, self.count, self.size, self.mask)
    for i = 0, self.mask do
        dbg(' %d', self.buf[i])
    end
    dbg('\n')
    --collectgarbage()
end

function mt:enq()
    -- self:asserts()
    if self.count >= self.mask  then
        return nil -- KLUDGE error() ?
    end
    local idx = bit_band((self.front + self.count), self.mask)
    self.count = self.count + 1
    -- self:asserts()
    return self.buf + idx
end

function mt:enq_value(x)
    -- self:asserts()
    if self.count >= self.mask  then
        return false -- KLUDGE error() ?
    end
    local idx = bit_band((self.front + self.count), self.mask)
    ffi_copy(self.buf + idx, x, self.element_size)
    self.count = self.count + 1
    -- self:asserts()
    return true
end

function mt:deq()
    -- self:asserts()
    if self.count == 0 then
        return nil
    end
    self.count = self.count - 1
    local idx = self.front
    self.front = bit_band(self.front + 1, self.mask)
    -- self:asserts()
    return self.buf + idx
end

function mt:head()
    -- self:asserts()
    if self.count == 0 then
        return nil
    end
    return self.buf + self.front
end

function mt:available(expand)
    if expand and (self.mask - self.count) == 0 then
        self:expand()
    end
    return self.mask - self.count
end

function mt:ptr(idx)
    return self.buf + bit_band(self.front + idx, self.mask)
end

function mt:clear()
    self.count = 0
end

function mt:delete(nbr)
    if nbr <= 0 then
        return 0
    elseif self.count < nbr then
        return -1, "delete request larger than count (" ..
            tostring(nbr) .. " > " .. tostring(self.count) .. ")"
    elseif self.front + nbr <= self.mask then 
        self.front = bit_band(self.front + nbr, self.mask)
    else
        self.front = nbr - (self.mask - self.front)
    end
    self.count = self.count - nbr
    return nbr
end

function mt:expand()
    if self.size >= 65536 then
        return
    end
    self.size = self.size * 2
    self.mask = self.size - 1
    local buf = C.realloc(self.buf, self.size * self.element_size)
    self.buf = buf
end

function mt:__len()
    return self.count
end

mt.__index = mt

local function ringbuf_ctor(t, is_struct)
    if not ctor_cache[t] then
        local rb_type = string.format('struct %s_ringbuf', t)
        ffi.cdef(string.format([[
            %s {
                int front;
                int count;
                int element_size;
                int size;
                int mask;
                %s scratch[1];
                %s *buf;
            };
        ]], rb_type, t, t))
        ctor_cache[t] = ffi.metatype(rb_type, mt)
    end
    return ctor_cache[t]
end

local m = {}

function m:ringbuf_size(max_count)
    local size = 1
    while size < max_count do
        size = size * 2
    end
    return size
end

function m:new(t, max_count)
    local ctor = ringbuf_ctor(t)
    local rb = ctor()
    rb.element_size = ffi.sizeof(t)
    rb.size = self:ringbuf_size(max_count)
    rb.mask = rb.size - 1
    local buf = C.calloc(rb.size, rb.element_size)
    rb.buf = buf
    --collectgarbage()
    return rb
end

return m
