local armada = require 'armada'
local s = require 'struct'
local ev = require 'ev'
local dbus = require 'dbus'
local u = require 'unix'
local mmio = require 'edaq_mmio'
local ioctl = require 'edaq_expmod_ioctl'
local channel_proxy = require 'channel_proxy'
local ts = require 'timestamps'

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

-- Constants
local first_command_port = 0xFF10
local last_command_port = 0xFFFE
local reset_command_port = 0xFF0F
local first_run_port = 0
local last_run_port = 0xFEFF
local exception_port = 0xFFFE   -- Must == last_command_port

----------------------------------------------------------------------

local expmod_proxy = class('expmod_proxy', channel_proxy)
function expmod_proxy:initialize(name, devname)
    self.name = name
    self.devname = devname
    self.cmd_actions = { }
    self.cmdports = {}
    self.cmdports[exception_port] = {}
    self.last_cmdport = first_command_port
    self.resp_handlers = {}

    self.cmd_dev = assert(u.open(devname, u.O_RDWR + u.O_NONBLOCK))
    ioctl.check_error(self.cmd_dev)  -- Clear old errors
    ioctl.filter(self.cmd_dev, first_command_port, last_command_port)

    self.cmd_watcher = ev.IO.new(function (loop, w, revents)
                                     self:cmd_interrupt()
                                 end, self.cmd_dev, ev.READ)
    self.cmd_watcher:start(ev.Loop.default)

    self.data_dev = assert(u.open(devname, u.O_RDWR + u.O_NONBLOCK))
    ioctl.check_error(self.data_dev)  -- Clear old errors
    ioctl.filter(self.data_dev, first_run_port, last_run_port)

    self.mmio = mmio

    local msg = self:cmd_block('GET NAME=SERIAL_NUMBER')
    assert(msg and msg[#msg]['ERRNO'] == 'OK')
    self.sn = msg[1]['SERIAL_NUMBER']

    msg = self:cmd_block('CHARACTERIZATION_READ')
    assert(msg and msg[#msg]['ERRNO'] == 'OK')
    table.remove(msg)
    self.characterization = {}
    self.characterization.raw_set = msg
    local channel = 0
    for _,line in ipairs(msg) do
        local assignment = line['ASSIGNMENT']
        if assignment then
            local k, v = assignment:match("([^%=]+)=(.+)")
            if k == 'CHANNEL' then
                channel = tonumber(v) + 1
            elseif k then
                if not self.characterization[channel] then
                    self.characterization[channel] = {}
                end
                self.characterization[channel][k] = v
            end
        end
    end

    msg = self:cmd_block('ECN_READ')
    assert(msg and msg[#msg]['ERRNO'] == 'OK')
    table.remove(msg)
    self.ecns = msg

    -- KLUDGE doesn't call superclass to allow stuff between
end

function expmod_proxy:enable_data_watcher()
    self.data_watcher = ev.IO.new(function (loop, w, revents)
                                      self:data_interrupt()
                                  end, self.data_dev, ev.READ)
    self.data_watcher:start(ev.Loop.default)
end

function expmod_proxy:send_command(...)
    local command = table.concat({...}, ' ')
    local cmdlen = #command + 2
    local portnum = self.last_cmdport
    self.last_cmdport = self.last_cmdport + 1
    if self.last_cmdport == last_command_port then
        self.last_cmdport = first_command_port
    end
    if self.cmdports[portnum] == nil then
        self.cmdports[portnum] = { command = command }
        local header = s.pack('<HH', cmdlen, portnum)
        if self.debug_expmod_com then
            print(string.format('%s FF%02X tx %s',
                  self.devname, portnum, command))
        end
        u.write(self.cmd_dev, header .. command)
        return portnum
    end
    error("Excessive outstanding commands to layer " .. self.devname)
end

function expmod_proxy:set_response_handler(handler, proxy)
    for portnum=first_run_port,last_run_port do
        if not self.resp_handlers[portnum] then
            self.resp_handlers[portnum] = { handler, proxy }
            return portnum
        end
    end
    error("Excessive response handlers on layer " .. self.devname)
end

function expmod_proxy:clear_response_handler(portnum)
    assert(self.resp_handlers[portnum])
    self.resp_handlers[portnum] = nil
end

function expmod_proxy:data_interrupt()
    local resp_port, resp = self:read_dev(self.data_dev)
    if self.resp_handlers[resp_port] then
        local handler, proxy = unpack(self.resp_handlers[resp_port])
        handler(proxy, resp)
        return
    end
    error(self.sn .. " sent packet with no handler on port " .. resp_port)
end

function expmod_proxy:process_exception()
    -- Default processing, can override in derived proxies
    local msg = self:cmd_block('GET NAME=EXCEPTION_MESSAGE')
    self:cmd('SET EXCEPTION_MESSAGE=')
    error(self.sn .. ' reports ' .. msg[1]['EXCEPTION_MESSAGE'])
end

function expmod_proxy:read_dev(dev)
    local header, errmsg = u.read(dev, 2)
    if not header then error(self.devname .. ' returned ' .. errmsg) end
    assert(#header == 2, "expmod layer out of sync, " .. self.devname)
    local resp_len = s.unpack('<H', header)
    assert(resp_len > 2 and resp_len <= 1024,
           "expmod layer returned bad length, " .. self.devname)
    local read_len = resp_len + (resp_len % 2)  -- Adjust for odd length
    local body, errmsg = u.read(dev, read_len)
    if not body then error(self.devname .. ' returned ' .. errmsg) end
    assert(#body == read_len, "expmod layer out of sync, " .. self.devname)
    local resp_port, resp = s.unpack(string.format('<Hc%d', resp_len - 2), body)
    return resp_port, resp
end

function expmod_proxy:poll_buffer()
    local resp_port, resp = self:read_dev(self.cmd_dev)
    if self.cmdports[resp_port] then
        if self.debug_expmod_com then
            print(string.format('%s %04X rx %s', self.devname, resp_port, resp))
        end
        local parsed_line = {}
        for pair in resp:gmatch("%S+") do
            local k, v = pair:match("([^%=]+)=(.+)")
            if not k then
                local errmsg = string.format('Parse error on %s, %s -> %s',
                                             self.sn or self.devname,
                                             resp, pair)
                error(errmsg)
            end
            parsed_line[k] = v
        end
        self.cmdports[resp_port][#self.cmdports[resp_port] + 1] =
            parsed_line
        if parsed_line['ERRNO'] then
            local command_response = self.cmdports[resp_port]
            if resp_port == exception_port then
                self.cmdports[resp_port] = {}
                self:process_exception()
            else
                self.cmdports[resp_port] = nil
                return resp_port, command_response
            end
        end
        return
    end
    error(self.sn .. " sent packet with no handler on port " .. resp_port)
end

function expmod_proxy:reset()
    local reset_command = "RESET"
    u.write(self.cmd_dev, s.pack('<HHs', #reset_command + 2,
                                 reset_command_port, reset_command))
    self.cmdports = {}
    self.resp_handlers = {}
end

function expmod_proxy:halt()
    self:reset()
end

function expmod_proxy:cmd_cb(cb, ...)
    table.insert(self.cmd_actions, { self:send_command(...), cb})
end

function expmod_proxy:cmd_block(...)
    local cmd_response = nil
    self:cmd_cb(function (msg)
                    cmd_response = msg
                end, ...)
    repeat
        ev.Loop.default:loop('once')
    until cmd_response
    return cmd_response
end

function expmod_proxy:cmd(...)
    self:cmd_cb(function (msg)
                    self:msg_validate(msg)
                end, ...)
end

function expmod_proxy:cmd_yield(...)
    local my_coro = assert(coroutine.running())
    self:cmd_cb(function (msg)
                    assert(coroutine.resume(my_coro, msg))
                end, ...)
    return assert(coroutine.yield())
end

function expmod_proxy:msg_validate(msg, ...)
    local catch_errors = {...}
    assert(msg)
    local result = assert(msg[#msg]['ERRNO'])
    for _,v in pairs(catch_errors) do
        if result == v then return v end
    end
    assert(result == 'OK',
           string.format('%s sent to %s, returned %s', msg.command,
                                                       self.sn,
                                                       msg[#msg]['ERRNO']))
end

function expmod_proxy:cmd_interrupt()
    local resp_port, msg = self:poll_buffer()
    if not msg then return end
    local portnum, cb = unpack(assert(table.remove(self.cmd_actions, 1)))
    if portnum ~= resp_port then
        local errmsg = string.format(
            "Command (%04X) / response (%04X) sequencing error on %s.\r\n",
            portnum, resp_port, self.sn or self.devname)
        errmsg = errmsg .. "Command was " .. msg.command .. "\r\n"
        error(errmsg)
    end
    cb(msg)
    if self.launch_coroutine then self:launch_coroutine() end
end

return expmod_proxy
