#!/usr/bin/lua

require('lj_include'){
    'sys/types',
    'unistd',
    'sched',
    'linux/sched',
}

local log = require 'log'
-- Open logging session for the entire process
-- NOTE: Lua modules, loaded via 'require', should *NOT* call "log:open()"
log:open('flagship-proxy', log.USER)

local ffi = require 'ffi'
local ev = require 'ev'
local u = require 'unix'
local cc = require 'connctrl'

local C = ffi.C

local m = require 'register_critical'
m.RegisterMe()

-- Increase process priority
C.nice(-5);

local sp = ffi.new('struct sched_param')
sp.__sched_priority = 50 -- Arbitrary for now
assert(C.sched_setscheduler(C.getpid(),
    bit.bor(C.SCHED_FIFO, C.SCHED_RESET_ON_FORK), sp) == 0)

local function dbg(fmt, ...)
    log:log(log.DEBUG, fmt, ...)
end

local proxies = {}
local proxy_map

if u.stat('/sys/class/edaq') then
    -- Build proxy map from probed eDAQ expmod layers
    local function sendcmd(devname, cmd)
        local cl = 'sendcmd '..devname..' 0xfff1 '..cmd
        local f = assert(io.popen(cl, 'r'))
        local results = assert(f:read())
        f:close()
        if results ~= 'Timeout' then return results end
        -- First retry
        f = assert(io.popen(cl, 'r'))
        results = assert(f:read())
        f:close()
        if results ~= 'Timeout' then return results end
        -- Second retry
        f = assert(io.popen(cl, 'r'))
        results = assert(f:read())
        f:close()
        return results
    end
    proxy_map = {}
    local instance_counts = {}
    for i=0,7 do
        local devname = '/dev/edaq_expmod' .. i
        if u.stat(devname) then
            if sendcmd(devname, 'reboot') == 'Timeout' then
                dbg(devname .. ' not responding, sending RESET command')
                sendcmd(devname, 'reset')
                os.execute('sleep 3')
                if sendcmd(devname, 'reboot') == 'Timeout' then
                    -- Reset ISA bus
                    dbg(devname .. ' not responding, resetting ISA bus')
                    os.execute('/sbin/e10io -modl 0x8048 0xffffffff 0x00000100')
                    os.execute('/sbin/e10io -modl 0x8048 0xfffffeff 0x00000000')
                    os.execute('sleep 3')
                    break
                end
            end
            sendcmd(devname, 'run_application')
        end
    end
    for i=0,7 do
        local devname = '/dev/edaq_expmod' .. i
        if u.stat(devname) then
            local sn = sendcmd(devname, 'get name=serial_number')
            local proxy = sn:match('^SERIAL_NUMBER=..(...)')
            local fw_version = sendcmd(devname, 'get name=fw_version')
            if proxy and fw_version:find('armada') then
                proxy = proxy:lower()
                instance_counts[proxy] = (instance_counts[proxy] or 0) + 1
                local instance = proxy .. instance_counts[proxy]
                proxy_map[#proxy_map + 1] = { proxy, instance, devname }
            end
        end
    end

    proxies.tcb = require 'tcb_proxy'
    proxies.itb = require 'tcb_proxy'
    proxies.brg = require 'analog_proxy'
    proxies.hls = require 'analog_proxy'
else
    if not u.stat('/sys/class/cc_class/cca3') then
        error("Driver not loaded or FPGA unconfigured")
    end

    proxy_map = {
        {'can', 'can1', '/dev/cca0'},
        {'can', 'can2', '/dev/cca1'},
        {'can', 'can3', '/dev/cca2'},
        {'aux', 'aux',  '/dev/cca3'},
        {'gps', 'gps',  '/dev/cca4'},
        {'dio', 'dio',  '/dev/cca5'},
    }

    proxies.can = require 'can_proxy'
    proxies.dio = require 'dio_proxy'
    proxies.gps = require 'gps_proxy'
    proxies.aux = require 'aux_proxy'

    for _,v in ipairs(proxy_map) do cc.controller_reset(v[3]) end
end

local ps = {}

for i, v in ipairs(proxy_map) do
    if proxies[v[1]] then
        ps[#ps + 1] = proxies[v[1]]:new(v[2], v[3])
    end
end

ev.Signal.new(function(loop) loop:unloop() end, u.SIGINT):start(ev.Loop.default)
ev.Signal.new(function(loop) loop:unloop() end, u.SIGTERM):start(ev.Loop.default)

local ok, err = xpcall(function() ev.Loop.default:loop() end, debug.traceback)

for _, p in ipairs(ps) do
    dbg('halt', p)
    p:halt()
end
if not ok then
    error(err .. "\n")
end
