require('lj_include'){
    'unistd',
    'linux/time',
    'sys/time',
}

local armada = require 'armada'
local ffi = require 'ffi'
local s = require 'struct'
local ts = require 'timestamps'
local u = require 'unix'
local ev = require 'ev'
local config = require 'system_config'

local C = ffi.C

local sample_clock = require 'sample_clock'
local gps = require 'gps'
local gpschar = require 'gpschar'
local membuf = require 'membuf'
local nmea_parser = require 'nmea_parser'

local channel_proxy = require 'channel_proxy'
local cca_handler = require 'cca_handler'

local dbus = require 'dbus'
local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

local __ = require '__'
local log = require 'log'

-- Function used to retrieve preferences for regulation of Si514 by GPS' PPS
-- Any failure returns values that cause the Si514 to be regulated by the PPS
-- signal of the GPS and the system/PTP time is NOT synchronized to GPS time
local function get_oscillator_control()
    -- FIXME:  Remove this bit of code to re-enable setting this via the GUI
    if config.gps_timing_regulation then
        if config.gps_timing_regulation == 'none' then
            return false, true
        elseif config.gps_timing_regulation == 'regulate_only' then
            return false, false
        elseif config.gps_timing_regulation == 'sync_to_gps_time' then
            return true, false
        end
    end
    -- Default is "none"
    do return false, true end
    -- FIXME:  END

    local sqlite3 = require 'lsqlite3'
    local cjson = require 'cjson'

    local db = sqlite3.open('/var/lib/armada/setup.db')
    if not db then
        log(log.WARNING, "Unable to open setup database")
        return false, true
    end

    local vm = db:prepare([[
        SELECT content from tsd_users
        WHERE deleted = '' AND db_name = 'system' and id = 'preferences';
    ]])
    if not vm then
        db:close()
        log(log.WARNING, "Failure in preparation of database query")
        return false, true
    end
    local r = {}
    while vm:step() == sqlite3.ROW do
        r[#r + 1] = vm:get_uvalues()
    end
    vm:finalize()
    db:close()

    if #r < 1 then
        log(log.WARNING, "Unable able to get preferences from setup database")
        return false, true
    end

    local t = cjson.decode(r[1])
    if not t.gps_timing_regulation or
        type(t.gps_timing_regulation) ~= 'string' then
        log(log.WARNING, "GPS oscillator control preference not in database")
        return false, true
    end
    if t.gps_timing_regulation == 'none' then
        return false, true
    elseif t.gps_timing_regulation == 'regulate_only' then
        return false, false
    elseif t.gps_timing_regulation == 'sync_to_gps_time' then
        return true, false
    end
    return false, false
end

local time_synced_to_gps, no_gps_regulation_of_oscillator =
    get_oscillator_control()

local function dbg(fmt, ...)
    log:log(log.DEBUG, fmt, ...)
end

ffi.cdef[[
    typedef struct field_descr_struct {
        size_t start;
        size_t length;
    } field_descr;
]]
local field_descr_type = ffi.typeof('field_descr[?]')
local MAX_NBR_FIELDS = 30;

----------------------------------------------------------------------

-- This function converts a string of ASCII decimal bytes to a
-- Lua number.  Note that the input string may contain a prefixed
-- minus character.  It may also contain a decimal point and
-- following digits.
--
-- This function returns the number, or nil if the input string is
-- found to be non-existent, or -1 if an error is found.
--
-- The input string is considered to be non-existent if 'dec_start'
-- has a negative value.
--
-- An error will be caused by 'in_buf' or 'dec_start' having a nil
-- value.  An error can also be caused by an ASCII non-decimal byte
-- within the input (sub)string, except that an ASCII period will
-- be accepted to indicate a number with a decimal fractional part.
-- However, in the case of the parameter 'dec_len' being <= zero or nil,
-- an ASCII non-decimal byte will merely indicate the end of the
-- input (sub)string to be converted.
--
-- Parameter 'in_buf' is the membuf object containing the string to
--     be converted.  'in_buf' is not modified by this function.  It
--     can not be nil.
-- Parameter 'dec_start' is the offset from the head of 'in_buf' at
--     which the string to be converted is found.  This value must
--     be defined.  If the value is < zero the given input string is
--     considered to be non-existent and a zero is returned.
-- Parameter 'dec_len' is the length in bytes of the string to
--     be converted.  If 'dec_len' has a value <= zero, the length
--     is assumed to be the remaining bytes in 'in_buf' after
--     'dec_start', up to the first ASCII non-decimal and non-decimal
--     point character.
function ascii_dec_to_number(in_buf, dec_start, dec_len)
    local returned_nbr = 0
    local whole_nbr = 0
    local nbr_fraction_digits = 0
    local fraction = 0
    local fraction_found = false
    local len_to_convert = 0
    local len_modified = false
    local polarity_factor = 1

    if not in_buf or not dec_start then
        dbg("ascii_dec_to_number input ERROR\n")
        returned_nbr = -1
    end

    if dec_start < 0 then
        return nil
    end

    if not dec_len or dec_len <= 0 then
        len_to_convert = #in_buf - dec_start
        if len_to_convert <= 0 then
            return 0
        end
        len_modified = true
    else
        len_to_convert = dec_len
    end

    -- check for prefixed minus sign
    local iterate_start = dec_start
    if in_buf.buf[iterate_start] == gpschar.CHAR_HYPHEN then
        polarity_factor = -1
	iterate_start = iterate_start + 1
    end

    -- iterate through all remaining digits (including any fraction)
    for i = iterate_start, dec_start + len_to_convert - 1, 1 do
        if in_buf.buf[i] >= gpschar.CHAR_ZERO and
           in_buf.buf[i] <= gpschar.CHAR_NINE then
            if not fraction_found then
                whole_nbr = whole_nbr * 10 +
                            (in_buf.buf[i] - gpschar.CHAR_ZERO)
            else
                fraction = fraction * 10 +
                            (in_buf.buf[i] - gpschar.CHAR_ZERO)
                nbr_fraction_digits = nbr_fraction_digits + 1
	    end
        elseif in_buf.buf[i] == gpschar.CHAR_PERIOD then
            if not fraction_found then
                fraction_found = true
            else
                if not len_modified then
                    returned_nbr = -1
                end
                break;
	    end
        else
            if not len_modified then
                returned_nbr = -1
            end
            break;
        end
    end

    -- assemble absolute value of number to return
    if returned_nbr == 0 then
        if nbr_fraction_digits > 0 then
            returned_nbr = whole_nbr +
                           (fraction / (10 ^ nbr_fraction_digits))
        else
            returned_nbr = whole_nbr
        end
    end

    return polarity_factor * returned_nbr
end

-- This function returns an array of C structures containing the
--     offsets and field lengths specifying the fields in the given
--     GPS input buffer.  The input fields are comma separated.
-- Parameter 'in_buf' is a membuf object containing the GPS input.
local function identify_fields(in_buf)
    -- KLUDGE - field_descr_type(MAX_NBR_FIELDS, {[0] = -1, -1}) doesn't
    -- seem to work.
    local field_descrs = field_descr_type(MAX_NBR_FIELDS)
    local nbr_of_fields = 0
    local search_offset = -1

    for i = 0, MAX_NBR_FIELDS - 1, 1 do
        local field_offset = search_offset + 1
        if field_offset >= #in_buf then
            break;
        end
        search_offset = in_buf:search(field_offset, ",", 1)
        nbr_of_fields = nbr_of_fields + 1
        field_descrs[i].start = field_offset
        if not search_offset then
            field_descrs[i].length = #in_buf - field_offset
            break;
        else
            field_descrs[i].length = search_offset - field_offset
        end
    end

    return field_descrs, nbr_of_fields
end

-- This function converts the specified membuf substring to a lua number
local function field_to_number(f, descr)
    return ascii_dec_to_number(f, descr.start, descr.length)
end

-- This function converts the specified membuf substring to a lua string
local function field_to_string(f, descr)
    return ffi.string(f:head() + descr.start, descr.length)
end

-- This function returns an index into the input array for the
--     start of the suffix substring ".0", or nil if such is
--     not found.
-- Parameter utc_date_time is a membuf object
local function find_trailing_0(utc_date_time)
    if utc_date_time == nil or #utc_date_time <= 0 then
        return nil
    end

    local suffix_index = nil
    local period_offset = #utc_date_time - 2
    local zero_offset = #utc_date_time - 1

    if utc_date_time.buf[period_offset] == gpschar.CHAR_PERIOD and
       utc_date_time.buf[zero_offset] == gpschar.CHAR_ZERO then
        suffix_index = period_offset
    end

    return suffix_index
end

local function fix_date(f, descr)
    if (descr.length ~= 6) then
        return nil
    end
    local date_buf = membuf:new(nil, 10)

    date_buf:append(gpschar.CHAR_TWO_STR, 1)
    date_buf:append(gpschar.CHAR_ZERO_STR, 1)
    date_buf:append(f:head() + descr.start + 4, 2)
    date_buf:append(gpschar.CHAR_HYPHEN_STR, 1)
    date_buf:append(f:head() + descr.start + 2, 2)
    date_buf:append(gpschar.CHAR_HYPHEN_STR, 1)
    date_buf:append(f:head() + descr.start, 2)
    return date_buf
end

local function fix_time(f, descr)
    if (descr.length ~= 8) then
        return nil
    end
    local time_buf = membuf:new(nil, 10)

    time_buf:append(f:head() + descr.start, 2)
    time_buf:append(gpschar.CHAR_COLON_STR, 1)
    time_buf:append(f:head() + descr.start + 2, 2)
    time_buf:append(gpschar.CHAR_COLON_STR, 1)
    time_buf:append(f:head() + descr.start + 4, descr.length - 4)
    return time_buf
end

-- Returns seconds (to decimal tenths) converted to Lua number
local function fix_seconds(f, descr)
    if (descr.length ~= 8) then
        return nil
    end

    local nbr = ascii_dec_to_number(f, descr.start + 4, 2)
    if nbr then
        nbr = nbr + (ascii_dec_to_number(f, descr.start + 7, 1) / 10)
    end
    return nbr
end

-- Returns membuf containing date followed by time
local function fix_date_time(f, fdd, fdt)
    local result_buf = fix_date(f, fdd)
    local time_buf = fix_time(f, fdt)
    if not result_buf or not time_buf then
        return nil
    end
    result_buf:append(time_buf)
    return result_buf
end

-- Returns latitude degrees.  Minutes are represented as decimal
-- fraction part of returned Lua number.  Degrees are negative if
-- parameter 'ns' equals "S".
local function fix_lat(f, descr, ns_descr)
    local lat_whole = ascii_dec_to_number(f, descr.start, 2)
    local lat_minutes = ascii_dec_to_number(f, descr.start + 2)
    local lat = nil

    if lat_whole and lat_minutes then
        lat = lat_whole + lat_minutes / 60.0
        if f.buf[ns_descr.start] == gpschar.CHAR_S then
            lat = -lat
        end
    end

    return lat
end

-- Returns longitude degrees.  Minutes are represented as decimal
-- fraction part of returned Lua number.  Degrees are negative if
-- parameter 'ew' equals "W".
local function fix_lon(f, descr, ew_descr)
    local lon_whole = ascii_dec_to_number(f, descr.start, 3)
    local lon_minutes = ascii_dec_to_number(f, descr.start + 3)
    local lon = nil

    if lon_whole and lon_minutes then
        lon = lon_whole + lon_minutes / 60.0
        if f.buf[ew_descr.start] == gpschar.CHAR_W then
            lon = -lon
        end
    end

    return lon
end

local function fix_mv(f, descr, ew_descr)
    local v = ascii_dec_to_number(f, descr.start)
    if v and f.buf[ew_descr.start] == gpschar.CHAR_E then
        return -v
    end
    return v
end

local gps_proxy = class('gps_proxy', channel_proxy)
function gps_proxy:initialize(name, devname)
    self.proxy_type = 'gps'
    self.clk_frequency = sample_clock.master_rate
    cca_handler.new(devname, self, gps, 'GPS_0.bin', self.clk_frequency)

    self:cmd_block('txdr_power_on_5v')

    self.nmea = nmea_parser:new()
    self.channels = {}
    self.external_names = {}
    self.internal_names = {}
    self.cur_sentence_ts = nil
    self.cur_sentence_max = 256
    self.cur_sentence = ffi.new('uint8_t[?]', self.cur_sentence_max)
    self.cur_sentence_pos = 0
    self.servo_ts = 0
    self.pps_ts = 0
    self.pps_ts_array = {}
    self.oscillator_control_state = 'none'
    self.state = 'gps_lock_init'
    self.time_set = true
    self.hb_sinks = nil

    -- Oscillator Control
    self.osc_collecting = false
    self.osc_pps_count = 0
    self.osc_servo = false
    self.osc_control = false
    self.osc_servo_ts = 0
    self.osc_pps_ts = 0
    self.osc_offset = 0

    self.sinks = {}

    -- GPS sync stuff
    self.pps_rising_events = 0
    self.last_offset = 1e30
    self.dt = .2        -- GARMIN ONLY!!
    self.Kp, self.Ki, self.Kd = 37, 10, 4
    self.integral = 0

    self.pps_rate = 5 -- Hz KLUDGE Garmin 5Hz hardcode!!!
    self.pps_interval = self.clk_frequency / self.pps_rate
    self.gprmc_channels_good = false -- KLUDGE for Garmin

    if config.debug_files then
        self.max_regulations = 1000
        self.num_regulations = 0
        self.regulation_skip_count = 0
        self.handle = assert(io.open('/tmp/gps_regulation.txt', 'w+'))
        self.handle:write('count\toffset\toutput\n')
    end

    channel_proxy.initialize(self, name)

    -- An event loop is required to complete the stuff in this function so we
    -- defer these items until the event loop is up and going.
    self.init_completion = ev.Timer.new(function (loop, idle, revents)
        -- At this point we need to start up the PPS events so we can get GPS
        -- sync going (if we have GPS lock).
        local now_40m = ts.ts_to_stamp(ts.get_ts())
        self:cmd('event_setup',
            { pps_falling = true, pps_rising = true })
        self:cmd('uart_rx_start',
            now_40m + self.clk_frequency, 2, 16, 512, '8N1', 38400)
        self:cmd('event_start',
            now_40m + self.clk_frequency, 2, 1, 128)
        idle:stop(ev.Loop.default)
    end, .001, 0)
    self.init_completion:start(ev.Loop.default)

    -- We need the servo() calls to not happen within the cca reader context
    -- because DBus (via SignalHeartbeat()) will screw us over horribly by
    -- nesting calls to gps_proxy:buffer(). So, we use a pipe as a queue for
    -- the offsets, thereby avoiding any libev recursion.
    self.servo_offset_buffer = ffi.new('double[1]')
    self.servo_pipe = ffi.new('int[2]')
    assert(C.pipe(self.servo_pipe) == 0)
    self.servo_w = ev.IO.new(self.servo_watcher_fn, self.servo_pipe[0], ev.READ)
    self.servo_w.proxy = self
    self.servo_w:start(ev.Loop.default)
end

function gps_proxy:queue_servo(offset, override_state)
    if self.state == 'gps_locked' or override_state then
        self.servo_offset_buffer[0] = offset
        local len = ffi.sizeof('double')
        assert(C.write(
            self.servo_pipe[1], self.servo_offset_buffer, len) == len)
    end
end

function gps_proxy.servo_watcher_fn(loop, w, revents)
    local self = w.proxy
    local len = ffi.sizeof('double')
    assert(C.read(self.servo_pipe[0], self.servo_offset_buffer, len) == len)
    self:servo(self.servo_offset_buffer[0])
end

function gps_proxy:debug_stop_check()
    if self.handle then
        if self.num_regulations == self.max_regulations then
            self.handle:close()
            self.handle = nil
        end
    end
end

function gps_proxy:write_debug(offset, output)
    if self.handle then
        self.num_regulations = self.num_regulations + 1
        self.handle:write(string.format("%s\t%s\t%s\n", self.num_regulations,
            offset, output))
    end
end

function gps_proxy:do_skip(offset)
    if self.handle then
        if self.regulation_skip_count > 0 then
            self.regulation_skip_count = self.regulation_skip_count - 1
            self.handle:write(string.format("%s\t%s\t0\n",
                -self.regulation_skip_count, offset))
            return true
        end
    end
end

function gps_proxy:time_synced_servo(offset)
    -- Always send the heartbeat if we have GPS lock.  This indicates to others
    -- that we have GPS sync regardless of whether or not the clock is being
    -- adjusted.
    armada.master:SignalHeartbeat('GPSServo')

--dbg(string.format("Offset: %f (ns)\n", offset * (1000000000 / self.clk_frequency)))

    if math.abs(offset) > 10000 then
        armada.timing:AdjustClockRate(0, 'GPS')
        local ns = offset * (1000000000 / self.clk_frequency)
        local s, ns = math.floor(ns / 1000000000), math.floor(ns % 1000000000)
        armada.timing:AdjustClock(s, ns, 'GPS')
        self.integral = 0
    else
        if self:do_skip(offset) then return end
        if self.last_offset > 1e29 then
            self.last_offset = offset
        end
        self:debug_stop_check()
        local previous_error = self.last_offset
        local error = offset
        self.integral = self.integral + (error * self.dt)
        local derivative = (error - previous_error) / self.dt
        local output = (self.Kp * error) + (self.Ki * self.integral) +
                       (self.Kd * derivative)
        armada.timing:AdjustClockRate(output, 'GPS')
        self.last_offset = offset
        self:write_debug(offset, output)
    end
end

function gps_proxy:time_not_synced_servo(offset)
    -- Always send the heartbeat if we have GPS lock.  This indicates to others
    -- that we have GPS sync regardless of whether or not the clock is being
    -- adjusted.
    armada.master:SignalHeartbeat('GPSServo')

    if self:do_skip(offset) then return end

    if self.last_offset > 1e29 then
        self.last_offset = offset
    end
    self:debug_stop_check()
    local previous_error = self.last_offset
    local error = offset
    self.integral = self.integral + (error * self.dt)
    local derivative = (error - previous_error) / self.dt
    local output = (self.Kp * error) + (self.Ki * self.integral) +
                   (self.Kd * derivative)
    armada.timing:AdjustClockRate(output, 'GPS')
    self.last_offset = offset
    self:write_debug(offset, output)
end

local function set_gps_regulation()
    if not no_gps_regulation_of_oscillator then
        if time_synced_to_gps then
            gps_proxy.servo = gps_proxy.time_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled with Linux Time Sync")
        else
            gps_proxy.servo = gps_proxy.time_not_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled")
        end
    else
        -- If no regulation of the oscillator via the GPS is desired make the
        -- servo do nothing
        gps_proxy.servo = function(self, offset) self.last_offset = offset end
        log:notice(__"GPS: Si514 PPS Servo Disabled")
    end
end

set_gps_regulation()

-- FIXME:  The function below MAY need changes to work properly using an
-- EGPS200.
function gps_proxy:buffer(msg)
    local element_count = msg.element_count
    local heartbeat = false
    local first_char_ts = false
    for j=1,element_count do
        if msg.heartbeat and j == element_count then
            heartbeat = msg.bufdata[j - 1].ts
            if self.hb_sinks then
                for cn, sink in pairs(self.hb_sinks) do
                    sink(heartbeat)
                end
            end
        else
            if msg.cmd == ffi.C.GPS_UART_RX_START then
                local c = msg.bufdata[j - 1].rxdata
                self.cur_sentence[self.cur_sentence_pos] = c
                self.cur_sentence_pos = self.cur_sentence_pos + 1
                if self.cur_sentence_pos >= self.cur_sentence_max then
                    self.cur_sentence_ts = nil
                    self.cur_sentence_pos = 0
                end
                if c == string.byte('\n') then
                    if self.cur_sentence_ts then
                        self:sentence(self.cur_sentence_ts,
                                      self.cur_sentence,
                                      self.cur_sentence_pos)
                    end
                    self.cur_sentence_ts = nil
                    self.cur_sentence_pos = 0
                elseif c == string.byte('$') then
                    self.cur_sentence_ts = msg.bufdata[j - 1].ts
                    self.cur_sentence[0] = c
                    self.cur_sentence_pos = 1
                end
            elseif msg.cmd == ffi.C.GPS_EVENT_START then
                local buffer = msg.bufdata[j - 1]
                local events = buffer.events
                if events.pps_rising == 1 and not self.osc_control then
                    -- If we got a PPS rising event then mark the GPS as locked
                    -- and start a monitoring timer to mark it unlocked if we
                    -- lose PPS rising events for more than 3 seconds.
                    if not self.gps_lock_monitor then
                        self.gps_lock_monitor =
                            ev.Timer.new(function (loop, timer, revents)
                                    timer:stop(ev.Loop.default)
                                    self.gps_lock_monitor = nil
                                    self.pps_rising_events = 0
                                end, 3, 0)
                        self.gps_lock_monitor:start(ev.Loop.default)
                    end
                    self.gps_lock_monitor:again(ev.Loop.default, 3)
                    self.pps_ts = buffer.ts

                    -- If we are NOT synchronizing time to GPS time then just
                    -- regulate the Si514 based on raw count.  FIXME:  This is
                    -- only for GARMIN.  Will need modification for Racetech
                    if not time_synced_to_gps then
                        self.pps_ts_array[#self.pps_ts_array + 1] = self.pps_ts
                        if #self.pps_ts_array == 6 then     -- GARMIN!
                            local delta = 40000000 -
                                tonumber(self.pps_ts_array[6] -
                                    self.pps_ts_array[1])
                            -- When the GPS is physically plugged in we seem to
                            -- get some false events that look like they are
                            -- switch bounce types of things.
                            local ppm = delta / 40
                            if math.abs(ppm) < 100 then
                                self:queue_servo(ppm)
                                table.remove(self.pps_ts_array, 1)
                            else
                                self.pps_ts_array = {}
                            end
                        end
                    end
                    self.pps_rising_events = self.pps_rising_events + 1
                elseif events.pps_rising == 1 and self.osc_control then
                    -- Oscillator Control Operations
                    if self.osc_pps_ts ~= 0 then
                        self.pps_rising_events = self.pps_rising_events + 1
                        self.osc_offset = self.pps_interval - tonumber(buffer.ts - self.osc_pps_ts)
                        -- Normalize offset to PPM of clock frequency
                        -- (self.clk_frequency / 1000000) = 40
                        local ppm = self.osc_offset / 40
                        if math.abs(ppm) < 10000 then
                            if self.osc_servo then
                                if buffer.ts ~= self.osc_servo_ts then
                                    self:queue_servo(ppm, true)
                                end
                                self.osc_servo_ts = buffer.ts + self.pps_interval
                            end
                            if self.osc_collecting then
                                self.osc_pps_data[#self.osc_pps_data + 1] = {
                                    ts = tonumber(buffer.ts),
                                    offset = tonumber(self.osc_offset * self.pps_rate),
                                    interval = tonumber(buffer.ts - self.osc_pps_ts),
                                }
                                self.osc_pps_count = self.osc_pps_count + 1
                            end
                        end
                    end
                    self.osc_pps_ts = buffer.ts
                end
                if events.cable_unplugged == 1 then
                    dbg("CABLE UNPLUGGED\n")
                end
                if events.cable_plugged == 1 then
                    dbg("CABLE PLUGGED\n")
                end
                if events.start_of_break == 1 then
                    dbg("START OF BREAK\n")
                end
                if events.end_of_break == 1 then
                    dbg("END OF BREAK\n")
                end
                if events.parity_error == 1 then
                    dbg("PARITY ERROR\n")
                end
                if events.frame_error == 1 then
                    dbg("FRAME ERROR\n")
                end
            end
        end
    end
end

function gps_proxy:start_comet_stream(w, channels, sink)
-- required channel_proxy callback function
end

local channel_specs -- KLUDGE

-- Function used to set parameters used to control if the Si514 is regulated by
-- the GPS and whether or not PTP/System time is synchronized to GPS time.
function gps_proxy:oscillator_control(params, cb)
    local control = (params and params[1]) or 'none_specified'
    control = control:lower()
    if control == 'none' then
        time_synced_to_gps = false
        no_gps_regulation_of_oscillator = true
    elseif control == 'regulate_only' then
        time_synced_to_gps = false
        no_gps_regulation_of_oscillator = false
    elseif control == 'sync_to_gps_time' then
        time_synced_to_gps = true
        no_gps_regulation_of_oscillator = false
    elseif control == 'get' then
        return {
            result = 'success',
            msg = 'OK',
            current_state = self.oscillator_control_state,
        }
    else
        return {
            result = 'error',
            msg = string.format("Unknown oscillator_control: %s", control),
        }
    end
    self.oscillator_control_state = control

    -- Make sure the servo get switched to the correct one if the user changes
    -- the regulation mode via the API.
    set_gps_regulation()

    return {
        result = 'success',
        msg = 'OK',
        current_state = self.oscillator_control_state,
    }
end

function gps_proxy:start_channel_stream(w, spec, md_sink, sink)
    local channel_sink
    local ch_name = self.internal_names[spec.connector]
    local start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
    spec._timescale = string.format("1/%.0f", self.clk_frequency)
    local proxy = self
    if channel_specs[ch_name] and channel_specs[ch_name].range_max then
        channel_sink = function(timestamp, name, value)
            sink(timestamp - start_stamp, s.pack('>d', tonumber(value)))
        end
        self.hb_sinks = self.hb_sinks or {}
        local v = spec.data_type == 'message_float32' and
            ffi.new('float[1]') or ffi.new('double[1]')
        v[0] = 0/0
        self.hb_sinks[spec.connector] = function(ts)
                sink(ts - start_stamp, v, ffi.sizeof(v))
            end
    else
        channel_sink = function(timestamp, name, value)
            sink(timestamp - start_stamp, tostring(value))
        end
    end
    local ch = spec.connector
    self.sinks[ch] = self.sinks[ch] or {}
    self.sinks[ch][w.socket] = channel_sink
    w.cleanup = function()
        self.sinks[ch][w.socket] = nil
        self.sinks[ch] = nil
        if self.hb_sinks and self.hb_sinks[ch] then
            self.hb_sinks[ch] = nil
        end
    end
    self:write_metadata(w, spec)
end

channel_specs = {
    utc_date_time={name="utc_date_time",
        description="UTC date and time of position fix"},
    geoidal_height={name="geoidal_height",
        description="Height of geoid above WGS84",
        measurand="Position",
        electrical_units="m",
        range_min=-1000, range_max=1000,
        invalid_data_output_value=1001},
    fix_quality={name="fix_quality",
        description="GPS positional fix quality ($GPGGA)",
        --  0=>"Fix not available"
        --  1=>"Non-differential GPS fix available"
        --  2=>"Differential GPS (WAAS) fix available"
        --  6=>"Estimated"
        -- NOTE:  considered valid unless == 0
        electrical_units="quality",
        range_min=0, range_max=32,
        invalid_data_output_value=255},
    fix_type={name="fix_type",
        description="GPS positional fix type ($GPGSA)",
        --  1=>"not available"
        --  2=>"2D"
        --  3=>"3D"
        electrical_units="quality",
        range_min=0, range_max=3,
        invalid_data_output_value=255},
    ground_heading={name="ground_heading",
        description="Course over ground",
        measurand="Orientation",
        electrical_units="degrees",
        range_min=0, range_max=360,
        invalid_data_output_value=361},
    ground_speed_knots={name="ground_speed_knots",
        description="Speed over ground in knots",
        measurand="Velocity",
        electrical_units="knots",
        range_min=0, range_max=1000,
        invalid_data_output_value=1001},
    altitude={name="altitude",
        description="Altitude relative to mean sea level",
        measurand="Position",
        electrical_units="m",
        range_min=-10000, range_max=100000,
        invalid_data_output_value=100001},
    horizontal_dilution_of_precision={name="horizontal_dilution_of_precision",
        description="Horizontal dilution of precision",
        measurand="Precision",
        range_min=0, range_max=100,
        invalid_data_output_value=101},
    latitude={name="latitude",
        description="Latitude",
        measurand="Position",
        electrical_units="degrees",
        range_min=-90, range_max=90,
        invalid_data_output_value=91},
    longitude={name="longitude",
        description="Longitude",
        measurand="Position",
        electrical_units="degrees",
        range_min=-180, range_max=180,
        invalid_data_output_value=181},
    magnetic_variation={name="magnetic_variation",
        description="Magnetic variation",
        electrical_units="degrees",
        range_min=-180, range_max=180,
        invalid_data_output_value=181},
    number_of_satellites_in_use={name="number_of_satellites_in_use",
        description="Number of satellites in use",
        electrical_units="satellites",
        range_min=0, range_max=24,
        invalid_data_output_value=255},
    position_dilution_of_precision={name="position_dilution_of_precision",
        description="Position dilution of precision",
        measurand="Precision",
        range_min=0, range_max=100,
        invalid_data_output_value=101},
    utc_seconds={name="utc_seconds",
        description="UTC seconds (fractional)",
        measurand="Time",
        electrical_units="seconds",
        range_min=0, range_max=60,
        invalid_data_output_value=61},
    vertical_dilution_of_precision={name="vertical_dilution_of_precision",
        description="Vertical dilution of precision",
        measurand="Precision",
        range_min=0, range_max=100,
        invalid_data_output_value=101},
}

-- Validation functions for GPS sentences.
--   Parameters:
--       nbr_sentence_fields == number of fields in a
--           well formed sentence;
--       fields == table of the fields included in the sentence
--           which are of interest to us.
local function validate_gpgga(nbr_sentence_fields, fields)
    local good = true
    if nbr_sentence_fields < 9 or
            not fields.utc_seconds or not fields.latitude or
            not fields.longitude or not fields.fix_quality or
            not fields.number_of_satellites_in_use or
            not fields.horizontal_dilution_of_precision or
            not fields.altitude or not fields.geoidal_height then
        good = false
    end
    return good
end

local function validate_gpgsa(nbr_sentence_fields, fields)
    local good = true
    if nbr_sentence_fields < 4 or not fields.fix_type or
            not fields.position_dilution_of_precision or
            not fields.horizontal_dilution_of_precision or
            not fields.vertical_dilution_of_precision then
        good = false
    end
    return good
end

local function validate_gprmc(nbr_sentence_fields, fields)
    local good = true
    if nbr_sentence_fields < 7 or
            not fields.utc_date_time or not fields.utc_seconds or
            not fields.latitude or not fields.longitude or
            not fields.ground_speed_knots or not fields.ground_heading or
            not fields.magnetic_variation then
        good = false
    end
    return good
end

local filters = {
    {
        provides = {
            "utc_seconds",
            "latitude",
            "longitude",
            "fix_quality",
            "number_of_satellites_in_use",
            "horizontal_dilution_of_precision",
            "altitude",
            "geoidal_height",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGGA"},
        },
        action = function(pkt)
            local f = pkt.fields
            local fdr = pkt.field_descrs
            return {
                utc_seconds = fix_seconds(f, fdr[0]),
                latitude = fix_lat(f, fdr[1], fdr[2]),
                longitude = fix_lon(f, fdr[3], fdr[4]),
                fix_quality = field_to_number(f, fdr[5]),
                number_of_satellites_in_use = field_to_number(f, fdr[6]),
                horizontal_dilution_of_precision = field_to_number(f, fdr[7]),
                altitude = field_to_number(f, fdr[8]),
                geoidal_height = field_to_number(f, fdr[10]),
            }, true, validate_gpgga
        end
    },
    {
        provides = {
            "fix_type",
            "position_dilution_of_precision",
            "horizontal_dilution_of_precision",
            "vertical_dilution_of_precision",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPGSA"},
        },
        action = function(pkt)
            local f = pkt.fields
            local fdr = pkt.field_descrs
            local nbr_fields = pkt.nbr_of_fields

            return {
                fix_type = field_to_number(f, fdr[1]),
                position_dilution_of_precision =
                    field_to_number(f, fdr[nbr_fields - 3]),
                horizontal_dilution_of_precision =
                    field_to_number(f, fdr[nbr_fields - 2]),
                vertical_dilution_of_precision =
                    ascii_dec_to_number(f, fdr[nbr_fields - 1].start)
            }, true, validate_gpgsa
        end
    },
    {
        provides = {
            "utc_date_time",
            "utc_seconds",
            "latitude",
            "longitude",
            "ground_speed_knots",
            "ground_heading",
            "magnetic_variation",
        },
        spec = {
            {field="protocol", exact_value="nmea"}, -- "invariant" on source?
            {field="sentence", exact_value="$GPRMC"},
        },
        action = function(pkt)
            local f = pkt.fields
            local fdr = pkt.field_descrs
            local gps_status = field_to_string(f, fdr[1])
            return {
                utc_date_time = fix_date_time(f, fdr[8], fdr[0]),
                utc_seconds = fix_seconds(f, fdr[0]),
                latitude = fix_lat(f, fdr[2], fdr[3]),
                longitude = fix_lon(f, fdr[4], fdr[5]),
                ground_speed_knots = field_to_number(f, fdr[6]),
                ground_heading = field_to_number(f, fdr[7]),
                magnetic_variation = fix_mv(f, fdr[9], fdr[10])
            }, gps_status == 'A', validate_gprmc
        end
    },
}

function gps_proxy:available_channels(hostname)
    local o = {}
    -- KLUDGE for raw NMEA
    o[1] = {
        name = 'raw_' .. hostname:gsub("%W", "_"),
        connector = '@' .. hostname,
        description = "Raw NMEA sentences",
        data_type = 'message_raw', -- KLUDGE message_ascii or message_utf8
        category = 'source/hbm_cx23r/msg_gps',
        valid_controls = {
            oscillator_control = true,
        },
    }
    o[2] = {
        name = 'control',
        connector = 'control.status@' .. hostname,
        description = "GPS Control Channel",
        data_type = 'message_raw',
        category = 'source/hbm_cx23r/gps',
    }
    self.external_names[''] = o[1].connector -- KLUDGE for raw NMEA
    self.internal_names[o[1].connector] = '' -- KLUDGE for raw NMEA
    local seen = {}
    for _, f in ipairs(filters) do
        for _, ch_name in ipairs(f.provides) do
            if not seen[ch_name] then
                seen[ch_name] = 1
                --log(log.WARNING, 'ch_name == {0}\n', ch_name)
                assert(channel_specs[ch_name])
                local ch
                if ch_name == 'utc_date_time' then
                    -- special case for utc_date_time
                    ch = {
                        message_rate = 5, -- KLUDGE Garmin-5Hz specific
                        category = 'source/hbm_cx23r/msg_gps',
                    }
                else
                    ch = {
                        message_rate = 5, -- KLUDGE Garmin-5Hz specific
                        category = 'source/hbm_cx23r/gps',
                    }
                end
                for k, v in pairs(channel_specs[ch_name]) do
                    ch[k] = v
                end
                ch.connector = ch_name .. "@" .. hostname
                ch.data_type = ch.range_max
                    and 'message_float64' or 'message_raw'
                o[#o + 1] = ch
                self.external_names[ch_name] = ch.connector
                self.internal_names[ch.connector] = ch_name
            end
        end
    end
    self.channels = o
    return self.channels
end

function gps_proxy:sentence(timestamp, text_start, text_len)
    -- dbg('sentence %f %s\n', tonumber(timestamp), text)
    timestamp = tonumber(timestamp) -- KLUDGE
    local name_map = self.external_names
    local sentences = self.nmea:nmea_filter{timestamp=timestamp,
                                            text=text_start,
                                            length=text_len}

    for i = 1, #sentences do
        local pkt = sentences[i]
        -- dbg('  sentence %d, %s, %f, %s', i, pkt, tonumber(pkt.timestamp), pkt.nmea)
        pkt.field_descrs, pkt.nbr_of_fields =
            identify_fields(pkt.fields)

        -- KLUDGE to spit raw NMEA
        local k = name_map['']
        local sinks = self.sinks[k]
        if sinks then
            for _, sink in pairs(sinks) do
                sink(timestamp, k, ffi.string(pkt.nmea:head(), pkt.nmea_len))
            end
        end

        for fi = 1, #filters do
            local f = filters[fi]
            local matched = true
            for si = 1, #f.spec do
                local spec = f.spec[si]
                local exact = spec.exact_value
                if exact and pkt[spec.field] ~= exact then
                    matched = false
                    break
                end
            end
            local o = nil
            local good_status, valid = true
            if matched then
                local validate = nil
                o, good_status, validate = f.action(pkt)
                valid = validate(pkt.nbr_of_fields, o)
                if not valid then
                    -- log invalid sentence
                    local logmsg = "invalid GPS sentence: " ..
                        tostring(pkt.sentence)
                    log(log.WARNING, logmsg)
                end
            end

            if matched and valid then
                local send_channels = true
                -- If we are synchronizing time to GPS time then we need to
                -- use the time synchronizing servo.
                local trailing_0 = find_trailing_0(o.utc_date_time)
                if trailing_0 ~= nil then

                    -- Mark the GPS as locked after we have received at least
                    -- 4 continous seconds of fixes.  Once locked, reset back
                    -- to init state if we ever go for more than 2 seconds
                    -- without a new fix.  All this does is prevent the servo
                    -- from running until we know we have valid PPS pulses!
                    if self.state == 'gps_lock_init' then
                        log:notice("GPS lock wait")
                        self.state = 'gps_lock_wait'
                        self.valid_fix_count = 1
                        self.lock_check_timer = ev.Timer.new(function(l, t, e)
                                if self.state == 'gps_lock_wait' and
                                    self.valid_fix_count >= 4 then
                                    log:notice("GPS locked")
                                    self.state = 'gps_locked'
                                    self.lock_monitor_timer =
                                        ev.Timer.new(function(l, t, e)
                                                log:notice("GPS lock lost")
                                                self.state = 'gps_lock_init'
                                                t:stop(l)
                                                self.lock_monitor_timer = nil
                                            end, 2, 2)
                                    self.lock_monitor_timer:start(
                                        ev.Loop.default)
                                else
                                    log:notice("<4 fixes in 4s - to init state")
                                    self.state = 'gps_lock_init'
                                end
                            end, 5, 0):start(ev.Loop.default)
                    elseif self.state == 'gps_lock_wait' then
                        self.valid_fix_count = self.valid_fix_count + 1
                    end
                    if self.state == 'gps_locked' and
                        self.lock_monitor_timer then
                        self.lock_monitor_timer:again(ev.Loop.default)
                    end

                    if time_synced_to_gps then
                        if self.pps_ts ~= self.servo_ts and
                            timestamp - self.pps_ts < self.pps_interval then
                            local tmp_ts = ts.membuf_to_ts(o.utc_date_time)
                            local fix_stamp = ts.ts_to_stamp(tmp_ts)
                            self:queue_servo(fix_stamp - self.pps_ts)
                            self.servo_ts = self.pps_ts
                        end
                    elseif not self.time_set and not self.dont_set_time then
                        self.time_set = true
                        local tmp_ts = ts.membuf_to_ts(o.utc_date_time)
                        local tv = ffi.new('struct timeval[1]')
                        tv[0].tv_sec = tmp_ts.s
                        tv[0].tv_usec = 0
                        C.settimeofday(tv, nil)
                    end
                end

                -- Temporary KLUDGE, until sink replaced
                if pkt.sentence == "$GPGGA" then
                    -- only send time info when $GPRMC channels
                    -- are valid and being sent
                    if self.gprmc_channels_good then
                        o.utc_seconds = tostring(o.utc_seconds)
                    else
                        o.utc_seconds = nil
                    end
                    o.latitude = tostring(o.latitude)
                    o.longitude = tostring(o.longitude)
                    if tonumber(o.fix_quality) == 0 then
                        -- the fix quality is bad, so only
                        -- allow fix quality values through
                        local tmp_o = o
                        o = {}
                        o["fix_quality"] = tmp_o.fix_quality
                        o["number_of_satellites_in_use"] =
                            tmp_o.number_of_satellites_in_use
                        o["horizontal_dilution_of_precision"] =
                            tmp_o.horizontal_dilution_of_precision
                    elseif tonumber(o.number_of_satellites_in_use) < 4 then
                        o["altitude"] = nil
                    end
                elseif pkt.sentence == "$GPRMC" then
                    if good_status then
                        self.gprmc_channels_good = true
                        local date_str = o.utc_date_time:extract(0, 10)
                        date_str = date_str .. "T" ..
                            o.utc_date_time:__tostring()
                        o.utc_date_time = date_str
                        o.utc_seconds = tostring(o.utc_seconds)
                        o.latitude = tostring(o.latitude)
                        o.longitude = tostring(o.longitude)
                        o.magnetic_variation =
                            tostring(o.magnetic_variation)
                    else
                        self.gprmc_channels_good = false
                        send_channels = false
                    end
                end

                if send_channels then
                    for name, v in pairs(o) do
                        if name and v then
                            local k = name_map[name]
                            local sinks = self.sinks[k]
                            if sinks then
                                for _, sink in pairs(sinks) do
                                    sink(timestamp, k, v)
                                end
                            end
                        end
                    end
                end
            end
        end
    end
end

function gps_proxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params, cb)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

-- Set the system time from the GPS.  Note, the response is deferred until
-- either the time is set or a timeout expires.  If the time is set, a success
-- message is returned, otherwise an error response is returned.
function gps_proxy:set_time_from_gps(params, cb)
    self.time_set = false
    local retry = 4
    ev.Timer.new(function(l, t, e)
            retry = retry - 1
            if retry == 0 then
                cb({
                    msg = 'Unable to set system time set to GPS time',
                    result = 'error',
                })
                t:stop(l)
            elseif self.time_set then
                cb({
                    msg = 'System time set to GPS time',
                    result = 'success',
                })
                t:stop(l)
            end
        end, 1, 1):start(ev.Loop.default)
end

function gps_proxy:osc_mode(params)
    if not params then
        return {
            msg = 'osc_mode request not specified',
            result = 'error',
        }
    end
    local msg
    if params[1] == 'start' then
        if not self.osc_control then
            -- Setup Oscillator Calibration Environment
            armada.timing:AdjustClockRate(0, 'osc_mode_on')
            log:notice(__"GPS: Entering Si514 Calibration Mode on {0}",
                       self.proxy_hostname)
            if self.gps_lock_monitor then
                self.gps_lock_monitor:stop(ev.Loop.default)
                self.gps_lock_monitor = nil
            end
            self.osc_control = true
            msg = string.format("osc_mode %s succeeded on %s",
                                params[1], self.proxy_hostname)
        else
            msg = string.format("osc_mode %s already executed on %s",
                                params[1], self.proxy_hostname)
        end
        return {
            msg = msg,
            result = 'success',
        }
    elseif params[1] == 'stop' then
        if self.osc_control then
            -- Restore Default Oscillator Environment
            armada.timing:AdjustClockRate(0, 'osc_mode_off')
            log:notice(__"GPS: Leaving Si514 Calibration Mode on {0}",
                       self.proxy_hostname)
            self.osc_collecting = false
            self.osc_servo = false
            self.last_offset = 1e30
            self.integral = 0
            self.pps_rising_events = 0
            self.dt = .2 -- GARMIN ONLY!!
            self.pps_rate = 5
            self.pps_interval = self.clk_frequency / self.pps_rate
            self.Kp, self.Ki, self.Kd = 37, 10, 4
            self.osc_control = false
            set_gps_regulation()
            msg = string.format("osc_mode %s succeeded on %s",
                                params[1], self.proxy_hostname)
        else
            msg = string.format("osc_mode %s already executed on %s",
                                params[1], self.proxy_hostname)
        end
        return {
            msg = msg,
            result = 'success',
        }
    else
        return {
            msg = string.format("osc_mode %s failed on %s",
                                params[1], self.proxy_hostname),
            result = 'error',
        }
    end
end

function gps_proxy:osc_read(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_read not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    return {
        m_divider = armada.timing:OscillatorRead(),
        result = 'success',
    }
end

function gps_proxy:osc_write(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_write not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not params[1] then
        return {
            msg = string.format("osc_write value not specified on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local ppm = tonumber(params[1])
    return {
        m_divider = armada.timing:OscillatorWrite(ppm),
        result = 'success',
    }
end

function gps_proxy:osc_collect(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_collect not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not params[1] then
        return {
            msg = string.format("osc_collect count not specified on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local count = tonumber(params[1])
    self.osc_collecting = false
    self.osc_servo = false
    self.osc_offset = 0
    self.osc_pps_data = {}
    self.osc_pps_count = 0
    self.osc_pps_ts = 0
    self.last_offset = 1e30
    self.integral = 0
    self.pps_rising_events = 0
    if params[2] then
        if params[2] == "on" then
            self.osc_servo = true
            if params[3] and params[4] and params[5] then
                self.Kp = params[3]
                self.Ki = params[4]
                self.Kd = params[5]
            end
            if params[6] then
                -- params[6] is the rate at which PPS rising edges
                -- will arrive at the GPS PPS connector pin.
                self.pps_rate = params[6]
                self.pps_interval = math.ceil(self.clk_frequency / params[6])
                self.dt = 1 / params[6]
            else
                return {
                    msg = string.format(
                                "osc_collect PPS interval not specified on %s",
                                self.proxy_hostname),
                    result = 'error',
                }
            end
            gps_proxy.servo = gps_proxy.time_not_synced_servo
            log:notice(__"GPS: Si514 PPS Servo Enabled")
        elseif params[2] == "off" then
            self.osc_servo = false
            gps_proxy.servo = function(self, offset)
                                  self.last_offset = offset
                              end
            log:notice(__"GPS: Si514 PPS Servo Disabled")
            if params[3] then
                -- params[3] is the rate at which PPS rising edges
                -- will arrive at the GPS PPS connector pin.
                self.pps_rate = params[3]
                self.pps_interval = math.ceil(self.clk_frequency / params[3])
                self.dt = 1 / params[3]
            else
                return {
                    msg = string.format(
                                "osc_collect PPS interval not specified on %s",
                                self.proxy_hostname),
                    result = 'error',
                }
            end
        else
            return {
                msg = string.format(
                                "osc_collect servo state not specified on %s",
                                self.proxy_hostname),
                result = 'error',
            }
        end
    end
    self.osc_collecting = true
    repeat
        ev.Loop.default:loop('once')
    until self.osc_pps_count == count
    return {
        data = self.osc_pps_data,
        msg = string.format("osc_collect %s succeeded on %s",
                            params[1], self.proxy_hostname),
        result = 'success',
    }
end

function gps_proxy:osc_store(params)
    if not self.osc_control then
        return {
            msg = string.format("osc_store not in osc_mode on %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end
    local offset = 0
    if params and params[1] then
        offset = armada.timing:OscillatorOffset(tonumber(params[1]))
    end
    return {
        msg = string.format("osc_store %d succeeded on %s",
                            offset, self.proxy_hostname),
        result = 'success',
    }
end

return gps_proxy
