#!/usr/bin/lua

require('lj_include'){
    'unistd',
    'sys/types',
    'signal',
    'net/if',
}

local ffi = require 'ffi'
local C = ffi.C
local ev = require 'ev'
local struct = require 'struct'
require 'middleclass'
local cjson = require 'cjson.safe'

local version = {}
local f = io.open('/etc/version.json')
if f then
    version = cjson.decode(f:read('*a')) or {}
    f:close()
end

local net_ffi = require 'net_ffi'

local log = require 'log'

-- Deal with command line options
local ipv6, announcement_period = ...

ipv6 = (ipv6 and true) or false
announcement_period = announcement_period or 5
if announcement_period <= 0 then
    announcement_period = 1
end

-- Wait until the external network interface is up
local nf = require 'net_functions'
nf.wait_carrier('any', 'ignore', {'eth0'})

ffi.cdef[[
enum {
    CRC_MASK = 0xff,
};

typedef uint32_t        in_addr_t;

char *inet_ntoa(struct in_addr in);
]]

-- Function used to determine whether the specified interface has ANY IPv4
-- addresses assigned to it.
local function have_ipv4_addr()
    local addrs = net_ffi.getifaddrs('eth0')
    if addrs and #addrs > 0 then return true end
end

local DEFAULT_QUANTUMX_DOMAIN_ADDRESS_V6 = "ff02::1"
local DEFAULT_QUANTUMX_DOMAIN_ADDRESS = "239.255.77.76"
local MX_DISCOVERY_MULTICAST_PORT = "31416"

local htonl = ffi.abi('le') and bit.bswap or bit.tobit

local htonll_union =
    ffi.typeof('union { unsigned long long u64; unsigned long u32[2]; }')

local function htonll(x)
    if ffi.abi('le') then
        local v = htonll_union(x)
        v.u32[0], v.u32[1] = bit.bswap(v.u32[1]), bit.bswap(v.u32[0])
        return v.u64
    end
    return x
end

local function htons(x)
    if ffi.abi('le') then
        return bit.rshift(htonl(x), 16)
    end
    return htonl(x)
end
local ntohl = htonl
local ntohs = htons
local ntohll = htonll

local net = class('net')

-- Function used to get the IPv4 address of an interface
function net:get_ipv4_address(fd)
    local ifreq = ffi.new('struct ifreq[1]')
    ffi.copy(ifreq[0].ifr_ifrn.ifrn_name, self.iface_name)
    C.ioctl(fd, C.SIOCGIFADDR, ifreq)
    local addr = ffi.new('struct sockaddr_in')
    ffi.copy(addr, ifreq[0].ifr_ifru.ifru_addr, ffi.sizeof(addr))

    return addr.sin_addr
end

function net:ipv4_socket_init(port, fd)
    -- Need the IPv4 address of the specified interface
    self.myaddr = self:get_ipv4_address(fd)
    if self.myaddr.s_addr == 0 then
        return false
    end

    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    hints.ai_family = self.af
    local hst = self.mhst
    local r = C.getaddrinfo(hst, nil, hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return false
    end
    r = res[0]

    -- Extract the address out of the returned struct
    local a = ffi.new('struct sockaddr_in')
    ffi.copy(a, r.ai_addr, ffi.sizeof(a))

    -- Done with address info
    C.freeaddrinfo(res[0])

    -- Set up for setting socket options
    local mr = ffi.new('struct ip_mreq')
    mr.imr_multiaddr.s_addr = a.sin_addr.s_addr
    mr.imr_interface.s_addr = self.myaddr.s_addr

    -- Limit multicasting to interface with this address
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_MULTICAST_IF, mr, ffi.sizeof(mr))

    -- Join the multicast group
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_ADD_MEMBERSHIP, mr, ffi.sizeof(mr))

    -- Disable multicast loopback
    local enable = ffi.new('int[1]', {0})
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_MULTICAST_LOOP,
        enable, ffi.sizeof(enable))

    -- Set the hop limit
    local hop_limit = ffi.new('int[1]', {255})
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_MULTICAST_TTL,
        hop_limit, ffi.sizeof(hop_limit))

    return true
end

function net:ipv6_socket_init(port, fd)

    -- Only allow multicast on the interface we are bound to
    C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_MULTICAST_IF, self.idx,
        ffi.sizeof(self.idx))

    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    hints.ai_family = self.af
    local r = C.getaddrinfo(self.mhst, nil, hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return false
    end
    r = res[0]

    -- Extract the address out of the returned struct
    local fam, prt, flow, adr = struct.unpack('H H I4 c16',
        ffi.string(r.ai_addr, 24))

    -- Done with address info
    C.freeaddrinfo(res[0])

    -- Join the multicast group
    local mr = ffi.new('struct ipv6_mreq', {
            interface = self.idx,
        })
    for i = 0, 15 do
            mr.multiaddr.__in6_u.__u6_addr8[i] = string.byte(adr, i + 1)
    end
    local res = C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_JOIN_GROUP, mr,
        ffi.sizeof(mr))

    -- Disable multicast loopback
    local enable = ffi.new('int[1]', {0})
    C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_MULTICAST_LOOP,
        enable, ffi.sizeof(enable))

    -- Set multicast hop limit
    local ttl = ffi.new('int[1]', 255)
    C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_MULTICAST_HOPS, ttl,
        ffi.sizeof(ttl))

    return true
end

function net:socket_init(port)
    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    if tonumber(port) == 0 then
        hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    else
        hints.ai_flags = C.AI_PASSIVE + C.AI_NUMERICHOST + C.AI_NUMERICSERV
    end
    hints.ai_family = self.af
    local r = C.getaddrinfo(nil, tostring(port), hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return -1
    end
    r = res[0]

    -- Create a socket
    local fd = C.socket(r.ai_family, r.ai_socktype, r.ai_protocol)
    if fd < 0 then
        local e = ffi.errno()
        log:log(log.ERR,
            string.format("Can not create socket %d", e))
        return -1
    end

    -- Bind to the specific interface so we aren't sending everywhere
    local stat = C.setsockopt(fd, C.SOL_SOCKET, C.SO_BINDTODEVICE,
        self.iface_name, self.iface_name:len())
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR,
            string.format("Can not bind socket to iface %d", e))
        C.close(fd)
        return -1
    end

    -- Bind the socket
    stat = C.bind(fd, r.ai_addr, r.ai_addrlen)
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR, string.format("Can not bind socket %d", e))
        C.close(fd)
        return -1
    end

    -- Done with address info
    C.freeaddrinfo(res[0])

    if tonumber(port) == 0 then
        return fd
    end

    -- Allow immediate reuse of the addrees
    local enable = ffi.new('int[1]', {1})
    stat = C.setsockopt(fd, C.SOL_SOCKET, C.SO_REUSEADDR,
        enable, ffi.sizeof(enable))
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR, string.format("Can not set REUSEADDR %d", e))
        C.close(fd)
        return -1
    end

    -- Set the hop limit
    local hop_limit = ffi.new('int[1]', {32})
    stat = C.setsockopt(fd, C.IPPROTO_IP, C.IP_MULTICAST_TTL,
        hop_limit, ffi.sizeof(hop_limit))
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR, string.format("Can not set hop limit %d", e))
        C.close(fd)
        return -1
    end

    -- Do remainder of V4/V6 initialization
    if self.af == C.AF_INET6 then
        if self:ipv6_socket_init(port, fd) == false then
            C.close(fd)
            fd = -1
        end
    elseif self:ipv4_socket_init(port, fd) == false then
        C.close(fd)
        fd = -1
    end

    return fd
end

function net:receive(sock, buf)
    local ret = C.recvfrom(sock, buf.buf + buf.offset, buf.size - buf.offset,
        C.MSG_DONTWAIT, nil, nil)
    if ret < 0 then
        local errno = ffi.errno()
        if errno == C.EAGAIN or errno == C.EINTR then
            return 0
        end
        log:log(log.ERR, string.format("recvfrom fail: %s",
            ffi.string(C.strerror(errno))))
        return
    elseif ret == 0 then
        return
    end
    buf.offset = buf.offset + ret
    return ret
end

function net:shutdown()
    C.close(self.multicast_fd)
end

--[==[
function net:process_announcement(json)
    if json.method and json.method == 'announce' then
        local module_name
        local ip_addr
        local ipv6_addr
        local fwver

        -- Set of services that have to be available before we attempt
        -- to run a proxy for a particular quantum device.
        local smap = {
            daqStream = true,
            hbmProtocol = true,
            http = true,
            jetd = true,
            jetws = true,
            ssh = true,
        }

        -- Figure out which services are available
        if json.params.services then
            for _, v in ipairs(json.params.services) do
                if v.port and v.port ~= 0 then
                    smap[v] = nil
                end
            end
        end

        -- Only attempt start proxies on units that have all expected
        -- services up and going.
        if #smap == 0 then

            -- Prefer the ZCIP address otherwise default to other IPv4
            -- address if it exists
            if json.params.interface and json.params.interface.ipv4 then
                for _, v in ipairs(json.params.interface.ipv4) do
                    if v.address and v.address:match('^169%.254%.') then
                        ip_addr = v.address
                        break
                    end
                end
                ip_addr = ip_addr or
                    json.params.interface.ipv4[1].address
            elseif json.params.netSettings.interface.ipv4 then
                for _, v in ipairs(
                    json.params.netSettings.interface.ipv4) do
                    if v.address and v.address:match('^169%.254%.') then
                        ip_addr = v.address
                        break
                    end
                end
                ip_addr = ip_addr or
                    json.params.netSettings.interface.ipv4[1].address
            end
            if json.params.interface and json.params.interface.ipv6 and
                json.params.interface.ipv6[1].address then
                ipv6_addr = json.params.interface.ipv6[1].address
            elseif json.params.netSettings.interface.ipv6 and
                json.params.netSettings.interface.ipv6[1].address then
                ipv6_addr =
                    json.params.netSettings.interface.ipv6[1].address
            end
            if json.params.device.type then
                module_name = json.params.device.type
            end
            if json.params.device.firmware_version then
                fwver = json.params.device.firmware_version
            elseif json.params.device.firmwareVersion then
                fwver = json.params.device.firmwareVersion
            end
            if fwver and fwver:find('^3.') and module_name and
                not module_name:find('%d+B') then
                module_name = module_name .. 'B'
            end
            if module_name and ipv6_addr then
                ip_addr = ipv6_addr .. '%ethint'
                log:debug(string.format(
                    "QuantumX device %s found at address %s",
                    module_name, ip_addr))
                    self:start_proxy(module_name, ip_addr)
                self.mx_fw_ver = self.mx_fw_ver or {}
                if not self.mx_fw_ver[ip_addr] then
                    self.mx_fw_ver[ip_addr] = fwver
                end
            else
                log:log(log.ERR, string.format(
                    "Device %s does not have an IPv6 address", module_name))
            end
        else
            log:debug("QuantumX device but not all services were available")
        end
    end
end
]==]

local announce = {
    jsonrpc = '2.0',
    method = 'announce',
    params = {
        apiVersion = '1.0',
        device = {
            familyType = 'SoMatXR',
            firmwareVersion = version.version or 'unknown',
            hardwareId = 'CX23R_R1D',
            name = 'Somat CX23R',
            type = 'CX23R',
        },
        expiration = announcement_period * 3,
        netSettings = {
            interface = {
                name = 'eth0',
                type = 'ethernet',
                description = 'CX23R External network interface',
                configurationMethod = 'multiple',
            },
        },
        services = {
            {
                type = 'https',
                port = 443,
            },
            {
                type = 'ssh',
                port = 22,
            },
        },
    },
}

-- Get all of the IP addresses assigned to the external network interface
function net:get_addresses()
    local ns = announce.params.netSettings.interface
    ns.ipv4 = nil
    ns.ipv6 = nil
    local fh = io.popen('ip addr show dev eth0')
    if fh then
        for line in fh:lines() do
            local uuid = line:match('link/ether (.+) brd')
            local v, addr, mask = line:match('(inet%d?) (.-)/(%d+) ')
            if v and v == 'inet' and addr and mask then
                local n = 0
                for i = 1, mask do
                    n = bit.rshift(n, 1)
                    n = bit.bor(n, 0x80000000)
                end
                local netmask = C.inet_ntoa(ffi.new('struct in_addr', {n}))
                ns.ipv4 = ns.ipv4 or {}
                ns.ipv4[#ns.ipv4 + 1] = {
                    address = addr,
                    netmask = ffi.string(netmask),
                }
            elseif v and v == 'inet6' and addr and mask then
                ns.ipv6 = ns.ipv6 or {}
                ns.ipv6[#ns.ipv6 + 1] = {
                    address = addr,
                    prefix = tonumber(mask),
                }
            elseif uuid then
                announce.params.device.uuid = uuid:gsub(':', '')
            end
        end
        fh:close()
    end

    -- And the default IPv4 route
    local fh = io.popen('/sbin/route -n | grep eth0')
    if fh then
        for line in fh:lines() do
            local df, gw = line:match('^(0%.0%.0%.0)%s-(%d+%.%d+%.%d+%.%d+)')
            if df and gw then
                announce.params.netSettings.defaultGateway = {
                    ipv4Address = gw,
                }
            end
        end
        fh:close()
    end
end

function net:set_sendto_addr(host, port)
    local sa = ffi.new('char[8192]')
    local salen = ffi.new('socklen_t[1]', {ffi.sizeof(sa)})

    local hints = ffi.new('struct addrinfo[1]')
    hints[0].ai_flags = C.AI_NUMERICSERV + C.AI_NUMERICHOST
    hints[0].ai_socktype = C.SOCK_DGRAM
    hints[0].ai_family = C.AF_UNSPEC
    local ret = ffi.new('struct addrinfo *[1]')
    port = tostring(port)
    local status = C.getaddrinfo(host, port, hints, ret)

    assert(status == 0, string.format("Unable to get addr info: %s",
        ffi.string(C.gai_strerror(status))))

    if self.sendto_addr then
        C.free(self.sendto_addr)
    end
    self.sendto_addr = ffi.cast('struct sockaddr *',
        C.calloc(ret[0].ai_addrlen, ffi.sizeof('char')))
    ffi.copy(self.sendto_addr, ret[0].ai_addr, ret[0].ai_addrlen)
    self.sendto_addrlen = ret[0].ai_addrlen
    C.freeaddrinfo(ret[0])
end

function net:initialize(ipv6, iface_name)
    self.iface_name = iface_name or 'eth0'
    self.idx = ffi.new('int[1]')
    self.idx[0] = C.if_nametoindex(self.iface_name)

    if ipv6 then
        self.mhst = DEFAULT_QUANTUMX_DOMAIN_ADDRESS_V6
        self.af = C.AF_INET6
        self.multicast_fd = self:socket_init(MX_DISCOVERY_MULTICAST_PORT)
        if self.multicast_fd < 0 then
            return self
        end
    else
        self.mhst = DEFAULT_QUANTUMX_DOMAIN_ADDRESS
        self.af = C.AF_INET
        self.multicast_fd = self:socket_init(MX_DISCOVERY_MULTICAST_PORT)
        if self.multicast_fd < 0 then
            return self
        end
    end

    -- Set up address we are going to send announcments to
    self:set_sendto_addr(self.mhst, MX_DISCOVERY_MULTICAST_PORT)

    self.mcast_ibuf = {}
    self.mcast_ibuf.size = 65536
    self.mcast_ibuf.buf = ffi.new('char[?]', self.mcast_ibuf.size)
    self.mcast_ibuf.offset = 0

    self.mcast_rx_handler = ev.IO.new(function(loop, io, revents)
            local num_rx = self:receive(self.multicast_fd, self.mcast_ibuf)
            while self.mcast_ibuf.offset > 0 do
                local jstr =
                    ffi.string(self.mcast_ibuf.buf, self.mcast_ibuf.offset)
                local s, e = jstr:find('}{')
                if s and e then
                    jstr = string.sub(jstr, 1, s)
                    self.mcast_ibuf.offset = self.mcast_ibuf.offset - e
                    ffi.copy(self.mcast_ibuf.buf, self.mcast_ibuf.buf + e,
                        self.mcast_ibuf.offset)
                else
                    self.mcast_ibuf.offset = 0
                end
--[==[
print(jstr)
                local no_error, decode = pcall(cjson.decode, jstr)
                if no_error then
                    if decode.method and decode.method == 'announce' then
                        self:process_announcement(decode)
                    end
                else
                    log:log(log.WARNING,
                        string.format("Can not decode JSON\n%s", jstr))
                end
]==]
            end
        end, self.multicast_fd, ev.READ)

    self.mcast_rx_handler:start(ev.Loop.default)

    -- Start the send announcement timer
    self.send_announcement_timer = ev.Timer.new(function(loop, t, revents)
            self:get_addresses()
            local cj = cjson.encode(announce)
            if cj then
                C.sendto(self.multicast_fd, cj, #cj, 0,
                    self.sendto_addr, self.sendto_addrlen)
            end
        end, .1, announcement_period)
    self.send_announcement_timer:start(ev.Loop.default)

    self.initialized = true

    return self
end

function net:shutdown()
    if self.send_announcement_timer then
        self.send_announcement_timer:stop(ev.Loop.default)
    end
    if self.multicast_fd and self.multicast_fd > 0 then
        C.close(self.multicast_fd)
    end
end

local n

local timer = ev.Timer.new(function(l, t, e)
        if have_ipv4_addr() then
            n = net:new()
            if n and n.initialized then 
                t:stop(l)
            end
        end
    end, 1, 1)

--timer:start(ev.Loop.default)

nf.carrier_up('eth0', ev.Loop.default,
    function(iface, state)
        if state then   -- Link came up
            timer:start(ev.Loop.default)
        else            -- Link went down
            timer:stop(ev.Loop.default)
            if n then
                n:shutdown()
                n = nil
            end
        end
    end)

-- Run the event processing loop
ev.Loop.default:loop()
