--[[
    This provides the buffers used to support streaming channel data
    out to one or more clients.

    This is NOT a ring buffer.  It is expected to be drained completely
    every time there is a dequeue operation, so each dequeue resets
    the count to zero and there is no need for "front" or "back"
    indices.

    To reduce the amount of copying required, two header fields are
    reserved at the front of the queue's buffer.  This allows filling
    in header values as necessary and then the use of ffi.string()
    without any memory copying to prepend a header.  The second
    header field will contain the byte count of the queue's data
    (not including the header bytes) and will be kept up to date
    by the buffer object's code.

    Since the header fields' size may not be a multiple of the queue
    element size, the queue's buf pointer may not point to the first
    byte of the allocated buffer memory.  A separate pointer will be
    kept that does point to the start of the allocated buffer memory,
    for use in reallocations.
]]

local bit = require 'bit'
local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local ts = require 'timestamps'
require 'sample_types'

ffi.cdef[[
    void *malloc(size_t size);
    void *realloc(void *ptr, size_t size);
    void *memset(void *s, int c, size_t n);

    struct ntp_stamp {
        unsigned long   seconds;
        unsigned long   fraction;
    };
]]

local C = ffi.C

local bit_band = bit.band

local ctor_cache = {}
local mt = {}

local HEADER_LEN = 2 * ffi.sizeof("int")
assert(HEADER_LEN)

function mt:dump()
    local header = ffi.cast("unsigned int*", self.buf_head)
    local el_type = ffi.string(self.el_type)
    local dump_tbl = {
        count = self.count,
        element_size = self.element_size,
        size = self.size,
        sample_rate = self.sample_rate,
        enq_attempted = self.enq_attempted,
        subscribed = self.subscribed,
        element_type = el_type,
        alloc_ptr = self.alloc_ptr,
        header_ptr = self.buf_head,
        header_0 = header[0],
        header_1 = header[1],
        data_ptr = self.buf
    }
    io.stderr:write(string.format("%s", dbg.dump(dump_tbl)))
    dump_tbl = {}
    local buf_ptr = self.buf
    for i = 0, self.count - 1 do
        local byte = ffi.cast(el_type, self.buf[i])
        dump_tbl[#dump_tbl + 1] = string.format("%x", tonumber(byte))
    end
    io.stderr:write(string.format("***data***\n%s\n", dbg.dump(dump_tbl)))
end

function mt:nbr_subscribed()
  return self.subscribed
end

function mt:is_subscribed()
  return self.subscribed > 0
end

function mt:subscribe()
    self.subscribed = self.subscribed + 1
    return self.subscribed
end

function mt:unsubscribe()
    if self.subscribed > 0 then
        self.subscribed = self.subscribed - 1
    else
        self.subscribed = 0
    end
    if self.subscribed <= 0 then
        self:clear()
    end
    return self.subscribed
end

function mt:enq()
    self.enq_attempted = self.enq_attempted + 1
    if self.subscribed > 0 then
        if self:available(true) ~= 0 then
            local entry = self.buf + self.count
            self.count = self.count + 1
            return entry
        end
    end
end

function mt:_enq_sequential_value(x)
    self.enq_attempted = self.enq_attempted + 1
    if self.subscribed > 0 then
        if self:available(true) ~= 0 then
            self.buf[self.count] = x
            self.count = self.count + 1
            return true
        end
    end
end

function mt:adjust_attempted(x)
    self.enq_attempted = self.enq_attempted + x
end

function mt:_enq_message_value(x)
    self.enq_attempted = self.enq_attempted + 1
    if self.subscribed > 0 then
        if self:available(true) ~= 0 then
            ffi.copy(self.buf + self.count, x, self.element_size)

            -- Convert the TS to NTP format
            local v = ffi.cast('unsigned long long *', self.buf + self.count)
            local ntp = ffi.cast('struct ntp_stamp *', v)
            local s, f = math.modf(ts.stamp_to_epoch(tonumber(v[0])))
            ntp.seconds = s
            ntp.fraction = math.floor(f * 2^32)
            self.count = self.count + 1

            return true
        end
    end
end

-- Returns a pointer to the start of the queue's contents
-- and a byte count of the queues contents, including the
-- two 32 bit "header" fields.
function mt:get_blk()
    if self.count == 0 then
        return nil
    end
    assert(self.count > 0 and self.count <= self.size)
    -- update header's byte count of data length
    local header = ffi.cast("unsigned int*", self.buf_head)
    local data_len = self.count * self.element_size
    header[1] = data_len
    return self.buf_head, data_len + HEADER_LEN
end

-- Return amount of space left in buffer and expand if full and directed to
-- do so.
function mt:available(expand)
    assert(self.count >= 0 and self.count <= self.size)
    if expand and (self.size - self.count) == 0 then
        self:expand()
    end
    return self.size - self.count
end

-- Clear out the buffer
function mt:clear()
    self.count = 0

    -- The base of the allocated buffer is reserved for the HBM streaming
    -- protocol header.  Clear out the "count" portion of it.
    local header = ffi.cast("unsigned int *", self.buf_head)
    header[1] = 0
end

-- Allocate more space in the buffer
function mt:expand(num)
    if self.size >= 65536 then
        return
    end
    self.size = num or (self.size * 2)

    -- Amount of space needed for number of elements plus reserved space for
    -- HBM streaming protocol header
    local size_to_alloc = (self.size * self.element_size) + HEADER_LEN

    self.alloc_ptr = C.realloc(self.alloc_ptr, size_to_alloc)
    assert(self.alloc_ptr and self.alloc_ptr ~= nil)

    -- Set pointer to HBM streaming protocol header
    self.buf_head = ffi.cast('unsigned char *', self.alloc_ptr)

    -- Set ponter to elements
    self.buf = ffi.cast(ffi.string(self.el_type) .. "*",
        self.buf_head + HEADER_LEN)
end

-- Return number of elements in the buffer
function mt:__len()
    return self.count
end

mt.__index = mt

local function buf_ctor(t, dtype)
    assert(t)
    local streambuf_type = string.format('struct %s_streambuf', t)
    if not ctor_cache[streambuf_type] then
        ffi.cdef(string.format([[
            %s {
                int is_message;
                int enable;
                int count;
                int element_size;
                int size;
                double sample_rate;
                long long enq_attempted; // # attempted enqueues
                int subscribed;
                unsigned char el_type[64];
                void *alloc_ptr;          // allocated buffer memory
                unsigned char *buf_head;  // points at header
                %s *buf;                  // points to first element
            };
        ]], streambuf_type, dtype))
        ctor_cache[streambuf_type] = ffi.metatype(streambuf_type, mt)
    end
    return ctor_cache[streambuf_type]
end

local m = {}

local dtypes = {
    sequential_float32 = 'float',
    sequential_float64 = 'double',
    message_float32 = 'message_float32',
    message_float64 = 'message_float64',
}

function m:new(data_type, sample_rate)
    assert(data_type)

    local dtype = dtypes[data_type]
    assert(dtype)

    -- Set the enq function for the metatable based on the data type
    if data_type:find('^message') then
        mt.enq_value = mt._enq_message_value
    else
        mt.enq_value = mt._enq_sequential_value
    end

    -- Instantiate the metatype
    local stream_buf = buf_ctor(data_type, dtype)()

    -- Initialize all of the stream buffer stuff
    if data_type:find('^message') then
        stream_buf.is_message = 1
    else
        stream_buf.is_message = 0
    end
    ffi.copy(stream_buf.el_type, dtype)
    stream_buf.element_size = ffi.sizeof(dtype)
    stream_buf.count = 0
    stream_buf.enable = 0
    stream_buf.sample_rate = rational:new(sample_rate):tonumber()
    assert(stream_buf.sample_rate)
    stream_buf.enq_attempted = 0LL
    stream_buf.subscribed = 0

    -- This stream buffer may never be used, so start with a small
    -- buffer size.  This can be increased if/when the stream is
    -- selected for use.
    stream_buf:expand(1)

    return stream_buf
end

return m

