require 'middleclass'
local ev = require'ev'
local http_server = require 'http_server'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'
local cjson = require 'cjson.safe'
local sb = require 'hbm_stream_buf'

ffi.cdef[[
typedef struct hbm_stream_packet {
    unsigned long       hdr;
    unsigned long       length;
    unsigned char       data[0];
} hbm_stream_packet_t;

typedef struct hbm_small_stream_packet {
    unsigned long       hdr;
    unsigned char       data[256];
} hbm_small_stream_packet_t;

enum {
    HDR_DATA = 0x10000000,
    HDR_METAINFO = 0x20000000
};

enum {
    METATYPE_JSON = 1,
};

enum {
    WS_BINARY = 2,
};
]]

local ss = class('stream_server')

local jsonrpc_version = '2.0'

local api_version_json = cjson.encode({
    method = 'apiVersion',
    params = {'1.0'},
})

-- Function used to send short metadata packets on a websocket
function ss:_send_short_metadata(ws, pkt, id, data)
--print(id, data)
    local length = #data
    pkt[0].hdr = C.HDR_METAINFO + ((length + 4) * 1048576) + id
    ffi.cast('unsigned int *', pkt[0].data)[0] = C.METATYPE_JSON
    ffi.copy(pkt[0].data + 4, data, length)
    ws:send(ffi.string(pkt, length + 8), C.WS_BINARY)
end

-- Function used to send long metadata packets on a websocket
function ss:_send_long_metadata(ws, pkt, id, data)
    local length = #data
    pkt.hdr = C.HDR_METAINFO + id
    pkt.length = length + 4
    ffi.cast('unsigned int *', pkt.data)[0] = C.METATYPE_JSON
    ffi.copy(pkt.data + 4, data, length)
    ws:send(ffi.string(pkt, length + 12), C.WS_BINARY)
end

-- Function used to send metadata packets on a websocket
function ss:send_metadata(ws_info, id, data)
    if #data <= 252 then
        return self:_send_short_metadata(ws_info.ws, ws_info.s_pkt, id, data)
    else
        local len = #data + ffi.sizeof('hbm_stream_packet_t') + 4
        if len > ws_info.pkt_size then
            ws_info.pkt_size = len
            ws_info.pkt = ffi.cast('hbm_stream_packet_t *',
                ffi.new('char[?]', len))
        end
        return self:_send_long_metadata(ws_info.ws, ws_info.pkt, id, data)
    end
end

-- Function used to send the initial set of metadata on a websocket when a
-- client first connects to the websocket
function ss:send_stream_metadata(ws_info)
    local pkt = ws_info.s_pkt
    local ws = ws_info.ws

    -- Send the API version metadata with stream id 0
    self:send_metadata(ws_info, 0, api_version_json)

    -- Send the "init" metadata with stream id 0
    self:send_metadata(ws_info, 0, cjson.encode({
            method = 'init',
            params = {
                streamId = ws_info.id,
                supported = {
                    alive = {timeout = 2},
--                    fill = true,
                },
                commandInterfaces = {
                    ['jsonrpc-http'] = self.http_server_info,
                },
            },
        }))

    -- Send the "available" channels metadata with stream id 0
    self:send_metadata(ws_info, 0, cjson.encode({
            method = 'available',
            params = self.channel_list,
        }))
end

-- Function called when a new websocket client connects
function ss:ws_connect(ws)
    -- Generate the "alive" metadata packet for this stream
    local alive_pkt = ffi.new('hbm_small_stream_packet_t[1]')
    local data = cjson.encode({
        method = 'alive',
    })
    local length = #data
    alive_pkt[0].hdr = C.HDR_METAINFO + ((length + 4) * 1048576)  -- Stream ID 0
    ffi.cast('unsigned int *', alive_pkt[0].data)[0] = C.METATYPE_JSON
    ffi.copy(alive_pkt[0].data + 4, data, length)
    local apkt = ffi.string(alive_pkt, length + 8)

    local id = string.format("WS_STREAM_%d", math.random(10000))
    local client_info = {
        ws = ws,
        subscribed_channels = {},
        pkt_size = ffi.sizeof('hbm_stream_packet_t'),
        pkt = ffi.cast('hbm_stream_packet_t *',
            ffi.new('char[?]', ffi.sizeof('hbm_stream_packet_t'))),
        s_pkt = ffi.new('hbm_small_stream_packet_t[1]'),
        id = id,
    }
    self.ws_client_list[#self.ws_client_list + 1] = client_info

    -- Timer used to send "alive" metadata messages on the stream
    local hbt = ev.Timer.new(function(l, t, e)
            ws:send(apkt, C.WS_BINARY)
        end, 1, 1)
    hbt.info = self.ws_client_list[#self.ws_client_list]
    self.ws_client_list[#self.ws_client_list].hbt = hbt

    -- If everything is OK, send the "stream" metadata back on the stream
    self:send_stream_metadata(client_info)

    -- Start the heartbeat timer
    hbt:start(self.loop)
end

-- Function called when a websocket client disconnects
function ss:ws_close(ws)
    -- Remove the client information from the client info list
    for i, ci in ipairs(self.ws_client_list) do
        if ci.ws == ws then
            -- Make sure we "unsubscribe" from each of the channels subscribed
            -- on this connection
            for _, sig_name in ipairs(ci.subscribed_channels) do
                local cinfo = self.channel_info[sig_name]

                if cinfo then
                    cinfo.buffer:unsubscribe()
                end
            end

            -- Stop the heartbeats for this client
            if ci.hbt then ci.hbt:stop(self.loop) end
            table.remove(self.ws_client_list, i)
            return
        end
    end
end

-- Function used to get a client list entry by stream id
function ss:get_ws_client(stream_id)
    for _, c in ipairs(self.ws_client_list) do
        if c.id == stream_id then
            return c
        end
    end
end

-- Function used to extract the stream id from the methods used to subscribe or
-- unsubscribe channels on a stream
function ss:get_stream_id(method, req_id)
    local stream_id = method:match('^(.+)%.unsubscribe$') or
        method:match('^(.+)%.subscribe$')
    if stream_id then
        local client = self:get_ws_client(stream_id)
        if not client then
            return nil, cjson.encode({
                jsonrpc = jsonrpc_version,
                error = {
                    message = 'Stream ID not found',
                    code = -32601,
                },
                id = req_id,
            })
        end
        return stream_id
    end
    return nil, cjson.encode({
        jsonrpc = jsonrpc_version,
        error = {
            message = 'Unable to parse stream id from method',
            code = -32603,
        },
        method = method,
        id = req_id,
    })
end

-- Mapping of CX23 test engine datatypes to HBM streaming protocol datatypes
local value_types = {
    float = 'real32',
    double = 'real64',
    message_float32 = 'real32',
    message_float64 = 'real64',

    -- NOTE:  The entries below are here for completeness but no native data
    -- source on a CX23 supplies data in any of these formats.
    int32 = 's32',
    uint32 = 'u32',
    int64 = 's64',
    uint64 = 'u64',
}

-- Function called to process an incoming subscribe request for a stream
-- FIXME:  Does subscribing to a channel that has already been subscribed to
-- result in an error response or success?  And, if success, is a second
-- instance provided in the stream or is the existing instance used?
function ss:subscribe(json)
    local req_id = json.id
    local stream_id, err = self:get_stream_id(json.method, req_id)
    if not stream_id then
        return err
    end
    local failed_list
    local client = self:get_ws_client(stream_id)
    local ws = client.ws
    local ws_send = ws.send
    local subscribed_channels = client.subscribed_channels
    local pkt = client.s_pkt
    for _, sig_name in ipairs(json.params) do
        -- Check to see if the channel has already been subscribe to
        local already_subscribed
        for _, s in ipairs(subscribed_channels or {}) do
            if s == sig_name then
                already_subscribed = true
                break
            end
        end

        local cinfo = self.channel_info[sig_name]
        if cinfo and not already_subscribed then
            -- Send the subscribe metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'subscribe',
                    params = {sig_name},
                }))

            -- Send the signal rate metadata for this channel on the websocket
            -- client.
            -- FIXME:  What should be done here for "asynchronous" channels?
            if cinfo.buffer.sample_rate > 0 then
                local seconds, fractions =
                    math.modf(1.0/(cinfo.buffer.sample_rate / 5))
                fractions = fractions * (2^32)
                self:send_metadata(client, cinfo.slot, cjson.encode({
                        method = 'signalRate',
                        params = {
                            samples = 5,
                            delta = {
                                type = 'ntp',
                                seconds = math.floor(seconds),
                                fraction = math.floor(fractions),
                            },
                        },
                    }))
            end

            -- Send the signal "data" metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'data',
                    params = {
                        -- FIXME:  There is also a "TB" pattern for timestamp
                        -- per block.  That would be used for 'message_raw'.
                        pattern =
                            cinfo.buffer.is_message ~= 0 and 'TV' or 'V',
                        endian = 'big',
                        valueType =
                            value_types[ffi.string(cinfo.buffer.el_type)],
                        time = {
                            type = 'ntp',
                            size = 8,
                        },
                    },
                }))

            -- Send the signal first sample timestamp metadata for this channel
            -- on the websocket client
            -- FIXME:  What should be done here for "asynchronous" channels?
            if cinfo.buffer.sample_rate > 0 then
                local fsts = 
                    tonumber(cinfo.buffer.enq_attempted - cinfo.buffer.count) *
                    1.0 / tonumber(cinfo.buffer.sample_rate)
                local s, f = math.modf(fsts)
                self:send_metadata(client, cinfo.slot, cjson.encode({
                        method = 'time',
                        params = {
                            stamp = {
                                type = 'ntp',
                                seconds = s,
                                fraction = math.floor((f * (2^32))),
                            },
                        },
                    }))
            end

            -- Tell the system to start sending data for this channel on the
            -- appropriate websocket client
            subscribed_channels[#subscribed_channels + 1] = sig_name

            -- Increment the count of the number of clients using the buffer
            cinfo.buffer:subscribe()
        else
            failed_list = failed_list or {}
            failed_list[#failed_list + 1] = sig_name
        end
    end
    if failed_list then
        return cjson.encode({
            jsonrpc = jsonrpc_version,
            error = {
                message = 'Invalid params',
                code = -32602,
                data = failed_list,
            },
            id = req_id,
        })
    end
    return cjson.encode({
        jsonrpc = jsonrpc_version,
        result = {},
        id = req_id,
    })
end

function ss:get_subscribed_channel_by_name(clist, sig_name)
    for i, name in ipairs(clist) do
        if name == sig_name then
            return clist[i], i
        end
    end
end

-- Function called to process an incoming unsubscribe request for a stream
function ss:unsubscribe(json)
    local req_id = json.id
    local stream_id, err = self:get_stream_id(json.method, req_id)
    if not stream_id then
        return err
    end
    local failed_list
    local client = self:get_ws_client(stream_id)
    local ws = client.ws
    local ws_send = ws.send
    local pkt = client.s_pkt
    local subscribed_channels = client.subscribed_channels
    for _, sig_name in ipairs(json.params) do
        local e, i =
            self:get_subscribed_channel_by_name(subscribed_channels, sig_name)
        if e and i then
            local cinfo = self.channel_info[sig_name]
            -- Send the unsubscribe metadata for this channel on the websocket
            -- client
            self:send_metadata(client, cinfo.slot, cjson.encode({
                    method = 'unsubscribe',
                }))

            -- Tell the system to start sending data for this channel on the
            -- appropriate websocket client
            table.remove(subscribed_channels, i)

            -- Disable and clear the buffer for this channel if there are no
            -- more clients using the channel
            cinfo.buffer:unsubscribe()
        else
            failed_list = failed_list or {}
            failed_list[#failed_list + 1] = sig_name
        end
    end
    if failed_list then
        return cjson.encode({
            jsonrpc = jsonrpc_version,
            error = {
                message = 'Invalid params',
                code = -32602,
                data = failed_list,
            },
            id = req_id,
        })
    end
    return cjson.encode({
        jsonrpc = jsonrpc_version,
        result = {},
        id = req_id,
    })
end

-- Handler for HTTP requests coming in from the stream control server
function ss.http_handler(loop, w, http)
    local self = w.me
    if not http.uri:match('^' .. self.http_server_info.httpPath) then
        w.streambuf:write("HTTP/1.1 404 NOT FOUND\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif http.method ~= self.http_server_info.httpMethod then
        w.streambuf:write("HTTP/1.1 400 INVALID METHOD\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not http.headers or not http.headers.content_type or
        not http.headers.content_type:find('application/json') then
        w.streambuf:write("HTTP/1.1 400 INVALID CONTENT TYPE\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not http.content then
        w.streambuf:write("HTTP/1.1 400 EMPTY CONTENT\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local cj = cjson.decode(http.content)
    if not cj then
        w.streambuf:write("HTTP/1.1 400 UNABLE TO DECODE JSON\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    elseif not cj.jsonrpc or cj.jsonrpc ~= jsonrpc_version then
        w.streambuf:write("HTTP/1.1 400 INVALID RPC VERSION\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local method = cj.method:match('^.+%.(.+)$')
    if not method or not self[method] then
        w.streambuf:write("HTTP/1.1 400 UNRECOGNIZED REQUEST\r\n\r\n")
        w.server.http_write_shutdown(loop, w)
        return
    end
    local json = self[method](self, cj)
    w.streambuf:write(
        "HTTP/1.1 200 OK\r\n" ..
        "Content-Length: " .. #json + 2 .. "\r\n" ..
        "Content-Type:  application/json\r\n\r\n" ..
        json .. "\r\n")
    w.server.http_write_shutdown(loop, w)
end

-- Function used to add a channel to the servers list of channels available to
-- be served on the websocket connections.
function ss:add_channel(id, buffer)
    if not self.channel_info[id] then
        self.channel_list[#self.channel_list + 1] = id
        self.channel_info[id] = {
            buffer = buffer,
            slot = #self.channel_list,
        }
    end
end

-- Function used to send data from channels to each of the subscribed channels
-- on each of the websocket connections.
function ss:process_streams()
    local sid_list = {}

    -- Iterate through all of the clients
    for cid, cinfo in ipairs(self.ws_client_list) do
        -- Iterate through all of the subscribed channels for the client
        -- FIXME:  Maybe make this a function to help the compiler out a bit
        for _, sig_name in ipairs(cinfo.subscribed_channels) do
            -- Send all available data for this channel on the current client
            -- websocket connection.
            local chan_info = self.channel_info[sig_name]
            local pkt, len = chan_info.buffer:get_blk()
            if pkt then
                local sid = chan_info.slot
                pkt = ffi.cast('hbm_stream_packet_t *', pkt)
                pkt[0].hdr = C.HDR_DATA + sid
                cinfo.ws:send(ffi.string(pkt, len), C.WS_BINARY)
                sid_list[#sid_list + 1] = sig_name
            end
        end
    end

    -- Reset all that were sent
    for _, sid in ipairs(sid_list) do
        self.channel_info[sid].buffer:clear()
    end
end

-- Function used to initialize streaming service
function ss:initialize(loop, opts)
    -- Seed the random number generator
    math.randomseed(os.time())

    opts = opts or {}
    opts.rate = opts.rate or .05

    -- Save off stuff
    self.loop = loop
    self.ws_client_list = {}

    self.channel_list = {}
    self.channel_info = {}

    -- Create an event based HTTP server for stream control
    self.http_server = http_server:new{
        ev_loop = loop,
        handler = self.http_handler,
        add_to_watchers = {
            me = self,
        }
    }

    -- FIXME?:  Should this be limited to the "external" interface only?
    self.http_w = self.http_server:listen({address = '*', port = opts.cport })
    self.http_port = self.http_w.port

    -- Save HTTP info needed when websocket clients connect
    self.http_server_info = {
        httpPath = '/stream_control',
        httpMethod = 'POST',
        httpVersion = '1.1',
        port = self.http_w.port,
    }

    self.ws_port = opts.port

    -- create a event based websocket server and start listening
    self.ws_server = require'websocket.server_ev'.listen{
        -- FIXME:  Ideally we want to use port 0 (zero) so the the listen will
        -- occur on a random port.  However, there is no way to retrieve the
        -- port that gets used from the websocket object.
        port = self.ws_port or 0,
-- FIXME:  IPv4 addresses don't work
--        interface = '127.0.0.1',

        -- the protocols field holds
        --   key: protocol name
        --   value: callback on new connection
        protocols = {
            -- this callback is called, whenever a new client connects.
            -- ws is a new websocket instance
            streaming = function(ws)
                    self:ws_connect(ws)

                    -- This is called when the connection closes
                    ws:on_close(function()
                            self:ws_close(ws)
                        end)
                end
        }
    }

    local _ip, _port, _fam = self.ws_server:sock():getsockname()
    self.ws_port = _port

    -- Create a timer used to move sample data from the input buffers to the
    -- stream sockets
    self.data_writer = ev.Timer.new(function(l, t, e)
            self:process_streams()
        end, opts.rate, opts.rate):start(loop)
end

-- Terminate the websocket stuff
function ss:terminate(log, close_complete_func)
    -- For every client connected, close the connection and leave.  NOTICE
    -- this is intend for use just before using process termination
--[[
    if log then
        log:log(log.NOTICE, "Terminating HBM stream server")
    end
]]

    -- If there are no clients just return.  Call the callback function if one
    -- was given.
    if not self.ws_client_list or
        (self.ws_client_list and #self.ws_client_list == 0) then
        if close_complete_func then
            close_complete_func()
        end
        return
    end

    -- Close all of the client connections.  After the last client acknowledges
    -- the connection is closed, the callback is called (if given).
    for i, ci in ipairs(self.ws_client_list) do
        if i == #self.ws_client_list and close_complete_func then
            ci.ws:on_close(function()
                    self.ws_client_list[i] = nil
                    local num = 0
                    for _, c in pairs(self.ws_client_list) do
                        num = num + 1
                    end
                    if num == 0 then close_complete_func() end
                end)
        else
            ci.ws:on_close(function() end)
        end
        ci.ws:close(nil, 'Server terminating', .01)
    end
end

return ss
