require 'strict'
require 'middleclass'
local netdb = require 'network_db'
local net = require 'net'
local StreamBuffer = net.StreamBuffer
local cjson = require 'cjson'
local http_parser = require 'http_parser'
local ffi = require 'ffi'

-- A class library used to set the LED's on CX23 devices to specific states
local led_lib = class('led_lib')

function led_lib:_get_server(remote)
    if not remote then
        self.host = self.netdb:get_inside_network_hostname()
    else
        self.host = remote
    end
    remote = 'power.' .. self.host
    local host, port, ip, canon = self.netdb:resolve_service(remote)
    return ip, port
end

function led_lib:initialize(loop)
    self.loop = loop
    self.netdb = netdb:new()
    self.modes = {
        off = true,
        on = true,
        one_hz = true,
        two_hz = true,
        four_hz = true,
        eight_hz = true
    }
    self.colors = {
        red = true,
        green = true,
        blue = true,
        yellow = true
    }
end

function led_lib:_send_led_request(ip, port, mode, color, callback)
    local s = net.socket{ip = ip, port = port}
    local ok, socket = pcall(s.connect, s)
    if not ok then return end
    local sbuf = StreamBuffer:new(socket)
    local callback = callback
    sbuf.read_cb = function(loop)
        local avail, buf = sbuf:read()
        if avail then
            if avail == 0 then
                sbuf:close()
                return 0
            end
            local response, endpos =
                http_parser.response:match(ffi.string(buf, avail))
            if response then
                if response.response_code and
                    response.response_code == '200' then
                    local rh = response.headers
                    if rh and rh.content_length then
                        local cl = tonumber(rh.content_length)
                        if cl + endpos > avail + 1 then return 0 end
                        if rh.content_type == 'application/json' then
                            local json = ffi.string(buf + endpos - 1, cl)
                            if callback then callback(json) end
                            sbuf:close()
                            return avail
                        else
                            if callback then
                                callback(cjson.encode{
                                        result = 'error',
                                        msg = string.format(
                                            'Unexpected content type: %s',
                                            rh.content_type),
                                    })
                            end
                            sbuf:close()
                            return 0
                        end
                    else
                        if callback then
                            callback(cjson.encode{
                                    result = 'error',
                                    msg = 'Response with no content length',
                                })
                        end
                        sbuf:close()
                        return 0
                    end
                else
                    if callback then
                        callback(cjson.encode{
                                result = 'error',
                                msg = string.format("HTTP Response:  %s",
                                    response.response_code or "NONE"),
                            })
                    end
                    sbuf:close()
                    return 0
                end
            end
        end
        return 0
    end
    sbuf:start(self.loop)
    local json = cjson.encode{
        request = 'set_state',
        connector = string.format("%s_led_state.status@power.%s",
            color, self.host),
        params = {mode:lower()}
    }
    local req = string.format('POST /control HTTP/1.0\r\n' ..
        'Content-Length: %d\r\n' ..
        'Content-Type:  application/json\r\n\r\n' ..
        '%s\r\n',
        #json + 2, json)
    sbuf:write(req)
end

function led_lib:SetLED(mode, color, callback, remote)
    if not self.modes[mode:lower()] or not self.colors[color:lower()] then
        return
    end
    mode = mode:lower()
    color = color:lower()
    local ip, port = self:_get_server(remote)
    if ip and port then
        self:_send_led_request(ip, port, mode, color, callback)
    end
end

function led_lib:SetRedLED(mode, callback, remote)
    return self:SetLED(mode, 'red', callback, remote)
end

function led_lib:SetGreenLED(mode, callback, remote)
    return self:SetLED(mode, 'green', callback, remote)
end

function led_lib:SetBlueLED(mode, callback, remote)
    return self:SetLed(mode, 'blue', callback, remote)
end

function led_lib:SetYellowLED(mode, callback, remote)
    return self:SetLed(mode, 'yellow', callback, remote)
end

return led_lib
