-- listener()
--
-- Watch a channel or channels for activity.  All non-command inbound messages
-- are processed and individual structures are passed to helper functions
-- which are identified by functions attached to the specific channel object.
-- The member function 'element_action_func' will be called for each
-- structure received.  The parameters are the channel object, the name of the
-- ring buffer that the structure was received on, and the structure itself.
-- The member function 'heartbeat_func' will be called for each heartbeat
-- detected.  The parameters of the call are the channel object, the name of the
-- ring buffer that the heartbeat was detected on, and the timestamp of the
-- heartbeat.
--
-- The function will exit once a timeout is hit or the sample_count provided
-- is met or exceeded by the sum of the collected messages from all channels,
-- not including heartbeats.  Channel_objs must be a table list of channels,
-- timeout is a number in seconds, sample_count is an integer or nil if the
-- desired action is to exit on the first msg without a buffer key, i.e. a
-- command status.
--
-- Debugging is possible on a per-channel basis by simply defining a member
-- of the channel object called 'listener_debug' and setting it to true.
--


local function display_message(channel_obj, msg)
    io.write('MSG('..channel_obj.devname..','..msg.cmd..','..msg.status)
    for k,v in pairs(msg) do
        if k ~= 'cmd' and k ~= 'status' then
            if k == 'buffer' then
                io.write(',buffer{')
                local bytes_total = msg.bufdata_bytes
                local element_count = msg.element_count
                local heartbeat = msg.heartbeat
                if heartbeat then heartbeat = 'HB' else heartbeat = '-' end
                io.write(bytes_total..','..element_count..','..heartbeat..'}')
            else
                io.write(','..k..'='..tostring(v))
            end
        end
    end
    print(')')
end

local function listener(channel_objs, timeout, sample_count)
    local u = require 'unix'
    local ev = require 'ev'
    local msg
    local last_element = nil
    local buffer_detected = false
    local sample_count_hit = false
    local loop = ev.Loop.new()
    local running_total = 0
    local null_func = function() end
    local element_action_func = nil
    local heartbeat_func = nil
    for i=1,#channel_objs do
        if channel_objs[i].fd == nil then
            channel_objs[i].fd = u.open(channel_objs[i].devname, u.O_RDONLY)
        end
        if not channel_objs[i].sample_count then
            channel_objs[i].sample_count = 0
        end
        local element_action_func = channel_objs[i].element_action_func
        local heartbeat_func = channel_objs[i].heartbeat_func
        if element_action_func == nil then element_action_func = null_func end
        if heartbeat_func == nil then heartbeat_func = null_func end
        running_total = running_total + channel_objs[i].sample_count
        local process_buffer = function(channel_obj, msg)
            local name = channel_obj.ring_buffer_table[msg.pa]
            local element_count = msg.element_count
            local heartbeat = msg.heartbeat
            local element = nil
            for j=1,element_count do
                if heartbeat and j == element_count then
                    local ts = msg.bufdata[j - 1].ts
                    heartbeat_func(channel_obj, name, ts)
                    element_count = element_count - 1
                    break
                else
                    element = msg.bufdata[j - 1]
                    channel_obj.sample_count = channel_obj.sample_count + 1
                    element_action_func(channel_obj, name, element)
                end
            end
            return element, element_count
        end
        ev.IO.new(
            function(loop, watcher)
                local fake_proxy = {}
                local update_msg
                function fake_proxy.buffer(proxy, t)
                    update_msg = t
                end
                u.read(channel_objs[i].fd, 1)
                msg = channel_objs[i]:poll(fake_proxy)
                if msg and type(msg) ~= 'table' then msg = update_msg end
                if channel_objs[i].listener_debug then
                    display_message(channel_objs[i], msg)
                end
                if msg.bufdata then
                    local last_element, count = process_buffer(channel_objs[i], msg)
                    running_total = running_total + count
                    buffer_detected = true
                    if sample_count and running_total >= sample_count then
                        sample_count_hit = true
                        watcher:stop(loop)
                        loop:unloop()
                    end
                else if msg.msg then
                    -- async error
                    print('Error:'..msg.msg)
                else
                    -- command
                    watcher:stop(loop)
                    loop:unloop()
                end end
            end,
            channel_objs[i].fd, ev.READ):start(loop)
    end
    ev.Timer.new(
        function(loop, timer)
            loop:unloop()
        end,
        timeout):start(loop)
    loop:loop()
    if sample_count then
        if sample_count_hit then
            return last_element
        else
            return nil
        end
    else
        return msg
    end
end

return listener
