require 'strict'

local __ = require '__'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'

local config = require 'system_config'

--local debug_msgs_enabled = true
local debug_msgs_enabled = config.debug_msgs or nil

local C = ffi.C

ffi.cdef[[

void openlog(const char *ident, int option, int facility);
void syslog(int priority, const char *format, ...);
void closelog(void);
void vsyslog(int priority, const char *format, va_list ap);

enum {
    /* Option flags for openlog. */
    LOG_PID       = 0x01,   /* log the pid with each message */
    LOG_CONS      = 0x02,   /* log on the console if errors in sending */
    LOG_NDELAY    = 0x08,   /* don't delay open */
    LOG_PERROR    = 0x20,   /* log to stderr as well */

    /* priorities (these are ordered) */
    LOG_EMERG     = 0,      /* system is unusable */
    LOG_ALERT     = 1,      /* action must be taken immediately */
    LOG_CRIT      = 2,      /* critical conditions */
    LOG_ERR       = 3,      /* error conditions */
    LOG_WARNING   = 4,      /* warning conditions */
    LOG_NOTICE    = 5,      /* normal but significant condition */
    LOG_INFO      = 6,      /* informational */
    LOG_DEBUG     = 7,      /* debug-level messages */

    /* facility codes */
    LOG_KERN      = (0<<3), /* kernel messages */
    LOG_USER      = (1<<3), /* random user-level messages */
    LOG_MAIL      = (2<<3), /* mail system */
    LOG_DAEMON    = (3<<3), /* system daemons */
    LOG_AUTH      = (4<<3), /* security/authorization messages */
    LOG_SYSLOG    = (5<<3), /* messages generated internally by syslogd */
    LOG_LPR       = (6<<3), /* line printer subsystem */
    LOG_NEWS      = (7<<3), /* network news subsystem */
    LOG_UUCP      = (8<<3), /* UUCP subsystem */
    LOG_CRON      = (9<<3), /* clock daemon */
    LOG_AUTHPRIV  = (10<<3),/* security/authorization messages (private) */
    LOG_FTP       = (11<<3),/* ftp daemon */

    /* other codes through 15 reserved for system use */
    LOG_LOCAL0    = (16<<3),/* reserved for local use */
    LOG_LOCAL1    = (17<<3),/* reserved for local use */
    LOG_LOCAL2    = (18<<3),/* reserved for local use */
    LOG_LOCAL3    = (19<<3),/* reserved for local use */
    LOG_LOCAL4    = (20<<3),/* reserved for local use */
    LOG_LOCAL5    = (21<<3),/* reserved for local use */
    LOG_LOCAL6    = (22<<3),/* reserved for local use */
    LOG_LOCAL7    = (23<<3) /* reserved for local use */
};

    char *strerror(int errnum);

    int isatty(int);
]]

local mt = {}

mt.log_fh = false
mt.stdout_log = false
mt.enabled = true

function mt:open(ident, facility, option)
    option = option or C.LOG_PID
    C.openlog(ident, option, facility)
end

function mt:log(level, fmt, ...) -- KLUDGE add timestamp???
    if level == self.DEBUG and debug_msgs_enabled ~= true or
        not self.enabled then
        return
    end
    local msg = type(fmt) == 'table' and fmt or {fmt, ...}
    local msg_json = cjson.encode(msg)
    local formatted = __:format(msg)
    --C.syslog(level, '%s', msg_json) -- Waiting for downstream syslog support
    C.syslog(level, '%s', formatted)
    if self.stdout_log == false then
        self.stdout_log = C.isatty(1)
    end
    if self.stdout_log ~= 0 then
        io.stdout:write(formatted)
        io.stdout:write('\n')
        io.stdout:flush()
    end
    if self.log_fh then -- KLUDGE formatted_fh, json_fh ???
        self.log_fh:write(formatted)
        self.log_fh:write('\n')
        self.log_fh:flush()
    end
    --io.stderr:write(msg_json)
    return msg
end

function mt:notice(fmt, ...)
    return self:log(self.NOTICE, fmt, ...)
end

function mt:warn(fmt, ...)
    return self:log(self.WARNING, fmt, ...)
end

function mt:debug(fmt, ...)
    return self:log(self.DEBUG, fmt, ...)
end

function mt:error(fmt, ...)
    if fmt then
        self:log(self.ERR, fmt, ...)
    end
    local msg = self:log(self.ERR, 'Error: {0}\n', debug.traceback())
    -- error(__:format(msg))
    error('Fatal error', 3)
end

function mt:strerror(errnum)
    errnum = errnum or ffi.errno()
    return ffi.string(C.strerror(errnum))
end

function mt:enable(enable)
    self.enabled = enable
end

local m = {}

setmetatable(m, {
    __index = function(t, key)
        local v = rawget(t, key) or rawget(mt, key)
        if v ~= nil then
            return v
        end
        return C['LOG_' .. key]
    end,
    __call = function(t, level, fmt, ...) -- KLUDGE add timestamp???
        t:log(level, fmt, ...)
    end,
})

return m
