#!/usr/bin/lua

--! The master daemon.
--!
--! \file

--! Our D-Bus namespace.
--? namespace com.somat.armada

local ev = require 'ev'
local dbus = require 'dbus'
local sqlite3 = require 'lsqlite3'
local unix = require 'unix'
local timestamps = require 'timestamps'
local ffi = require 'ffi'
local C = ffi.C
local loop = ev.Loop.default
local ledlib = require('led_lib'):new(loop)
local cjson = require 'cjson.safe'
local sys_notify = require 'sys_notify'

ffi.cdef[[
typedef int     pid_t;

int kill(pid_t, int sig);
int mkdir(const char *path, int);
]]

C.mkdir('/tmp/proxy', 511)

local config = require 'system_config'

-- Log entries to keep
local keep = 100000

ev.Signal.new(function() end, unix.SIGPIPE):start(loop) -- ignore SIGPIPE

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()
assert(bus:bus_request_name('com.somat.armada.Logger'))

--! A D-Bus-accessible class for Armada logging.
--!
--! This class is exported over D-Bus from bus name
--! <tt>com.somat.armada.Logger</tt> and object path
--! <tt>/com/somat/aramda/Logger</tt>.
--? class com.somat.armada.Logger
local Logger = { }
--? alias Logger com.somat.armada.Logger
Logger.__index = Logger
local Logger_iface = dbus.Interface.new('com.somat.armada.Logger')

function Logger.new(db) --? private
    local self = setmetatable({ }, Logger)
    self.db = assert(sqlite3.open(db))
    self.insert =
        assert(self.db:prepare[[
            INSERT INTO log (time, component, level, message) 
            VALUES (?, ?, ?, ?)
        ]])
    return self
end

function Logger:maybe_open_transaction() --? private
    if not self.transaction_open then
        assert(self.db:exec('BEGIN'))
        local max_id
        for r in self.db:rows('SELECT MAX(id) FROM log') do
            max_id = tonumber(r[1])
        end
        if max_id then
            self.db:exec('DELETE FROM log WHERE id < ' ..
                         math.max(0, max_id - keep))
        end
        ev.Timer.new(function(loop, timer, revents)
                         assert(self.db:exec('COMMIT'))
                         self.transaction_open = false
                     end, 1):start(loop)
        self.transaction_open = true
    end
end

--! Handles setting the RED LED
-- KLUDGE:  Errors logged before the power daemon starts up result in the RED
-- LED NOT being turned on.  This basically waits for the system to be up
-- and if an error message was logged during boot time before the power daemon
-- was up and running, will make the RED LED turn on...yikes this is ugly!
-- Fifteen seconds was picked by experimentation.  Ten seconds was too short
-- and Thirty seconds seemed excessively long.
local error_logged
ev.Timer.new(function(l, t, e)
        if error_logged then
            pcall(ledlib.SetLED, ledlib, "on", "red")
            error_logged = false
        end
    end, 15, .1):start(loop)

--! Logs a message.
function Logger:Log(component, level, message)
    local secs, nsecs = unix.clock_gettime()
    local stamp = timestamps.epoch_to_iso(secs + nsecs * 10e-9)
    self:maybe_open_transaction()
    self.insert:reset()
    self.insert:bind_values(stamp, component, level, message)
    repeat until self.insert:step() == sqlite3.DONE
    if level == 'err' and not component:find('kern') then
        error_logged = true
    end
end
Logger_iface:add_method{'Log', 'sss', '', no_reply = true}

local logger = Logger.new('/var/lib/armada/log.db')
bus:export_object(logger, '/com/somat/armada/Logger', { Logger_iface })

local syslog_priorities = {
    [0] = 'emerg',
    [1] = 'alert',
    [2] = 'crit',
    [3] = 'err',
    [4] = 'warning',
    [5] = 'notice',
    [6] = 'info',
    [7] = 'debug'
}

local syslog_facilities = {
    [0] = 'kern',
    [1] = 'user',
    [2] = 'mail',
    [3] = 'daemon',
    [4] = 'auth',
    [5] = 'syslog',
    [6] = 'lpr',
    [7] = 'news',
    [8] = 'uucp',
    [9] = 'cron',
    [10] = 'authpriv',
    [11] = 'ftp',
    [12] = 'reserved0',
    [13] = 'reserved1',
    [14] = 'reserved2',
    [15] = 'reserved3',
    [16] = 'local0',
    [17] = 'local1',
    [18] = 'local2',
    [19] = 'local3',
    [20] = 'local4',
    [21] = 'local5',
    [22] = 'local6',
    [23] = 'local7',
}

local function syslog(msg, default, trim)
    local code, submsg = msg:match("^<(%d+)>(.*)")
    code = code or default
    msg = submsg or msg
    if trim then
        msg = msg:sub(trim)
    end
    local facility = math.floor(code / 8)
    local priority = code % 8

    -- Promote kernel unhandled signal messages (info) to error
    if (facility == 0 or facility == 'kern') and
        (priority == 6 or priority == 'info') and
        msg:find('unhandled signal') then
        priority = 3
        facility = 1
    end

    logger:Log('syslog.' .. (syslog_facilities[facility] or facility),
               syslog_priorities[priority] or priority,
               msg)
end

local klog_fd = assert(unix.open('/proc/kmsg', unix.O_RDONLY +
                                 unix.O_CLOEXEC + unix.O_NONBLOCK))
local klog_remaining = ""
ev.IO.new(function(loop, io, revents)
              local buf = unix.read(klog_fd, 8192)
              klog_remaining = klog_remaining .. buf
              repeat
                  local nl = klog_remaining:find('\n')
                  if nl then
                      syslog(klog_remaining:sub(1, nl - 1), 5) -- notice
                      klog_remaining = klog_remaining:sub(nl + 1)
                  end
              until not nl
          end, klog_fd, ev.READ):start(loop)

unix.unlink('/dev/log')
local log_fd = assert(unix.socket(unix.AF_UNIX, unix.SOCK_DGRAM +
                                  unix.SOCK_NONBLOCK + unix.SOCK_CLOEXEC, 0))
assert(unix.bind(log_fd, unix.new_sockaddr_un('/dev/log')))
assert(unix.chmod('/dev/log', 0x1b6)) -- KLUDGE 0666
ev.IO.new(function(loop, io, revents)
              local buf = unix.recv(log_fd, 8192)
              syslog(buf, 8 + 5, 17)
          end, log_fd, ev.READ):start(loop)

-- UDP logging for channel boards, eventually:
local res =
    assert(unix.getaddrinfo(nil, 12345,
                            unix.AI_PASSIVE, unix.AF_INET6, unix.SOCK_DGRAM))
local r = res[1]
local udp_fd = assert(unix.socket(r.family, r.socktype, r.protocol))
assert(unix.setsockopt_int(udp_fd, unix.SOL_SOCKET, unix.SO_REUSEADDR, 1))
assert(unix.bind(udp_fd, r.addr))
ev.IO.new(function(loop, io, revents)
              local buf, addr = assert(unix.recvfrom(udp_fd, 8192))
              local ip, port =
                  assert(unix.getnameinfo(addr, unix.NI_NUMERICHOST +
                                          unix.NI_NUMERICSERV))
              local pri, msg = buf:match("^<(%d+)>(.*)")
              pri = pri and pri % 8 or 5
              msg = msg or buf
              logger:Log('udp.[' .. ip .. ']', syslog_priorities[pri], msg)
          end, udp_fd, ev.READ):start(loop)

assert(bus:bus_request_name('com.somat.armada.Master'))

--! A D-Bus-accessible class for Armada system services.
--!
--! This class is exported over D-Bus from bus name
--! <tt>com.somat.armada.Master</tt> and object path
--! <tt>/com/somat/aramda/Master</tt>.
--? class com.somat.armada.Master
Master = { }
--? alias Master com.somat.armada.Master
Master.__index = Master
local Master_iface = dbus.Interface.new('com.somat.armada.Master')

local critical_process_list = {}

function Master:RegisterCriticalProcess(name, pid)
    if not critical_process_list[pid] then
        critical_process_list[pid] = {
            name = name,
            pid = ffi.new('pid_t', tonumber(pid))
        }
    end
end
Master_iface:add_method{'RegisterCriticalProcess', 'ss', ''}

function Master:UnregisterCriticalProcess(pid)
    critical_process_list[pid] = nil
end
Master_iface:add_method{'UnregisterCriticalProcess', 's', ''}

Master_iface:add_signal{'SystemCriticalProcessDeath', 's'}


-- If we are still running after 1 minute, remove the reboot limiter file
ev.Timer.new(function(l, t, e)
        os.execute('rm -f /var/etc/armada/reboots.lua 2>/dev/null')
        t:stop(l)
    end, 60, 0):start(ev.Loop.default)

-- Function used to send system wide notifications
local function do_notification(name, msg, do_reboot)
    name = string.match(name, '.+[\\/](.+)$') or name
    local sn = sys_notify:new(string.format(
        "System Critical Process '%s' Died", name))
    if not sn then return end
    sn:critical_error(msg)
    if do_reboot then
        ev.Timer.new(function(l, t, e)
                os.execute('sync;sync;/usr/bin/reboot error')
            end, .25, 0):start(ev.Loop.default)
    end
end

-- Function used to determine if there are any locally running tests
local function is_test_running()
    local ndb = sqlite3.open('/var/lib/armada/network.db')
    if ndb then
        for row in ndb:rows([[
            SELECT rdata FROM unified
            WHERE owner = '_restart-armada-tests'
            AND local_source != ''
        ]]) do
            return true
        end
    end
end


-- Timer used to check to make sure processes that have registered themselves
-- as system critical processes are running.  This is done once per second.
local crit_watcher = ev.Timer.new(function(l, t, e)
        local respawn_list = {}
        for k, cp in pairs(critical_process_list) do
            if C.kill(cp.pid, 0) < 0 then
                logger:Log('master', 'err',
                    string.format("System critical process %s died", cp.name))
                if cp.respawn then
                    table.insert(respawn_list, cp)
                end

                -- If a test is running when a critical process dies, just
                -- reboot the system and let the interrupted test restart
                -- mechanism do it's thing.
                if is_test_running() then
                    t:stop(l)
                    do_notification(cp.name, "The critical process died " ..
                        "while a test was running.  The system will reboot " ..
                        "and if interrupted test restarts are enabled and " ..
                        "the test has not reached the interrupted test " ..
                        "restart limit, the test will be restarted.  After " ..
                        "reboot you will be given the opportunity to cancel " ..
                        "test restart if it is active.", true)
                    return
                end

                if ledlib then
                    -- If the ptp daemon died, make sure the blue LED is off
                    if cp.name:find('ptp') then
                        pcall(ledlib.SetLED, ledlib, "off", "blue")
                    end
                    pcall(ledlib.SetLED, ledlib, "eight_hz", "red")
                end

                critical_process_list[k] = nil
                if not config.override_critical_process_death_reboot then
                    local reboot_count = 0
                    local rb = loadfile('/var/etc/armada/reboots.lua')
                    if rb then
                        reboot_count = rb() or 0
                    end

                    -- If we have rebooted less than 5 times, then do it again
                    if reboot_count < config.reboot_retry_limit then
                        reboot_count = reboot_count + 1
                        os.execute(string.format(
                            'echo "return %d" > /var/etc/armada/reboots.lua',
                            reboot_count))

                        -- Let the GUI know (if it is attached)
                        do_notification(cp.name, string.format(
                            "System will attempt automatic reboot %d of %d.",
                            reboot_count, config.reboot_retry_limit), true)
                    else
                        -- Cancel all interrupted test restarts
                        os.execute('killall -SIGUSR1 restart-tests')

                        -- Let the GUI know (if it is attached) but dont reboot
                        do_notification(cp.name, string.format(
                            "The system appears to be in an infinite reboot " ..
                            "cycle.  No more attempts to reboot will be " ..
                            "attempted because more than %d reboot attempts " ..
                            "have failed.  This situation can be caused " ..
                            "by an interrupted test restart that is failing " ..
                            "or by a problem with the CX23 device itself.  " ..
                            "Any active interrupted test restarts will not " ..
                            "be performed at this time because the system " ..
                            "may be in an unstable state.  Please call HBM " ..
                            "customer service for assistance.",
                            config.reboot_retry_limit), false)
                        os.execute('rm /var/etc/armada/reboots.lua 2>/dev/null')
                    end
                else
                    -- Let the GUI know (if it is attached) but dont reboot
                    do_notification(cp.name,
                        "Automatic reboot upon detection of critical " ..
                        "process death is disabled.  The system is likely " ..
                        "in an unstable state.  Reboot is highly " ..
                        "recommended", false)
                end
            end
        end
        if #respawn_list > 0 then
            for _, cp in ipairs(respawn_list) do
                Master:Spawn(cp.name, cp.args, false, true)
            end
        end
    end, 1, 1)
crit_watcher:start(ev.Loop.default)

-- This signal is sent when the system shutdown sequence is initiated.  When
-- that is happening it is normal for critical processes to get killed off and
-- we don't want a pile of things ending up in the log or as notifications when
-- those things are normal.
ev.Signal.new(function(l, s, e)
        crit_watcher:stop(l)
        crit_watcher:clear_pending(l)
        s:stop(l)
    end, unix.SIGUSR1):start(ev.Loop.default)


--! Spawns a process, logging its stdout and stderr.
--!
--! The process will be owned by the \c master process.
--!
--! \param path The full path to the executable to spawn.
--! \param args The arguments for exec, including the executable name.
--! \param is_not_critical Does not impact LED state
--! \param respawn Spawn the executable again if it dies
function Master:Spawn(path, args, is_not_critical, respawn)
    local stdout_read, stdout_write = assert(unix.pipe())
    local stderr_read, stderr_write = assert(unix.pipe())
    local stderr_level = 'err'
    local stdout_level = 'info'
    local _args = {}
    for i, v in ipairs(args) do
        local m = v:match('std.*_level:(.*)')
        if m then
            if v:find('stderr_level') then stderr_level = m
            elseif v:find('stdout_level') then stdout_level = m
            end
        else
            _args[#_args + 1] = v
        end
    end
    args = _args
    local pid = assert(unix.fork())
    if pid == 0 then
        loop:fork()
        local dev_null = unix.open('/dev/null', unix.O_RDONLY)
        unix.close(stdout_read)
        unix.close(stderr_read)
        unix.dup2(dev_null, 0)
        unix.dup2(stdout_write, 1)
        unix.dup2(stderr_write, 2)
        unix.close(dev_null)
        unix.close(stdout_write)
        unix.close(stderr_write)
        assert(unix.execv(path, args))
    else
        logger:Log('master', 'info',
                   'Spawned ' .. table.concat(args, ' ') .. ' as pid ' .. pid)
        if not is_not_critical then
            critical_process_list[pid] = {
                name = path,
                pid = ffi.new('pid_t', tonumber(pid)),
                args = args,
                respawn = respawn,
            }
        end
        unix.close(stdout_write)
        unix.close(stderr_write)
        unix.setfd(stdout_read, unix.FD_CLOEXEC)
        unix.setfl(stdout_read, unix.O_NONBLOCK)
        unix.setfd(stderr_read, unix.FD_CLOEXEC)
        unix.setfl(stderr_read, unix.O_NONBLOCK)
        local basename = path:gsub("^.*/", "")
        local hdr = string.format("%s[%d]", basename, pid)
        ev.IO.new(function(loop, io, revents)
                      local buf = unix.read(stdout_read, 8192)
                      if buf == "" then
                          unix.close(stdout_read)
                          io:stop(loop)
                      elseif buf and stdout_level ~= 'none' then
                          logger:Log(hdr, stdout_level, buf)
                      end
                  end, stdout_read, ev.READ):start(loop)
        ev.IO.new(function(loop, io, revents)
                      local buf = unix.read(stderr_read, 8192)
                      if buf == "" then
                          unix.close(stderr_read)
                          io:stop(loop)
                      elseif buf and stderr_level ~= 'none' then
                          logger:Log(hdr, stderr_level, buf)
                      end
                  end, stderr_read, ev.READ):start(loop)
        return pid
    end
end
Master_iface:add_method{'Spawn', 'sas', 'd'}

function Master:SpawnNonCritical(path, args)
    return self:Spawn(path, args, true)
end
Master_iface:add_method{'SpawnNonCritical', 'sas', 'd'}

Master_iface:add_signal{'Heartbeat', 's'}

function Master:SignalHeartbeat(from)
    self:Heartbeat(from)
end
Master_iface:add_method{'SignalHeartbeat', 's', ''}

local master = setmetatable({ }, Master)
bus:export_object(master, '/com/somat/armada/Master', { Master_iface })

local spawned_stdout_mode = 'none'
--local spawned_stdout_mode = 'debug'

-- Get all of the network interfaces configured.
master:SpawnNonCritical("/usr/bin/network-config",
    { "network-config", "stdout_level:" .. spawned_stdout_mode})
-- When network configuration was part of the network deamon, there was a 5
-- second delay between when the interface were set up and when the network
-- daemon proper stuff was started.  I don't know why but that 5 second delay
-- seems to be required so I've added it here.
os.execute('sleep 5')

master:Spawn("/usr/bin/network",
    { "network", "stdout_level:" .. spawned_stdout_mode })
master:Spawn("/usr/bin/timing",
    { "timing", "stdout_level:" .. spawned_stdout_mode })
master:SpawnNonCritical("/usr/bin/pulser",
    { "pulser", "stdout_level:" .. spawned_stdout_mode })
master:Spawn("/usr/bin/system-proxy",
    { "system-proxy", "stdout_level:" .. spawned_stdout_mode })
--[[
master:Spawn("/usr/bin/network-proxy",
    { "network-proxy", "stdout_level:" .. spawned_stdout_mode })
]]
if not(config.target == 'edaq') then
    master:Spawn("/usr/bin/storage",
        { "storage", "stdout_level:" .. spawned_stdout_mode })

    -- Run the ptp runner.  This basically fires up PTP daemons on inside
    -- network interfaces when the interface has a link.  If the link goes away
    -- the daemon is killed and then restarted when the link returns.
    master:Spawn("/usr/bin/ptpv2d-ffi",
        { "ptpv2d-ffi", "eth1", "eth2", "stdout_level:" .. spawned_stdout_mode })

    -- Run the QX finder.  This spawns qx proxies for each qx device that
    -- shows up on the network and kills any qx proxies for devices that
    -- disappear from the network.  Note, proxies will only be started for
    -- qx devices that are reachable (in a network sense).
    master:Spawn("/usr/bin/mx-manager",
        { "mx-manager", "stdout_level:" .. spawned_stdout_mode })

    -- Run the camera manager.  This spawns camera-proxies for each camera that
    -- shows up on the network and kills any camera-proxies for cameras that
    -- disappear from the network.  Note, proxies will only be started for
    -- cameras that are reachable (in a network sense).

    -- XXX Per Ticket #3498.  Removing invocation of camera-manager process to
    -- prevent unnecessary use of system resources.  Later to be reinstated
    -- when feature is slated for release.
    -- master:Spawn("/usr/bin/camera-manager",
    --     { "camera-manager", "stdout_level:" .. spawned_stdout_mode })

    -- Run the HBM discovery daemon to announce us to whatever happens to be
    -- listening for us
    master:Spawn("/usr/bin/hbm-discovery",
        { "hbm-discovery", "stdout_level:" .. spawned_stdout_mode }, true)
end

-- Move all of the stuff from the update fpga script's log to the log database
ev.Timer.new(function(l, t, e)
        local ulf = io.open('/tmp/update-fpga.log', 'r')
        if ulf then
            while true do
                local line = ulf:read('*l')
                if not line then break end
                logger:Log('update-fpga', 'info', line)
            end
            ulf:close()
        end
    end, 1, 0):start(loop)

loop:loop()
