--! \file

local common = require 'mx.device_common'
local underscore = require 'underscore'

local m = {}

-- XML used to do initial configuration of channels when proxy starts up
local default_xml = [[<?xml version="1.0" encoding="UTF-8"?>
                <Amplifier>
                    <Connector#conn_num#>
                        <Mode choice="AnalogInChannel">
                            <AnalogInChannel>
                                <Sensor>
                                    <TedsOptions>
                                        <UseTeds>2</UseTeds>
                                    </TedsOptions>
                                </Sensor>
                                <SyncSignal address="1">
                                    <OutputRate>10</OutputRate>
                                </SyncSignal>
                                <OutputUnit>
                                    <UnitId>701</UnitId>
                                </OutputUnit>
                            </AnalogInChannel>
                        </Mode>
                    </Connector#conn_num#>
                </Amplifier>
]]

-- Function used to return the XML used to set the channel to it's default
-- configuration
function m:get_default_xml(channel_num)
    return default_xml:gsub('#conn_num#', tostring(channel_num))
end

m.json2xpath = {}
underscore.extend(m.json2xpath, common.json2xpath)
m.json2xpath.sensor_input.type =
    '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Thermocouple/ThermocoupleType'


-- List of parameters in a channel setup that we do NOT want sent off to the MX
-- device
m.ignore_list = {}
underscore.extend(m.ignore_list, common.ignore_list)

-- List of parameters returned by MX device that we don't verify after setting
-- up a channel.
m.dont_verify_list = {}
underscore.extend(m.dont_verify_list, common.dont_verify_list)

function m:get_default_spec()
    return self.default_spec
end

m.phase_delay = {}
m.phase_delay.fixed = 398
m.phase_delay.bessel_iir = {}
m.phase_delay.butterworth_iir = {}
m.phase_delay.linear_phase_fir = {}
m.phase_delay.butterworth_fir = {}

function m:sensor_input_type(p, fromxml)
    return m.translate({
            ['0'] = 'b',
            ['1'] = 'e',
            ['2'] = 'j',
            ['3'] = 'k',
            ['4'] = 'n',
            ['5'] = 'r',
            ['6'] = 's',
            ['7'] = 't',
            ['8'] = 'c',
            ['9'] = 'd',
            ['10'] = 'g',
            ['11'] = 'l',
            ['12'] = 'u',
            ['13'] = 'a',
        }, p, fromxml)
end

m.translator = {
    sensor_input_type = m.sensor_input_type
}

function m:to_external(spec)
    spec.calibration = spec.calibration or {}
    spec.calibration.define_source_units = {
        purpose = 'define_source_units',
        applied = true,
        units = '°C',
    }
    spec.calibration.sensor_scaling = {
        purpose = 'sensor_scaling',
        mode = 'defined_slope_intercept',
        slope = '1',
        intercept = '0',
        applied = true,
        units = '°C',
    }
    return spec
end

function m:to_internal(spec)
    spec.calibration = spec.calibration or {}
    spec.calibration.define_source_units = {
        purpose = 'define_source_units',
        applied = true,
        units = '°C',
    }
    spec.calibration.sensor_scaling = {
        purpose = 'sensor_scaling',
        mode = 'internal',
        units = '°C',
    }
    return spec
end

-- Specification of the default settings of this channel type
m.default_spec = {
    sample_rate = '10',
    sensor_input = {
        mode = 'thermocouple',
        type = 'k',
    },
    calibration = {
        define_source_units = {
            purpose = 'define_source_units',
            applied = true,
            units = '°C',
        },
        sensor_scaling = {
            purpose = 'sensor_scaling',
            mode = 'defined_slope_intercept',
            slope = '1',
            intercept = '0',
            applied = true,
            units = '°C',
        },
        -- NOTE:  We aren't using the unit converter here because it doesn't
        -- work in user defined mode (at this point).  So...the test engine
        -- gets to deal with unit conversion and zero balancing.
        units_conversion = {
            purpose = 'units_conversion',
            mode = 'defined_slope_intercept',
            applied = false,
            units = '°C',
            slope = '1',
            intercept = '0',
        },
        rezero = {
            purpose = 'rezero',
            mode = 'defined_slope_intercept',
            intercept = '0.00000',
            applied = false,
        },
    },
    digital_filter_type = 'bessel_iir',
    digital_filter_attenuation_frequency = '1',
    sensor = {
        use_teds = 2,
    },
}

return m
