require('lj_include')(
    'string',
    'errno',
    'sys/uio'
)

require 'middleclass'
local net = require 'net'
local StreamBuffer = net.StreamBuffer
local ev = require 'ev'
local ffi = require 'ffi'
local C = ffi.C
local cjson = require 'cjson'
local http_parser = require 'http_parser'
local config = require 'system_config'
local sys_notify = require 'sys_notify'

config.mx_data_rate = tonumber(config.mx_data_rate or 20) or 20

local Stream = class('Stream')

local float_size = ffi.sizeof('float')

ffi.cdef[[
// Data stream definitions
enum {
    MSG_TYPE_SIGNAL = 1,
    MSG_TYPE_METAINFO = 2,
};

enum {
    METAINFO_TYPE_JSON = 1,
};

typedef struct qx_short_hdr {
    unsigned int        h;
} qx_short_hdr_t;

typedef struct qx_long_hdr {
    unsigned int        h;
    unsigned int        size;
} qx_long_hdr_t;

struct sink_hdr {
    unsigned int        len;
    unsigned long long  ts;
} __attribute__((packed));

typedef struct sink_hdr sink_hdr_t;

enum {
    RB_BUF_SIZE = 80000,        // 20K samples
};

typedef struct sample_buffer {
    int         bytes;
    int         offset;
    struct iovec iov[2];
    sink_hdr_t  sink_hdr[1];
    union {
        char    cbuf[RB_BUF_SIZE];
        short   sbuf[RB_BUF_SIZE / sizeof(short)];
        int     ibuf[RB_BUF_SIZE / sizeof(int)];
        long    lbuf[RB_BUF_SIZE / sizeof(long)];
        float   fbuf[RB_BUF_SIZE / sizeof(float)];
        double  dbuf[RB_BUF_SIZE / sizeof(double)];
    } samples;
} sample_buffer_t;

typedef struct channel_info {
    unsigned int        in_use;
    char                reference[128];
    double              first_sample_ts;
    double              mx_first_sample_ts;
    struct {
        unsigned long   seconds;
        unsigned long   fraction;
    } mx_first_sample_ntp;
    double              original_first_sample_ts;
    double              start_ts;
    unsigned long long  samples_to_te;
    int                 sink_fd;
    unsigned long long  start_sample_num;
    unsigned long long  end_sample_num;
    unsigned long long  total_samples;
    double              sample_rate;
    double              sample_period;
    sample_buffer_t     sample_buffer;
    float               comet_data;
} channel_info_t;

enum {
    MAX_SUBSCRIBED_CHANNELS = 50,
};

enum {
    GOT_INIT = 1,
    GOT_AVAILABLE = 2,
    GOT_VERSION = 4,
    GOT_ALL = 7,
};

]]

local mx_epoch = os.time(
{
    year = 2000,
    month = 1,
    day = 1,
    hour = 0,
    min = 0,
    sec = 0,
})

local subscribed_channels =
    ffi.new('channel_info_t[?]', C.MAX_SUBSCRIBED_CHANNELS)

local read_buffer_size = 20000 * 20 * 16 / 10

local id_to_reference_map = {
    [1] = 'AnalogIn_Connector1.Signal2',
    [2] = 'AnalogIn_Connector1.Signal1',
    [4] = 'AnalogIn_Connector2.Signal2',
    [5] = 'AnalogIn_Connector2.Signal1',
    [7] = 'AnalogIn_Connector3.Signal2',
    [8] = 'AnalogIn_Connector3.Signal1',
    [10] = 'AnalogIn_Connector4.Signal2',
    [11] = 'AnalogIn_Connector4.Signal1',
    [13] = 'AnalogIn_Connector5.Signal2',
    [14] = 'AnalogIn_Connector5.Signal1',
    [16] = 'AnalogIn_Connector6.Signal2',
    [17] = 'AnalogIn_Connector6.Signal1',
    [19] = 'AnalogIn_Connector7.Signal2',
    [20] = 'AnalogIn_Connector7.Signal1',
    [22] = 'AnalogIn_Connector8.Signal2',
    [23] = 'AnalogIn_Connector8.Signal1',
    [25] = 'AnalogIn_Connector9.Signal2',
    [26] = 'AnalogIn_Connector9.Signal1',
    [28] = 'AnalogIn_Connector10.Signal2',
    [29] = 'AnalogIn_Connector10.Signal1',
    [31] = 'AnalogIn_Connector11.Signal2',
    [32] = 'AnalogIn_Connector11.Signal1',
    [34] = 'AnalogIn_Connector12.Signal2',
    [35] = 'AnalogIn_Connector12.Signal1',
    [37] = 'AnalogIn_Connector13.Signal2',
    [38] = 'AnalogIn_Connector13.Signal1',
    [40] = 'AnalogIn_Connector14.Signal2',
    [41] = 'AnalogIn_Connector14.Signal1',
    [43] = 'AnalogIn_Connector15.Signal2',
    [44] = 'AnalogIn_Connector15.Signal1',
    [46] = 'AnalogIn_Connector16.Signal2',
    [47] = 'AnalogIn_Connector16.Signal1',
}

local reference_to_id_map = {}
for i, v in pairs(id_to_reference_map) do
    reference_to_id_map[v] = i
end

local function get_stream_id(ref)
    return reference_to_id_map[ref]
end

function Stream:dbg_msg(msg, ...)
    self.log:log(self.log.DEBUG, msg, ...)
end

function Stream:error(msg)
    self.log:log(self.log.ERR, msg)
end

function Stream:_decode_message(buf, size)
    local consumed = ffi.sizeof('unsigned int')
    -- Do we have enough for the header?
    if size < consumed then return 0 end

    local hdr = ffi.cast('unsigned int *', buf)
    self.msg.signal_number = bit.band(hdr[0], 0xfffff)
    self.msg.size = bit.band(bit.rshift(hdr[0], 20), 0xff)
    local metainfo_offset = 1
    if self.msg.size == 0 then
        consumed = consumed + ffi.sizeof('unsigned int')
        -- Do we have enough for the header and optional size?
        if size < consumed then return 0 end
        self.msg.size = hdr[1]
        metainfo_offset = 2
    end

    -- Do we have the whole message?
    if size < consumed + self.msg.size then return 0 end

    self.msg.type = bit.band(bit.rshift(hdr[0], 28), 0x3)
    self.msg.data = buf + consumed
    consumed = consumed + self.msg.size
    self.msg.metadata_type = nil
    if self.msg.type == C.MSG_TYPE_METAINFO then
        self.msg.metadata_type = hdr[metainfo_offset]
        self.msg.data = self.msg.data + ffi.sizeof('unsigned int')
        self.msg.size = self.msg.size - ffi.sizeof('unsigned int')
    end

    -- Return how much we've used
    return consumed
end

function Stream:_process_stream_metainfo(str)
    if str:match('"method":"alive"') then
        self.got_heartbeat = true
        return
    end
    local json = cjson.decode(str)
    if not json then
        self:error(string.format("Unable to decode JSON: %s", str))
        self.sbuf:shutdown()
    elseif json.method then
        if json.method == 'version' or json.method == 'apiVersion' then
            self.stream_version = json.params[1]
            self.got = bit.bor(self.got or 0, C.GOT_VERSION)
            if self.got and self.got == C.GOT_ALL and self.ready_cb then
                self.ready_cb(self.channel_list)
            end
        elseif json.method == 'init' then
            self.stream_id = json.params.streamId
            self.http_info = json.params.commandInterfaces['jsonrpc-http']
            if self.http_info.port:lower() == 'http' then
                self.http_info.port = 80
            else
                self:error(string.format("Unknown control channel port %s",
                    self.http_info.port))
                self.sbuf:shutdown()
            end
            self.http_info.jsonrpc = "2.0"
            self.channel_list = nil
            self.got = bit.bor(self.got or 0, C.GOT_INIT)
            if self.got and self.got == C.GOT_ALL and self.ready_cb then
                self.ready_cb(self.channel_list)
            end
        elseif json.method == 'available' then
            if not self.channel_list then self.channel_list = {} end
            for _, c in ipairs(json.params) do
                self.channel_list[#self.channel_list + 1] = c
            end
            self.got = bit.bor(self.got or 0, C.GOT_AVAILABLE)
            if self.got and self.got == C.GOT_ALL and self.ready_cb then
                self.ready_cb(self.channel_list)
            end
        elseif json.method == 'unavailable' then
            for _, c in ipairs(json.params) do
                for i, ic in ipairs(self.channel_list) do
                    if c == ic then
                        table.remove(self.channel_list, i)
                        break
                    end
                end
            end
        elseif json.method == 'connect' then
            -- I don't think we care about this
        elseif json.method == 'disconnect' then
            -- I don't think we care about this
        elseif json.method == 'time' then
            -- We don't care about this one
        elseif json.method == 'alive' then
            -- Reset the connection monitor timer so it doesn't fire
            self.got_heartbeat = true
        elseif json.method == 'fill' then
            local f = tonumber(json.params[1])
            if f > 97 then
                self:error(string.format("QuantumX buffers are full"))
                self.sbuf:shutdown()
                os.exit(-2)
            elseif f > 75 then
                self.log:log(self.log.WARNING, string.format(
                    "Warning Quantum device buffers are %d%% full", f))
            end
        else
            self:error(string.format("Unrecognized JSON: %s", str))
            self.sbuf:shutdown()
        end
    else
        self:error(string.format("Unrecognized JSON: %s", str))
        self.sbuf:shutdown()
    end
end

function Stream:_ntp_to_seconds(to)
    if to.type ~= 'ntp' then
        self:error(string.format("Time object is not NTP"))
        self.sbuf:shutdown()
        return
    end
    local t = tonumber(to.seconds)
    if not to.subFraction then
        t = t + (tonumber(to.fraction) / 0x100000000)
    else
        local v = bit.bor(bit.lshift(
            ffi.new('unsigned long long', to.fraction), 32),
            ffi.new('unsigned long long', to.subFraction))
        t = t + tonumber(v) / tonumber(0x10000000000000000)
--[[
print(string.format("Subfraction:  %8.8X %8.8X %.30f %.60f", to.fraction,
to.subFraction, t, 1/t))
]]
    end
    return t
end

function Stream:_ntp_to_ts(to, no_offset)
    if no_offset then return self:_ntp_to_seconds(to) end
    
    -- Convert to seconds and add in the diffence in seconds between the MX
    -- and CX epochs
    return (self:_ntp_to_seconds(to) + mx_epoch)
end

function Stream:get_stats(ref)
    local signo = reference_to_id_map[ref]
    if not signo then return end
    local siginfo = subscribed_channels[signo]
    if siginfo.in_use == 0 then return end
    return tonumber(siginfo.samples_to_te), tonumber(siginfo.start_sample_num),
        tonumber(siginfo.total_samples)
end

function Stream:set_start_time(start_ts, ref)
    if start_ts == 0 and not ref then
        for i = 0, C.MAX_SUBSCRIBED_CHANNELS - 1 do
            subscribed_channels[i].sink_fd = -1
            subscribed_channels[i].start_ts = 0
        end
        return
    end
    local signo = reference_to_id_map[ref]
    if not signo then return end
    local siginfo = subscribed_channels[signo]
    siginfo.start_ts = start_ts
end

function Stream:attach_sink(ref, sink_fd, end_ts)
    local signo = reference_to_id_map[ref]
    if not signo then return end
    local siginfo = subscribed_channels[signo]
    if siginfo.in_use == 0 then return end

    siginfo.samples_to_te = 0

    -- TODO:  Ending tests
    siginfo.end_sample_num = end_ts or -1
    siginfo.sink_fd = sink_fd
end

function Stream:detach_sink(ref)
    local signo = reference_to_id_map[ref]
    if not signo then return end
    local siginfo = subscribed_channels[signo]
    if siginfo.in_use == 0 then return end

    siginfo.start_ts = 0
    siginfo.sink_fd = -1
    siginfo.samples_to_te = 0
    siginfo.start_sample_num = 0
    siginfo.end_sample_num = 0
end


function Stream:test_start_abort(msg, siginfo)
    self.set_start_time(0)
--    self.qx.start_ts = nil
    self.subscribe_cb = {}

    local emsg
    if siginfo then
        siginfo.sink_fd = -1
        siginfo.samples_to_te = 0
        siginfo.start_sample_num = 0
        siginfo.end_sample_num = 0
        local ref = ffi.string(siginfo.reference)
        emsg = string.format("%s: %s", ref, msg)
    else
        emsg = string.format("%s", msg)
    end

    -- KLUDGE:  If more than one quantum proxy is involved in the test AND the
    -- test start is aborted (due to there not being enough time allowed to
    -- complete channel setup or the like), proxies can die with segfaults.
    -- This is intended to be a short term work around until multi-stage start
    -- is implemented.
    if not self.abort_timer then
        self.abort_timer = ev.Timer.new(function(l, t, e)
            self:error(emsg)
            t:stop(l)
            self.abort_timer = nil
            end, 1, 1)
        self.abort_timer:start(self.loop)
    else
        self.abort_timer:again(self.loop)
    end
end

function Stream:_get_num_samples_to_discard(siginfo)
    -- Convert the CX timebase to MX timebase
    local new_start_ts = siginfo.start_ts - mx_epoch

    -- Calculate the number of samples between the first sample timestamp and
    -- the test start time.
    local f_start_sample_num =
        ((new_start_ts - siginfo.original_first_sample_ts) /
            siginfo.sample_period)
    local start_sample_num = math.floor(f_start_sample_num + .5)

    local samples_between_first_sample_ts_and_epoch =
        siginfo.original_first_sample_ts / siginfo.sample_period

    local function log_info(level, xtra_msg)
        local msg = string.format("\n" ..
            "%s" ..
            "\tReference: '%s'\n" ..
            "\tTS of 1st wanted sample (CX - MX EPOCH): %.9f\n" ..
            "\tUnmodifed 1st sample TS (NTP): %.0f (s) %.0f (frac)\n" ..
            "\tUnmodifed 1st sample TS (double): %.9f\n" ..
            "\tFudge Factor: %.9f\n" ..
            "\tRaw first sample ts adjusted by fudge factor: %.9f\n" ..
            "\tNumber of samples between mx epoch and afsts: %.9f\n" ..
            "\tTime diff between rounded samples time and epoch (us): %.5f\n" ..
            "\tSample Period: %.9f\n" ..
            "\tSample Rate:   %.9f\n" ..
            "\tSamples to discard to get to 1st wanted (float): %.9f\n" ..
            "\tSamples to discard to get to 1st wanted (rounded): %f",
            xtra_msg or '',
            ffi.string(siginfo.reference),
            new_start_ts,
            siginfo.mx_first_sample_ntp.seconds, -- Unmodified 1st sample NTP
            siginfo.mx_first_sample_ntp.fraction, -- Unmodified 1st sample NTP
            siginfo.mx_first_sample_ts, -- Unmodified 1st sample TS
            self.ts_adjust_seconds,         -- Fudge factor in seconds
            siginfo.original_first_sample_ts,  -- Raw 1st sample ts with fudge
            samples_between_first_sample_ts_and_epoch,
            (samples_between_first_sample_ts_and_epoch -
                math.floor(samples_between_first_sample_ts_and_epoch + 0.5)) *
                siginfo.sample_period * 1000000,
            siginfo.sample_period,  -- Sample period
            1.0 / siginfo.sample_period,    -- Sample Rate
            f_start_sample_num,     -- Raw number of samples to discard
            start_sample_num)      -- Rounded number of samples to discard
        if level and level == self.log.ERR then
            self:error(msg)
        else
            self.log:log(level or self.log.INFO, msg)
        end
    end

    -- Make sure the first sample timestamp is before or at the test start time
    if start_sample_num < 0 then
        log_info(self.log.ERR,
            "\tERROR - First sample timestamp is after test start time\n")
        self:error("Test aborted")
        return 0
    end

    if config.mx_phase_debug_messages then
        log_info()
    end

    -- Issue an error if the number of samples between the MX epoch and
    -- the afsts is not within a very small fraction of an integer boundary
    if math.abs((samples_between_first_sample_ts_and_epoch -
        math.floor(samples_between_first_sample_ts_and_epoch + 0.5)) *
        siginfo.sample_period * 1000000) > .25 then
        log_info(self.log.ERR,
            "\tERROR - First sample timestamp is out of tolerance\n")
        local sn = sys_notify:new('MX Device Error')
        sn:critical_error(
            'First sample timestamp from device is out of tolerance', false)
        -- FIXME:  Exit so the proxy will restart.  Supposedly doing so will
        -- somehow magically make the out of tolerance error go away.  What
        -- really needs to happen here is to restart the test after the proxy
        -- comes back up!
        os.exit(-1)
    end

    return start_sample_num
end

function Stream:_process_channel_metainfo(str, _signo)
    local json = cjson.decode(str)
    if not json then
        self:error(string.format("Unable to decode JSON: %s", str))
        self.sbuf:shutdown()
        return
    end
    local signo = _signo or self.msg.signal_number
    if json.method then
        if json.method == 'subscribe' then
            json.subscribe = json.params[1]
        elseif json.method == 'unsubscribe' then
            json.unsubscribe = true
        end
    end
    if json.subscribe then
--        self:dbg_msg(string.format("Subscribed to: %s", json.subscribe))
        C.strcpy(subscribed_channels[signo].reference, json.subscribe)
        subscribed_channels[signo].first_sample_ts = 0
        subscribed_channels[signo].sample_period = 0
        subscribed_channels[signo].samples_to_te = 0
        subscribed_channels[signo].start_sample_num = 0
        subscribed_channels[signo].end_sample_num = 0
        subscribed_channels[signo].total_samples = 0
        subscribed_channels[signo].in_use = 1
        subscribed_channels[signo].sink_fd = -1
        if self.subscribe_cb[json.subscribe] then
            self.subscribe_cb[json.subscribe](signo, json.subscribe)
        end

        if config.mx_data_rate and config.mx_data_rate == 0 and
            not self.data_processor[signo] then
            self.data_processor[signo] = true
            self.data_processor_info[#self.data_processor_info + 1] = {
                signo = signo,
                func = function(signo)
                    local siginfo = subscribed_channels[signo]
                    local sb = siginfo.sample_buffer
                    local offset = sb.offset
                    if offset > 0 then
                        -- How many samples do we have
                        local num_samples = sb.offset / ffi.sizeof('float')

                        -- Make last sample in buffer available to comet stream
                        siginfo.comet_data = sb.samples.fbuf[num_samples - 1]

                        -- If a sink is attached, send the data to the sink
                        if siginfo.sink_fd > 0 then
                            local d = sb.samples.fbuf
                            if siginfo.total_samples <
                                siginfo.start_sample_num then
                                -- Trivially reject all samples because they
                                -- all fall BEFORE the first sample we want
                                num_samples = 0
                            elseif siginfo.total_samples - num_samples >=
                                siginfo.start_sample_num then
                                -- Trivially accept all samples because they
                                -- all fall AFTER the first sample we want
                            else
                                -- Some of the samples are BEFORE the first
                                -- sample we want and some are after.
                                local ns = siginfo.total_samples -
                                    siginfo.start_sample_num
                                d = d + (num_samples - ns)
                                num_samples = ns
                            end
                            if num_samples > 0 then
                                local len = (num_samples * ffi.sizeof('float'))
                                sb.sink_hdr[0].len = len +
                                    ffi.sizeof('unsigned long long')
                                sb.sink_hdr[0].ts = siginfo.samples_to_te
                                sb.iov[1].iov_base = d
                                sb.iov[1].iov_len = len
                                local tlen = len + sb.iov[0].iov_len
                                local sent =
                                    C.writev(siginfo.sink_fd, sb.iov, 2)
                                if sent < 0 then
                                    local eno = ffi.errno()
                                    if eno == C.EAGAIN then
                                        -- Retry the write next time
                                        return

                                    -- These are the errors that may naturally
                                    -- occur when the MX device goes away.
                                    elseif eno ~= C.ECONNRESET and
                                        eno ~= C.EINVAL and
                                        eno ~= C.EBADF and
                                        eno ~= C.EPIPE then
                                        self:error(string.format(
                                            "writev error on %d: %d",
                                            siginfo.sink_fd,
                                            eno))
                                    end
                                    siginfo.sink_fd = -1
                                    return
                                elseif sent == 0 then
                                    siginfo.sink_fd = -1
                                    return
                                elseif sent ~= tlen then
                                    -- FIXME:  Should deal with this properly!
                                    self.log:log(self.log.ERR, string.format(
[[Short write: %d -> %d (%d)
This is usually an indication of an overloaded system]],
sent, tlen, siginfo.sink_fd))
                                    C.close(siginfo.sink_fd)
                                    siginfo.sink_fd = -1
                                    return
                                end
            
                                -- Increment count of samples
                                siginfo.samples_to_te =
                                    siginfo.samples_to_te + num_samples
                            end
                        end

                        -- Flush the buffer
                        sb.offset = 0
                    end
                end
            }
        elseif not self.rb_timer[signo] then
            -- Timer used to send data from the buffer to the comet
            -- servers and the test engine.
            self.rb_timer[signo] = ev.Timer.new(function(l, t, e)
                    local siginfo = subscribed_channels[signo]
                    local sb = siginfo.sample_buffer
                    local offset = sb.offset
                    if offset > 0 then
                        -- How many samples do we have
                        local num_samples = sb.offset / ffi.sizeof('float')

                        -- Make last sample in buffer available to comet stream
                        siginfo.comet_data = sb.samples.fbuf[num_samples - 1]

                        -- If a sink is attached, send the data to the sink
                        if siginfo.sink_fd > 0 then
                            local d = sb.samples.fbuf
                            if siginfo.total_samples <
                                siginfo.start_sample_num then
                                -- Trivially reject all samples because they
                                -- all fall BEFORE the first sample we want
                                num_samples = 0
                            elseif siginfo.total_samples - num_samples >=
                                siginfo.start_sample_num then
                                -- Trivially accept all samples because they
                                -- all fall AFTER the first sample we want
                            else
                                -- Some of the samples are BEFORE the first
                                -- sample we want and some are after.
                                local ns = siginfo.total_samples -
                                    siginfo.start_sample_num
                                d = d + (num_samples - ns)
                                num_samples = ns
                            end
                            if num_samples > 0 then
                                local len = (num_samples * ffi.sizeof('float'))
                                sb.sink_hdr[0].len = len +
                                    ffi.sizeof('unsigned long long')
                                sb.sink_hdr[0].ts = siginfo.samples_to_te
                                sb.iov[1].iov_base = d
                                sb.iov[1].iov_len = len
                                local tlen = len + sb.iov[0].iov_len
                                local sent =
                                    C.writev(siginfo.sink_fd, sb.iov, 2)
                                if sent < 0 then
                                    local eno = ffi.errno()
                                    if eno == C.EAGAIN then
                                        -- Retry the write next time
                                        return

                                    -- These are the errors that may naturally
                                    -- occur when the MX device goes away.
                                    elseif eno ~= C.ECONNRESET and
                                        eno ~= C.EINVAL and
                                        eno ~= C.EBADF and
                                        eno ~= C.EPIPE then
                                        self:error(string.format(
                                            "writev error on %d: %d",
                                            siginfo.sink_fd,
                                            eno))
                                    end
                                    t.siginfo.sink_fd = -1
                                    return
                                elseif sent == 0 then
                                    t.siginfo.sink_fd = -1
                                    return
                                elseif sent ~= tlen then
                                    -- FIXME:  Should deal with this properly!
                                    self.log:log(self.log.ERR, string.format(
[[Short write: %d -> %d (%d)
This is usually an indication of an overloaded system]],
sent, tlen, siginfo.sink_fd))
                                    C.close(siginfo.sink_fd)
                                    t.siginfo.sink_fd = -1
                                    return
                                end
            
                                -- Increment count of samples
                                siginfo.samples_to_te =
                                    siginfo.samples_to_te + num_samples
                            end
                        end

                        -- Flush the buffer
                        sb.offset = 0
                    end
                end,
                1.0 / (config.mx_data_rate or 5.0),
                1.0 / (config.mx_data_rate or 5.0))
            self.rb_timer[signo].siginfo = subscribed_channels[signo]
            self.rb_timer[signo].signo = signo
            self.rb_timer[signo]:start(self.loop)
        end

        return
    elseif json.unsubscribe then
--        self:dbg_msg(string.format("Unsubscribed from: %s",
--            id_to_reference_map[signo]))
        subscribed_channels[signo].in_use = 0
        return
    end
    if json.method then
        if json.method == 'time' then
            json.time = json.params
        elseif json.method == 'signalRate' then
            json.signal_rate = json.params
        elseif json.method == 'data' then
            json.data = json.params
            json.data.value_type = json.data.valueType
        end
    end
    local siginfo = subscribed_channels[signo]
    if siginfo.in_use == 0 then
        self:error(string.format("No signal info for %d", signo))
        if self.sbuf then
            self.sbuf:shutdown()
        end
        return
    elseif json.time then
        siginfo.first_sample_ts = self:_ntp_to_ts(json.time.stamp)
        siginfo.mx_first_sample_ts = self:_ntp_to_ts(json.time.stamp, true)
        siginfo.mx_first_sample_ntp.seconds = ffi.new('unsigned long',
                tonumber(json.time.stamp.seconds))
        siginfo.mx_first_sample_ntp.fraction = ffi.new('unsigned long',
                tonumber(json.time.stamp.fraction))

        -- Adjust the first sample time stamp by the MX device's analog guard
        -- filter delay plus some "fudge" factor that appears to be one 4801Hz
        -- sample period.
        local save_f =
            ffi.new('unsigned long', tonumber(json.time.stamp.fraction))
        local new_f = save_f + ffi.new('long', self.ts_adjust)
        if self.ts_adjust >= 0 then
            if tonumber(new_f) < tonumber(save_f) then
                local s = tonumber(json.time.stamp.seconds) + 1
                json.time.stamp.seconds = string.format("%.f", s)
            end
        else
            if tonumber(new_f) > tonumber(save_f) then
                local s = tonumber(json.time.stamp.seconds) - 1
                json.time.stamp.seconds = string.format("%.f", s)
            end
        end
        json.time.stamp.fraction = string.format("%.f", tonumber(new_f))
        siginfo.original_first_sample_ts =
            self:_ntp_to_ts(json.time.stamp, true)


        siginfo.total_samples = 0
        local sigref = ffi.string(siginfo.reference)
        self:dbg_msg(string.format("First sample TS for %s: %12.9f",
            sigref, siginfo.first_sample_ts))

        if siginfo.sample_period > 0 then
            local num_samples =
                math.floor(siginfo.first_sample_ts / siginfo.sample_period)
            local real_ts = num_samples * siginfo.sample_period
            local ts_offset = real_ts - siginfo.first_sample_ts
            self:dbg_msg(string.format("TS offset for %s: %12.9f",
                sigref, ts_offset))
--            siginfo.first_sample_ts = siginfo.first_sample_ts + ts_offset
local proposed_first_sample_ts = siginfo.first_sample_ts + ts_offset
            self:dbg_msg(string.format(
                "Adjusted (proposed) first sample TS for %s: %12.9f",
                sigref, proposed_first_sample_ts))
--                sigref, siginfo.first_sample_ts))
        end

        if siginfo.sample_period > 0 and siginfo.start_ts > 0 then

            -- If debug files stuff is enabled, dump the raw start time stamps
            -- for each channel in the test to a file.
            if config.debug_files then
                if self.dump_info and self.dump_info.file and
                    self.dump_info.reference_list then
                    if not self.dump_info.fh then
                        self.dump_info.fh = io.open(self.dump_info.file, 'w+')
                        if not self.dump_info.fh then
                            self.dump_info = nil
                            self.log:log(self.log.WARNING,
                                "Unable to create dump info file")
                        else
                            self.dump_info.fh:write(
                                "Unit\tChan\tSeconds(u32)\tFraction(u32)\n")
                        end
                    end
                    if self.dump_info and self.dump_info.fh then
                        if self.dump_info.reference_list and
                            self.dump_info.reference_list[sigref] then
                        end
                        local cnum = sigref:match("Connector(%d+)%.") or sigref
                        self.dump_info.fh:write(
                            string.format("%s\t%s\t%.0f\t%.0f\n",
                            self.dump_info.unit, cnum,
                            tonumber(siginfo.mx_first_sample_ntp.seconds),
                            tonumber(siginfo.mx_first_sample_ntp.fraction)))
                        self.dump_info.reference_list[sigref] = nil
                        local count = 0
                        for _, _ in pairs(self.dump_info.reference_list) do
                            count = count + 1
                        end
                        if count == 0 then
                            self.dump_info.fh:close()
                            self.dump_info = nil
                        end
                    end
                end
            end
            siginfo.start_sample_num = self:_get_num_samples_to_discard(siginfo)
        end
    elseif json.signal_rate then
        local t = self:_ntp_to_seconds(json.signal_rate.delta)
        if t > 0 then
            siginfo.sample_rate = tonumber(json.signal_rate.samples) / t

            -- KLUDGE imprecise sample periods when device is in decimal data
            -- rate mode.
            if self.drd and self.drd == 'decimal' then
                local real_sr = {
                    20000, 10000, 5000, 2500, 2000, 1000, 600,
                    500, 200, 100, 50, 20, 10, 5, 2, 1
                }
                for _, rt in ipairs(real_sr) do
                    if math.abs(siginfo.sample_rate - rt) <= 1 then
                        siginfo.sample_rate = tonumber(rt)
                        break
                    end
                end
            end

            siginfo.sample_period = 1 / siginfo.sample_rate
            self:dbg_msg(string.format("Sample rate for %s: %12.9f Hz",
                ffi.string(siginfo.reference), siginfo.sample_rate))
        else
            self:dbg_msg(string.format("Signal rate delta for %s is 0",
                ffi.string(siginfo.reference)))
        end

        if siginfo.first_sample_ts > 0 then
            local num_samples =
                math.floor(siginfo.first_sample_ts / siginfo.sample_period)
            local real_ts = num_samples * siginfo.sample_period
            local ts_offset = real_ts - siginfo.first_sample_ts
            self:dbg_msg(string.format("TS offset for %s: %12.9f",
                ffi.string(siginfo.reference), ts_offset))
--            siginfo.first_sample_ts = siginfo.first_sample_ts + ts_offset
local proposed_first_sample_ts = siginfo.first_sample_ts + ts_offset
            self:dbg_msg(string.format(
                "Adjusted (proposed) first sample TS for %s: %12.9f",
                ffi.string(siginfo.reference), proposed_first_sample_ts))
--                    ffi.string(siginfo.reference), siginfo.first_sample_ts))
        end
        if siginfo.first_sample_ts > 0 and siginfo.start_ts > 0 then
            siginfo.start_sample_num = self:_get_num_samples_to_discard(siginfo)
        end
    elseif json.data then
        if json.data.pattern ~= 'V' then
            self:error(string.format("Unexpected channel data pattern %s",
                json.data.pattern))
            self.sbuf:shutdown()
        elseif json.data.endian ~= 'big' then
            self:error(string.format("Unexpected channel data endianess %s",
                json.data.endian))
        elseif json.data.value_type ~= 'real32' then
            self:error(string.format("Unexpected channel data type %s",
                json.data.value_type))
        elseif json.data.time.type ~= 'ntp' then
            self:error(string.format("Unexpected channel data time type %s",
                json.data.time.type))
        end
    else
        self:error(string.format("Unknown channel metainfo JSON: %s", str))
        self.sbuf:shutdown()
    end
end

function Stream:_process_metainfo()
    if self.msg.metadata_type == C.METAINFO_TYPE_JSON then
        local str = ffi.string(self.msg.data, self.msg.size)
        if self.msg.signal_number == 0 then
            self:_process_stream_metainfo(str)
        else
            self:_process_channel_metainfo(str)
        end
    else
        self:error(string.format("Unexpected metadata type: %d",
            self.msg.metadata_type))
        self.sbuf:shutdown()
    end
end

function Stream:process_data()
    -- Get amount of data available and pointer to buffer
    local avail, buf = self.sbuf:read()

    local total_consumed = 0
    while total_consumed < avail do

        -- Calculate remaining bytes in buffer
        local remaining = avail - total_consumed

        -- Minimum amount needed
        local consumed = ffi.sizeof('qx_short_hdr_t')

        -- Do we have enough for the header?
        if remaining < consumed then break end

        -- Get pointer to raw data
        local b = buf + total_consumed

        -- Assume short header first
        local hdr = ffi.cast('qx_short_hdr_t *', b)
        local hsize = bit.band(bit.rshift(hdr.h, 20), 0xff)
        local data = b + ffi.sizeof('qx_short_hdr_t')

        -- If size is 0 (zero) then we have a long header
        if hsize == 0 then
            consumed = ffi.sizeof('qx_long_hdr_t')

            -- Do we have enough for the long header ?
            if avail < consumed then break end
            hdr = ffi.cast('qx_long_hdr_t *', b)
            hsize = hdr.size
            data = data + ffi.sizeof('uint32_t')
        end

        -- Do we have the whole message?
        if remaining < consumed + hsize then break end

        -- Get pointer to data
        local htype = bit.rshift(hdr.h, 28)
        local hsigno = bit.band(hdr.h, 0xfffff)

        if htype == C.MSG_TYPE_METAINFO then
            local metadata_type = ffi.cast('int *', data)[0]
            data = data + ffi.sizeof('unsigned int')
            local size = hsize - ffi.sizeof('unsigned int')

            if metadata_type == C.METAINFO_TYPE_JSON then
                local str = ffi.string(data, size)
                if hsigno == 0 then
                    self:_process_stream_metainfo(str)
                else
                    self:_process_channel_metainfo(str, hsigno)
                end
            else
                self:error(string.format("Unexpected metadata type: %d",
                    metadata_type))
                self.sbuf:shutdown()
                break
            end
        else
            local subinfo = subscribed_channels[hsigno]
            local sb = subinfo.sample_buffer
            local offset = sb.offset
            local len = hsize
            local num_samples = len / ffi.sizeof('float')

            -- If there is  more data available than can be put into the buffer
            -- flush the buffer first.
            if len > sb.bytes - offset then
                local w = self.rb_timer[hsigno]
                if w then
                    local tfunc = w:callback()
                    tfunc(ev.Loop.default, w, 0)
                    w:again(ev.Loop.default)
                end
                offset = sb.offset
            end

            -- If there is still too much data, then our buffer is simply too
            -- small and we must exit.
            if len > (sb.bytes - offset) then
                self:error("Input buffer is full.  " ..
                    "This is usually and indication that the data consumer " ..
                    "(test engine) is not able to keep up with the amount " ..
                    "of data being supplied by the channel proxy(s).")
                os.exit(-40)
            end

            ffi.copy(sb.samples.cbuf + offset, data, len)
            sb.offset = offset + len

            subinfo.total_samples = subinfo.total_samples + num_samples
        end

        -- Adjust count of total we've consumed
        total_consumed = total_consumed + consumed + hsize
    end

    -- Process all of the data received
    self:do_process_data()

    return total_consumed
end

function Stream:send_http_request(data, callback, lloop)
    local s = net.socket{ip = self.ip, port = self.http_info.port}
    local ok, socket = pcall(s.connect, s)
    if not ok then
        self:error(string.format("Can not connect to %s on port %s",
            self.ip, self.http_info.port))
--        self.sbuf:shutdown()
        callback(self, "{}")
        return
    end
    local sbuf = StreamBuffer:new(socket)
    sbuf.read_cb = function(loop)
        local avail, buf = sbuf:read()
        if avail then
            local response, endpos =
                http_parser.response:match(ffi.string(buf, avail))
            if response then
                if response.response_code and
                    response.response_code == '200' then
                    local rh = response.headers
                    if rh and rh.content_length then
                        local cl = tonumber(rh.content_length)
                        if cl + endpos > avail + 1 then return 0 end
                        if rh.content_type and
                            rh.content_type == 'application/json' then
                            local json = ffi.string(buf + endpos - 1, cl)
                            sbuf.read_cb = nil
                            callback(self, json)
                            sbuf:close()
                            return avail
                        else
                            self:error(string.format(
                                "Unexpected content type: %s", rh.content_type))
                            sbuf:close()
                        end
                    else
                        self:error("No content_length in response")
                        sbuf:close()
                    end
                else
                    self:error(string.format("HTTP response:  %s",
                        response.response_code or "NONE"))
                    sbuf:close()
                end
            end
        end
        return 0
    end
    sbuf:start(lloop or self.loop)
    local req
    if self.http_info.method and self.http_info.method == 'POST' then
        req = string.format('%s %s HTTP/1.0\r\n' ..
            'Content-Length: %d\r\n' ..
            'Content-Type:  application/json\r\n\r\n' ..
            '%s\r\n',
            self.http_info.method, self.http_info.absolute_path,
            #data + 2, data)
    elseif self.http_info.httpMethod and
        self.http_info.httpMethod == 'POST' then
        req = string.format('%s %s HTTP/%s\r\n' ..
            'Content-Length: %d\r\n' ..
            'Content-Type:  application/json\r\n\r\n' ..
            '%s\r\n',
            self.http_info.httpMethod, self.http_info.httpPath,
            self.http_info.httpVersion,
            #data + 2, data)
    elseif self.http_info.method then
        req = string.format('%s %s HTTP/1.0\r\n\r\n%s\r\n',
            self.http_info.method, self.http_info.absolute_path, data)
    else
        req = string.format('%s %s HTTP/1.0\r\n\r\n%s\r\n',
            self.http_info.httpMethod, self.http_info.httpPath, data)
    end
    sbuf:write(req)
end

function Stream:get_channel_list()
    return self.channel_list
end

function Stream:_find_channel_ref(ref)
    if self.channel_list then
        for _, r in ipairs(self.channel_list) do
            if r == ref then return true end
        end
    end
end


function Stream:get_comet_data(ref)
    local sid = get_stream_id(ref)
    if not sid then return end
    if subscribed_channels[sid].in_use == 0 then return end
    if math.abs(subscribed_channels[sid].comet_data) > 100000000 then return end
    return subscribed_channels[sid].comet_data or nil
end

function Stream:_read_http_response_subscribe(str)
end

function Stream:subscribe_channels(ref_list)
    local channels = {}
    for _, v in pairs(ref_list) do
        if not self:_find_channel_ref(v) then
            self:error(string.format(
                "Can not find channel ref \"%s\" in device channel list",
                chan_sub_info.reference))
                self.sbuf:shutdown()
                return
        end
        local sid = get_stream_id(v)
        if not sid then
            self:error(string.format(
                "Can not find channel sid for reference \"%s\"", v))
                self.sbuf:shutdown()
                return
        end
        channels[#channels + 1] = v
    end
    if #channels > 0 then
        local json = cjson.encode({
            jsonrpc = self.http_info.jsonrpc,
            method = string.format('%s.subscribe', self.stream_id),
            params = channels,
            id = math.random(65535),
        })
        self:send_http_request(json, self._read_http_response_subscribe)
    end
end

function Stream:_resubscribe_read_http_response_unsubscribe(str)
    local json = cjson.decode(str)
    if json and json.id then
        local ref = id_to_reference_map[json.id]
        if ref then
            -- Now subscribe to the channel
            local method = string.format('%s.subscribe', self.stream_id)
            local _json = cjson.encode({
                jsonrpc = self.http_info.jsonrpc,
                method = method,
                params = { ref },
                id = json.id,
            })
            local status = self:send_http_request(_json,
                self._read_http_response_subscribe)
            return
        end
        self:error(string.format("No reference for ID: %d", json.id))
    else
        self:error(string.format(
            "Response to unsubscribe does not appear to be JSON\n" ..
            "MESSAGE \"%s\"", str))
    end
    self.sbuf:shutdown()
end

function Stream:unsubscribe_all(ref_list, cb)
    -- Unsubscribe all listed channels
    local method = string.format('%s.unsubscribe', self.stream_id)
    local json = cjson.encode({
        jsonrpc = self.http_info.jsonrpc,
        method = method,
        params = ref_list,
        id = math.random(256),
    })
    self:send_http_request(json, function(self, str)
            if cb then cb(str) end
        end)
end

function Stream:subscribe_all(ref_list, cb)
    -- We are subscribed to this channel, unsubscribe from it
    local method = string.format('%s.subscribe', self.stream_id)
    local json = cjson.encode({
        jsonrpc = self.http_info.jsonrpc,
        method = method,
        params = ref_list,
        id = math.random(256),
    })
    local status = self:send_http_request(json, function(self, str)
            if cb then cb(str) end
        end)
end

function Stream:resubscribe_channel(ref)
    local sid = get_stream_id(ref)
    if not sid then
        self:error(string.format(
            "Can not find channel sid for reference \"%s\"", v))
        if cb then cb() end
        return
    end
    if subscribed_channels[sid].in_use > 0 then
        -- We are subscribed to this channel, unsubscribe from it
        local method = string.format('%s.unsubscribe', self.stream_id)
        local json = cjson.encode({
            jsonrpc = self.http_info.jsonrpc,
            method = method,
            params = { ref },
            id = sid,
        })
        local status = self:send_http_request(json,
            self._resubscribe_read_http_response_unsubscribe)

        return
    end

    -- Just subscribe to the channel if we are not already subscribed to it
    local method = string.format('%s.subscribe', self.stream_id)
    local json = cjson.encode({
        jsonrpc = self.http_info.jsonrpc,
        method = method,
        params = { ref },
        id = sid,
    })
    local status = self:send_http_request(json,
        self._read_http_response_subscribe)
end

function Stream:_read_http_response_channel_list(str)
    local json = cjson.decode(str)
    if json and json.result then
        self.channel_list = json.result
        if self.ready_cb then
            self.ready_cb(self.channel_list)
        end
        return
    end
    self:error(string.format(
        "Response to list request does not appear to be JSON\n" ..
        "MESSAGE \"%s\"", str))
    self.sbuf:shutdown()
end

function Stream:shutdown()
    if self.connection_monitor_timer then
        self.connection_monitor_timer:stop(self.loop)
        self.connection_monitor_timer = nil
    end
    if self.sbuf then
        self.sbuf.stop_cb = nil
        self.sbuf.read_cb = nil
        self.sbuf:shutdown()
        self.sbuf = nil
    elseif self.socket then
        self.socket:close()
    end
end

function Stream:data_rate_domain(drd)
    self.drd = drd
end

--function Stream:initialize(loop, ip, port, logger, ecallback, read_rate)
function Stream:initialize(ip, logger, loop, read_rate, ts_adjust)
    self.log = logger
    self.loop = loop
    self.msg = {}
    self.rb_timer = {}
    self.ip = ip
    self.ts_adjust = ts_adjust or 0
    self.ts_adjust_seconds = self.ts_adjust / 0x100000000

    if config.mx_data_rate and config.mx_data_rate == 0 then
        self.data_processor_info = {}
        self.data_processor = {}
        self.do_process_data = function()
                for _, dpi in ipairs(self.data_processor_info) do
                    dpi.func(dpi.signo)
                end
            end
    else
        self.do_process_data = function() return end
    end

    for i = 0, C.MAX_SUBSCRIBED_CHANNELS - 1 do
        subscribed_channels[i].in_use = 0
        subscribed_channels[i].sample_buffer.bytes = C.RB_BUF_SIZE
        subscribed_channels[i].sample_buffer.iov[0].iov_base =
            subscribed_channels[i].sample_buffer.sink_hdr
        subscribed_channels[i].sample_buffer.iov[0].iov_len =
            ffi.sizeof(subscribed_channels[i].sample_buffer.sink_hdr[0])
    end

    local s = net.socket{ip = ip, port = '7411'}
    local ok, socket = pcall(s.connect, s)
    if not ok then
        self:error(string.format("Unable to connect to port 7411 at %s", ip))
        return
    else
        self:dbg_msg(string.format(
            "Connected to data stream on host %s port 7411", ip))
    end
    self.socket = socket
    self.subscribe_cb = {}
    self.end_ts = 0
    if read_rate then
        self.sbuf =
            StreamBuffer:new(self.socket, read_buffer_size, read_rate)
    else
        self.sbuf =
            StreamBuffer:new(self.socket, read_buffer_size)
    end
    self.sbuf.read_cb = function(loop)
        return self:process_data()
    end

    self.sbuf.stop_cb = function()
        self:dbg_msg(string.format("Stream connection to %s closing", ip))
        if self.connection_monitor_timer then
            self.connection_monitor_timer:stop(loop)
            self.connection_monitor_timer:clear_pending(loop)
        end
        self.sbuf = nil
    end

    -- KLUDGE FOR QUANTUMX NEW FIRMWARE CONNECTION SHUTDOWN BADNESS!!
    self.sbuf.shutdown = function()
            self.sbuf:close()
            self:dbg_msg(string.format("Stream connection to %s closing", ip))
            if self.connection_monitor_timer then
                self.connection_monitor_timer:stop(loop)
                self.connection_monitor_timer:clear_pending(loop)
            end
            self.sbuf = nil
        end

    -- We delay the start of this for a while because if there are a large
    -- number of MX devices being controlled by this particular CX23 unit, the
    -- boot time CPU load can be enough to cause this to catch.  So, we delay
    -- the first check long enough to account for all of that possible initial
    -- CPU load!
    self.connection_monitor_timer = ev.Timer.new(function(l, t, re)
            if not self.got_heartbeat and
                not self.firmware_update_in_progress then
                local sn = sys_notify:new('MX Device Error')
                sn:critical_error(
                    'Heartbeats on stream connection have stopped', false)
                self:dbg_msg("Heartbeats on stream connection stopped")
                self.log:log(self.log.NOTICE,
                    "Stream heartbeats stopped - exiting proxy!")
                os.exit(-100)
            end
            self.got_heartbeat = false
        end, 60, 5)

    self.connection_monitor_timer:start(loop)
    self.sbuf:start(loop)

    return self
end

return Stream
