#!/usr/bin/lua

require('lj_include'){
    'unistd',
    'sys/types',
    'signal',
    'net/if',
}

local ffi = require 'ffi'
local C = ffi.C
local ev = require 'ev'
local dbus = require 'dbus'
local struct = require 'struct'
require 'middleclass'
local cjson = require 'cjson'
local config = require 'system_config'

local net_ffi = require 'net_ffi'

local log = require 'log'
log:open('mx-manager', log.USER)

-- Deal with command line options
local update_rate, ipv6 = ...

ipv6 = (ipv6 and true) or false
update_rate = tonumber(update_rate) or 15
if update_rate < 15 then
    update_rate = 15
end


-- Figure out which MX firmware version the MX device *should* be using.
-- FIXME:  Take out support for old ".qxi" files sometime in the future
local expected_mx_fw_ver
local mx_fw_file
local f = io.popen("ls -lt /usr/share/quantum-firmware/qxall* 2>/dev/null")
if f then
    for line in f:lines() do
        local e = line:match('qxall_(.+)%.qxi$')
        if e then
            expected_mx_fw_ver = e:gsub('-', '.')
            mx_fw_file = line:match('(/usr/share.+)')
            break
        end
    end
    f:close()
else
    local f = io.popen("ls -lt /usr/share/quantum-firmware/*.hbg 2>/dev/null")
    if f then
        for line in f:lines() do
            local e = line:match('firmware/(%d+-%d+-%d+-%d+)')
            if e then
                expected_mx_fw_ver = e:gsub('-', '.')
                mx_fw_file = line:match('(/usr/share.+)')
                break
            end
        end
        f:close()
    end
end

-- Wait until any of the internal network carriers are up
local nf = require 'net_functions'
nf.wait_carrier('any', 'ignore', {'eth1', 'eth2'})

-- Need services provided by master
local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()
local master = assert(bus:auto_proxy('com.somat.armada.Master',
                                     '/com/somat/armada/Master'))
local network = assert(bus:auto_proxy_wait('com.somat.armada.Network',
                                     '/com/somat/armada/Network'))

ffi.cdef[[
enum {
    CRC_MASK = 0xff,
};
]]

-- Function used to retrieve the ZCIP assigned IPv4 address of the internal
-- network interface
local function get_zcip_addr()
    local addrs = net_ffi.getifaddrs('ethint')
    for _, addr in ipairs(addrs) do
        if addr:find('^169%.254%.') then
            return addr
        end
    end
end

-- Get the ZCIP of the internal network interface
local my_zcip_addr

local DEFAULT_QUANTUMX_DOMAIN_ADDRESS_V6 = "ff02::1"
local DEFAULT_QUANTUMX_DOMAIN_ADDRESS = "239.255.77.76"
local MX_DISCOVERY_MULTICAST_PORT = "31416"

local htonl = ffi.abi('le') and bit.bswap or bit.tobit

local htonll_union =
    ffi.typeof('union { unsigned long long u64; unsigned long u32[2]; }')

local function htonll(x)
    if ffi.abi('le') then
        local v = htonll_union(x)
        v.u32[0], v.u32[1] = bit.bswap(v.u32[1]), bit.bswap(v.u32[0])
        return v.u64
    end
    return x
end

local function htons(x)
    if ffi.abi('le') then
        return bit.rshift(htonl(x), 16)
    end
    return htonl(x)
end
local ntohl = htonl
local ntohs = htons
local ntohll = htonll

local net = class('net')

-- Function used to get the IPv4 address of an interface
function net:get_ipv4_address(fd)
    local ifreq = ffi.new('struct ifreq[1]')
    ffi.copy(ifreq[0].ifr_ifrn.ifrn_name, self.iface_name)
    C.ioctl(fd, C.SIOCGIFADDR, ifreq)
    local addr = ffi.new('struct sockaddr_in')
    ffi.copy(addr, ifreq[0].ifr_ifru.ifru_addr, ffi.sizeof(addr))

    return addr.sin_addr
end

function net:ipv4_socket_init(port, fd)
    -- Need the IPv4 address of the specified interface
    self.myaddr = self:get_ipv4_address(fd)
    if self.myaddr.s_addr == 0 then
        return false
    end

    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    hints.ai_family = self.af
    local hst = self.mhst
    local r = C.getaddrinfo(hst, nil, hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return false
    end
    r = res[0]

    -- Extract the address out of the returned struct
    local a = ffi.new('struct sockaddr_in')
    ffi.copy(a, r.ai_addr, ffi.sizeof(a))

    -- Done with address info
    C.freeaddrinfo(res[0])

    -- Set up for setting socket options
    local mr = ffi.new('struct ip_mreq')
    mr.imr_multiaddr.s_addr = a.sin_addr.s_addr
    mr.imr_interface.s_addr = self.myaddr.s_addr

    -- Limit multicasting to interface with this address
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_MULTICAST_IF, mr, ffi.sizeof(mr))

    -- Join the multicast group
    C.setsockopt(fd, C.IPPROTO_IP, C.IP_ADD_MEMBERSHIP, mr, ffi.sizeof(mr))

    return true
end

function net:ipv6_socket_init(port, fd)

    -- Only allow multicast on the interface we are bound to
    C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_MULTICAST_IF, self.idx,
        ffi.sizeof(self.idx))

    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    hints.ai_family = self.af
    local r = C.getaddrinfo(self.mhst, nil, hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return false
    end
    r = res[0]

    -- Extract the address out of the returned struct
    local fam, prt, flow, adr = struct.unpack('H H I4 c16',
        ffi.string(r.ai_addr, 24))

    -- Done with address info
    C.freeaddrinfo(res[0])

    -- Join the multicast group
    local mr = ffi.new('struct ipv6_mreq', {
            interface = idx,
        })
    for i = 0, 15 do
            mr.multiaddr.__in6_u.__u6_addr8[i] = string.byte(adr, i + 1)
    end
    local res = C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_JOIN_GROUP, mr,
        ffi.sizeof(mr))

    -- Set multicast hop limit
    local ttl = ffi.new('int[1]', 255)
    C.setsockopt(fd, C.IPPROTO_IPV6, C.IPV6_MULTICAST_HOPS, ttl,
        ffi.sizeof(ttl))

    return true
end

function net:socket_init(port)
    local res = ffi.new('struct addrinfo *[1]')
    local hints = ffi.new('struct addrinfo')
    hints.ai_socktype = C.SOCK_DGRAM
    if tonumber(port) == 0 then
        hints.ai_flags = C.AI_NUMERICHOST + C.AI_NUMERICSERV
    else
        hints.ai_flags = C.AI_PASSIVE + C.AI_NUMERICHOST + C.AI_NUMERICSERV
    end
    hints.ai_family = self.af
    local r = C.getaddrinfo(nil, tostring(port), hints, res)
    if r ~= 0 then
        log:log(log.ERR,
            string.format("getaddrinfo error: %s", C.gai_strerror(r)))
        return -1
    end
    r = res[0]

    -- Create a socket
    local fd = C.socket(r.ai_family, r.ai_socktype, r.ai_protocol)
    if fd < 0 then
        local e = ffi.errno()
        log:log(log.ERR,
            string.format("Can not create socket %d", e))
        return -1
    end

    -- Bind to the specific interface so we aren't sending everywhere
    local stat = C.setsockopt(fd, C.SOL_SOCKET, C.SO_BINDTODEVICE,
        self.iface_name, self.iface_name:len())
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR,
            string.format("Can not bind socket to iface %d", e))
        C.close(fd)
        return -1
    end

    -- Bind the socket
    stat = C.bind(fd, r.ai_addr, r.ai_addrlen)
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR, string.format("Can not bind socket %d", e))
        C.close(fd)
        return -1
    end

    -- Done with address info
    C.freeaddrinfo(res[0])

    if tonumber(port) == 0 then
        return fd
    end

    -- Allow immediate reuse of the addrees
    local enable = ffi.new('int[1]', {1})
    stat = C.setsockopt(fd, C.SOL_SOCKET, C.SO_REUSEADDR,
        enable, ffi.sizeof(enable))
    if stat < 0 then
        local e = ffi.errno()
        log:log(log.ERR, string.format("Can not set REUSEADDR %d", e))
        C.close(fd)
        return -1
    end

    -- Do remainder of V4/V6 initialization
    if self.af == C.AF_INET6 then
        if self:ipv6_socket_init(port, fd) == false then
            C.close(fd)
            fd = -1
        end
    elseif self:ipv4_socket_init(port, fd) == false then
        C.close(fd)
        fd = -1
    end

    return fd
end

function net:receive(sock, buf)
    local ret = C.recvfrom(sock, buf.buf + buf.offset, buf.size - buf.offset,
        C.MSG_DONTWAIT, nil, nil)
    if ret < 0 then
        local errno = ffi.errno()
        if errno == C.EAGAIN or errno == C.EINTR then
            return 0
        end
        log:log(log.ERR, string.format("recvfrom fail: %s",
            ffi.string(C.strerror(errno))))
        return
    elseif ret == 0 then
        return
    end
    buf.offset = buf.offset + ret
    return ret
end

function net:shutdown()
    C.close(self.multicast_fd)
end

local function make_mx_hostname(name, ip)
    local a, b = ip:gsub('%%ethint', ''):match('.-::.-:.-:(.-):(.-)$')
    a = tonumber('0x' .. a) % 256
    b = tonumber('0x' .. b)
    if not name:find('B$') then name = name .. 'B' end
    return string.format("%s-%02x%04x", name, a, b):lower()
end

-- Get the internal network hostname
local net_db = require 'network_db'
local ndb = net_db:new()
local my_internal_hostname = ndb:get_inside_network_hostname()

local crc_8_table = ffi.new('unsigned char[256]', {
    0x00, 0x5E, 0xBC, 0xE2, 0x61, 0x3F, 0xDD, 0x83,
    0xC2, 0x9C, 0x7E, 0x20, 0xA3, 0xFD, 0x1F, 0x41,
    0x9D, 0xC3, 0x21, 0x7F, 0xFC, 0xA2, 0x40, 0x1E,
    0x5F, 0x01, 0xE3, 0xBD, 0x3E, 0x60, 0x82, 0xDC,
    0x23, 0x7D, 0x9F, 0xC1, 0x42, 0x1C, 0xFE, 0xA0,
    0xE1, 0xBF, 0x5D, 0x03, 0x80, 0xDE, 0x3C, 0x62,
    0xBE, 0xE0, 0x02, 0x5C, 0xDF, 0x81, 0x63, 0x3D,
    0x7C, 0x22, 0xC0, 0x9E, 0x1D, 0x43, 0xA1, 0xFF,
    0x46, 0x18, 0xFA, 0xA4, 0x27, 0x79, 0x9B, 0xC5,
    0x84, 0xDA, 0x38, 0x66, 0xE5, 0xBB, 0x59, 0x07,
    0xDB, 0x85, 0x67, 0x39, 0xBA, 0xE4, 0x06, 0x58,
    0x19, 0x47, 0xA5, 0xFB, 0x78, 0x26, 0xC4, 0x9A,
    0x65, 0x3B, 0xD9, 0x87, 0x04, 0x5A, 0xB8, 0xE6,
    0xA7, 0xF9, 0x1B, 0x45, 0xC6, 0x98, 0x7A, 0x24,
    0xF8, 0xA6, 0x44, 0x1A, 0x99, 0xC7, 0x25, 0x7B,
    0x3A, 0x64, 0x86, 0xD8, 0x5B, 0x05, 0xE7, 0xB9,
    0x8C, 0xD2, 0x30, 0x6E, 0xED, 0xB3, 0x51, 0x0F,
    0x4E, 0x10, 0xF2, 0xAC, 0x2F, 0x71, 0x93, 0xCD,
    0x11, 0x4F, 0xAD, 0xF3, 0x70, 0x2E, 0xCC, 0x92,
    0xD3, 0x8D, 0x6F, 0x31, 0xB2, 0xEC, 0x0E, 0x50,
    0xAF, 0xF1, 0x13, 0x4D, 0xCE, 0x90, 0x72, 0x2C,
    0x6D, 0x33, 0xD1, 0x8F, 0x0C, 0x52, 0xB0, 0xEE,
    0x32, 0x6C, 0x8E, 0xD0, 0x53, 0x0D, 0xEF, 0xB1,
    0xF0, 0xAE, 0x4C, 0x12, 0x91, 0xCF, 0x2D, 0x73,
    0xCA, 0x94, 0x76, 0x28, 0xAB, 0xF5, 0x17, 0x49,
    0x08, 0x56, 0xB4, 0xEA, 0x69, 0x37, 0xD5, 0x8B,
    0x57, 0x09, 0xEB, 0xB5, 0x36, 0x68, 0x8A, 0xD4,
    0x95, 0xCB, 0x29, 0x77, 0xF4, 0xAA, 0x48, 0x16,
    0xE9, 0xB7, 0x55, 0x0B, 0x88, 0xD6, 0x34, 0x6A,
    0x2B, 0x75, 0x97, 0xC9, 0x4A, 0x14, 0xF6, 0xA8,
    0x74, 0x2A, 0xC8, 0x96, 0x15, 0x4B, 0xA9, 0xF7,
    0xB6, 0xE8, 0x0A, 0x54, 0xD7, 0x89, 0x6B, 0x35
})

local function get_crc_byte(_crc, _byte)
    local idx = bit.bxor(bit.band(_crc, C.CRC_MASK), _byte)
    _crc = bit.bxor(bit.band(bit.rshift(_crc, 8), C.CRC_MASK), crc_8_table[idx])
    return(_crc)
end

local function get_crc_8(p, len)
    local crc = ffi.new('unsigned int', 0)
    while len > 0 do
        crc = get_crc_byte(tonumber(crc), tonumber(p[0]))
        p = p + 1
        len = len - 1
    end
    return crc
end

local unclaimed_count = {}

function net:proxy_running(ip)
    self.proxy_pid = self.proxy_pid or {}
    if not self.proxy_pid[ip] then
        return
    elseif C.kill(self.proxy_pid[ip], 0) < 0 then
        self.proxy_pid[ip] = nil
        return
    end
    return true
end

local function get_proxy(ip)
    local fh = io.popen(string.format('ps -o pid,args | grep %s', ip))
    if not fh then return end
    for line in fh:lines() do
        local pid, _ip = line:match('(%d+) luajit .+proxy.lua (.+)')
        if pid and _ip == ip then
            fh:close()
            return pid
        end
    end
    fh:close()
end

local function start_proxy(ip)
    return(get_proxy(ip) or
        master:Spawn('/usr/bin/luajit',
        {
            'luajit',
            '/usr/share/lua/5.1/mx_proxy.lua',
            ip,
            'stderr_level:none',
            'stdout_level:none',
        }))
end

function net:start_proxy(name, ip)

    -- Generate this MX's hostname
    local owner = make_mx_hostname(name, ip)

    if owner then
        -- Check to see if this MX is in my local database.  If
        -- this MX device is in my local table, then I own the
        -- device!
        local rrs = network:QueryLocalRR(owner, 'A', '*', '*')
        if rrs and #rrs > 0 then
            log:debug(string.format("I am owner of %s", owner))
            -- If the MX has been rebooted, it may have a
            -- different ZCIP address.  If so, reset the
            -- database entry to reflect that change.
            local rr = rrs[1]
            if rr[3] ~= ip then
                log:debug(string.format("MX %s has new ip address %s -> %s",
                    owner, rr[3], ip))
                network:ModifyRRs({{owner, 'A', rr[3]}}, {{owner, 'A', ip}})
            end
            -- Is in my local table, check to see if the proxy
            -- for it is running and if not start it.
            if not self:proxy_running(ip) then
                log:debug(string.format(
                    "Starting MX proxy for %s at address %s", owner, ip))
                self.proxy_pid[ip] = start_proxy(ip)
            end

        -- If this MX is not in the master table then it is
        -- not "owned" by a CX23 device.  Based on the MX
        -- device's serial number and the number of CX23
        -- devices in the system, figure out who should own
        -- the device.
        else
            local mrrs = network:QueryRR(owner, 'A', '*', '*')

            -- Device is NOT in the master table
            if not mrrs or #mrrs == 0 then
                log:debug(string.format("my_internal_hostname=%s",
                    my_internal_hostname))
                -- Get a list of ALL of the CX23's in the system
                -- FIXME:  Do these need to be sorted?
                local cx23s = network:QueryRR('*', 'A', '*', '*')
                -- Only do this if there are CX23's in the
                -- master database.
                if cx23s and #cx23s > 0 then
                    local cx23_list = {}
                    for _i, cx23 in ipairs(cx23s) do
                        if cx23[1]:find('^cx23r-') then
                            log:debug(string.format("%s: %s", _i, cx23[1]))
                            cx23_list[#cx23_list + 1] = cx23[1]
                        end
                    end

                    -- Create a hash based on the
                    -- Quantums hostname
                    local p = ffi.new('char[?]', owner:len(), owner)
                    local selected = (get_crc_8(p, owner:len()) %
                        #cx23_list) + 1
                    log:debug(string.format("selected owner of %s is %s",
                        owner, cx23_list[selected]))

                    -- If the CX23 that should be controlling
                    -- this unit is me then add the RR to the
                    -- database and start the proxy.
                    if cx23_list[selected] ==
                        my_internal_hostname then
                        log:debug(string.format("Selected owner of %s is me",
                            owner))
                        network:AddRR(owner, 'A', ip)
                        if not self:proxy_running(ip) then
                            log:debug(string.format(
                                "Starting MX proxy for %s at %s", owner, ip))
                            self.proxy_pid[ip] = start_proxy(ip)
                        end
                    else
                        -- If the unit that should be
                        -- controlling this MX device has not
                        -- started doing so after a few
                        -- discovery passes, claim it for
                        -- myself.
                        unclaimed_count[owner] = unclaimed_count[owner] or 0
                        unclaimed_count[owner] = unclaimed_count[owner] + 1

                        -- If the MX device has not been
                        -- claimed for at least 4 discovery
                        -- passes by the CX23 that should have
                        -- claimed it, claim it for myself!
                        if unclaimed_count[owner] > 4 then
                            network:AddRR(owner, 'A', ip)
                            if not self:proxy_running(ip) then
                                log:debug(string.format(
                                    "Starting proxy for "
                                    .. "unclaimed %s at ip "
                                    .. "address %s", owner, ip))
                                self.proxy_pid[ip] = start_proxy(ip)
                            end
                        end
                    end
                end
            -- Device is in the master table, clear unclaimed
            -- count for the device
            else
                unclaimed_count[owner] = nil
            end
        end
    end
end

function net:process_announcement(json)
    if json.method and json.method == 'announce' then
        local module_name
        local ip_addr
        local ipv6_addr
        local fwver

        -- Set of services that have to be available before we attempt
        -- to run a proxy for a particular quantum device.
        local smap = {
            daqStream = true,
            hbmProtocol = true,
            http = true,
            jetd = true,
            jetws = true,
            ssh = true,
        }

        -- Figure out which services are available
        if json.params.services then
            for _, v in ipairs(json.params.services) do
                if v.port and v.port ~= 0 then
                    smap[v] = nil
                end
            end
        end

        -- Only attempt start proxies on units that have all expected
        -- services up and going.
        if #smap == 0 then

            -- Prefer the ZCIP address otherwise default to other IPv4
            -- address if it exists
            if json.params.interface and json.params.interface.ipv4 then
                for _, v in ipairs(json.params.interface.ipv4) do
                    if v.address and v.address:match('^169%.254%.') then
                        ip_addr = v.address
                        break
                    end
                end
                ip_addr = ip_addr or
                    json.params.interface.ipv4[1].address
            elseif json.params.netSettings.interface.ipv4 then
                for _, v in ipairs(
                    json.params.netSettings.interface.ipv4) do
                    if v.address and v.address:match('^169%.254%.') then
                        ip_addr = v.address
                        break
                    end
                end
                ip_addr = ip_addr or
                    json.params.netSettings.interface.ipv4[1].address
            end
            if json.params.interface and json.params.interface.ipv6 and
                json.params.interface.ipv6[1].address then
                ipv6_addr = json.params.interface.ipv6[1].address
            elseif json.params.netSettings.interface.ipv6 and
                json.params.netSettings.interface.ipv6[1].address then
                ipv6_addr =
                    json.params.netSettings.interface.ipv6[1].address
            end
            if json.params.device.type then
                module_name = json.params.device.type
            end
            if json.params.device.firmware_version then
                fwver = json.params.device.firmware_version
            elseif json.params.device.firmwareVersion then
                fwver = json.params.device.firmwareVersion
            end
            if fwver and fwver:find('^3.') and module_name and
                not module_name:find('%d+B') then
                module_name = module_name .. 'B'
            end
            if module_name and ipv6_addr then
                ip_addr = ipv6_addr .. '%ethint'
                log:debug(string.format(
                    "QuantumX device %s found at address %s",
                    module_name, ip_addr))
                    self:start_proxy(module_name, ip_addr)
                self.mx_fw_ver = self.mx_fw_ver or {}
                if not self.mx_fw_ver[ip_addr] then
                    self.mx_fw_ver[ip_addr] = fwver
                end
            else
                log:log(log.ERR, string.format(
                    "Device %s does not have an IPv6 address", module_name))
            end
        else
            log:debug("QuantumX device but not all services were available")
        end
    end
end

function net:initialize(ipv6, iface_name)
    self.iface_name = iface_name or 'ethint'
    self.idx = ffi.new('int[1]')
    self.idx[0] = C.if_nametoindex(self.iface_name)

    if ipv6 then
        self.mhst = DEFAULT_QUANTUMX_DOMAIN_ADDRESS_V6
        self.af = C.AF_INET6
        self.multicast_fd = self:socket_init(MX_DISCOVERY_MULTICAST_PORT)
        if self.multicast_fd < 0 then
            log:log(log.ERR, "Unable to create multicast socket")
            return
        end
    else
        self.mhst = DEFAULT_QUANTUMX_DOMAIN_ADDRESS
        self.af = C.AF_INET
        self.multicast_fd = self:socket_init(MX_DISCOVERY_MULTICAST_PORT)
        if self.multicast_fd < 0 then
            log:log(log.ERR, "Unable to create multicast socket")
            return
        end
    end

    self.mcast_ibuf = {}
    self.mcast_ibuf.size = 65536
    self.mcast_ibuf.buf = ffi.new('char[?]', self.mcast_ibuf.size)
    self.mcast_ibuf.offset = 0

    self.mcast_rx_handler = ev.IO.new(function(loop, io, revents)
            local num_rx = self:receive(self.multicast_fd, self.mcast_ibuf)
            while self.mcast_ibuf.offset > 0 do
                local jstr =
                    ffi.string(self.mcast_ibuf.buf, self.mcast_ibuf.offset)
                local s, e = jstr:find('}{')
                if s and e then
                    jstr = string.sub(jstr, 1, s)
                    self.mcast_ibuf.offset = self.mcast_ibuf.offset - e
                    ffi.copy(self.mcast_ibuf.buf, self.mcast_ibuf.buf + e,
                        self.mcast_ibuf.offset)
                else
                    self.mcast_ibuf.offset = 0
                end
                local no_error, decode = pcall(cjson.decode, jstr)
                if no_error then
                    if decode.method and decode.method == 'announce' then
                        self:process_announcement(decode)
                    end
                else
                    log:log(log.WARNING,
                        string.format("Can not decode JSON\n%s", jstr))
                end
            end
        end, self.multicast_fd, ev.READ)

    self.mcast_rx_handler:start(ev.Loop.default)

    return self
end

local n

ev.Timer.new(function(l, t, e)
        my_zcip_addr = get_zcip_addr()
        if my_zcip_addr then
            n = net:new(ipv6)
            if n then
                t:stop(l)
            end
        end
    end, 1, 1):start(ev.Loop.default)

-- Run the event processing loop
ev.Loop.default:loop()
