#!/usr/bin/lua

--! Network(s) configuration
--!
--! This script does all of the network configuration
--!
--! \file
local sqlite3 = require 'lsqlite3'
local net = require 'net'
local cjson = require 'cjson'
local bit = require 'bit'
local nf = require 'net_functions'
local ffi = require 'ffi'
local C = ffi.C
local ev = require 'ev'

local log = require 'log'
log:open('network-config', log.USER)

ffi.cdef[[
int sleep(int);
]]

local config = require 'system_config'

local ext_if = config.network.ext_if
local int_if = config.network.int_if

-- Function used to get information from the sql database
local function sql_get(db, sql)
    local vm = db:prepare(sql)
    if not vm then return end
    local result = {}
    while vm:step() == sqlite3.ROW do
        result[#result + 1] = vm:get_uvalues()
    end
    vm:finalize()
    if #result > 0 then
        return result, true
    end
    return
end

--[=[
Function used to get external network interface configuration from a
database.  NOTE, this code assumes the database is of the "tracked" variety.

This is what the JSON from the database will look like
    extif_config = [[
        {
            "id":"Bullshit",
            "type":"Dynamic",
            "network":"Outside",
            "ip_address":"192.168.1.1/24",
            "gateway":null,
            "primary_dns":null,
            "secondary_dns":null,
            "subset_mask":null,
            "broadcast":"192.168.1.255"
        }]]
    db_hostname = [[
        {
            "id":"Bullshit",
            "hostname":"cx-23-106",
        }
    ]]
]=]
local function get_outside_netconfig_from_db(database)
    local db = sqlite3.open(database)
    if not db then
        error("Unable to open database to get network configuration\n")
        return
    end

    -- Get the most recent hostname
    -- NOTE, make up default hostname based on existing config
    local _hostname = string.format('{"hostname":"%s"}', config.hostname)
    local r, ok = sql_get(db, [[
        SELECT content
        FROM tsd_users
        WHERE db_name = 'networks' AND id = 'hostname' AND deleted = ''
        ORDER BY added DESC LIMIT 1;
    ]])

    if ok then
        _hostname = r[1]
    end

    local t = cjson.decode(_hostname)
    if t.hostname and type(t.hostname) == 'string' then
        config.hostname = t.hostname
    end

    -- Get the most recent active network configuration
    local r, ok = sql_get(db, [[
        SELECT content
        FROM tsd_users
        WHERE db_name = 'networks' AND id != 'hostname' AND deleted = ''
            AND content LIKE ('%"network":"outside"%')
            AND content LIKE ('%"active":true%')
        ORDER BY added DESC LIMIT 1;
    ]])

    local extif_config = ok and r[1]
    if not extif_config then
        log:debug(
            "Unable to find ext network interface configuration in database")
        db:close()
        return
    end

    db:close()
    -- Get the two items above from the database

    local t = cjson.decode(extif_config)
    if t.type and type(t.type) == 'string' then
        config.outside_network = {}
        if t.type:lower() == 'static' then
            local bits
            config.outside_network.mode = 'static'
            if t.ip_address and type(t.ip_address) == 'string' then
                config.outside_network.address, bits =
                    t.ip_address:match('(.+)/(%d+)')
                if not config.outside_network.address and not bits then
                    config.outside_network.address = t.ip_address
                end
            end
            -- If subnet mask is specified it overrides /xx syntax above
            if t.subnet_mask and type(t.subnet_mask) == 'string' then
                bits = nil
                config.outside_network.netmask = t.subnet_mask
            end
            if t.broadcast and type(t.broadcast) == 'string' then
                config.outside_network.broadcast = t.broadcast
            end
            if bits then
                local m = bit.lshift(0xffffffff, 32 - bits)
                local mask = ''
                for i = 24, 0, -8 do
                    mask = mask ..  string.format("%d",
                        tonumber(bit.band(bit.rshift(m, i), 0xff)))
                    if i ~= 0 then
                        mask = mask .. '.'
                    end
                end
                config.outside_network.netmask = mask
            else
                bits = 0
                local c = {t.subnet_mask:match('^(%d-)%.(%d-)%.(%d-)%.(%d+)$')}
                if #c == 4 then
                    local mask = 0
                    local bits = 0
                    for _, v in ipairs(c) do
                        for i = 0, 7 do
                            if bit.band(v, 0x80) == 0x80 then
                                bits = bits + 1
                            end
                            v = bit.lshift(v, 1)
                        end
                    end
                end
            end
            config.outside_network.dns = {}
            for _, dns in ipairs{'primary_dns', 'secondary_dns'} do
                if t[dns] and type(t[dns]) == 'string' then
                    config.outside_network.dns[#config.outside_network.dns + 1]
                        = t[dns]
                end
            end
            if t.gateway and type(t.gateway) == 'string' then
                config.outside_network.gateway = t.gateway
            end
        else
            config.outside_network.mode = 'dynamic'
        end
    else
        -- Network type is present but not set
        config.outside_network = nil
    end
end

local function get_inside_netconfig_from_db(database)
    local db = sqlite3.open(database)
    if not db then
        error("Unable to open database to get network configuration\n")
        return
    end

    -- Get the most recent active network configuration
    local r, ok = sql_get(db, [[
        SELECT content
        FROM tsd_users
        WHERE db_name = 'networks' AND id != 'hostname' AND deleted = ''
            AND content LIKE ('%"network":"inside"%')
            AND content LIKE ('%"active":true%')
        ORDER BY added DESC LIMIT 1;
    ]])

    local intif_config = ok and r[1]
    if not intif_config then
        log:debug(
            "Unable to find int network interface configuration in database")
        db:close()
        return
    end

    db:close()
    -- Get the two items above from the database

    local t = cjson.decode(intif_config)
    if t.type and type(t.type) == 'string' then
        config.quantumx = config.quantumx or {}
        config.quantumx.network = config.quantumx.network or {}
        config.quantumx.network.interface = int_if
        if t.type:lower() == 'static' then
            local bits
            config.quantumx.network.mode = 'static'
            if t.ip_address and type(t.ip_address) == 'string' then
                config.quantumx.network.address, bits =
                    t.ip_address:match('(.+)/(%d+)')
                if not config.quantumx.network.address and not bits then
                    config.quantumx.network.address = t.ip_address
                end
            end
            -- If subnet mask is specified it overrides /xx syntax above
            if t.subnet_mask and type(t.subnet_mask) == 'string' then
                bits = nil
                config.quantumx.network.netmask = t.subnet_mask
            end
            if t.broadcast and type(t.broadcast) == 'string' then
                config.quantumx.network.broadcast = t.broadcast
            end
            if bits then
                local m = bit.lshift(0xffffffff, 32 - bits)
                local mask = ''
                for i = 24, 0, -8 do
                    mask = mask ..  string.format("%d",
                        tonumber(bit.band(bit.rshift(m, i), 0xff)))
                    if i ~= 0 then
                        mask = mask .. '.'
                    end
                end
                config.quantumx.network.netmask = mask
            end
            if t.gateway and type(t.gateway) == 'string' then
                config.quantumx.network.gateway = t.gateway
            end
        else
            -- Dynamic configuration is not allowed on inside network
            config.quantum.network = nil
        end
    else
        -- Network type is present but not set
        config.quantumx.network = nil
    end
end

-- TODO:  Something appropriate needs to be done here for the armadized eDAQ
-- Get external network interface configuration from database if it's available
if config.target == 'armada' then
    local old_config = config.outside_network
    get_outside_netconfig_from_db('/var/lib/armada/setup.db')
    if not config.outside_network then
        config.outside_network = old_config
    end
    old_config = config.quantumx.network
    get_inside_netconfig_from_db('/var/lib/armada/setup.db')
    if not config.quantumx.network then
        config.quantumx.network = old_config
    end
else
    error("TODO:  Need to do something here for armadized eDAQ\n")
end


-- Set the hostname for the outside netork
os.execute('hostname '.. config.hostname)


-- Bring up the external network interface
os.execute("ip link set up dev " .. ext_if)

function setup_internal_network_bridge()
    if nf.file_exists("/sys/class/net/ethint") then
        -- This setup has been done before.  Do only what is needed, if
        -- anything, to get networking going.
    else
        -- Create the bridge device
        os.execute("brctl addbr ethint")

        -- Create the TUN device that makes everything work when there is no
        -- physical network connection on any of the inside networks.
        os.execute("/usr/bin/tun-util &")
        os.execute("ip link set up dev tun0")

        -- Bring up ALL of the inside network devices, attach them to the
        -- bridge, and set up ebtables entries to cause PTPv2 packets to be
        -- routed normally rather be sent to the bridge.  This is needed
        -- because hardware timestamps do NOT propagate through the bridge.
        -- This allows the PTPv2 daemon to use the network ports directly.
        for _, dev in ipairs{'eth1', 'eth2'} do
            if nf.file_exists("/sys/class/net/" .. dev .. "/carrier") then
                os.execute("ip link set up dev " .. dev)
                os.execute("brctl addif ethint " .. dev)
                os.execute("ebtables -t broute -A BROUTING -p ipv6 -i " ..
                    dev .. " " ..
                    "--ip6-protocol 17 --ip6-destination-port 319:320 " ..
                    "--ip6-source-port 319:320 -j DROP")
            end
        end

        -- Finally, add the TUN device to the bridge.
        os.execute("brctl addif ethint tun0")

        -- The bridge must be in promiscuous mode for IPv4 multicast to work
        -- through it.  This is used for QuantunmX discovery!
        os.execute("/sbin/ifconfig ethint promisc")
    end
end

if config.target == 'armada' then
    setup_internal_network_bridge()
end

-- Setup the IPv6 address of the inside network based on the MAC address of
-- the inside network interface
local int_prefix = "fda9:aac8:be27:0"
local int_index, int_mac, int_ll_ip
do
    local ipfh = io.popen("ip -o link")
    for line in ipfh:lines() do
        local index, mac = 
            line:match("^(%d+): " .. int_if .. ": .*link/ether ([^ ]+)")
        if mac then
            int_index = index
            int_mac = mac
        end
    end
    ipfh:close()
end
assert(int_mac)
int_mac = int_mac:gsub(":", "")
local ipstub = 
    int_prefix .. ":" .. 
    int_mac:sub(1, 4) .. ":" ..
    int_mac:sub(5, 8) .. ":" ..
    int_mac:sub(9, 12) .. ":"
os.execute("ip link set up dev " .. int_if)
local int_ip = ipstub .. '0'
os.execute("ip addr add " .. int_ip .. "/64 dev " .. int_if)
repeat
    local tentative = false
    local ipfh = io.popen("ip -o addr")
    for line in ipfh:lines() do
        local llip =
            line:match("^%d+: " .. int_if .. ".*inet6 ([^ ]+)/64 scope link")
        if llip then
            int_ll_ip = llip .. '%' .. int_if
        end
        if line:match(int_if .. ".*inet6 " .. int_ip ..
                      "/64 scope global tentative") then
            tentative = true
        end
    end
    ipfh:close()
    C.sleep(1)
until int_ll_ip and not tentative

local json = {
    internal_hostname = config.internal_hostname,
    internal_mac = int_mac,
    internal_if_index = int_index,
    internal_ip = int_ip,
    internal_ll_ip = int_ll_ip,
}
local jsonf = assert(io.open('/var/run/network_config.json.tmp', 'w'))
assert(jsonf:write(cjson.encode(json)))
jsonf:close()
assert(os.rename('/var/run/network_config.json.tmp',
                 '/var/run/network_config.json'))

-- If this is an CX23, configure the external network
if ext_if and config.target ~= 'edaq' then
    if not config.outside_network then
        -- If no outside network configuration exists, default the IP address
        -- to something like what we do on an eDAQ.
        -- FIXME:  This really should be replaced with uPNP auto IP stuff.
        os.execute('ip addr add 192.168.100.101/24 broadcast ' .. 
            '192.168.100.255 dev ' .. ext_if .. ' 2>/dev/null')
    elseif config.outside_network.mode and
        config.outside_network.mode ~= 'dynamic' and
        config.outside_network.address and
        config.outside_network.netmask and
        config.outside_network.broadcast then
        local a, b, c, d = string.match(config.outside_network.netmask,
            '(%d+).(%d+).(%d+).(%d+)')
        local m = bit.lshift(a or 0, 24) + bit.lshift(b or 0, 16) +
            bit.lshift(c or 0, 8) + (d or 0)
        local bits = 32
        while bit.band(m, 1) == 0 do
            bits = bits - 1
            m = bit.rshift(m, 1)
        end
        os.execute('ip addr add ' .. config.outside_network.address .. '/' ..
            bits .. ' broadcast ' .. config.outside_network.broadcast ..
            ' dev ' .. ext_if .. ' 2>/dev/null')

        -- Gateway, remove existing first, then set gateway
        local cmd = string.format(
            "/sbin/route del default gw 0.0.0.0 dev %s 2>/dev/null", ext_if)
        while os.execute(cmd) == 0 do end
        if config.outside_network.gateway then
            local cmd = string.format(
                '/sbin/route add default gw %s dev %s metric 0 2>/dev/null',
                config.outside_network.gateway,
                ext_if)
            os.execute(cmd)
        end

        -- Name servers
        local rc = io.open('/etc/resolv.conf.new', 'w')
        if config.outside_network.dns then
            for i = 1, #config.outside_network.dns do
                local ns = config.outside_network.dns[i]
                rc:write(string.format('nameserver %s\n', ns))
            end
        end
        rc:close()
        os.rename('/etc/resolv.conf.new', '/etc/resolv.conf')
    else
        -- Do configuration of external interface
        os.execute('/usr/bin/ext_if_config ' .. config.hostname .. ' ' ..
            ext_if .. ' &')
    end
end

-- If this is a CX23 run zcip on the inside network
if config.target ~= 'edaq' then
    -- UPnP route
    os.execute('/sbin/route add -host 239.255.255.250 dev ethint 2>/dev/null')
    -- Axis Camera route (Not zcip related, but this is a convenient place)
    os.execute('/sbin/route add -host 224.0.0.251 dev ethint 2>/dev/null')
    -- Zeroconf
    os.execute('/sbin/zcip ethint /usr/etc/network-zconf-script.lua')
end

-- If this is a CX23 AND inside network configuration is specified, configure
-- it as an additional IP address on the inside network interface.  NOTE - This
-- is only here to maintain backward compatibility and may be removed at some
-- time in the future.
if config.quantumx and config.quantumx.network and config.target ~= 'edaq' then
    local qx_net = config.quantumx.network
    qx_net.netmask = qx_net.netmask or '255.255.255.0'

    local bits = qx_net.address:match("/(%d+)")
    if not bits then
        local a, b, c, d = string.match(qx_net.netmask,
            '(%d+).(%d+).(%d+).(%d+)')
        local m = bit.lshift(a or 0, 24) + bit.lshift(b or 0, 16) +
            bit.lshift(c or 0, 8) + (d or 0)
        bits = 32
        while bit.band(m, 1) == 0 do
            bits = bits - 1
            m = bit.rshift(m, 1)
        end
    else
        qx_net.address = qx_net.address:match("^(.*)/")
    end
    os.execute('ip addr add ' .. qx_net.address .. '/' ..
        bits .. ' broadcast ' .. qx_net.broadcast ..
        ' dev ' .. int_if .. ' 2>/dev/null')
end

-- If we are NOT an eDAQ then we need to make sure the right thing happens when
-- link status changes occur on the two network ports that are used in the
-- bridge.
if config.target ~= 'edaq' then

    os.execute('echo 0 >/sys/devices/virtual/net/tun0/brport/multicast_router')

    local current_bridge_port_state = {}

    local sys_paths = {}

    -- Figure out the path in the sys file system for this stuff because it may
    -- be different based on what hardware we are running on.
    -- E.G.  Old prototype hardware or new hardware.
    for _, iface in ipairs({'eth1', 'eth2'}) do
        local fp =
            io.popen('find /sys/devices/e0000000.immr -type d -name ' ..  iface)
        if fp then
            for line in fp:lines() do
                if line:find('ethernet/net/' .. iface) then
                    sys_paths[iface] = line
                    break
                end
            end
            fp:close()
        end
    end

    local function set_multicast_router(iface)
        local path = sys_paths[iface]
        if path then
            repeat
                local cmd =
                    string.format("cat %s/brport/multicast_router", path)
                local fp = io.popen(cmd)
                if fp then
                    for line in fp:lines() do
                        if not line:match('2') then
                            cmd = string.format(
                                "echo 2 > %s/brport/multicast_router", path)
                            os.execute(cmd)
                            break
                        else
                            return
                        end
                    end
                    fp:close()
                end
                C.sleep(1)
            until false
        end
    end

    local function disconnect_port(iface)
        log:debug(string.format("Removing %s from bridge", iface))
        local cmd = string.format(
            '/usr/sbin/brctl delif ethint %s 2>/dev/null', iface)
        os.execute(cmd)
        current_bridge_port_state[iface] = nil
    end
    
    local function connect_port(iface)
        if not current_bridge_port_state[iface] then
            log:debug(string.format("Adding %s to bridge", iface))
            local cmd = string.format(
                '/usr/sbin/brctl addif ethint %s 2>/dev/null', iface)
            os.execute(cmd)
            current_bridge_port_state[iface] = true
        end
        set_multicast_router(iface)
    end

    local function link_status_callback(iface, state)
        if state then
            log:debug(string.format("Acquired link on interface %s", iface))

            -- These MUST be redirect stderr to /dev/null because if the
            -- thing being set is already set to what we are attempting to set
            -- it to, the stupid 'ethtool' prints a message to stderr!
            local cmd = string.format(
                '/usr/sbin/ethtool -C %s rx-frames 50 2>/dev/null', iface)
            os.execute(cmd)
            local cmd = string.format(
                '/usr/sbin/ethtool -C %s rx-usecs 5000 2>/dev/null', iface)
            os.execute(cmd)
            connect_port(iface)
        else
            log:debug(string.format("Lost link on interface %s", iface))
            disconnect_port(iface)
        end
    end

    nf.carrier_up('eth1', ev.Loop.default, link_status_callback)
    nf.carrier_up('eth2', ev.Loop.default, link_status_callback)

    ev.Loop.default:loop()
end
