#!/usr/bin/lua

local armada = require 'armada'
local ev = require 'ev'
local ffi = require 'ffi'
local C = ffi.C
local nf = require 'net_functions'
local ts = require 'timestamps'
local channel_proxy = require 'channel_proxy_noffi'
local S = require 'syscall'
local nl = require 'syscall.nl'

local dd = ffi.new('double[1]')

local networkDev = class('networkDev', channel_proxy)

local chan_info = {
    {
        base_name = 'duplex',
        description = 'Duplex Mode',
        electrical_units = '',
        get_func = function(d, ii, obj) return ii.duplex end,
        data_type = 'message_raw',
    },
    {
        base_name = 'addr',
        description = 'MAC Address',
        electrical_units = '',
        get_func = function(d, ii, obj) return tostring(d.macaddr) end,
        data_type = 'message_raw',
    },
    {
        base_name = 'link_speed',
        description = 'Link Speed',
        electrical_units = 'Mb',
        get_func = function(d, ii, obj) return ii.speed end,
        data_type = 'message_float64',
    },
    {
        base_name = 'link_state',
        description = 'Link State',
        electrical_units = '',
        get_func = function(d, ii, obj) return ii.link_up end,
        data_type = 'message_raw',
    },
    {
        base_name = 'rx_bytes',
        description = 'RX Bytes',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.rx_bytes) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_errors',
        description = 'RX Errors',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.rx_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_bytes',
        description = 'TX Bytes',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.tx_bytes) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_errors',
        description = 'TX Errors',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.tx_errors) end,
        data_type = 'message_float64',
    },

    -- Additional channels that could be included if need be
--[[
    {
        base_name = 'mtu',
        description = 'MTU',
        electrical_units = 'Bytes',
        get_func = function(d, ii, obj) return tonumber(d.mtu) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'ifindex',
        description = 'Interface Index',
        electrical_units = 'Unit',
        get_func = function(d, ii, obj) return obj.ifindex end,
        data_type = 'message_float64',
    },
    {
        base_name = 'flags',
        description = 'Operational Flags',
        electrical_units = '',
        get_func = function(d, ii, obj) return tostring(d.flags) end,
        data_type = 'message_raw',
    },
    {
        base_name = 'collisions',
        description = 'Collisions',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.collisions) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_crc_errors',
        description = 'RX CRC Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_crc_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_dropped',
        description = 'RX Dropped',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.rx_dropped) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_fifo_errors',
        description = 'RX FIFO Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_fifo_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_frame_errors',
        description = 'RX Frame Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_frame_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_missed_errors',
        description = 'RX Missed Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_missed_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_overrun_errors',
        description = 'RX Overrun Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_over_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'rx_packets',
        description = 'RX Packets',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.rx_packets) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_aborted_errors',
        description = 'TX Aborted Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.tx_aborted_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_carrier_errors',
        description = 'TX Carrier Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.tx_carrier_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_compressed',
        description = 'TX Compressed',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.tx_compressed) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_dropped',
        description = 'TX Dropped',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.tx_dropped) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_fifo_errors',
        description = 'TX FIFO Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.tx_fifo_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_heartbeat_errors',
        description = 'TX Heartbeat Errors',
        electrical_units = 'Count',
        get_func =function(d, ii, obj)
            return tonumber(d.stats.tx_heartbeat_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_packets',
        description = 'TX Packets',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.tx_packets) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'tx_window_errors',
        description = 'TX Window Errors',
        electrical_units = 'Count',
        get_func =
            function(d, ii, obj) return tonumber(d.stats.tx_window_errors) end,
        data_type = 'message_float64',
    },
    {
        base_name = 'multicast',
        description = 'Multicast',
        electrical_units = 'Count',
        get_func = function(d, ii, obj) return tonumber(d.stats.multicast) end,
        data_type = 'message_float64',
    },
]]
}

local nl_sock, err = nl.socket("route", {})
local t, pt, s, c = S.t, S.pt, S.s, S.c

local link_info

-- Timer used to occasionally get the netlink information
ev.Timer.new(function(l, timer, e)
        if not nl_sock then
            nl_sock, err = nl.socket("route", {})
        end
        if not nl_sock then
            link_info = nil
            return
        end
        local k = t.sockaddr_nl()
        local ok, err = nl.write(nl_sock, k, c.RTM.GETLINK,
            c.NLM_F.REQUEST + c.NLMSG_GETLINK.DUMP, nil, t.rtgenmsg,
            {rtgen_family = c.AF.PACKET})
        if not ok then
            link_info = nil
            return
        end
        local r, err = nl.read(nl_sock, k, nil, true)
        if not r then
            link_info = nil
            return
        end
        link_info = r
    end, 1/5, 1/5):start(ev.Loop.default)


-- Function used to gather information and send it to sinks
local function collect_from_proc(loop, timer)
    local info = timer.info

    local timestamp = ts.ts_to_stamp(ts.get_ts())
    if not timestamp then return end

    local nl_link = link_info[info.iface]
    local iface_info = nf.get_info(info.iface, info.fd)
    if not nl_link then return end
    for i = 1, #info.channels do
        local ci = info.channels[i]
        ci.timestamp = timestamp
        local d = ci.get_func(nl_link, iface_info, info)
        if ci.data_type == 'message_float64' and d then
            for _, sink in pairs(ci.comet_sinks) do
                sink(timestamp, ci.connector, d)
            end
            if ci.channel_sink then
                dd[0] = d
                ci.channel_sink(timestamp - ci.start_stamp, dd,
                    ffi.sizeof(dd))
            end
        elseif ci.data_type == 'message_raw' then
            for _, sink in pairs(ci.comet_sinks) do
                sink(timestamp, ci.connector, d)
            end
            if ci.channel_sink then
                ci.channel_sink(timestamp - ci.start_stamp, d, d:len())
            end
        end
    end
end

function networkDev:initialize(iface)
    self.comet_data = {}
    self.iface = iface

    -- Deep copy the channel information
    self.channels = {}
    for i = 1, #chan_info do
        local ci = chan_info[i]
        self.channels[i] = {}
        for k, v in pairs(ci) do
            self.channels[i][k] = v
        end
    end
    self.proxy_type = 'networkDev'
    channel_proxy.initialize(self, iface)

    for _, chan in ipairs(self.channels) do
        chan.comet_sinks = {}
        chan.channel_sink = nil
        chan.timestamp = 0
        chan.message_rate = self.rate or 2
    end

    self.ifindex = -1
    local i = nl.getlink()
    for k, v in pairs(i) do
        if v.name == iface then
            self.ifindex = tonumber(k)
            break
        end
    end

    self.rate = self.rate or 2
    self.interval = 1.0 / self.rate
    self.timer = ev.Timer.new(collect_from_proc, self.interval, self.interval)
    self.fd = C.socket(C.LOCAL_AF_INET, C.LOCAL_SOCK_DGRAM, 0)
    self.timer.info = self
    self.timer:start(ev.Loop.default)
end

function networkDev:available_channels(hostname)
    local c = {}
    for _, chan in ipairs(self.channels) do
        chan.connector = chan.base_name ..  '.status@' .. self.proxy_hostname
        chan.name = self.iface .. '_' .. chan.base_name
        chan.category = 'source/hbm_cx23r/msg_status'
        chan.is_logical = false
        c[#c + 1] = {}
        for k, v in pairs(chan) do c[#c][k] = v end
        c[#c].get_func = nil
    end
    return c
end

function networkDev:start_comet_stream(w, chans, sink)
    for _, chan in ipairs(chans) do
        for _, v in ipairs(self.channels) do
            if chan == v.connector then
                v.comet_sinks[w.socket] = sink
                w.cleanup = function()
                    for _, v in ipairs(self.channels) do
                        v.comet_sinks[w.socket] = nil
                    end
                end
                break
            end
        end
    end
end

function networkDev:start_channel_stream(w, spec, md_sink, sink)
    for _, chan in ipairs(self.channels) do
        if chan.connector == spec.connector then
            chan.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
            spec._timescale = self.interval
            spec.data_type = chan.data_type
            self.rate = chan.message_rate
            chan.channel_sink = sink
            chan.timestamp = 0
            w.cleanup = function()
                chan.channel_sink = nil
            end
            md_sink(spec)
            return
        end
    end
    self.interval = 1.0 / self.rate
    self.timer:again(ev.Loop.default, self.interval)
end

local i = nl.getlink()

for _, iface in ipairs({'eth0', 'eth1', 'eth2'}) do
    if i[iface] then
        networkDev:new(iface)
    end
end

ev.Loop.default:loop()
