#!/usr/bin/lua

local armada = require 'armada'
local ev = require 'ev'
local ffi = require 'ffi'
local ts = require 'timestamps'
local channel_proxy = require 'channel_proxy'

local dd = ffi.new('double[1]')

local powerMicroDev = class('powerMicroDev', channel_proxy)

local chan_info = {
    [1] = {
        name = 'input_voltage',
        description = 'Input Voltage',
        electrical_units = 'Volts',
        slot = 'ivolts',
        ts_block = 'Status',
    },
    [2] = {
        name = 'capacitor_bank_voltage',
        description = 'Capacitor Bank Voltage',
        electrical_units = 'Volts',
        slot = 'cvolts',
        ts_block = 'Status',
    },
    [3] = {
        name = 'supply_3_3v',
        description = '3.3 V Supply Voltage',
        electrical_units = 'Volts',
        slot = 'volts_3_3',
        ts_block = 'Status',
    },
    [4] = {
        name = 'supply_5_0v',
        description = '5.0 Volt Supply Voltage',
        electrical_units = 'Volts',
        slot = 'volts_5_0',
        ts_block = 'Status',
    },
    [5] = {
        name = 'supply_1_9v',
        description = '1.9 Volt Supply Voltage',
        electrical_units = 'Volts',
        slot = 'volts_1_9',
        ts_block = 'Status',
    },
    [6] = {
        name = 'supply_1_0v',
        description = '1.0 Volt Supply Voltage',
        electrical_units = 'Volts',
        slot = 'volts_1_0',
        ts_block = 'Status',
    },
    [7] = {
        name = 'temperature',
        description = 'Internal Temperature',
        electrical_units = 'Degrees C',
        slot = 'temp',
        ts_block = 'Status',
    },
    [8] = {
        name = 'cap_charge_level',
        description = 'Capacitor Bank Charge Level',
        electrical_units = '%',
        slot = 'cap_level',
        ts_block = 'Status',
    },
    [9] = {
        name = 'cap_charger_state',
        description = 'Capacitor Charger State',
        electrical_units = '',
        slot = 'charger_state',
        ts_block = 'Status',
    },
    [10] = {
        name = 'blue_led_state',
        description = 'Blue LED State',
        electrical_units = '',
        slot = 'blue_led_state',
        ts_block = 'LED State',
    },
    [11] = {
        name = 'green_led_state',
        description = 'Green LED State',
        electrical_units = '',
        slot = 'green_led_state',
        ts_block = 'LED State',
    },
    [12] = {
        name = 'yellow_led_state',
        description = 'Yellow LED State',
        electrical_units = '',
        slot = 'yellow_led_state',
        ts_block = 'LED State',
    },
    [13] = {
        name = 'red_led_state',
        description = 'Red LED State',
        electrical_units = '',
        slot = 'red_led_state',
        ts_block = 'LED State',
    },
}

function powerMicroDev:set_logger(logger)
    self.logger = logger
end

function powerMicroDev:log(msg, ...)
    if self.logger then
        self.logger:log(self.logger.DEBUG, string.format(msg, ...))
    end
end

function powerMicroDev:process_info(msg)
    if msg.cmd == 'Status' then
        self.comet_data.temp = msg.temperature
        self.comet_data.ivolts = msg.ivolts
        self.comet_data.cvolts = msg.cvolts
        self.comet_data.volts_3_3 = msg.volts_3
        self.comet_data.volts_5_0 = msg.volts_5
        self.comet_data.volts_1_0 = msg.volts_1
        self.comet_data.volts_1_9 = msg.volts_1_9
        self.comet_data.cap_level = msg.cap_level
        self.comet_data.charger_state = msg.charger_state
    elseif msg.cmd == 'LED State' then
        self.comet_data.blue_led_state = msg.led_state.blue
        self.comet_data.green_led_state = msg.led_state.green
        self.comet_data.red_led_state = msg.led_state.red
        self.comet_data.yellow_led_state = msg.led_state.yellow
    end
    self.timestamps[msg.cmd] = ts.get_ts()
end

function powerMicroDev:initialize()

    self.comet_data = {}
    self.timestamps = {}
    self.message_rate = 5

    self.channels = chan_info
    self.proxy_type = 'powerMicroDev'
    channel_proxy.initialize(self, 'power')

    for _, chan in ipairs(chan_info) do
        chan.comet_sinks = {}
        chan.channel_sink = nil
        chan.channel_name = chan.name .. '.status@' .. self.proxy_hostname
        chan.timestamp = 0
        chan.connector = chan.channel_name
    end

    self.update_timer_handler = function(loop, timer, revents)
        for _, chan in ipairs(self.channels) do
            local slot = self.comet_data[chan.slot]
            if slot then
                local _ts = self.timestamps[chan.ts_block]
                if _ts then
                    local timestamp = ts.ts_to_stamp(_ts)
                    for _, sink in pairs(chan.comet_sinks) do
                        sink(timestamp, chan.channel_name, slot)
                    end
                    if chan.channel_sink then
                        dd[0] = slot
                        chan.channel_sink(timestamp - chan.start_stamp, dd,
                            ffi.sizeof(dd))
                    end
                end
            end
        end
    end
    self.update_timer = ev.Timer.new(self.update_timer_handler,
        1/self.message_rate, 1/self.message_rate)
    self.update_timer:start(ev.Loop.default)
end

function powerMicroDev:set_state(connector, params)
    if self.set_led_state then
        local color = connector:match("(.*)_led_state")
        if color then
            self.set_led_state(params[1], color)
            return {
                result = 'success',
                msg = string.format("Changed %s LED state to %s", color,
                    params[1]),
            }
        end
        return {
            result = 'error',
            msg = string.format("Can not determine color from '%s'",
                connector),
        }
    end
    return {
        result = 'error',
        msg = 'set_led_state not defined',
    }
end

function powerMicroDev:get_state(connector, params)
    if self.get_led_state then
        local color = connector:match("(.+)_led_state")
        if color then
            local state = self.get_led_state(color)
            return {
                result = 'success',
                msg = 'LED State Retrieved',
                params = {state},
            }
        end
        return {
            result = 'error',
            msg = string.format("Can not determine color from '%s'",
                connector),
        }
    end
    return {
        result = 'error',
        msg = 'get_led_state not defined',
    }
end

function powerMicroDev:set_led_state_func(func)
    self.set_led_state = func
end

function powerMicroDev:get_led_state_func(func)
    self.get_led_state = func
end

function powerMicroDev:available_channels(hostname)
    local ci = {}
    for _, v in ipairs(self.channels) do
        v.connector = v.name .. '.status@' .. self.proxy_hostname
        v.data_type = 'message_float64'
        v.message_rate = self.message_rate
        v.category = 'source/hbm_cx23r/power'

        ci[#ci + 1] = {}
        for k, v2 in pairs(v) do
            ci[#ci][k] = v2
        end

        if v.name:find('led_state') then
            for _, req in ipairs{'set_state', 'get_state'} do
                v[req] = self[req]
                local _ci = ci[#ci]
                _ci.requests = _ci.requests or {}
                _ci.requests[#_ci.requests + 1] = req
            end
        end
    end
    return ci
end

function powerMicroDev:start_comet_stream(w, chans, sink)
    for _, chan in ipairs(chans) do
        for _, v in ipairs(self.channels) do
            if chan == v.channel_name then
                v.comet_sinks[w.socket] = sink
                w.cleanup = function()
                    for _, v in ipairs(self.channels) do
                        v.comet_sinks[w.socket] = nil
                    end
                end
                break
            end
        end
    end
end

function powerMicroDev:start_channel_stream(w, spec, md_sink, sink)
    for _, chan in ipairs(self.channels) do
        if chan.connector == spec.connector then
            chan.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
            chan.channel_sink = sink
            spec.message_rate = self.message_rate
            w.cleanup = function()
                chan.channel_sink = nil
            end
            md_sink(spec)
            return
        end
    end
end

function powerMicroDev:control(loop, w, http, spec)
    if not spec.request then
        return {
            result = 'error',
            msg = 'Control with no request specified',
        }
    end
    if not spec.connector then
        return {
            result = 'error',
            msg = 'Control with no connector specified',
        }
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not c[spec.request] or
                type(c[spec.request]) ~= 'function' then
                return {
                    result = 'error',
                    msg = string.format("Invalid request %s on '%s'",
                        spec.request, c.connector)
                }
            end
            return self[spec.request](self, c.connector, spec.params)
        end
    end
    return {
        result = 'error',
        msg = string.format("Connector '%s' not found", spec.connector),
    }
end

return powerMicroDev
