#!/usr/bin/lua

local ev = require 'ev'
local dbus = require 'dbus'
local ffi = require 'ffi'
local C = ffi.C
local cjson = require 'cjson'   -- Needed by network_db
local ts = require 'timestamps'
local pm_control = require 'pm-control'
local netdb = require 'network_db'
local config = require 'system_config'
local log = require 'log'
local sqlite3 = require 'lsqlite3'
local stop_tests = require 'tests_killer'

require('lj_include'){
    'time',
}


ffi.cdef[[
typedef int     pid_t;

void sync(void);
int umount2(const char *, int);
int kill(pid_t, int);
unsigned int sleep(unsigned int);
int nanosleep(const struct local_timespec *req, struct local_timespec *rem);

typedef long    time_t;

struct local_timespec {
    time_t      tv_sec;
    long        tv_nsec;
};

enum {
    MNT_FORCE = 1,
};

enum {
    SIGINT = 2,
    SIGUSR1 = 10,
    SIGUSR2 = 12,
};
]]

-- Get logging going
log:open('powerd', log.USER)
log:log(log.INFO, "Power daemon starting.")

local log_db = sqlite3.open('/var/lib/armada/log.db')

local function log_direct(msg)
    local stamp = ts.epoch_to_iso(ts.get_ts().s)
    local req = string.format(
        [[INSERT INTO log (time, component, level, message)
        VALUES ('%s', 'powerd', 'notice', '%s');]], stamp, msg)
    log_db:exec(req)
end

local power_micro_proxy

local pmc = pm_control:new(log, function(msg)
        if not power_micro_proxy then return end
        power_micro_proxy:process_info(msg)
    end, 5)

local power_micro = pmc.pm
local shutdown_in_progress


-- 500ms sleeptime - I tried shorter times but they were not reliable
local sleep_time = ffi.new('struct local_timespec', {0, 500000000})

-- Shutdown all PTP instances
local function shutdown_ptp()
    local lines = io.popen('ps | grep ptpv2d | grep -v grep')
    for line in lines:lines() do
        local pid = line:match("^ +(%d+) ")
        if pid and tonumber(pid) > 1 then
            local p = ffi.new('pid_t', tonumber(pid))
            C.kill(p, 15)
        end
    end
    lines:close()
end

-- Funtion used to continue the shutdown sequence after tests have been stopped
function power_micro:_continue_stop_all()
    -- Kill off ptp daemon(s) so the blue LED can be turned off and stay off
    shutdown_ptp()

    -- Kill the flagship proxy
    os.execute('/etc/init.d/flagship-proxy stop >/dev/null 2>&1')

    -- Unload the connctrl driver
    os.execute('/sbin/rmmod connctrl')

    -- Dismount the data storage drive
    C.umount2(config.storage.path, C.MNT_FORCE)
end

-- Shutdown everything that needs to be shutdown before powering off
function power_micro:_stop_all(emergency, cb)
    if emergency then
        stop_tests(true)
        self:_continue_stop_all()
        if cb then cb() end
    else
        stop_tests(false, function(status)
                self:_continue_stop_all()
                if cb then cb() end
            end)
    end
end

local master_pid
local fh = io.popen('ps -o pid,args | grep master | grep -v grep')
if fh then
    for line in fh:lines() do
        master_pid = tonumber(line:match('(%d+) .+/master$'))
        break
    end
    fh:close()
end

function power_micro:_shutdown(msg, green_rate)
    if not shutdown_in_progress then
        -- Let master know we are starting the shutdown sequence.  This is
        -- mostly so master doesn't log/notify when processes start getting
        -- killed off as part of the shutdown process.
        if master_pid then
            C.kill(ffi.new('pid_t', master_pid), C.SIGUSR1)
        end

        local sl = self.SetLed
        self.SetLed = function() end

        -- Start the green LED flashing
        sl(self, green_rate, "green")
        sl(self, "off", "red")
        sl(self, "off", "yellow")
        sl(self, "off", "blue")

        -- Log the fact that the power button was pressed
        log_direct(msg)

        -- Stop stuff that needs to be stopped before turning off channel power
        self:_stop_all(msg == "Power Lost" and true or false, function()
                -- Now that everything is in a state where by it should be
                -- safe to power down, request that to happen
                self:RequestSystemShutdown()

                -- If we don't see the ACK from this within 1 second, force it
                self.req_watcher = ev.Timer.new(function(l, t, e)
                        self:ShutdownAck(0)
                    end, 1, 1)
                self.req_watcher:start(ev.Loop.default)
            end)

    end
    shutdown_in_progress = true
end

-- Callback function called when the user presses the power button.
function power_micro:ButtonDepressed()
    os.execute('touch /var/lib/armada/powered-off;sync')
    self:_shutdown("Power Button Depressed", "four_hz")
end


-- Callback function called  after the power micro acknowledges a shutdown
-- request.
function power_micro:ShutdownAck(mode)
    if mode == 0 then
        self.req_watcher:stop(ev.Loop.default)
        self:ChannelPowerOff()

        -- If we don't see the ACK from this within 1 second, force it
        self.req_watcher = ev.Timer.new(function(l, t, e)
                self:ChannelPowerAck()
            end, 1, 1)
        self.req_watcher:start(ev.Loop.default)
    end
end

-- Callback function called after the power micro acknowledges the carrier
-- power off req
function power_micro:ChannelPowerAck()
    if not self.in_chan_power_ack then
        self.in_chan_power_ack = true

        self.req_watcher:stop(ev.Loop.default)
        self.req_watcher = nil

        -- Tell the PM that it is now safe to turn off the power.
        -- The PM turns the power off 8 seconds after it gets this request.
        -- That should be plenty of time for the kill below to do it's thing.
        self:ShutdownFinished()

        -- Signal the system that we are going to power off
        C.kill(ffi.new('pid_t', 1), C.SIGUSR2)

        -- We are done, power is going down whether we like it or not!
        ev.Loop.default:unloop()
    end
end


-- Callback function called when the remote power state changes.
function power_micro:RemotePowerStateChange(state)
    os.execute('touch /var/lib/armada/powered-off;sync')
    self:_shutdown("Remote Power State Changed", "eight_hz")
end

-- Callback function called when power is lost
function power_micro:PowerLost(mode)
    os.execute('touch /var/lib/armada/powered-off;sync')
    log_direct("Power Lost")

    -- Need a small bit of time to get this message into the log before going
    -- down.
    ev.Timer.new(function(l, t, e)
            self:AckPowerLost()
            self:_shutdown("Power Lost", "two_hz")
        end, .5, 0):start(ev.Loop.default)
end

-- Callback function called when a restore credentials message is received
function power_micro:RestoreAdminCredentials()
    -- Log occurrance
    log:notice("Administrator login and password reset")

    -- Reset administrator login and password to default values
    os.execute('/usr/bin/db-initialize reset-admin-only')
end

-- Function used to get information from the sql database
local function sql_exec(db, sql)
    local ok, result = pcall(function() return db:exec(sql) end)
    if ok then
        return result, ok
    else
        return {ok=false, error=result}, ok
    end
end

-- Function used to get information from the sql database
local function sql_get(db, sql)
    local vm = db:prepare(sql)
    if not vm then return end
    local result = {}
    while vm:step() == sqlite3.ROW do
        result[#result + 1] = vm:get_uvalues()
    end
    vm:finalize()
    if #result > 0 then
        return result, true
    end
    return
end

-- Function to reset the default network configuration back to factory setting
-- This is done by deleting the bootstrap 'default' configuration
-- then reset the active status to false on all remaining networks
-- and finally inserting the default network configuration
function power_micro:ResetNetworkParameters()
    local db = sqlite3.open('/var/lib/armada/setup.db')
    if not db then
        log:log(log.ERR, "Unable to open network configuration database")
        return
    end
    -- delete the default record if it already exists
    local r, ok = sql_exec(db, [[
        DELETE FROM tsd_users
        WHERE db_name = 'networks' AND id = 'default' AND deleted = '';
    ]])
    -- update other network configurations to active = false
    r, ok = sql_get(db, [[ SELECT content from tsd_users
                           WHERE db_name = 'networks' AND id != 'hostname' AND deleted = '';]])
    if ok and r then
        for _, v in ipairs(r) do
            if v then
                local c = cjson.decode(v)
                if c and c.active then
                    c.active = false;
                    local query = string.format(
                        [[ UPDATE tsd_users  SET content = '%s'
                           WHERE db_name = 'networks' AND id = '%s' AND deleted = '';]],  cjson.encode(c), c.id)
                    sql_exec(db, query)
                end
            end
        end
    end
    -- create the default network configuration
    local default_network = require 'default_network'
    local req = string.format(
        [[INSERT INTO tsd_users (db_name, id, content, added)
            VALUES ('networks', 'default', '%s', '%s')]],
            cjson.encode(default_network), os.date('!%FT%H:%M:%S', os.time()))
    r, ok = sql_exec(db, req)
    db:close()
    if ok then
        log:log(log.NOTICE, "Network settings reset to factory default")
        return true
    end
    log:log(log.ERR, "Unable to reset network settings to factory default")
end

-- Attach signal hanelers
for _, sig in ipairs{C.SIGINT, C.SIGQUIT, C.SIGTERM} do
    ev.Signal.new(function(loop, sig, re) loop:unloop() end, sig):start(
        ev.Loop.default)
end

local counter = 0

local function chase_leds()
    local led_map = {'blue', 'green', 'yellow', 'red'}
    local led = led_map[(counter % 4) + 1]
    for _, l in ipairs(led_map) do
        local mode = 'off'
        if l == led then
            mode = 'on'
        end
        power_micro:SetLed(mode, l)
    end
end

local ptp_monitor
local ptp_states = {}
local got_gps
local blue_led_state

local function set_blue_led(state)
    if not blue_led_state or blue_led_state ~= state then
        power_micro:SetLed(state, "blue")
    end
end

local function operate_blue_led()
    -- If neither of the PTP daemons have registered a state yet, then we are
    -- entirely unsynchronized or synchronized to GPS.
    if not ptp_states.eth1 and not ptp_states.eth2 then
        if not got_gps then
            set_blue_led("off")
        else
            set_blue_led("one_hz")
        end
        got_gps = false
        return
    end

    -- If we are NOT synchronized to GPS and one of the internal interfaces is
    -- synchronized to PTP and the other is a PTP master, we are in boundary
    -- clock mode, so indicate that.
    if ptp_states.eth1 and ptp_states.eth2 and not got_gps then
        if ptp_states.eth1 == 'PTPMaster' and
            ptp_states.eth2 == 'PTPSync' then
            set_blue_led('eight_hz')
            return
        elseif ptp_states.eth1 == 'PTPSync' and
            ptp_states.eth2 == 'PTPMaster' then
            set_blue_led('eight_hz')
            return
        end
    end

    -- If ANY PTP instance is in slave mode then we are synchronized to PTP
    for _, v in pairs(ptp_states) do
        if v == 'PTPSync' then
            -- PTP synchronized slave
            set_blue_led("on")
            got_gps = false
            return
        end
    end

    -- If none of the PTP instances are in slave mode AND at least one of the
    -- PTP instances is in master mode, then we are a PTP master are are either
    -- synchronized to GPS or not.
    for _, v in pairs(ptp_states) do
        if v == 'PTPMaster' then
            if got_gps then
                -- GPS Sync'd master
                set_blue_led("one_hz")
            else
                -- Unsynchronized master
                set_blue_led("two_hz")
            end
            got_gps = false
            return
        end
    end

    -- The PTP daemon is running, we are not MASTER, and we are not sync'd.
    set_blue_led("off")
    got_gps = false
end

if config.target ~= 'armada' then
    operate_blue_led = nil
end

ev.Timer.new(function(l, t, e)
        if counter < 30 then
            chase_leds()
        elseif counter >= 30 and counter < 40 then
            power_micro:SetLed("off", "blue")
            power_micro:SetLed("off", "red")
            power_micro:SetLed("off", "yellow")
            power_micro:SetLed("on", "green")
        end

        -- Operate the blue led based on received PTP and GPS servo heartbeats
        -- Note that this can not be done faster than the maximum rate at which
        -- GPS heartbeats arrive (1Hz).
        if counter >= 60 and operate_blue_led and counter % 20 == 0 then
            operate_blue_led()
        end
        counter = counter + 1
    end, .1, .1):start(ev.Loop.default)


local bus = dbus.bus_get('system')
local master
local network

-- Don't block this process waiting on master to come available!
ev.Timer.new(function(l, t, e)
    if not master then
        master = bus:auto_proxy('com.somat.armada.Master',
            '/com/somat/armada/Master')
        if master then
            master.Heartbeat:bind(function(from)
                if from == "GPSServo" then
                    got_gps = true
                elseif from:match(":PTP") then
                    local iface, state = from:match("(eth%d+):(.*)")
                    if state ~= 'PTPOff' then
                        ptp_states[iface] = state
                    else
                        ptp_states[iface] = nil
                    end
                end
            end)
        end
    end
    if not network then
        network = bus:auto_proxy('com.somat.armada.Network',
            '/com/somat/armada/Network')
        if network then
            power_micro_proxy = require('power_micro_proxy'):new()
            power_micro_proxy:set_logger(log)
            power_micro_proxy:set_led_state_func(function(state, color)
                    power_micro:SetLed(state, color)
                end)

            -- Get initial LED state to proxy
            local msg = {
                cmd = 'LED State',
                led_state = {}
            }
            for _, led in ipairs{'blue', 'green', 'yellow', 'red'} do
                msg.led_state[led] = pmc:get_led(led)
            end
            power_micro_proxy:process_info(msg)

        end
    end
    if master and network then t:stop(l) end
end, 10, 1):start(ev.Loop.default)

-- When some entity in the system wants the system shutdown, it sends SIGPWR
-- to the powerd process.
ev.Signal.new(function(l, s, e)
        power_micro:_shutdown('System requested power down', "one_hz")
    end, C.SIGPWR):start(ev.Loop.default)

-- SIGUSR1 is used to put the LEDs into a state to indicate that the system is
-- doing something like rebooting or the like.
ev.Signal.new(function(l, s, e)
        local sl = power_micro.SetLed
        power_micro.SetLed = function() end

        log_direct('Reboot requested')

        -- Start the green LED flashing
        sl(power_micro, "eight_hz", "green")
        sl(power_micro, "off", "red")
        sl(power_micro, "off", "yellow")
        sl(power_micro, "off", "blue")
    end, C.SIGUSR1):start(ev.Loop.default)
        

-- Run the loop
ev.Loop.default:loop()
