require('lj_include'){
    'stdlib',
    'string',
    'unistd',
    'sys/socket',
}

require 'strict'
local __ = require '__'
local bit = require 'bit'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'
local log = require 'log'
local membuf = require 'membuf'
local sample_types = require 'sample_types'
local time_bracket = require 'time_bracket'
local sys_notify = require 'sys_notify'
local config = require 'system_config'

local C = ffi.C

local bit_band = bit.band
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

-- local htonl = ffi.abi("le") and bit.bswap or bit.tobit
-- The following version doesn't abort the JIT compile
local htonl = ffi.abi("le") and bit.bswap or function(x) return x end
local ntohl = htonl

ffi.cdef[[
    struct proxy_buffer {
        struct membuf inbuf;
        struct time_bracket bracket;
        int i;
        int sample_size;
        int timestamped_samples;
        int fd;
        ssize_t already_read;
        int inhibit_exception;
        int failed;
        int read_errno;
        ssize_t read_ret;
        timestamp timestamp[1];
        char *connector_name;
        float64_t xform_scale;
        float64_t xform_offset;
        int exit_on_proxy_error;
    };
]]

local mt = {}

local metadata_marker = bit.lshift(1LL, 63)

local function tohex(inbuf)
    return tostring(inbuf):gsub("(.)", function(c)
        return string.format("%02X", string.byte(c))
            end)
end

function mt:read_from_proxy(test_engine, buf_type)
    if self.already_read == 0 then
        self.inbuf:assure_available()
        self.already_read = C.read(self.fd, self.inbuf:tail(),
            self.inbuf:available())
        if self.already_read <= 0 then
            self.failed = 1
            self.read_ret = self.already_read
            if self.already_read < 0 then
                self.read_errno = ffi.errno()
            end
            C.shutdown(self.fd, C.SHUT_RDWR)
            C.close(self.fd)
            if self.inhibit_exception == 0 then
                self:read_exception()
            end
            return nil
        end
        self.inbuf:append(nil, self.already_read)
        self.inbuf:assure_available() -- Expand if we read a full buffer
        self.i = 0
    end
    if self.i + ffi_sizeof'proxy_length_header' > #self.inbuf then
        return self:reset()
    end
    -- KLUDGE maybe copy instead of cast for unalignededness issues
    local ph = ffi_cast('proxy_header *', self.inbuf:head() + self.i)
    local len = ntohl(ph.length)
    local ni = self.i + ffi_sizeof'proxy_length_header' + len
    if ni > #self.inbuf then
        if ni > self.inbuf.max_size then
            if len > 20000000 then
                local hex = tohex(self.inbuf)
                log(log.ERR,
                    __'Test {0} proxy {1}: Bad length {2} pos {3} in "{4}"',
                    test_engine.test_id, ffi_string(self.connector_name),
                    len, self.i, hex)
                self.failed = 1
                return nil
            end
            self.inbuf:expand(ni)
        end
        return self:reset()
    end

    local payload_len = len - ffi_sizeof'timestamp'
    -- We do ffi.copy() here to avoid traps on unaligned 64-bit access
    -- KLUDGE WRONG needs ntohll() for timestamp if portable to LE
    ffi_copy(self.timestamp, ph.timestamp, ffi_sizeof(self.timestamp))
    if self.timestamp[0] == metadata_marker then
        local ptr = self.inbuf:head() + self.i + ffi_sizeof'proxy_header'
        local json = ffi_string(ptr, payload_len)
        self.i = ni
        local md, err = cjson.decode(json)
        if err then
            log(log.ERR,
                __'Test {0} Bad channel metadata from {1}: {2} {3}',
                test_engine.test_id, ffi_string(self.connector_name),
                err, cjson.encode(tostring(self.inbuf)))
            self.failed = 1
            return nil
        end
        if md.error_message then
            if self.exit_on_proxy_error ~= 0 then os.exit(-10) end
            log(log.ERR,
                __'Test {0} Error message from proxy {1}: "{2}"',
                test_engine.test_id, ffi_string(self.connector_name),
                md.error_message)
            self.failed = 1
            return nil
        end
        return md
    end

    local sp = ffi_cast(buf_type, ph)
    local n = (self.timestamped_samples == 0 and payload_len or len) /
        self.sample_size
    self.i = ni
    -- KLUDGE disable "assertion" below once sufficiently tested
    if bit_band(ffi_cast('int32_t', sp.sample), 3) ~= 0 and
            (self.sample_size == 4 or self.sample_size == 8) then
        log(log.ERR,
            __'Test {0} Proxy {1}: Bad alignment of {2} for sample_size {3}',
            test_engine.test_id, ffi_string(self.connector_name), sp.sample,
            self.sample_size)
        self.failed = 1
        return nil
    end
    return sp.sample, n, self.timestamp[0]
end

function mt:read_exception()
    local sn = sys_notify:new("Test Unexpectedly Stopped")
    sn:critical_error(string.format("Error reading from '%s'.  " ..
        "This usually indicates the proxy serving this channel has exited.",
        ffi_string(self.connector_name)), true)
    if self.read_ret == 0 then
        log:error(__'{0} unexpectedly disconnected from test\n',
            ffi_string(self.connector_name))
    else
        log:error(__'Failed proxy read on connector "{0}", errno={1} "{2}"\n',
            ffi_string(self.connector_name),
            self.read_errno, log:strerror(self.read_errno))
    end
end

function mt:reset()
    self.inbuf:remove(0, self.i)
    self.already_read = 0
    self.i = 0
    return nil
end

mt.__index = mt

function mt:__gc()
    self.inbuf:__gc()
    C.free(self.connector_name)
    self.connector_name = nil
end

local mt_ctype = ffi.metatype('struct proxy_buffer', mt)

local m = {}

function m:new(buffer_size, sample_size, connector_name, timestamped_samples,
        sample_rate)
    local pb = mt_ctype()
    pb.inbuf:initialize(buffer_size)
    pb.bracket:initialize()
    pb.bracket.sample_rate = sample_rate or 0
    pb.sample_size = sample_size
    pb.exit_on_proxy_error = 1
    connector_name = tostring(connector_name) or ''
    pb.connector_name = C.malloc(#connector_name + 1)
    ffi_copy(pb.connector_name, connector_name)
    pb.timestamped_samples = timestamped_samples and timestamped_samples or 0
    pb.xform_scale = 1.0
    pb.xform_offset = 0.0
    return pb
end

return m
