local ffi = require'ptpv2-ffi/cdata'
local C = ffi.C

local m = {}

local function tx_ts_cb(ts, raw)
end

-- Function used to send an announce message
function m:send()
    local ci = self.clock_info
    local size = ci.sizeof_announce_msg

    ffi.fill(ci.ibuf, size + 2)
    self:build_msg_header()
    local msg = ffi.cast('announce_msg_t *', ci.ibuf)
    msg.hdr.transport_specific_message_type = C.PTP_MSG_TYPE_ANNOUNCE
    msg.hdr.message_length = size
    msg.hdr.sequence_id = C.htons(ci.sent_announce_sequence_id)
    msg.hdr.control_field = 0x5
    msg.hdr.log_message_interval = ci.log_announce_interval

    msg.current_utc_offset = C.htons(ci.current_utc_offset)
    ffi.copy(msg.grandmaster, ci.grandmaster, ffi.sizeof(msg.grandmaster))

    msg.steps_removed = C.htons(ci.steps_removed)
    msg.time_source = ci.time_source

    if ci.transport_is_ipv6 ~= 0 then
        size = size + 2
    end
    
    local sock = self[C.GENERAL_PORT]
    local r, e = pcall(sock.sendto, sock, ci.ibuf, size, tx_ts_cb)
    if not r then
        self:log("error", string.format("send_sync(): sendto() failed - %s", e))
        self:to_state(C.PTP_STATE_FAULTY)
        return
    end
    ci.sent_announce_sequence_id = ci.sent_announce_sequence_id + 1
end

function m:do_record_update()
    local ci = self.clock_info
    if ci.state == C.PTP_STATE_LISTENING or ci.state == C.PTP_STATE_PASSIVE or
        ci.state == C.PTP_STATE_SLAVE or ci.state == C.PTP_STATE_MASTER then

        local state = self:best_master_clock()
        if state ~= ci.state then
            self:to_state(state)
        end
    end
end

function m:faulty(time)
    -- FIXME: Recover from faulty state?
end

-- Function to reset the announce receipt timer
function m:reset_announce_timer()
    local ci = self.clock_info
    self.announce_receipt_timer:again(self.loop,
        ci.announce_receipt_timeout * math.pow(2, ci.log_announce_interval))
end

function m:passive(time)
    local ci = self.clock_info
    local hdr = ffi.cast('msg_hdr *', ci.ibuf)
    if ci.passive_port_identity.port_number ==
        hdr.source_port_identity.port_number and
        C.memcmp(ci.passive_port_identity.clock_identity,
            hdr.source_port_identity.clock_identity,
            ffi.sizeof(ci.passive_port_identity.clock_identity)) == 0 then
        m.reset_announce_timer(self)
    end
    self:add_foreign()
    m.do_record_update(self)
end

function m:uncalibrated(time)
    if self:is_from_current_master() then
        m.reset_announce_timer(self)
    end
    self:add_foreign()
    m.do_record_update(self)
end

function m:slave(time)
    m.do_record_update(self)
    if self:is_from_current_master() then
        local ci = self.clock_info
        local msg = ffi.cast('announce_msg_t *', ci.ibuf)
        local hdr = ffi.cast('msg_hdr_t *', ci.ibuf)

        hdr.message_length = C.ntohs(hdr.message_length)
        hdr.correction_field = self:fixup_correction(hdr.correction_field)
        hdr.sequence_id = C.ntohs(hdr.sequence_id)
        hdr.source_port_identity.port_number =
            C.ntohs(hdr.source_port_identity.port_number)

        msg.origin_timestamp.seconds.msb =
            C.ntohs(msg.origin_timestamp.seconds.msb)
        msg.origin_timestamp.seconds.lsb =
            C.ntohl(msg.origin_timestamp.seconds.lsb)
        msg.origin_timestamp.nanoseconds =
            C.ntohl(msg.origin_timestamp.nanoseconds)
        msg.current_utc_offset = C.ntohs(msg.current_utc_offset)
        msg.grandmaster.clock_quality.offset_scaled_log_variance =
            C.ntohs(msg.grandmaster.clock_quality.offset_scaled_log_variance)
        msg.steps_removed = C.ntohs(msg.steps_removed)
        self:s1(hdr, msg)
    else
        self:add_foreign()
    end
    m.reset_announce_timer(self)
end

function m:listen(time)
    self:add_foreign()
    m.do_record_update(self)
end

-- Table of handler functions to handle announce messages in specfic states
m.handle_state = {}
m.handle_state[C.PTP_STATE_LISTENING] = m.listen
m.handle_state[C.PTP_STATE_PRE_MASTER] = m.handle_state[C.PTP_STATE_LISTENING]
m.handle_state[C.PTP_STATE_MASTER] = m.handle_state[C.PTP_STATE_LISTENING]
m.handle_state[C.PTP_STATE_SLAVE] = m.slave
m.handle_state[C.PTP_STATE_UNCALIBRATED] = m.uncalibrated
m.handle_state[C.PTP_STATE_FAULTY] = m.faulty
m.handle_state[C.PTP_STATE_PASSIVE] = m.passive

-- Function used to handle ALL incoming announce messages
function m:handle(time, from_me)
    if from_me then
        return
    end
    local state = self.clock_info.state
    local f = m.handle_state[state]
    if f then
        f(self, time)
    end
end

return m
