require 'strict'

local __ = require '__'
local bit_compiler = require 'bit_compiler'
local cc_parser = require 'cc-parser'
local cjson = require 'cjson'
local compiler_edge = require 'compiler_edge'
local compiler_node = require 'compiler_node'
local edge_buf = require 'edge_buf'
local log = require 'log'
local network_db = require 'network_db'
local setup_spec = require 'setup_spec'
local pcre = require 'pcre_ffi'
local rational = require 'rational'
local sample_types = require 'sample_types'
local sie_writer = require 'sie_writer'
local sqlite = require 'sqlite_ffi'
local statics = require 'setup_compiler_statics'
local uuid = require 'uuid'
local sample_clock = require 'sample_clock'

require 'middleclass'

log:open('setup-compiler', log.USER)

local function errorf(fmt, ...)
    local msg = string.format(fmt, ...)
    log:log(log.ERR, msg)
    error(msg)
end

local function warn(fmt, ...)
    local msg = string.format(fmt, ...)
    log:log(log.WARNING, msg)
    io.stderr:write(msg)
end

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

local source_re = pcre:new('^source(/.*)?$')
local filter_re = pcre:new('^filter(/.*)?$')
local internal_re = pcre:new('^internal(/.*)?$')
local filter_expression_re = pcre:new('^filter/expression(/.*)?$')
local sink_re = pcre:new('^sink(/.*)?$')

local m = class('setup_compiler'):include(require 'template')
m:include(require 'compiler_oc')

m.default_sie_buffer_size = 8192

local function boolean(x, default)
    if x == 'true' or x == true then
        return true
    end
    if x == 'false' or x == false then
        return false
    end
    if x == nil then
        if default ~= nil then
            return default
        end
        return nil
    end
    errorf("'%s' must be a boolean true or false value", x)
end

local func_compile = {}

local function func_args(self, node, ...)
    local args = {...}
    local t = {}
    for i = 1, #args do
        t[#t + 1] = self:walk(node, args[i])
    end
    return t
end

local function base_sample(v)
    if type(v) ~= 'string' then
        return v
    end
    return v:gsub('%.value%[0%]$', '')
end

function func_compile.xbit(self, node, ...)
    local name = self.func_name -- Crucial, will change during walk()
    local a = func_args(self, node, ...)
    local val = table.remove(a, 1)
    local prefix, unit = name:match('^(.)(.*)$')
    val = base_sample(val)
    local comp = bit_compiler['compile_' .. unit](bit_compiler, prefix, val, a)
    return string.format('%s --[[ %s(%s) ]]',
        comp, name, table.concat(a, ', '))
end

function func_compile.can_message(self, node, ...)
    local a = func_args(self, node, ...)
    if #a > 1 then
        errorf('Non-hex-constant can_message() not yet supported')
    end
    local name = self:oc_can_constant(a[1])
    return name
end

local lua_op_map = {
    ['||'] = 'or',
    ['&&'] = 'and',
    ['!='] = '~=',
    ['!'] = 'not',
}

local func_map = {
    abs = 'math.abs',

    floor = 'math.floor',
    ceil = 'math.ceil',

    sin = 'math.sin',
    cos = 'math.cos',
    tan = 'math.tan',

    sinh = 'math.sinh',
    cosh = 'math.cosh',
    tanh = 'math.tanh',

    asin = 'math.asin',
    acos = 'math.acos',
    atan = 'math.atan',
    atan2 = 'math.atan2',

    log = 'math.log',
    log10 = 'math.log10',

    max = 'math.max',
    min = 'math_min',

    exp = 'math.exp',
    ldexp = 'math.ldexp',

    pow = 'math.pow',

    fmod = 'math.fmod',
    modf = 'math.modf',

    pi = function(self, node) return 'math.pi' end,

    ['nil'] = function(self, node) return 'nil' end,

    time = function(self, node) return '_time' end,
    sample_count = function(self, node) return '(_count + 1)' end,

    float = function(self, node, expr)
        return string.format("((%s) and 1 or 0)", self:walk(node, expr))
    end,

    integrate = function(self, node, expr)
        local exprnode = mk_computed_node(expr, node.child_md)
        local intnode = {
            md = { _type = 'IntegratorNode' },
            inputs = { { exprnode, 1 } }
        }
        table.insert(nodes, intnode)
        return add_input(node, intnode)
    end,

    decimate = function(self, node, expr, factor)
        local exprnode = mk_computed_node(expr, node.child_md)
        factor = cc.getnumber(factor)
        local decnode = {
            md = { _type = 'DecimatorNode', _factor = factor },
            inputs = { { exprnode, 1 } }
        }
        table.insert(nodes, decnode)
        return add_input(node, decnode)
    end,

    timeshift = function(self, node, expr, shift)
        local exprnode = mk_computed_node(expr, node.child_md)
        shift = cc.getnumber(shift)
        local shiftnode = {
            md = { _type = 'TimeshiftNode', _shift = shift },
            inputs = { { exprnode, 1 } }
        }
        table.insert(nodes, shiftnode)
        return add_input(node, shiftnode)
    end,

    bit_match = function(self, node, ...)
        local a = func_args(self, node, ...)
        for i = 1, #a do
            a[i] = base_sample(a[i])
        end
        if #a == 2 then
            return string.format('bit_match(%s, %s)', a[1], a[2])
        elseif #a == 3 then
            return string.format('bit_match_masked(%s, %s, %s)',
                a[1], a[2], a[3])
        else
            errorf('Wrong number of arguments (%d) for bit_match()', #a)
        end
    end,

    sbit = func_compile.xbit,
    ubit = func_compile.xbit,
    sbyte = func_compile.xbit,
    ubyte = func_compile.xbit,
    can_message = func_compile.can_message,
}

function m:walk(node, expr)
    if expr[1] == 'number' then
        return expr[2]
    elseif expr[1] == 'message' then
        --return 'message("' .. expr[2] .. '")'
        local mess = '"' .. expr[2]:gsub("..", function(s) return string.format("\\x%s", s) end) .. '"'
        return mess
    elseif expr[1] == 'name' then
        if not node.input_names[expr[2]] then
            errorf('Unknown node.input_names["%s"]', expr[2])
        end
        return node.input_names[expr[2]]
    elseif expr[1] == 'symbol' then
        return expr[2] -- KLUDGE???
    elseif expr[1] == 'op' then
        return lua_op_map[expr[2]] or expr[2]
    elseif expr[1] == 'unary' then
        return (lua_op_map[expr[2]] or expr[2]) .. self:walk(node, expr[3])
    elseif expr[1] == 'funcall' then
        local func_name = expr[2][2]
        local mapped = func_map[func_name]
        if type(mapped) == 'table' then
            -- KLUDGE do something here...
        end
        if type(mapped) == 'string' then
            local args = { }
            for i = 1, #expr - 2 do
                args[i] = self:walk(node, expr[i + 2])
            end
            return mapped .. '(' .. table.concat(args, ', ') .. ')'
        elseif type(mapped) == 'function' then
            local args = { }
            for i = 1, #expr - 2 do
                args[i] = expr[i + 2]
            end
            self.func_name = func_name
            return mapped(self, node, unpack(args))
        else
            error('unknown function ' .. expr[2][2])
        end
    else
        local parts = { }
        for i = 1, #expr - 1 do
            parts[i] = self:walk(node, expr[i + 1])
        end
        return '(' .. table.concat(parts, ' ') .. ')'
    end
end

local function extend(t, ...)
    for _, nt in ipairs{...} do
        for k, v in pairs(nt) do
            if t[k] == nil then
                t[k] = v
            end
        end
    end
    return t
end

local function std_trig_mode(mode)
    if mode == "always_on" or
       mode == "trigger" or
       mode == "gate" or
       mode == "one_shot" then
        return true
    else
        return false
    end
end

function m:err(fmt, ...)
    table.insert(self.errors, string.format(fmt, ...))
end

function m:add_status_channels()
    local system_time_cn
    local net_db = network_db:new()
    local hn = net_db:get_inside_network_hostname()
    net_db:close()
    local tc = string.format('system_clock.status@system.%s', hn)
    if not self.json:find(tc, 1, true) then
        -- Add system time channel
        self.original_setup[#self.original_setup + 1] = {
            category = 'source/hbm_cx23r/msg_status',
            connector = tc,
            collect = false,
            description = 'System Clock',
            name = 'system_clock',
            output_units = 'seconds',
            is_logical = false,
            data_type = 'message_float64',
            live_data_stream = false,
        }
        system_time_cn = 'system_clock'
    else
        for _, v in ipairs(self.original_setup) do
            if v.connector == tc then
                system_time_cn = v.name
                break
            end
        end
    end
    if not system_time_cn then return end

    -- Add the elapsed run time channel
    self.original_setup[#self.original_setup + 1] = {
        __input_channels_0 = system_time_cn,
        category = 'internal/elapsed_run_time',
        collect = false,
        description = 'Elapsed Test Run Time',
        input_channels = {system_time_cn},
        name = 'elapsed_test_run_time.status',
        output_units = 'seconds',
        electrical_units = 'seconds',
        live_data_stream = true,
        is_logical = false,
        data_type = 'message_float64',
    }

    -- Add the SIE file size channel
    self.original_setup[#self.original_setup + 1] = {
        __input_channels_0 = system_time_cn,
        category = 'internal/sie_size',
        collect = false,
        description = 'SIE File Size',
        input_channels = {system_time_cn},
        name = 'sie_file_size.status',
        output_units = 'bytes',
        electrical_units = 'bytes',
        live_data_stream = true,
        is_logical = false,
        data_type = 'message_float64',
    }
end

function m:initialize(init)
    self.errors = {}
    self:template_init()
    for k, v in pairs(init) do
        self[k] = v
    end
    self.outfh = io.stdout
    local args = self.args
    local i = 1
    while i <= #args do
        local arg = args[i]
        if arg == '-o' or arg == '--output' then
            i = i + 1
            self.output_file = args[i]
            self.outfh = assert(io.open(self.output_file, 'w'))
        elseif arg == '-s' or arg == '--settings' then
            i = i + 1
            self.settings = args[i]
        elseif arg == '--dump' then
            self.jit_dump = true
        else
            self.setup_file = args[i]
        end
        i = i + 1
    end

    if self.setup_file then
        self.json = self:read_file(self.setup_file)
        self.file_base = self.setup_file:gsub('%.[^%.]*$', '')
    end
    if self.json then
        self.original_setup = cjson.decode(self.json)
        -- FIXME:  This should only be done for non-digital display type tests
        self:add_status_channels()
    end
    self.setup = setup_spec:normalize(self.original_setup)
    do
        local summary_idx
        for idx, item in ipairs(self.setup) do
            if item.id == 'summary' or item.category == 'summary' then
                summary_idx = idx
                break
            end
        end
        if not summary_idx then
            summary_idx = 1
            table.insert(self.setup, summary_idx, {})
        end
        local defaults = {
            category = 'summary',
            id = 'summary',
            sample_rate_domain = 'hbm_legacy', -- KLUDGE for Quantums
        }
        extend(self.setup[summary_idx], defaults)
    end
    self.pretty_json = setup_spec:export(self.setup)

    self.buffers = {}
    self.nodes = {}
    self.edges = {}
    self.stream_nodes = {}
    self.setup_aux = {}
    self.modules = {}
    self.cache = {}

    self:defaults()
    self:traverse()
    self:dot_graph('post_traverse')
    self:dot_graph()
    self:compile()

    for _, tag in ipairs{'preamble', 'reqr', 'op_locals', 'decl',
            'proxy_buffers', 'proxy_readers', 'finalize', 'process',
            'edge_bufs', 'hbm_stream_bufs', 'init_1', 'init_2',
            'init_3', 'postamble'} do
        self.outfh:write('---- ' .. tag .. ' ----\n')
        self.outfh:write(self:render_buffer(tag))
        self.outfh:write('\n')
    end
end

function m:read_file(fname)
    local fh = assert(io.open(fname))
    local content = fh:read('*a')
    fh:close()
    return content
end

local function dot_name(n)
    n = n:gsub('^(%x%x%x%x%x%x%x%x)%-%x%x%x%x%-%x%x%x%x%-%x%x%x%x%-'
        .. '%x%x%x%x%x%x%x%x%x%x%x%x$', '%1...')
    return n
end

function m:dot_graph(graph_name)
    if not self.dot_fh then
        if not self.file_base then
            return
        end
        self.dot_fh = io.open(self.file_base .. '.dot', 'w')
    end
    if not graph_name then
        self.dot_fh:close()
        return
    end
    self:dot('digraph %s {\n', graph_name)
    self:dot('    node[fontsize=10,fontname="Helvetica"];\n')
    self:dot('    edge[fontsize=8,fontname="Helvetica"];\n')
    for idx, edge in ipairs(self.edges) do
        assert(idx == edge.idx)
        self:dot('    "%s" -> "%s" [label="(e%s)\\n%s &#x2192; %s"];\n',
            edge.from_idx, edge.to_idx, edge.idx, edge.from_idx, edge.to_idx)
    end
    local sources, sinks = {}, {}
    for _, node in ipairs(self.nodes) do
        local name = dot_name(node.meta.name)
        local style = ''
        local cat = node.meta.category
        if cat:match('^source') then
            style = 'shape=box,style=filled,color="#CCFFCC",'
            sources[#sources + 1] = node.idx
        elseif cat:match('^sink') then
            style = 'shape=box,style=filled,color="#FFCCCC",'
            sinks[#sinks + 1] = node.idx
        else
            style = 'shape=box,style=filled,color="#CCCCFF",'
        end
        self:dot('    "%s" [%slabel="%s [n%s]\\n%s\\n%s%s"];\n',
            node.idx, style, name, node.idx, node.meta.category,
            node.data_type, node.meta.data_type and ' *' or '')
    end
    for _, list in ipairs{sources, sinks} do
        self:dot('    subgraph { rank = same;')
        for _, idx in ipairs(list) do
            self:dot(' "%s";', idx)
        end
        self:dot('}\n')
    end
    self:dot('}\n')
end

function m:dot(fmt, ...)
    self.dot_fh:write(string.format(fmt, ...))
end

-- Important for below!!!
math.randomseed(0)
-- Deterministic "UUID" generator so that we can compare output with diff
function m:uuid()
    -- return uuid.generate()
    return string.format('%.08x-%.04x-4%.03x-%.1x%.03x-%.08x%.04x',
        math.random(0, 2^32 - 1),
        math.random(0, 2^16 - 1),
        math.random(0, 2^12 - 1),
        math.random(0x8, 0xB),
        math.random(0, 2^12 - 1),
        math.random(0, 2^32 - 1),
        math.random(0, 2^16 - 1))
end

function m:node(meta, extra)
    extra = extra or {}
    extra.setup_item = extra.setup_item or extend({}, meta)
    local cat = meta.category
    if extra.try_module then
        if self.modules[cat] == nil then
            pcall(function()
                self.modules[cat] = require('setup_compiler/' .. cat)
            end)
            self.modules[cat] = self.modules[cat] or false
        end
        if not self.modules[cat] then
            return nil
        end
        extra.module = self.modules[cat]
        extra.try_module = nil
    end
    local node = compiler_node:new(self, extend({}, meta), extra)
    node.cal_xforms = self:normalize_calibration(meta.calibration)
    return node
end

function name_walker(t, names)
    local first = 2
    if t[1] == 'name' then
        local add = true
        for i = 1, #names do
            if names[i] == t[2] then
                add = false
                break
            end
        end
        if add then
            names[#names + 1] = t[2]
        end
    elseif t[1] == 'funcall' then
        first = 3 -- skip function name
    end
    for i = first, #t do
        if type(t[i]) == 'table' then
            name_walker(t[i], names)
        end
    end
    return names
end

function m:defaults()
    local net_db = network_db:new()
    local channel_defaults = net_db:available_channels('connector')
    net_db:close()

    for k, v in pairs(channel_defaults) do
        -- KLUDGE need warnings-as-errors flag!!!
        if not v.data_type then
            warn('MISSING data_type: %s => %s\n', v.connector, cjson.encode(v))
        elseif v.data_type:match('^sequential_') and not v.sample_rate then
            warn('MISSING sample_rate: %s => %s\n', v.connector,
                cjson.encode(v))
        end
    end

    self.summary_item = {}
    self.summary_idx = -1
    self.box_map = {}

    for i, item in ipairs(self.setup) do
        item.collect = boolean(item.collect)
        if item.category == 'summary' and self.summary_idx == -1 then
            self.summary_item = item
            self.summary_idx = i
        end

        local aux = {}
        if item.connector then
            -- KLUDGE we really need to do resolve_hostname() or something here
            local box = item.connector:match('^[^@]*@[^%.]+%.(.*)$')
            local is_internal
            if not box then
                box = item.connector:match('^.*%.internal$')
                if box then is_internal = true end
            end
            self.box_map[box] = self.box_map[box] or {}
            table.insert(self.box_map[box], item.connector)
            aux.box = box
            if not is_internal then
                local dch = channel_defaults[item.connector]
                if not item.connector:match('^@') then
                    if not dch and not item.connector:match('@aux') and
                                   not item.connector:match('@can') and
                                   not item.connector:match('@gps') then
                        errorf('No channel found: %s', item.connector)
                    end
                end
                extend(item, dch)
            end
            if item.category == nil then
                item.category = 'source'
            end

            if not item.data_type then
                self:err('MISSING data_type: %s => %s',
                    item.connector, cjson.encode(item))
            elseif item.data_type:match('^sequential_') and
                    not item.sample_rate then
                self:err('MISSING sample_rate: %s => %s',
                    item.connector, cjson.encode(item))
            end
        end

        if item.collect == nil and (source_re:match(item.category) or
                filter_re:match(item.category)) then
            item.collect = true
        end

        -- Turn degenerate "calibration": {...} case into [{...}]
        if item.calibration and not item.calibration[1] then
            item.calibration = {item.calibration}
        end

        -- KLUDGE Allow item.trigger_channel
        if item.trigger_channel and item.control_channels == nil then
            item.control_channels = {item.trigger_channel}
        end

        -- Turn degenerate "ch" into ["ch"]
        for _, tag in ipairs{'control_channels', 'input_channels'} do
            if item[tag] and type(item[tag]) ~= 'table' then
                item[tag] = {item[tag]}
            end
        end

        self.setup_aux[i] = aux
    end

    for i = 1, #self.setup do
        local item = self.setup[i]
        local aux = self.setup_aux[i]

        if self.box_map[aux.box] then
            item._channel_count = #self.box_map[aux.box]
        end
        for _, k in ipairs{'sample_rate_domain', 'purpose'} do
            if item[k] == nil and self.summary_item[k] then
                item[k] = self.summary_item[k]
            end
        end
    end

    -- KLUDGE this should be done as part of generalized validation
    -- after defaulting
    self.unique_columns = {name={}}
    for i, item in pairs(self.setup) do
        for key, col in pairs(self.unique_columns) do
            if item[key] ~= nil then
                if col[item[key]] ~= nil then
                    self:err('Duplicate %s %s: %s, %s',
                        key, item[key], cjson.encode(col[item[key]]),
                        cjson.encode({i, item.id}))
                else
                    self.unique_columns[key][item[key]] = {i, item.id}
                end
            end
        end
    end

    if #self.errors > 0 then
        error(table.concat(self.errors, '\n') .. '\n')
    end
end

function m:dump_state(msg)
    msg = msg or ''
    for _, k in ipairs{'sources', 'nodes', 'edges'} do
        dbg("%s%s => %s\n", msg, k, setup_spec:export(self[k]))
    end
end

function m:get_node(spec)
    if type(spec) == 'table' and self.nodes[spec.idx] == spec then
        return spec
    end
    if type(spec) == 'string' then
        -- KLUDGE maybe cache eventually
        for idx = 1, #self.nodes do
            if self.nodes[idx].meta.name == spec then
                return self.nodes[idx]
            end
        end
        errorf('No node with name "%s"', spec)
    elseif type(spec) == 'number' then
        if not self.nodes[spec] then
            errorf('No node with index "%d"', spec)
        end
        return self.nodes[spec]
    end
    errorf('No node with specifier "%s"', spec)
end

function m:connect(from_node, to_node)
    return compiler_edge:new(self, from_node, to_node)
end

function m:insert_node(node, old_edge)
    local from_node = self:get_node(old_edge.from_idx)
    local to_node = self:get_node(old_edge.to_idx)
    local new_edge = compiler_edge:new(self, node, to_node)
    old_edge.to_idx = node.idx
end

function m:prepend_node(node, prepended_node)
    -- Insert appended_node between node and its output nodes
    for _, edge in ipairs(self.edges) do
        if edge.to_idx == node.idx then
            edge.to_idx = prepended_node.idx
        end
    end
    self:connect(prepended_node, node)
end

function m:append_node(node, appended_node)
    -- Insert appended_node between node and its output nodes
    for _, edge in ipairs(self.edges) do
        if edge.from_idx == node.idx then
            edge.from_idx = appended_node.idx
        end
    end
    self:connect(node, appended_node)
end

function m:parse_expression(item, expr)
    -- First we do a "macro" pass
    -- The general form of a macro looks like this:
    -- @{channel_name.property_name}
    -- @{channel_name.property_name.subproperty_name}
    -- @{channel_name.property_name.subproperty_name.subproperty_name}
    -- subproperty_name can be a numeric array index
    -- Special cases
    -- @{.property_name} -- For this channel {i.e. item passed in}
    -- @{channel_name.1} -- Same as @{channel_name.input_channels.1}
    -- @{.1} -- Same as @{.input_channels.1}
    -- KLUDGE Currently this is all static. Lookup of runtime metadata
    -- would be better (or essential) in some cases. Proper dynamic lookup
    -- would require declaring the value at runtime and then assigning the
    -- value only once all metadata had come in from the proxies. Or by
    -- keeping it updated in the metadata handlers.

    expr = expr or item.expression
    repeat -- Loop until all macros containing macros get evaluated
        local old_expr = expr
        expr = expr:gsub('@{(.-)}',
            function(v)
                local path = {}
                while v ~= '' do
                    local p = v:match('^([^%.]*)%.?')
                    v = v:sub(#p + 2)
                    path[#path + 1] = tonumber(p) or p
                end
                -- Allow @(.1) for input_channels[1], etc.
                if type(path[2]) == 'number' then
                    table.insert(path, 2, 'input_channels')
                end
                local value = (path[1] == '')
                    and item or self.unique_columns.name[path[1]]
                assert(value ~= nil) -- KLUDGE proper error here
                for i = 2, #path do
                    value = value[path[i]]
                    assert(value ~= nil) -- KLUDGE proper error here
                end
                return value
            end)
    until old_expr == expr
    return cc_parser.parse_expression(expr)
end

function m:traverse()
    local in_chs = {}
    local nodes_needing_dt = {}
    local is_seq_num_rx = pcre:new('^(sequential_(u?int|float)\\d+)$')
    for i, item in pairs(self.setup) do
        local node = self:node(item, {
            try_module = true,
            setup_item = extend({}, item)
        })
        if node then
            in_chs[node.idx] = item.input_channels
            -- need to check later that these nodes have
            -- the correct data_type value
            nodes_needing_dt[#nodes_needing_dt + 1] = node
            if sink_re:match(item.category) then
                node.sink_post_propagate = true
            end
        elseif source_re:match(item.category) then
            node = self:node(item)
        elseif internal_re:match(item.category) then
            node = self:node(item)
        elseif filter_expression_re:match(item.category) then
            node = self:node(item)
            node.parse_tree = self:parse_expression(item)
            in_chs[node.idx] = name_walker(node.parse_tree, {})
            if not node.data_type and not node.meta.data_type and
               not node.is_message then
                nodes_needing_dt[#nodes_needing_dt + 1] = node
            end
        elseif sink_re:match(item.category) then
            node = self:node(item)
            in_chs[node.idx] = item.input_channels
            node.sink_post_propagate = true
        end
        if node and item.collect then
            local sink_node = self:node({category = 'sink/sie/time_history'},
                {setup_item=extend({}, item)})
            in_chs[sink_node.idx] = {item.name}
            if node.is_sequential and node.is_numeric then
                sink_node.node_to_collect = node
            end
        end
    end

    -- Connect
    for _, node in ipairs(self.nodes) do
        for i, name in ipairs(in_chs[node.idx] or {}) do
            self:connect(assert(self:get_node(name)), node)
        end
        in_chs[node.idx] = nil
    end
    in_chs = nil

    -- Assign data_type values for sequential nodes which
    -- don't have one.  "Upstream" nodes are checked for int/float
    -- and type size.
    for i, node in ipairs(nodes_needing_dt) do
        if node.module and node.module.is_sequential or
                filter_expression_re:match(node.meta.category) then
            local dt = node:upstream_meta('data_type')
            if #dt > 0 then
                node.meta.data_type = sample_types.to_sequential_type(dt[1])
            else
                errorf("no data type found upstream of or for '%s'",
                    node.meta.name)
            end
        end
    end

    -- Automagic sink/comet_stream insertion
    -- "live_data_stream": false is a master switch to turn all live data off
    --
    -- Truth table for "live_data_stream":
    --           T F U channel
    -- summary T L   L
    --         F
    --         U L
    --
    -- No live stream if node.internal_name set
    if not self.summary_item.live_data_stream then
        self.summary_item.live_data_stream = true
    end
    if self.summary_item.live_data_stream ~= false then
        local sum = self.summary_item.live_data_stream
        local comet_node
        for idx, node in ipairs(self.nodes) do
            local ch = node.meta.live_data_stream
            if not node.is_sink and not node.internal_name and
                    (sum or ch) and ch ~= false then
                if not comet_node then
                    local meta = {
                        category = 'sink/comet_stream',
                        data_type = 'message_raw', -- KLUDGE for propagate()
                    }
                    comet_node = self:node(meta, {try_module = true})
                end
                self:connect(node, comet_node)
            end
        end
    end

    -- Automagic sequentializer insertion
    for idx = 1, #self.nodes do
        local node = self.nodes[idx]
        if not node.is_sink and node.is_message and node.meta.sample_rate then
            local meta = extend({
                category = 'filter/sequentializer',
                collect = false,
                data_type = node.data_type:gsub('^message_', 'sequential_'),
            }, node.meta)
            local seq = assert(self:node(meta, {try_module = true}))
            node.meta.sample_rate = nil
            -- sequentializer node must get the name, not the messsage node
            node:set_internal_name()
            self:append_node(node, seq)
        end
    end

    -- Automagic sink/hbm_stream insertion
    -- "live_data_stream": false is a master switch to turn all live data off
    --
    -- Truth table for "live_data_stream":
    --           T F U channel
    -- summary T L   L
    --         F
    --         U L
    --
    -- No live stream if node.internal_name set
    if self.summary_item.live_data_stream ~= false then
        local sum = self.summary_item.live_data_stream
        local stream_node
        for idx, node in ipairs(self.nodes) do
            local ch = node.meta.live_data_stream
            if node.is_streamable and (sum or ch) and ch ~= false then
                if not stream_node then
                    local meta = {
                        category = 'sink/hbm_stream',
                        data_type = 'message_raw', -- KLUDGE for propagate()
                    }
                    stream_node = self:node(meta, {try_module = true})
                end
                self:connect(node, stream_node)
                self.stream_nodes[#self.stream_nodes + 1] = node
            end
        end
    end

    -- Automagic multi-sink splitting
    for idx = 1, #self.nodes do
        local node = self.nodes[idx]
        if node.sink_post_propagate then
            local edges = node.input_edges
            local md = extend({}, node.meta)
            for i = 1, #edges do
                local edge = edges[i]
                local in_ch = edge.from_node.meta.name
                local meta = extend({
                    somat_input_channel = in_ch,
                    name = in_ch .. '@' .. md.name,
                    input_channels = {in_ch},
                }, md, edge.from_node.meta)
                if node.module then
                    -- needed for TAL, at least
                    meta.label_name = md.name
                end
                if i == 1 then
                    node.meta = meta
                    node.setup_item = extend({}, meta)
                    node.setup_item.name = meta.name
                elseif node.module then
                    local dup_node = self:node(meta, {
                        try_module = true,
                        setup_item = extend({}, meta)
                    })
                    edge.to_idx = dup_node.idx
                else
                    local sink = assert(self:node(meta))
                    edge.to_idx = sink.idx
                end
            end
        end
    end

    -- Automagic trigger insertion
    for idx = 1, #self.nodes do
        local node = self.nodes[idx]
        local md = node.meta
        if md.category == 'filter/integrator' and  -- FIX use general form below
                md.trigger_channel then
            local trig_node = assert(self:get_node(md.trigger_channel))
            table.insert(node.meta.input_channels, md.trigger_channel)
            self:connect(trig_node, node)
        elseif std_trig_mode(md.trigger_mode) and
                md.category ~= 'filter/trigger' and
                md.trigger_mode ~= 'always_on' then
            local meta = extend({
                category = 'filter/trigger',
                collect = false,
            }, node.meta)
            local trig = assert(self:node(meta, {try_module = true}))
            self:prepend_node(node, trig)
            for _, name in ipairs(md.control_channels) do
                table.insert(trig.meta.input_channels, name)
                self:connect(assert(self:get_node(name)), trig)
            end
        elseif not std_trig_mode(md.trigger_mode) and
               md.trigger_channel then
            local trig_node = assert(self:get_node(md.trigger_channel))
            node.trig_edge = self:connect(trig_node, node)
        end
    end

    -- Automagic fanout insertion
    for idx = 1, #self.nodes do
        local node = self.nodes[idx]
        local outs = node.output_edges
        if #outs > 1 and node.module and not node.module.multi_fanout then
            local meta = {
                category = 'filter/fanout',
                collect = false,
                name = node.meta.name
            }
            local fanout = assert(self:node(meta, {try_module = true}))
            -- fanout node must get the name, not the module node
            node:set_internal_name()
            self:append_node(node, fanout)
        end
    end

    -- Automagic sink/null insertion
    for idx = 1, #self.nodes do
        local node = self.nodes[idx]
        if not node.is_sink and #node.output_edges == 0 then
            local meta = {
                category = 'sink/null',
            }
            local sink_null = assert(self:node(meta, {try_module = true}))
            self:append_node(node, sink_null)
        end
    end

    -- Propagate
    for idx, node in ipairs(self.nodes) do
        if node.is_source then
            self:propagate(nil, node)
        end
    end

    -- Calibration recombobulation
    -- First, move all calibration from source node to destination node
    for idx, node in ipairs(self.nodes) do
        if node:only_sink_node() then
            local outs = node.output_nodes
            assert(#outs == 1)
            local src = node.cal_xforms
            local dst = outs[1].cal_xforms
            for i = #src, 1, -1 do
                table.insert(dst, 1, table.remove(src, i))
            end
        end
    end
    for idx, node in ipairs(self.nodes) do
        self:finalize_calibration(node)
    end
end

function m:finalize_calibration(node)
    local m = 1
    local b = 0

    local xforms = node.cal_xforms
    for i = 1, #xforms do
        local xform = xforms[i]
        b = xform.m * b + xform.b
        m = xform.m * m
    end

    node.xform_expr = ''
    if m ~= 1 or b ~= 0 then
        node.finalized_xform = {linear = true, m = m, b = b}
        node.xform_expr = string.format(' * %.18e + %.18e', m, b)
        local item = node.meta
        if item.range_min then
            item._input_min = (item.range_min - b) / m
        end
        if item.range_max then
            item._input_max = (item.range_max - b) / m
        end
    end
end

function m:normalize_calibration(cals)
    local xform = {}
    cals = cals or {}
    for i = 1, #cals do
        xform[#xform + 1] = self:normalize_calibration_step(cals[i], true)
    end
    return xform
end

function m:normalize_calibration_step(in_cal, nil_if_identity)
    assert(type(in_cal) == 'table')
    local cal = setup_spec:normalize(in_cal)

    cal.linear = true
    local mode = cal.mode

    if mode == 'defined_two_point' or mode == 'experimental_two_point' then
        local e = cal.electrical
        local p = cal.physical
        cal.m = (p[2] - p[1]) / (e[2] - e[1])
        cal.b = p[1] - cal.m * e[1]
    elseif mode == 'defined_zero_span' or mode == 'shunt' then
        cal.m = cal.physical_span / cal.electrical_span
        cal.b = -cal.m * cal.electrical_zero
    elseif mode == 'defined_slope_intercept' or mode == nil then
        cal.m = cal.slope or 1
        cal.b = cal.intercept or 0
    elseif mode ~= 'strain_gage' and  -- strain_gage pass thru (for MX)
           mode ~= 'internal' then
        errorf('Unknown calibration.mode: %s calibration=%s',
            mode, cjson.encode(in_cal))
    end

    if cal.applied == true then
        return nil
    end
    if cal.m == 1 and cal.b == 0 then
        cal.identity = true
        if nil_if_identity then
            return nil
        end
    end
    return cal
end

m.propagate_keys = {
    'invalid_data_output_value',
    'sample_rate',
    'stale_data_expiration_time',
}

function m:propagate(input_node, node)
    if input_node then
        for _, k in ipairs(self.propagate_keys) do
            if node.meta[k] == nil then
                node.meta[k] = input_node.meta[k]
                if node.setup_item then
                    node.setup_item[k] = input_node.meta[k]
                end
            end
        end
    end
    if node.meta.sample_rate then
        -- allow for sample_rate adjustments caused by "node"
        -- being an "up" or "down" sampler computed channel
        node.meta.factor = tonumber(node.meta.factor)
        if node.meta.adjusted_sample_rate then
            node.meta.sample_rate = node.meta.adjusted_sample_rate
        elseif node.meta.factor then
            local r = rational:new(node.meta.sample_rate):tonumbers()
            if node.meta.category == 'filter/up_sampler' then
                node.meta.sample_rate = r * node.meta.factor
            elseif node.meta.category == 'filter/down_sampler' then
                node.meta.sample_rate = r / node.meta.factor
            end
        end
        node.float_sample_rate =
            rational:new(node.meta.sample_rate):tonumbers()
    end
    if node.is_message then
        local timescale = 1 / sample_clock.master_rate -- KLUDGE!!!
        node.meta._timescale = timescale
        if node.setup_item then
            node.setup_item._timescale = timescale
        end
    end
    for _, output_node in ipairs(node.output_nodes) do
        self:propagate(node, output_node)
    end
end

function m:escape_string(s)
    return s -- KLUDGE FOR NOW
end

function m:compile()
    if self.jit_dump or self.summary_item.jit_dump then
        self:template({output = 'preamble'}, [==[
        |require('jit.dump').start('+rsxT', '(=file_base=).dump')
        ]==], self)
    end
    self:template({output = 'preamble'}, [==[
    |collectgarbage('setpause',2000)
    |local pristine_setup = [========[
    |(=json=)]========]
    |local normalized_setup = [========[
    |(=pretty_json=)]========]
    ]==], self)
    self:template({output = 'preamble'}, statics.preamble, {})

    for _, edge in ipairs(self.edges) do
        local from_node = edge.from_node
        local buffer_time = 0.5 -- KLUDGE constant
        local rb_size = edge_buf:ringbuf_size(math.max(8192,
            -- KLUDGE try 'message_rate' in here as well for device defaults!!!
            (from_node.float_sample_rate or 4096) * buffer_time)) -- KLUDGE
        if from_node.is_message and not from_node.is_fixed_size then
            -- KLUDGE different rb_size for fixed-size message channels ????
            -- KLUDGE need size hints from channel metadata ???
            --datatype = 'uint8_t' -- KLUDGE?
            --rb_size = 0x8000 -- KLUDGE!!!
        end
        if not from_node:only_sink_node() then
            self:template({output = 'edge_bufs'}, [==[
            |bufs[(=edge.idx=)] = edge_buf:new('(=edge.data_type=)',
            |+   (=rb_size=)) -- (=edge.from_name=) -> (=edge.to_name=)
            ]==], {
                edge = edge,
                rb_size = rb_size,
            })
        end
    end

    self.sw = sie_writer:new{file_name='/dev/null',
        setup_compiler_mode=true, max_bufs=#self.nodes}
    self.sie_buf_idx = self.sw.last_buf + 1

    self.omit_sie = false -- KLUDGE eventually change for default
    for i, node in ipairs(self.nodes) do
        if node.is_sink_sie and not node.sie_meta then
            local item = extend({}, node.setup_item)
            item.data_type = node.data_type
            node.sie_meta = self.sw:channel(item)
            node.sie_buf, node.sie_buf_idx = self:oc_sie_buf()
            self.omit_sie = false
        end
    end
    -- Right now, only allow omit_sie to be forced from the summary item
    if self.summary_item.omit_sie ~= nil then
        self.omit_sie = self.summary_item.omit_sie == true
    end

    self:oc_finalize([==[
    |function m:finalize()
    ]==])

    self:compile_init()
    self:compile_process()

    self:oc_finalize([==[
    |end -- finalize()
    ]==])

    self:template({output = 'postamble'}, statics.postamble, {})
end

function m:compile_init()

    self:default_buffer('init_2')
    for i, node in ipairs(self.nodes) do
        if node.module then
            assert(node.module)
            node.mod = node.module:cm_new{compiler = self, node = node}
            if node.mod.cm_declare then
                node.mod_name = node.mod:cm_declare()
            end
            if node.mod.cm_initialize then
                self:push_indent(4)
                local tv = {}
                tv.mod_initialize = node.mod:cm_initialize(node.mod_name)
                self:template([==[
                |(=mod_initialize=)
                ]==], tv, node)
                self:pop_indent()
            end
        elseif node.is_source then
            local ch = node.meta
            assert(ch.connector)
            -- KLUDGE this sizing will only work for sequential data
            local buffer_time = 0.5 -- KLUDGE BROKEN BAD CONSTANT
            local tv = {
                buffer_time = buffer_time,
                cjson_md = cjson.encode(ch),
                data_type = ch.data_type,
                buf_size = 8192, -- HIDEOUS BAD KLUDGE - need hint
            }
            if node.is_sequential and node.is_fixed_size then
                tv.buf_size = math.max(8192,
                    math.ceil(node.float_sample_rate * tv.buffer_time))
            end

            if node.is_sequential then
                if node:only_sink_node() then
                    self:oc_source_sequential_to_sie(node, tv)
                elseif node.is_fixed_size then
                    self:oc_source_sequential_to_bufs(node, tv)
                else
                    errorf('NYI: unhandled case for %s', tv.data_type)
                end
            elseif node.is_message then
                if node.is_fixed_size then
                    if node:only_sink_node() and node.has_heartbeats then
                        self:oc_source_fixed_message_to_sie(node, tv)
                    elseif node:only_sink_node() then
                        self:oc_source_fixed_message_no_hb_to_sie(node, tv)
                    else
                        self:oc_source_fixed_message_to_bufs(node, tv)
                    end
                else
                    if node:only_sink_node() then
                        self:oc_source_var_message_to_sie(node, tv)
                    else
                        self:oc_source_var_message_to_bufs(node, tv)
                    end
                end
            else
                errorf('NYI: unhandled case for %s', tv.data_type)
            end
        elseif node.is_sink_sie then
            self:oc_init_sink_sie(node, {})
        end
    end

    self:default_buffer('init_1')

    -- Make the stream buffers
    for h_idx, n in ipairs(self.stream_nodes) do
        self:template({output = 'hbm_stream_bufs'}, [==[
        |hbm_bufs[(=idx=)] = hbm_stream_buf:new('(=data_type=)',
        |+   (=sample_rate=))  -- (=name=)
        ]==], {
            idx = h_idx,
            data_type = n.data_type,
            name = n.meta.name or 'unknown',
            sample_rate = n.sample_rate or n.meta.sample_rate or 0
        })
    end

    local start_stream_server = self.summary_item.purpose ~=
        "digital_display"

    self:template([==[
    |function m:init(args)
    |    local md
    |    local node
    |    self.setup = cjson.decode(normalized_setup)
    |    self.summary = self.setup[(=summary_idx=)] or {}
    |    self.test_id = args[3] or self.summary.test_id or uuid.generate()
    |    self.test_name = self.summary.test_name or 'untitled_setup'
    |    self.test_title = self.summary.test_title or 'Untitled Setup'
    |    self.sie_name = args[1] or self.test_name .. '.sie'
    |    self.friendly_sie_name = args[4] or self.sie_name
    |    self.specified_test_duration = tonumber(args[2])
    |    self.interrupted_restart = (args[5] and true) or nil
    |    self:system_init((=start_stream_server=))
    |    self.stop_latch = quorum_latch:new(#proxy_buffers)
    ]==], {start_stream_server = start_stream_server}, self)

    if not self.omit_sie then
        self:template([==[
        |    self.sw = sie_writer:new{file_name=self.sie_name,
        |+       max_bufs=(=sie_buf_idx=)}
        |    sie_bufs = self.sw.bufs
        |    self.sw:group_tag('somat:sxr_setup',
        |+       'application/json', pristine_setup)
        |    self.sw:group_tag_from_file('somat:system_version',
        |+       'application/json', '/etc/version.json')
        |    self.sw:group_tag_from_file('somat:compiled_test_program',
        |+       'text/plain', source_filename)
        ]==], self)
    end

    self:template([==[
    |    -- KLUDGE set start time from arguments here
    |    log:notice(__'test_start "{0}" {1} at {2}', self.test_name,
    |+       self.test_id, start_time)
    |    net_db:begin_transaction()
    ]==])

    self:default_buffer('init_3')
    self:template([==[
    |    net_db:commit()
    |    net_db:close()
    ]==])
    if not self.omit_sie then
        self:template([==[
        |    self.sw:test_header(0, 1, 'Fake Run Description', start_time)
        |    self.sw:xml()
        ]==])
    end
    self:template([==[
    |    local duration = self.specified_test_duration or
    |        self.summary.test_duration
    |    if duration then
    |        self:schedule_test_stop(duration)
    |    end
    |end -- m:init()
    |
    ]==])
    self:default_buffer(nil)
end

function m:compile_process()

    self:default_buffer('process')

    self:template([==[
    |function m:process(done)
    |    local more = false
    ]==])
    self:indent('    ')
    for i, node in ipairs(self.nodes) do
        if not node.is_source and not node.only_sink_cleanup then
            self:template([==[
            |do -- ((=meta.category=)) (=meta.name=)
            ]==], node)
            if node.module then
                self:oc_process_module(node)
            elseif node.is_message and node.is_fixed_size then
                self:oc_process_fixed_message(node)
            elseif node.is_message and not node.is_fixed_size then
                self:oc_process_variable_message(node)
            elseif node.is_sequential and node.is_fixed_size then
                self:oc_process_fixed_sequential(node)
            else
                errorf('NYI: unhandled case in process()')
            end
            self:template([==[
            |end
            ]==])
        end
    end
    self:indent('')
    self:template([==[
    |    return more
    |end
    ]==])

    self:default_buffer(nil)
end

return m
