-- down_sampler.lua - Implements the Down Sampler computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local sample_clock = require 'sample_clock'

local math_min = math.min

local ctor_cache = {}

local mt = {}


function mt:dump()
    dbg('out_count=%lld', self.out_count)
end

function mt:run(in_buf, count, out_buf)
    -- Decimate input values by 'factor' times
    -- (i.e., only put every 'factor' input value
    --  in the output ringbuffer)
    local doable = math_min(count, out_buf:available())
    for i = 0, doable - 1 do
        local sample = in_buf:deq()
        -- decimation_counter keeps track of when to allow an
        -- input sample to be passed to out_buf.
        if sample and sample.value then
            if self.decimation_counter % self.factor == 0 then
                out_buf:enq_value(sample.value)
                self.decimation_counter = 1
            else
                self.decimation_counter = self.decimation_counter + 1
            end
        end
    end
    -- Return true if any samples were output from this node
    return doable > 0
end

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type = string.format('struct %s_down_sampler', t)
        ffi.cdef(string.format([[
            %s {
                float64_t factor;
                float64_t decimation_counter;
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, factor)
    local down_sampler
    local ctor = get_ctor(data_type)
    down_sampler = ctor()
    down_sampler.factor = tonumber(factor)
    down_sampler.decimation_counter = down_sampler.factor
    return down_sampler
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    if args.node.meta.factor < 1 then
        error('INVALID down sample factor: %s < 1\n',
              tostring(args.node.meta.factor))
    end
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- this filter only makes sense when attached to one input node
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1) -- KLUDGE
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+   '(=factor=)')
    ]==], self.node, self.node.meta, {})
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count, (=output_edges.1.buf=)
    |+       )) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
