
local dbg = require 'dbg'
local ffi = require 'ffi'

local math_min = math.min

-- per issue #2439, SomatXR Integrator needs to support the same
-- triggering options as the eDAQ Integrator, for at least all real
-- (i.e., floating point) input channels.  Summary of the functionality
-- required (copied from issue #2439):
--
-- Trigger / Reset Options - A (logical) trigger channel can be
-- specified for either or both of the following two purposes.
--
-- 1) Integrate Only When True - Check this option to have the
--    integration suppressed when the trigger channel is FALSE.
--
-- 2) Enable Triggered Reset - Check this option to have the Integrator
--    reset (back to the specified Initial Value) whenever the trigger
--    channel satisfies the selected Reset Mode.
--
-- There are three trigger reset modes available.
--
--     a) When True - Reset whenever the trigger channel is TRUE.
--
--     b) On False-True Edge - Reset when the trigger channel
--        transitions from FALSE to TRUE.
--
--     c) After True-False Edge - Reset on the sample after the trigger
--        channel transitions from TRUE to FALSE. (If the Integrate Only
--        When True option is not selected, then the output sample after
--        the True to False edge is the sum of the Initial Value and the
--        scaled input sample.)
--
--     Note - When both the Integrate Only When True and the Enable
--     Triggered Reset options are selected, the Reset Mode is limited
--     to After True-False Edge.
--
-- Reset on Sum Exceeds - Check this option to have the Integrator
--    reset to the Initial Value whenever the absolute value of the
--    Integrator sum exceeds the (positive) value specified in the
--    associated edit field.
--
-- Note - If the Reset on Sum Exceeds option is selected then none of
-- the other Trigger / Reset Options can be used.

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('value=%f, triggered=%s', self.value[0], self.triggered)
end

-- KLUDGE run_sequential_(), run_message() ???
-- KLUDGE or mt_sequential:run(), mt_message:run()

-- sequential only case
function mt:run(in_buf, trig_buf, out_buf)
    local doable =
        math_min(in_buf.count, out_buf:available())
    if trig_buf then
        doable = math_min(doable, trig_buf.count)
    end
    local reset_mode = ffi.string(self.triggered_reset_mode)
    for i = 0, doable - 1 do
        local pre_reset = false
        local post_reset = false
        local integrate = true
        local sample = in_buf:deq()
        local prev_value = self.value[0]
        local integrated_value = prev_value + sample.value[0] *
            self.scale_factor
        -- Reset On Sum Exceeds trumps other trigger/reset options
        if self.reset_when_sum_exceeds ~= 0 then
            if math.abs(integrated_value) > self.sum_maximum_value then
                pre_reset = true
            end
        elseif trig_buf then
            local prev_triggered = self.triggered
            local trigger_input = assert(trig_buf:deq(), "No trigger sample")
            self.triggered = (trigger_input.value[0] ~= 0) and 1 or 0
            if self.integrate_only_when_true ~= 0 then
                integrate = (self.triggered ~= 0)
            end
            if self.enable_triggered_reset ~= 0 then
                if reset_mode == 'when_true' then
                    pre_reset = (self.triggered ~= 0)
                elseif reset_mode == 'on_false_to_true_edge' then
                    pre_reset = (prev_triggered == 0) and
                        (self.triggered ~= 0)
                elseif reset_mode == 'after_true_to_false_edge' then
                    post_reset = (prev_triggered ~= 0) and
                        (self.triggered == 0)
                end
            end
        end
        -- set output value
        if pre_reset then
            self.value[0] = self.initial_value
        elseif integrate then
            self.value[0] = integrated_value
        else
            self.value[0] = prev_value
        end
        local os = out_buf:enq()
        os[0].value[0] = self.value[0]
        if post_reset then
            self.value[0] = self.initial_value
        end
    end
    -- Return true if any samples were output from this node
    return doable > 0
end

local function get_ctor(t)
    if not ctor_cache[t] then
        --The following isn't used for the length of the
        --triggered_reset_mode array (a literal is used instead) -
        --couldn't make it work with luajit:
        --local max_mode_size = math.max(#'when_true',
        --                               #'on_false_to_true_edge',
        --                               #'after_true_to_false_edge')
        local struct_type = string.format('struct %s_integrator', t)
        ffi.cdef(string.format([[
            %s {
                float64_t initial_value;
                float64_t scale_factor;
                float64_t sum_maximum_value;
                uint8_t   reset_when_sum_exceeds;
                uint8_t   integrate_only_when_true;
                uint8_t   enable_triggered_reset;
                uint8_t   triggered_reset_mode[30];
                float64_t value[1];
                int       triggered;
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, initial_value, scale_factor,
               reset_when_sum_exceeds, sum_maximum_value,
               integrate_only_when_true, enable_triggered_reset,
               triggered_reset_mode)
    -- KLUDGE make sure data type is sequential here until message integrator
    local ctor = get_ctor('sequential', initial_value, scale_factor)
    local obj = ctor()
    obj.initial_value = tonumber(initial_value) or 0
    obj.scale_factor = tonumber(scale_factor) or 1
    obj.sum_maximum_value = tonumber(sum_maximum_value) or 0
    obj.reset_when_sum_exceeds = reset_when_sum_exceeds or 0
    obj.integrate_only_when_true = integrate_only_when_true or 0
    obj.enable_triggered_reset = enable_triggered_reset or 0
    if triggered_reset_mode and
       type(triggered_reset_mode) == 'string' then
        ffi.copy(obj.triggered_reset_mode, triggered_reset_mode)
    end
    obj.value[0] = obj.initial_value
    obj.triggered = 0
    return obj
end

-- setup-compiler module support

local cm_mt = {__index = m}

-- KLUDGE m:cm_check(), validate, etc...???

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- KLUDGE need self.compiler:assert(cond, __"msg", arg, arg, ...)
    assert(self.node.is_numeric, "Node must be a numeric data type")
    assert(self.node.is_sequential, "Node must be a sequential data type")
    assert(#self.node.input_edges >= 1
        and #self.node.input_edges <= 2) -- KLUDGE
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1) -- KLUDGE

    local tv =
        {initial_value = tonumber(self.node.meta.initial_value) or 0,
         scale_factor = tonumber(self.node.meta.scale_factor) or 1,
         reset_when_sum_exceeds =
             self.node.meta.reset_when_sum_exceeds or false,
         sum_maximum_value =
             self.node.meta.sum_maximum_value or 0, -- validated later
         integrate_only_when_true =
             self.node.meta.integrate_only_when_true or false,
         enable_triggered_reset =
             self.node.meta.enable_triggered_reset or false,
         triggered_reset_mode =
             self.node.meta.triggered_reset_mode or false}
                

    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+ (=initial_value=), (=scale_factor=), (=reset_when_sum_exceeds=),
    |+ (=sum_maximum_value=), (=integrate_only_when_true=),
    |+ (=enable_triggered_reset=), '(=triggered_reset_mode=)')
    ]==], tv, self.node, self.node.meta)
    return name, value
end

function m:cm_run(name)
    local output_edges = self.node.output_edges
    local input_edges = self.node.input_edges
    -- KLUDGE--v if multiple non-trigger input channels are ever
    --           allowed the following must be fixed.
    local actual_input_buf = input_edges[1] and input_edges[1].buf
    assert(actual_input_buf, "Integrator channel has no input buffer")
    local trigger_buf = nil
    if self.node.meta.trigger_channel then
        trigger_buf = input_edges[#input_edges].buf
        assert(trigger_buf, "Trigger channel has no input buffer")
    end

    -- Perform some input validation.  See "SomatXR Test Setup
    -- File Format" document for details about valid Integrator
    -- Specific Computed Channel Parameter combinations.
    assert(tonumber(self.node.meta.scale_factor) ~= 0,
           "scale_factor must not be zero")
    if self.node.meta.reset_when_sum_exceeds then
        assert(not trigger_buf and
               not self.node.meta.enable_triggered_reset and
               not self.node.meta.integrate_only_when_true,
               "Cannot use other trigger/reset options when " ..
               "Reset On Sum Exceeds selected")
        local max = tonumber(self.node.meta.sum_maximum_value)
        local initial = tonumber(self.node.meta.initial_value) or 0
        assert(max and (max > initial),
               string.format("Reset On Sum Exceeds requires maximum " ..
                   "value (%s) greater than initial value (%i)",
                   max, initial))
    else
        if trigger_buf then
            assert(self.node.meta.enable_triggered_reset or
                   self.node.meta.integrate_only_when_true,
                   "Having a trigger channel requires either Enable" ..
                   " Triggered Reset or Integrate Only When True set")
            if self.node.meta.integrate_only_when_true and
               self.node.meta.enable_triggered_reset then
                -- The following restriction is required by issue #2439,
                -- if not by the "SomatXR Test Setup File Format"
                -- document.
                assert(self.node.meta.triggered_reset_mode ==
                           'after_true_to_false_edge',
                       "only After True-False Edge reset mode " ..
                       "allowed when both Integrate Only When True " ..
                       "and Enable Triggered Reset selected")
            end
        else
            assert(not self.node.meta.enable_triggered_reset,
                   "triggered reset not allowed without a " ..
                   "trigger channel")
            assert(not self.node.meta.integrate_only_when_true,
                   "Integrate Only When True not allowed without a " ..
                   "trigger channel")
        end
        if self.node.meta.enable_triggered_reset then
            local mode = self.node.meta.triggered_reset_mode
            assert(mode == 'when_true' or
                   mode == 'on_false_to_true_edge' or
                   mode == 'after_true_to_false_edge',
                   string.format("Invalid triggered reset mode: %s", mode))
        else
            assert(self.node.meta.triggered_reset_mode == nil,
                   "triggered reset mode not allowed unless Enable " ..
                   "Triggered Reset is true")
        end
    end

    local tv = {cm_name = name,
                actual_input_buf = actual_input_buf,
                trigger_buf = trigger_buf or false,
                output_edge_buf = output_edges[1].buf}
    local code = self.compiler:template({output = false, inline = true}, [==[
    |            more = ((=cm_name=):run((=actual_input_buf=), (=trigger_buf=),
    |+                  (=output_edge_buf=))) or more
    ]==], tv, self.node)
    return code
end

return m
