require "strict"

require('lj_include'){
    'stdio',
    'stdlib',
    'string',
    'unistd', 
    'sys/types',
    'signal',
    'linux/time',
    'sys/time',
    'sys/resource', 
} 

local dbg = require 'dbg'
local ffi = require 'ffi'

--[====[
From: https://devzero.somat.com/redmine/issues/3002

    The Interactive Trigger computed channel will largely emulate the
    eDAQ version of this computed channel.

Requirements document:  Interactive Trigger Computed Channel.pdf (attached
to https://devzero.somat.com/redmine/issues/3002).
--]====]

local C = ffi.C

-- Table containing executable test module.
local mt = {}

local struct_def = [[
    void * calloc(size_t nmemb, size_t size);
    void free(void *ptr);

    struct interactive_trigger {
        uint8_t  invert;
        uint8_t  triggered;
        uint8_t* id;
    };
]]

ffi.cdef(struct_def)
local ctor = ffi.metatype('struct interactive_trigger', {__index = mt})

local network = nil     -- set by m:new()
local static_obj = nil  -- set by m:new()

-- debugging utility function
function mt:dump()
    dbg('triggered=%s, id=%s, invert=%s',
        tonumber(self.triggered) == 0 and "off" or "on",
        ffi.string(self.id),
        tonumber(self.invert) == 0 and "false" or "true")
end

-- Get current trigger status from network database.
function mt:query_trigger()
    local triggered = 0
    local owner = '_interactive-trigger-' .. ffi.string(self.id)
    local rec = static_obj:get_rr(owner)
    local trig_val = false
    if rec then
        trig_val = rec[3]:match("^.-:%s*(.*)$")
        if self.invert ~= 0 then
            -- invert sense of trigger (i.e., "true" really means "false")
            triggered = trig_val == 'true' and 0 or 1
        else
            triggered = trig_val == 'true' and 1 or 0
        end
    else
        -- Take improper number of records for the specified trigger to
        -- mean a trigger condition of "false".  No record existing could
        -- be normal, meaning trigger hasn't been used.  Too many records
        -- existing is an error condition, which we'll take to mean "not
        -- triggered" (for now).
        if self.invert ~= 0 then
            -- invert sense of trigger (i.e., "true" really means "false")
            triggered = 1
        else
            triggered = 0
        end
    end
    return triggered
end

-- Execute interactive_trigger object with any new input.
function mt:run(in_buf, out_buf)
    local count = in_buf.count
    if count > 0 then
        self.triggered = self:query_trigger()
        in_buf:clear()  -- don't need the sample values (they just
                        -- indicate nbr of output values needed)
        for i = 1, count do
            out_buf:enq().value[0] = self.triggered
        end
    end
end

function mt:finalize()
    -- free all C-allocated memory
    ffi.C.free(self.id)
end

-- Table used during test generation.
local m = {}

-- Returns a new interactive_trigger object
function m:new(id, invert, network_obj, static_obj_in)
    static_obj = static_obj or static_obj_in
    network = network or network_obj
    local obj = ctor()
    obj.triggered = 0
    -- C-style allocation used here to avoid strange behavior
    -- if cdata constructor is used instead:
    obj.id = ffi.C.calloc(#id + 1, ffi.sizeof('uint8_t'))
    ffi.copy(obj.id, id)
    obj.invert = invert == 'true' and 1 or 0
    return obj
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1, "Fan-in failed")
    assert(#self.node.output_edges == 1, "Fan-out failed")
    assert(not self.node.meta.invert_trigger or
        (tostring(self.node.meta.invert_trigger) == 'false' or
         tostring(self.node.meta.invert_trigger) == 'true'),
        "Invalid value for attribute 'invert_trigger' (" ..
        tostring(self.node.meta.invert_trigger) .. ")")
    assert(self.node.meta.trigger_index,
        "Value for attribute 'trigger_index' does not exist")
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=trigger_index=)',
    |+  '(=invert_trigger=)', network, m)  -- network defined in statics
    ]==], self.node, self.node.meta)
    return name, value
end

function m:cm_initialize()
    local code = self.compiler:template({output = false}, [==[
    |m:bind_changed_rrs()
    |m:monitor_rr('_interactive-trigger-' .. '(=trigger_index=)')
    |m:update_monitored_rr_values('_interactive-trigger-' .. '(=trigger_index=)')
    ]==], self.node, self.node.meta)
    return code
end

function m:cm_run(name)
    local code = self.compiler:template({output = false}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=output_edges.1.buf=))) or more
    ]==], {cm_name = name}, self.node)
    return code
end

function m:cm_finalize(name)
    local code = self.compiler:template({output = false}, [==[
    |(=cm_name=):finalize()
    ]==], {cm_name = name})
    return code
end

return m
