local dbg = require 'dbg'
local ffi = require 'ffi'
local sample_clock = require 'sample_clock'

local math_min = math.min

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('out_count=%lld', self.out_count)
end

function mt:run(in_buf, count, out_buf)
    local out_available = out_buf:available()
    local doable = math_min(count, out_available)
    for i = 0, doable - 1 do
        local sample = in_buf:deq()
        while self.clock.current_clock < sample.timestamp[0] do
            self.clock:tick()
            if self.clock.current_clock > self.stale_after then
                self.value[0] = self.stale_value[0]
            end
            if sample.timestamp[0] >= 0 then
                local os = out_buf:enq()
                assert(os, 'buffer overflow on resample')
                os[0] = self.value[0]
            end
        end
        self.last_stamp = sample.timestamp[0]
        if not sample:is_heartbeat() then
            -- How stale_after is set depends on exactly how stale_clocks
            -- is calculated.
            self.stale_after = self.clock.next_clock + self.stale_clocks -- ???
            self.value[0].value[0] = sample.value[0]
        end
    end
    -- Return true if any samples were output from this node
    return doable > 0
end

local function seq_ctor(t)
    if not ctor_cache[t] then
        local seq_type = string.format('struct %s_seq', t)
        ffi.cdef(string.format([[
            %s {
                sample_clock clock;
                %s value[1];
                %s stale_value[1];
                int64_t stale_clocks;
                int64_t last_stamp;
                int64_t stale_after;
                int64_t out_count;
                //int64_t hb_value;
            };
        ]], seq_type, t, t))
        ctor_cache[t] = ffi.metatype(seq_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, rate, stale_after, stale_value, start_stamp)
    local ctor = seq_ctor(data_type)
    local seq = ctor()
    seq.clock:init(rate, sample_clock.master_rate)
    -- There are various ways of calculating stale_clocks depending on
    -- how it's used at runtime.
    seq.stale_clocks
        = math.floor(sample_clock.master_rate * tonumber(stale_after)) - 1
    local sv = tonumber(stale_value)
    if sv then
        seq.stale_value[0].value[0] = sv
    else
        error('Non-numeric sequentializer not yet supported')
    end
    seq.value[0] = seq.stale_value[0]
    return seq
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1) -- KLUDGE
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1) -- KLUDGE
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+   '(=sample_rate=)',
    |+   '(=stale_data_expiration_time=)',
    |+   '(=invalid_data_output_value=)')
    ]==], self.node, self.node.meta)
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count, (=output_edges.1.buf=)
    |+       )) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
