-- time_base_shifter.lua - Implements the Time Base Shifter
-- computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local sample_clock = require 'sample_clock'

local math_min = math.min

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('out_count=%lld', self.out_count)
end

function mt:run(in_buf, count, out_buf)
    local out_available = out_buf:available()
    if count < 1 then
        return false  -- no more left to do
    elseif out_available < 1 then
        return true   -- more to do, but can't do it now
    end

    -- NOTE:  A shift of zero is disallowed, causing an error
    --        to be thrown by the setup compiler (before this
    --        is ever executed).
    if self.direction_flag > 0 then
        -- shift direction is "lag", so write to the output buffer
        -- the requested number of copies of the 1st input sample
        -- (plus one, to account for the zero'th sample)
        local in_index_start = 0
        local nbr_1st_out = 0
        if (self.shift_count + 1 - self.nbr_samples_shifted) > 0 then
            nbr_1st_out =
                math_min(out_available, self.shift_count + 1 -
                                        self.nbr_samples_shifted)
            if self.nbr_samples_shifted == 0 then
                self.prefix_sample = in_buf:deq()[0]
                in_index_start = 1
            end
            for i = 1, nbr_1st_out do
                out_buf:enq_value(self.prefix_sample)
            end
            self.nbr_samples_shifted =
                self.nbr_samples_shifted + nbr_1st_out 
        end
        -- write as many other input samples to the output buffer
        -- as we can
        local input_doable = math_min(count - in_index_start,
                                      out_available - nbr_1st_out)
        for in_index = in_index_start, input_doable - 1 do
            local sample = in_buf:deq()[0]
            out_buf:enq_value(sample.value)
        end
    elseif self.direction_flag < 0 then
        -- shift direction is "lead", so ignore input samples until
        -- the shift "period" is exceeded
        local doable = math_min(count, out_available)
        for in_index = 0, doable - 1 do
            local sample = in_buf:deq()[0]
            if self.nbr_samples_shifted < self.shift_count then
                self.nbr_samples_shifted =
                    self.nbr_samples_shifted + 1
            else
                out_buf:enq_value(sample.value)
            end
        end
    end
    -- Return true since samples were output from this node
    return true
end

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type =
            string.format('struct %s_time_base_shifter', t)
        ffi.cdef(string.format([[
            %s {
                float64_t  shift_count;
                float64_t  direction_flag;
                float64_t  nbr_samples_shifted;
                %s         prefix_sample;
            };
        ]], struct_type, t))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, shift_count, shift_direction)
    local time_base_shifter
    local ctor = get_ctor(data_type)
    time_base_shifter = ctor()
    time_base_shifter.shift_count = tonumber(shift_count)
    if shift_direction == "lead" then
        time_base_shifter.direction_flag = -1
    else
        time_base_shifter.direction_flag = 1
    end
    time_base_shifter.nbr_samples_shifted = 0
    ffi.fill(time_base_shifter.prefix_sample, ffi.sizeof(data_type))
    return time_base_shifter
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    if tonumber(args.node.meta.shift_count) <= 0 or
       tonumber(args.node.meta.shift_count) > 10000 then
        error(string.format('INVALID shift value: %s\n',
              tostring(args.node.meta.shift_count)))
    end
    if args.node.meta.shift_direction ~= "lead" and
       args.node.meta.shift_direction ~= "lag" then
        error(string.format('INVALID shift direction: %s\n',
              tostring(args.node.meta.shift_direction)))
    end
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- this filter only makes sense when attached to one input node
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges > 0)
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+   '(=shift_count=)', '(=shift_direction=)')
    ]==], self.node, self.node.meta, {})
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count, (=output_edges.1.buf=)
    |+       )) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
