-- time_channel.lua - Implements the Time computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local dioconst = require 'dioconst'

local math_min = math.min

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('time=%lld', self.time)
end

function mt:run(in_buf, out_buf)
    local out_available = out_buf:available()
    local doable = math_min(in_buf.count, out_available)
    for i = 0, doable - 1 do
        local sample = in_buf:deq()
        if not sample:is_heartbeat() then
            local os = out_buf:enq()
            assert(os, 'buffer overflow in time channel')
            os[0].value[0] = tonumber(self.time) *
                tonumber(self.scale_factor)
            local hs
            self.time = self.time + 1
        end
    end
    -- Return true if any samples were output from this node
    return doable > 0
end

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type = string.format('struct %s_time_channel', t)
        ffi.cdef(string.format([[
            %s {
                uint64_t  time;
                float64_t scale_factor;
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:is_sequential()
    return true
end

function m:new(data_type, scale_factor)
    local ctor = get_ctor(data_type)
    local time_channel = ctor()
    time_channel.time = 0
    time_channel.scale_factor = scale_factor
    return time_channel
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    local attributes = args or {}
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1)
    local sample_rate =
        rational:new(self.node.meta.sample_rate):tonumbers()
    assert(sample_rate and sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(self.node.meta.sample_rate)))
    local tv = {data_type = self.node.data_type or
        self.node.meta.data_type,
        category = self.node.meta.category,
        scale_factor = self.node.meta._timescale or 1 / sample_rate}
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+    (=scale_factor=))
    ]==], tv)
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=output_edges.1.buf=))) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
