
local ffi = require 'ffi'
local math = require 'math'
local rational = require 'rational'

--[====[
From: https://devzero.somat.com/redmine/issues/2997

    The Timed Trigger computed channel will largely emulate the eDAQ
    version of this computed channel. However, it will be simplified
    in that it will not support conditional options on the delay and
    sustain time periods (at least in the first offering).

Requirements document:  Timed Trigger Computed Channel.pdf (attached
to https://devzero.somat.com/redmine/issues/2997).

--]====]

local mt = {}

ffi.cdef( [[
    struct timed_trigger {
        int64_t start_mode;
        int64_t delay_samples;
        int64_t end_samples;
        int64_t elapsed_after_event_start;
        int64_t prev_input;
        int64_t state;
        int64_t output_value;
    };
]] )

local ctor = ffi.metatype('struct timed_trigger', {__index = mt})

local WHEN_TRUE = 1
local ON_FALSE_TO_TRUE_EDGE = 2
local AFTER_TRUE_TO_FALSE_EDGE = 3

local ST_LOOKING_FOR_TRIGGER = 1
local ST_DELAY_WAIT = 2
local ST_OUTPUT_TRIGGER = 3

function mt:dump()
end

local m = {}

-- Returns a new timed_trigger object.
function m:new(data_type, sample_rate, start_mode,
               delay_period, sustain_period  )
    local rate = rational:new(sample_rate):tonumbers()
    local obj = ctor()
    obj.start_mode = (start_mode == "when_true" and
        WHEN_TRUE) or (start_mode ==
        "on_false_to_true_edge" and ON_FALSE_TO_TRUE_EDGE) or
        (start_mode == "after_true_to_false_edge" and
        AFTER_TRUE_TO_FALSE_EDGE)
    obj.delay_samples =
        math.floor(rate * delay_period + 0.5)
    if tonumber(obj.start_mode) == AFTER_TRUE_TO_FALSE_EDGE then
        -- account for the "after" part of this mode
        obj.delay_samples = obj.delay_samples + 1
    end
    local sustain_samples =
        math.floor(rate * sustain_period + 0.5)
    sustain_samples = (sustain_samples == 0) and 1 or sustain_samples
    obj.end_samples = obj.delay_samples + sustain_samples
    obj.elapsed_after_event_start = 0  -- at sample rate
    obj.prev_input = 0
    if tonumber(obj.start_mode) == ON_FALSE_TO_TRUE_EDGE then
        -- prevent spurious trigger at first input
        obj.prev_input = 1
    end
    obj.state = ST_LOOKING_FOR_TRIGGER
    obj.output_value = 0
    return obj
end

-- Execute timed trigger object with any new input.
function mt:run(in_buf, out_buf)
    local count = in_buf.count
    local sample = nil
    local out_value = 0
    for i = 1, count do
        sample = in_buf:deq()
        if sample and sample.value then
            -- Note that the following state engine can progress
            -- through the states from first state to last without
            -- any delay between transitions.  This is only true
            -- once per loop iteration - in other words, a transition
            -- to an earlier state requires another loop iteration.

            self.output_value = 0  -- pre-set

            if tonumber(self.state) == ST_LOOKING_FOR_TRIGGER then
                if tonumber(self.start_mode) == WHEN_TRUE then
                    if tonumber(sample.value[0]) ~= 0 then
                        self.state = ST_DELAY_WAIT
                        self.elapsed_after_event_start = 0
                    end
                elseif tonumber(self.start_mode) == ON_FALSE_TO_TRUE_EDGE then
                    if tonumber(self.prev_input) == 0 and
                            tonumber(sample.value[0]) ~= 0 then
                        self.state = ST_DELAY_WAIT
                        self.elapsed_after_event_start = 0
                    end
                elseif tonumber(self.start_mode) ==
                        AFTER_TRUE_TO_FALSE_EDGE then
                    if tonumber(self.prev_input) ~= 0 and
                            tonumber(sample.value[0]) == 0 then
                        self.state = ST_DELAY_WAIT
                        self.elapsed_after_event_start = 0
                    end
                end
            end

            if tonumber(self.state) == ST_DELAY_WAIT then
                if self.elapsed_after_event_start >= self.delay_samples then
                    self.state = ST_OUTPUT_TRIGGER
                end
            end

            if tonumber(self.state) == ST_OUTPUT_TRIGGER then
                if self.elapsed_after_event_start < self.end_samples then
                    self.output_value = 1
                else  -- finished with this triggered occurrence
                    self.state = ST_LOOKING_FOR_TRIGGER
                    self.output_value = 0
                end
            end

            -- update ticker for this triggered occurrence
            self.elapsed_after_event_start =
                self.elapsed_after_event_start + 1

            self.prev_input = sample.value[0]
            -- briefly reuse sample.value for convenience
            sample.value[0] = self.output_value
            out_buf:enq_value(sample.value)
        end
    end
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- KLUDGE need self.compiler:assert(cond, __"msg", arg, arg, ...)
    assert(self.node.is_numeric, "Node must be a numeric data type")
    assert(self.node.is_sequential, "Node must be a sequential data type")
    assert(#self.node.input_edges == 1,
        string.format("invalid number of input_edges:  %s",
        tostring(#self.node.input_edges))) -- KLUDGE
    assert(#self.node.output_edges >= 1,
        string.format("invalid number of output_edges:  %s",
        tostring(#self.node.output_edges))) -- KLUDGE
    assert(self.node.meta.trigger_start_mode == "when_true" or
        self.node.meta.trigger_start_mode == "on_false_to_true_edge" or
        self.node.meta.trigger_start_mode == "after_true_to_false_edge",
        "Invalid trigger_start_mode value: " ..
        tostring(self.node.meta.trigger_start_mode))
    assert(self.node.meta.trigger_delay_period >= 0,
        "Invalid trigger_delay_period value: " ..
        tostring(self.node.meta.trigger_delay_period))
    assert(self.node.meta.trigger_sustain_period >= 0,
        "Invalid trigger_sustain_period value: " ..
        tostring(self.node.meta.trigger_sustain_period))

    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+ '(=sample_rate=)', '(=trigger_start_mode=)',
    |+ (=trigger_delay_period=), (=trigger_sustain_period=))
    ]==], self.node, self.node.meta)
    return name, value
end

function m:cm_run(name)
    local tv = {}
    tv.output_edges = self.node.output_edges
    tv.cm_name = name
    local code = self.compiler:template({output = false}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+        (=output_edges.1.buf=))) or more
    ]==], tv, self.node)
    return code
end

return m
