local dbg = require 'dbg'
local ffi = require 'ffi'

--[====[
    From: https://devzero.somat.com/redmine/issues/2176

    We need to extend support to include the same triggering options that
    are available in the eDAQ (i.e., "always_on", "trigger", "gate", and
    "one_shot").

    Always On: Do not use triggering. Data sampling is always on from the
    start of the test.

    Trigger: Data sampling starts when the trigger channel becomes TRUE.
    Once the trigger channel is TRUE, data sampling runs continuously,
    irrespective of any future changes in the trigger channel.

    Gate: Data sampling occurs if and only if the trigger channel is TRUE.
    Data sampling stops when the trigger channel is FALSE.

    One Shot: Take a single data sample when the trigger channel transitions
    from FALSE to TRUE or if the trigger channel is TRUE on the first sample
    of any run.

    In the eDAQ, only logical channels are allowed for the trigger channel,
    but since there is currently no way to designate that a channel is a
    logical channel, any channel can be used for the trigger channel in the
    short term at least.
--]====]

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('triggered=%s', self.triggered)
end

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type = string.format('struct %s_trigger', t)
        ffi.cdef(string.format([[
            %s {
                int triggered;
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

m.multi_fanout = true

function m:new(data_type)
    local ctor = get_ctor('')
    local obj = ctor()
    obj.triggered = 0
    return obj
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- KLUDGE need self.compiler:assert(cond, __"msg", arg, arg, ...)
    assert(self.node.is_numeric, "Node must be a numeric data type")
    assert(self.node.is_sequential, "Node must be a sequential data type")
    assert(#self.node.input_edges >= 1
        and #self.node.input_edges <= 2) -- KLUDGE
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1) -- KLUDGE
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)')
    ]==], self.node, self.node.meta)
    return name, value
end

function m:cm_run(name)
    local tv = {cm_name = name}
    local o = {}
    local input_edges = self.node.input_edges
    local output_edges = self.node.output_edges
    local trigger_mode = self.node.meta.trigger_mode
    local control_edge = input_edges[#input_edges]
    tv.control_buf = control_edge.buf
    tv.control_from_name = control_edge.from_name
    o[#o + 1] = self.compiler:template({output = false}, [==[
    |    -- trigger_mode '(=meta.trigger_mode=)'
    |    local count = (=processable_count=)
    |    for i = 0, count - 1 do
    |        local v = (=control_buf=):deq()
    |+           -- trigger channel (=control_from_name=)
    ]==], tv, self.node)
    tv.v = 'v[0].value[0]'
    if trigger_mode == 'always_on' then
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |        if true then -- KLUDGE this node should have been elided
        ]==], tv, self.node)
    elseif trigger_mode == 'gate' then
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |        if (=v=) ~= 0 then
        ]==], tv, self.node)
    elseif trigger_mode == 'one_shot' then
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |        local do_one_shot = false
        |        if (=cm_name=).triggered == 0 and (=v=) ~= 0 then
        |            do_one_shot = true
        |            (=cm_name=).triggered = 1
        |        elseif (=v=) == 0 then
        |            (=cm_name=).triggered = 0
        |        end
        |        if do_one_shot then
        ]==], tv, self.node)
    elseif trigger_mode == 'trigger' then
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |        if (=cm_name=).triggered == 0 and (=v=) ~= 0 then
        |            (=cm_name=).triggered = 1
        |        end
        |        if (=cm_name=).triggered ~= 0 then
        ]==], tv, self.node)
    end
    -- KLUDGE the following assumes input_edges[i] corresponds to 
    -- output_edges[i]. Naming corresponding edges would probably
    -- be better. In any case, this is not currently an issue as
    -- no existing sinks allow more than one input edge.
    for i = 1, #input_edges - 1 do
        tv.in_edge = input_edges[i]
        tv.out_edge = output_edges[i]
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |            (=out_edge.buf=):enq()[0] = (=in_edge.buf=):deq()[0]
        ]==], tv, self.node)
    end
    o[#o + 1] = self.compiler:template({output = false}, [==[
    |        else
    ]==], tv, self.node)
    for i = 1, #input_edges - 1 do
        tv.in_edge = input_edges[i]
        tv.out_edge = output_edges[i]
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |            (=in_edge.buf=):deq()
        ]==], tv, self.node)
    end
    o[#o + 1] = self.compiler:template({output = false}, [==[
    |        end
    |    end
    |    more = more or count > 0
    ]==], tv, self.node)
    return table.concat(o)
end

return m
