-- up_sampler.lua - Implements the Up Sampler computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local sample_clock = require 'sample_clock'
local dioconst = require 'dioconst'

local math_min = math.min

local ctor_cache = {}

local mt = {}

function mt:dump()
    dbg('out_count=%lld', self.out_count)
end

function mt:run(in_buf, count, out_buf)
    local out_available = out_buf:available()
    local doable = math_min(count, out_available)
    for i = 0, doable - 1 do
        local sample = in_buf:deq()
        for j = 0, self.factor - 1 do
            out_buf:enq_value(sample.value)
        end
    end
    -- Return true if any samples were output from this node
    return doable > 0
end

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type = string.format('struct %s_up_sampler', t)
        ffi.cdef(string.format([[
            %s {
                float64_t factor;
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, factor)
    local up_sampler
    local ctor = get_ctor(data_type)
    up_sampler = ctor()
    up_sampler.factor = tonumber(factor)
    return up_sampler
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    if args.node.meta.factor < 1 then
        error('INVALID up sample factor: %s < 1\n',
              tostring(args.node.meta.factor))
    else
        local max_factor = dioconst.data_rate_maximums
                           [args.compiler.summary_item['sample_rate_domain']]
        if args.node.meta.factor > max_factor then
            warn('TOO LARGE factor: %s, max valid is %s\n',
                 tostring(args.node.meta.factor),
                 tostring(self.summary_item['sample_rate_domain']))
        end
    end
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- this filter only makes sense when attached to one input node
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1) -- KLUDGE
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+   '(=factor=)')
    ]==], self.node, self.node.meta, {})
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count, (=output_edges.1.buf=)
    |+       )) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
