-- time_channel.lua - Implements the Time computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local ts = require 'timestamps'

local math_min = math.min

local ctor_cache

local mt = {}

function mt:dump()
    dbg('elapsed time=%f', self.last_time - self.start_time)
end

-- Note, we don't really use the time values coming from the system time
-- channel here.  We actually get the times from the hardware so system time
-- changes that don't propagate to the system time don't interfer with the
-- elapsed run time.
function mt:run(in_buf, out_buf)
    local doable = 0
    while #in_buf > 0 do
        local out_available = out_buf:available()
        local doable = math_min(1, out_available)
        for i = 0, doable - 1 do
            local sample = in_buf:deq()
            if not sample then break end
            local os = out_buf:enq()
            assert(os, 'buffer overflow in ert channel')
            if self.start_time == 0 then
                self.start_time = ts.get_ts().s
            end
            self.last_time = ts.get_ts().s
            os[0].value[0] = self.last_time - self.start_time
            if self.end_time > 0 and self.last_time >= self.end_time then
                os[0].value[0] = self.end_time - self.start_time
            end
            if os[0].value[0] < 0 then os[0].value[0] = 0 end
        end
    end
    return doable > 0
end

local function get_ctor()
    if not ctor_cache then
        local struct_type = string.format('struct ert_channel')
        ffi.cdef(string.format([[
            %s {
                float64_t start_time;
                float64_t end_time;
                float64_t last_time;
            };
        ]], struct_type))
        ctor_cache = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache
end

local m = {}

function m:new(start_time, end_time)
    local ctor = get_ctor()
    local ert = ctor()
    ert.start_time = start_time
    ert.end_time = end_time or 0
    return ert
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    local attributes = args or {}
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1)
    local tv = {data_type = self.node.data_type or
        self.node.meta.data_type,
        category = self.node.meta.category}
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new(ts.iso_to_epoch(start_time), end_time)
    ]==], tv)
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=output_edges.1.buf=))) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
