-- digital_output.lua - Implements the Digital Out channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local C = ffi.C
local rational = require 'rational'
local net = require 'net'
local StreamBuffer = net.StreamBuffer
local cjson = require 'cjson'
local http_parser = require 'http_parser'
local network_db = require 'network_db'
local netdb = network_db:new()
local ev = require 'ev'
local log = require 'log'

local math_min = math.min

local ctor_cache = {}

local on = true
local off = false

-- constant values used in this file
local state = {
    ON = 0,
    OFF = 1,
    NO_ACTION = 2
}
local mode = {
    SET_UNLATCHED = 0,
    CLEAR_UNLATCHED = 1,
    SET_LATCHED = 2,
    CLEAR_LATCHED = 3
}

local power = {
    DO_NOT_SET = 0,
    OFF = 1, 
    FIVE_VOLTS = 2,
    TWELVE_VOLTS = 3
}

local power_to_str = { "off", "5v", "12v" }

-- Function used to retrieve the ZCIP assigned IPv4 address of the internal
-- network interface.  (This should probably be put in a utilities or net file?)
local function get_zcip_addr()
    local addrs = net.getifaddrs('ethint')
    for _, addr in ipairs(addrs) do
        if addr:find('^169%.254%.') then
            return addr
        end
    end
end

local mt = {}

function mt:dump()
    return dbg([=[output_port_nbr=%lld,
        output_mode=%lld, power_on=%lld, started=%lld,
        latched=%lld, initial_state=%lld, final_state=%lld]=],
        self.output_port_nbr, self.output_mode,
        self.power_on, self.started,
        self.latched, self.initial_state, self.final_state)
end

function mt:_log_err(msg)
    log.error(msg)
end

-- The function mt:write_output_port(true/false) sends a request to
-- the dio_proxy via the dio output socket to turn on (true) or
-- off (false) the dio output port.
--
-- Structure of json sent to dio_proxy:
--    {"connector":"<connector name>","request":"set_output",
--     "params":["<output port number>","<setting>"]}
-- where "<setting>" is either "on" or "off".
--
function mt:write_output_port(on)
    local json
    if on then
        json = {connector = ffi.string(self.connector),
            request = "set_output",
            params = {tostring(self.output_port_nbr), "on"}}
    else
        json = {connector = ffi.string(self.connector),
            request = 'set_output',
            params = {tostring(self.output_port_nbr), 'off'}}
    end
    local enc_json = cjson.encode(json)
    ffi.copy(self.json_buf, enc_json, #enc_json)
    self.json_buf[#enc_json] = 0
    C.write(self.sock, self.json_buf, #enc_json + 1)
end

-- The function mt:run(inbuf) dequeues and processes all the
--     current entries in "inbuf".
--
-- Also, the first time mt:run is called it will perform the
--    following test initialization as needed:
--    1)  If the test specifies a transducer power setting a request
--    will be sent to the dio_proxy.  The structure of the json sent:
--    {"connector":"<connector name>","request":"transducer_power",
--     "params":["<power setting>"]}
--    where "<power setting>" is "off", "5v", or "12v".
--    2)  A request will be sent to the dio_proxy to set the
--    output port as specified for the initial state, via a call to
--    self:write_output_port().
function mt:run(in_buf)
    -- Need to do "first time" (aka startup) processing?
    if self.started == 0 then
        self.started = 1
        if self.power_status ~= 0 then
            -- Set the transducer power as requested
            local json = {connector = ffi.string(self.connector),
                request = 'transducer_power',
                params = {power_to_str[self.power_status]}}
            local enc_json = cjson.encode(json)
            ffi.copy(self.json_buf, enc_json, #enc_json)
            self.json_buf[#enc_json] = 0
local tmpnbr =            C.write(self.sock, self.json_buf, #enc_json + 1)
        end
        -- Start the port in the user-specified state
        if self.initial_state == state.ON then
            self:write_output_port(on)
        else
            -- the default initial port state is closed (i.e., off)
            self:write_output_port(off)
        end
    end

    -- Determine update value for digital out port,
    -- processing all input samples available.
    local in_count = in_buf.count
    if in_count <= 0 then
        return in_buf.count > 0
    end
    local input_true = false
    for i = 1, in_count do
        local sample = in_buf:deq()
        -- a single true input means the whole input sample
        -- window must be treated as true
        if not input_true and sample and sample.value then
            input_true = sample.value[0] ~= 0
        end
    end
    if self.latched ~= 0 then
        -- Output has been latched, so no need to do any further
        -- processing.
        return in_buf.count > 0
    end
    local port_on = false
    -- determine desired port status
    if self.output_mode == mode.CLEAR_UNLATCHED then
        port_on = not input_true
    elseif self.output_mode == mode.SET_LATCHED then
        if input_true then
            port_on = true
            self.latched = 1
        end
    elseif self.output_mode == mode.CLEAR_LATCHED then
        if input_true then
            port_on = false
            self.latched = 1
        end
    else  -- "set_sink_unlatched" is the default mode
        port_on = input_true
    end
    -- set port status as specified
    self:write_output_port(port_on)

    return in_buf.count > 0
end

function mt:cleanup()
    -- Leave the port in the user-specified state
    if self.final_state == state.ON then
        self:write_output_port(on)
    elseif self.final_state == state.OFF then
        self:write_output_port(off)
    end -- the default final port state is "no action"
end

-- (The following buffer sizes are multiples of 8 to avoid
--     any possible problems due to byte alignment.)
-- Size of buffer used to save the connector name, the
--     value of which is sent as part of the json packet
--     requesting an output port change.
ffi.cdef[[
    static const int CONN_NAME_BUF_SIZE = 64;
]]
-- Size of buffer containing the json packet send to
--     request a write (change) to a DIO output port.
ffi.cdef[[
    static const int JSON_BUF_SIZE = 128;
]]

local function get_ctor(t)
    if not ctor_cache[t] then
        local struct_type = string.format('struct %s_digital_output', t)
        ffi.cdef(string.format([[
            %s {
                uint8_t     connector[CONN_NAME_BUF_SIZE];
                float64_t   output_port_nbr;        // treat as constant
                float64_t   output_mode;            // treat as constant
                float64_t   initial_state;          // treat as constant
                float64_t   final_state;            // treat as constant
                float64_t   power_status;           // treat as constant
                float64_t   started;          // "first time" flag
                float64_t   sock;             // socket for dio_proxy
                float64_t   latched;
                uint8_t     json_buf[JSON_BUF_SIZE];
            };
        ]], struct_type))
        ctor_cache[t] = ffi.metatype(struct_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(data_type, output_port_nbr, output_mode, power_status,
               initial_state, final_state, connector)
    assert(#connector < C.CONN_NAME_BUF_SIZE,
        "connector (" .. tostring(connector) .. ") name too long")
    local samples_to_check = 1
    local ctor = get_ctor(data_type)
    local digital_output = ctor()

    -- set fields used to send dio output ports requests to dio_proxy
    local orig_host = connector:gsub('^.*%@', '')
-- FIX when proxy "from-test-to-proxy" addresses are published:
--    local ctrl_host, ctrl_port, ctrl_ip
--        = assert(netdb:resolve_service(orig_host))
--    assert(connector:gsub('^.*%@dio[.]', '') == ctrl_host,
--        string.format("Invalid host for connector (%s ~= %s)",
--        tostring(connector:gsub('^.*%@dio[.]', '')),
--        tostring(ctrl_host)))
--    local s = net.socket{ip = ctrl_ip, port = ctrl_port}:connect()
    local internal_addr = get_zcip_addr()
    local s = net.socket{ip = internal_addr,
        port = tonumber(65535 - 100)}:connect()
    if not s then
        log.error("net.socket connect failed")
        return
    end
    ffi.copy(digital_output.connector, connector)
    digital_output.sock = s.sock
    digital_output.output_port_nbr = output_port_nbr
    digital_output.output_mode = output_mode
    digital_output.power_status = power_status
    digital_output.started = 0
    digital_output.latched = 0      -- default to port unlatched
    digital_output.initial_state = initial_state
    digital_output.final_state = final_state

    return digital_output
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    -- this filter only makes sense when attached to one input node
    assert(#self.node.input_edges == 1)
    local from_node_name = self.node.input_edges[1].from_node.meta.name
    local from_node = assert(self.compiler:get_node(from_node_name))
    assert(from_node.meta.is_logical,
        "input channel must be of logical type")
    assert(self.node.meta.connector, "missing output connector")
    local output_port = string.match(self.node.meta.connector,
        "^output_(%d)")
    output_port = tonumber(output_port)
    assert(output_port and output_port >= 1 and output_port <= 2,
        "invalid output connector")
    -- convert string parameters to number constants, for ease of
    -- storage in digital_output's cdef
    local output_mode_enum = 0
    local output_mode = self.node.meta.output_mode or "set_sink_unlatched"
    if output_mode == "set_sink_unlatched" then
        output_mode_enum = mode.SET_UNLATCHED 
    elseif output_mode == "clear_sink_unlatched" then
        output_mode_enum = mode.CLEAR_UNLATCHED 
    elseif output_mode == "set_sink_latched" then
        output_mode_enum = mode.SET_LATCHED 
    elseif output_mode == "clear_sink_latched" then
        output_mode_enum = mode.CLEAR_LATCHED 
    else
        error('INVALID mode value: %s\n',
              tostring(self.node.meta.output_mode))
    end
    local power_status_enum = 0
    local power_status = self.node.meta.transducer_power
    if power_status == "off" then
        power_status_enum = power.OFF 
    elseif power_status == "5v" then
        power_status_enum = power.FIVE_VOLTS 
    elseif power_status == "12v" then
        power_status_enum = power.TWELVE_VOLTS 
    elseif power_status then
        error('INVALID transducer power value: %s\n',
              tostring(self.node.meta.transducer_power))
    end
    local final_state_enum = 0
    local final_state = self.node.meta.final_state or "no_action"
    if final_state == "high" then
        final_state_enum = state.OFF
    elseif final_state == "low" then
        final_state_enum = state.ON
    elseif final_state == "no_action" then
        final_state_enum = state.NO_ACTION 
    else
        error('INVALID "final state" value: %s\n',
              tostring(self.node.meta.final_state))
    end
    local initial_state_enum = 0
    local initial_state = self.node.meta.initial_state or "high"
    if initial_state == "high" then
        initial_state_enum = state.OFF
    elseif initial_state == "low" then
        initial_state_enum = state.ON
    else
        error('INVALID "initial state" value: %s\n',
              tostring(self.node.meta.initial_state))
    end
    -- generate code to declare the module
    local tv = {
        output_port_nbr = output_port,
        output_mode = output_mode_enum,
        power_status = power_status_enum,
        initial_state = initial_state_enum,
        final_state = final_state_enum
        }
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new('(=data_type=)',
    |+   (=output_port_nbr=), (=output_mode=), (=power_status=),
    |+   (=initial_state=), (=final_state=), '(=connector=)')
    ]==], tv, self.node, self.node.meta)
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=))) or more
    ]==], {cm_name = name}, self.node)
    return code
end

function m:cm_finalize(name)
    local code = self.compiler:template({output = false}, [==[
    |    (=cm_name=):cleanup()
    ]==], {cm_name = name})
    return code
end

return m
