local cjson = require 'cjson'
local ffi = require 'ffi'
local rational = require 'rational'

local dbg = require 'dbg'

local m = require 'setup_compiler/sink/sie/burst_history_runtime'

-- setup-compiler module support

local cm_mt = {__index = m}

-- This table (as output code written to the test file) is
-- used to cache identifiers related to a data type, to avoid
-- some unnecessary duplication in the sie XML.
local cached_ids_decl = 'local cached_burst_hist_ids = {}'
-- This flag is used to prevent multiple occurrences of the
-- 'cached_ids_decl' code.
local mod_declared = false

-- returns a new module object used to create Burst History
-- data mode objects
function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

-- creates test runtime code that instantiates a Burst History
-- data mode object
function m:cm_declare()
    -- #self.node.input_edges > 1 due to trigger channel
    assert(self.node.meta.sample_rate, "Missing sample_rate")
    assert(self.node.trig_edge, "Missing trigger channel")
    assert(tonumber(self.node.meta.burst_pre_trigger_time),
        "Invalid burst_pre_trigger_time")
    assert(tonumber(self.node.meta.burst_post_trigger_time),
        "Invalid burst_post_trigger_time")
    assert(self.node.meta.trigger_mode == "when_true" or
           self.node.meta.trigger_mode == "on_false_to_true_edge",
        "Missing or invalid trigger_mode")

    local tv = {}
    tv.data_type = self.node.data_type

    -- ensure the 'cached_burst_hist_ids' table is declared in test code
    if not mod_declared then
        mod_declared = true
        self.compiler:oc_buffered('decl', {no_lhs = true},
            cached_ids_decl)
    end

    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)_runtime'):new('(=data_type=)',
    |+ cached_burst_hist_ids)
    ]==], tv, self.node.meta)

    return name, value
end

-- creates test runtime code that initializes a Burst History
-- data mode object
function m:cm_initialize(mod_name)
    local input_node = self.compiler:get_node(self.node.meta.input_channels[1])
    local trigger_node = self.compiler:get_node(self.node.meta.trigger_channel)
    local code = {}
    local tv = {}

    tv.input_channel_name = input_node.meta.name
    tv.trigger_channel_name = trigger_node.meta.name
    tv.channel_description =
        self.node.sie_meta['core:description'] or ""
    tv.sample_rate = rational:new(self.node.meta.sample_rate):tonumbers()
    tv.scale_factor = 1 / tv.sample_rate
    tv.pre_count = tonumber(self.node.meta.burst_pre_trigger_time) *
        tv.sample_rate
    local post_count = tonumber(self.node.meta.burst_post_trigger_time) *
        tv.sample_rate
    tv.max_burst_count = tv.pre_count + post_count + 1
    tv.sie_buf_size = 8192
    tv.sie_element_size = ffi.sizeof(self.node.data_type) +
        ffi.sizeof('float64_t')

    tv.data_bits = self.node.sie_meta['somat:data_bits']
    tv.data_format = self.node.sie_meta['somat:data_format']
    tv.dim0_label = self.node.sie_meta._dims[1]['core:label'] or 'Time'
    tv.dim0_units = self.node.sie_meta._dims[1]['core:units'] or 'Seconds'
    local x_label = self.node.sie_meta._dims[2]['core:label'] or ''
    local tmp_label = x_label:match('^(.-)@.*$')
    x_label = tmp_label and tmp_label or x_label
    tv.dim1_label = x_label
    tv.dim1_units = self.node.sie_meta._dims[2]['core:units'] or
        self.node.physical_units or ''
    tv.dim1_units = tostring(tv.dim1_units)
    tv.cjson_setup_meta = self.node.sie_meta['somat:armada_input_metadata']

    self.node.sie_buf, self.node.sie_buf_idx = self.compiler:oc_sie_buf()

    code[#code + 1] = self.compiler:template({output = false}, [==[
    |
    |do
    |    local mod = (=mod_name=)
    |    local name = "(=name=)" -- KLUDGE
    |
    |    local tv = {}
    |    tv.datamode_name = name
    |    tv.scale_factor = (=scale_factor=)
    |    tv.input_channel_name = "(=input_channel_name=)" -- KLUDGE
    |    tv.channel_description = "(=channel_description=)"
    |    tv.data_type = "(=data_type=)"
    |    tv.data_bits = (=data_bits=)
    |    tv.data_format = "(=data_format=)"
    |    tv.dim0_label = "(=dim0_label=)"
    |    tv.dim0_units = "(=dim0_units=)"
    |    tv.dim1_label = "(=dim1_label=)"
    |    tv.dim1_units = "(=dim1_units=)"
    |
    |    local burst_base = mod:base_channel(self.sw, tv)
    |
    |    tv.name = name
    |    tv.sample_rate = (=sample_rate=)
    |    tv.base = burst_base
    |    tv.ch_id = self.sw:next_id('ch')
    |    self.sw:new_buffer((=sie_buf_idx=), nil,
    |        (=sie_buf_size=), (=sie_element_size=),
    |        (=sizeof'timestamp'=)) --  => (=_sie_name=)
    |    tv.grp = (=sie_buf=).group
    |    mod:burst_channel(self.sw, tv)
    |    mod:initialize((=max_burst_count=), (=sie_buf=),
    |+       (=pre_count=), "(=trigger_mode=)", (=scale_factor=),
    |+       "(=input_channel_name=)")
    |end
    ]==], tv, self.node.sie_meta, self.node, self.node.meta)

    return table.concat(code)
end

-- creates test runtime code that calls (executes) a Burst History
-- data mode object
function m:cm_run(name)
    local code = self.compiler:tmpl([==[
    |        (=cm_name=):run((=input_edges.1.buf=), (=trig_edge.buf=))
    ]==], {cm_name = name}, self.node)
    return code
end

-- creates test runtime code to finalize a Burst History data mode object
function m:cm_finalize(name)
    local code = self.compiler:template({output = false}, [==[
    |do
    |    (=cm_name=):finalize()
    |    (=sie_buf=):flush()
    |end
    ]==], {cm_name = name}, self.node)
    return code
end

return m
