local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log'
local sample_types = require 'sample_types'
local sie_writer = require 'sie_writer'

local dbg = require 'dbg'

local RING_BUF_EXTRA_ALLOWED = 1100
local RING_BUF_PRUNE = RING_BUF_EXTRA_ALLOWED - 100

local circular_buf_t_str = "burst_%s_circular_buf_t"
local circular_buf_def_str = [[
    typedef struct burst_%s_circular_buf_struct {
        %s *buf;
    } %s; ]]
local circ_cdefined = {}

-- runtime burst_history module support

local ctor_cache = {}

-- Used to create a decoder and a base channel description in the SIE
-- file which can be extended by multiple Burst History data mode
-- channels. (NOTE that the time dimension's scale is not defined here.)
-- The decoder definition takes into account that each row of the
-- burst history data contains 3 values:
-- 1) timestamp (counting from test start) of the collected data value,
-- 2) the collected data value,
-- 3) the time offset of the collected data value from the event which
--    triggered the burst collection.
-- <sample/> returns those three values in v0, v1, v2.
local burst_base_channel = [==[
    |<decoder id="(=decoder_id=)">
    | <read var="v0" bits="64" type="int" endian="big"/>
    | <read var="v2" bits="64" type="int" endian="big"/>
    | <loop var="v0">
    |  <read var="v1" bits="(=data_bits=)" type="(=c_data_type=)" endian="big"/>
    |  <sample/>
    |  <set var="v2" value="{$v2 + 1}"/>
    | </loop>
    |</decoder>
    | <ch id="(=ch_id=)" name="(=data_type=)@burst_history_base"
    |+ private="base">
    |  <tag id="somat:datamode_name">(=datamode_name=)</tag>
    |  <tag id="somat:datamode_type">burst_history</tag>
    |  <tag id="somat:data_bits">(=data_bits=)</tag>
    |  <tag id="somat:data_format">(=data_format=)</tag>
    |  <tag id="core:description">Base Channel for (=data_type=)</tag>
    |  <tag id="data_type">(=data_type=)</tag>
    |  <tag id="core:schema">somat:burst</tag>
    |  <dim index="0">
    |   <tag id="core:description">Burst History Time</tag>
    |   <tag id="core:label">(=dim0_label=)</tag>
    |   <tag id="core:units">(=dim0_units=)</tag>
    |   <data decoder="(=decoder_id=)" v="0"/>
    |   <xform scale="(=scale_factor=)" offset="0"/>
    |  </dim>
    |  <dim index="1">
    |   <tag id="core:description">Burst History Data</tag>
    |   <tag id="core:label">(=dim1_label=)</tag>
    |   <tag id="core:units">(=dim1_units=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |  </dim>
    |  <dim index="2">
    |   <tag id="core:description">Burst Index</tag>
    |   <tag id="core:label">index</tag>
    |   <tag id="core:units">seconds</tag>
    |   <data decoder="(=decoder_id=)" v="2"/>
    |  </dim>
    | </ch>
    ]==]

-- used to create a Burst History data mode channel in SIE file
-- KLUDGE - test id needs fixing?
local burst_history_channel = [==[
    | <ch test="0" id="(=ch_id=)" base="(=base=)"
    |+ group="(=grp=)" name="(=name=)">
    |  <tag id="somat:input_channel">(=input_channel_name=)</tag>
    |  <tag id="core:sample_rate">(=sample_rate=)</tag>
    |  <tag id="core:description">(=channel_description=)</tag>
    | </ch>
    ]==]

ffi.cdef[[
    typedef struct
        {
            timestamp stamp;
            timestamp offset;
        } bhist_timestamp_t;
]]

local mt = {}

-- create a base channel description in the SIE file, which is
-- extended by multiple Burst History data mode channels
function mt:base_channel(sw, itv)
    local tv = {}
    tv.data_type = self.data_type

    if not self.cached_ids.base_channel_id or
       not self.cached_ids.base_channel_id[self.data_type] then
        if not self.cached_ids.base_channel_id then
            self.cached_ids.base_channel_id = {}
            self.cached_ids.decoder_id = {}
        end
        if not self.cached_ids.base_channel_id[self.data_type] then
            self.cached_ids.base_channel_id[self.data_type] =
                sw:next_id('ch')
            self.cached_ids.decoder_id[self.data_type] =
                sw:next_id('decoder')
        end
        tv.decoder_id = self.cached_ids.decoder_id[self.data_type]
        tv.ch_id = self.cached_ids.base_channel_id[self.data_type]
        local c_type = string.match(self.data_type, "^.-_(.-)%d+$")
        assert(c_type, "Missing decoder type")
        tv.c_data_type = c_type
        tv.scale_factor = self.scale_factor
        sw:xml()
        sw:xml(burst_base_channel, tv, itv)
        sw:xml()
    else
        tv.ch_id = self.cached_ids.base_channel_id[self.data_type]
    end

    return tv.ch_id
end

-- create a Burst History data mode channel description in the SIE file
function mt:burst_channel(sw, itv)
    local decoder_id = self.cached_ids.decoder_id[self.data_type]
    sw:xml()
    sw:xml(burst_history_channel, itv, {decoder_id = decoder_id})
    sw:xml()
    return itv.ch_id
end

function mt:dump()
    return string.format("data type = %s, sie buf = %s\n",
        self.data_type, tostring(self.sie_buf))
end

-- assigns the attributes for this Burst History data mode object
-- which need to be assigned at test runtime
function mt:initialize(max_count, sie_buf, pre_count, trigger_mode,
        scale_factor, input_channel_name)
    -- nbr of data values in a burst
    self.max_count = max_count
    -- SIE buffer
    self.sie_buf = sie_buf
    -- nbr of data values in a burst *before* the triggering event
    self.pre_count = pre_count
    -- trigger mode - note the mode values are unique to burst
    -- history data mode
    self.trigger_mode = trigger_mode
    -- inverse of sample rate
    self.scale_factor = scale_factor or 1
    -- used for error reports
    self.input_channel_name = input_channel_name
    -- circular buffer used to store pre-trigger data before it is used;
    --     use Lua values for count and index since there is no need to
    --     include them in ctype (we aren't using bit-wise operations)
    self.rbuf_front = 0
    self.rbuf_back = 0
    self.rbuf_element_size = ffi.sizeof(self.data_type)
    local buf_t_str = string.format(circular_buf_t_str, self.data_type)
    if not circ_cdefined[buf_t_str] then
        ffi.cdef(string.format(circular_buf_def_str, self.data_type,
            self.data_type, buf_t_str))
        circ_cdefined[buf_t_str] = 1
    end
    self.rbuf = ffi.new(buf_t_str)
    -- pre_count number of incoming data values are stored in rbuf
    -- (allocate one extra element to allow not using a "count" field)
    self.rbuf.buf = C.calloc(self.pre_count + 1,
        ffi.sizeof(self.data_type))
    -- a count of "time ticks" from start of test
    self.total_count = 0
    -- flag indicating a burst has been triggered
    self.triggered = false
    -- sample count at start of current burst (irrelevant if
    -- there is no current burst)
    self.burst_start = 0
    -- previous trigger input value, for use when trigger mode
    -- is 'on_false_to_true_edge'
    self.prev_trig_input = 0
    -- flag indicating a burst is currently being output
    self.writing_burst = false
    -- used to keep track of current sample's offset from the
    -- triggering event's timestamp
    self.sample_offset = 0
    -- number of samples already output in current burst
    self.nbr_output = 0
    -- ctype buffer used to build a sie_buf timestamp entry
    self.calc_stamp_buf = ffi.new('bhist_timestamp_t')
    -- ctype buffer used to build a sie_buf entry
    self.entry_type = string.format("bhist_%s_entry_t", self.data_type)
    ffi.cdef(string.format("typedef struct {%s value[1];} %s",
        self.data_type, self.entry_type))
    self.sample_buf = ffi.new(self.entry_type)
    self.entry_type_size = ffi.sizeof(self.entry_type)
end

-- calculates and sets triggered state, and returns the triggered state
function mt:calc_trigger(value)
    if self.trigger_mode == 'when_true' then
        self.triggered = tonumber(value) ~= 0
    elseif self.trigger_mode == 'on_false_to_true_edge' then
        if self.prev_trig_input == 0 and tonumber(value) ~= 0 then
            self.triggered = true
        else
            self.triggered = false
        end
    end

    return self.triggered
end

-- puts value's entry in sie_buf, while ensuring SIE "burst" blocks
-- start with a timestamp
function mt:enq_to_sie(value)
    self.sie_buf:flush_unless_available(self.entry_type_size)
    if self.sie_buf:pos() == 0 then
        -- Calculate timestamp from scaled input count, then
        -- write it and trigger offset at start of each block.
        self.calc_stamp_buf.stamp = self.burst_start + self.nbr_output
        self.calc_stamp_buf.offset = self.sample_offset + self.nbr_output
        self.sie_buf:enq(self.calc_stamp_buf,
            ffi.sizeof'bhist_timestamp_t')
    end
    ffi.copy(self.sample_buf.value, value,
        self.rbuf_element_size)
    if self.sie_buf:enq(self.sample_buf, self.entry_type_size) then
        local msg = string.format("SIE buf enq error: " ..
            "%s at time %s\n",
            tostring(self.input_channel_name),
            tostring(self.sample_offset))
        log:log(log.WARNING, msg)
    end
    self.nbr_output = self.nbr_output + 1
end

local ts = require 'timestamps'
-- executes the Burst History data mode at test runtime
function mt:run(in_buf, trig_buf)
    local doable = math.min(in_buf.count, trig_buf.count)
    if doable <= 0 then
        return false
    end

    local in_value
    local trig_value
    local newly_triggered = false

    for i = 0, doable - 1 do
        in_value = in_buf:deq().value
        trig_value = trig_buf:deq().value
        -- when triggered, stays triggered until burst data is
        -- written to file
        assert(self.rbuf_front <= self.pre_count, "invalid buf front index")
        assert(self.rbuf_back <= self.pre_count, "invalid buf back index")
        local buf_count = 0
        if self.rbuf_back > self.rbuf_front then
            buf_count = self.rbuf_back - self.rbuf_front
        elseif self.rbuf_back < self.rbuf_front then
            buf_count = self.rbuf_back + self.pre_count -
                self.rbuf_front + 1
        end
        if not self.writing_burst then
            newly_triggered = self:calc_trigger(trig_value[0])
            if newly_triggered then
                self.writing_burst = true
                self.nbr_output = 0
                self.sample_offset =
                    -math.min(buf_count, self.pre_count)
                self.burst_start = self.total_count + self.sample_offset
            end
        end
        self.prev_trig_input = tonumber(trig_value[0])

        if self.writing_burst then
            if newly_triggered and self.pre_count ~= 0 then
                -- transfer pre-trigger burst data to sie buffer
                -- NOTE:  This code is set up to avoid the need for
                --   updating the front pointer for each buffer entry.
                --   This is possible since we are emptying the
                --    buffer here.
                local top_buf_count =
                    math.min(self.pre_count - self.rbuf_front + 1,
                             buf_count)
                local bottom_buf_count = buf_count > top_buf_count
                    and buf_count - top_buf_count or false
                for i = self.rbuf_front,
                        self.rbuf_front + top_buf_count - 1 do
                    self:enq_to_sie(self.rbuf.buf[i])
                end
                if bottom_buf_count then
                    for i = 0, bottom_buf_count - 1 do
                        self:enq_to_sie(self.rbuf.buf[i])
                    end
                end
                -- circular buffer is now empty, so reset "pointers"
                self.rbuf_front = 0
                self.rbuf_back = 0
            end
            newly_triggered = false

            self:enq_to_sie(in_value)

            if self.nbr_output >= self.max_count then
                self.writing_burst = false
                self.sie_buf:flush()
            end
        else
            -- not writing out a burst, so keep up to 'pre_count'
            -- values in the internal circular queue
            ffi.copy(self.rbuf.buf + self.rbuf_back, in_value,
                self.rbuf_element_size)
            self.rbuf_back = self.rbuf_back >= self.pre_count and
                0 or (self.rbuf_back + 1)
            if buf_count >= self.pre_count then
                -- circular buffer was full, so drop oldest entry
                self.rbuf_front = self.rbuf_front >= self.pre_count and
                    0 or (self.rbuf_front + 1)
            end
        end

        -- total_count is # of input samples from test start
        self.total_count = self.total_count + 1
    end

    return true
end

-- called by the runtime test's "finalize" method; ensures
-- all this Burst History data mode channel's SIE data is
-- written to file
function mt:finalize()
    if self.writing_burst then
        -- write sie buffer to file
        self.sie_buf:flush()
    end
    if self.rbuf.buf ~= 0 then
        C.free(self.rbuf.buf)
    end
    self.rbuf = nil
end

local m = {}

-- returns a new Burst History data mode object
function m:new(data_type, cached_ids)
    local th_obj = {data_type = data_type, cached_ids = cached_ids}
    local new_obj = setmetatable(th_obj, {__index = mt})
    new_obj.data_type = data_type
    return new_obj
end

return m
