local cjson = require 'cjson'
local ffi = require 'ffi'

local aux_data = {}
local ctor_cache = {}

local histogram_bin_channel = [==[
    |<ch id="(=ch_id=)" private="bin">
    | <tag id="core:description">Generic histogram bin min/max values
    |+   lookup table</tag>
    | <dim index="0"><data decoder="(=bin_decoder_id=)" v="0"/></dim>
    | <dim index="1"><data decoder="(=bin_decoder_id=)" v="1"/></dim>
    | <dim index="2"><data decoder="(=bin_decoder_id=)" v="2"/></dim>
    |</ch>
    ]==]

local histogram_decoder_1d = [==[
    |<decoder id="(=decoder_id=)">
    | <read var="dimcount" bits="32" type="uint" endian="big"/>
    | <read var="bincount1" bits="32" type="uint" endian="big"/>
    | <loop var="v1" start="0" end="{$bincount1}">
    |  <read var="v0" bits="(=bin_bits=)" type="uint" endian="big"/>
    |  <sample/>
    | </loop>
    |</decoder>
    ]==]

local mt = {}

function mt:dump()
    return string.format('    bin_count=%s, lo_bound=%s, hi_bound=%s, ' ..
        'bound_scale=%s, ' ..
        'bound_span=%s\n',
        self.bin_count, self.lo_bound, self.hi_bound, self.bound_scale,
        self.bound_span)
end

function mt:initialize()
end

-- function mt:run_evenly_divided_sequential(in_buf, count)
-- function mt:run_evenly_divided_message(in_buf, count)

function mt:run_evenly_divided_generic(in_buf, count --[[, sbuf???]])
    for i = 0, count - 1 do
        local sample = in_buf:deq()
        if not sample:is_heartbeat() then
            local t = sample.value[0]
            local bin = math.floor((t - self.lo_bound) * self.bound_scale) + 1
            bin = math.max(0, bin)
            bin = math.min(self.bin_count - 1, bin)
            self.bins[bin] = self.bins[bin] + 1
            if self.bins[bin] == 0 then
                self.bins[bin] = bit.bnot(self.bins[bin])
            end
        end
    end
    return false -- No output from a sink
end

function mt:finalize() -- sbuf arg???
end

function mt:aux_id()
    local aux_id = tonumber(ffi.cast('int', ffi.cast('void *', self)))
    return aux_id
end

function mt:aux()
    return aux_data[self:aux_id()]
end

function mt:set_aux(aux)
    aux_data[self:aux_id()] = aux
end

function mt:histogram_1d_channel(itv)
    local aux = self:aux()
    local sw = aux.compiler.sw
    local tv = {}
    tv.ch_id = itv.ch_id or sw:next_id('ch')
    tv.grp = itv.grp or sw:next_id('group') -- KLUDGE WRONG probably...
    sw:xml('<ch test="0" id="(=ch_id=)" base="(=base=)" group="(=grp=)" name="(=name=)"/>\n',
        itv, tv)
    return tv.ch_id
end

function mt:get_histogram_bin_base_channel(bounds_bits)
    local aux = self:aux()
    local sw = aux.compiler.sw

    if sw.hist_bounds_base_channel then
        return sw.hist_bounds_base_channel
    end

    bounds_bits = bounds_bits or 64
    local bounds_type = '_histogram_bin_bounds' .. bounds_bits

    local tv = {}
    tv.bin_decoder_id = sw:decoder_id(bounds_type)
    tv.ch_id = sw:next_id('ch'),
    sw:xml()
    sw:xml(histogram_bin_channel, tv)
    sw:xml()
    sw.hist_bounds_base_channel = tv.ch_id
    return tv.ch_id
end

function mt:histogram_bounds_channel(base_ch_id, group, name,
        bounds, bounds_size)
    local aux = self:aux()
    local sw = aux.compiler.sw

    local tv = {}
    tv.i = ''
    tv.base = base_ch_id
    tv.grp = group
    tv.ch_id = sw:next_id('ch')
    tv.name = name

    sw:xml('(=i=)<ch id="(=ch_id=)" base="(=base=)" group="(=grp=)"'
        .. ' name="(=name=)#bin_1" private="bin"/>\n', tv)
    return tv.ch_id
end

function mt:histogram_1d_base_channel(itv)
    local aux = self:aux()
    local sw = aux.compiler.sw

    local tv = {}
    tv.base_id = sw:next_id('ch') -- KLUDGE CACHE
    tv.decoder_id = sw:next_id('decoder') -- KLUDGE CACHE

    sw:xml(histogram_decoder_1d, tv, itv)
    sw:xml()

    sw:xml([==[
    | <ch id="(=base_id=)" name="(=input_channel_name=)@(=datamode_name=)"
    |+     private="base">
    |  <tag id="somat:datamode_name">(=datamode_name=)</tag>
    |  <tag id="somat:datamode_type">time_at_level</tag>
    |  <tag id="somat:data_bits">(=bin_bits=)</tag>
    |  <tag id="somat:data_format">uint</tag>
    |  <tag id="somat:input_channel">(=input_channel_name=)</tag>
    |  <tag id="core:description">(=input_channel_description=)</tag>
    |  <tag id="core:sample_rate">(=sample_rate=)</tag>
    |  <tag id="core:schema">somat:histogram</tag>
    |  <dim index="0">
    |   <tag id="core:description"></tag>
    |   <tag id="core:label">Counts</tag>
    |   <tag id="core:units"></tag>
    |   <data decoder="(=decoder_id=)" v="0"/>
    |  </dim>
    |  <dim index="1">
    |   <tag id="core:description">Axis 0 Bin Min</tag>
    |   <tag id="core:label"></tag>
    |   <tag id="somat:range_min">(=range_min=)</tag>
    |   <tag id="somat:range_max">(=range_max=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |   <xform index_ch="(=index_ch=)" index_dim="1"/>
    |  </dim>
    |  <dim index="2">
    |   <tag id="core:description">Axis 0 Bin Max</tag>
    |   <tag id="core:label"></tag>
    |   <tag id="somat:range_min">(=range_min=)</tag>
    |   <tag id="somat:range_max">(=range_max=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |   <xform index_ch="(=index_ch=)" index_dim="2"/>
    |  </dim>
    | </ch>
    ]==], tv, itv)
    sw:xml()

    return tv.base_id
end

local function meta_ctor(t)
    if not ctor_cache[t] then
        local hist_type = string.format('struct %s_hist', t)
        local spec = string.gsub([[
            struct $t_hist_sie {
                $t dimcount; // KLUDGE placeholder
                $t bin_count; // KLUDGE placeholder
            };
            struct $t_hist {
                int bin_count;
                $t *bins;
                double *bounds;
                /* For evenly_divided special case */
                double lo_bound;
                double hi_bound;
                double bound_scale;
                double bound_span;
            };
        ]], '%$(%w+)', {t = t})
        ffi.cdef(spec)
        ctor_cache[t] = ffi.metatype(hist_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(bin_type)
    local ctor = meta_ctor(bin_type)
    local obj = ctor()
    return obj
end

return m
