local cjson = require 'cjson'
local ffi = require 'ffi'
local rational = require 'rational'

local dbg = require 'dbg'

local m = {}

-- setup-compiler module support

local cm_mt = {__index = m}

-- This table (as output code written to the test file) is
-- used to cache identifiers related to a data type, to avoid
-- some unnecessary duplication in the sie XML.
local cached_ids_decl = 'local cached_time_hist_ids = {}'
-- This flag is used to prevent multiple occurrences of the
-- 'cached_ids_decl' code.
local mod_declared = false

-- returns a new module object used to create Time History
-- data mode objects
function m:cm_new(...)
    local obj = {}
    for _, nt in ipairs{...} do
        for k, v in pairs(nt) do
            if obj[k] == nil then
                obj[k] = v
            end
        end
    end
    return setmetatable(obj, cm_mt)
end

-- creates test runtime code that instantiates a Time History
-- data mode object
function m:cm_declare()
    assert(#self.node.input_edges == 1) -- KLUDGE

    local tv = {}
    tv.data_type = self.node.data_type

    -- ensure the 'cached_time_hist_ids' table is declared in test code
    if not mod_declared then
        mod_declared = true
        self.compiler:oc_buffered('decl', {no_lhs = true},
            cached_ids_decl)
    end

    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)_runtime'):new('(=data_type=)',
    |+ cached_time_hist_ids)
    ]==], tv, self.node.meta)

    return name, value
end

-- creates test runtime code that initializes a Time History
-- data mode object
function m:cm_initialize(mod_name)
    local input_node = self.compiler:get_node(self.node.meta.input_channels[1])
    local code = {}
    local tv = {}

    tv.input_channel_name = self.node.meta.input_channels[1]
    tv.input_channel_name = input_node.meta.name
    tv.channel_description =
        self.node.sie_meta['core:description'] or ""
    tv.sie_buf_size = tostring(self.compiler.default_sie_buffer_size)
    tv.sample_rate = rational:new(self.node.meta.sample_rate):tonumbers()

    tv.data_bits = self.node.sie_meta['somat:data_bits']
    tv.data_format = self.node.sie_meta['somat:data_format']
    tv.schema = self.node.sie_meta['core:schema']
    tv.dim0_label = self.node.sie_meta._dims[1]['core:label'] or 'Time'
    tv.dim0_units = self.node.sie_meta._dims[1]['core:units'] or 'Seconds'
    tv.dim1_label = self.node.sie_meta._dims[2]['core:label'] or ''
    tv.dim1_units = self.node.sie_meta._dims[2]['core:units'] or
        self.node.output_units or ''
    tv.dim1_units = tostring(tv.dim1_units)

    self.node.sie_buf, self.node.sie_buf_idx = self.compiler:oc_sie_buf()

    code[#code + 1] = self.compiler:template({output = false}, [==[
    |
    |do
    |    local mod = (=mod_name=)
    |    local name = "(=name=)" -- KLUDGE
    |    local input_channel_name = "(=input_channel_name=)" -- KLUDGE
    |    local channel_description = "(=channel_description=)"
    |+       -- KLUDGE
    |
    |    local hist_base =
    |        mod:base_channel(self.sw,
    |        {datamode_name = name,
    |         data_bits = (=data_bits=),
    |         data_format = "(=data_format=)",
    |         schema = "(=schema=)",
    |         input_channel_name = input_channel_name,
    |         channel_description = channel_description,
    |         data_type = "(=data_type=)",
    |         dim0_label = "(=dim0_label=)",
    |         dim0_units = "(=dim0_units=)",
    |         dim1_label = "(=dim1_label=)",
    |         dim1_units = "(=dim1_units=)"
    |        })
    |
    |    local tv = {}
    |    tv.name = input_channel_name .. "@" .. name
    |    tv.input_channel_name = input_channel_name
    |    tv.channel_description = channel_description
    |    tv.sample_rate = (=sample_rate=)
    |    tv.base = hist_base
    |    tv.ch_id = self.sw:next_id('ch')
    |    self.sw:new_buffer((=sie_buf_idx=), nil, (=sie_buf_size=),
    |        ffi.sizeof'(=data_type=)', ffi.sizeof'timestamp')
    |+   --  => (=_sie_name=)
    |    tv.grp = (=sie_buf=).group
    |    tv.time_scale = 1 / (=sample_rate=)
    |    mod:mode_channel(self.sw, tv)
    |    mod:initialize{sie_buf = (=sie_buf=)}
    |end
    ]==], tv, self.node.sie_meta, self.node, self.node.meta)

    return table.concat(code)
end

-- creates test runtime code that calls (executes) a Time History
-- data mode object
function m:cm_run(name)
    local code = self.compiler:tmpl([==[
    |        (=cm_name=):run((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count,
    |+       ffi.sizeof'(=data_type=)')
    ]==], {cm_name = name}, self.node)
    return code
end

-- creates test runtime code to finalize a Time History data mode object
function m:cm_finalize(name)
    local code = self.compiler:template({output = false}, [==[
    |do
    |    (=cm_name=):finalize()
    |end
    ]==], {cm_name = name}, self.node)
    return code
end

return m
