local cjson = require 'cjson'
local ffi = require 'ffi'

local dbg = require 'dbg'

-- runtime time_history module support

local ctor_cache = {}

-- used to create a base channel description in the SIE file
-- which can be extended by multiple Time History data mode channels
-- (NOTE that the time dimension's scale is not defined here)
local th_base_channel = [==[
    | <ch id="(=ch_id=)" name="(=data_type=)@time_history_base"
    |+ private="base">
    |  <tag id="somat:datamode_name">(=datamode_name=)</tag>
    |  <tag id="somat:datamode_type">time_history</tag>
    |  <tag id="data_type">(=data_type=)</tag>
    |  <tag id="somat:data_bits">(=data_bits=)</tag>
    |  <tag id="somat:data_format">(=data_format=)</tag>
    |  <tag id="core:description">Base Channel for (=data_type=)</tag>
    |  <tag id="core:schema">(=schema=)</tag>
    |  <dim index="0">
    |   <tag id="core:description">Time History Time</tag>
    |   <tag id="core:label">(=dim0_label=)</tag>
    |   <tag id="core:units">(=dim0_units=)</tag>
    |   <data decoder="(=decoder_id=)" v="0"/>
    |  </dim>
    |  <dim index="1">
    |   <tag id="core:description">Time History Data</tag>
    |   <tag id="core:label">(=dim1_label=)</tag>
    |   <tag id="core:units">(=dim1_units=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |  </dim>
    | </ch>
    ]==]

-- used to create a Time History data mode channel in SIE file
-- KLUDGE - test id needs fixing?
local time_history_channel = [==[
    | <ch test="0" id="(=ch_id=)" base="(=base=)"
    |+ group="(=grp=)" name="(=name=)">
    |  <tag id="somat:input_channel">(=input_channel_name=)</tag>
    |  <tag id="core:sample_rate">(=sample_rate=)</tag>
    |  <tag id="core:description">(=channel_description=)</tag>
    |  <dim index="0">
    |   <xform scale="(=time_scale=)" offset="0"/>
    |  </dim>
    | </ch>
    ]==]

local mt = {}

-- create a base channel description in the SIE file, which is
-- extended by multiple Time History data mode channels
function mt:base_channel(sw, itv)
    local tv = {}
    tv.decoder_id = sw:decoder_id(self.data_type)
    tv.data_type = self.data_type

    if not self.cached_ids.base_channel_id or
       not self.cached_ids.base_channel_id[self.data_type] then
        if not self.cached_ids.base_channel_id then
            self.cached_ids.base_channel_id = {}
        end
        if not self.cached_ids.base_channel_id[self.data_type] then
            self.cached_ids.base_channel_id[self.data_type] =
                sw:next_id('ch')
        end
        tv.ch_id = self.cached_ids.base_channel_id[self.data_type]
        sw:xml()
        sw:xml(th_base_channel, tv, itv)
        sw:xml()
    else
        tv.ch_id = self.cached_ids.base_channel_id[self.data_type]
    end

    return tv.ch_id
end

-- create a Time History data mode channel description in the SIE file
function mt:mode_channel(sw, itv)
    sw:xml()
    sw:xml(time_history_channel, itv)
    sw:xml()
    return itv.ch_id
end

function mt:dump()
    return string.format("data type = %s, sie buf = %s\n",
        self.data_type, tostring(self.sie_buf))
end

-- assigns the attributes for this Time History data mode object
-- which need to be assigned at test runtime
function mt:initialize(...)
    for _, nt in ipairs{...} do
        for k, v in pairs(nt) do
            if self[k] == nil then
                self[k] = v
            end
        end
    end
end

-- executes the Time History data mode at test runtime
function mt:run(in_buf, count, data_el_size)
    if (count <= 0) then
        return false
    end

    local sbuf = self.sie_buf
    sbuf.timestamp[1] = sbuf.timestamp[0]
    if sbuf:pos() == 0 then
        sbuf:enq(sbuf.timestamp, ffi.sizeof'timestamp')
    end
    for i = 0, count - 1 do
        local sample = in_buf:deq()
        if sbuf:enq(sample.value, data_el_size) then
            -- :enq returns true when enqueue failed!
            sbuf:flush()
            sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + i
            sbuf:enq(sbuf.timestamp, ffi.sizeof'timestamp')
            sbuf:enq(sample.value, data_el_size)
        end
    end
    sbuf.timestamp[0] = tonumber(sbuf.timestamp[1]) + count
    return false
end

-- called by the runtime test's "finalize" method; ensures
-- all this Time History data mode channel's SIE data is
-- written to file
function mt:finalize()
    self.sie_buf:flush()
end

local m = {}

-- returns a new Time History data mode object
function m:new(data_type, cached_ids)
    local th_obj = {data_type = data_type, cached_ids = cached_ids}
    local new_obj = setmetatable(th_obj, {__index = mt})
    return new_obj
end

return m
