local m = {}

m.preamble = [==[
local jit = require 'jit'
jit.opt.start('maxtrace=4000', 'maxmcode=8192', 'sizemcode=64', 'hotloop=56')

require('lj_include'){
    'stdlib',
    'string',
    'sys/types',
    'unistd',
    'sys/stat',
    'sys/socket',
    'fcntl',
    'signal',
}

require 'strict'

local __ = require '__'
local bit = require 'bit'
local cjson = require 'cjson.safe'
local dbg = require 'dbg'
local dbus = require 'dbus'
local ev = require 'ev'
local ffi = require 'ffi'
local edge_buf = require 'edge_buf'
local http_server = require 'http_server'
local log = require 'log'
local message_can = require 'message_can'
local net = require 'net'
local network_db = require 'network_db'
local proxy_buffer = require 'proxy_buffer'
local quorum_latch = require 'quorum_latch'
local mx_sta = require 'mx_sta'
local sample_clock = require 'sample_clock'
local sample_types = require 'sample_types'
local sie_writer = require 'sie_writer'
local time_bracket = require 'time_bracket'
local ts = require 'timestamps'
local uuid = require 'uuid'
local hbm_stream_buf = require 'hbm_stream_buf'
local hbm_stream_server = require 'hbm_stream_server'

local bit_band = bit.band
local bit_bor = bit.bor
local bit_lshift = bit.lshift
local bit_rshift = bit.rshift
local bit_tobit = bit.tobit
local ffi_sizeof = ffi.sizeof
local math_min = math.min

ffi.cdef[[
    struct sockaddr_un {
        sa_family_t sun_family;
        char        sun_path[108];
    };
]]

local test_engine

local source_filename = debug.getinfo(1).short_src

log:open('test-engine', log.USER)

local function printf(fmt, ...)
    io.stdout:write(string.format(fmt, ...))
end

local function errorf(fmt, ...)
    error(string.format(fmt, ...))
end

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

local htonl = ffi.abi("le") and bit.bswap or bit.tobit

local loop = ev.Loop.default

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

local network = assert(bus:auto_proxy_wait('com.somat.armada.Network',
                                           '/com/somat/armada/Network'))
local net_db = network_db:new()

local hostname = net_db:get_module_hostname()
local ip = net_db:get_module_ip()

local config = require 'system_config'

-- NOTE:  The start time delay can NEVER be less the the amount of time it
-- takes to setup a DIO channel because DIO channels will NOT return data if
-- the start time they are given has already passed!
local start_time_delay = config.start_time_delay or 6

-- Note, this should work regardless of whether or not system time is being
-- synchronized to GPS time because this start time is based off of current
-- PTP time.
local start_time = mx_sta.get_adjusted_start_time(
    ts.epoch_to_iso(ts.get_ts().s + start_time_delay),
    cjson.decode(normalized_setup), log)

local args = {...}

local duration = tonumber(args[2])
local end_time
if duration then
    end_time = ts.iso_to_epoch(start_time) + duration
end

local start_timestamp = ts.iso_to_stamp(start_time)

local decl = {}
local proxy_buffers = {}
local proxy_readers = {}
local bufs = {}
local comet_message = {}
local hbm_bufs = {}
local sie_bufs

local m = {}

m.nodes = {}

local C = ffi.C

local function change_group_id(group)
    local f = io.open('/etc/group')
    for line in f:lines() do
        local g, id = line:match("(.-):.-:(.-):")
        if g == group then
            C.setegid(tonumber(id))
            break
        end
    end
    f:close()
end

if config.enable_test_core_dumps then
    require('core')()
else
    change_group_id('acq')
end

local NaN = 0/0

local function hexdump(p, len)
    local hp = ffi.cast('uint8_t *', p)
    local t = {}
    for i = 0, len do
        t[#t + 1] = string.format('%.02X', hp[i])
    end
    return table.concat(t, ' ')
end

local function bit32_unsigned_fixup(n)
    return (n >= 0) and n or n + 0x100000000
end

local function bit_match(msg, value)
    if msg:byte_length() < value:byte_length() then
        return nil
    end
    for i = 0, value:byte_length() - 1 do
        if msg.value[i] ~= value.value[i] then
            return nil
        end
    end
    return msg
end

local function bit_match_masked(msg, value, mask)

    if msg:byte_length() < value:byte_length() then
        return nil
    end
    for i = 0, value:byte_length() - 1 do
        -- KLUDGE precompute masked value!
        if bit_band(msg.value[i], mask.value[i])
                ~= bit_band(value.value[i], mask.value[i]) then
            return nil
        end
    end
    return msg
end

local function check_stop_bracket(engine, bracket, vote_id)
    if bracket.stopped == 1 then
        -- Make the above test a one-shot
        bracket.stopped = 2
        if engine.stop_latch:set_vote(vote_id) then
            engine:stop_test() -- KLUDGE
        end
    end
end

function m.normal_stop(loop, sig, revents)
    test_engine:schedule_test_stop()
end

function m.restartable_normal_stop(loop, sig, revents)
    test_engine.restart_test = true
    test_engine:schedule_test_stop()
end

function m.emergency_stop(loop, sig, revents)
    test_engine.restart_test = true
    test_engine:stop_test()
end

function m.comet_cleanup(w)
    local self = test_engine
    for i = 1, #self.comet_streams do
        if self.comet_streams[i] == w then
            table.remove(self.comet_streams, i)
            break
        end
    end
    if #self.comet_streams == 0 then
        self.comet_timer:stop(loop)
    end
end

function m:config_nginx_stream(hash)
--    self.stream_server = hbm_stream_server:new(loop)
    self.stream_server = hbm_stream_server:new(loop, {port = 8080, cport = 8081})
    local hash = hash:match('^(.+)%._armada%-tests')
    -- Create the upstream (websocket) server configuration
    local ch = io.open('/tmp/nginx/conf.d/' .. hash .. '_upstream.conf', 'w')
    ch:write(string.format(
[[    upstream %s_stream_server {
        server  127.0.0.1:%s;
    }
]], hash, self.stream_server.ws_port))
    ch:close()

    -- Create the HTTP connection upgrade configuration
    ch = io.open('/tmp/nginx/conf.d/' .. hash .. '_stream.conf', 'w')
    ch:write(string.format(
[[ location /%s/stream_data {
            proxy_pass http://%s_stream_server;
            proxy_http_version 1.1;
            proxy_set_header Upgrade $http_upgrade;
            proxy_set_header Connection "upgrade";
            proxy_set_header Host $host;

            proxy_set_header X-Real-IP $remote_addr;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Forwarded-Proto http;

            proxy_redirect off;
        }
 location /%s/stream_control {
            proxy_pass http://127.0.0.1:%s/stream_control;
        }
]], hash, hash, hash, self.stream_server.http_port))
    ch:close()

    -- Tell NGINX to re-read it's configuration files.
    ch = io.popen('ps -o pid,args | grep nginx | grep master | grep -v grep')
    if ch then
        for line in ch:lines() do
            local pid = line:match('(%d+) nginx: master')
            if pid then
                C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGHUP)
                break
            end
        end
        ch:close()
    end
end

function m:system_init(start_stream_server)
    local sigs = {
        SIGPIPE = false,
        SIGINT = false,
        SIGTERM = false,
        SIGPWR = true,
--        SIGQUIT = true,
    }

    test_engine = self
    self.rr_changes_bound = false
    self.monitored_rr_owners = {}  -- identifiers of rrs of interest
    self.monitored_rr = {}  -- current state of rrs of interest

    for signame, emergency in pairs(sigs) do
        local fn = emergency and self.emergency_stop or self.normal_stop
        local sig = ev.Signal.new(fn, C[signame])
        sig:start(loop)
    end
    ev.Signal.new(self.restartable_normal_stop, C.SIGQUIT):start(loop)

    self.test_srvname = self.test_id:gsub('[:.]', '-') .. '._armada-tests'
    local test_srvname = self.test_srvname

    local http_listener = http_server:new{}
    local http_w = http_listener:listen{address = ip}
    function http_listener.handler(loop, w, http)
        local resp = '404 Not Found'

        if http.uri:match('^/stop') then
--            self:stop_test(http.uri:match('stop_time=([^&]+)'))
            self:stop_test()
            resp = '200 OK'
        elseif http.uri:match("^/comet_stream$") then
            w.streambuf:write(
                "HTTP/1.0 200 OK\r\n"
                .. "Content-Type: application/octet-stream\r\n"
                .. "\r\n"
            )
            -- Ignore requested channels for now
            if #self.comet_streams == 0 then
                self.comet_timer:start(loop)
            end
            table.insert(self.comet_streams, w)
            w.cleanup = self.comet_cleanup
            return
        end
        local msg =
            'HTTP/1.0 '..resp..'\r\n'
         .. 'Content-Type: text/plain\r\n'
         .. '\r\n'
        C.write(w.socket, msg, #msg)
        w.server.http_write_shutdown(loop, w)
    end

    -- If this test is being started because the system was rebooted while the
    -- test was running, then start a timer that will make the test NOT exit
    -- when any proxy connection sends back an error message after about 10
    -- seconds.
    if self.interrupted_restart then
        ev.Timer.new(function(l, t, e)
                for _, pb in ipairs(proxy_buffers) do
                    pb.exit_on_proxy_error = 0
                end
            end, 10, 0):start(loop)     -- KLUDGE: Hardcoded 10 seconds!
    else
        -- If not in restart interrupted test mode, immediately switch the
        -- proxy buffers to log errors from the proxys and exit normally.
        for _, pb in ipairs(proxy_buffers) do
            pb.exit_on_proxy_error = 0
        end
    end

    local l = cjson.encode({
                id = self.test_id,
                name = self.test_name,  
                title = self.test_title,
                raw_sie_file = self.sie_name,
                -- Base friendly name without run number
                friendly_sie_name =
                    self.friendly_sie_name:match('(.+)%-%d%d%d') or
                    self.friendly_sie_name,
                setup = cjson.decode(pristine_setup),
                duration = duration})

    local add_rrs = 
        {{'_armada-tests', 'PTR', test_srvname},
        {test_srvname, 'SRV', '0 0 ' .. http_w.port .. ' ' .. hostname},
        {test_srvname, 'TXT', 'id: ' .. self.test_id},
        {test_srvname, 'TXT', 'name: ' .. self.test_name},
        {test_srvname, 'TXT', 'title: ' .. self.test_title},
        {test_srvname, 'TXT', 'start-time: ' .. start_time},
        {test_srvname, 'TXT', 'raw-data-file: ' .. self.sie_name},
        {test_srvname, 'TXT', 'data-file: ' .. self.friendly_sie_name},
        {test_srvname, 'TXT', 'setup-file: ' .. pristine_setup}}

    -- Only add this one if resets are not overridden
    if not config.override_sys_error_resets then
        add_rrs[#add_rrs + 1] = {'_restart-armada-tests', 'TXT', l}

        -- Function called on normal test stop
        self.db_cleanup = function()
                if not self.restart_test then
                    self:modify_rrs({{'_restart-armada-tests', 'TXT', l}}, {})
                end
            end
    end
    self:modify_rrs({}, add_rrs)

    if start_stream_server then
        -- Configure NGINX so it can proxy the websocket stream for this test
        self:config_nginx_stream(test_srvname)
    end

    local f = io.open(self.sie_name:gsub('%.sie$', '%.txt'), 'w')
    if f then
        f:write(self.friendly_sie_name)
        f:close()
    end
end

function m:modify_rrs(del, ins)
    return network:ModifyRRs(del, ins)
end

function m:bind_changed_rrs()
    if not self.rr_changes_bound then
        network.ChangedRRs:bind(
            function(removed, added)
                self:changed_rr(removed, added)
            end
        )
        self.rr_changes_bound = true
    end
end

-- NOTE:  BEWARE of multiple rrs with one owner!
--        This code does not currently handle that case,
--        other than to save (only) the rr with the highest
--        index.
function m:changed_rr(removed, added)
    for i,rr in ipairs(removed) do
        if self.monitored_rr_owners[rr[1]] then
            self.monitored_rr[rr[1]] = nil
        end
    end
    for i,rr in ipairs(added) do
        if self.monitored_rr_owners[rr[1]] then
            self.monitored_rr[rr[1]] = rr
            log:notice(__'"{0}" set to "{1}"',
                rr[1], rr[3])
        end
    end
end

-- NOTE:  BEWARE of multiple rrs with one owner!
--        This code does not currently handle that case,
--        other than to update (only) the rr with the highest
--        index.
function m:update_monitored_rr_values(owner)
    local recs = network:QueryRR(owner, 'TXT', '*', '*')
    for i = 1, #recs do
        self.monitored_rr[owner] = recs[i]
    end
end

function m:monitor_rr(owner)
    self.monitored_rr_owners[owner] = 1  -- any "true" value will do
end

function m:get_rr(owner)
    return self.monitored_rr[owner]
end

function m:run(done)
    repeat until self:process(done) ~= true
end

function m:close_sie_file()
    if self.sw then
        self.sw:close()
        self.sw = nil
    end
end

function m.stop_failed_handler(loop, timer, revents)
    local self = test_engine
    log(log.ERR, __'stop_test "{0}" {1} at {2}, stop forced at {3}',
        self.test_name, self.test_id, self.stop_stamp_iso or __'UNSCHEDULED',
        ts.epoch_to_iso(ts.get_ts().s))
    loop:unloop()
end

function m:schedule_test_stop(requested_duration)
    local now = ts.get_ts().s
    local start_epoch = ts.stamp_to_epoch(start_timestamp)
    local ok_duration = math.ceil(now + 1 - start_epoch)
    local duration = math.max(tonumber(requested_duration) or 0, ok_duration)
    local stop_stamp = duration * sample_clock.master_rate

    self.stop_stamp_iso = ts.stamp_to_iso(start_timestamp + stop_stamp)

    log(log.INFO, __'schedule_test_stop "{0}" {1} at {2}',
        self.test_name, self.test_id, self.stop_stamp_iso)

    for i = 1, #proxy_buffers do
        local sb = proxy_buffers[i].bracket
        sb.stop_stamp = (sb.sample_rate == 0)
            and stop_stamp or duration * sb.sample_rate
    end

    -- Kill 5 seconds from now no matter what
    local kill_at = math.ceil(start_epoch + duration + 15 - now)
    self.stop_failed_timer = ev.Timer.new(self.stop_failed_handler, kill_at, 0)
    self.stop_failed_timer:start(loop)
end

function m:stop_test()
    log:notice(__'stop_test "{0}" {1} at {2}, now = {3}',
        self.test_name, self.test_id, self.stop_stamp_iso or __'UNSCHEDULED',
        ts.epoch_to_iso(ts.get_ts().s))
    if self.db_cleanup then self:db_cleanup() end
    loop:unloop()
end

function m:create_input_node(json_md, reader_fn, pbuf)
    local node = {}
    table.insert(self.nodes, node)

    node.json_md = json_md
    node.md = cjson.decode(json_md)

    assert(node.md.connector)
--dbg('create_input_node, connector=%s\n', tostring(node.md.connector))

    node.orig_host = node.md.connector:gsub('^.*%@', '')

--dbg('orig_host=%s\n', tostring(node.orig_host))

    local type_name = assert(node.md.data_type)

    node.host, node.port, node.ip
        = assert(net_db:resolve_service(node.orig_host))

--[[
    node.s = net.socket{ip = node.ip, port = node.port}:connect_nb(loop,
        function(s)
            if s then
                node.s = s
                local w = ev.IO.new(reader_fn, node.s.sock, ev.READ)
                pbuf.fd = node.s.sock
                node.reader = w
                w.node = node
                w:start(loop)

                node.md.start_time = start_time
                node.md._test_id = self.test_id
                local out_md = cjson.encode(node.md)

                local msg = 'POST /channel_stream HTTP/1.0\r\n'
                    .. 'Content-length: '..#out_md..'\r\n\r\n'
                    .. out_md
                C.write(node.s.sock, msg, #msg)
            end
        end)
]]

    -- Function called when connect completes with or without error
    local function handle_connect_complete(s)
        if s then
            node.s = s
            local w = ev.IO.new(reader_fn, node.s.sock, ev.READ)
            pbuf.fd = node.s.sock
            node.reader = w
            w.node = node
            w:start(loop)

            node.md.start_time = start_time
            node.md._test_id = self.test_id
            local out_md = cjson.encode(node.md)

            local msg = 'POST /channel_stream HTTP/1.0\r\n'
                .. 'Content-length: '..#out_md..'\r\n\r\n'
                .. out_md
            C.write(node.s.sock, msg, #msg)
        else
            log:log(log.ERR, string.format("Unable to connec to channel %s",
                node.md.name))
        end
    end

    -- Function called when UNIX domain connection completes with or without
    -- error
    local function handle_unix_connect_complete(s)
        if s then
            return handle_connect_complete(s)
        end
        node.s = net.socket{ip = node.ip, port = node.port}:connect_nb(loop,
            handle_connect_complete)
    end

    local sock = net.socket{
        family = C.AF_UNIX,
        type = C.SOCK_STREAM,
        flags = 0}

    -- Setup for connect via UNIX domain socket
    local sa = ffi.new('struct sockaddr_un[1]')
    sa[0].sun_family = C.AF_UNIX
    local name = node.md.connector:match('^.*%@(.+)$')
    ffi.copy(sa[0].sun_path, '/tmp/proxy/' .. name)

    sock.addrs = ffi.new('struct addrinfo')
    sock.addrs.ai_flags = 0
    sock.addrs.ai_family = C.AF_UNIX
    sock.addrs.ai_socktype = C.SOCK_STREAM
    sock.addrs.ai_protocol = 0
    sock.addrs.ai_addrlen = ffi.sizeof('struct sockaddr_un')
    sock.addrs.ai_addr = ffi.cast('struct sockaddr *', sa)
    sock.addrs.ai_next = nil

    node.s = sock:connect_nb(loop, handle_unix_connect_complete)

    comet_message[node.md.name] = node.md.comet_message
end

]==]

m.postamble = [==[


local ok, err = pcall(function()
    local starter
    starter = ev.Idle.new(function(loop, idle, revents)
        m:init(args)
        starter:stop(loop)
    end)
    starter:start(loop)
    loop:loop()

    m:run(true)
    m:finalize()
    m:close_sie_file()
    if m.stream_server then
        m.stream_server:terminate(log, function()
                os.exit(0)
            end)
        loop:loop()
    end
end)

if err then
    log(log.ERR, err)
    m:close_sie_file()
    if m.stream_server then
        m.stream_server:terminate(log, function()
                os.exit(1)
            end)
        loop:loop()
    end
    os.exit(1)
end

os.exit(0)
]==]

return m
