#!/usr/bin/luajit

package.path = "../share/lua/?.lua;" .. package.path

local ffi = require "ffi"
local libsie = require 'libsie_ffi'

local args = {...}

local function usage()
    io.stdout:write("Usage: sie_dump.lua <file> [<options>]\n\n")
    io.stdout:write("If no options are given, will dump to string the entire"..
                    " SIE file (May be very large!)\n\n")
    io.stdout:write("Options:\n")
    io.stdout:write("         --taglist         Return a list of all"..
                    " available tag names\n")
    io.stdout:write("         --tag  <tagname>  Return a particular tag\n")
    io.stdout:write("         --type <tagtype>  Return all tag names of a"..
                    " particular type\n")
    io.stdout:write("                             {File, Test, Channel,"..
                    " Dimension}\n\n")
    os.exit(0)
end

if #args == 0 then
    usage()
end

local opt = {}
do
    local i = 1
    while i <= #args do
        local a = args[i]
        if a == '--tag' then
            i = i + 1
            opt.target_tag = assert(args[i])
        elseif a == '--taglist' then
            opt.all_tags = true
        elseif a == '--type' then
            i = i + 1
            local types = {["file"] = 1,    ["test"] = 2,
                           ["channel"] = 3, ["dimension"] = 4}
            opt.type = types[string.lower(assert(args[i]))]
            if not opt.type then usage() end
        else
            opt.filename = args[i]
        end
        i = i + 1
    end
end

assert(opt.filename)

local function printf(fmt, ...)
    io.stdout:write(string.format(fmt, ...))
end

if opt.target_tag then
    printf = function() end
end

local tag_value = ffi.new("const char *[?]", 1)
local tag_size = ffi.new("size_t[?]", 1)
local function print_tag(tag, prefix)
    local name = libsie.sie_tag_get_id(tag)
    local ok = libsie.sie_tag_get_value_b(tag, tag_value, tag_size)
    if ok ~= 0 then
        local tag_name = ffi.string(name)
        if opt.target_tag == tag_name then
            io.stdout:write(ffi.string(tag_value[0], tag_size[0])..'\n')
            --os.exit(0)
        end
        local value_size = tag_size[0]
        if value_size > 50 then
            printf("%s'%s': long tag of %g bytes.\n",
                prefix, tag_name, tonumber(value_size))
            libsie.sie_free(ffi.cast("void *", tag_value[0]))
            return
        end
        printf("%s'%s': '%s'\n", prefix, tag_name,
            ffi.string(tag_value[0], tag_size[0]))
        libsie.sie_free(ffi.cast("void *", tag_value[0]))
    end
end


local context = libsie.sie_context_new()
local file = libsie.sie_file_open(context, opt.filename)
if file == nil then
    local exception = libsie.sie_get_exception(context)
    io.stderr:write(string.format("Something bad happened:\n  %s\n",
        ffi.string(libsie.sie_verbose_report(exception))))
    os.exit(1)
end

local printfold = printf

printf("File '%s':\n", opt.filename)

if opt.type and opt.type ~= 1 then
    printf = function() end
end

local tag_iterator = libsie.sie_get_tags(file)
while true do
    local tag = libsie.sie_iterator_next(tag_iterator)
    if tag == nil then
        break
    end
    print_tag(tag, "  File tag ")
end
libsie.sie_release(tag_iterator)
printf = printfold

local test_iterator = libsie.sie_get_tests(file)
while true do
    local test = libsie.sie_iterator_next(test_iterator)
    if test == nil then
        break
    end
    local id = libsie.sie_get_id(test)
    if not opt.type or (opt.type and opt.type > 1) then
        printf("  Test id %g\n", tonumber(id))
    end

    if opt.type and opt.type ~= 2 then
        printf = function() end
    end
    local tag_iterator = libsie.sie_get_tags(test)
    while true do
        local tag = libsie.sie_iterator_next(tag_iterator)
        if tag == nil then
            break
        end
        print_tag(tag, "    Test tag ")

    end
    libsie.sie_release(tag_iterator)
    printf = printfold

    local channel_iterator = libsie.sie_get_channels(test)
    while true do
        local channel = libsie.sie_iterator_next(channel_iterator)
        if channel == nil then
            break
        end
        if not opt.type or (opt.type and opt.type > 2) then
            printf("    Channel id %g, '%s':\n",
                tonumber(libsie.sie_get_id(channel)),
                ffi.string(libsie.sie_get_name(channel)))
        end
        
        if opt.type and opt.type ~= 3 then
            printf = function() end
        end

        local tag_iterator = libsie.sie_get_tags(channel)
        while true do
            local tag = libsie.sie_iterator_next(tag_iterator)
            if tag == nil then
                break
            end
            print_tag(tag, "      Channel tag ")
        end
        libsie.sie_release(tag_iterator)
        printf = printfold

        local dimension_iterator = libsie.sie_get_dimensions(channel)
        while true do
            local dimension = libsie.sie_iterator_next(dimension_iterator)
            if dimension == nil then
                break
            end
            
            if opt.type and opt.type ~= 4 then
                printf = function() end
            end
            printf("      Dimension index %g:\n",
                tonumber(libsie.sie_get_index(dimension)))

            local tag_iterator = libsie.sie_get_tags(dimension)
            while true do
                local tag = libsie.sie_iterator_next(tag_iterator)
                if tag == nil then
                    break
                end
                print_tag(tag, "        Dimension tag ")
            end
            libsie.sie_release(tag_iterator)
        end
        libsie.sie_release(dimension_iterator)
        printf = printfold

        local spigot = libsie.sie_attach_spigot(channel)
        while true and not opt.all_tags and not opt.type do
            local output = libsie.sie_spigot_get(spigot)
            if output == nil then
                break
            end
            local num_dims = tonumber(libsie.sie_output_get_num_dims(output))
            local num_rows = tonumber(libsie.sie_output_get_num_rows(output))
            printf("num_dims=%s, num_rows=%s\n", num_dims, num_rows)
            --[[
            printf("      Data block %g, %g dimensions, %g rows:\n",
                tonumber(libsie.sie_output_get_block(output)),
                tonumber(num_dims),
                tonumber(num_rows))
            --]]
            local outs = libsie.sie_output_get_struct(output)
            for row = 0, num_rows - 1 do
                --printf("        Row %g: ", row);
                for dim = 0, num_dims - 1 do
                    if dim ~= 0 then
                        printf(", ")
                    end
                    if outs.dim[dim].type == libsie.SIE_OUTPUT_FLOAT64 then
                        --printf("%.15g", tonumber(outs.dim[dim].float64[row]))
                        local num = tonumber(outs.dim[dim].float64[row])
                        if num % 1 ~= 0 then
                            printf("%.9f", num)
                        else
                            printf("%d", num)
                        end
                    elseif outs.dim[dim].type == libsie.SIE_OUTPUT_RAW then
                        local uchar_p = ffi.cast("const unsigned char *",
                            outs.dim[dim].raw[row].ptr)
                        local size = tonumber(outs.dim[dim].raw[row].size)
                        local printable = true
                        for i = 0, size - 1 do
                            local c = uchar_p[i]
                            if c < 0x20 or c > 0x7F then
                                if c ~= 0x09 and c ~= 0x0A and c ~= 0x0D then
                                    printable = false
                                    break
                                end
                            end
                        end
                        if printable then
                            printf("%s", ffi.string(uchar_p, size))
                        elseif size > 32 then
                            printf("(raw data of size %g.)", tonumber(size))
                        else
                            for byte = 0, size - 1 do
                                printf("%02x", tonumber(uchar_p[byte]))
                            end
                        end
                    end
                end
                printf("\n")
            end
        end
        libsie.sie_release(spigot)
    end
    libsie.sie_release(channel_iterator)
end
libsie.sie_release(test_iterator)

local channel_iterator = libsie.sie_get_channels(file)
while true do
    local channel = libsie.sie_iterator_next(channel_iterator)
    if channel == nil then
        break
    end
    --[[
    printf("  Channel id %g, '%s' ", tonumber(libsie.sie_get_id(channel)),
        ffi.string(libsie.sie_get_name(channel)))
    local test = libsie.sie_get_containing_test(channel)
    if test ~= nil then
        printf("is contained in test id %g.\n",
            tonumber(libsie.sie_get_id(test)))
    else
        printf("is not in a test.\n")
    end
    --]]
    libsie.sie_release(test)
end
libsie.sie_release(channel_iterator)

if libsie.sie_check_exception(context) ~= nil then
    local exception = libsie.sie_get_exception(context)
    io.stderr:write(string.format("Something bad happened:\n  %s\n",
        ffi.string(libsie.sie_verbose_report(exception))))
    libsie.sie_release(exception)
end

libsie.sie_release(file)

local leaked_objects = libsie.sie_context_done(context)
if leaked_objects ~= 0 then
    io.stderr:write(string.format("Warning:  Leaked %d SIE objects!\n",
        leaked_objects))
end

os.exit(0)
