require 'strict'

require('lj_include'){
    'sys/types',
    'sys/stat',
    'fcntl',
    'sys/socket',
    'sys/uio',
    'unistd',
}

local __ = require '__'
local bit = require 'bit'
local cjson = require 'cjson'
local log = require 'log'
local membuf = require 'membuf'
local ffi = require 'ffi'
local rational = require 'rational'
local time_bracket = require 'time_bracket'
local zlib = require 'zlib'

require 'middleclass'

local C = ffi.C


local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local math_floor = math.floor

local function hexify(s)
    return string.gsub(s, "(.)", function(c)
        return string.format(" %02X", string.byte(c))
    end)
end

local htonl = ffi.abi("le") and bit.bswap or function(x) return x end
local ntohl = htonl

local function html_escape(str)
    return tostring(str):gsub('[&"<]', {
        ['&'] = '&amp;',
        ['"'] = '&quot;',
        ['<'] = '&lt;'
    })
end

local function get_rational(x)
    local slash = (type(x) == 'string') and x:find('/')
    if slash then
        local num = tonumber(x:sub(1, slash - 1))
        local denom = tonumber(x:sub(slash + 1))
        return num / denom, num, denom
    end
    local num = tonumber(x)
    return num, num, 1
end

ffi.cdef[[
    typedef int64_t timestamp;

    typedef struct __attribute((packed)) {
        uint32_t size;
        uint32_t group;
        uint32_t magic;
    } sie_block_header;

    typedef struct __attribute((packed)) {
        uint32_t crc;
        uint32_t size;
    } sie_block_trailer;

    typedef struct __attribute((packed)) {
        uint64_t offset;
        uint32_t group;
    } sie_index_entry;

    typedef struct sie_buffer {
        int idx;
        int fd;
        int group;
        char *file_name;
        int element_size;
        int payload_overhead;
        timestamp timestamp[2]; /* 0 = actual, 1 = scratch */
        sie_index_entry index_entry[1]; /* Only for index buffer */
        struct sie_buffer *index_buf;
        struct iovec vec[3];
        sie_block_header header;
        struct membuf payload;
        sie_block_trailer trailer;
        struct time_bracket *bracket;
        double total_written;
    } sie_buffer;
]]

local sie_overhead = 20
local sie_magic = 0x51EDA7A0

local mt = {}

function mt:initialize(fd, group, max_payload, element_size, payload_overhead)
    self.fd = fd
    self.group = group
    self.element_size = element_size
    self.payload_overhead = payload_overhead
    self.header.group = htonl(group)
    self.header.magic = htonl(sie_magic)
    self.payload:initialize(max_payload)
    self.vec[0].iov_base = self.header
    self.vec[0].iov_len = ffi_sizeof(self.header)
    self.vec[2].iov_base = self.trailer
    self.vec[2].iov_len = ffi_sizeof(self.trailer)
    self.total_written = 0
end

function mt:pos()
    return #self.payload
end

function mt:block_len()
    return #self.payload + sie_overhead
end

function mt:enq(buf, size)
    if size > self.payload:available() then
        -- If there's only a small amount buffered already (i.e.
        -- payload headers), and the buffer is too small to fit
        -- what's being enqueued, then expand.
        if #self.payload < 256 then
            self.payload:expand(#self.payload + size)
        else
            return true -- Enqueue FAILED
        end
    end
    self.total_written = self.total_written + size
    self.payload:append(buf, size)
    -- Non-intuitive boolean sense here so that the caller
    -- can write "if sbuf:enq(...) then handle_enq_failure() end"
    return false -- OK, no problem with the enqueue
end

function mt:append(buf, size)
    if not size then
        size = #buf
    end
    if self:enq(buf, size) then
        self:flush()
        if self:enq(buf, size) then
            return false
        end
    end
    return #self.payload
end

function mt:finalize()
    local block_size = self:block_len()
    local block_size_be = htonl(block_size)
    self.header.size = block_size_be
    local crc = zlib.crc32(0, nil, 0)
    crc = zlib.crc32(crc, ffi_cast('char *', self.header),
        ffi_sizeof(self.header))
    crc = zlib.crc32(crc, self.payload:head(), #self.payload)
    self.trailer.crc = htonl(crc)
    self.trailer.size = block_size_be
end

function mt:flush(force)
    if #self.payload <= self.payload_overhead and not force then
        return 0
    end
    self:write_index(self.header.group, self:block_len())
    self:finalize()
    local len = self:block_len()
    -- local off = tonumber(C.lseek64(self.fd, 0, C.SEEK_CUR))
    -- KLUDGE won't work non-blocking, needs error handling
    self:write_block()
    -- assert(off + len == tonumber(C.lseek64(self.fd, 0, C.SEEK_CUR)))
    self.payload:clear()
    return len
end

function mt:write_index(group, len)
    if self.index_buf == nil then
        return -- Don't index the index blocks
    end
    self = self.index_buf
    assert(self.index_buf == nil)
    assert(ffi_sizeof(self.index_entry) == 12)
    self.index_entry[0].group = group
    if self.payload:available() < ffi_sizeof(self.index_entry) then
        -- flush() but without infinite loop danger
        self:finalize()
        self:write_block()
        self.index_entry[0].offset =
            self.index_entry[0].offset + self:block_len()
        self.payload:clear()

        -- We don't index index blocks on the grounds of gratuitous
        -- redundancy, but if we did, it would be like this:
        -- self.index_entry[0].group = self.header.group
        -- self:enq(self.index_entry, ffi_sizeof(self.index_entry))
        -- self.index_entry[0].group = group
    end
    assert(self:enq(self.index_entry, ffi_sizeof(self.index_entry)) == false)
    self.index_entry[0].offset = self.index_entry[0].offset + len
end

function mt:write_block()
    self.vec[1].iov_base = self.payload:head()
    self.vec[1].iov_len = #self.payload
    local wrote = C.writev(self.fd, self.vec, 3)
    local len = self:block_len()
    if wrote ~= len then
        local fname = ffi.string(self.file_name)
        if wrote == -1 and ffi.errno() ~= C.ENOSPC then
            log:error(__'SIE {0} write() error: {1}', fname, log:strerror())
        else
            if wrote > 0 then
                -- Short write, so undo it
                local pos = C.lseek64(self.fd, 0, C.SEEK_END)
                if pos == -1 then
                    log:error(__'SIE {0} lseek64() error: {1}',
                        fname, log:strerror())
                end
                if C.ftruncate64(self.fd, pos - wrote) ~= 0 then
                    log:error(__'SIE {0} ftruncate64() error: {1}',
                        fname, log:strerror())
                end
            end
            -- KLUDGE This is a temporary measure pending full crisis
            -- buffer support, test continuation on filesystem full, etc.
            log:error('Filesystem full writing SIE file {0}', fname)
        end
    end
    return wrote
end

function mt:bytes_left()
    return self.payload:available()
end

function mt:flush_unless_available(bytes)
    if self.payload:available() < bytes then
        self:flush()
        return true
    end
    return false
end

function mt:elements_available()
    return math_floor(self:bytes_left() / self.element_size)
end

local sie_buffer_ctor = ffi.metatype('sie_buffer', {__index = mt})

local m = class('sie_writer'):include(require 'template')

m.statics = {
    sie_header = [==[
    |<?xml version="1.0" encoding="UTF-8"?>
    |<sie version="1.0" xmlns="http://www.somat.com/SIE">
    |<!-- SIE format standard definitions: -->
    | <!-- SIE stream decoder: -->
    | <decoder id="0">
    |  <loop>
    |   <read var="size" bits="32" type="uint" endian="big"/>
    |   <read var="group" bits="32" type="uint" endian="big"/>
    |   <read var="syncword" bits="32" type="uint" endian="big"
    |+      value="0x51EDA7A0"/>
    |   <read var="payload" octets="{$size - 20}" type="raw"/>
    |   <read var="checksum" bits="32" type="uint" endian="big"/>
    |   <read var="size2" bits="32" type="uint" endian="big" value="{$size}"/>
    |  </loop>
    | </decoder>
    | <tag id="sie:xml_metadata" group="0" format="text/xml"/>
    |
    | <!-- SIE index block decoder:  v0=offset, v1=group -->
    | <decoder id="1">
    |  <loop>
    |   <read var="v0" bits="64" type="uint" endian="big"/>
    |   <read var="v1" bits="32" type="uint" endian="big"/>
    |   <sample/>
    |  </loop>
    | </decoder>
    | <tag id="sie:block_index" group="1" decoder="1"/>
    |
    |<!-- Stream-specific definitions begin here: -->
    |
    |<!--
    | Element Reference:
    |     ch = channel
    |    dim = dimension
    |  xform = transform
    |-->
    | <tag id="core:version">1.0</tag>
    | <tag id="somat:version">1.0</tag>
    ]==],
    types = {
        _histogram_bin_bounds32 = {
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <read var="v1" bits="32" type="float" endian="big"/>
            | <loop var="v0">
            |  <read var="v2" bits="32" type="float" endian="big"/>
            |  <sample/>
            |  <set var="v1" value="{$v2}"/>
            | </loop>
            |</decoder>
            ]==]},
        _histogram_bin_bounds64 = {
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <read var="v1" bits="64" type="float" endian="big"/>
            | <loop var="v0">
            |  <read var="v2" bits="64" type="float" endian="big"/>
            |  <sample/>
            |  <set var="v1" value="{$v2}"/>
            | </loop>
            |</decoder>
            ]==]},
        sequential_float32 = {
            sie_core_schema = 'somat:sequential',
            somat_datamode_type = 'time_history',
            somat_data_bits = 32,
            somat_data_format = 'float',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <read var="v0" bits="64" type="int" endian="big"/>
            | <loop var="v0">
            |  <read var="v1" bits="32" type="float" endian="big"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        sequential_float64 = {
            sie_core_schema = 'somat:sequential',
            somat_datamode_type = 'time_history',
            somat_data_bits = 64,
            somat_data_format = 'float',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <read var="v0" bits="64" type="int" endian="big"/>
            | <loop var="v0">
            |  <read var="v1" bits="64" type="float" endian="big"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        sequential_int16 = {
            sie_core_schema = 'somat:sequential',
            somat_datamode_type = 'time_history',
            somat_data_bits = 16,
            somat_data_format = 'int',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <read var="v0" bits="64" type="int" endian="big"/>
            | <loop var="v0">
            |  <read var="v1" bits="16" type="int" endian="big"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        message_can =  {
            sie_core_schema = 'somat:message',
            somat_datamode_type = 'message_log',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <loop>
            |  <read var="v0" bits="64" type="int" endian="big"/>
            |  <seek from="current" offset="12"/>
            |  <read var="ff" bits="8" type="uint"/>
            |  <seek from="current" offset="-13"/>
            |  <read var="v1" octets="{4 + ($ff &amp; 15)}" type="raw"/>
            |  <seek from="current" offset="{9 - ($ff &amp; 15)}"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        message_float32 =  {
            sie_core_schema = 'somat:message-numeric',
            somat_datamode_type = 'message_log',
            somat_data_bits = 32,
            somat_data_format = 'float',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <loop>
            |  <read var="v0" bits="64" type="int" endian="big"/>
            |  <read var="v1" bits="32" type="float" endian="big"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        message_float64 =  {
            sie_core_schema = 'somat:message-numeric',
            somat_datamode_type = 'message_log',
            somat_data_bits = 64,
            somat_data_format = 'float',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <loop>
            |  <read var="v0" bits="64" type="int" endian="big"/>
            |  <read var="v1" bits="64" type="float" endian="big"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
        message_raw =  {
            sie_core_schema = 'somat:message',
            somat_datamode_type = 'message_log',
            decoder = [==[
            |<decoder id="(=decoder_id=)">
            | <loop>
            |  <read var="size" bits="32" type="uint" endian="big"/>
            |  <read var="v0" bits="64" type="int" endian="big"/>
            |  <read var="v1" octets="{$size - 8}" type="raw"/>
            |  <sample/>
            | </loop>
            |</decoder>
            ]==]},
    },
}

function m:initialize(args)
    self.max_bufs = args.max_bufs + 5 -- XML + index + group_tag + slop
    self.bufs = ffi_new('sie_buffer[?]', self.max_bufs)
    self.last_buf = -1
    self.next_ids = {
        group = 0,
        test = 0,
        ch = 0,
        decoder = 2,
    }
    self.decoder_ids = {}
    self:template_init()
    if args.fd then
        self.fd = args.fd
        self.file_name = args.file_name or string.format("fd(%d)", self.fd)
    elseif args.file_name then
        local flags = bit.bor(C.O_CREAT, C.O_WRONLY, C.O_TRUNC, C.O_APPEND,
            C.O_LARGEFILE)
        -- ffi.new because open(), open64() declared as varargs functions
        local mode = ffi.new('mode_t', 0x1B6) -- rw-rw-rw  (0x124 = r--r--r--)
        self.fd = C.open(args.file_name, flags, mode)
        assert(self.fd >= 0)
        self.file_name = args.file_name
    else
        error('file_name or fd unspecified')
    end
    self.c_file_name = ffi_new('char[?]', #self.file_name + 1, self.file_name)
    self.xml_group = self:next_id('group')
    self.index_group = self:next_id('group')

    self.xml_buf_idx = 0
    self.index_buf_idx = 1
    self.scratch_buf_idx = 2

    -- The index new_buffer() is first so that index_buf gets set properly
    -- in all succeeding new_buffer() calls
    self:new_buffer(self.index_buf_idx, self.index_group,
        8192, ffi_sizeof'sie_index_entry')
    self.bufs[self.index_buf_idx].index_buf = nil   -- KLUDGE?

    -- KLUDGE make dynamic maybe
    self:new_buffer(self.xml_buf_idx, self.xml_group, 256 * 1024)
    self:new_buffer(self.scratch_buf_idx, 0 --[[placeholder]], 1024, 1)

    self.index_entry = ffi_new('sie_index_entry')

    if args.setup_compiler_mode then
        self.out_xml = {}
    else
        self:xml(self.statics.sie_header)
    end

    -- KLUDGE prime the C cache with C.close()
    -- LuaJIT interpreter needs to run a C.close() before the compiler can
    -- compile a C.close(). Have seen trace aborts on C.close not in cache.
    do
        local fd = C.open('/dev/null', C.O_RDONLY, 0)
        C.close(fd)
    end
end

function m:amount_written()
    local total = 0
    for i = 0, self.max_bufs - 1 do
        total = total + self.bufs[i].total_written
    end
    return math.floor(total)
end

function m:next_id(t)
    local id = self.next_ids[t]
    self.next_ids[t] = id + 1
    return id
end

function m:new_buffer(idx, group, bytes, sample_size, payload_overhead)
    group = group or self:next_id('group')
    if not idx then
        self.last_buf = self.last_buf + 1
        idx = self.last_buf
    end
    self.last_buf = math.max(idx, self.last_buf)
    assert(self.last_buf < self.max_bufs)
    local buf = self.bufs[idx]
    buf:initialize(self.fd, group, bytes, sample_size or 1,
        payload_overhead or 0)
    buf.idx = idx
    buf.file_name = self.c_file_name
    buf.index_buf = self.index_buf_idx and self.bufs[self.index_buf_idx] or nil
    return idx
end

function m:xml(template, ...)
    if not template then
        self.bufs[self.xml_buf_idx]:flush()
        return
    end
    template = self:template({output = false}, template, ...)
    if self.out_xml then
        self.out_xml[#self.out_xml + 1] = template
    else
        self.bufs[self.xml_buf_idx]:append(template)
    end
end

function m:emit(buf_name, data, len)
end

function m:close()
    self.bufs[self.xml_buf_idx]:flush()
    for i = 2, self.last_buf do
        self.bufs[i]:flush()
    end
    self.bufs[self.index_buf_idx]:flush()
    -- KLUDGE if a socket, need to do shutdown dance
    -- C.shutdown(self.fd, C.SHUT_WR) -- KLUDGE check for errors
    -- Then wait for read side to shutdown
    assert(C.close(self.fd) == 0)

    -- KLUDGE deal with GC stuff here
end

function m:begin_test(meta)
    self:xml()
    self.test_id = self:next_id('test')
    self:xml('\n<test id="(=test_id=)">\n', {test_id=self.test_id})
end

function m:decoder_id(datatype)
    local id = self.decoder_ids[datatype]
    if id then
        return id
    end
    self:xml() -- KLUDGE blocking test
    local id = self:next_id('decoder')
    self.decoder_ids[datatype] = id
    self:xml(assert(self.statics.types[datatype].decoder), {decoder_id=id})
    self:xml() -- KLUDGE blocking test
    return id
end

function m:end_test()
    self:xml('</test>\n')
    self:xml()
end

function m:test_header(test_id, run_number, run_description, start_time)
    self:xml([[
 <test id="(=test_id=)">
  <tag id="core:test_count">(=run_number=)</tag>
  <tag id="core:description">(=run_description=)</tag>
  <tag id="core:start_time">(=start_time=)</tag>
 </test>
]], {
        test_id = test_id or 0,
        run_number = run_number or 1,
        run_description = run_description or 'None',
        start_time = start_time or '0000-01-01T00:00:00-0000',
    })
end

function m:channel(item)
    local sample_rate = item.sample_rate
    if sample_rate then
        sample_rate = rational:new(item.sample_rate):tonumbers()
    end
    if not item._timescale and sample_rate and
            item.data_type:match('^sequential_') then
        item._timescale = 1 / sample_rate
    end
    local dt = self.statics.types[item.data_type]

    local md = {
        _sie_test = item._sie_test or 0, -- KLUDGE
        _sie_ch = self:next_id('ch'),
        _sie_group = self:next_id('group'),
        _sie_decoder = self:decoder_id(item.data_type),
        _sie_name = item.name,
        data_type = item.data_type,

        ['somat:armada_input_metadata'] = cjson.encode(item),
        ['somat:connector'] = item.connector,
        ['somat:input_channel'] = item.somat_input_channel or item.name,
        ['core:description'] = item.description,
        ['core:sample_rate'] = sample_rate,
        ['core:schema'] = dt.sie_core_schema,
        ['somat:datamode_type'] = dt.somat_datamode_type,
        ['somat:data_bits'] = dt.somat_data_bits,
        ['somat:data_format'] = dt.somat_data_format,
        ['somat:datamode_name'] = 'dm', -- KLUDGE!!!
        _dims = {
            {
                ['core:label'] = item.time_label or 'Time',
                ['core:units'] = item.time_units or 'Seconds',
                _scale = item._timescale
            }, {
                ['core:label'] = item.name,
                ['core:units'] = item.output_units
            }
        }
    }

    self:xml('<ch test="(=_sie_test=)" id="(=_sie_ch=)" name="(=_sie_name=)" group="(=_sie_group=)">\n',
        md)
    for k, v in pairs(md) do
        if k:sub(1, 1) ~= '_' then
            self:xml(' <tag id="(=k=)">(=v=)</tag>\n',
                {k=html_escape(k), v=html_escape(v)})
        end
    end
    for i, dim in ipairs(md._dims) do
        md._i = i - 1
        self:xml(' <dim index="(=_i=)">\n', md)
        for k, v in pairs(dim) do
            if k:sub(1, 1) ~= '_' then
                self:xml('  <tag id="(=k=)">(=v=)</tag>\n',
                    {k=html_escape(k), v=html_escape(v)})
            end
        end
        self:xml('  <data decoder="(=_sie_decoder=)" v="(=_i=)"/>\n', md)
        if dim._scale or dim._offset then
            local t = {scale = 1, offset = 0}
            if dim._scale then t.scale = get_rational(dim._scale) end
            if dim._offset then t.offset = get_rational(dim._offset) end
            self:xml('  <xform scale="(=scale=)" offset="(=offset=)"/>\n', t)
        end
        self:xml(' </dim>\n')
    end
    self:xml('</ch>\n')

    return md
end

function m:group_tag(id, format, data, len, indent, group)
    local gtb = self.bufs[self.scratch_buf_idx]
    indent = indent or ' '
    group = group or self:next_id('group')
    gtb.header.group = htonl(group)

    self:xml('(=i=)<tag id="(=id=)" group="(=grp=)" format="(=fmt=)"/>\n', {
        i = indent,
        id = id,
        grp = group,
        fmt = format,
    })
    len = len or #data

    gtb.payload:clear()
    self.bufs[self.xml_buf_idx]:flush()

    gtb.payload:expand(len)
    gtb:enq(data, len)
    gtb:flush()

    return group
end

function m:group_tag_from_file(id, format, filename)
    local content = ''
    local fh = io.open(filename)
    if fh then
        content = fh:read('*a')
        fh:close()
    end
    self:group_tag(id, format, content)
end

return m
