require 'middleclass'
local ffi = require 'ffi'
local C = ffi.C
local bit = require 'bit'

ffi.cdef[[
typedef int ssize_t;

ssize_t read(int, void *, size_t);
int close(int);

typedef long    time_t;

struct timespec {
    time_t      tv_sec;
    long        tv_nsec;
};

struct itimerspec {
    struct timespec     it_interval;
    struct timespec     it_value;
};

int timerfd_create(int, int);
int timerfd_settime(int, int, const struct itimerspec *, struct itmerspec *);
int timerfd_gettime(int, struct itimerspec *);

enum {
    CLOCK_MONOTONIC = 1,
};

enum {
    TFD_CLOEXEC = 02000000,
    TFD_NONBLOCK = 04000,
};

struct epoll_event {
    uint32_t    events;
    uint64_t    data;
};

int epoll_wait(int, struct epoll_event *, int, int);
int epoll_create(int);
int epoll_create1(int);
int epoll_ctl(int, int, int, struct epoll_event *);

enum {
    EPOLL_CTL_ADD = 1,
    EPOLL_CTL_DEL,
    EPOLL_CTL_MOD,
};

enum {
    EPOLLIN = 1,
    EPOLLPRI = 2,
    EPOLLOUT = 4,
    EPOLLERR = 8,
    EPOLLHUP = 0x10,
    EPOLLRDNORM = 0x40,
    EPOLLRDBAND = 0x80,
    EPOLLWRNORM = 0x100,
    EPOLLWRBAND = 0x200,
    EPOLLMSG = 0x400,
    EPOLLRDHUP = 0x2000,
};

struct signalfd_siginfo {
    uint32_t    ssi_signo;
    int32_t     ssi_errno;
    int32_t     ssi_code;
    uint32_t    ssi_pid;
    uint32_t    ssi_uid;
    int32_t     ssi_fd;
    uint32_t    ssi_tid;
    uint32_t    ssi_band;
    uint32_t    ssi_overrun;
    uint32_t    ssi_trapno;
    int32_t     ssi_status;
    int32_t     ssi_int;
    uint64_t    ssi_ptr;
    uint64_t    ssi_utime;
    uint64_t    ssi_stime;
    uint64_t    ssi_addr;
    uint8_t     pad[48];
};

enum {
    SFD_NONBLOCK = 04000,
};

enum {
    _SIGSET_NWORDS = (1024 / (8 * sizeof(unsigned long int))),
};

typedef struct {
    unsigned long int   val[_SIGSET_NWORDS];
} sigset_t;

int signalfd(int, const sigset_t *, int);
int sigprocmask(int how, const sigset_t *, sigset_t *);
int sigemptyset(sigset_t *);
int sigaddset(sigset_t *, int);

enum {
    SIG_BLOCK = 0,
};

enum {
    SIGHUP = 1,
    SIGINT = 2,
    SIGQUIT = 3,
    SIGILL = 4,
    SIGTRAP = 5,
    SIGABRT = 6,
    SIGIOT = 6,
    SIGBUS = 7,
    SIGFPE = 8,
    SIGKILL = 9,
    SIGUSR1 = 10,
    SIGSEGV = 11,
    SIGUSR2 = 12,
    SIGPIPE = 13,
    SIGALRM = 14,
    SIGTERM = 15,
    SIGSTKFLT = 16,
    SIGCHLD = 17,
    SIGCONT = 18,
    SIGSTOP = 19,
    SIGTSTP = 20,
    SIGTTIN = 21,
    SIGTTOU = 22,
    SIGURG = 23,
    SIGXCPU = 24,
    SIGXFSZ = 25,
    SIGVTALRM = 26,
    SIGPROF = 27,
    SIGWINCH = 28,
    SIGIO = 29,
    SIGPOLL = SIGIO,
    SIGPWR = 30,
    SIGSYS = 31,
    SIGUNUSED = 31,
    SIGRTMIN = 32,
};

time_t time(time_t *);

enum {
    EVBACKEND_SELECT = 1,
    EVBACKEND_POLL = 2,
    EVBACKEND_EPOLL = 4,
};
]]

-- Why did I do this?  Because EAGAIN is defined in some LUA FFI modules that
-- may use this module and not defined in others.
local EAGAIN = 11

-- Event class
local ev = class('ev')

-- Loop class
ev.Loop = class('ev.Loop')

function ev.Loop:initialize()
    self.rdepth = 0
    self.fd = C.epoll_create1(0)
    if self.fd < 0 then
        error("Error creating epoll instance", ffi.errno())
    end
    self.timeout_ms = 100
    self.max_revents = 20
    self.revents = ffi.new('struct epoll_event[?]', self.max_revents)
    self.watchers = { }
    self.idle_watchers = { }
    self.ep_event = ffi.new('struct epoll_event[1]')
    return self
end

function ev.Loop:start(watcher)
    if not watcher.fd then
        self.idle_watchers[#self.idle_watchers + 1] = watcher
        return
    end
    local f = assert(tonumber(watcher.fd))
    local mode = C.EPOLL_CTL_ADD
    local events = 0
    local wl = self.watchers[f]
    if not wl then
        self.watchers[f] = {}
        wl = self.watchers[f]
    else
        mode = C.EPOLL_CTL_MOD
    end
    wl[#wl + 1] = watcher
    for _, v in pairs(wl) do
        events = bit.bor(events, v.events)
    end
    self.ep_event[0].events = events
    self.ep_event[0].data = watcher.fd
    local status = C.epoll_ctl(self.fd, mode, watcher.fd, self.ep_event)
    if status < 0 then
        error("Unable to add FID to epoll", ffi.errno())
    end
end

function ev.Loop:_loop_exit_check()
    for _, _ in pairs(self.idle_watchers) do
        return
    end

    -- If there are ANY watchers still attached to this loop, keep running
    for _, _ in pairs(self.watchers) do
        return
    end
    self.loop_exit = true
end

function ev.Loop:stop(watcher)
    if not watcher.fd then
        for i, v in pairs(self.idle_watchers) do
            if v == watcher then
                self.idle_watchers[i] = nil
                self:_loop_exit_check()
                return
            end
        end
        return
    end

    local f = assert(tonumber(watcher.fd))
    local events

    -- If the watcher list for this FID doesn't exist, it means start was never
    -- called for that watcher and thus the FID was never added to the FID list
    -- for epoll.  At this point we can just return.
    if not self.watchers[f] then
        return
    end

    -- First remove this watcher from the list
    for i, v in ipairs(self.watchers[f]) do
        if v == watcher then
            self.watchers[f][i] = nil
            if #self.watchers[f] == 0 then
                self.watchers[f] = nil
            end
            break
        end
    end

    -- Now check to see if there are any watchers left in the list that are
    -- using this same FID
    if self.watchers[f] then
        local ev = ffi.new('struct epoll_event[1]')
        ev[0].events = 0
        ev[0].data = watcher.fd
        for i, v in ipairs(self.watchers[f]) do
            ev[0].events = bit.bor(ev[0].events, v.events)
        end
        C.epoll_ctl(self.fd, C.EPOLL_CTL_MOD, watcher.fd, ev)
    else
        C.epoll_ctl(self.fd, C.EPOLL_CTL_DEL, watcher.fd, nil)
    end

    self:_loop_exit_check()
end


function ev.Loop:loop()
    self.rdepth = self.rdepth + 1
    self.iterations = 0
    self.loop_exit = false
    while self.loop_exit == false do
        self.iterations = self.iterations + 1
        local status = C.epoll_wait(self.fd, self.revents, self.max_revents,
            self.timeout_ms)
        self.time = C.time(nil)
        if status < 0 then
            error("Error from epoll_wait", ffi.errno())
        elseif status > 0 then
            for i = 0, status - 1 do
                local re = self.revents[i]
                local fid = tonumber(re.data)

                -- Walk the list of watchers and call the callbacks for each
                -- watcher attached to the FID who has an event pending.
                -- NOTE, I don't use for...ipairs here because it is really
                -- slow compared to just walking over the list starting at
                -- 1 (one).
                local wlist = self.watchers[fid]
                for j = 1, #wlist do
                    local w = wlist[j]
                    if re.events == w.events then
                        if w.cb then
                            w.cb(self, w, w.events)
                        end
                    end
                end
            end
        else
            for _, w in ipairs(self.idle_watchers) do
                w.cb(self, w, 0)
            end
        end
    end
    self.rdepth = self.rdepth - 1
end

function ev.Loop:now()
    if not self.time then
        self:update_now()
    end
    return self.time
end

function ev.Loop:update_now()
    self.time = C.time(nil)
end

function ev.Loop:iteration()
    return self.iterations or 0
end

function ev.Loop:depth()
    return self.rdepth or 0
end

function ev.Loop:unloop()
    self.loop_exit = true
end

function ev.Loop:backent()
    return C.EVBACKEND_EPOLL
end

function ev.Loop.new(...)
    local instance = ev.Loop:allocate()
    instance:initialize(...)
    return instance
end

ev.Loop.default = ev.Loop.new()
ev.Loop.default.is_default = true

function ev.Loop:is_default()
    return self.is_default or false
end


-- Watcher class
ev.watcher = class('ev.watcher')

function ev.watcher:initialize(fd, events, callback)
    self.fd = fd
    self.events = events
    self.cb = callback
    self.active = false
    self.pending = false
end

function ev.watcher:start(loop)
    self.active = true
    loop:start(self)
end

function ev.watcher:stop(loop)
    self.active = false
    loop:stop(self)
end

function ev.watcher:callback(new_callback)
    local c = self.cb
    self.cb = new_callback or self.cb
    return c
end

function ev.watcher:is_active()
    return self.active
end

function ev.watcher:is_pending()
    return self.pending
end


-- IO watcher class
ev.IO = class('ev.IO', ev.watcher)

function ev.IO.new(...)
    local instance = ev.IO:allocate()
    instance:initialize(...)
    return instance
end

function ev.IO:initialize(callback, fd, events)
    ev.watcher.initialize(self, fd, events, callback)
    return self
end

function ev.IO:getfd()
    return self.fd
end


-- Signal watcher class
ev.Signal = class('ev.Signal', ev.watcher)

function ev.Signal.new(...)
    local instance = ev.Signal:allocate()
    instance:initialize(...)
    return instance
end

function ev.Signal:initialize(callback, signo)
    self.signo = signo
    self.is_daemon = false
    self.siginfo = ffi.new('struct signalfd_siginfo[1]')
    self.siginfo_size = ffi.sizeof(self.siginfo)
    self.sigset = ffi.new('sigset_t[1]')
    self.sigset_size = ffi.sizeof(self.sigset)
    self.callback = callback
    ffi.fill(self.sigset, self.sigset_size)
    local word = (signo - 1) / (8 * ffi.sizeof('unsigned long int'))
    local mask = bit.lshift(1,
        ((signo - 1) % (8 * ffi.sizeof('unsigned long int'))))
    self.sigset[0].val[word] = bit.bor(self.sigset[0].val[word], mask)
    local fd = C.signalfd(-1, self.sigset, C.SFD_NONBLOCK)
    if fd < 0 then
        error("signalfd failure ", ffi.errno())
    end
    C.sigprocmask(C.SIG_BLOCK, self.sigset, nil)

    ev.watcher.initialize(self, fd, C.EPOLLIN, function(loop, w, ev)
            local amount_read = C.read(self.fd, self.siginfo, self.siginfo_size)
            if amount_read ~= self.siginfo_size then
                local e = ffi.errno()
                if amount_read < 0 and e == EAGAIN then
                    return
                end
                error("Error reading from signalfd" .. e)
            end
            if self.signo == self.siginfo[0].ssi_signo then
                if self.callback then self.callback(loop, w, ev) end
            end
        end)
    return self
end

function ev.Signal:stop(loop)
    C.close(self.fd)
    ev.watcher.stop(self, loop)
end

function ev.Signal:start(loop, is_daemon)
    self.is_daemon = is_daemon or self.is_daemon
    ev.watcher.start(self, loop)
end


-- Idle watcher class
ev.Idle = class('ev.Idle', ev.watcher)

function ev.Idle.new(...)
    local instance = ev.Idle:allocate()
    instance:initialize(...)
    return instance
end

function ev.Idle:initialize(callback)
    self.cb = callback
    self.active = false
    self.pending = false
    return self
end

function ev.Idle:start(loop)
    ev.watcher.start(self, loop)
end

function ev.Idle:stop(loop)
    ev.watcher.stop(self, loop)
end


-- Timer watcher class
ev.Timer = class('ev.Timer', ev.watcher)

function ev.Timer.new(...)
    local instance = ev.Timer:allocate()
    instance:initialize(...)
    return instance
end

function ev.Timer:initialize(callback, initial_timeout, interval_timeout)
    self.itimerspec = ffi.new('struct itimerspec')
    self.ritmerspec = ffi.new('struct itimerspec[1]')

    local interval_timeout = interval_timeout or 0

    initial_timeout = initial_timeout * 1e9
    interval_timeout = interval_timeout * 1e9

    self.itimerspec.it_interval.tv_sec = math.floor(interval_timeout / 1e9)
    self.itimerspec.it_interval.tv_nsec = math.floor(interval_timeout % 1e9)

    self.itimerspec.it_value.tv_sec = math.floor(initial_timeout / 1e9)
    self.itimerspec.it_value.tv_nsec = math.floor(initial_timeout % 1e9)

    local fd =
        C.timerfd_create(C.CLOCK_MONOTONIC, C.TFD_NONBLOCK + C.TFD_CLOEXEC)
    if fd < 0 then
        print("Unable to create timer", ffi.errno())
        return
    end

    self.buf = ffi.new('unsigned long long[1]')
    self.buf_size = ffi.sizeof(self.buf)

    self.callback = callback
    ev.watcher.initialize(self, fd, C.EPOLLIN, function(loop, timer, events)
        local amount_read = C.read(self.fd, self.buf, self.buf_size)
        if amount_read ~= self.buf_size then
            local e = ffi.errno()
            if e == EAGAIN then return end
            error("Error reading from timerfd", ffi.errno())
        end
        if self.callback then self.callback(loop, timer, events) end
    end)

    return self
end

function ev.Timer:stop(loop)
    self.itimerspec.it_value.tv_sec = 0
    self.itimerspec.it_value.tv_nsec = 0
    self:again(loop, 0)
    ev.watcher.stop(self, loop)
    C.close(self.fd)
end

function ev.Timer:start(loop, is_daemon)
    self:again(loop)
    ev.watcher.start(self, loop)
end

function ev.Timer:again(loop, new_interval)
    if new_interval ~= nil then
        new_interval = new_interval * 1e9

        self.itimerspec.it_interval.tv_sec = math.floor(new_interval / 1e9)
        self.itimerspec.it_interval.tv_nsec = math.floor(new_interval % 1e9)
    end
    C.timerfd_settime(self.fd, 0, self.itimerspec, self.ritimerspec)
end

function ev.Timer:callback(new_callback)
    local c = self.callback
    self.callback = new_callback or self.callback
    return c
end

ev.READ = C.EPOLLIN
ev.WRITE = C.EPOLLOUT

ev.SIGHUP = C.SIGHUP
ev.SIGINT = C.SIGINT
ev.SIGQUIT = C.SIGQUIT
ev.SIGILL = C.SIGILL
ev.SIGTRAP = C.SIGTRAP
ev.SIGABRT = C.SIGABRT
ev.SIGIOT = C.SIGIOT
ev.SIGBUS = C.SIGBUS
ev.SIGFPE = C.SIGFPE
ev.SIGKILL = C.SIGKILL
ev.SIGUSR1 = C.SIGUSR1
ev.SIGSEGV = C.SIGSEGV
ev.SIGUSR2 = C.SIGUSR2
ev.SIGPIPE = C.SIGPIPE
ev.SIGALRM = C.SIGALRM
ev.SIGTERM = C.SIGTERM
ev.SIGSTKFLT = C.SIGSTKFLT
ev.SIGCHLD = C.SIGCHLD
ev.SIGCONT = C.SIGCONT
ev.SIGSTOP = C.SIGSTOP
ev.SIGTSTP = C.SIGTSTP
ev.SIGTTIN = C.SIGTTIN
ev.SIGTTOU = C.SIGTTOU
ev.SIGURG = C.SIGURG
ev.SIGXCPU = C.SIGXCPU
ev.SIGXFSZ = C.SIGXFSZ
ev.SIGVTALRM = C.SIGVTALRM
ev.SIGPROF = C.SIGPROF
ev.SIGWINCH = C.SIGWINCH
ev.SIGIO = C.SIGIO
ev.SIGPOLL = C.SIGPOLL
ev.SIGPWR = C.SIGPWR
ev.SIGSYS = C.SIGSYS
ev.SIGUNUSED = C.SIGUNUSED
ev.SIGRTMIN = C.SIGRTMIN

return ev
