--! SQLite bindings for LuaJIT.
--! \file

--! \mainpage
--! Start with the #sqlite class.

require 'strict'
require('lj_include')('sqlite3')

local ffi = require 'ffi'

local C = ffi.C
local sqlite3_ffi = ffi.load('libsqlite3.so.0')

local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_string = ffi.string

local function ok(obj, v)
    if v ~= C.SQLITE_OK then
        local db = obj:db()
        error(string.format("SQLite error %d: %s\n", v, db:errmsg()))
    end
end

local stmt_cache = {}

--! A SQLite database connection.
--? class sqlite_db
local sqlite_db = {}

--! Returns the raw database handle.
function sqlite_db:db()
    return self
end

--! Closes the database.
function sqlite_db:close()
    for stmt, _ in pairs(stmt_cache) do
        stmt:finalize()
    end
    return sqlite3_ffi.sqlite3_close(self)
end

--! Get error message
function sqlite_db:errmsg()
    return ffi_string(sqlite3_ffi.sqlite3_errmsg(self))
end

--! Busy timeout
function sqlite_db:busy_timeout(secs)
    return sqlite3_ffi.sqlite3_busy_timeout(self, secs * 1000)
end

--! Evaluates one or more queries in a string.
--! \param q The SQL queries to evaluate
function sqlite_db:eval(q)
    while q and #q do
        local stmt, tail = self:prepare(q)
        local result = stmt:exec()
        ok(self, stmt:finalize())
        q = tail
    end
end

--! Evaluates a single query in a string.
--! \param q the SQL query to evaluate
function sqlite_db:exec(q, ...)
    local stmt = self:prepare(q)
    local result = stmt:exec(...)
    ok(self, stmt:finalize())
    return result
end

--! Creates a new prepared statement (#sqlite_stmt).
--! \param q the SQL query to evaluate
--! \return two values: the prepared statement, and any leftover SQL
function sqlite_db:prepare(q)
    local _stmt = ffi_new('sqlite3_stmt *[1]')
    local _tail = ffi_new('const char *[1]')
    local pq = ffi_cast('const char *', q)
    ok(self, sqlite3_ffi.sqlite3_prepare_v2(self, pq, #q, _stmt, _tail))
    local tail = nil
    local tail_len = #q - (_tail[0] - pq)
    if tail_len > 0 then
        tail = ffi_string(_tail[0], tail_len)
    end
    local stmt = _stmt[0]
    stmt:initialize()
    return stmt, tail
end

ffi.metatype('sqlite3', { __index = sqlite_db })

--! A prepared statement.
--? class sqlite_stmt
local sqlite_stmt = {}

--! Returns the raw database handle.
function sqlite_stmt:db()
    return sqlite3_ffi.sqlite3_db_handle(self)
end

function sqlite_stmt:bind_value(n, v)
    if type(n) ~= 'number' then
        local bind_names = self:bind_names()
        local name = n
        n = bind_names[name]
        if not n then
            error(string.format( -- KLUDGE do something better here
                'sqlite3 bind_value() of unknown placeholder "%s"\n', name))
        end
    end
    local t = type(v)
    if t == 'number' then
        sqlite3_ffi.sqlite3_bind_double(self, n, v)
    elseif t == 'string' then
        sqlite3_ffi.sqlite3_bind_text(self, n, v, #v,
            -- #define SQLITE_TRANSIENT ((sqlite3_destructor_type)-1)
            ffi_cast('sqlite3_destructor_type', -1))
    elseif t == 'nil' then
        sqlite3_ffi.sqlite3_bind_null(self, n)
    else
        error(string.format( -- KLUDGE do something better here
            'sqlite3 bind_value() of unhandled type %s\n', t))
    end
end

function sqlite_stmt:bind(a)
    for i, v in pairs(a) do
        self:bind_value(i, v)
    end
end

function sqlite_stmt:column_count()
    return sqlite3_ffi.sqlite3_column_count(self)
end

function sqlite_stmt:column_name(i)
    return ffi_string(sqlite3_ffi.sqlite3_column_name(self, i))
end

function sqlite_stmt:column_names()
    local names = stmt_cache[self].column_names
    if not names then
        names = {}
        local cols = self:column_count()
        for i = 0, cols - 1 do
            names[i] = self:column_name(i)
        end
        stmt_cache[self].column_names = names
    end
    return names
end

function sqlite_stmt:bind_names()
    local names = stmt_cache[self].bind_names
    if not names then
        names = {}
        local count = sqlite3_ffi.sqlite3_bind_parameter_count(self)
        for i = 1, count do
            local cname = sqlite3_ffi.sqlite3_bind_parameter_name(self, i)
            if cname == nil then
                names[i] = i
            elseif cname[0] == 0x3F then -- '?'
                names[tonumber(ffi_string(cname + 1))] = i
            else
                names[ffi_string(cname + 1)] = i
            end
        end
        stmt_cache[self].bind_names = names
    end
    return names
end

function sqlite_stmt:column_value(n)
    local t = sqlite3_ffi.sqlite3_column_type(self, n)
    if t == C.SQLITE_INTEGER or t == C.SQLITE_FLOAT then
        return tonumber(sqlite3_ffi.sqlite3_column_double(self, n))
    elseif t == C.SQLITE_NULL then
        return nil
    else
        local bytes = sqlite3_ffi.sqlite3_column_bytes(self, n)
        local blob = sqlite3_ffi.sqlite3_column_blob(self, n)
        return ffi_string(ffi_cast('const char *', blob), bytes)
    end
end

function sqlite_stmt:exec(values)
    local cols = self:column_count()
    local col_names = self:column_names()
    local result = {}
    if not values or type(values[1]) ~= 'table' then
        if values then
            self:bind(values)
        end
        self:churn(result, cols, col_names)
    else
        for _, v in ipairs(values) do
            self:bind(v)
            self:churn(result, cols, col_names)
        end
    end
    return result
end

function sqlite_stmt:finalize()
    stmt_cache[self] = nil
    return sqlite3_ffi.sqlite3_finalize(self)
end

function sqlite_stmt:initialize()
    stmt_cache[self] = {}
end

function sqlite_stmt:reset()
    return sqlite3_ffi.sqlite3_reset(self)
end

function sqlite_stmt:churn(result, cols, col_names)
    while true do
        local got = self:step()
        if got == C.SQLITE_ROW then
            local row = {}
            for i = 0, cols - 1 do
                row[col_names[i]] = self:column_value(i)
            end
            result[#result + 1] = row
        elseif got == C.SQLITE_DONE then
            break
        elseif got == C.SQLITE_BUSY then
            -- KLUDGE do callback here ???
            break
        else
            error(string.format('sqlite3_step() error %g = %s\n',
                tonumber(got), self:db():errmsg()))
        end
    end
    ok(self, self:reset())
end

function sqlite_stmt:step()
    return sqlite3_ffi.sqlite3_step(self)
end

ffi.metatype('sqlite3_stmt', { __index = sqlite_stmt })

--! It's a sqlite class.
--? class sqlite
local sqlite = { }

--! Creates a new database connection (#sqlite_db).
function sqlite:new(db_file, opts)
    opts = opts or {wal = true}
    local _db = ffi_new('sqlite3 *[1]')
    assert(sqlite3_ffi.sqlite3_open(db_file, _db) == C.SQLITE_OK)
    if opts.wal and db_file ~= ':memory:' then
        local r = _db[0]:exec('PRAGMA journal_mode = WAL;')
        assert(r[1].journal_mode == 'wal', 'Error setting journal_mode = WAL')
        _db[0]:exec('PRAGMA synchronous = FULL;')
    end
    return _db[0]
end

return sqlite
