local base = _G

local assert, error, io, loadstring, pairs, print, string, table, type
    = assert, error, io, loadstring, pairs, print, string, table, type

local lpeg = require "lpeg"
local lfs = require "lfs"

local      P,      R,      S,      V
    = lpeg.P, lpeg.R, lpeg.S, lpeg.V
local      C,      Carg,      Cb,      Cc,      Cf,      Cg,      Cmt,      Ct
    = lpeg.C, lpeg.Carg, lpeg.Cb, lpeg.Cc, lpeg.Cf, lpeg.Cg, lpeg.Cmt, lpeg.Ct

module("st_compiler")

local m = {}
local metat = { __index = m }

function new(self, o)
    o = o or {}
    base.setmetatable(o, metat)
    self.__index = self
    return o
end

local function node(tag, v)
    return Ct(Cg(Cc(tag), "type") * v)
end

function m:parse(template, delimiters)
    if not delimiters then
        delimiters = self.delimiters
    end
    if not delimiters then
        delimiters = {"$", "$"}
    end

    local ldelim = P(delimiters[1])
    local rdelim = P(delimiters[2])
    local whitespace = S(" \t\r\n")
    local ws = whitespace^1
    local ows = whitespace^0
    local id_start = R("az", "AZ", "__", "//")
    local id_rest = P(id_start + R("09"))
    local function id(tag)
        return ows * Cg(id_start * id_rest^0, tag) * ows
    end
    local function reserved(word)
        return ows * P(word) * ows
    end
    -- KLUDGE quoted_string wrong - placeholder for now - needs quote escapes
    local quoted_string = P("\"") * Cg((1 - S("\r\n\""))^0, "content") * P("\"")
    local lparen = ows * "(" * ows
    local rparen = ows * ")" * ows
    local function if_content(name, excludes)
        return Cg(Ct((V"element" - (ldelim * ows * (excludes)))^0), name)
    end

    local st_grammar = P{
        'template'; -- Start rule

        template = node("template",
              V"element"^0),
        element =
              V"ifstat"
            + V"text"
            + V"comment"
            + V"expr_tag"
            + V"error",
        text = node("text",
              Cg((1 - ldelim)^1, "content")),
        comment = node("comment",
              ldelim * "!"
            * Cg((1 - ("!" * rdelim))^0, "content")
            * "!" * rdelim),
        expr_tag =
              ldelim * V"expr" * V"render" * rdelim,
        subtemplate = Ct(
              ows
            * "{" * (V"subtemplate_text" + V"expr_tag" + V"comment")^0 * "}"
            * ows),
        subtemplate_text = node("text",
              Cg((1 - (ldelim + "}"))^1, "content")),
        ifstat = node("if_statement", Cg(Ct(
            Ct(
              ldelim * reserved("if")
            * lparen * V"conditional" * rparen * rdelim
            * if_content("content",
                reserved("elseif") + reserved("else") + reserved("endif"))
            )
            * (Ct(
                  ldelim * reserved("elseif")
                * lparen * V"conditional" * rparen * rdelim
                * if_content("content",
                    reserved("elseif") + reserved("else") + reserved("endif"))
              ))^0
            ), "conditionals")
            * (ldelim * reserved("else") * rdelim
                * if_content("else_clause", reserved("endif")))^-1
            * ldelim * reserved("endif") * rdelim
            ),
        conditional = Cg(Ct(
              V"and_conditional" * ("||" * V"and_conditional")^0 * V"render"
            ), "condition"),
        and_conditional =
              V"not_conditional" * ("&&" * V"not_conditional")^0,
        not_conditional =
              ows * node("cond_not",
                  "!" * Cg(Ct(V"member_expr" * V"render"), "content"))
            + V"member_expr",
        expr =
              V"map_expr",
        map_expr =
              V"member_expr" * ("," * V"member_expr")^0
            * (
                  node("map", ":" * ows * V"template_ref")
                  -- KLUDGE stuff missing here
              )^0,
        member_expr =
              V"call_expr" * (
                  node("attribute_member", "." * id("content"))
                + node("attribute_indirect",
                      "."
                    * lparen
                    * Cg(Ct(V"map_expr" * V"render"), "content")
                    * rparen)
              )^0,
        call_expr =
              id("func") * lparen * V"expr" * rparen
              -- KLUDGE missing (super.) below
            + node("invoke_direct", V"template_name" * V"args")
              -- KLUDGE missing '@' stuff here
            + V"primary",
        primary = ows * (
              node("invoke_direct", V"template_name" * V"args")
            + V"attribute"
            + node("string", quoted_string)
            + V"subtemplate"
            -- + V"list" KLUDGE MISSING
              -- KLUDGE The following two rules are a single rule in the
              -- StringTemplate ANTLR grammar - maybe this can be done better
            + node("invoke_indirect",
                lparen * Cg(Ct(V"expr" * V"render"), "content") * rparen
                * V"args"^1)
            + lparen * V"expr" * rparen
            ) * ows,
        expr_no_comma =
              V"member_expr" * ( ":" * V"template_ref" )^-1,
        template_name =
              id("template"),
        args = Cg(
              Ct(lparen * (V"arg" * ("," * V"arg")^0)^-1 * rparen),
            "args"),
        arg = Ct(
              id("name")
            * "=" * Cg(Ct(V"expr_no_comma" * node("render", "")), "value")),
        template_ref =
              id("template") * V"args"
            + Cg(V"subtemplate", "subtemplate")
            + lparen * Cg(Ct(V"expr" * V"render"), "indirect_template") * rparen
                * V"args"^0,
        attribute =
              node("attribute_direct", V"attribute_name"),
              -- + node("attribute_indirect", "(" * V"attribute_name" * ")"),
        attribute_name =
              id("content"),
        options = Cg(
              Ct(ows * ";" * V"option" * ("," * V"option")^0),
            "options"),
        option = Ct(
              id("name") * "=" * Cg(Ct(V"expr"), "value")),
        render =
              node("render", V"options"^-1),
        error = C(P(1)^1) / "SYNTAX ERROR: `%0`"
    }

    return lpeg.match(st_grammar, template)
end

local function sourcify(str)
    for i = 1, 10 do
        local middle = string.rep("=", i)
        local delim = "]" .. middle .. "]"
        if not string.match(str, delim) then
            return "[" .. middle .. "[\n" .. str .. delim
        end
    end
    assert(false == "Can't escape string: " .. str)
end

local function stringify(str)
    -- KLUDGE WRONG MUST ESCAPE "
    return "\"" .. str .. "\""
end

function m:out(...)
    self._out[#self._out + 1] = string.format(...) 
end

function m:begin_template(name)
    if not name then
        name = #self.templates + 1
    end
    self.templates[name] = {}

    self.out_stack[#self.out_stack + 1] = name
    self._out = self.templates[name]
end

function m:end_template()
    local name = self.out_stack[#self.out_stack]
    self.out_stack[#self.out_stack] = nil
    self._out = self.templates[self.out_stack[#self.out_stack]]
    -- Make the template into a single string
    self.templates[name] = { table.concat(self.templates[name]) }
    -- So that we can optimize it away if we already have one
    if type(name) == "number" then
        local tpl = self.templates[name][1]
        local cached = self.template_cache[tpl]
        if cached then
            self.templates[name] = nil
            name = cached
        else
            self.template_cache[tpl] = name
        end
    end
    return self.templates[name], name
end

local function option_value(tree, name)
    if tree.options then
        for i = 1, #tree.options do
            local opt = tree.options[i]
            if opt.name == name then
                return opt.value
            end
        end
    end
    return nil
end

local builders -- forward reference

local function build(self, tree)
    for i = 1, #tree do
        local v = tree[i]
        local fn = builders[v.type]
        if fn then
            fn(self, v)
        else
            error("UNHANDLED type " .. v.type)
        end
    end
end

local function expr(self, v)
    self:begin_template()
    build(self, v)
    local tpl, tpl_name = self:end_template()
    local out = string.format("templates[%d]", tpl_name)
    return out
end

local function render_expr_to(self, e, lhs)
    self:out("    push_writer()\n")
    self:out("    %s()\n", expr(self, e))
    self:out("    %s = pop_writer()\n", lhs)
end

local function set_args(self, v)
    if v.args then
        for i = 1, #v.args do
            local a = v.args[i]
            self:out("    push_writer()\n")
            self:out("    %s()\n", expr(self, a.value))
            self:out("    set(%s, pop_writer())\n", stringify(a.name))
        end
    end
end

builders = {
    attribute_direct = function(self, v)
        self:out("    id = {%s}\n", stringify(v.content))
        self:out("    v = get(id)\n")
        end,
    attribute_indirect = function(self, v)
        render_expr_to(self, v.content, "id[#id + 1]")
        self:out("    v = get(id)\n")
        end,
    attribute_member = function(self, v)
        self:out("    id[#id + 1] = %s\n", stringify(v.content))
        self:out("    v = get(id)\n")
        end,
    comment = function(self, v)
        -- Do nothing
        end,
    cond_not = function(self, v)
        self:begin_template()
        self:out("    if not if_cond(%s) then\n", expr(self, v.content))
        self:out("        writer(\"1\")\n")
        self:out("    end\n")
        local tpl, tpl_name = self:end_template()
        self:out("    templates[%d]()\n", tpl_name)
        end,
    if_statement = function(self, v)
        local if_op = "if"
        for i = 1, #v.conditionals do
            local cond = v.conditionals[i]
            self:out("    %s if_cond(%s) then\n",
                if_op, expr(self, cond.condition))
            self:out("        %s()\n", expr(self, cond.content))
            if_op = "elseif"
        end
        if v.else_clause then
            self:out("    else\n")
            self:out("        %s()\n", expr(self, v.else_clause))
        end
        self:out("    end\n")
        end,
    invoke_direct = function(self, v)
        self:out("    push_context()\n")
        set_args(self, v)
        self:out("    templates[%s]()\n", stringify(v.template))
        self:out("    pop_context()\n")
        end,
    invoke_indirect = function(self, v)
        self:out("    push_context()\n")
        set_args(self, v)
        self:out("    push_writer()\n")
        self:out("    %s()\n", expr(self, v.content))
        self:out("    templates[pop_writer()]()\n")
        self:out("    pop_context()\n")
        end,
    map = function(self, v)
        self:out("    do\n")
        self:out("    nv = {}\n")
        self:out("    push_context()\n")
        set_args(self, v)
        if v.indirect_template then
            self:out("    push_writer()\n")
            self:out("    %s()\n", expr(self, v.indirect_template))
            self:out("    local ind_tpl = pop_writer()\n")
        end
        self:out("    for i = 1, #v do\n")
        self:out("        push_context()\n") -- KLUDGE necessary?
        self:out("        r_set(\"i\", i)\n")
        self:out("        r_set(\"i0\", i - 1)\n")
        self:out("        r_set(\"it\", v[i])\n")
        self:out("        push_writer()\n")
        if v.subtemplate then
            self:out("        %s()\n", expr(self, v.subtemplate))
        elseif v.indirect_template then
            self:out("        templates[ind_tpl]()\n")
        else
            self:out("        templates[%s]()\n", stringify(v.template))
        end
        self:out("        nv[i] = pop_writer()\n")
        self:out("        pop_context()\n") -- KLUDGE necessary?
        self:out("    end\n")
        self:out("    pop_context()\n")
        self:out("    v = nv\n")
        self:out("    end\n")
        end,
    render = function(self, v)
        local sep = option_value(v, "separator")
        local sep_out = "nil"
        if sep then
            sep_out = expr(self, sep)
        end
        self:out("    render(v, %s)\n", sep_out)
        end,
    string = function(self, v)
        self:out("    writer(%s)\n", sourcify(v.content))
        end,
    text = function(self, v)
        self:out("    writer(%s)\n", sourcify(v.content))
        end,
}

local header = [=[
    local m, output

    local chunks

    local function writer(...)
        output[#output + 1] = table.concat({...})
    end

    local function push_writer()
        chunks[#chunks + 1] = output
        output = {}
    end

    local function pop_writer()
        local out = table.concat(output)
        output = chunks[#chunks]
        chunks[#chunks] = nil
        return out
    end

    local function push_context()
        m[#m + 1] = {}
    end

    local function pop_context()
        m[#m] = nil
    end

    local function set(spec, value)
        m[#m][spec] = value
    end

    local function r_set(spec, value) -- For StringTemplate reserved words
        m[#m][spec] = value -- not handled any differently for now
    end

    local function if_cond(tpl_ref)
        push_writer()
        tpl_ref()
        local cond = pop_writer()
        if cond and cond ~= "" then
            return true
        end
        return false
    end

    local function get(spec)
        local v
        for d = #m, 1, -1 do
            local tm = m[d]
            if type(spec) == "table" then
                v = tm
                for i = 1, #spec do
                    v = v[spec[i]]
                    if not v then
                        break
                    end
                end
            else
                v = tm[spec]
            end
            if v then
                if type(v) ~= "table" then
                    v = { v }
                end
                return v
            end
        end
        return {}
    end

    local function render(v, separator)
        if not v then
            return
        end
        for i = 1, #v do
            if separator and i > 1 then
                separator()
            end
            writer(v[i])
        end
    end

]=]

local trailer = [=[

    return function(model, start)
        m = {model}
        output = {}
        chunks = {}

        if not start then
            start = 1
        end

        if not templates[start] then
            return nil
        end

        templates[start]()
        return table.concat(output)
    end
]=]

local function emit_template_function(src, name, statements)
    if type(name) == "number" then
        src[#src + 1] = string.format("[%d] = function()\n", name)
    else
        src[#src + 1] = string.format("[\"%s\"] = function()\n", name)
    end
    src[#src + 1] = "    local id, nv, v\n"
    for i = 1, #statements do
        src[#src + 1] = statements[i]
    end
    src[#src + 1] = "end,\n"
end

function m:compile(args)
    self.out_stack = {}
    self.templates = {}
    self.template_cache = {}
    local tree = nil
    if args.debug then
        tree = {}
    end

    self.delimiters = args.delimiters
    if args.templates then
        self.source_templates = args.templates
    end
    if args.template then
        self:begin_template()
        build(self, self:parse(args.template, args.delimiters))
        self:end_template()
    elseif args.start_template then
        -- KLUDGE ignore start_template arg for now
    end

    for name, tpl in pairs(self.source_templates) do
        self:begin_template(name)
        local parsed = self:parse(tpl, args.delimiters)
        build(self, parsed)
        if tree then
            tree[name] = parsed
        end
        self:end_template()
    end

    local src = {}
    src[#src + 1] = header

    src[#src + 1] = "local templates -- forward reference\n"
    src[#src + 1] = "templates = {\n"
    for name, tpl in pairs(self.templates) do
        emit_template_function(src, name, tpl)
    end
    src[#src + 1] = "} -- templates\n"

    src[#src + 1] = trailer

    local lua_source = table.concat(src)
    -- KLUDGE luac step here?
    local fn = assert(loadstring(lua_source))()
    return fn, lua_source, tree, self.templates
end

function m:find_file_templates(paths)
    local templates = {}
    for i = 1, #paths do
        local path = paths[i]
        for file in lfs.dir(path) do
            local tpl_name = string.match(file, "^(.*)%.st$")
            if tpl_name and not templates[tpl_name] then
                local f = io.open(path .. "/" .. file)
                templates[tpl_name] = f:read("*a")
                f:close()
            end
        end
    end
    return templates
end
