#!/usr/bin/lua

local ev = require 'ev'
local loop = ev.Loop.default
local dbus = require 'dbus'
local u = require 'unix'
local struct = require 'struct'
local class = require 'class'
local lxp = require 'lxp'
local lfs = require 'lfs'
local net = require 'net_noffi'
local log = require 'log'
local sqlite3 = require 'lsqlite3'
local sqlite = require 'sqlite_ffi'
local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
int inotify_init1(int flags);
int inotify_add_watch(int fd, const char *pathname, uint32_t mask);
int close(int fd);
int read(int fd, void *buf, int count);

enum {
    IN_CREATE = 0x100,
    IN_DELETE = 0x200,
    IN_MOVED_TO = 0x80,
    IN_CLOEXEC = 02000000,
};

struct inotify_event {
    int         wd;
    uint32_t    mask;
    uint32_t    cookie;
    uint32_t    len;
    char        name[];
};
]]

log:open('storage', log.USER)

local db = sqlite:new('/var/lib/armada/setup.db')

local SieMerge = class()
local SieMergeInput = class()

function SieMergeInput:init()
    local function cb(f)
        return function(p, ...) f(self, ...) end
    end
    self.xml_parser = lxp.new({
                                  StartElement = cb(self.xml_start_element),
                                  EndElement = cb(self.xml_end_element),
                                  CharacterData = cb(self.xml_character_data)
                              })
    self.stack = { { } }
    self.remap = {
        test = { },
        ch = { },
        decoder = { },
        group = { }
    }
end

function SieMergeInput:add_sie_block(block)
    local group = struct.unpack(">I4", block:sub(5, 9))
    -- print('got group', group)
    local payload = block:sub(13, -9)
    if group == 0 then
        -- print('parse', payload:len(), payload)
        self.xml_parser:parse(payload)
    else
        local remap = self.remap.group[tostring(group)]
        -- print('remap', remap)
        if remap then
            self.parent:output_block(remap, payload)
        end
    end
end

local tinsert, tremove, getn = table.insert, table.remove, table.getn

function SieMergeInput:xml_start_element(tag, attr)
    local newelement = {tag = tag, attr = attr}
    -- print('start', tag, getn(self.stack))
    tinsert(self.stack, newelement)
end

function SieMergeInput:xml_end_element(tag)
    local element = tremove(self.stack)
    assert(element.tag == tag)
    local level = getn(self.stack)
    -- print('end', tag, level)
    if level > 2 then
        tinsert(self.stack[level], element)
    else
        self:process_toplevel(element)
    end
end

function SieMergeInput:xml_character_data(text)
    local element = self.stack[getn(self.stack)]
    if element.tag ~= 'tag' then
        return
    end
    local n = getn(element)
    if type(element[n]) == "string" then
        element[n] = element[n] .. text
    else
        tinsert(element, text)
    end
end

local function wrap_path(el)
    local out = el
    if type(el) == 'table' then
        if el.attr.dim then
            out = { out; tag = 'dim', attr = { 'id'; id = el.attr.dim } }
            el.attr.dim = nil
        end
        if el.attr.ch then
            out = { out; tag = 'ch', attr = { 'id'; id = el.attr.ch } }
            el.attr.ch = nil
        end
        if el.attr.test then
            out = { out; tag = 'test', attr = { 'id'; id = el.attr.test } }
            el.attr.test = nil
        end
        local to_remove = { }
        for i,v in ipairs(el.attr) do
            if v == 'dim' or v == 'ch' or v == 'test' then
                table.insert(to_remove, i)
            end
        end
        repeat
            local i = table.remove(to_remove)
            if i then
                table.remove(el.attr, i)
            end
        until not i
        for i,v in ipairs(el) do
            el[i] = wrap_path(v)
        end
    end
    return out
end

function SieMergeInput:process_toplevel(el)
    el = wrap_path(el)
    if (el.tag == 'tag' and (el.attr.id == 'sie:xml_metadata' or
                             el.attr.id == 'sie:block_index')) or
       (el.tag == 'decoder' and (el.attr.id == '0' or el.attr.id == '1')) then
        return
    end
    self:process_element(el)
    self:remap_xml(el)

    local out = { }
    self:output_xml(el, function(text) table.insert(out, text) end)
    self.parent:output_xml(table.concat(out))
end

function SieMergeInput:process_element(el)
    if el.tag == 'tag' and el.attr.group then
        self:remap_id('group', el.attr.group)
    elseif el.tag == 'test' then
        self:remap_id('test', el.attr.id)
    elseif el.tag == 'ch' then
        self:remap_id('ch', el.attr.id)
        if el.attr.group then
            self:remap_id('group', el.attr.group)
        end
    elseif el.tag == 'dim' then
        if el.attr.group then
            self:remap_id('group', el.attr.group)
        end
    elseif el.tag == 'decoder' then
        self:remap_id('decoder', el.attr.id)
    end

    for i,v in ipairs(el) do
        if type(v) == 'table' then
            self:process_element(v)
        end
    end
end

local remappings = {
    ch = { base = 'ch', group = 'group', id = 'ch' },
    data = { decoder = 'decoder' },
    decoder = { id = 'decoder' },
    dim = { group = 'group' },
    tag = { decoder = 'decoder', group = 'group' },
    test = { id = 'test' },
    xform = { index_ch = 'ch' }
}

function SieMergeInput:remap_xml(el)
    local remap = remappings[el.tag]
    if remap then
        for i,v in pairs(el.attr) do
            local remap_as = remap[v]
            if remap_as then
                -- print('remap_xml', v, el.attr[v], '->',
                --       self.remap[remap_as][el.attr[v]])
                el.attr[v] = self.remap[remap_as][el.attr[v]]
            end
        end
    end

    for i,v in ipairs(el) do
        if type(v) == 'table' then
            self:remap_xml(v)
        end
    end
end

local function html_escape(str)
    return str:gsub('[&"<]', {
                        ['&'] = '&amp;',
                        ['"'] = '&quot;',
                        ['<'] = '&lt;'
                    })
end

function SieMergeInput:output_xml(el, outfunc, indent)
    if type(el) == 'table' then
        outfunc('<' .. el.tag)
        for i,v in ipairs(el.attr) do
            outfunc(' ' .. v .. '="' .. html_escape(el.attr[v]) .. '"')
        end
        if #el == 0 then
            outfunc('/>')
        else
            outfunc('>')
            for i,v in ipairs(el) do
                self:output_xml(v, outfunc)
            end
            outfunc('</' .. el.tag .. '>')
        end
    else
        outfunc(html_escape(el))
    end
end

function SieMergeInput:remap_id(type, from, to)
    -- print('remap_id', type, from, to)
    if not self.remap[type][from] then
        if not to then
            to = self.parent:next_id(type)
        end
        self.remap[type][from] = tostring(to)
    end
end

local alien = require 'alien'
local sie = alien.load('libsie.so.0')

sie.sie_context_new:types('pointer')
sie.sie_context_done:types('int', 'pointer')
sie.sie_release:types('void', 'pointer')
sie.sie_writer_new:types('pointer', 'pointer', 'callback', 'pointer')
sie.sie_writer_write_block:types('int', 'pointer', 'uint', 'string', 'ulong')
local function sie_writer_write_block(writer, group, string)
    return sie.sie_writer_write_block(writer, group, string, string:len())
end
sie.sie_writer_xml_header:types('int', 'pointer')
sie.sie_writer_xml_string:types('int', 'pointer', 'string', 'ulong')
local function sie_writer_xml_string(writer, string)
    return sie.sie_writer_xml_string(writer, string, string:len())
end
sie.sie_writer_flush_xml:types('int', 'pointer')
sie.sie_writer_next_id:types('uint', 'pointer', 'int')
local types = { group = 0, test = 1, ch = 2, decoder = 3 }
local function sie_writer_next_id(writer, type)
    return sie.sie_writer_next_id(writer, types[type])
end

local sie_ctx = sie.sie_context_new()

function SieMerge:init()
    self.file = assert(io.open(self.filename, 'w'))
    local function outfunc(user, data, size)
        local data = alien.tostring(data, size)
        assert(self.file:write(data))
        self.file:flush()
        return size
    end
    self.outcb = alien.callback(outfunc, 'ulong',
                                'pointer', 'pointer', 'ulong')
    self.writer = sie.sie_writer_new(sie_ctx, self.outcb, nil)
    sie.sie_writer_xml_header(self.writer)
end

function SieMerge:output_xml(xml)
    sie_writer_xml_string(self.writer, xml)
    sie.sie_writer_flush_xml(self.writer)
end

function SieMerge:output_block(group, payload)
    -- print('output_block', group, payload:len())
    sie_writer_write_block(self.writer, group, payload)
end

function SieMerge:close()
    sie.sie_release(self.writer)
    self.file:close()
end

function SieMerge:next_id(type)
    return sie_writer_next_id(self.writer, type)
end

function SieMerge:new_input()
    return SieMergeInput:new{parent = self}
end

-- local files = {...}
-- local outfile = table.remove(files, 1)

-- local merge = SieMerge:new{filename = outfile}
-- for i,file in ipairs(files) do
--     local f = assert(io.open(file))
--     local input = merge:new_input()
--     local buf = ''
--     repeat
--         local new = f:read(8192) or ''
--         local size
--         buf = buf .. new
--         repeat
--             if buf:len() >= 20 then
--                 size = struct.unpack('>I4', buf:sub(1, 4))
--                 if buf:len() >= size then
--                     input:add_sie_block(buf:sub(1, size))
--                     buf = buf:sub(size + 1)
--                 end
--             end
--         until buf:len() < 20 or buf:len() < size
--     until new == ''
-- end
-- merge:close()

-- _armada-storage      PTR ppc1._armada-storage
-- ppc1._armada-storage SRV 0 0 55124 ppc1
-- ppc1._armada-storage TXT "capacity: 404040404040"
-- ppc1._armada-storage TXT "free: 202020202020"

-- _armada-sie          PTR f1.ppc1._armada-sie
-- f1.ppc1._armada-sie  SRV 0 0 55124 ppc1
-- f1.ppc1._armada-sie  TXT "name: My Test Name.sie"
-- f1.ppc1._armada-sie  TXT "size: 10203040"
-- f1.ppc1._armada-sie  TXT "modified: 2010-10-10 10:10:10.1010"

local config = require 'system_config'

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

local network = assert(bus:auto_proxy_wait('com.somat.armada.Network',
                                           '/com/somat/armada/Network'))

local open_files = { }
local open_rfiles = { }

local ip = nil -- KLUDGE network:GetModuleIP()
local server = net.socket{ip = ip}:listen()
function server:acceptor(loop, sock)
    local sbuf = net.streambuf(sock)

    function sbuf:close_cb(loop, dir, source)
        -- print(self.socket, 'shutdown', dir, source)
        self:shutdown()
    end

    function sbuf:read_cb(loop)
        local nl = self.readbuf:find('\n')
        if nl then
            local cmdline = self:read(nl)
            -- print(self.socket, 'cmdline', cmdline)
            local cmd, name = cmdline:match("([^ ]+) (.*)\n")
            if cmd == 'OPEN' then
                open_files[name] = SieMerge:new{
                    filename = config.storage.path .. '/' .. name
                }
            elseif cmd == 'OPENR' then
                open_rfiles[name] = io.open(config.storage.path .. '/' .. name)
            elseif cmd == 'CLOSE' and open_rfiles[name] then
                open_rfiles[name]:close()
                open_rfiles[name] = nil
            elseif cmd == 'CLOSE' and open_files[name] then
                open_files[name]:close()
                open_files[name] = nil
            elseif cmd == 'DATA' and open_rfiles[name] then
                local got = open_rfiles[name]:read(16384)
                if got then
                    self.sbuf:write(got, #got)
                    return
                end
                open_rfiles[name]:close()
                open_rfiles[name] = nil
            elseif cmd == 'DATA' and open_files[name] then
                self.input = open_files[name]:new_input()
                self.read_cb = self.data_size_cb
                self.read_wm = 20
                return
            end
            self:shutdown()
        else
            self.read_wm = self.readbuf:len() + 1
        end
    end

    function sbuf:data_size_cb(loop)
        -- print(self.socket, 'size')
        self.read_wm = struct.unpack('>I4', self.readbuf:sub(1, 4))
        self.read_cb = self.data_payload_cb
    end

    function sbuf:data_payload_cb(loop)
        -- print(self.socket, 'payload', self.read_wm)
        self.input:add_sie_block(self:read(self.read_wm))
        self.read_wm = 20
        self.read_cb = self.data_size_cb
    end

    sbuf:start(loop)
end
server:start(loop)

local hostname = network:GetModuleHostname()
local storage_srvname = hostname .. '._cx23r-storage'
local files_srvname = hostname .. '._armada-files'
network:ModifyRRs({},
                  {{ '_cx23r-storage', 'PTR', storage_srvname },
                   { storage_srvname, 'SRV',
                     '0 0 ' .. server.port .. ' ' .. hostname }})

ev.Signal.new(function() print('SIGPIPE') end, u.SIGPIPE):start(loop) -- ignore SIGPIPE
ev.Signal.new(function()
                  print('SIGINT')
                  for k,v in pairs(open_files) do
                      print('closing', k)
                      v:close()
                  end
                  loop:unloop()
              end, u.SIGINT):start(loop)


local function sql_exec(db, query)
    local ok, result = pcall(function() return db:exec(query) end)
    if ok then
        return result, ok
    end
    return {ok=false, error=result}, ok
end

-- Function used to retrieve the "friendly" name of an SIE file
local function get_friendly_name(id)
    if not db then
        db = sqlite:new('/var/lib/armada/setup.db')
    end
    -- Get the friendly name for the file
    local query = string.format(
        [[SELECT content from "tsd_users" WHERE db_name == 'sie_name'
        AND deleted == '' AND id == '%s';]], id)
    local r, ok = sql_exec(db, query)
    if ok and r then
        for _, v in ipairs(r) do
            if v.content then
                local c = cjson.decode(v.content)
                if c.friendly_name then
                    return c.friendly_name .. '.sie'
                end
            end
        end
    end
end

local files = { }
local next_fnum = 1
local fs_stats = {}
local function timer_fn(loop, w)
    local delrr = {}
    local addrr = {}

    local old_fs_stats = fs_stats
    fs_stats = {}
    local stat = u.statvfs(config.storage.path)
    fs_stats.capacity = stat.blocks * stat.frsize
    fs_stats.available = stat.bavail * stat.frsize
    for k, v in pairs(fs_stats) do
        local ov = old_fs_stats[k]
        if ov ~= v then
            if ov ~= nil then
                delrr[#delrr + 1] = {storage_srvname, 'TXT', k .. ': ' .. ov}
            end
            addrr[#addrr + 1] = {storage_srvname, 'TXT', k .. ': ' .. v}
        end
    end

    local old_files = files
    files = { }
    for name in lfs.dir(config.storage.path) do
        if name:match('%.sie$') then
            files[name] = old_files[name]
            old_files[name] = nil
            local stat =
                lfs.attributes(config.storage.path .. '/' .. name)
            if not files[name] then
                files[name] = {
                    mod = 0,
                    srvname =
                        'f' .. next_fnum .. '.' .. files_srvname,
                    rr = { }
                }
                next_fnum = next_fnum + 1
            end
            local f = files[name]
            if stat and stat.modification ~= f.mod then
                f.mod = stat.modification
                local mod = os.date('!%Y-%m-%d %H:%M:%S', f.mod)
                for i = 1, #f.rr do
                    delrr[#delrr + 1] = f.rr[i]
                end
                -- Get the "friendly" name for the file
                local friendly_name =
                    get_friendly_name(name:match('^(.+)%.sie$'))
                if friendly_name then
                    f.rr = {
                        { '_armada-files', 'PTR', f.srvname },
                        { f.srvname, 'SRV',
                        '0 0 ' .. server.port .. ' ' .. hostname },
                        { f.srvname, 'TXT', 'name: ' .. friendly_name },
                        { f.srvname, 'TXT', 'size: ' .. stat.size },
                        { f.srvname, 'TXT', 'modified: ' .. mod },
                        { f.srvname, 'TXT', 'hash: ' ..
                            name:match('^(.+)%.sie$') },
                    }
                    for i = 1, #f.rr do
                        addrr[#addrr + 1] = f.rr[i]
                    end
                end
            end
        end
    end
    for k,f in pairs(old_files) do
        for i = 1, #f.rr do
            delrr[#delrr + 1] = f.rr[i]
        end
    end
    network:ModifyRRs(delrr, addrr)
end
ev.Timer.new(timer_fn, 0.1, 5):start(loop)

-- Monitor for SIE files showing up or going away from the storage directory
-- so changes will be noticed more quickly
local ifd = C.inotify_init1(C.IN_CLOEXEC)
if ifd > 0 then
    local fd = C.inotify_add_watch(ifd, config.storage.path,
        C.IN_CREATE + C.IN_DELETE + C.IN_MOVED_TO)
    if fd > 0 then
        local buf = ffi.new('char[8192]')
        local ine = ffi.cast('struct inotify_event *', buf)
        ev.IO.new(function(l, io, e)
                local len = C.read(ifd, buf, ffi.sizeof(buf))
                if len < 0 then
                    C.close(fd)
                    C.close(ifd)
                    io:stop(l)
                end
                local name_len = len - ffi.sizeof('struct inotify_event')
                if name_len then
                    if ffi.string(ine.name, name_len):find('%.sie') then
                        timer_fn(l)
                    end
                end
            end, ifd, ev.READ):start(loop)
    else
        C.close(ifd)
    end
end

loop:loop()
