#!/usr/bin/lua

local armada = require 'armada'
local ev = require 'ev'
local lpeg = require 'lpeg'
local ffi = require 'ffi'
local ts = require 'timestamps'
local channel_proxy = require 'channel_proxy'
local unix = require 'unix'

local config = require 'system_config'

local category = 'unknown'
if config.target == 'armada' then category = 'source/hbm_cx23r/msg_status' end
if config.target == 'edaq' then category = 'source/somat_edaq_msg_status' end

local dd = ffi.new('double[1]')

local systemDev = class('systemDev', channel_proxy)

local x100000000 = math.pow(2, 32)
local string_gsub = string.gsub
local string_match = string.match
local string_sub = string.sub

local      P,      R,      S,      V
    = lpeg.P, lpeg.R, lpeg.S, lpeg.V
local      C,      Carg,      Cb,      Cc,      Cf,      Cg,      Cmt,      Ct
    = lpeg.C, lpeg.Carg, lpeg.Cb, lpeg.Cc, lpeg.Cf, lpeg.Cg, lpeg.Cmt, lpeg.Ct

local function maybe(p) return p^-1 end

local whitespace = S(" \t")
local ws = whitespace^1
local lf = P"\n"
local integer = R("09")^1
local dot = '.'
local number = integer * maybe(dot*integer)

local loadavg_parser = P{
    'lines'; -- Start rule
    lines = Ct((
        V"loadavg_line"
        )^0),
    loadavg_line = Cg(number, "one_minute_loadavg")
        * ws * Cg(number, "loadavg")
        * ws * Cg(number, "fifteen_minute_loadavg")
        * (1 - lf)^0 * lf,
}

local stat_parser = P{
    'lines'; -- Start rule
    lines = Ct((
        V"cpu_line"
      + V"other_line"
        )^0),
    cpu_line = "cpu"
        * ws * Cg(integer, "cpu_user")
        * ws * Cg(integer, "cpu_niced")
        * ws * Cg(integer, "cpu_system")
        * ws * Cg(integer, "cpu_idle")
        * ws * Cg(integer, "cpu_iowait")
        * ws * Cg(integer, "cpu_irq")
        * ws * Cg(integer, "cpu_softirq")
        * (1 - lf)^0 * lf,
    other_line = (1 - lf)^0 * lf,
}

local meminfo_parser = P{
    'lines';
    lines = Ct((
        V"MemTotal_line"
      + V"MemFree_line"
      + V"Buffers_line"
      + V"Cached_line"
      + V"other_line"
        )^0),
    MemTotal_line = "MemTotal:"
        * ws * Cg(integer, "memory_used")
        * (1 - lf)^0 * lf,
    MemFree_line = "MemFree:"
        * ws * Cg(integer, "memory_free")
        * (1 - lf)^0 * lf,
    Buffers_line = "Buffers:"
        * ws * Cg(integer, "memory_buffered")
        * (1 - lf)^0 * lf,
    Cached_line = "Cached:"
        * ws * Cg(integer, "memory_cached")
        * (1 - lf)^0 * lf,
    other_line = (1 - lf)^0 * lf,
}

local chan_info = {
    [1] = {
        name = 'memory_used',
        description = 'Used Memory',
        electrical_units = 'Kb',
    },
    [2] = {
        name = 'memory_free',
        description = 'Free Memory',
        electrical_units = 'Kb',
    },
    [3] = {
        name = 'memory_buffered',
        description = 'Buffered Memory',
        electrical_units = 'Kb',
    },
    [4] = {
        name = 'memory_cached',
        description = 'Cached Memory',
        electrical_units = 'Kb',
    },
    [5] = {
        name = 'cpu_user',
        description = 'CPU USer Time',
        electrical_units = '%',
    },
    [6] = {
        name = 'cpu_system',
        description = 'CPU System Time',
        electrical_units = '%',
    },
    [7] = {
        name = 'cpu_idle',
        description = 'CPU Idle Time',
        electrical_units = '%',
    },
    [8] = {
        name = 'cpu_iowait',
        description = 'CPU I/O Wait Time',
        electrical_units = '%',
    },
    [9] = {
        name = 'cpu_irq',
        description = 'CPU IRQ Time',
        electrical_units = '%',
    },
    [10] = {
        name = 'cpu_softirq',
        description = 'CPU Soft IRQ Time',
        electrical_units = '%',
    },
    [11] = {
        name = 'system_clock',
        description = 'System Real-Time Clock',
        electrical_units = 's',
    },
    [12] = {
        name = 'timestamp',
        description = 'Sample Clock Timestamp',
        electrical_units = 'ticks',
    },
    [13] = {
        name = 'loadavg',
        description = 'CPU Load Average',
        electrical_units = '',
    },
    [14] = {
        name = 'control',
        description = 'System Control Channel',
        electrical_units = '',
    },
}


local function collect_from_proc(loop, timer)
    local self = timer.server
    local fh

    fh = io.open("/proc/uptime")
    local uptime_file = fh:read("*a")
    fh:close()

    fh = io.open("/proc/stat")
    local stat_file = fh:read("*a")
    fh:close()

    fh = io.open("/proc/meminfo")
    local meminfo_file = fh:read("*a")
    fh:close()

    fh = io.open("/proc/loadavg")
    local loadavg_file = fh:read("*a")
    fh:close()

    local stat, end_pos = stat_parser:match(stat_file)
    local mstat, _ = meminfo_parser:match(meminfo_file)
    local ldavg, _ = loadavg_parser:match(loadavg_file)

    local o = {}

    if self.old_stat then
        local old_stat = self.old_stat
        local total = 0
        for k, v in pairs(stat) do
            local diff = v - old_stat[k]
            o[k] = diff
            total = total + diff
        end
        for k, v in pairs(o) do
            o[k] = 100.0 * v / total
        end
        for k, v in pairs(mstat) do
            o[k] = v
        end
        if o['memory_used'] and o['memory_free'] then
            o['memory_used'] = o['memory_used'] - o['memory_free']
        end
        for k, v in pairs(ldavg) do
            o[k] = v
        end
    end
    self.old_stat = stat

    local secs, nsecs = unix.clock_gettime()
    local timestamp = ts.ts_to_stamp(ts.get_ts())
    if not timestamp then return end

    o.system_clock = secs + nsecs * 1e-9
    o.timestamp = timestamp

    for _, chan in ipairs(self.channels) do
        local d = o[chan.name]
        if d ~= nil then
            d = tonumber(d)
            for _, sink in pairs(chan.comet_sinks) do
                sink(timestamp, chan.channel_name, d)
            end
            if chan.channel_sink then
                dd[0] = d
                chan.channel_sink(timestamp - chan.start_stamp, dd, 8)
            end
        end
    end
end

function systemDev:initialize()
    self.comet_data = {}
    self.external_names = {}

    self.channels = chan_info
    self.proxy_type = 'systemDev'
    self.rate = self.rate or 5
    channel_proxy.initialize(self, 'system')

    for _, chan in ipairs(chan_info) do
        chan.comet_sinks = {}
        chan.channel_sink = nil
        chan.channel_name = chan.name .. '.status@' .. self.proxy_hostname
        chan.timestamp = 0
        chan.connector = chan.channel_name
        chan.data_type = 'message_float64'
        chan.category = category
    end

    self.interval = 1.0 / self.rate
    self.timer = ev.Timer.new(collect_from_proc, self.interval, self.interval)
    self.timer.server = self
    self.timer:start(ev.Loop.default)
end

function systemDev:available_channels(hostname)
    for _, ci in ipairs(chan_info) do
        ci.connector = ci.name .. ".status@" .. self.proxy_hostname
        if ci.name == 'control' then
            ci.data_type = 'message_raw'
        else
            ci.data_type = 'message_float64'
        end
        ci.message_rate = self.rate
        ci.is_logical = false
        self.external_names[ci.name] = ci.connector
    end
    return chan_info
end

function systemDev:start_comet_stream(w, chans, sink)
--print(string.format("CW: %d", w.socket))
    for _, chan in ipairs(chans) do
        for _, v in ipairs(self.channels) do
            if chan == v.channel_name then
                v.comet_sinks[w.socket] = sink
                w.cleanup = function()
                    for _, v in ipairs(self.channels) do
                        v.comet_sinks[w.socket] = nil
                    end
                end
                break
            end
        end
    end
end

function systemDev:start_channel_stream(w, spec, md_sink, sink)
--print(string.format("CSW: %d", w.socket))
    for _, chan in ipairs(self.channels) do
        if chan.connector == spec.connector then
            chan.start_stamp = ts.ts_to_stamp(ts.iso_to_ts(spec.start_time))
            chan.channel_sink = sink
            w.cleanup = function()
                chan.channel_sink = nil
            end
            md_sink(spec)
            return
        end
    end
end

function systemDev:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            msg = string.format('Control request not specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    if not spec.connector then
        return {
            msg = string.format('Control with no connector specified on %s',
                                self.proxy_hostname),
            result = 'error',
        }
    end
    for _, c in ipairs(self.channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    msg = string.format("Control request %s invalid on %s",
                                        spec.request, c.connector),
                    result = 'error',
                }
            end
            return self[spec.request](self, spec.params)
        end
    end
    return {
        msg = string.format("Control connector %s not found",
                            spec.connector),
        result = 'error',
    }
end

--[[
function systemDev:command(params)
    if not params[1] then
        return {
            msg = string.format("system command not specified for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end

    if not params[2] then
        return {
            msg = string.format("system command mode not specified for %s",
                                self.proxy_hostname),
            result = 'error',
        }
    end

    local mode = params[2]

    if mode ~= 'r' and mode ~= 'w' and mode ~= 'b' and mode ~= 't' then
        return {
            msg = string.format("system command mode \'%s\' not valid for %s",
                                mode, self.proxy_hostname),
            result = 'error',
        }
    end

    local file = io.popen(params[1], mode)
    local output = file:read('*all')
    file:close()

    return {
        msg = string.format("%s:%s", self.proxy_hostname, output),
        result = 'success',
    }
end
]]

systemDev:new()

ev.Loop.default:loop()
