local ffi = require 'ffi'
local sample_types = require 'sample_types'

local m = {}

function m:template_init(v)
    self.template_indent = ''
    self.template_indent_stack = {}
end

function m:indent(v)
    local old = self.template_indent
    if v then
        assert(type(v) == 'string')
        self.template_indent = v
    end
    return old
end

function m:push_indent(v)
    -- v = '    ' -- absolute indent value
    -- v = number, add or subract spaces from current value
    table.insert(self.template_indent_stack, self.template_indent)
    if type(v) == 'number' then
        v = string.rep(' ', #self.template_indent + v)
    end
    self.template_indent = v
    return v
end

function m:pop_indent(v)
    local old_v = self.template_indent
    self.template_indent = assert(table.remove(self.template_indent_stack))
    return old_v
end

local function lookup(var, values_list)
    local path = {}
    for v in var:gmatch('([^%.]+)%.?') do
        path[#path + 1] = v
    end
    for _, values in ipairs(values_list) do
        local v = values
        for i, part in ipairs(path) do
            assert(type(v) == 'table')
            part = tonumber(part) or part
            v = v[part]
            if v == nil then
                break
            end
            if type(v) ~= 'table' and i == #path then
                return tostring(v)
            end
        end
    end
    return nil
end

function m:template(template, ...)
    local values_list = {...}
    local settings = {}
    if settings.inline then
        settings.indent = ''
    end
    if type(template) == 'table' then
        settings = template
        template = table.remove(values_list, 1)
        if settings.indent then
            settings.old_indent = self:indent(settings.indent)
        end
    end

    local out = template

    -- First, deal with |?var and |!var conditionals
    out = out:gsub("%|([%?%!])([%w_%.]+)(.-\n[ ]*)",
        function(op, var, line)
            local spc = string.rep(' ', #var + 1)
            local v = lookup(var, values_list)
            if op == '!' then
                v = not v
            end
            return v and '|' .. spc .. line or ''
        end)

    local old_out
    repeat
        old_out = out
        out = out:gsub("%(=([%w_%.']+)=%)",
            function(match)
                local dt = match:match("^sizeof'(.-)'$")
                if dt then
                    local size = assert(ffi.sizeof(dt))
                    return string.format('%s --[[sizeof(%s)]]', size, dt)
                end
                local v = lookup(match, values_list)
                if v ~= nil then
                    return v
                end
                error(string.format("template error, no value for '%s'", match))
            end)
    until out == old_out

    out = out:gsub('^[ ]*%|', self.template_indent)
    out = out:gsub('\n[ ]*%|%+[ ]*', ' ')
    out = out:gsub('%s*%|%*[ ]*\n', '\n')
    out = out:gsub('\n[ ]*%|', '\n' .. self.template_indent)
    out = out:gsub('[ ]+$', '')
    if settings.inline then
        out = out:gsub('^%s*(.-)%s*$', '%1')
    end
    if settings.old_indent then
        self:indent(settings.old_indent)
    end
    if settings.output == false then
        return out
    end
    local dest = settings.output or self.buffers._default or 'out'
    self.buffers[dest] = self.buffers[dest] or {}
    self.buffers[dest][#self.buffers[dest] + 1] = out
end

function m:tmpl(template, ...)
    return self:template({inline = true, output = false}, template, ...)
end

function m:default_buffer(buf)
    self.buffers._default = buf
end

function m:render_buffer(buf_name)
    local buf = self.buffers[buf_name or self.buffers._default] or {}
    local nk, sk = {}, {}
    for k, _ in pairs(buf) do
        if tonumber(k) then
            nk[#nk + 1] = k
        else
            sk[#sk + 1] = k
        end
    end

    table.sort(nk)
    table.sort(sk)

    local o = {}
    for i = 1, #nk do
        o[#o + 1] = buf[nk[i]]
    end
    for i = 1, #sk do
        o[#o + 1] = buf[sk[i]]
    end
    return table.concat(o)
end

function m:clear_buffer(buf)
    local buf = buf or self.buffers._default
    self.buffers[buf] = {}
end

function m:twr(fmt, ...)
    local dest = self.buffers._default
    self.buffers[dest] = self.buffers[dest] or {}
    self.buffers[dest][#self.buffers[dest] + 1] =
        string.format(string.gsub(fmt, '%%', '%%%%'), ...)
end

function m:twri(fmt, ...)
    local dest = self.buffers._default
    self.buffers[dest] = self.buffers[dest] or {}
    self.buffers[dest][#self.buffers[dest] + 1] =
        self.template_indent ..  string.format(fmt, ...)
end

local function emit_lua(self, v, depth)
    depth = (depth or 0) + 1
    local t = type(v)
    if t == 'nil' or t == 'number' then
        self:twr(tostring(v))
    elseif t == 'string' then
        local fv = v:gsub("(['%z-\x1F\x7F-\xFF])",
            function(x) return string.format('\\x%.02x', string.byte(x)) end)
        self:twr("'" .. fv .. "'")
    elseif t == 'table' then
        local nk, sk = {}, {}
        for k, v in pairs(v) do
            if tonumber(k) then
                nk[#nk + 1] = k
            else
                sk[#sk + 1] = k
            end
        end
        table.sort(nk)
        table.sort(sk)
        local ordered = #sk == 0 and nk[1] == 1 and nk[#nk] == #nk
        if ordered then
            local i = 1
            while ordered and i <= #nk do
                ordered = nk[i] == i
                i = i + 1
            end
        end
        self:twr('{\n')
        self:push_indent(4)
        if ordered then
            -- {v1, v2, v3 ... vn}
            for k = 1, #nk do
                self:twri('')
                emit_lua(self, v[k], depth)
                self:twr(',\n')
            end
        else
            for i = 1, #nk do
                self:twri('[%s] = ', nk[i])
                emit_lua(self, v[nk[i]], depth)
                self:twr(',\n')
            end
            for i = 1, #sk do
                local k = sk[i]
                if k:match('^[%a_][%w_]*$') then
                    self:twri('%s = ', k)
                else
                    self:twri('[')
                    emit_lua(self, k, depth)
                    self:twr('] = ')
                end
                emit_lua(self, v[k], depth)
                self:twr(',\n')
            end
        end
        self:pop_indent()
        self:twri('}')
        if depth == 1 then
            self:twr('\n')
        end
    else
        error("render_lua() can't render " .. t)
    end
end

function m:render_lua(v)
    local old_buf = self.buffers._default
    local tb = 'ELBUF'
    self:default_buffer(tb)
    emit_lua(self, v, 0)
    local out = self:render_buffer()
    self:clear_buffer()
    self:default_buffer(old_buf)
    return out
end

return m
