require 'strict'

-- interactive_trigger.lua - Implements the Interactive Trigger
-- computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local sample_types = require 'sample_types'
local rational = require 'rational'
local frame = require 'test_engine/frame'
local netdba = require 'test_engine/network_db_access'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_fill = ffi.fill
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('tot_count=%lld, scale_factor=%lld',
        self.tot_count, self.scale_factor)
end

-- Performs any needed post instantiation initialization.
-- Parameters:
--     SIE writer - NOT USED by this module!
--     restarted - true when the test has been restarted
--         due to a reboot - don't want to reset the trigger
--         value at startup in that case
-- Return values:  true
function mt:setup(sie_writer, restarted)
    netdba:initialize()
    netdba:monitor_rr(self.trigger_name)
    netdba:bind_changed_rrs()
    netdba:update_monitored_rr_values(self.trigger_name)

    if not restarted then
        -- "reset" interactive trigger in database
        netdba:change_rr(self.trigger_name,
            self.trigger_name .. ":" .. "false")
    end

    return true
end

-- Get current trigger status from network database.
-- Returns:  true if trigger is "on", false otherwise
function mt:query_trigger()
    local triggered =false 
    local rec = netdba:get_rr(self.trigger_name)
    local trig_val = false
    if rec then
        trig_val = rec[3]:match("^.-:%s*(.*)$")
        if self.invert_trigger then
            -- invert sense of trigger (i.e., "true" really means "false")
            triggered = trig_val == 'false' and true or false
        else
            triggered = trig_val == 'true' and true or false
        end
    else
        -- Take no record for the specified trigger to mean a trigger
        -- condition of "false".  No record existing could be normal,
        -- meaning trigger hasn't been used.
        if self.invert_trigger then
            -- invert sense of trigger (i.e., "true" really means "false")
            triggered = true
        else
            triggered = false
        end
    end
    return triggered
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    local work_to_do, thrown_away =
        utils.there_is_work_to_do(self.in_pipe, self.pipe_list)
    if not work_to_do then
        self.tot_count = self.tot_count + thrown_away
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local of        -- will reference created output frame
    local out_vals  -- will reference "pointer" to output frame's data
    local _         -- we don't need 'in_vals' "pointer"
    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        -- create frame for output
        of, _, out_vals = utils.create_output_frame(f.maxsize, f,
            self.tot_count, self.out_frame_sample_size)
        -- set output values
        if self:query_trigger() then
            ffi_fill(out_vals, f.maxsize * self.out_frame_sample_size, 1)
            for i = 0, f.count - 1 do
                out_vals[i] = 1
            end
        else
            ffi_fill(out_vals, f.maxsize * self.out_frame_sample_size)
        end

        -- finish with frame
        of.count = f.count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        utils.put_frame(of, self.pipe_list)
        -- delete input frame, get next
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local interactive_trigger = setmetatable({}, {__index = mt})
    interactive_trigger.sie_collect_idx = first_sie_index
    utils.set_common_meta(interactive_trigger, stanza)
    -- validate needed metadata, and create input/output pipe
    -- structures
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or
        stanza.summary.input_channels or {}
    assert(#input_channels == 1, "Interactive Trigger computed " ..
        "channel only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    interactive_trigger.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(interactive_trigger.in_pipe, "failed to create input pipe")
    interactive_trigger.pipe_list = Pipe_new_pipe_list()

    -- get remaining needed metadata
    --     output frame element size
    assert(stanza.data_type, "Missing data type")
    interactive_trigger.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(interactive_trigger.out_frame_sample_size and
           interactive_trigger.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(interactive_trigger.out_frame_sample_size)))
    --     invert trigger
    interactive_trigger.invert_trigger = stanza.invert_trigger
    assert(not interactive_trigger.invert_trigger or
        (tostring(interactive_trigger.invert_trigger) == 'false' or
         tostring(interactive_trigger.invert_trigger) == 'true'),
        "Invalid value for attribute 'invert_trigger' (" ..
        tostring(stanza.invert_trigger) .. ")")
    interactive_trigger.invert_trigger =
        tostring(interactive_trigger.invert_trigger) == "true" or false
    --     trigger index
    local trig_index = stanza.trigger_index
    assert(trig_index,
        "Value for attribute 'trigger_index' does not exist")
    interactive_trigger.trigger_name = '_interactive-trigger-' ..
        tostring(trig_index)

    -- used for frames' time stamps (just a sample count for
    -- this module)
    interactive_trigger.tot_count = 0

    return interactive_trigger, 1
end

return m
