require 'strict'

-- smooth.lua - Implements the Smoothing Filter computed channel.
--
-- The filter_length is the number of samples to be averaged,
-- amnd must always be an odd number of samples between 3 and 201.

local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'
local log = require 'log'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame


local mt = {}

function mt:dump()
    dbg('filter length=%d', self.filter_length)
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    if Utils_put_frame(self.of, self.pipe_list) then
        self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
        self.out_vals = ffi_cast(Utils_frame_data_type(
            self.out_frame_sample_size), self.of.buf)
    end
    self.of.hdr.ts = self.tot_count
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    local taps = self.filter_length        -- the filter length
    local new_tap = 0                      -- newest value
    local old_tap = 0                      -- oldest value in cbuf
    local f = Pipe_get_frame(self.in_pipe) -- input frame
    -- will reference "pointer" to input frame's data
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
    local inp_index = 0                    -- Index to input frame???
    while f ~= nil do
        while not self.cbuffull and inp_index < f.count do
            new_tap = in_vals[inp_index]
            inp_index = inp_index + 1
            self.cbuf_sum = self.cbuf_sum + new_tap
            self.cbuf[self.cbuf_index] = new_tap
            self.cbuf_index = self.cbuf_index + 1
            --backfill with initial average
            if self.cbuf_index == taps then
                self.cbuf_index = 0
                self.cbuffull = true
                for i = 0,(taps-1)/2 do
                    self.out_vals[self.out_index] = self.cbuf_sum/taps
                    self.out_index = self.out_index + 1
                    if self.out_index == self.bufsize then
                        self:flush_output_frame()
                    end
                end
            end
        end

        while self.cbuffull and inp_index < f.count do
            new_tap = in_vals[inp_index]
            inp_index = inp_index + 1
            old_tap = self.cbuf[self.cbuf_index]
            self.cbuf[self.cbuf_index] = new_tap
            self.cbuf_index = self.cbuf_index + 1
            if self.cbuf_index == taps then
                self.cbuf_index = 0
            end
            self.cbuf_sum = self.cbuf_sum + new_tap - old_tap
            self.out_vals[self.out_index] = self.cbuf_sum/taps
            self.out_index  = self.out_index + 1
            if self.out_index == self.bufsize then
                self:flush_output_frame()
            end
        end
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
        if f ~= nil then
            in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)
            inp_index = 0
        end
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local smooth = setmetatable({}, {__index = mt})
    smooth.sie_collect_idx = first_sie_index
    utils.set_common_meta(smooth, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Smoothing filter computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    smooth.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(smooth.pipe_rate and smooth.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(smooth.pipe_rate))
    --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    smooth.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(smooth.sample_rate and smooth.sample_rate > 0,
        "Invalid sample rate: " .. tostring(smooth.sample_rate))
    --smooth.out_frame_sample_size = ffi_sizeof('double *')
    smooth.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    --   time_base_shifter 'filter_length'
    smooth.filter_length = tonumber(stanza.filter_length)
    assert(smooth.filter_length and
        smooth.filter_length >= 3 and
        smooth.filter_length <= 201 and
        smooth.filter_length % 2 == 1,
        "INVALID filter length value: " ..
        tostring(stanza.filter_length))

    -- create input/output pipe structures
    smooth.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(smooth.in_pipe, "failed to create input pipe")
    smooth.pipe_list = Pipe_new_pipe_list()
    assert(smooth.pipe_list, "failed to create (output) pipe list")
    
    --local variables needed not set from stanza.
    smooth.tot_count = 0
    smooth.bufsize = frame.bufsize(smooth.sample_rate, smooth.pipe_rate)
    -- output frame
    smooth.of = Frame_create(smooth.bufsize, smooth.out_frame_sample_size)
    -- "pointer" to output frame data
    smooth.out_vals = ffi_cast(Utils_frame_data_type(
        smooth.out_frame_sample_size), smooth.of.buf)
    -- circular buffer to compute average
    smooth.cbuf = ffi.new("double["..smooth.filter_length.."]", {})
    smooth.cbuffull = false -- keep track whether cbuf is full
    smooth.cbuf_sum = 0     -- the current sum of cbuf / taps
    smooth.cbuf_index = 0   -- current index in cbuf
    smooth.out_index  = 0   -- current index in output frame of

    return smooth, 1
end

return m
