require 'strict'

-- time_base_shifter.lua - Implements the Time Base Shifter
-- computed channel.
--
-- The output time can be shifted to "lead" or to "lag".
-- "lead" means that input samples will be discarded to
-- make the output time zero appear shifted to the left
-- (before) with respect to the input time zero.
-- "lag" means that the first input sample will be repeated
-- to make the output time zero appear shifted to the right
-- (after) with respect to the input time zero.

local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local sample_clock = require 'sample_clock'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list


local mt = {}

function mt:dump()
    dbg('shift count=%lld, shift direction=%s',
        self.sxr_shift_count, tostring(self.shift_direction))
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Puts input frame's data into output frame.  The output frame is
--     assumed to have be partially filled, so this function fills
--     the remainder of the output frame and puts the remaining
--     input data into a newly created partial output frame.
-- Parameter:  'f' is an input frame
-- Returns:  
function mt:put_in_partial(f)
    -- Make sure we have an output frame to use.
    if not self.partial_of then
        self.partial_of =
            utils.create_output_frame(f.maxsize, f, self.ts,
                self.out_frame_sample_size)
    end

    -- set up data buf "pointers"
    local sample_type = utils.frame_data_type(f.element_size)
    local in_vals = ffi_cast(sample_type, f.buf)
    local out_vals = ffi_cast(sample_type, self.partial_of.buf)

    -- fill the remainder of the partial output frame
    local in_buf_index = 0
    for i = self.partial_of.count, f.count - 1 do
        out_vals[i] = in_vals[in_buf_index]
        in_buf_index = in_buf_index + 1
    end
    self.partial_of.count = f.count

    -- put filled output frame in pipe(s)
    utils.put_frame(self.partial_of, self.pipe_list)
    self.ts = self.ts + self.partial_of.count
    self.partial_of, in_vals, out_vals =
            utils.create_output_frame(f.maxsize, f, self.ts,
                self.out_frame_sample_size)

    -- put remaining input samples in new partial output buf
    local out_buf_index = 0
    for i = in_buf_index, f.count - 1 do
        out_vals[out_buf_index] = in_vals[i]
        out_buf_index = out_buf_index + 1
    end
    self.partial_of.count = out_buf_index
end

-- Processes incoming frames for "lead" direction.
-- It writes full outgoing frames, and saves partially filled
-- outgoing frames for later filling.
function mt:write_lead(f)
    -- ignore input samples until the shift count is exceeded
    if self.tot_input_count < self.sxr_shift_count then
        -- Need to shift, ignoring input samples until the
        -- shift count is exceeded.
        -- Make sure we have an output frame to use.
        if not self.partial_of then
            self.partial_of =
                utils.create_output_frame(f.maxsize, f, self.ts,
                    self.out_frame_sample_size)
        end

        if (self.tot_input_count + f.count) > self.sxr_shift_count then
            -- use some of the input in this frame
            local sample_type = utils.frame_data_type(f.element_size)
            local in_vals = ffi_cast(sample_type, f.buf)
            local out_vals = ffi_cast(sample_type, self.partial_of.buf)
            local in_buf_start = self.sxr_shift_count - self.tot_input_count
            local out_buf_index = 0
            for i = in_buf_start, f.count - 1 do
                out_vals[out_buf_index] = in_vals[i]
                out_buf_index = out_buf_index + 1
            end
            self.partial_of.count = out_buf_index
        --  else use none of the input in this frame
        end
    else
        -- process input samples normally (the shift count is exceeded)
        self:put_in_partial(f)
    end
    self.tot_input_count = self.tot_input_count + f.count
end

-- Processes incoming frames for "lag" direction.
-- It writes full outgoing frames, and saves partially filled
-- outgoing frames for later filling.
function mt:write_lag(f)
    if f.count <= 0 then
        return  -- no input to process (is this an error?)
    end

    if self.tot_out_count < self.sxr_shift_count then
        -- Need to shift.
        -- Make sure we have an output frame to use.
        if not self.partial_of then
            self.partial_of =
                utils.create_output_frame(f.maxsize, f, self.ts,
                    self.out_frame_sample_size)
        end
        -- set up data buf "pointers"
        local sample_type = utils.frame_data_type(f.element_size)
        local in_vals = ffi_cast(sample_type, f.buf)
        local out_vals = ffi_cast(sample_type, self.partial_of.buf)

        -- repeat the first input sample to output "shift
        -- count" times (the regular output of input buf
        -- will take care of the "+ 1" write of first input sample
        local first_out = in_vals[0]
        local out_count
        while self.tot_out_count < self.sxr_shift_count do
            -- fill partial output frame, as much as possible
            out_count = math_min(f.count,
                self.sxr_shift_count - self.tot_out_count)
            for i = 0, out_count - 1 do
                out_vals[i] = first_out
            end
            self.partial_of.count = out_count
            self.tot_out_count = self.tot_out_count + out_count
            if self.partial_of.count >= f.count then
                -- put output frame in output pipe(s)
                utils.put_frame(self.partial_of, self.pipe_list)
                self.ts = self.ts + self.partial_of.count
                self.partial_of, in_vals, out_vals =
                    utils.create_output_frame(f.maxsize, f, self.ts,
                        self.out_frame_sample_size)
            end
        end
    end

    self:put_in_partial(f)
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    local work_to_do, thrown_away =
        utils.there_is_work_to_do(self.in_pipe, self.pipe_list)
    if not work_to_do then
        self.ts = self.ts + thrown_away
        return
    end

    local f = Pipe_get_frame(self.in_pipe)

    -- Process each frame in input pipe.  There is only one input pipe.
    while f ~= nil do
        if self.shift_direction == "lead" then
            self:write_lead(f)
        else
            self:write_lag(f)
        end
        -- delete input frame, get next one
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local time_base_shifter = setmetatable({}, {__index = mt})
    time_base_shifter.sie_collect_idx = first_sie_index
    utils.set_common_meta(time_base_shifter, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Time Base Shifter computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    --   output frame element size
    assert(stanza.data_type, "Missing data type")
    time_base_shifter.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(time_base_shifter.out_frame_sample_size and
           time_base_shifter.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(time_base_shifter.out_frame_sample_size)))
    --   time_base_shifter 'shift_count'
    time_base_shifter.sxr_shift_count = tonumber(stanza.shift_count)
    assert(time_base_shifter.sxr_shift_count and
        time_base_shifter.sxr_shift_count > 0 and
        time_base_shifter.sxr_shift_count <= 10000,
        "INVALID shift value: " ..
        tostring(stanza.shift_count))
    --   time_base_shifter 'shift_direction'
    time_base_shifter.shift_direction = stanza.shift_direction
    assert(time_base_shifter.shift_direction and
        (time_base_shifter.shift_direction == "lead" or
        time_base_shifter.shift_direction == "lag"),
        "INVALID shift direction: " ..
        tostring(stanza.shift_direction))

    -- create input/output pipe structures
    time_base_shifter.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(time_base_shifter.in_pipe, "failed to create input pipe")
    time_base_shifter.pipe_list = Pipe_new_pipe_list()
    assert(time_base_shifter.pipe_list, "failed to create (output) pipe list")

    time_base_shifter.nbr_samples_shifted = 0

    -- used for "lead" direction to determine how many input samples
    -- to discard
    time_base_shifter.tot_input_count = 0
    -- used for "lag" direction to determine when to stop padding
    -- with the first input sample value
    time_base_shifter.tot_out_count = 0

    -- used for frames' time stamps (just a sample count for
    -- this module)
    time_base_shifter.ts = 0

    return time_base_shifter, 1
end

return m
