require 'strict'

-- burst_history.lua - Implements the Burst History data mode.
--
-- Implementation note:
--     This data mode uses its own particular trigger mechanism,
--     not the "generic" datamode triggering mechanism.

local cjson = require 'cjson'
local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sample_types = require 'sample_types'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local C = ffi.C
local ffi_cast = ffi.cast
local ffi_cdef = ffi.cdef
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local math_min = math.min

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_add_frame     = pipe.add_frame
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame

ffi.cdef[[
    void *calloc(size_t nmemb, size_t size);
    void free(void *ptr);
]]

local circular_buf_t_str = "burst_%s_circular_buf_t"
local circular_buf_def_str = [[
    typedef struct burst_%s_circular_buf_struct {
        %s  *buf;
    } %s; ]]
local circ_cdefined = {}

local BH_DATAMODE = 0x42


local mt = {}

function mt:dump()
    --dbg('tot_count=%lld', self.tot_count)
end

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Implements "gate" trigger, returning true whenever the input value
-- is non-zero.
-- Parameter:
--     'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is non-zero, false otherwise
function mt:gate_trigger(trig_data)
    return trig_data ~= 0
end

-- Implements "edge" trigger, returning true whenever the input value
-- is non-zero and the previous trigger channel value was zero.
-- Parameter:
--     'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is transitions from zero to non-zero,
--     false otherwise
function mt:edge_trigger(trig_data)
    if tonumber(trig_data) == 0 then
        self.prev_trig_on = false
        return false
    else  -- trig_data ~= 0  
        if self.prev_trig_on then
            return false
        else
            self.prev_trig_on = true
            return true
        end
    end
end

-- Creates and initializes a circular buffer for each input channel
-- for use in storing pre-trigger data.
function mt:init_circ_bufs()
    -- create circular buffers for storing pre-trigger input data
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        local data_type = info.stanza.data_type
        assert(data_type, "Missing data type for " ..
            tostring(info.name))
        local cbuf_type = string.format(circular_buf_t_str, data_type)
        if not circ_cdefined[cbuf_type] then
            ffi_cdef(string.format(circular_buf_def_str, data_type,
                data_type, cbuf_type))
            circ_cdefined[cbuf_type] = 1
        end
        info.circular_buf = ffi_new(cbuf_type)
        -- need room for pre_count number of incoming data values
        -- in circular_buf.buf
        info.circular_buf.buf =
            C.calloc(self.pre_count, ffi_sizeof(data_type))
    end
    -- use same front and back indices for all circular buffers,
    -- since all input channels will be basing their bursts off
    -- of the same trigger channel
    self.circ_buf_front = 0
    self.circ_buf_back = 0
    self.circ_buf_count = 0
end

-- Puts sample in the input frames into the circular buffers used
-- to store pre-trigger history.  Stores up to "pre-count" input
-- samples.  If the number of input samples is greater than the
-- number required for the pre-trigger history, the most recent
-- input samples are stored.
-- Parameters:
--    'input_offset' is the input frame data index to start at
--    'last_offset' is the input frame data index to end at
function mt:store_pre_trig_history(input_offset, last_offset)
    local count = last_offset - input_offset + 1
    if count >= self.pre_count then
        -- this is the simple case:  just fill circular buffer
        -- with the "tail end" of the incoming data
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            local iv = ffi_cast(sample_type, infr.buf)
            local hv = ffi_cast(sample_type,
                                self.chan_info[i].circular_buf.buf)
            local input_index = last_offset - self.pre_count + 1
            for j = 0, self.pre_count - 1 do
                hv[j] = iv[input_index]
                input_index = input_index + 1
            end
        end
        self.circ_buf_back = 0
        self.circ_buf_front = 0
        self.circ_buf_count = self.pre_count
    elseif count > 0 then
        -- not enough new samples to fill circular buffer, so
        -- must store them all
        local hist_index
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            local iv = ffi_cast(sample_type, infr.buf)
            local hv = ffi_cast(sample_type,
                            self.chan_info[i].circular_buf.buf)
            hist_index = self.circ_buf_back
            for j = input_offset, last_offset do
                hv[hist_index] = iv[j]
                hist_index = (hist_index + 1) % self.pre_count
            end
        end
        self.circ_buf_count =
            math_min(self.circ_buf_count + count, self.pre_count)
        self.circ_buf_back = hist_index
        if self.circ_buf_count >= self.pre_count then
            self.circ_buf_front = self.circ_buf_back
        end
    end
end

-- Copies each input channel's circular buffer contents into an
-- output frame and puts the frame in the output pipe list.
-- Parameters:
--    'current_time' timestamp for trigger sample
function mt:write_pre_trig_history(current_time)
    if self.circ_buf_count <= 0 then
        return
    end

    -- put requested pre-trigger history in an
    -- output frame for each input channel,
    -- and put output frames in pipe lists
    local count = math_min(self.circ_buf_count, self.pre_count)
    for i = 1, #self.chan_info do
        -- create output frame
        local info = self.chan_info[i]
        local infr = info.in_frame
        local sample_type = utils.frame_data_type(infr.element_size)
        local iv = ffi_cast(sample_type, infr.buf)
        local hv = ffi_cast(sample_type, info.circular_buf.buf)
        local outfr = Frame_create(count, infr.element_size)
        outfr.hdr.ts = current_time - count
        outfr.offset = ffi_cast("int64_t", -count)
        local ov = ffi_cast(sample_type, outfr.buf)
        -- copy history from circular buffer
        local hist_index = self.circ_buf_front
        for j = 0, count - 1 do
            ov[j] = hv[hist_index]
            hist_index = (hist_index + 1) % self.pre_count
        end
        outfr.count = count
        utils.put_frame(outfr, info.pipe_list)
    end

    self.burst_written = self.burst_written + count 
    self.pre_trig_written = count

    -- empty the pre-trigger buffers
    self.circ_buf_count = 0
    self.circ_buf_back = 0
    self.circ_buf_front = 0
end

-- Finds the proper SIE data format to use:
--     “int”, “uint”, or “float”.
-- Paramater:
--     'ch_info' contains information about the input channel
-- Returns:  'float'
-- NOTE - FUTURE - needs to be changed to support integer types
local function get_data_format(ch_info)
    return 'float'
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- create meta data table for this SIE channel
        info.stanza.connector = ""
        info.stanza.description = self.stanza.description or ""
        info.stanza.sie_core_schema = 'somat:burst'
        info.stanza.somat_datamode_type = 'burst_history'
        info.stanza.somat_data_bits =
            sample_types.bits(info.stanza.data_type)
        info.stanza.somat_data_format = 'float'
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = info.name .. '@' .. self.name
        -- The following table specializes the data "dimensions"
        -- for a "burst" SIE channel.
        info.stanza.dims = {
                {
                  label = info.stanza.time_label or 'Time',
                  units = info.stanza.time_units or 'Seconds',
                  xform_scale = 1 / self.real_sample_rate,
                  xform_offset = 0
                }, {
                  label = info.stanza.name,
                  units = info.stanza.output_units
                }, {
                  label = 'Offset',
                  units = 'Sample'
                }
            }
        info.stanza.decoder_type = "_burst_" ..
            tostring(info.stanza.somat_data_format) ..
            tostring(info.stanza.somat_data_bits)
        local md = sie_utils.sie_channel(self.sw, info.stanza)
        -- write the meta data for the SIE channel
        sie_utils.write_channel(self.sw, md)
        -- get node for this input channel
        local in_node = info.in_node
        -- Attach one end of pipe to input channel node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, 8192,
            ffi_sizeof(info.data_type), 8, p,
            md.data_type:find('^message_') and true or nil,
            BH_DATAMODE)
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    got_frames = got_frames and Pipe_has_frames(self.trig_pipe)
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
        self.trig_frame = Pipe_get_frame(self.trig_pipe)
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
        self.trig_frame = nil
    end
    return got_frames
end

-- Checks trigger frame for next trigger occurrence, starting
-- at given index.
-- Paramater:
--     'start_index' is where to start looking in trigger frame
--     'max' is where to stop looking in trigger frame (inclusive)
-- Returns:  index of next input sample to process, and true
--     if a trigger was found (false otherwise for second value)
function mt:look_for_trig(start_index, max)
    local trigger_found = false
    local index = start_index
    local trig_vals =
        ffi_cast(utils.frame_data_type(
            self.trig_frame.element_size), self.trig_frame.buf)
    -- find next trigger condition in trigger frame
    for i = start_index, max do
        if self:trigger(trig_vals[i]) then
            index = i
            trigger_found = true
            break
        end
    end
    if trigger_found then
        self.trigger_time = self.tot_count + index
        if self.pre_count > 0 and index > start_index then
            -- store any pre-trigger samples as history
            self:store_pre_trig_history(start_index, index - 1)
        end
    else
        -- have "processed" all input samples
        index = max + 1
        if self.pre_count > 0 then
            -- store all given samples as history
            self:store_pre_trig_history(start_index, max)
        end
    end
    return index, trigger_found
end

-- Processes incoming frames, producing outgoing frames.
-- Outline of logic:
--     if we have frames from all input pipes and the trigger pipe
--         set 'next input index' to 0 -- (same variable used for all frames)
--         set 'count' to first input frame's count value
--         loop until 'next input index' >= 'count'
--             if not in "triggered" state
--                 check for a trigger condition in the trigger frame
--                 if trigger condition found
--                     set "triggered" state on
--                     if pre-trigger history required
--                         for each input frame
--                             copy pre-trigger history from the
--                                 pre-trigger buffer to an output frame
--                             empty the pre-trigger buffer
--                             put the output frame in outgoing pipe list.
--                             NOTE:  pre-trigger history is never
--                                        duplicated in output frames!
--                         increment 'burst used' count by pre-trigger output
--                         set 'next input index' to the trigger index value
--                 else if a pre-trigger history is required
--                     for each input frame
--                         put as many input samples as necessary in
--                             the pre-trigger history buffer
--                     -- done with these frames
--                     set 'next input index' to 'count'
--             end
--             if in "triggered" state
--                 for each input frame
--                     copy as many samples as necessary into an output frame
--                         starting from the 'next' index
--                     put the output frame in outgoing pipe list.
--                 set 'next' to index of first input sample that
--                     hasn't been copied
--                 if number of samples output in the current "burst" 
--                         is the requested size of the burst
--                     reset to "not triggered" state
--         end  -- frame sample loop
--         get next set of input and trigger frames, process them
function mt:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        local input_index = 0
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- loop until all input frame samples are processed
        while input_index < count do
            if self.triggered then
                -- The "triggered" state may have been set by earlier
                -- handling of a non-triggered state for this set of
                -- input frames, or may have been carried over from a
                -- previous set of input frames.
                if self.pre_count > 0 then
                    self:write_pre_trig_history(self.trigger_time)
                end
                local left_to_write = self.burst_size - self.burst_written
                local left_in_frame = count - input_index
                local out_count = math_min(left_to_write, left_in_frame)
                local trig_offset = tonumber(self.burst_written -
                    self.pre_trig_written)
                for i = 1, #self.chan_info do
                    local info = self.chan_info[i]
                    local infr = info.in_frame
                    local sample_type = utils.frame_data_type(infr.element_size)
                    local iv = ffi_cast(sample_type, infr.buf)
                    local outfr = Frame_create(out_count,
                        ffi_sizeof(info.data_type))
                    local ov = ffi_cast(sample_type, outfr.buf)
                    local in_index = input_index
                    -- copy to output frame buf
                    for j = 0, out_count - 1 do
                        ov[j] = iv[in_index]
                        in_index = in_index + 1
                    end
                    outfr.count = out_count
                    outfr.hdr.ts = self.tot_count + input_index
                    outfr.offset = ffi_cast("int64_t", trig_offset)
                    utils.put_frame(outfr, info.pipe_list)
                end
                input_index = input_index + out_count
                self.burst_written = self.burst_written + out_count 
                if self.burst_written >= self.burst_size then
                    -- finished this burst, so flush sie bufs
                    -- and reset some volatiles
                    self.sw:flush_these(self.sie_collect_idx,
                        #self.chan_info)
                    self.triggered = false
                    self.pre_trig_written = 0
                    self.burst_written = 0
                end
            else
                input_index, self.triggered =
                    self:look_for_trig(input_index, count - 1)
            end
        end  -- while input is left in frames to process

        self.tot_count = self.tot_count + count

        -- don't need the processed input frames any longer
        Frame_delete(self.trig_frame)
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

-- Cleans up at test conclusion.
-- Paramater:
--     'sie_writer' is SIE writer object - NOT USED HERE
function mt:finalize(sie_writer)
-- FIX ME note:  The following cleanup (e.g., calls of C.free)
-- causes problems at this point (v1.3.3.05) because run() is
-- called once more by run-test after this function is called.
-- That results in already freed memory being reused, which
-- causes random havoc/crashes.  The immediate low-risk fix
-- is to not do the C.free calls.  Eventually we will want
-- to change run-test.lua so that run() is not called after
-- this function is called.  Then the C.free calls can be
-- added back.
--    for i = 1, #self.chan_info do
--        if self.chan_info[i].circular_buf then
--            C.free(self.chan_info[i].circular_buf.buf)
--        end
--    end
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local burst_history = setmetatable({}, {__index = mt})
    burst_history.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    burst_history.sie_collect_idx = first_sie_index
    burst_history.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Burst History data mode " ..
        "must have at least one input channel")
    --     node name
    burst_history.name = stanza.name
    assert(burst_history.name, "Missing node name")
    --     sample rate
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        burst_history.sample_rate = node_meta.sample_rate
        if burst_history.sample_rate then
            burst_history.real_sample_rate =
                rational:new(burst_history.sample_rate):tonumber()
            burst_history.sample_rate =
                math.floor(burst_history.real_sample_rate)
            if burst_history.sample_rate > 0 then
                break
            end
        end
    end
    assert(burst_history.sample_rate and burst_history.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(burst_history.sample_rate)))
    --     pre- and post-trigger counts
    assert(stanza.burst_pre_trigger_time, "Missing pre-trigger time")
    local pre_trig_time = tonumber(stanza.burst_pre_trigger_time)
    assert(pre_trig_time >= 0, "Invalid pre-trigger time: \"" ..
        tostring(stanza.burst_pre_trigger_time) .. "\"")
    burst_history.pre_count = pre_trig_time * burst_history.sample_rate
    assert(stanza.burst_post_trigger_time, "Missing post-trigger time")
    local post_trig_time = tonumber(stanza.burst_post_trigger_time)
    assert(post_trig_time >= 0, "Invalid post-trigger time: \"" ..
        tostring(stanza.burst_post_trigger_time) .. "\"")
    burst_history.post_count = post_trig_time * burst_history.sample_rate
    burst_history.burst_size = burst_history.pre_count +
        burst_history.post_count + 1
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode and
        (trigger_mode == "when_true" or
         trigger_mode == "on_false_to_true_edge"),
        "Missing or invalid trigger mode: \"" ..
        tostring(trigger_mode) .. "\"")
    local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Burst History requires a trigger channel")
    local trig_node = node_list[trig_chan]
    assert(trig_node, "Node for trigger channel " ..
        tostring(trig_chan) .. " not found")
    burst_history.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(burst_history.trig_pipe, "failed to create trigger pipe")
    burst_history.trigger = trigger_mode == "when_true" and
        burst_history.gate_trigger or burst_history.edge_trigger
    --     create input pipe structures
    local input_node
    local node_meta
    for i = 1, #input_channels do
        input_node = node_list[input_channels[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        --     data type and cdefs
        local data_type = node_meta.data_type
        assert(data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        -- channel info array
        burst_history.chan_info[i] =
            {name = input_channels[i],
             in_node = input_node,
             in_pipe = Pipe_attach_pipe(input_node.pipe_list),
             pipe_list = {},
             stanza = node_meta,
             data_type = data_type}
    end

    if burst_history.pre_count > 0 then
        burst_history:init_circ_bufs()
    end

    -- initialize volatile values
    burst_history.triggered = false
    burst_history.prev_trig_on = false
    burst_history.burst_written = 0
    burst_history.pre_trig_written = 0

    -- used for frames' time stamps (just a sample count for
    -- this module)
    burst_history.tot_count = 0

    return burst_history, #input_channels  -- 1 sie buf per input chan
end

return m
