local cjson = require 'cjson'
local ffi = require 'ffi'
local C = ffi.C 
local Pipe = require 'test_engine/pipe'
local Frame = require 'test_engine/frame'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local Pipe_attach_pipe = Pipe.attach_pipe
local Pipe_get_frame = Pipe.get_frame
local Pipe_has_frames = Pipe.has_frames
local Frame_delete = Frame.delete

local math_floor = math.floor
local math_min = math.min
local math_max = math.max

local ctor_cache = {}

local histogram_bin_channel = [==[
    |<ch id="(=ch_id=)" private="bin">
    | <tag id="core:description">Generic histogram bin min/max values
    |+   lookup table</tag>
    | <dim index="0"><data decoder="(=bin_decoder_id=)" v="0"/></dim>
    | <dim index="1"><data decoder="(=bin_decoder_id=)" v="1"/></dim>
    | <dim index="2"><data decoder="(=bin_decoder_id=)" v="2"/></dim>
    |</ch>
    ]==]

local histogram_decoder_1d = [==[
    |<decoder id="(=decoder_id=)">
    | <read var="dimcount" bits="32" type="uint" endian="big"/>
    | <read var="bincount1" bits="32" type="uint" endian="big"/>
    | <loop var="v1" start="0" end="{$bincount1}">
    |  <read var="v0" bits="(=bin_bits=)" type="uint" endian="big"/>
    |  <sample/>
    | </loop>
    |</decoder>
    ]==]

local mt = {}

-- function mt:run_evenly_divided_sequential(in_buf, count)
-- function mt:run_evenly_divided_message(in_buf, count)

function mt:run_evenly_divided_generic(num)
    local hist = self.hist[num]
    local f = Pipe_get_frame(hist.in_pipe)
    local bin
    local t
    while f ~= nil do
        local sptr = ffi_cast(f.element_size == 4 and 'float *' or 'double *',
            f.buf)
        for i = 0, f.count - 1 do
            t = sptr[i]
            bin = math_floor((t - hist.lo_bound) * hist.bound_scale) + 1
            bin = math_max(0, bin)
            bin = math_min(hist.bin_count - 1, bin)
            hist.bins[bin] = hist.bins[bin] + 1
            if hist.bins[bin] == 0 then
                hist.bins[bin] = bit.bnot(hist.bins[bin])
            end
        end
        Frame_delete(f)
        f = Pipe_get_frame(hist.in_pipe)
    end
end

function mt:run()
    local num = self.num_hist - 1
    for i = 0, num do
        if not Pipe_has_frames(self.hist[i].in_pipe) then return end
    end
    for i = 0, num do
        self:run_evenly_divided_generic(i)
    end
end

function mt:finalize(sie_writer)
    if self.final_called ~= 0 then return end
    self.final_called = 1
    local num = self.num_hist - 1
    local bin_bits = 64
    local hdr = ffi.new('uint32_t[2]')
    hdr[0] = 3
    for i = 0, num do
        local hist = self.hist[i]
        local buf = sie_writer.bufs[hist.sie_idx]
        hdr[1] = hist.bin_count
        buf:enq(hdr, ffi_sizeof('uint32_t') * 2)
        buf:enq(hist.bins, hdr[1] * bin_bits / 8)
        buf:flush(true)
    end
end

function mt:histogram_1d_channel(itv, sw)
    local tv = {}
    tv.ch_id = itv.ch_id or sw:next_id('ch')
    tv.grp = itv.grp or sw:next_id('group') -- KLUDGE WRONG probably...
    sw:xml('<ch test="0" id="(=ch_id=)" base="(=base=)" group="(=grp=)" name="(=name=)"/>\n', itv, tv)
    return tv.ch_id
end

function mt:get_histogram_bin_base_channel(bounds_bits, sw)
    if sw.hist_bounds_base_channel then
        return sw.hist_bounds_base_channel
    end

    bounds_bits = bounds_bits or 64
    local bounds_type = '_histogram_bin_bounds' .. bounds_bits

    local tv = {}
    tv.bin_decoder_id = sw:decoder_id(bounds_type)
    tv.ch_id = sw:next_id('ch'),
    sw:xml()
    sw:xml(histogram_bin_channel, tv)
    sw:xml()
    sw.hist_bounds_base_channel = tv.ch_id
    return tv.ch_id
end

function mt:histogram_bounds_channel(base_ch_id, group, name, sw)

    local tv = {
        i = '',
        base = base_ch_id,
        grp = group,
        ch_id = sw:next_id('ch'),
        name = name,
    }

    sw:xml('(=i=)<ch id="(=ch_id=)" base="(=base=)" group="(=grp=)"'
        .. ' name="(=name=)#bin_1" private="bin"/>\n', tv)
    return tv.ch_id
end

function mt:histogram_1d_base_channel(itv, sw)
    local tv = {}
    tv.base_id = sw:next_id('ch') -- KLUDGE CACHE
    tv.decoder_id = sw:next_id('decoder') -- KLUDGE CACHE

    sw:xml(histogram_decoder_1d, tv, itv)
    sw:xml()

    sw:xml([==[
    | <ch id="(=base_id=)" name="(=input_channel_name=)@(=datamode_name=)"
    |+     private="base">
    |  <tag id="somat:datamode_name">(=datamode_name=)</tag>
    |  <tag id="somat:datamode_type">time_at_level</tag>
    |  <tag id="somat:data_bits">(=bin_bits=)</tag>
    |  <tag id="somat:data_format">uint</tag>
    |  <tag id="somat:input_channel">(=input_channel_name=)</tag>
    |  <tag id="core:description">(=input_channel_description=)</tag>
    |  <tag id="core:sample_rate">(=sample_rate=)</tag>
    |  <tag id="core:schema">somat:histogram</tag>
    |  <dim index="0">
    |   <tag id="core:description"></tag>
    |   <tag id="core:label">Counts</tag>
    |   <tag id="core:units"></tag>
    |   <data decoder="(=decoder_id=)" v="0"/>
    |  </dim>
    |  <dim index="1">
    |   <tag id="core:description">Axis 0 Bin Min</tag>
    |   <tag id="core:label">(=input_channel_name=) ((=units=))</tag>
    |   <tag id="somat:range_min">(=range_min=)</tag>
    |   <tag id="somat:range_max">(=range_max=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |   <xform index_ch="(=index_ch=)" index_dim="1"/>
    |  </dim>
    |  <dim index="2">
    |   <tag id="core:description">Axis 0 Bin Max</tag>
    |   <tag id="core:label">(=input_channel_name=) ((=units=))</tag>
    |   <tag id="somat:range_min">(=range_min=)</tag>
    |   <tag id="somat:range_max">(=range_max=)</tag>
    |   <data decoder="(=decoder_id=)" v="1"/>
    |   <xform index_ch="(=index_ch=)" index_dim="2"/>
    |  </dim>
    | </ch>
    ]==], tv, itv)
    sw:xml()

    return tv.base_id
end

-- Create all of the SIE buffers needed by this sink
function mt:setup(sie_writer, is_restart)
    for i = 0, self.num_hist - 1 do
        local hist = self.hist[i]

        local name = ffi_string(hist.hist_name)
        local input_channel_name = ffi_string(hist.name)
        local input_channel_description = ffi_string(hist.description)
        local sample_rate = hist.sample_rate

        local bounds_bits = 64
        local bounds_base =
            self:get_histogram_bin_base_channel(bounds_bits, sie_writer)

        local idx = sie_writer:new_buffer(sie_writer.scratch_buf_idx, nil,
            (hist.bin_count + 1) * bounds_bits / 8, bounds_bits / 8, 0)

        -- Write this stuff
        local bbuf = sie_writer.bufs[idx]
        bbuf:append(hist.bounds, (hist.bin_count + 1) * bounds_bits / 8)
        bbuf:flush(true)

        local bounds_ch = self:histogram_bounds_channel(bounds_base,
            bbuf.group, name, sie_writer)

        local bin_bits = 64
        local hist_base = self:histogram_1d_base_channel({
            datamode_name = name,
            input_channel_name = input_channel_name,
            input_channel_description = input_channel_description,
            sample_rate = sample_rate,
            bin_bits = bin_bits,
            index_ch = bounds_ch,
            range_min = hist.lo_bound,
            range_max = hist.hi_bound,
            units = ffi_string(hist.units),
        }, sie_writer)

        local dims = 1

        sie_writer:new_buffer(hist.sie_idx, nil,
            hist.bin_count * bin_bits / 8, bin_bits / 8,
            ffi_sizeof('uint64_t') * (dims + 1))
        local buf = sie_writer.bufs[hist.sie_idx]

        local tv = {
            name = input_channel_name .. '@' .. name,
            base = hist_base,
            grp = buf.group,
        }

        sie_writer:reserve_space(
            ((hist.bin_count + 2) * (bin_bits / 8)) +   -- bin space
            (ffi_sizeof('uint32_t') * 2) + -- header space
            (ffi_sizeof('uint32_t') * 5), -- SIE overheader
            function() self:finalize(sie_writer) end)

        self:histogram_1d_channel(tv, sie_writer)
    end
end

local function meta_ctor(t)
    if not ctor_cache[t] then
        local tal_type = string.format('struct %s_tal', t)
        local spec = string.gsub([[
            struct $t_hist_sie {
                $t dimcount; // KLUDGE placeholder
                $t bin_count; // KLUDGE placeholder
            };
            struct $t_hist {
                int             bin_count;
                $t              *bins;
                double          *bounds;
                /* For evenly_divided special case */
                double          lo_bound;
                double          hi_bound;
                double          bound_scale;
                double          bound_span;
                int             sie_idx;
                char            *name;
                char            *description;
                char            *hist_name;
                char            *units;
                double          sample_rate;
                struct pipe     *in_pipe;
            };
            struct $t_tal {
                int             num_hist;
                int             final_called;
                struct $t_hist  *hist;
            };
        ]], '%$(%w+)', {t = t})
        ffi.cdef(spec)
        ctor_cache[t] = ffi.metatype(tal_type, {__index = mt})
    end
    return ctor_cache[t]
end

require 'middleclass'

local m = class('time_at_level_1d')

function m:setup(sie_writer, is_restart)
    self.tal:setup(sie_writer, is_restart)
end

function m:run()
    if self.trigger_module then self.trigger_module:run() end
    self.tal:run()
end

--[[
function m:finalize(sie_writer)
    self.tal:finalize(sie_writer)
end
]]

function m:initialize(stanza, node_list, sie_idx, stanza_list)
    local bin_type = stanza.bin_type or 'uint64_t'
    local ctor = meta_ctor(bin_type)

    self.sie_buffers_needed = #stanza.input_channels

    self.tal = ctor()
    self.tal.num_hist = #stanza.input_channels
    self.tal.hist = ffi_cast(string.format('struct %s_hist *', bin_type),
        C.calloc(self.tal.num_hist,
        ffi_sizeof(string.format('struct %s_hist', bin_type))))
    local objs = self.tal.hist

    -- If triggered then instantiate a new trigger module
    self.trigger_module = stanza.trigger_mode ~= 'always_on' and
        require('test_engine/modules/trigger/' .. stanza.trigger_mode):new(
            stanza.input_channels, stanza.trigger_channel, node_list)

    for i = 0, #stanza.input_channels - 1 do
        local hb = stanza.histogram_bins[i + 1]
        local obj = objs[i]
        local ic = stanza.input_channels[i + 1]

        obj.sie_idx = sie_idx + i

        -- Add 2 extra bins for underflow and overflow
        obj.bin_count = hb.number_of_bins + 2
        obj.bins = ffi_cast(string.format('%s *', bin_type),
            C.calloc(obj.bin_count, ffi_sizeof(bin_type)))
        obj.bounds = ffi_cast('double *',
            C.calloc(obj.bin_count + 1, ffi_sizeof('double')))
        obj.name = ffi_cast('char *',
            C.calloc(ic:len() + 1, ffi_sizeof('char')))
        ffi_copy(obj.name, ic)
        local units = stanza_list[ic].output_units or ''
        obj.units = ffi_cast('char *', C.calloc(units:len() + 1,
            ffi_sizeof('char')))
        ffi_copy(obj.units, units)

        local desc = stanza_list[ic].description or ic or
            'No channel description'

        obj.description = ffi_cast('char *', C.calloc(desc:len() + 1,
            ffi_sizeof('char')))

        ffi_copy(obj.description, desc)

        obj.hist_name = ffi_cast('char *',
            C.calloc(stanza.name:len() + 1, ffi_sizeof('char')))
        ffi_copy(obj.hist_name, stanza.name)

        if not self.trigger_module then
            -- Not triggered...attach directly to input channel
            obj.in_pipe = Pipe_attach_pipe(node_list[ic].pipe_list)
        else
            -- Triggered...attach to gated channel
            obj.in_pipe =
                Pipe_attach_pipe(self.trigger_module.chan_info[i + 1].pipe_list)
        end

        obj.sample_rate = rational(stanza_list[ic].sample_rate):tonumber()

        -- Evenly divided bins
        if hb.type == 'evenly_divided' then
            obj.lo_bound = hb.low_bound
            obj.hi_bound = hb.high_bound
            obj.bound_span = obj.hi_bound - obj.lo_bound
            obj.bound_scale = (obj.bin_count - 2) / obj.bound_span
            obj.bounds[0] = -1/0
            obj.bounds[obj.bin_count] = 1/0
            for i = 0, obj.bin_count - 2 do
                obj.bounds[i + 1] = obj.lo_bound + i / obj.bound_scale
            end
    
        -- User defined bins
        elseif hb.type == 'user_defined' then
            error("TAL Datamode with user defined bins is not yet supported")
    
        -- Unknown bin type
        else
            error(string.format("TAL with unknown bin type: %s", hb.type))
        end
    end

    return self
end

return m
