require 'strict'

-- time_history.lua - Implements the Triggered Time History
-- data mode.
--
-- Implementation note:
--     Data modes in general handle setting up their own trigger
--     mechanisms, where such are called for.  For those (such as
--     this one) which use the "generic" triggering mechanism a
--     trigger.lua module will be implemented.  The trigger module
--     will be given all this data mode instance's input channels,
--     along with the trigger channel, and the trigger module will
--     then provide this data mode instance with an input pipe
--     for each input channel.  This data mode instance will then
--     simply write each input pipe's frames to the associated
--     sie buffer pipe.

local cjson = require 'cjson'
local dbg = require 'dbg'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local log = require 'log'
local notify = require 'sys_notify'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sample_types = require 'sample_types'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_add_frame     = pipe.add_frame
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame

local mt = {}

function mt:dump()
    --dbg('tot_count=%lld', self.tot_count)
end

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Finds the proper SIE data format to use:
--     “int”, “uint”, or “float”.
-- Paramater:
--     'ch_info' contains information about the input channel
-- Returns:  'float'
-- NOTE - KLUDGE - needs to be changed to support integer types
local function get_data_format(ch_info)
    return 'float'
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- create meta data table for this SIE channel
        info.stanza.sie_core_schema = 'somat:sequential'
        info.stanza.somat_datamode_type = 'time_history'
        info.stanza.somat_data_bits =
            sample_types.bits(info.stanza.data_type)
        info.stanza.somat_data_format = 'float'
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = info.name .. '@' .. self.name
        -- The following table specializes the data "dimensions"
        -- for a "sequential" SIE channel.
        info.stanza.dims = {
                {
                  label = info.stanza.time_label or 'Time',
                  units = info.stanza.time_units or 'Seconds',
                  xform_scale = 1 / self.sample_rate,
                  xform_offset = 0
                }, {
                  label = info.stanza.name,
                  units = info.stanza.output_units
                }
            }
        info.stanza.decoder_type = info.stanza.data_type
        local md = sie_utils.sie_channel(self.sw, info.stanza)
        -- write the meta data for the SIE channel
        sie_utils.write_channel(self.sw, md)
        -- get node for this input channel
        local in_node = info.in_node
        -- Attach one end of pipe to input channel node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, 8192,
            in_node.frame_sample_size, 8, p,
            md.data_type:find('^message_') and true or nil)
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Processes incoming frames, producing outgoing frames.
function mt:run()
    if self.trigger_module then
        -- trigger node feeds this node
        self.trigger_module:run()
    end

    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

         -- get input frames data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            self.chan_info[i].invals =
                ffi_cast(sample_type, infr.buf)
        end
        -- create output frames and get data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            local outfr =
                Frame_create(infr.maxsize, infr.element_size)
            outfr.hdr.ts = self.tot_count
            self.chan_info[i].out_frame = outfr
            self.chan_info[i].outvals =
                ffi_cast(sample_type, outfr.buf)
        end

        -- fill output frames, and put them in "sie buffer" pipes
        for i = 1, #self.chan_info do
            local iv = self.chan_info[i].invals
            local ov = self.chan_info[i].outvals
            for j = 0, count - 1 do
                ov[j] = iv[j]
            end
            self.chan_info[i].out_frame.count = count
            -- put output frame in output pipe(s)
            utils.put_frame(self.chan_info[i].out_frame,
                self.chan_info[i].pipe_list)
        end

        self.tot_count = self.tot_count + count

        -- don't need the processed input frames any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local time_history = setmetatable({}, {__index = mt})
    time_history.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    time_history.sie_collect_idx = first_sie_index
    time_history.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Time History data mode " ..
        "must have at least one input channel")
    --     node name
    time_history.name = stanza.name
    assert(time_history.name, "Missing node name")
    --     sample rate
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        time_history.sample_rate = node_meta.sample_rate
        if time_history.sample_rate then
            time_history.sample_rate =
                rational:new(time_history.sample_rate):tonumbers()
            if time_history.sample_rate > 0 then
                break
            end
        end
    end
    assert(time_history.sample_rate and time_history.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(time_history.sample_rate)))
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    if trigger_mode == "always_on" then
        --     create input pipe structures
        local input_node
        local node_meta
        for i = 1, #input_channels do
            input_node = node_list[input_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            node_meta = meta_list[input_channels[i]]
            assert(node_meta, "metadata stanza for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            time_history.chan_info[i] =
                {name = input_channels[i],
                 in_node = input_node,
                 in_pipe = Pipe_attach_pipe(input_node.pipe_list),
                 pipe_list = {},
                 sample_size = input_node.sample_size,
                 stanza = node_meta}
        end
    else
        --     create trigger module instance
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            local sn = notify:new(__'Test Error')
            local msg = string.format(__'Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name)
            sn:error(msg)
            log:log(log.ERR, msg)
        end
        local trigger_module =
            mod:new(input_channels, trig_chan, node_list)
        if not trigger_module then
            local sn = notify:new(__'Test Error')
            local msg = string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name)
            sn:error(msg)
            log:log(log.ERR, msg)
        end
        assert(#trigger_module.chan_info ==
               #input_channels, "Number of trigger output pipes " ..
               "does not equal number of input channels")
        --     create input pipe structures using
        --     trigger module output pipes as this instance's
        --     input pipes
        local input_node
        for i = 1, #input_channels do
            input_node = node_list[input_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            local node_meta = meta_list[input_channels[i]]
            assert(node_meta, "metadata stanza for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            time_history.chan_info[i] =
                {name = input_channels[i],
                 in_node = input_node,
                 in_pipe = Pipe_attach_pipe(
                     trigger_module.chan_info[i].pipe_list),
                 pipe_list = {},
                 sample_size = input_node.sample_size,
                 stanza = node_meta}
        end
        time_history.trigger_module = trigger_module
    end
 
    -- used for frames' time stamps (just a sample count for
    -- this module)
    time_history.tot_count = 0

    return time_history, #input_channels  -- 1 sie buf per input chan
end

return m
