require 'strict'

-- trig_common.lua - Provides a framework for triggers
-- used by a data mode.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list

local mt = {}

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    got_frames = got_frames and Pipe_has_frames(self.trig_pipe)
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
        self.trig_frame = Pipe_get_frame(self.trig_pipe)
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
        self.trig_frame = nil
    end
    return got_frames
end

-- Filters input frame(s) based on the value of the trigger frame.
--     If a trigger frame sample is not zero the corresponding
--     samples from the input frames are copied to their output
--     frames.  If the trigger sample is zero, the corresponding
--     samples from the input frames are ignored.
function mt:run()
    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in trigger and input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        So it doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

        -- get input and trigger frames data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            self.chan_info[i].invals =
                ffi_cast(sample_type, infr.buf)
        end
        local trig_vals =
            ffi_cast(utils.frame_data_type(self.trig_frame.element_size),
                     self.trig_frame.buf)
        -- create output frames and get data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local outfr = Frame_create(infr.maxsize, infr.element_size)
            local sample_type = utils.frame_data_type(
                self.chan_info[i].in_frame.element_size)
            outfr.hdr.ts = infr.hdr.ts
            self.chan_info[i].out_frame = outfr
            self.chan_info[i].outvals =
                ffi_cast(sample_type, outfr.buf)
        end

        -- process each trigger and input sample
        local out_index = 0
        for in_index = 0, count - 1 do
            if self.core_trig:triggered(trig_vals[in_index]) then
                for findex = 1, #self.chan_info do
                    self.chan_info[findex].outvals[out_index] =
                        self.chan_info[findex].invals[in_index]
                end
                out_index = out_index + 1
            end
        end

        -- only send output frames if they have some data
        if out_index > 0 then
            -- set count for each output frame and put it in pipe;
            -- also create a new output frame for each input channel
            for i = 1, #self.chan_info do
                local of = self.chan_info[i].out_frame
                of.count = out_index
                utils.put_frame(of, self.chan_info[i].pipe_list)
            end
        else
            -- delete output frames, since we aren't relying on the
            -- next output frames being the same size
            for i = 1, #self.chan_info do
                Frame_delete(self.chan_info[i].out_frame)
            end
        end

        -- don't need the processed input frames any longer
        Frame_delete(self.trig_frame)
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        have_all_frames = self:get_frames()
    end
end

local m = {}

-- Creates a new trigger object.
-- Parameters:
--   'in_chans'  input channel names
--   'trig_chan' trigger channel name
--   'node_list' list of previously instantiated channel nodes
-- Returns:
--   the new instance object
function m:new(in_chans, trig_chan, node_list, core_trig)
    assert(in_chans, "missing first parameter")
    assert(trig_chan, "missing second parameter")
    assert(node_list, "missing third parameter")
    assert(core_trig, "missing core trigger object parameter")

    local trig_common = setmetatable({}, {__index = mt})

    -- get input trigger pipe
    local trig_node = node_list[trig_chan]
    assert(trig_node, "Node for trigger channel " ..
        tostring(trig_chan) .. " not found")
    trig_common.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(trig_common.trig_pipe, "failed to create trigger pipe")

    -- set info for each input channel
    trig_common.chan_info = {}
    for i = 1, #in_chans do
        local input_node = node_list[in_chans[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(in_chans[i]) .. "\" not found")
        trig_common.chan_info[i] =
            {name = in_chans[i],
             in_pipe = Pipe_attach_pipe(input_node.pipe_list),
             pipe_list = Pipe_new_pipe_list()}
    end

    trig_common.core_trig = core_trig

    trig_common.trigger_on = false
    return trig_common
end

return m
