require 'strict'
local ffi = require 'ffi'
local C = ffi.C
local frame = require 'test_engine/frame'

local ffi_sizeof = ffi.sizeof
local ffi_cast = ffi.cast

local Frame_delete = frame.delete

ffi.cdef[[
    struct pipe {
        struct pipe     *next;          // Next frame in list
        struct pipe     *prev;          // Previous frame in list
        struct frame    *frame_list;    // Frame list
        struct frame    *last_frame;    // Last frame in frame list
    };

    void *calloc(size_t nmemb, size_t size);
    void free(void *ptr);
]]

local m = {}

-- Create a pipe
function m._create()
    local p = ffi_cast('struct pipe *', C.calloc(1, ffi_sizeof('struct pipe')))
    if p == nil then
        error("Can not allocate memory for pipe")
        return
    end
    return p
end

-- Delete a pipe
function m._delete(pipe)
    if not pipe or pipe == nil then
        error("Unable to delete pipe:  argument required")
    end
    -- First delete any frames that are still in the pipe
    m.flush_frames(pipe)
    C.free(pipe)
end

-- Creates a new pipe and attaches it to the specified pipe list
function m.attach_pipe(pipe_list)
    if not pipe_list or pipe_list == nil then
        error("Unable to attach new pipe to pipe list:  argument required")
    end
    local p = m._create()
    local pl = pipe_list[0]
    if pl == nil then
        pipe_list[0] = p
        return p
    end
    while pl.next ~= nil do
        pl = pl.next
    end
    pl.next = p
    p.prev = pl
    return p
end

-- Detaches a pipe from a pipe list
function m.detach_pipe(pipe_list, pipe)
    if not pipe_list or pipe_list == nil then
        error("Unable to detach pipe from pipe list:  pipe list required")
    end
    if not pipe or pipe == nil then
        error("Unable to detach:  pipe not given")
    end
    local p = pipe_list[0]
    if pipe == p then
        -- First node in list
        pipe_list[0] = pipe.next
    end

    -- Fix up links for previous and next nodes if they exist
    if pipe.next ~= nil then
        pipe.next.prev = pipe.prev
    end
    if pipe.prev ~= nil then
        pipe.prev.next = pipe.next
    end

    m._delete(pipe)
end

-- Add a frame to the beginning of the frame list
function m.add_frame(pipe, f)
    if not pipe or pipe == nil then
        error("Unable to add frame to pipe:  pipe not specified")
    end
    if not f or f == nil then
        error("Unable to add frame to pipe:  frame not specified")
    end
    if pipe.frame_list == nil and pipe.last_frame == nil then
        -- Frame list is empty
        pipe.frame_list = f
        pipe.last_frame = f
        f.next = nil
        f.prev = nil
    else
        -- Add frame to beginning of list
        f.next = pipe.frame_list
        f.prev = nil
        if f.next ~= nil then
            f.next.prev = f
        end
        pipe.frame_list = f
    end
end

-- Get the last frame from the frame list
function m.get_frame(pipe)
    if not pipe or pipe == nil then
        error("Unable to get frame from pipe:  pipe not specified")
    end
    local f = pipe.last_frame
    if f ~= nil and f.prev ~= nil then
        -- Frames exist in list prior to last frame
        pipe.last_frame = f.prev
        f.prev.next = nil
    elseif f ~= nil then
        -- Only frame in the list
        pipe.last_frame = nil
        pipe.frame_list = nil
    end
    return f
end

-- Flush and delete all frames from a pipe
function m.flush_frames(pipe)
    if not pipe or pipe == nil then
        error("Unable to flush frames from pipe:  pipe not specified")
    end
    local f = m.get_frame(pipe)
    while f ~= nil do
        Frame_delete(f)
        f = m.get_frame(pipe)
    end
end

-- Return true if pipe has frames, false otherwise
function m.has_frames(pipe_list)
    if not pipe_list or pipe_list == nil then
        error("Unable to determine if pipe has frames:  pipe not specified")
    end
    return (pipe_list.frame_list ~= nil) and true or false
end

-- Create a new pipe list
function m.new_pipe_list()
    local pl = ffi.cast('struct pipe **',
        C.calloc(1, ffi_sizeof('struct pipe *')))
    return pl
end

return m
