require 'strict'
local cjson = require 'cjson.safe'
local ffi = require 'ffi'
local net = require 'net'
local membuf = require 'membuf'
local sample_clock = require 'sample_clock'
local sample_types = require 'sample_types'
local message_can = require 'message_can'
local Frame = require 'test_engine/frame'
local Pipe = require 'test_engine/pipe'
local ev = require 'ev'
local C = ffi.C
local notify = require 'sys_notify'

ffi.cdef[[
    struct sockaddr_un {
        sa_family_t     sun_family;
        char            sun_path[108];
    };
    char *strerror(int errnum);
]]

local ffi_sizeof = ffi.sizeof
local ffi_copy = ffi.copy
local ffi_string = ffi.string
local ffi_cast = ffi.cast

local math_min = math.min

local Frame_create = Frame.create
local Frame_clone = Frame.clone
local Frame_delete = Frame.delete
local Frame_tail = Frame.tail

local Pipe_add_frame = Pipe.add_frame

local ctor_cache = {}

local mt = {}

-- Read whatever data is available on the connection to the proxy and stuff the
-- data into a buffer for processing by a reader!
function mt:_read_from_proxy()
    local ar = self.already_read
    if ar == 0 then
        local ibuf = self.inbuf
        ibuf:assure_available()
        ar = C.read(self.fd, ibuf:tail(), ibuf:available())
        if ar < 0 then
            self.read_errno = ffi.errno()
            if self.read_errno == C.EAGAIN then
                self.read_errno = 0
                return
            end
            self.read_ret = ar
            self.failed = 1
            C.shutdown(self.fd, C.SHUT_RDWR)
            C.close(self.fd)
            error(string.format("Error on connection to channel %s:  %s",
                ffi.string(self.channel),
                ffi.string(C.strerror(self.read_errno))))
            return
        elseif ar == 0 then
            self.read_errno = C.ECONNRESET
            self.read_ret = ar
            self.failed = 1
            C.shutdown(self.fd, C.SHUT_RDWR)
            C.close(self.fd)
            error(string.format(
                "Channel %s has unexpectedly disconnected from the test",
                ffi.string(self.channel)))
            return
        end
        self.already_read = ar
        ibuf:append(nil, ar)
        ibuf:assure_available()
        self.i = 0
    end
end

local plh_size = ffi_sizeof'proxy_length_header'
local ts_size = ffi_sizeof'timestamp'

function mt:reset()
    self.inbuf:remove(0, self.i)
    self.already_read = 0
    self.i = 0
    return nil
end

function mt:_process_read_data()
    local ibuf = self.inbuf
    local ib_count = #ibuf
    if self.i + plh_size > ib_count then
        return self:reset()
    end

    local ph = ffi_cast('proxy_header *', ibuf:head() + self.i)
    local len = ph.length
    local ni = self.i + plh_size + len
    if ni > ib_count then
        if ni > ibuf.max_size then
            ibuf:expand(ni)
        end
        return self:reset()
    end

    local payload_len = len - ts_size
    ffi_copy(self.timestamp, ph.timestamp, ffi_sizeof(self.timestamp))

    local sp = ffi_cast(ffi_string(self.buf_type), ph + 1)
    local n = payload_len / self.sample_size
    self.i = ni

    -- Original had some more error checking goop here to check alignment

    return sp, n, self.timestamp[0]
end

function mt:_process_timestamped_read_data()
    local ibuf = self.inbuf
    local ib_count = #ibuf
    if self.i + plh_size > ib_count then
        return self:reset()
    end

    local ph = ffi_cast('proxy_header *', ibuf:head() + self.i)
    local len = ph.length
    local ni = self.i + plh_size + len
    if ni > ib_count then
        if ni > ibuf.max_size then
            ibuf:expand(ni)
        end
        return self:reset()
    end

    local payload_len = len - ts_size
    ffi_copy(self.timestamp, ph.timestamp, ffi_sizeof(self.timestamp))

    local sp = ffi_cast(ffi_string(self.buf_type),
            ffi_cast('char *', ph) + plh_size)
    local n = len / self.sample_size
    self.i = ni

    -- Original had some more error checking goop here to check alignment

    return sp, n, self.timestamp[0]
end

-- Read what is in the input buffer, perform test engine based scaling, and
-- move the resulting data into frame pipes for downstream consumption.
function mt:proxy_reader()
    self:_read_from_proxy()
    local sample_size = self.sample_size
    local scale = self.scale
    local offset = self.offset
    local st = sample_size == 4 and 'float *' or 'double *'
    while true do
        local ptr, num, ts = self:_process_read_data()

        -- No data...we are done
        if ptr == nil then
            return
        end

        -- No where to send data...just keep track of timestamps
        if self.pipe_list[0] == nil then
            self.sample_count = self.sample_count + num
            return
        end

        -- This is where the data is "framed" and complete frames are pushed
        -- to each of the pipes attached to this proxy node
        local count
        while num > 0 do
            if self.current_frame == nil then
                self.current_frame =
                    Frame_create(self.frame_elements, sample_size)
                self.current_frame.hdr.ts = self.sample_count
            end
            local frame = self.current_frame
            count = math_min(frame.maxsize - frame.count, num)

            -- Perform any test engine side scaling/offset
            local to = ffi_cast(st, Frame_tail(frame))
            local from = ffi_cast(st, ptr)
            for i = 0, count - 1 do
                to[i] = from[i] * scale + offset
            end

            ptr = ptr + count
            num = num - count
            frame.count = frame.count + count
            self.sample_count = self.sample_count + count
            if frame.count >= frame.maxsize then
                -- This is where frames are cloned and shoved into pipes
                local p = self.pipe_list[0]
                Pipe_add_frame(p, frame)
                while p.next ~= nil do
                    Pipe_add_frame(p.next, Frame_clone(frame))
                    p = p.next
                end
                self.current_frame = nil
            end
        end
    end
end

-- Read what is in the input buffer and move it to frame pipes for downstream
-- consumption.
function mt:proxy_reader_no_scaling()
    self:_read_from_proxy()
    local sample_size = self.sample_size
    while true do
        local ptr, num, ts = self:_process_read_data()

        -- No data...we are done
        if ptr == nil then
            return
        end

        -- No where to send data...just keep track of timestamps
        if self.pipe_list[0] == nil then
            self.sample_count = self.sample_count + num
            return
        end

        -- This is where the data is "framed" and complete frames are pushed
        -- to each of the pipes attached to this proxy node
        local count
        while num > 0 do
            if self.current_frame == nil then
                self.current_frame =
                    Frame_create(self.frame_elements, sample_size)
                self.current_frame.hdr.ts = self.sample_count
            end
            local frame = self.current_frame
            count = math_min(frame.maxsize - frame.count, num)

            -- Copy data to frame
            ffi_copy(Frame_tail(frame), ptr, count * sample_size)

            ptr = ptr + count
            num = num - count
            frame.count = frame.count + count
            self.sample_count = self.sample_count + count
            if frame.count >= frame.maxsize then
                -- This is where frames are cloned and shoved into pipes
                local p = self.pipe_list[0]
                Pipe_add_frame(p, frame)
                while p.next ~= nil do
                    Pipe_add_frame(p.next, Frame_clone(frame))
                    p = p.next
                end
                self.current_frame = nil
            end
        end
    end
end

-- Function used to read from proxies that supply raw data
function mt:raw_proxy_reader()
    self:_read_from_proxy()
    while true do
        local ptr, num, ts = self:_process_timestamped_read_data()

        -- If no data, no pipes to send data to, or ts < 0 we are done
        if ptr == nil or self.pipe_list[0] == nil or ts < 0 then return end

        if self.current_frame == nil then
            -- Need enough space for the "raw" message (ts + data) and the
            -- "length" of the ts + data (num value above).
            self.current_frame = Frame_create(1, num + ffi_sizeof('uint32_t'))
            self.current_frame.hdr.ts = ts
        end
        local frame = self.current_frame
        frame.count = 1

        -- Move the data to the frame
        local b = ffi_cast('message_raw_buf *', frame.buf)
        b.length = num
        ffi_copy(b.timestamp, ptr, num)

        -- This is where frames are cloned and shoved into pipes
        local p = self.pipe_list[0]
        Pipe_add_frame(p, frame)
        while p.next ~= nil do
            Pipe_add_frame(p.next, Frame_clone(frame))
            p = p.next
        end
        self.current_frame = nil
    end
end

-- Function used to read from proxies that supply CAN messages
function mt:can_message_proxy_reader()
    self:_read_from_proxy()
    while true do
        local ptr, num, ts = self:_process_timestamped_read_data()

        -- If no data, no pipes to send data to, or ts < 0 we are done
        if ptr == nil or self.pipe_list[0] == nil or ts < 0 then return end

        if self.current_frame == nil then
            -- Need enough space for "num" CAN messages
            self.current_frame = Frame_create(num, ffi_sizeof('message_can'))
            self.current_frame.hdr.ts = ts
        end
        local frame = self.current_frame

        -- Move the data to the frame
        local b = ffi_cast('message_can *', frame.buf)
        ptr = ffi_cast('message_can *', ptr)
        local msgs = 0
        for i = 0, num - 1 do
            -- If the message is NOT a heartbeat, add it to the frame
            if bit.band(ptr.value[12], 0x20) == 0 then
                ffi_copy(b, ptr, ffi_sizeof('message_can'))
                msgs = msgs + 1
                b = b + 1
            end
            ptr = ptr + 1
        end

        -- If there is nothing but heartbeats through away the frame, we r done
        if msgs == 0 then
            Frame_delete(frame)
            self.current_frame = nil
            return
        end

        -- Set the frame information
        frame.count = msgs
        frame.maxsize = msgs

        -- This is where frames are cloned and shoved into pipes
        local p = self.pipe_list[0]
        Pipe_add_frame(p, frame)
        while p.next ~= nil do
            Pipe_add_frame(p.next, Frame_clone(frame))
            p = p.next
        end
        self.current_frame = nil
    end
end

-- Read what is in the input buffer and move it to "frames" on the output side
function mt:sequentialized_proxy_reader()
    self:_read_from_proxy()
    local ftype = ffi_string(self.buf_type):match('^.+_(float%d+)$')
    local st = ftype == 'float32' and 'float' or 'double'
    while true do
        local sample, num, ts = self:_process_timestamped_read_data()

        -- If no data or no pipes to send data to, we are done
        if sample == nil then return end

        if self.pipe_list[0] == nil then
            self.sample_count = self.sample_count + num
            return
        end

        -- Sequentialize data going into pipe(s)
        for i = 0, num - 1 do
            -- Get a sample from the input buffer
            local _sample = sample[i]

            -- Put as many samples as will fit into the frame or the remainder
            -- of what is left, whichever is lower
            while self.clock.current_clock < _sample.timestamp[0] do
                self.clock:tick()
                if self.clock.current_clock > self.stale_after then
                    self.value.value[0] = self.stale_value.value[0]
                end
                if _sample.timestamp[0] >= 0 then
                    -- Get frame into which data will be generated
                    local frame = self.current_frame
                    if frame == nil then
                        frame =
                            Frame_create(self.frame_elements, ffi_sizeof(st))
                        if frame ~= nil then
                            frame.hdr.ts = self.sample_count
                        end
                    end
                    if frame == nil then
                        error(string.format(
                            "Unable to get frame for channel %s",
                            ffi.string(self.channel)))
                    end
                    self.current_frame = frame
                    ffi_cast(st .. '*', Frame_tail(frame))[0] =
                        self.value.value[0]
                    frame.count = frame.count + 1
                    if frame.count == frame.maxsize then
                        -- This is where frames are cloned and shoved into pipes
                        local p = self.pipe_list[0]
                        if p ~= nil then
                            Pipe_add_frame(p, frame)
                            while p.next ~= nil do
                                Pipe_add_frame(p.next, Frame_clone(frame))
                                p = p.next
                            end
                            self.current_frame = nil
                        else
                            -- No pipes, just reuse the frame
                            frame.count = 0
                        end
                    end
                end
                self.sample_count = self.sample_count + 1
            end
            self.last_stamp = _sample.timestamp[0]
            if not _sample:is_heartbeat() then
                self.stale_after = self.clock.next_clock + self.stale_clocks
                self.value.value[0] =
                    _sample.value[0] * self.scale + self.offset
            end
        end
    end
end

function mt:__gc()
    self.inbuf:__gc()
    if self.channel ~= nil then C.free(self.channel) end
end

local function proxy_ctor(t, d)
    if not ctor_cache[t] then
        local pb_type = string.format('struct %s_proxy_buffer', t)
        ffi.cdef(string.format([[
            %s {
                struct membuf   inbuf;
                struct frame    *current_frame;
                struct pipe     *pipe_list[1];
                int             i;
                int             sample_size;
                int             frame_sample_size;
                int             frame_elements;
                int             timestamped_samples;
                int             fd;
                ssize_t         already_read;
                int             failed;
                int             read_errno;
                ssize_t         read_ret;
                timestamp       timestamp[1];
                char            *channel;
                char            buf_type[128];
                int             sie_idx;
                // This is for scale and offset performed the test engine
                double          scale;
                double          offset;
                // The stuff below this point is for sequentializing messages
                sample_clock    clock;
                %s              value;
                %s              stale_value;
                timestamp       last_stamp;
                int64_t         stale_clocks;
                int64_t         stale_after;
                int64_t         sample_count;
                int             is_logical;
            };
        ]], pb_type, d, d))
        ctor_cache[t] = ffi.metatype(pb_type, {__index = mt})
    end
    return ctor_cache[t]
end

local m = {}

function m:new(loop, md, host, port, ip, pipe_rate)
    if md.sample_rate then md.sample_rate = tonumber(md.sample_rate) end

    local is_seq = (md.data_type:find('^message_float') and md.sample_rate and
        md.sample_rate > 0 and true)

    local ct_type = md.data_type

    local buf_type = md.data_type
    local sample_size = ffi_sizeof(md.data_type)
    if md.data_type:find('^sequential_float32$') then
        buf_type = 'float'
        sample_size = 4
    elseif md.data_type:find('^sequential_float64$') then
        buf_type = 'double'
        sample_size = 8
    elseif md.data_type == 'message_raw' then
        buf_type = 'char'
        sample_size = 1
    elseif md.data_type == 'message_can' then
        buf_type = 'message_can_buf'
        sample_size = ffi_sizeof('message_can')
    end
    if is_seq then
        ct_type = 'sequentialized_' .. ct_type
    end

    local ctor = proxy_ctor(ct_type, buf_type)
    local pn = ctor()

    pn.channel = C.malloc(md.name:len())
    ffi_copy(pn.channel, md.name)
    pn.scale = md.scale or 1.0
    pn.offset = md.offset or 0.0

    if md.data_type:find('^message_') then
        pn.timestamped_samples = 1
        if is_seq then
            pn.clock:init(md.sample_rate, sample_clock.master_rate)
            pn.stale_clocks = math.floor(
                sample_clock.master_rate *
                tonumber(md.stale_data_expiration_time)) - 1
            local sv = tonumber(md.invalid_data_output_value)
            if sv then
                pn.stale_value.value[0] = sv
            else
                error(string.format(
                    'Can sequentialize non-numeric message channel %s',
                    md.name))
            end
            pn.value = pn.stale_value
            if md.data_type:find('^message_float32$') then
                pn.frame_sample_size = 4
            else
                pn.frame_sample_size = 8
            end
        end
    else
        pn.timestamped_samples = 0
    end

    pipe_rate = pipe_rate or 5

    if md.data_type ~= 'message_raw' and md.data_type ~= 'message_can' then
        -- Create the frames that will be used to move data downstream to the
        -- rest of the test engine
        local frame_entries = math.floor(md.sample_rate / pipe_rate)
        if frame_entries < 1 then frame_entries = 1 end
        pn.frame_elements = frame_entries
        if not is_seq then
            pn.current_frame = Frame_create(pn.frame_elements, sample_size)
            pn.frame_sample_size = sample_size
        end

        -- Input buffer is large enough to hold 1 seconds worth of data or 128
        -- samples from an unsequentialized message channel.
        pn.inbuf:initialize((sample_size or 0) *
            (md.sample_rate or md.message_rate or 128))
    else
        pn.inbuf:initialize(8192)       -- Start INBUF at 8k for raw messages
    end
    pn.sample_size = sample_size or 0
    ffi_copy(pn.buf_type, buf_type .. ' *')

    local reader_fn
    if is_seq then
        -- Sequentialized numeric message reader
        reader_fn = pn.sequentialized_proxy_reader
    elseif pn.timestamped_samples > 0 then
        if md.data_type:find('_raw$') then
            -- Raw message reader
            reader_fn = pn.raw_proxy_reader
        elseif md.data_type:find('_can$') then
            -- CAN message reader
            reader_fn = pn.can_message_proxy_reader
        else
            local sn = notify:new(__'Not Yet Implemented')
            sn:error(
                __'Non-sequentialized message channels not supported yet')
            -- Numeric message reader
            reader_fn = pn.message_proxy_reader
        end
    elseif pn.scale == 1.0 and pn.offset == 0 then
        -- Sequential reader with no scaling in test engine
        reader_fn = pn.proxy_reader_no_scaling
    else
        -- Sequential reader with scaling in test engine
        reader_fn = pn.proxy_reader
    end


    -- If the connection is remote, use IP otherwise use UNIX domain socket
    local sock = net.socket{
        family = C.AF_UNIX,
        type = C.SOCK_STREAM,
        flags = 0}

    -- Function called when the connection to a channel proxy completes
    local function handle_connect_complete(s)
        if not s then return end
        pn.fd = s.sock
        local w = ev.IO.new(function(l, i, e) reader_fn(pn) end,
            s.sock, ev.READ)
        self.reader_watcher = w
        w:start(loop)
        local out_md = cjson.encode(md)
        local msg = string.format(
            "POST /channel_stream HTTP/1.0\r\n" ..
            "Content-length: %d\r\n\r\n%s", #out_md, out_md)
        local written = C.write(s.sock, msg, #msg)
        if written ~= #msg then
            error(string.format("Unable to send channel setup request for %s",
                md.name))
        end
    end

    -- Function called when UNIX domain socket connect finishes
    local function handle_unix_connect_complete(s)
        if s then
            return handle_connect_complete(s)
        end
        -- UNIX domain connect failed, try IP domain
        sock:close()
        net.socket{ip = ip, port = port}:connect_nb(
            loop, handle_connect_complete)
    end

    -- Setup for connect via UNIX domain socket
    local sa = ffi.new('struct sockaddr_un[1]')
    sa[0].sun_family = C.AF_UNIX
    local name = md.connector:match('^.*%@(.+)$')
    ffi.copy(sa[0].sun_path, '/tmp/proxy/' .. name)

    sock.addrs = ffi.new('struct addrinfo')
    sock.addrs.ai_flags = 0
    sock.addrs.ai_family = C.AF_UNIX
    sock.addrs.ai_socktype = C.SOCK_STREAM
    sock.addrs.ai_protocol = 0
    sock.addrs.ai_addrlen = ffi.sizeof('struct sockaddr_un')
    sock.addrs.ai_addr = ffi.cast('struct sockaddr *', sa)
    sock.addrs.ai_next = nil

    sock:connect_nb(loop, handle_unix_connect_complete)

    -- Set "is_logical" flag from metadata
    pn.is_logical = (md.is_logical == "true" and true) or
        (type(md.is_logical) == "boolean" and md.is_logical) or false

    return pn
end

return m
