require 'strict'

-- sie_utils.lua - Utility functions for the test engine to use in
--     setting up sie file XML.

local cjson = require 'cjson'
local rational = require 'rational'
-- local sie_writer = require 'test_engine/sie_writer'


local m = {}


-- Converts certain characters in a string into HTML compatible
-- values.  (This function was copied from sie_writer.lua)
-- Parameters:
--     'str'  the string to be modified
-- Returns:  the modified string
local function html_escape(str)
    return tostring(str):gsub('[&"<]', {
        ['&'] = '&amp;',
        ['"'] = '&quot;',
        ['<'] = '&lt;'
    })
end

-- Converts the input into a rational number.
-- (This function was copied from sie_writer.lua)
local function get_rational(x)
    local slash = (type(x) == 'string') and x:find('/')
    if slash then
        local num = tonumber(x:sub(1, slash - 1))
        local denom = tonumber(x:sub(slash + 1))
        return num / denom, num, denom
    end
    local num = tonumber(x)
    return num, num, 1
end

-- Creates a metadata table that can be used to set up an
-- SIE channel by passing it to m.write_channel().  It can
-- be subsequently modified further if necessary before calling
-- m.write_channel().
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md_in' is metadata table containing the info needed
--     to create a channel.
-- Returns:  a new metadata table ready for the addition of
--     "dim" attributes (and additional specialization as needed)
function m.sie_channel(sw, md_in)
    assert(md_in.data_type, "data_type is required")
    assert(md_in.sample_rate, "sample_rate is required")
    assert(md_in.sie_core_schema, "SIE core schema is required")
    assert(md_in.somat_datamode_type, "Somat datamode type is required")
    assert(md_in.somat_data_bits, "Somat data bits is required")
    assert(md_in.somat_data_format, "Somat data format is required")
    assert(md_in.datamode_name, "data mode must have name")

    local sample_rate = rational:new(md_in.sample_rate):tonumbers()

    local md_out = {
        _sie_test = md_in._sie_test or 0, -- KLUDGE
        _sie_ch = sw:next_id('ch'),
        _sie_group = sw:next_id('group'),
        _sie_decoder = sw:decoder_id(md_in.decoder_type),
        _sie_name = md_in.sie_name or md_in.name,
        data_type = md_in.data_type,
        xform_scale = md_in.xform_scale,
        xform_offset = md_in.xform_offset,
        ['somat:armada_input_metadata'] = cjson.encode(md_in),
        ['somat:connector'] = md_in.connector,
        ['somat:input_channel'] = md_in.somat_input_channel or md_in.name,
        ['core:description'] = md_in.description,
        ['core:sample_rate'] = sample_rate,
        ['core:schema'] = md_in.sie_core_schema,
        ['somat:datamode_type'] = md_in.somat_datamode_type,
        ['somat:data_bits'] = md_in.somat_data_bits,
        ['somat:data_format'] = md_in.somat_data_format,
        ['somat:datamode_name'] = md_in.datamode_name,
        _dims = {}
    }

    -- simple copy of dimension info to output table
    for i, dim in ipairs(md_in.dims) do
        md_out._dims[i] =
            {['core:label'] = dim.label, ['core:units'] = dim.units,
             _scale = dim.xform_scale, _offset = dim.xform_offset}
    end

    return md_out
end

-- Modifies a previously constructed table (which should have been
-- acquired by calling m.sie_channel()) to make it a metadata table
-- that can be used to set up a Burst History SIE channel.
-- Parameters:
--     'metadata'  a table containing the info needed to set up
--         the desired SIE channel
--     'md' is a metadata table such as returned by m.sie_channel()
-- Returns:  'md', modified by adding the "dim" attributes needed
--     for SIE channels of somat:burst data schema
--[[ function m.sie_burst_dims(md)
    md._dims = {
        {
            ['core:label'] = metadata.time_label or 'Time',
            ['core:units'] = metadata.time_units or 'Seconds',
            _scale = md.xform_scale,
            _offset = md.xform_offset,
        }, {
            ['core:label'] = metadata.name,
            ['core:units'] = metadata.output_units
        }, {
            ['core:label'] = "Offset",
            ['core:units'] = 'Samples',
        }
    }

    return md
end
]]

-- Writes XML for a channel header using an sie_writer.
-- Parameters:
--     'sw' is the sie_writer object to use
--     'md' is a metadata table containing all the information
--         needed to write the channel's XML
-- No value is returned.
function m.write_channel(sw, md)
    -- Without the following call lua doesn't seem able to garbage
    -- collect fast enough when writing XML for some tests with a lot
    -- of channels, though perhaps smaller chunks of garbage collection
    -- could be done here if necessary.
    collectgarbage()

    sw:xml('<ch test="(=_sie_test=)" id="(=_sie_ch=)" ' ..
        'name="(=_sie_name=)" group="(=_sie_group=)">\n', md)
    for k, v in pairs(md) do
        if k:sub(1, 1) ~= '_' then
            sw:xml(' <tag id="(=k=)">(=v=)</tag>\n',
                {k=html_escape(k), v=html_escape(v)})
        end
    end
    for i, dim in ipairs(md._dims) do
        md._i = i - 1
        sw:xml(' <dim index="(=_i=)">\n', md)
        for k, v in pairs(dim) do
            if k:sub(1, 1) ~= '_' then
                sw:xml('  <tag id="(=k=)">(=v=)</tag>\n',
                    {k=html_escape(k), v=html_escape(v)})
            end
        end
        sw:xml('  <data decoder="(=_sie_decoder=)" v="(=_i=)"/>\n', md)
        if dim._scale or dim._offset then
            local t = {scale = 1, offset = 0}
            if dim._scale then t.scale = get_rational(dim._scale) end
            if dim._offset then t.offset = get_rational(dim._offset) end
            sw:xml('  <xform scale="(=scale=)" offset="(=offset=)"/>\n', t)
        end
        sw:xml(' </dim>\n')
    end
    sw:xml('</ch>\n')
end

return m
