require 'strict'

local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'

-- test_eng_utils.lua - Utility functions for test engine and modules.

local ffi = require 'ffi'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof
local Frame_clone = frame.clone
local Frame_create = frame.create
local Frame_delete = frame.delete
local Pipe_add_frame = pipe.add_frame
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames


local m = {}

-- Returns the size of the given data type in bytes, followed by
-- the C type implied by the given data type.
function m.sample_size(data_type)
    local sample_size = ffi_sizeof(data_type)
    local c_type = nil

    if data_type:find('^sequential_float32$') then
        c_type = 'float'
        sample_size = 4
    elseif data_type:find('^sequential_float64$') then
        c_type = 'double'
        sample_size = 8
    elseif data_type == 'message_raw' then
        c_type = 'char'
        sample_size = 1
    elseif data_type:find('^message_float32$') then
        c_type = 'float'
        sample_size = 4
    elseif data_type:find('^message_') then
        c_type = 'double'
        sample_size = 8
    end

    return sample_size, c_type
end

-- Returns a C floating point pointer type with the given byte size.
function m.frame_data_type(sample_size)
    return sample_size == 4 and 'float *' or 'double *'
end

-- When given a valid SIE buffer 'index' value, creates an SIE buffer
-- object with the given attributes and creates a frame pipe between
-- it and the given pipe_list.
function m.sie_collect_buf(sie_writer, index, sample_size,
                           stanza, pipe_list)
    -- do nothing if not given a valid sie buf index
    if index and type(index) == 'number' then
        -- Write the channel metadata to the SIE file
        stanza.description = stanza.description ~= nil and
            stanza.description or ""
        stanza.output_units = stanza.output_units ~= nil and
            stanza.output_units or ""
        local md = sie_writer:channel(stanza)
        -- Create 'collect' pipe and attach one to pipe list
        local p = Pipe_attach_pipe(pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        sie_writer:new_buffer(index, md._sie_group,
            8192, sample_size, 8, p,
            stanza.data_type:find('^message_') and true or nil)
    end
end

-- This function is used to "short circuit" a module's processing
-- when there is no processing to be done.  It checks that the
-- given input pipe contains at least one frame, and that the
-- given pipe list contains at least one pipe.  If a trigger
-- channel exists it is also checked for at least one frame.
-- NOTE:  If there are input frames but no output pipes, the
--        input frames WILL BE DELETED to prevent them backing
--        up in the input pipe.  Also the trigger frames if
--        such exist.  Care is taken to ensure the same number
--        of input and trigger frames are deleted.
-- Parameters:
--     'in_pipe' is the pipe to check for frames
--     'pipe_list' is the list of output pipes to check for a pipe
--     'trig_pipe' is an optional pipe used for trigger values
-- Returns:  true if there is work to do, false otherwise,
--     and the number of input samples thrown away (this will
--     be zero when there is at least one pipe in 'pipe_list')
function m.there_is_work_to_do(in_pipe, pipe_list, trig_pipe)
    local work_to_do = Pipe_has_frames(in_pipe) and
                       (trig_pipe == nil or Pipe_has_frames(trig_pipe))
    if not work_to_do then
        return false, 0
    end
    local thrown_away = 0

    -- If there are frames in the input pipe but there are no
    -- output pipes, throw away the data in the input pipe.
    if pipe_list[0] == nil then
        work_to_do = false
        if trig_pipe then
            local tf = Pipe_get_frame(trig_pipe)
            local f = Pipe_get_frame(in_pipe)
            while (f ~= nil) and (tf ~= nil) do
                thrown_away = thrown_away + f.count
                Frame_delete(f)
                Frame_delete(tf)
                f = Pipe_get_frame(in_pipe)
                tf = Pipe_get_frame(trig_pipe)
            end
        else
            local f = Pipe_get_frame(in_pipe)
            while f ~= nil do
                thrown_away = thrown_away + f.count
                Frame_delete(f)
                f = Pipe_get_frame(in_pipe)
            end
        end
    end

    return work_to_do, thrown_away
end

-- This function provides objects needed when a module writes
-- data generated from one input frame to an output frame.
-- It returns an empty output frame of the specified size
-- using the element size of the given input frame, as well as
-- data pointers.
-- NOTE!  The input frame is not to be modified in ANY way!
-- Parameters:
--     'fsize' is size of the to-be-created frame
--     'in_frame' is the frame whose element size is to be used
--     'ts' is the time stamp to be set in the output frame
--     'output_val_size' is the element size to be used for the
--         output frame - if not present, the input frame element
--         size will be used
-- Returns:  created frame, pointer to start of in_frame's data,
--     pointer to start of created frame's data
function m.create_output_frame(fsize, in_frame, ts, output_val_size)
    local of_el_size = output_val_size or in_frame.element_size
    local of = Frame_create(fsize, of_el_size)
    of.hdr.ts = ts
    local sample_type = m.frame_data_type(in_frame.element_size)
    local in_vals = ffi_cast(sample_type, in_frame.buf)
    if output_val_size then
        sample_type = m.frame_data_type(output_val_size)
    end
    local out_vals = ffi_cast(sample_type, of.buf)

    return of, in_vals, out_vals
end

-- Puts the given frame to the first pipe in the given
-- output pipe list, and puts a clone of the frame in the
-- remaining pipes in the list.
-- Parameters:
--     'of' is the frame to put in the first pipe
--     'pipe_list' is the list of pipes in which 'of' and
--         its clones is to be put
-- Returns:  true if pipe_list has a pipe, false otherwise
function m.put_frame(of, pipe_list)
    if pipe_list[0] ~= nil then
        local op = pipe_list[0]
        Pipe_add_frame(op, of)
        while op.next ~= nil do
            Pipe_add_frame(op.next, Frame_clone(of))
            op = op.next
        end
        return true
    end
    return false
end

-- Creates an SIE buffer object for a datamode and creates a
-- frame pipe between it and the given pipe_list.
-- Parameters:
--     'xml' is the channel XML to be written to the SIE file
-- Returns:  created "SIE buffer"
function m.create_datamode_sie_buf(xml, sie_writer, index,
        sample_size, stanza, pipe_list)
    -- check validity of sie buf index
    assert(index and type(index) == 'number',
        "INVALID sie buf index: " .. tostring(index))

    -- Write the channel metadata to the SIE file
    sie_writer:xml(xml)
    -- Create pipe for "sie buf" frames and attach to pipe list
    local p = Pipe_attach_pipe(pipe_list)
    -- Create the SIE buffer and attach the pipe to it
    sie_writer:new_buffer(index, nil,
        8192, sample_size, 8, p,
        stanza.data_type:find('^message_') and true or nil)
end

-- Returns a boolean value based on the given metadata value which
--   can be a "boolean" string or an actual boolean value.  Any
--   other value is automatically converted to false.
-- Parameters:
--     'meta' is the metadata value to be converted to boolean
-- Returns:  boolean, based on the value of 'meta'
function m.xlate_meta_boolean(meta)
    if type(meta) == 'number' then return meta ~= 0 and true or false end
    return (meta == "true" and true) or
           (type(meta) == "boolean" and meta) or false
end

-- Sets some common instance attributes from the given metadata values
-- Parameters:
--     'instance' module instance whose attributes are to be set
--     'stanza' module's metadata (from decoded setup json)
function m.set_common_meta(instance, stanza)
    instance.is_logical = m.xlate_meta_boolean(stanza.is_logical)
end

return m
