--! \file

local armada = require 'armada'
local ffi = require 'ffi'
local C = ffi.C

require('lj_include'){
    'fcntl',
    'unistd',
}

ffi.cdef[[
void *mmap(void *adr, size_t length, int prot, int flags, int fd, off_t offset);

enum {
    PROT_READ = 1,
    PROT_WRITE = 2,
    PROT_EXEC = 4,
};

enum {
    MAP_SHARED = 1,
};
]]

local config = require 'system_config'

--! Various timestamp utilities.
--? namespace timestamps
local m = {}
--? alias m timestamps

function m.epoch_to_iso(n)
    local epoch = math.floor(n)
    local fraction = n - epoch
    local date = os.date('!%Y-%m-%dT%H:%M:%S', epoch)
    return date .. string.format('%.9f', fraction):sub(2)
end

function m.iso_to_epoch(iso)
    -- KLUDGE doesn't work in non-GMT zones
    local time, fraction = { }, nil
    time.year, time.month, time.day, time.hour, time.min, time.sec, fraction =
        iso:match('(%d%d%d%d)%-(%d%d)%-(%d%d)T(%d%d)%:(%d%d)%:(%d%d)(%.?%d*)')
    local epoch = os.time(time)
    fraction = fraction or 0
    return epoch + fraction
end

--! A class for dealing with high-precision absolute timestamps.
--? class m.ts
m.ts = {}
m.ts.__index = m.ts
function m.ts:new(s, ns)
    if type(s) == 'table' then
        return m.ts:super()(s)
    elseif ns then
        ns = ffi.cast("int64_t", ns)
        local ns_per_s = ffi.cast("int64_t", 1000000000)
        local s_offset = ns / ns_per_s
        local ns_remainder = ns % ns_per_s
        s = s + s_offset
        if ns_remainder < 0 then
            ns = 1000000000 + ns_remainder
            s = s - 1
        else
            ns = ns_remainder
        end
        s = tonumber(s)
        ns = tonumber(ns)
        return setmetatable({s = s, ns = ns}, m.ts)
    else
        return setmetatable({s = math.floor(s), ns = (s % 1) * 1000000000},
                            m.ts)
    end
end

function m.ts:tonumber()
    return self.s + self.ns / 1000000000
end

function m.ts:tons()
    return self.s * 1000000000 + self.ns
end

function m.ts.__add(a, b)
    return m.ts:new(a.s + b.s, a.ns + b.ns)
end

function m.ts.__sub(a, b)
    return m.ts:new(a.s - b.s, a.ns - b.ns)
end

function m.ts.__eq(a, b)
    return a.s == b.s and a.ns == b.ns
end

function m.ts.__lt(a, b)
    return a.s <= b.s and a.ns < b.ns
end

local initialized = false
local ptp_clockrate
local sys_clockrate, sys_to_ns = 40000000, 25
local xsec, xnsec, xstamp = 0, 0, m.ts:new(0, 0)

local function set_info(xts, xtns, sysctr, sysrate, ptpctr, ptprate)
    xsec = xts
    xnsec = xtns
    xstamp = sysctr
    sys_clockrate = sysrate
    ptp_clockrate = ptprate
    sys_to_ns = 1000000000 / sys_clockrate
end

local function init()
    if not initialized then
        local xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate =
            armada.timing:GetClockRelationships()
        set_info(xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate)
        initialized = true

        -- Adjust stuff if/when time jumps for any reason
        armada.timing.TimeJumped:bind(
            function(xts, xtns, sysctr, sysrate, ptpctr, ptprate)
                set_info(xts, xtns, sysctr, sysrate, ptpctr, ptprate)
            end)
    end
end


function m.epoch_to_stamp(epoch)
    init()
    local xepoch = xsec + xnsec / 1000000000
    local diff = epoch - xepoch
    local stampdiff = diff * sys_clockrate
    return xstamp + stampdiff
end

function m.stamp_to_epoch(stamp)
    init()
    local xepoch = xsec + xnsec / 1000000000
    local stampdiff = stamp - xstamp
    local diff = stampdiff / sys_clockrate
    return xepoch + diff
end

function m.ts_to_stamp(ts)
    init()
    local xts = m.ts:new(xsec, xnsec)
    local diff = ts - xts
    local stampdiff = diff:tons() / sys_to_ns
    return xstamp + stampdiff
end

function m.stamp_to_ts(stamp)
    init()
    local xts = m.ts:new(xsec, xnsec)
    local stampdiff = stamp - xstamp
    local diff = stampdiff * sys_to_ns
    return xts + m.ts:new(0, diff)
end

function m.stamp_to_iso(stamp)
    return m.epoch_to_iso(m.stamp_to_epoch(stamp))
end

function m.ts_to_iso(n)
    init()
    local date = os.date('!%Y-%m-%dT%H:%M:%S', n.s)
    return string.format('%s.%09d', date, n.ns)
end

function m.iso_to_ts(iso)
    init()
    -- KLUDGE doesn't work in non-GMT zones
    local time, fraction = { }, nil
    time.year, time.month, time.day, time.hour, time.min, time.sec, fraction =
        iso:match('(%d%d%d%d)%-(%d%d)%-(%d%d)T(%d%d)%:(%d%d)%:(%d%d)%.?(%d*)')
    local epoch = os.time(time)
    local ns = 0
    if fraction then
        local frac_digits = fraction:len()
        ns = fraction * 10^(9 - frac_digits)
    end
    return m.ts:new(epoch, ns)
end

function m.iso_to_stamp(iso)
    return m.ts_to_stamp(m.iso_to_ts(iso))
end

-- This function converts a byte string contained in a membuf object
-- into a m.ts object.
--
-- Parameter time_buf is the membuf object containing the byte string
-- to be converted.  The byte string has the following format:
--     2013-07-1120:01:15.2
-- The following offsets are used to access the relevant fields in
-- the input byte string:
--     year offset --> 0
--     month offset --> 5
--     day offset --> 8
--     hours offset --> 10
--     minutes offset --> 13
--     seconds offset --> 16
--     tenths of seconds offset --> 19 (if exists in input)
--     hundredths of seconds offset --> 20 (if exists in input)
--
-- If time_buf is not a valid value an empty m.ts object is returned.
function m.membuf_to_ts(time_buf)
    if time_buf == nil or #time_buf < 18 then
        return m.ts:new(0, 0) -- error return kludge !!
    end

    init()
    -- KLUDGE doesn't work in non-GMT zones
    local time, fraction = { }, nil
    time.year = ffi.string(time_buf:head(), 4)
    time.month = ffi.string(time_buf:head() + 5, 2)
    time.day = ffi.string(time_buf:head() + 8, 2)
    time.hour = ffi.string(time_buf:head() + 10, 2)
    time.min = ffi.string(time_buf:head() + 13, 2)
    time.sec = ffi.string(time_buf:head() + 16, 2)
    if #time_buf >= 20 then
        fraction = ffi.string(time_buf:head() + 19, 1)
        if #time_buf > 21 then
            fraction = fraction ..
                       ffi.string(time_buf:head() + 20, 1)
        end
    end
    local epoch = os.time(time)
    local ns = 0
    if fraction then
        local frac_digits = fraction:len()
        ns = fraction * 10^(9 - frac_digits)
    end
    return m.ts:new(epoch, ns)
end

local fd = C.open('/dev/mem', C.O_RDWR)
local map = ffi.cast('unsigned int *',
    C.mmap(nil, 0x100000, C.PROT_READ + C.PROT_WRITE, C.MAP_SHARED,
        fd, ffi.new('unsigned int', 0xe0000000)))
C.close(fd)

local function get_local_timestamp()
    local ts_hi = map[0x0061c / 4]
    local ts_lo = map[0x0062c / 4]
    local ts_hi_reread = map[0x0061c / 4]
    if ts_hi ~= ts_hi_reread then
        ts_lo = map[0x0062c / 4]
        ts_hi = ts_hi_reread
    end
    return (tonumber(ts_hi) * 2^32) + tonumber(ts_lo)
end

function m.get_ts()
    init()
    local count = get_local_timestamp()
    -- Convert to ns
    count = count * sys_to_ns
    -- Convert to s and ns
    local s, ns = xsec + math.floor(count / 1000000000),
        xnsec + math.floor(count % 1000000000)
    -- Normalize
    while ns >= 1000000000 do
        ns = ns - 1000000000
        s = s + 1
    end
    return m.ts:new(s, ns)
end

return m
