#!/usr/bin/lua

require('lj_include'){
    'sys/types',
    'unistd',
    'sched',
    'linux/sched',
}

require 'middleclass'
local u = require 'unix'
local ev = require 'ev'
local loop = ev.Loop.default
local dbus = require 'dbus'
local ts = require 'timestamps'
local channel_proxy = require 'channel_proxy'
local ffi = require 'ffi'
local C = ffi.C

-- Increase process priority (flagship[-5], ptpv2[-5])
C.nice(-6)

local sp = ffi.new('struct sched_param')
sp.__sched_priority = 55
assert(C.sched_setscheduler(C.getpid(),
    bit.bor(C.SCHED_FIFO, C.SCHED_RESET_ON_FORK), sp) == 0)

local log = require 'log'

local dd = ffi.new('double[1]')

local config = require 'system_config'

ev.Signal.new(function() end, u.SIGPIPE):start(loop) -- ignore SIGPIPE
ev.Signal.new(function() loop:unloop() end, u.SIGINT):start(loop)

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

log:log(log.INFO, "Timing daemon starting.")

local mmio = require(config.target..'_mmio')
mmio:initialize(log)

local sys_rate = 40 * 1000000
local sys_ns = 1000000000 / sys_rate

local ptp_mult = 8
local ptp_rate = sys_rate * ptp_mult
local ptp_ns = 1000000000 / ptp_rate

local Timing = { }
Timing.__index = Timing
local Timing_iface = dbus.Interface.new('com.somat.armada.Timing')

if config.target == 'armada' then
    function Timing:SendSyncPulse()
        return mmio:send_sync_pulse()
    end
    Timing_iface:add_method{'SendSyncPulse', '', 'x'}
end

local function compute_wall_offset(sys, ptp, s, ns)
    local wall_clock = ts.ts:new(s, math.floor(ns / sys_ns) * sys_ns)
    return wall_clock - ts.ts:new(0, sys * sys_ns)
end

local rate_adjustment = 0
local servos = { }
local wall_offset = compute_wall_offset(mmio:sample_timers())
local last_adjuster
local osc_mode = false

function Timing:AdjustClockRate(ppb, adjuster)
    if adjuster == 'osc_mode_on' or adjuster == 'osc_mode_off' then
        if adjuster == 'osc_mode_on' then
            osc_mode = true
        else
            osc_mode = false
        end
        return
    end
    if osc_mode then
        if adjuster == 'PTP' then
            return
        end
    end
    if adjuster == 'GPS' and last_adjuster == 'PTP' then
        last_adjuster = nil
        return
    end
    last_adjuster = adjuster
    ppb = mmio:adjust_oscillator(ppb)
    if math.floor(math.abs(ppb - rate_adjustment)) > 50000 then
        for i, servo in ipairs(servos) do
            servo.ign_integral = 5
        end
    end
    rate_adjustment = ppb
    for i, servo in ipairs(servos) do
        servo:adjust(servo.last_adjustment)
    end
end
Timing_iface:add_method{'AdjustClockRate', 'ds', '', no_reply = true}

function Timing:AdjustClock(sec, nsec, adjuster)
    if adjuster == 'GPS' and last_adjuster == 'PTP' then
        last_adjuster = nil
        return
    end
    last_adjuster = adjuster
    nsec = math.floor(nsec / sys_ns + 0.5) * sys_ns
    local adj = ts.ts:new(sec, nsec)
    u.adjtimex{modes = u.ADJ_SETOFFSET, time = { sec = adj.s, nsec = adj.ns }}
    wall_offset = wall_offset + adj
    local xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate =
        self:GetClockRelationships()
    self:TimeJumped(xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate)
end
Timing_iface:add_method{'AdjustClock', 'xis', ''}

Timing_iface:add_signal{'TimeJumped', 'xixdxd'}

function Timing:SetTimestampDelta(delta)
    self:UpdateTimestampDelta(delta)
end
Timing_iface:add_signal{'UpdateTimestampDelta', 'x'}

function Timing:GetClockRelationships()
    local xt_sec, xt_nsec, sys_ctr = wall_offset.s, wall_offset.ns, 0
    local ptp_ctr = sys_ctr * ptp_mult
    return xt_sec, xt_nsec, sys_ctr, sys_rate, ptp_ctr, ptp_rate
end
Timing_iface:add_method{'GetClockRelationships', '', 'xixdxd'}

function Timing:OscillatorRead()
    return mmio:oscillator_read()
end
Timing_iface:add_method{'OscillatorRead', '', 's'}

function Timing:OscillatorWrite(value)
    return mmio:oscillator_write(value)
end
Timing_iface:add_method{'OscillatorWrite', 'x', 's'}

function Timing:OscillatorOffset(value)
    return mmio:oscillator_offset(value)
end
Timing_iface:add_method{'OscillatorOffset', 'd', 'd'}

local timing = setmetatable({ }, Timing)
bus:export_object(timing, '/com/somat/armada/Timing', { Timing_iface })
assert(bus:bus_request_name('com.somat.armada.Timing'))

local Servo = class('Servo')
function Servo:initialize(name, dt)
    self.name = name
    self.dt = dt
    self.Kp = 10
    self.Ki = 3
    self.Kd = 0
    self.ign_integral = 0
    self.last_adjustment = 0
    self:reset()
end

function Servo:reset()
    self.count = 0
    self.init = { }
    self.integral = nil
    self.last_offset = nil
end

function Servo:run(time, offset)
    if self.count < 10 then
        if self.count == 0 then
            self.init.time = time
            self.init.offset = offset
        else
            local time_diff = time - self.init.time
            local offset_diff = offset - self.init.offset
            self.integral = offset_diff / time_diff
            if self.debug then
                log:debug('%s %f %f %f', self.name, tonumber(time_diff),
                    tonumber(offset_diff), tonumber(self.integral))
            end
        end
        self.ign_integral = 10
    else
        local previous_error = self.last_offset
        local error_term = offset
        if self.ign_integral > 0 then
            self.ign_integral = self.ign_integral - 1
        else
            self.integral = self.integral + (error_term * self.dt * self.Ki)
        end
        local derivative = (error_term - previous_error) / self.dt
        local output =
            (self.Kp * error_term) + self.integral + (self.Kd * derivative)

        local extra = self:adjust(-output)
        if self.debug then
            log:debug(string.format('adjust %s %10.2f %10.2f %10.2f %s',
                self.name, error_term, self.integral, output,
                tostring(extra)))
        end
    end
    self.last_offset = offset
    self.count = self.count + 1
end

local dt = 0.1
local ptp_servo = Servo:new('ptp', dt)
function ptp_servo:adjust(ppb)
    self.last_adjustment = ppb
    ppb = ppb + rate_adjustment
    return mmio:adjust_ptp_counter(ppb)
end
table.insert(servos, ptp_servo)

local function do_sync(loop, w)
    local sys, ptp, s, ns = mmio:sample_timers()
    local sys_scaled = sys * ptp_mult
    local offset = ptp - sys_scaled
    ptp_servo:run(sys / sys_rate, offset * ptp_ns)
end
ev.Timer.new(do_sync, 0.001, dt):start(loop)


-- Apply the oscillator characterization.  NOTE:  If this is attempt too soon
-- after reboot, the writes to the Si514 device fail with I2C errors.  That is
-- why application is deferred via a timer for a bit.
ev.Timer.new(function(l, t, e)
        mmio:apply_oscillator_characterization()
    end, 1, 0):start(loop)

loop:loop()
