local video = [[
local camera = {
    ['axis'] = require 'axis_camera',
    ['vivotek'] = require 'vivotek_camera'
}

local net = require 'net_ffi'
local ffi = require 'ffi'
local sio = require 'sio'
local C = ffi.C

-- GNU C Library 'libc'

ffi.cdef[=[
    typedef unsigned int size_t;
    int sprintf(char *str, const char *format, ...);
    int atoi(const char *nptr);
    int sscanf(const char *str, const char *format, ...);
    char *strstr(const char *haystack, const char *needle);
    int strncmp(const char *s1, const char *s2, size_t n);
]=]

ffi.cdef[=[
    typedef struct video_runtime {
        uint32_t frame_length;
        uint8_t *jpeg;
        uint8_t *buffer;
        uint8_t *buffer_ptr;
        uint32_t buffer_length;
        char boundary[70];
    } video_runtime_t;
]=]

local function frame_update(camera, run)
    if (run[0].buffer_ptr - run[0].jpeg) >= run[0].frame_length then
        local offset = run[0].buffer_ptr - (run[0].jpeg + run[0].frame_length)
        camera:lock()
        camera:channel_write(run[0].jpeg, run[0].frame_length)
        ffi.copy(run[0].buffer, run[0].jpeg + run[0].frame_length, offset)
        run[0].buffer_ptr = run[0].buffer + offset
        run[0].frame_length = 0
        camera:unlock()
    end
end

local function video(camera)
    local sd = -1
    local state
    local run = ffi.new('video_runtime_t [1]')
    local n = 131072
    local rbuf = ffi.new('char [?]', n)
    local ptr = ffi.new('char *', nil)
    local rfds = ffi.new('fd_set [1]')
    local tv = ffi.new('struct timeval [1]')
    local boundary = ffi.new('char *')
    local header = ffi.new('char *')
    local ctype = ffi.new('char *')
    local clen = ffi.new('char *')
    local mime = ffi.new('char *')
    local http = ffi.new('char *')
    local http_response = {
        version = ffi.new('char [4]'),
        code = ffi.new('char [4]'),
        reason = ffi.new('char [33]')
    }
    local sres, hres, vres = camera:spec_image_size()
    run[0].buffer_length = hres * vres
    local buffer = ffi.new('uint8_t [?]', run[0].buffer_length)
    run[0].buffer = buffer
    run[0].buffer_ptr = buffer
    run[0].frame_length = 0
    while true do
        state = camera:state()
        if state == C.VIDEO_ERROR then
            os.exit(1)
        elseif state == C.VIDEO_SHUTDOWN then
            break
        elseif state == C.VIDEO_RETRY then
            camera:disconnect()
            camera:state(C.VIDEO_CONNECT)
        elseif state == C.VIDEO_CONNECT then
            sd = camera:connect()
            if sd < 0 then
                camera:state(C.VIDEO_ERROR)
            else
                camera:state(C.VIDEO_REQUEST)
            end
        elseif state == C.VIDEO_REQUEST then
            if camera:request() < 0 then
                camera:state(C.VIDEO_ERROR)
            else
                camera:state(C.VIDEO_STREAMING)
            end
        elseif state == C.VIDEO_STREAMING then
            -- consume MIME stream until state change or socket error
            local rc
            local bsize
            repeat
                sio.FD_ZERO(rfds)
                sio.FD_SET(sd, rfds)
                tv[0].tv_sec = 1
                tv[0].tv_usec = 200000
                if camera:state() ~= C.VIDEO_STREAMING then
                    break
                end
                rc = C.select(sd + 1, rfds, nil, nil, tv)
                if rc > 0 then
                    if camera:state() ~= C.VIDEO_STREAMING then
                        break
                    end
                    n = C.recv(sd, rbuf, ffi.sizeof(rbuf), 0)
                    if n <= 0 then
                        io.stderr:write(camera:name() .. ":video(): recv " ..
                                 ffi.string(C.strerror(ffi.errno())) .. "\n")
                        camera:state(C.VIDEO_ERROR)
                        break
                    end
                elseif rc == -1 then
                    io.stderr:write(camera:name() .. ":video(): select " ..
                             ffi.string(C.strerror(ffi.errno())) .. "\n")
                    camera:state(C.VIDEO_ERROR)
                    break
                else
                    if camera:state() == C.VIDEO_SHUTDOWN then
                        break
                    end
                    io.stdout:write(camera:name() .. ":video(): timeout\n")
                    camera:state(C.VIDEO_SHUTDOWN)
                    break
                end
                if http == nil then
                    http = C.strstr(rbuf, "HTTP/")
                    if http ~= nil then
                        ptr = C.strstr(http, "\r\n")
                        if ptr ~= nil then
                            ptr[0] = 0
                            C.sscanf(http, "HTTP/%s %s %32[^\n]s",
                                            http_response.version,
                                            http_response.code,
                                            http_response.reason)
                            ptr[0] = 13
                            if ffi.string(http_response.code) ~= '200' then
                                io.stderr:write("video: HTTP/" ..
                                    ffi.string(http_response.version) .. " " ..
                                    ffi.string(http_response.code) .. " " ..
                                    ffi.string(http_response.reason) .. "\n")
                                camera:state(C.VIDEO_ERROR)
                                break
                            end
                        end
                    end
                end
                if mime == nil then -- check for valid boundary definition
                    mime = C.strstr(rbuf, "boundary=")
                    if mime ~= nil then
                        ptr = C.strstr(mime, "\r\n")
                        if ptr ~= nil then
                            ptr[0] = 0
                            mime = mime + 9
                            if C.strncmp(mime, "--", 2) == 0 then
                                C.sprintf(run[0].boundary, "%s", mime)
                            else
                                C.sprintf(run[0].boundary, "--%s", mime)
                            end
                            ptr[0] = 13
                        end
                    end
                else -- process buffer
                    bsize = (run[0].buffer_ptr + n) - buffer
                    if bsize < run[0].buffer_length then
                        ffi.copy(run[0].buffer_ptr, rbuf, n);
                        run[0].buffer_ptr = run[0].buffer_ptr + n
                        if run[0].frame_length == 0 then
                            boundary = C.strstr(buffer, run[0].boundary)
                            ctype = C.strstr(buffer, "Content-Type: image/jpeg")
                            clen = C.strstr(buffer, "Content-Length: ")
                            if boundary ~= nil and ctype ~= nil and clen ~= nil then
                                clen = clen + 16
                                header = C.strstr(clen, "\r\n\r\n")
                                if header ~= nil then
                                    run[0].frame_length = C.atoi(clen)
                                    run[0].jpeg = header + 4
                                    frame_update(camera, run)
                                end
                            end
                        else
                            frame_update(camera, run)
                        end
                    else
                        -- In this case we have received a large amount
                        -- of data and no image has been properly
                        -- staged. Since the conection is functioning,
                        -- stop and send the request again. Works
                        -- consistently.
                        run[0].buffer = buffer
                        run[0].buffer_ptr = buffer
                        run[0].frame_length = 0
                        camera:state(C.VIDEO_RETRY)
                    end
                end
            until camera:state() ~= C.VIDEO_STREAMING
        end -- if state
    end -- while true
end -- video

local function thread(x)
    local data = ffi.cast('camera_data_t *', x)
    local vendor = ffi.string(data[0].vendor)
    video(camera[vendor]:setup(data))
end

thread_func_ptr = tonumber(ffi.cast('intptr_t',
                                    ffi.cast('void *(*)(void *)', thread)))
]]

return video
